#include <meshing.hpp>
#include <visual.hpp>

namespace netgen
{
  using namespace netgen;

class VisualSceneDrawTet : public VisualScene
{
public:
  int sing_vert[4];
  int sing_edge[6];
  int sing_face[4];
  int drawsectionplanes;
  int drawtetfaces;
  int drawtetedges;

public:  
  VisualSceneDrawTet ()
  {
    int i;
    for (i = 0; i < 4; i++)
      sing_vert[i] = 0;
    for (i = 0; i < 6; i++)
      sing_edge[i] = 0;
    for (i = 0; i < 4; i++)
      sing_face[i] = 0;
    drawsectionplanes = 0;
    drawtetfaces = 1;
    drawtetedges = 1;
  }

  virtual ~VisualSceneDrawTet ()
  {
    ;
  }

  virtual void DrawScene ()
  {
    cout << "draw " << endl;

    int i, j, k;
    static bool first = 1;

    if (first)
      {
	center = Point<3> (0.333, 0.333, 0.333);
	rad = 1;
	CalcTransformationMatrices();
	
	first = 0;
      }

    cout << "sing vertices : ";
    for (i = 0; i < 4; i++)
      cout << sing_vert[i] << " ";
    cout << endl;

    cout << "sing edges    : ";
    for (i = 0; i < 6; i++)
      cout << sing_edge[i] << " ";
    cout << endl;

    cout << "sing faces    : ";
    for (i = 0; i < 4; i++)
      cout << sing_face[i] << " ";
    cout << endl;

    static const int points[4][3] = 
      { { 1, 0, 0 },
	{ 0, 1, 0 },
	{ 0, 0, 1 },
	{ 0, 0, 0 } };

    static const int edges[6][2] = 
      { { 0, 1 },
	{ 0, 2 },
	{ 0, 3 }, 
	{ 1, 2 },
	{ 1, 3 },
	{ 2, 3 } };

    static const int faces[4][3] = 
      { { 1, 3, 2 },
	{ 0, 2, 3 },
	{ 0, 3, 1 },
	{ 0, 1, 2 } };

    static const int oppedges[6][2] = // points not on the edge
      { { 2, 3 },
	{ 1, 3 },
	{ 1, 2 }, 
	{ 0, 3 },
	{ 0, 2 },
	{ 0, 1 } };

    static const int edge2faces[6][2] =
      { { 2, 3 },
	{ 1, 3 },
	{ 1, 2 },
	{ 0, 3 },
	{ 0, 2 },
	{ 0, 1 } };
   


    glClearColor(backcolor, backcolor, backcolor, 1.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    
    glEnable (GL_COLOR_MATERIAL);
    glLineWidth (1.0f);

    SetLight();

    glPushMatrix();
    glMultMatrixf (transformationmat);
    
     // mark sing vertices
    glPointSize( 10.0 );
    glBegin (GL_POINTS);
    glColor3f (0.0f, 0.0f, 1.0f);  // blau
    for (i = 0; i < 4; i++)
      if (sing_vert[i])
	glVertex3f (points[i][0], points[i][1], points[i][2]);
    glEnd ();

    // mark sing edges
    glColor3f (0.0f, 0.0f, 1.0f);  // blau
    glLineWidth (3.0f);
    for (i = 0; i < 6; i++)
      if (sing_edge [i])
	{
	  glBegin (GL_LINES);
	  glVertex3f (points[edges[i][0]][0],
		      points[edges[i][0]][1],
		      points[edges[i][0]][2]);
	  glVertex3f (points[edges[i][1]][0],
		      points[edges[i][1]][1],
		      points[edges[i][1]][2]);
	  glEnd ();
	}


    if (drawtetedges)
      {
	glColor3f (0.0f, 0.0f, 0.0f);  // schwarz
	glLineWidth (1.0f);
	for (i = 0; i < 4; i++)
	  {
	    Point<3> p[3];
	    for (j = 0; j < 3; j++)
	      for (k = 0; k < 3; k++)
		p[j](k) = points[faces[i][j]][k];
	    
	    glBegin(GL_LINES);
	    for (j = 0; j < 3; j++)
	      glVertex3f (p[j](0), p[j](1), p[j](2));
	    glEnd();
	  }
      }
    

    glShadeModel (GL_SMOOTH);
    glDisable (GL_COLOR_MATERIAL);
    glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    if (drawtetfaces)
      {
	glBegin(GL_TRIANGLES);
	for (i = 0; i < 4; i++)
	  {
	    if (sing_face [i])
	      {
		float mat_col[] = { 1, 0, 0, 0.9 };
		glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_col);
		//	  glColor3f (1.0f, 0.0f, 0.0f);  // rot
	      }
	    else
	      {
		float mat_col[] = { 0, 1, 0, 0.9 };
		glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_col);
		//	  glColor3f (0.0f, 1.0f, 0.0f);  // gruen
	      }
	    Point<3> p[3];
	    for (j = 0; j < 3; j++)
	      for (k = 0; k < 3; k++)
		p[j](k) = points[faces[i][j]][k];
	    
	    Vec<3> nv = Cross (p[1]-p[0], p[2]-p[0]);
	    glNormal3f (nv(0), nv(1), nv(2));
	    for (j = 0; j < 3; j++)
	      glVertex3f (p[j](0), p[j](1), p[j](2));
	  }
	glEnd ();
      }
    
    if (drawsectionplanes) // draw section planes
      {
	double dist = 0.15;

	int vplane[] = {0, 0, 0, 0};
	int eplane[] = {0, 0, 0, 0, 0, 0};
	int fplane[] = {0, 0, 0, 0};

	glBegin(GL_TRIANGLES);
 
	float mat_col[] = { 0, 0, 1, 0.8 }; // blue
	glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_col);

	for (i = 0; i < 4; i++) // faces
	  if (sing_face [i]) // "near to face" plane
	    {
	      fplane[i] = 1;

	      Point<3> p[4]; // three points of the face + opposite
	      for (j = 0; j < 3; j++)
		for (k = 0; k < 3; k++)
		  p[j](k) = points[faces[i][j]][k];
	      for (k = 0; k < 3; k++)
		p[3](k) = points[i][k]; // opposite point
	      
	      Point<3> q[3]; // three points of the section plane
	      for (j = 0; j < 3; j++)
		q[j] = Point<3>((1-dist) * Vec<3>(p[j]) + dist*Vec<3>(p[3]));
	      
	      // draw section plane
	      Vec<3> nv = Cross (q[1]-q[0], q[2]-q[0]);
	      glNormal3f (nv(0), nv(1), nv(2));
	      for (j = 0; j < 3; j++)
		glVertex3f (q[j](0), q[j](1), q[j](2));

	      glNormal3f (-nv(0), -nv(1), -nv(2));
	      for (j = 0; j < 3; j++)
		glVertex3f (q[j](0), q[j](1), q[j](2));
	  
	    } else fplane[i] = 0;

	glEnd();

	
	mat_col[0] = 1;
	mat_col[1] = 1;
	mat_col[2] = 0; // yellow
	glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_col);
	

	for (i = 0; i < 6; i++) // edges
	  if (sing_edge[i] &&                // "near to edge" plane
	      !(fplane[edge2faces[i][0]] && fplane[edge2faces[i][1]]))
	    {
	      cout << fplane[edge2faces[i][0]] << endl;
	      cout << fplane[edge2faces[i][1]] << endl;
	      eplane[i] = 1;

	      glBegin(GL_POLYGON);
	      Point<3> p[4]; // two points of the edge + two opposite points
	      for (j = 0; j < 2; j++)
		for (k = 0; k < 3; k++)
		  p[j](k) = points[edges[i][j]][k];
	      for (j = 0; j < 2; j++)
		for (k = 0; k < 3; k++)
		  p[j+2](k) = points[oppedges[i][j]][k];
	      
	      Point<3> q[4]; // three points of the section plane
	      q[0] = Point<3>(dist * Vec<3>(p[2]) + (1-dist)*Vec<3>(p[0]));
	      q[1] = Point<3>(dist * Vec<3>(p[3]) + (1-dist)*Vec<3>(p[0]));
	      q[2] = Point<3>(dist * Vec<3>(p[3]) + (1-dist)*Vec<3>(p[1]));
	      q[3] = Point<3>(dist * Vec<3>(p[2]) + (1-dist)*Vec<3>(p[1]));
	      
	      // draw section plane
	      Vec<3> nv = Cross (q[1]-q[0], q[2]-q[0]);
	      glNormal3f (nv(0), nv(1), nv(2));
	      for (j = 0; j < 4; j++)
		glVertex3f (q[j](0), q[j](1), q[j](2));

	      glEnd();
	    } else eplane[i] = 0;

	glBegin(GL_TRIANGLES);
	
	
	mat_col[0] = 1;
	mat_col[1] = 1;
	mat_col[2] = 1; // white
	glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_col);
	

	for (i = 0; i < 4; i++) // vertices
	  if (sing_vert [i])                // "near to vertex" plane
	    {
	      /*
	      if (fplane[vert2faces[i][0]] & // all 3 face planes set
		  fplane[vert2faces[i][1]] &
		  fplace[vert2faces[i][2]])
		{
		  vplane[i] = 0; continue;
		}

	           
	      for (j = 0; j < 2; j++) // check 3 pairs of neighboring faces
		for (k = j+1; k < 3; k++)
		  if (fplane[vert2faces[i][j]] && fplane[vert2faces[i][k]])
		    // two face planes set && third not (cf. above)
		    {
		      // find the two edges in vertex i which are normal to one of the faces
		      for (l = 0; l < 6; l++)
			{
			  if (eplane[l]) continue;
			}
					   }
	      if (fplane[vert2faces[i][0]] & // all 3 face planes set
		  fplane[vert2faces[i][1]] &
		  fplace[vert2faces[i][2]])
		{
		  vplane[i] = 0; continue;
		}
	      */

	      vplane[i] = 1;

	      Point<3> p[4]; // three points of the opposite face + vertex itself
	      for (j = 0; j < 3; j++)
		for (k = 0; k < 3; k++)
		  p[j](k) = points[faces[i][j]][k];
	      for (k = 0; k < 3; k++)
		p[3](k) = points[i][k]; // opposite point
	      
	      Point<3> q[3]; // three points of the section plane
	      for (j = 0; j < 3; j++)
		q[j] = Point<3>(dist * Vec<3>(p[j]) + (1-dist)*Vec<3>(p[3]));
	      
	      // draw section plane
	      Vec<3> nv = Cross (q[1]-q[0], q[2]-q[0]);
	      glNormal3f (nv(0), nv(1), nv(2));
	      for (j = 0; j < 3; j++)
		glVertex3f (q[j](0), q[j](1), q[j](2));
	  
	    } else vplane[i] = 0;

	glEnd ();
      }

    
    glPopMatrix();


    DrawCoordinateCross ();
    DrawNetgenLogo ();
    glFinish();
  }
};


VisualSceneDrawTet vsdrawtet;


int Ng_DrawTet (ClientData clientData,
		Tcl_Interp * interp,
		int argc, tcl_const char *argv[])
{
  cout << "Ng_DrawTet" << endl;
  if (argc == 4 && strcmp (argv[1], "vertex") == 0)
    {
      int vnr = atoi (argv[2]);
      int val = atoi (argv[3]);
      cout << "vertex, vnr = " << vnr << ", val = " << val << endl;
      vsdrawtet.sing_vert[vnr] = val;
    }
  if (argc == 4 && strcmp (argv[1], "edge") == 0)
    {
      int enr = atoi (argv[2]);
      int val = atoi (argv[3]);
      cout << "edge, enr = " << enr << ", val = " << val << endl;
      vsdrawtet.sing_edge[enr] = val;
    }
  if (argc == 4 && strcmp (argv[1], "face") == 0)
    {
      int fnr = atoi (argv[2]);
      int val = atoi (argv[3]);
      cout << "face, enr = " << fnr << ", val = " << val << endl;
      vsdrawtet.sing_face[fnr] = val;
    }
  if (argc == 3 && strcmp (argv[1], "drawsectionplanes") == 0)
    {
      int val = atoi (argv[2]);
      vsdrawtet.drawsectionplanes  = val;
      cout << "drawsectionplanes = " << val << endl;
    }
  if (argc == 3 && strcmp (argv[1], "drawtetfaces") == 0)
    {
      int val = atoi (argv[2]);
      vsdrawtet.drawtetfaces  = val;
      cout << "drawtetfaces = " << val << endl;
    }
  if (argc == 3 && strcmp (argv[1], "drawtetedges") == 0)
    {
      int val = atoi (argv[2]);
      vsdrawtet.drawtetedges  = val;
      cout << "drawtetedges = " << val << endl;
    }
  return TCL_OK;
}


int Ng_MarkConstraint (ClientData clientData,
		Tcl_Interp * interp,
		int argc, tcl_const char *argv[])
{
    static const int points[4][3] = 
      { { 1, 0, 0 },
	{ 0, 1, 0 },
	{ 0, 0, 1 },
	{ 0, 0, 0 } };

    static const int edges[6][2] = 
      { { 0, 1 },
	{ 0, 2 },
	{ 0, 3 }, 
	{ 1, 2 },
	{ 1, 3 },
	{ 2, 3 } };

    static const int faces[4][3] = 
      { { 1, 3, 2 },
	{ 0, 2, 3 },
	{ 0, 3, 1 },
	{ 0, 1, 2 } };

    static const int face2edges[4][3] =
      { { 4, 5, 3 },
	{ 1, 5, 2 },
	{ 2, 4, 0 },
	{ 0, 3, 1 } };
    /*
    static const int face2edgeorients[4][3] =
      { {  1, -1, -1 },
	{  1,  1, -1 },
	{  1, -1, -1 },
	{  1,  1, -1 } };
    */

  int i;

  cout << "Ng_MarkConstraint" << endl;

  // mark all edges with two singular vertices
  for (i = 0; i < 6; i++)
    if (vsdrawtet.sing_vert [edges[i][0]] &&
	vsdrawtet.sing_vert [edges[i][1]]) vsdrawtet.sing_edge[i] = 1;

  // mark all faces with two or more singular edges
  for (i = 0; i < 4; i++)
    if (vsdrawtet.sing_edge [face2edges[i][0]] +
	vsdrawtet.sing_edge [face2edges[i][1]] + 
	vsdrawtet.sing_edge [face2edges[i][2]] > 1) vsdrawtet.sing_face[i] = 1;
	
  return TCL_OK;
}
}


int AddOn_Init (Tcl_Interp * interp)
{
  cout << "initialize ng-addon" << endl;
  netgen::AddVisualizationScene ("drawtet", &netgen::vsdrawtet);

  Tcl_CreateCommand (interp, "Ng_DrawTet", netgen::Ng_DrawTet,
		     (ClientData)NULL,
		     (Tcl_CmdDeleteProc*) NULL);

  Tcl_CreateCommand (interp, "Ng_MarkConstraint", netgen::Ng_MarkConstraint,
		     (ClientData)NULL,
		     (Tcl_CmdDeleteProc*) NULL);

}

