/**
   High Order Finite Element Space for H(Curl) 
  */
#include <comp.hpp>
#include <fem.hpp> 

namespace ngcomp 
{
  using namespace ngcomp; 

  HCurlHighOrderFESpaceAlt ::  
  HCurlHighOrderFESpaceAlt (const MeshAccess & ama, const Flags & aflags)
    : FESpace (ama, aflags),
      flags(aflags)
  {
    int ndom = ma.GetNDomains();
    gradientdomains.SetSize (ndom);
    gradientdomains.Set();

    gradientboundaries.SetSize (ma.GetNBoundaries());
    gradientboundaries.Set();

    fn=int(flags.GetNumFlag("face",1));

    if (flags.NumListFlagDefined("gradientdomains"))
      {
	const ARRAY<double> & graddomains = flags.GetNumListFlag ("gradientdomains");
	for (int i = 0; i < gradientdomains.Size(); i++)
	  if (!graddomains[i])
	    gradientdomains.Clear(i);
	cout << endl << " ******* gradientdomains" << gradientdomains << endl;
      }
    if (flags.NumListFlagDefined("gradientboundaries"))
    {
      const ARRAY<double> & gradbounds = flags.GetNumListFlag ("gradientboundaries");
      for (int i = 0; i < gradientboundaries.Size(); i++)
	if (!gradbounds[i])
	  gradientboundaries.Clear(i);
      cout << endl << " ******* gradientboundaries" << gradientboundaries << endl;
    }
    
    if(flags.GetDefineFlag("nograds"))
      {
	gradientdomains.Clear();   
	nograds = 1; 
      } 
       
    low_order_space = new  NedelecFESpace (ma, 0, dimension, iscomplex);
    
        
    var_order = flags.GetDefineFlag("variableorder");
    
    dom_order_min.SetSize(ndom); 
    dom_order_max.SetSize(ndom); 
	
    uniform_order_inner = int (flags.GetNumFlag ("orderinner", -1));
    uniform_order_face = int (flags.GetNumFlag ("orderface", -1));
    uniform_order_edge = int (flags.GetNumFlag ("orderedge", -1));
	
    if(var_order)
      {
	dom_order_min = INT<3> (0,0,0); 
	dom_order_max = INT<3> (99,99,99); 

	rel_order = int (flags.GetNumFlag ("relorder", -1)); 

	if(flags.NumListFlagDefined("dom_order"))
	  {
	    const ARRAY<double> & dom_order = flags.GetNumListFlag ("dom_order");
	    for (int i=0;i<ndom;i++)
	      for (int l=0;l<3;l++)
		{
		  dom_order_min[i][l] = int(dom_order[i]);
		  dom_order_max[i][l] = int(dom_order[i]); 
		}
	  }
	
	if (flags.NumListFlagDefined("dom_order_min_x"))
	  {
	    const ARRAY<double> & dom_order_min_x = flags.GetNumListFlag ("dom_order_min_x");
	    for (int i=0;i<ndom;i++)
	      dom_order_min[i][0] = int(dom_order_min_x[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_max_x"))
	  {
	    const ARRAY<double> & dom_order_max_x = flags.GetNumListFlag ("dom_order_max_x");
	    for (int i=0;i<ndom;i++)
	      dom_order_max[i][0] = int(dom_order_max_x[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_min_y"))
	  {
	    const ARRAY<double> & dom_order_min_y = flags.GetNumListFlag ("dom_order_min_y");
	    for (int i=0;i<ndom;i++)
	      dom_order_min[i][1] = int(dom_order_min_y[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_max_y"))
	  {
	     const ARRAY<double> & dom_order_max_y = flags.GetNumListFlag ("dom_order_max_y");
	    for (int i=0;i<ndom;i++)
	      dom_order_max[i][1] = int(dom_order_max_y[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_min_z"))
	  {
	    const ARRAY<double> & dom_order_min_z = flags.GetNumListFlag ("dom_order_min_z");
	    for (int i=0;i<ndom;i++)
	      dom_order_min[i][2] = int(dom_order_min_z[i]); 
	  }
	if (flags.NumListFlagDefined("dom_order_max_z"))
	  {
	   const ARRAY<double> &  dom_order_max_z = flags.GetNumListFlag ("dom_order_max_z");
	    for (int i=0;i<ndom;i++)
	      dom_order_max[i][2] = int(dom_order_max_z[i]); 
	  }

	
	
      }
    else
      {
	dom_order_min = INT<3> (order,order,order); 
	dom_order_max = INT<3> (order,order,order); 
      }
    
    
    smoother = int (flags .GetNumFlag("smoothing",7));
    //  level_adapted_order = flags.GetDefineFlag("level_adpated_order"); 

    
    augmented = int (flags.GetNumFlag ("augmented", 0));

    /*    if (flags.GetDefineFlag ("optext"))
	  { 
	  trig = new HCurlHighOrderTrig<TrigExtensionMin> (order);
	  tet =  new HCurlHighOrderTet<TrigExtensionMin> (order);
	  prism =  new HCurlHighOrderPrism<TrigExtensionMin> (order);
	  }
	  else
    */

    trig = new HCurlHighOrderTrig<IntegratedLegendreMonomialExt> (order);
    tet =  new HCurlHighOrderTet<IntegratedLegendreMonomialExt> (order);
    prism =  new HCurlHighOrderPrism<IntegratedLegendreMonomialExt> (order);
    quad = new HCurlHighOrderQuad<IntegratedLegendreMonomialExt> (order);
    hex = new HCurlHighOrderHex<IntegratedLegendreMonomialExt> (order);
    segm = new HCurlHighOrderSegm<IntegratedLegendreMonomialExt> (order); 
    pyramid = new HCurlHighOrderPyr<IntegratedLegendreMonomialExt> (order);
    
      
    /* dynamic_cast<HCurlHighOrderFiniteElement<1>*> (segm) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<2>*> (trig) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (tet) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (prism) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<2>*> (quad) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (pyramid) -> SetAugmented (augmented);
    dynamic_cast<HCurlHighOrderFiniteElement<3>*> (hex) -> SetAugmented (augmented);*/ 


    // Evaluator for shape tester 
    static ConstantCoefficientFunction one(1);
    if (ma.GetDimension() == 2)
      {
	ARRAY<CoefficientFunction*> coeffs(1);
	coeffs[0] = &one;
	evaluator = GetIntegrators().CreateBFI("massedge", 2, coeffs);
      }
    else if(ma.GetDimension() == 3) 
      {
	ARRAY<CoefficientFunction*> coeffs(1); 
	coeffs[0] = &one;
	evaluator = GetIntegrators().CreateBFI("massedge",3,coeffs); 
	boundary_evaluator = GetIntegrators().CreateBFI("robinedge",3,coeffs); 
	
      }

  }
  
  HCurlHighOrderFESpaceAlt :: ~HCurlHighOrderFESpaceAlt ()
  {
    ;
  }
  
  FESpace * HCurlHighOrderFESpaceAlt :: 
  Create (const MeshAccess & ma, const Flags & flags)
  {
    /*
    int order = int(flags.GetNumFlag ("order", 0));
    int dim = int(flags.GetNumFlag ("dim", 1));
    bool iscomplex = flags.GetDefineFlag ("complex");

    if (order <= 0)
      return new NedelecFESpace (ma,0, dim, iscomplex);      // Space with MG   
    else       
    */
    return new HCurlHighOrderFESpaceAlt (ma, flags);
  }
  
  void HCurlHighOrderFESpaceAlt :: Update()
  {
    if (order < 0) 
      throw Exception("HCurlHighOrderFESpaceAlt::Update() order < 0 !" ) ;
    
    low_order_space -> Update();

    nv = ma.GetNV();
    nel = ma.GetNE();
    int nse = ma.GetNSE();
    
    switch(ma.GetDimension())
      {
      case 2: 
	ned = ma.GetNEdges();
	nfa = 0;  
	break; 
      case 3: 
	ned = ma.GetNEdges();
	nfa = ma.GetNFaces();
	break; 
      }
    
    order_edge.SetSize (ned);   
    order_face.SetSize (nfa); 
    order_inner.SetSize (nel);
    usegrad_edge.SetSize (ned);                                
    usegrad_face.SetSize (nfa); 
    usegrad_cell.SetSize (nel);
    fine_edge.SetSize(ned); 
    fine_face.SetSize(nfa); 
     
    order_edge = 0; 
    order_face = INT<2> (0,0);
    order_inner = INT<3> (0,0,0);
    fine_edge = 0; 
    fine_face = 0; 
     
    usegrad_edge = 1;                                
    usegrad_face = 1; 
    usegrad_cell = 1; 

    ARRAY<int> eledges, elfaces, vnums;
    int dim = ma.GetDimension();

    /* int minuniform = 0; 
    if(uniform_order_inner !=-1) minuniform = uniform_order_inner;
    if(uniform_order_face !=-1)
      if(minuniform > 0) minuniform = min2(uniform_order_face,minuniform);
      else minuniform = uniform_order_face; 
    if(uniform_order_edge !=-1)
      if(minuniform > 0) minuniform = min2(uniform_order_edge,minuniform);
      else minuniform = uniform_order_edge; */

    for(int i=0; i< nel; i++) 
      if(gradientdomains[ma.GetElIndex(i)]) 
	{
	  ma.GetElEdges(i,eledges);
	  
	  for(int j=0;j<eledges.Size();j++)
	    usegrad_edge[eledges[j]]=1;
	  
	  if(ma.GetDimension()==3)
	    {
	      ma.GetElFaces(i,elfaces);
	      
	      for(int j=0;j<elfaces.Size();j++)
		usegrad_face[elfaces[j]]=1; 
	    }
	  usegrad_cell[i] = 1; 
	} 

    for(int i=0; i<nse && gradientboundaries.Size(); i++)
      if(gradientboundaries[ma.GetSElIndex(i)])
	{
	  ma.GetSElEdges(i,eledges);
	  for(int j=0; j<eledges.Size();j++)
	    usegrad_edge[eledges[j]] = 1;
	
	  if(ma.GetDimension()==3)
	    usegrad_face[ma.GetSElFace(i)] = 1;
	}


     for (int i = 0; i < nel; i++)
      {
	int index = ma.GetElIndex(i);
	if (!DefinedOn (index)) continue;
	
	INT<3> el_orders = ma.GetElOrders(i);

	for(int j=0;j<3;j++)
	  el_orders[j] = int(max2(dom_order_min[index][j],min2(dom_order_max[index][j],el_orders[j]+rel_order)));
		

	ELEMENT_TYPE eltype=ma.GetElType(i); 
	const FACE * faces = ElementTopology::GetFaces (eltype);
	const EDGE * edges = ElementTopology::GetEdges (eltype);
	const POINT3D * points = ElementTopology :: GetVertices (eltype);
	ma.GetElVertices (i, vnums);
	ma.GetElEdges (i, eledges);		
	ma.GetElFaces(i, elfaces);

	for(int j=0;j<eledges.Size();j++) fine_edge[eledges[j]] = 1; 
	if (dim == 3)
	  for(int j=0;j<elfaces.Size();j++) fine_face[elfaces[j]] = 1; 

	if(uniform_order_edge==-1)
	  {
	    for(int j=0;j<eledges.Size();j++)
	      for(int k=0;k<dim;k++)
		if(points[edges[j][0]][k] != points[edges[j][1]][k])
		  { 
		    order_edge[eledges[j]] = el_orders[k];
		    break; 
		  }
	  }
	else  
	  for(int j=0;j<eledges.Size();j++) 
	    order_edge[eledges[j]] = uniform_order_edge;  


	if(dim==3)
	  { 
	    if(uniform_order_face==-1)
	      for(int j=0;j<elfaces.Size();j++)
		{
		  if(faces[j][3]==-1) 
		    {
		      
		      order_face[elfaces[j]][0] = int(max2(order_face[elfaces[j]][0], 
							   el_orders[0]));
		      order_face[elfaces[j]][1] = order_face[elfaces[j]][0]; 
		    }
		  else //quad_face
		    {
		      int fmax = 0;
		      for(int k = 1; k < 4; k++) 
			if(vnums[faces[j][k]] > vnums[faces[j][fmax]]) fmax = k;   
		      INT<2> f((fmax+3)%4,(fmax+1)%4); 
		      if(vnums[faces[j][f[1]]] > vnums[faces[j][f[0]]]) swap(f[0],f[1]);
		      
		      for(int l=0;l<2;l++)
			for(int k=0;k<3;k++)
			  if(points[faces[j][fmax]][k] != points[faces[j][f[l] ]][k])
			    {
			      order_face[elfaces[j]][l] = int(max2(order_face[elfaces[j]][l], 
				 				   el_orders[k]));
			      break; 
			    } 
		    }  
		}
	     else
	       for(int j=0;j<elfaces.Size();j++) order_face[elfaces[j]] = INT<2> (uniform_order_face,uniform_order_face);
	  }
	

	if(uniform_order_inner==-1)
	  for(int j=0;j<dim;j++)
	    order_inner[i][j] = int(max2(order_inner[i][j],el_orders[j]));
	else 
	  order_inner[i] = INT<3> (uniform_order_inner,uniform_order_inner,uniform_order_inner); 

	/*
	int minorder = el_orders[0]; 
	for(int j=1;j<2;j++)
	  if(minorder > el_orders[j]) minorder = el_orders[j];
       
	if(uniform_order_inner != -1 && uniform_order_face != -1 && uniform_order_edge !=-1)	   
	  minorder=minuniform; 
	else
	  minorder = min2(minuniform,minorder); 
	   
	for(int j=0;j<nv; j++) 
	  if(order_avertex[vnums[j]]==0) 
	    order_avertex[j] = minorder+1;  // +1 compatible with  H1Space
	  else 
	    if(minorder < order_avertex[vnums[j]])
	    order_avertex[vnums[j]] = minorder+1; 	 */ 
	
      }
   
    /*
    // skip dofs on Dirichlet boundary hack [JS]
    int nsel = ma.GetNE();
    for (i = 0; i < nsel; i++)
      {
	  ma.GetSElEdges (i, eledges);
	  int face = ma.GetSElFace (i);
	  
	  for (j = 0; j < eledges.Size(); j++)
	    order_edge[eledges[j]] = 0;
	  order_face[face] = 0;
      }
    */
      
    //  for(i=0;i<ned;i++) order_edge[i] *= usegrad_edge[i]; 
    
     // pyramids 
    for ( int i = 0; i < nel; i++)
      { 
	if (!DefinedOn (ma.GetElIndex(i))) continue;
	ELEMENT_TYPE eltype=ma.GetElType(i); 
	if(eltype == ET_PYRAMID)
	  {
	    /*
	    const FACE * faces = ElementTopology::GetFaces (eltype);
	    const EDGE * edges = ElementTopology::GetEdges (eltype);

	    ma.GetElEdges (i, eledges);		
	    ma.GetElFaces(i, elfaces);

	    for(int j=0;j<8;j++)  
	      order_edge[eledges[j]] = order; 
	    
	    for(int j=0;j<4;j++) order_face[elfaces[j]] = INT<2> (order,order); 
	    */
	    //order_face[elfaces[4]] = INT<2> (0,0); 

	    order_inner[i] = INT<3>(0,0,0); 
	  }
	
      }

    *testout << " hcurl fine_edge " << fine_edge << endl; 
    *testout << " hcurl fine_face " << fine_face << endl; 

    for (int i=0;i<ned;i++) if(!fine_edge[i]) order_edge[i] = 0; 
    for (int i=0;i<nfa;i++) if(!fine_face[i]) order_face[i] = INT<2> (0,0); 

    // usegrad_cell = 0;

    UpdateDofTables(); 
  }
		
  void HCurlHighOrderFESpaceAlt :: UpdateDofTables()
  {
    nv = ma.GetNV();
    ned = ma.GetNEdges();
    nfa = ma.GetNFaces();
    nel = ma.GetNE();
    ARRAY<int> pnums; 
    int i; 
    
    ndof = ned; // Nedelec (order = 0) !!   
    if(augmented == 1 ) ndof+=nv; 
       
    first_edge_dof.SetSize(ned+1); 
    for (i = 0; i < ned; i++)
      {
	first_edge_dof[i] = ndof;
	if(order_edge[i]>0)
	  ndof += usegrad_edge[i]*order_edge[i]     - 1;
      }
    first_edge_dof[ned] = ndof;

    if (ma.GetDimension() == 3)
      {
	first_face_dof.SetSize(nfa+1);
	for (i=0; i< nfa; i++) 
	  { 
	    first_face_dof[i] = ndof; 
	    ma.GetFacePNums(i,pnums);  
	    INT<2> p = order_face[i]; 
	    switch(pnums.Size())
	      {
	      case 3: //Triangle   
		if(p[0]>1)
		  ndof += ((usegrad_face[i]+1)*p[0] + 2)*(p[0]-1)/2  ; //  -1;  
		break; 
	      case 4: //Quad 
		if(p[0]>=0 && p[1]>=0)
		  ndof += (usegrad_face[i]+1)*p[0]*p[1] + p[0] + p[1]; 
		break; 
	      }
	  } 
	first_face_dof[nfa] = ndof;     
      }

    first_inner_dof.SetSize(nel+1);
    for (i=0; i< nel; i++)
      {
	first_inner_dof[i] = ndof;
	INT<3> p = order_inner[i];
	switch(ma.GetElType(i)) 
	  {
	  case ET_TRIG:
	    if(p[0]>1)
	      ndof += ((usegrad_cell[i]+1)*p[0] + 2) * (p[0]-1) /2; //       -(p[0]-1);
	    ndof += 3;
	    break; 
	  case ET_QUAD: 
	    if(p[0]>=0 &&  p[1]>=0) 
	      ndof += (usegrad_cell[i]+1) * p[0] * p[1] + p[0] + p[1]; 
	    ndof += 4;
	    break; 
	  case ET_TET: 
	    if(p[0]>2)
	      ndof += ((usegrad_cell[i] + 2) *  p[0] + 3) * (p[0]-2) * (p[0]-1) / 6; 
	    ndof += 6;
	    // ndof += 4;
	    break; 
	  case ET_PRISM:
	    if(p[0]>1 &&  p[2]>=0) 
	      ndof += (usegrad_cell[i]+2)*p[2] * p[0]*(p[0]-1)/2
		+ (p[0]-1)*p[2] + p[0]*(p[0]-1)/2; 
	    ndof += 9;
	    break; 
	  case ET_HEX:
	    if(p[0]>=0 && p[1]>=0 && p[2] >=0)
	      ndof += (usegrad_cell[i] + 2)* p[0]*p[1]*p[2] +  p[0]*p[1] 
		+ p[0]*p[2] + p[1]*p[2];
	    break; 
	  case ET_PYRAMID:
	    ndof += 0; //(usegrad_cell[i] + 2)* p[0]*p[1]*p[2] +  p[0]*p[1] + p[0]*p[2] + p[1]*p[2];
	    break; 
	  }
      }
    first_inner_dof[nel] = ndof;    
    
    (*testout) << " HCURL " << endl; 
    (*testout) << "augmented " << augmented << endl; 
    (*testout) << "ndof  = " << ndof << endl; 
    (*testout) << "hc first_edge   = " << first_edge_dof << endl;
    (*testout) << "hc first_face   = " << first_face_dof << endl;
    (*testout) << "hc first_inner  = " << first_inner_dof << endl;
    (*testout) << " order = " << order << endl; 
    (*testout) << "order_edge HC = " << order_edge << endl;
    (*testout) << "order_face HC = " << order_face << endl;
    (*testout) << "order_inner HC = " << order_inner << endl;
      
  }
  
  const FiniteElement & HCurlHighOrderFESpaceAlt :: GetFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe = 0;
    
    switch (ma.GetElType(elnr))
      {
      case ET_TET:
	fe = tet; break;
      case ET_PYRAMID: 
	fe = pyramid; break;
      case ET_PRISM:
	fe = prism; break;
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      case ET_HEX: 
	fe = hex; break; 
      default:
	fe = 0;
      }
    
    if (!fe)
      {
	stringstream str;
	str << "HCurlHighOrderFESpaceAlt " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    ArrayMem<int,12> vnums; // calls GetElPNums -> max 12 for PRISM12
    ma.GetElVertices(elnr, vnums);

    if(ma.GetDimension() == 2) 
      {	
	HCurlHighOrderFiniteElement<2> * hofe = 
	  dynamic_cast<HCurlHighOrderFiniteElement<2>*> (fe);
	
    	ArrayMem<int, 4> ednums, ord_edge, ug_edge;
	
	ma.GetElEdges(elnr, ednums);
	
	ord_edge.SetSize (ednums.Size());
	ug_edge.SetSize (ednums.Size()); 
	
	for (int j = 0; j < ednums.Size(); j++)
	  {
	    ord_edge[j] = order_edge[ednums[j]];
	    ug_edge[j]  = usegrad_edge[ednums[j]]; 
	  }

	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (ord_edge);
	hofe -> SetOrderInner (order_inner[elnr]);

	hofe -> SetUsegradEdge (ug_edge); 
	hofe -> SetUsegradCell (usegrad_cell[elnr]); 
	//	hofe -> SetAugmented(augmented);
	// hofe -> SetOrderVertex (order_vert);

	hofe -> ComputeNDof();
      }   
    else if (ma.GetDimension() == 3) 
      {
	HCurlHighOrderFiniteElement<3> * hofe = 
	  dynamic_cast<HCurlHighOrderFiniteElement<3>*> (fe);

	ArrayMem<int, 12> ednums, ord_edge, ug_edge;
	ArrayMem<int, 6> fanums,  ug_face; 
	ArrayMem<INT<2>, 6> ord_face; 
	
	ma.GetElEdges(elnr, ednums);
	ma.GetElFaces(elnr, fanums);
	
	ord_edge.SetSize (ednums.Size());
	ord_face.SetSize (fanums.Size());
	ug_edge.SetSize (ednums.Size()); 
	ug_face.SetSize (fanums.Size()); 
	
	for (int j = 0; j < ednums.Size(); j++)
	  {
	    ord_edge[j] = order_edge[ednums[j]];
	    ug_edge[j] = usegrad_edge[ednums[j]]; 
	  }
	for (int j = 0; j < fanums.Size(); j++)
	  {
	    ord_face[j] = order_face[fanums[j]];
	    ug_face[j] = usegrad_face[fanums[j]];
	  }

	//	hofe -> SetAugmented(augmented); 
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (ord_edge);
	hofe -> SetOrderFace (ord_face);
	hofe -> SetOrderInner (order_inner[elnr]);

	hofe -> SetUsegradEdge (ug_edge); 
	hofe -> SetUsegradFace (ug_face); 
	hofe -> SetUsegradCell (usegrad_cell[elnr]); 

	//	hofe -> SetAugmented(augmented); 

	/*	ArrayMem<int, 8> order_vert;
	for (int j = 0; j < vnums.Size(); j++)
	  order_vert[j] = order_avertex[vnums[j]];
	hofe -> SetOrderVertex (order_vert);
	*/
	

	hofe -> ComputeNDof();
      }

    return *fe;
  }
 
  const FiniteElement & HCurlHighOrderFESpaceAlt :: GetSFE (int selnr, LocalHeap & lh) const
  {
    int i, j;

    FiniteElement * fe = 0;
    
    switch (ma.GetSElType(selnr))
      {
      case ET_SEGM: 
	fe = segm; break; 
      case ET_TRIG:
	fe = trig; break;
      case ET_QUAD:
	fe = quad; break;
      default:
	fe = 0;
      }
    
    if (!fe)
      {
	stringstream str;
	str << "HCurlHighOrderFESpaceAlt " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetSElType(selnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }

    ArrayMem<int, 8> vnums;
    ArrayMem<int, 4> ednums, ord_edge, ug_edge; 
       
    ma.GetSElVertices(selnr, vnums);

    if(fe == segm) 
      {
	HCurlHighOrderFiniteElement<1> * hofe =
	  dynamic_cast<HCurlHighOrderFiniteElement<1>*> (fe);
	
	hofe -> SetVertexNumbers (vnums);
	
	ma.GetSElEdges(selnr, ednums);
	hofe -> SetOrderInner (order_edge[ednums[0]]-1);
	hofe -> SetUsegradCell (usegrad_edge[ednums[0]]);
	hofe -> ComputeNDof();
      }
    else 
      {     
	HCurlHighOrderFiniteElement<2> * hofe =
	  dynamic_cast<HCurlHighOrderFiniteElement<2>*> (fe);
      
	ma.GetSElEdges(selnr, ednums);
	
	ord_edge.SetSize (ednums.Size());
	ug_edge.SetSize (ednums.Size()); 
	
	for (j = 0; j < ednums.Size(); j++)
	  {
	    ord_edge[j] = order_edge[ednums[j]]; 
	    ug_edge[j] = usegrad_edge[ednums[j]];
	  }
	
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (ord_edge);
	hofe -> SetOrderInner (INT<3> (order_face[ma.GetSElFace(selnr)][0],order_face[ma.GetSElFace(selnr)][1],0));
      
	hofe -> SetUsegradEdge(ug_edge); 
	hofe -> SetUsegradCell(usegrad_face[ma.GetSElFace(selnr)]); 

	/*ArrayMem<int, 8> order_vert;
	for (int j = 0; j < vnums.Size(); j++)
	  order_vert[j] = order_avertex[vnums[j]];
	  hofe -> SetOrderVertex (order_vert);*/


	hofe -> ComputeNDof();
	
      }
    
    return *fe;
  }
 


  int HCurlHighOrderFESpaceAlt :: GetNDof () const
  {
    return ndof;
  }

  void HCurlHighOrderFESpaceAlt :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    (*testout) << "dofs of el nr " << elnr << endl;
    // Save phase-functions the way 
    // (1*e1),.. (1*e_ne)  
    // (p_t)*tang_e1, ... p_t*tang_ne
    // p_n * el1, p_i * el1, ... , p_n * nel_n , p_i *el_nel 

    ARRAY<int> vnums, ednums, fnums;
    int i, j;
    int first,next; 

    ma.GetElEdges (elnr, ednums);
    if(ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fnums); 
    else 
      fnums.SetSize(0); 
    dnums.SetSize(0);
 
    if(order < 0) 
      throw Exception(" HCurlHighOrderFESpaceAlt :: GetDofNrs() order < 0 "); 

    //Nedelec0
    for (i = 0; i < ednums.Size(); i++) 
      dnums.Append (ednums[i]);
    
    //edges
    int firstinner = first_inner_dof[elnr];

    for (i = 0; i < ednums.Size(); i++) 
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1]; 
	
	for (j = first; j < next; j++)
	  dnums.Append (j);

	dnums.Append (firstinner);
	firstinner++;
	//	dnums.Append (firstinner);
	//	firstinner++;
      }
    
    // faces 
    for(i=0; i<fnums.Size(); i++)     
      { 
	first = first_face_dof[fnums[i]]; 
	next = first_face_dof[fnums[i]+1];

	if (next > first)
	  {
	    dnums.Append (firstinner);
	    firstinner++;
	  }
	for(j=first ; j<next; j++) 
	  dnums.Append(j); 
      } 
    
    //inner 
    first = firstinner;
    next = first_inner_dof[elnr+1]; 

    for(j=first; j<next; j++) 
      dnums.Append(j);

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
  }
 

  void HCurlHighOrderFESpaceAlt :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }

    ARRAY<int> vnums, ednums, fnums;
    int i, j;
    int first,next; 

    ma.GetElEdges (elnr, ednums);
    if(ma.GetDimension() == 3)
      ma.GetElFaces (elnr, fnums); 
    else 
      fnums.SetSize(0); 
    dnums.SetSize(0);

 
    if(order < 0) 
      throw Exception(" HCurlHighOrderFESpaceAlt :: GetDofNrs() order < 0 "); 

    //Nedelec0
    for (i = 0; i < ednums.Size(); i++) 
      dnums.Append (ednums[i]);
    
    //edges
    for (i = 0; i < ednums.Size(); i++) 
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1]; 
	
	for (j = first; j <next; j++)
	  dnums.Append (j);
      }
    // faces 

    for(i=0; i<fnums.Size(); i++)     
      { 
	first = first_face_dof[fnums[i]]; 
	next = first_face_dof[fnums[i]+1];
	for(j=first ; j<next; j++) 
	  dnums.Append(j); 
      } 

    if (!DefinedOn (ma.GetElIndex (elnr)))
      dnums = -1;
  }
 


  void HCurlHighOrderFESpaceAlt :: GetSDofNrs (int selnr, ARRAY<int> & dnums) const
  {
    // Save shape-functions the way 
    // (1*e1),.. (1*e_ne)  
    // (p_t)*tang_e1, ... p_t*tang_ne
    // p_n * el1, p_i * el1, ... , p_n * nel_n , p_i *el_nel 

    ARRAY<int> vnums, ednums;
    int fnum; 
    int i, j;
    int first,next; 

    ma.GetSElEdges (selnr, ednums);
    fnum = ma.GetSElFace (selnr); 

    dnums.SetSize(0);

    if(order <0) throw (" HCurlHighOrderFESpaceAlt :: GetSDofNrs() order < 0 "); 
    //Nedelec
    for (i = 0; i < ednums.Size(); i++) 
      dnums.Append (ednums[i]);

    //edges_tang
    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_edge_dof[ednums[i]];
	next = first_edge_dof[ednums[i]+1]; 
	for (j = first; j <next; j++)
	  dnums.Append (j);
	dnums.Append (-1);
	//	dnums.Append (-1);
      }

    if(first_face_dof.Size()>1)		       
      {
	// dnums.Append (-1);
	// inner = normal + bubbles 
	first = first_face_dof[fnum];
	next = first_face_dof[fnum+1]; 
	for(j=first; j<next; j++) 
	  dnums.Append(j);
      }
    
    if (!DefinedOnBoundary (ma.GetSElIndex (selnr)))
      dnums = -1;
  }


  void HCurlHighOrderFESpaceAlt ::
SetGradientDomains (const BitArray & adoms)
{
  gradientdomains = adoms;
}

static double InnerProduct (const Vec<3> & a, const Vec<3> & b)
{
  return a(0)*b(0) + a(1)*b(1) + a(2)*b(2);
}

void HCurlHighOrderFESpaceAlt ::
SetGradientBoundaries (const BitArray & abnds)
{
  gradientboundaries = abnds;
}

  
  Table<int> * HCurlHighOrderFESpaceAlt :: 
  CreateSmoothingBlocks (const Flags & precflags) const
  {
    int i, j, k, first;
    int ncnt; 
    
    int ii; 

    //    int SmoothingType=7; 
    int SmoothingType = smoother; 

    
    // 1: ned-block, ho-edes, inner 
    // 2: neds, ho-edges, inner 
    // 3: ned+ho-edges, inner
    // 4: ho-edges, inner (no ned Block)
    // 5: vertices with lo-edges, ho-edges, inner
    // 6: ho-edges, faces+inner 
    // 7: Clustering: ho-edges, faces, inner
    // 8: vertices with lo+ho edges, faces, inner 
    // 9: vertex augmented
    // 10: type 7 + AFW (for Markus)

    
    cout << "SmoothingType " << SmoothingType << endl; 
    
    ARRAY<int> vnums,elnums; 
    ARRAY<int> orient; 
    ARRAY<int> ednums, fanums, faorient;
    
    /*
    for(i=0; i< nel ; i++)
      ma.GetElVertices (i, vnums);
    
    for(i=0; i< nel; i++)  
      ma.GetElEdges (i,vnums,orient);
	      
    int pn1,pn2; 
    for(i=0; i< ned; i++) 
      ma.GetEdgePNums (i,pn1,pn2);
    */

    switch(SmoothingType) 
      {
      case 1: 
	// 1 RT Block 
	ncnt = 1 + first_edge_dof.Size() -1 + first_inner_dof.Size() -1; 	    
	break; 
	
      case 2: 
	// ned RT Blocks 
	ncnt = ned + first_edge_dof.Size() -1 + first_inner_dof.Size() -1; 	    
	break; 
	
      case 3: 
	// ned RT+ edges_normal Blocks   
	ncnt = ned + first_inner_dof.Size() -1; 	  
	break; 
      case 4: 
	// no  RT Block 
	ncnt = first_edge_dof.Size() -1 + first_inner_dof.Size() -1; 	    
	break; 
	
      case 5: 
	// RT in divergence blocks 
	ncnt = nv + first_edge_dof.Size() -1 + first_inner_dof.Size() -1; 
	break; 
      case 6:
	ncnt = ned + nfa;
	break;
      case 7: 
	ncnt = nv + ned + nfa + nel;
	break;
      case 8: 
	ncnt = nv + ned + nfa + nel;
	break;
      case 9: 
	ncnt = nv + ned + nfa + nel;
	break;
      case 10:
	ncnt = nv + ned + nfa + nel; 
	break;
      default: 
	cout << " Error in HCurlHOFESpaceAlt :: CreateSmoothingBlocks no SmoothingType " << endl; 
	ncnt =0; 
      } 

    ARRAY<int> cnt(ncnt); 
    cnt = 0;
    
    switch (SmoothingType)
      {
      case 1: 
	ii = 0;  
	cnt[ii] = ned; //RT
	ii++;
	for(i=0; i< first_edge_dof.Size()-1;i++ )
	  cnt[ii++] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< first_inner_dof.Size()-1;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 	
	break; 
      case 2: 
	ii = 0;  
	for(i=0; i< ned; i++)  cnt[ii++] = 1; 
	for(i=0; i< first_edge_dof.Size()-1;i++ )
	  cnt[ii++] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< first_inner_dof.Size()-1;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 
	break; 
      case 3: 
	ii = 0;  
	for(i=0; i< ned; i++)  cnt[i] = 1; 
	for(i=0; i< first_edge_dof.Size()-1;i++ )
	  cnt[i] += first_edge_dof[i+1] - first_edge_dof[i];
	ii=ned;
	for(i=0; i< first_inner_dof.Size()-1;i++)
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 
	break; 
      case 4: 
	ii = 0;  
	for(i=0; i< first_edge_dof.Size()-1;i++ )
	  cnt[ii++] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< first_inner_dof.Size()-1;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 	
	break; 	
      case 5: 
	int v1, v2; 
	for(i=0; i< nv; i++) cnt[i]=0;  
	for(i=0; i< ned; i++) 
	  {
	    ma.GetEdgePNums (i, v1, v2);
	    cnt[v1]++; 
	    cnt[v2]++;
	  }
	ii = nv; 
	for(i=0; i< first_edge_dof.Size()-1;i++ )
	  cnt[ii++] = first_edge_dof[i+1] - first_edge_dof[i]; 
	for(i=0; i< first_inner_dof.Size()-1;i++ )
	  cnt[ii++] = first_inner_dof[i+1] - first_inner_dof[i]; 	
	break; 	
      case 6:
	{
	  for (i = 0; i < ned; i++)
	    cnt[i] = first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[ned+i] = first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < nel; i++)
	    {	   
	      if(ma.GetDimension() == 3) ma.GetElFaces (i, fanums, faorient);
	      else fanums.SetSize(0); 
	      for (j = 0; j < fanums.Size(); j++)
		cnt[ned+fanums[j]] += first_inner_dof[i+1]-first_inner_dof[i];
	    }
	  break;
	}
      case 7:
	{
	  /*  
	  for (i = 0; i < ned; i++)
	    cnt[i] = first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[ned+i] = first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < nel; i++)
	    cnt[ned+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i];
	  break;
	  */
	  for (i = 0; i < ned; i++)
	    cnt[ma.GetClusterRepEdge(i)] +=
	      first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[ma.GetClusterRepFace(i)] += 
	      first_face_dof[i+1] - first_face_dof[i];
	  if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	      cnt[ma.GetClusterRepElement(i)] += 
		first_inner_dof[i+1] - first_inner_dof[i];




	  /*
	  for(i=0; i<nfa; i++)
	    {
	      ma.GetFaceEdges(i,ednums);
	      if(ednums.Size() != 4) continue;
	      
	      int nr_self = ma.GetClusterRepFace(i);

	      ARRAY<int> rep;
	      for(j=0; j<ednums.Size(); j++)
		{
		  int nr = ma.GetClusterRepEdge(ednums[j]);
		  if(nr != nr_self && !rep.Contains(nr))
		    rep.Append(nr);
		}
	      for(k=0; k<rep.Size(); k++)
		cnt[rep[k]] += first_face_dof[i+1] - first_face_dof[i];
	    }
	  */
	      

	  


	  break;
	}
      case 8:
	{
	  for (i = 0; i < ned; i++)
	    {
	      int pn1, pn2;
	      ma.GetEdgePNums (i,pn1,pn2);
	      cnt[pn1] += 1 + first_edge_dof[i+1] - first_edge_dof[i];
	      cnt[pn2] += 1 + first_edge_dof[i+1] - first_edge_dof[i];
	    }
	  for (i = 0; i < nfa; i++)
	    {
	      cnt[nv+ned+i] = first_face_dof[i+1] - first_face_dof[i];
	    }
	  if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	      cnt[nv+ned+nfa+i] = first_inner_dof[i+1] - first_inner_dof[i];
	  else
	    for (i = 0; i < nel; i++)
	      cnt[nv+ned+nfa+i] = 0;
	  break;
	}
      case 9:
	{
	  for (i = 0; i < nv; i++)
	    cnt[i] = 1;
	  for (i = 0; i < ned; i++)
	    cnt[nv+i] = first_edge_dof[i+1] - first_edge_dof[i];
	  for (i = 0; i < nfa; i++)
	    cnt[nv+ned+i] = first_face_dof[i+1] - first_face_dof[i];
	  for (i = 0; i < nel; i++)
	    {
	      ma.GetElEdges (i,ednums,orient);
	      for (k = 0; k < ednums.Size(); k++)
		cnt[nv+ednums[k]] += first_inner_dof[i+1] - first_inner_dof[i];

	      ma.GetElFaces (i,fanums,orient);
	      for (k = 0; k < fanums.Size(); k++)
		cnt[nv+ned+fanums[k]] += first_inner_dof[i+1] - first_inner_dof[i];
	    }
	  break;
	}
     case 10:
	{
	  for (i = 0; i < ned; i++)
	    {
	      int pn1, pn2;
	      ma.GetEdgePNums (i,pn1,pn2);
	      pn1 = ma.GetClusterRepVertex(pn1);
	      pn2 = ma.GetClusterRepVertex(pn2);
	      cnt[pn1]++;
	      if (pn1 != pn2)
		cnt[pn2]++;
	      cnt[ma.GetClusterRepEdge(i)] +=
		first_edge_dof[i+1] - first_edge_dof[i];
	    }
	  for (i = 0; i < nfa; i++)
	    cnt[ma.GetClusterRepFace(i)] += 
	      first_face_dof[i+1] - first_face_dof[i];
	  if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	      cnt[ma.GetClusterRepElement(i)] += 
		first_inner_dof[i+1] - first_inner_dof[i];
	  break;
	}
      }
                
 
    Table<int> & table = *new Table<int> (cnt); 
    
   switch(SmoothingType) 
      {
      case 1: 
	ii=0; 
	for(i=0; i<ned; i++) table[ii][i] = i; //RT 
	ii++;
	for(i=0; i<first_edge_dof.Size()-1; i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	    ii++; 
	  } 
	for(i=0; i<first_inner_dof.Size()-1; i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 	
	break; 
	
      case 2: 
	ii=0; 
	for(i=0; i<ned; i++) table[ii++][0] = i; //RT 
	for(i=0; i<first_edge_dof.Size() -1;i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	    ii++; 
	  } 
	for(i=0; i<first_inner_dof.Size()-1;i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 
	break; 

      case 3: 
	ii=0; 
	for(i=0; i<ned; i++) 
	  { 
	    table[ii++][0] = i; //RT 
	  }
	for(i=0; i<first_edge_dof.Size()-1;i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=1; j< cnt[i]; j++)
	      table[i][j] = j+first-1;  
	  } 
	for(i=0; i<first_inner_dof.Size()-1;i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 
	break; 
      case 4: 
	ii=0; 
	for(i=0; i<first_edge_dof.Size()-1; i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	    ii++; 
	  } 
	for(i=0; i<first_inner_dof.Size()-1; i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] =  j+ first; 
	    ii++;
	  } 	
	break;
      case 5:
	int v1,v2; 
	for(i=0; i< ned; i++) 
	  {
	    ma.GetEdgePNums (i, v1, v2);
	    table[v1][--cnt[v1]] = i; 
	    table[v2][--cnt[v2]] = i; 
	  }
	ii = nv; 
	for(i=0; i<first_edge_dof.Size()-1; i++) 
	  {
	    first = first_edge_dof[i]; 
	    for(j=0; j< cnt[ii]; j++)
	      table[ii][j] = j+first; 
	    ii++; 
	  } 
	for(i=0; i<first_inner_dof.Size()-1; i++) 
	  { 
	    first = first_inner_dof[i]; 
	    for(j=0; j< cnt[ii]; j++) 
	      table[ii][j] = j+ first; 
	    ii++;
	  } 
	break; 
      case 6:
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[i][cnt[i]++] = j;
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[ned+i][cnt[ned+i]++] = j;
	  for (i = 0; i < nel; i++)
	    {
	      if(ma.GetDimension() == 3)  ma.GetElFaces (i, fanums, faorient);
	      else fanums.SetSize(0); 
	      for (j = 0; j < fanums.Size(); j++)
		for (k = first_inner_dof[i]; k < first_inner_dof[i+1]; k++)
		  table[ned+fanums[j]][cnt[ned+fanums[j]]++] = k;
	    }
	  break;
	}
      case 7:
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    {
	      int nr = ma.GetClusterRepEdge (i);
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  for (i = 0; i < nfa; i++)
	    {
	      int nr = ma.GetClusterRepFace (i);
	      for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
 		table[nr][cnt[nr]++] = j;
	    }
	  /*
	  for (i = 0; i < nel; i++)
	    {
	      int nr = ma.GetClusterRepElement (i);
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  */
	  if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	      {
		int nr = ma.GetClusterRepElement (i);
		for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }


	  /*
	  for(i=0; i<nfa; i++)
	    {	      
	      ma.GetFaceEdges(i,ednums);
	      if(ednums.Size() != 4) continue;
	      
	      int nr_self = ma.GetClusterRepFace(i);

	      ARRAY<int> rep;
	      for(j=0; j<ednums.Size(); j++)
		{
		  int nr = ma.GetClusterRepEdge(ednums[j]);
		  if(nr != nr_self && !rep.Contains(nr))
		    rep.Append(nr);
		}
	      for(k=0; k<rep.Size(); k++)
		for(int l=first_face_dof[i]; l<first_face_dof[i+1]; l++)
		  table[rep[k]][cnt[rep[k]]++] = l;
	    }
	  */



	  
	  /*
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[i][cnt[i]++] = j;
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[ned+i][cnt[ned+i]++] = j;
	  for (i = 0; i < nel; i++)
	    for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
	      table[ned+nfa+i][cnt[ned+nfa+i]++] = j;
	  */ 
	

	  break;
	}
      case 8:
	{
	  cnt = 0;
	  for (i = 0; i < ned; i++)
	    {
	      int pn1, pn2;
	      ma.GetEdgePNums (i,pn1,pn2);	      
	      table[pn1][cnt[pn1]++] = i;
	      table[pn2][cnt[pn2]++] = i;
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		{
		  table[pn1][cnt[pn1]++] = j;
		  table[pn2][cnt[pn2]++] = j;
		}
	    }
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[nv+ned+i][cnt[nv+ned+i]++] = j;
	  if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	      for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		table[nv+ned+nfa+i][cnt[nv+ned+nfa+i]++] = j;
	  break;
	}
      case 9:
	{
	  cnt = 0;
	  for (i = 0; i < nv; i++)
	    table[i][cnt[i]++] = nv+i;
	  for (i = 0; i < ned; i++)
	    for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
	      table[nv+i][cnt[nv+i]++] = j;
	  for (i = 0; i < nfa; i++)
	    for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
	      table[nv+ned+i][cnt[nv+ned+i]++] = j;
	  for (i = 0; i < nel; i++)
	    {
	      ma.GetElEdges (i,ednums,orient);
	      for (k = 0; k < ednums.Size(); k++)
		{
		  int dof = nv+ednums[k];
		  for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		    table[dof][cnt[dof]++] = j;
		}
	      ma.GetElFaces (i,fanums,orient);
	      for (k = 0; k < fanums.Size(); k++)
		{
		  int dof = nv+ned+fanums[k];
		  for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		    table[dof][cnt[dof]++] = j;
		}
	    }
	 // (*testout) << "table = " << table << endl;
	  break;
	}

      case 10:
	{
	  cnt = 0; 
	  for (i = 0; i < ned; i++)
	    {
	      int pn1, pn2;
	      ma.GetEdgePNums (i,pn1,pn2);
	      pn1 = ma.GetClusterRepVertex(pn1);
	      pn2 = ma.GetClusterRepVertex(pn2);
	      table[pn1][cnt[pn1]++] = i;
	      if (pn1 != pn2)
		table[pn2][cnt[pn2]++] = i;

	      int nr = ma.GetClusterRepEdge (i);
	      for (j = first_edge_dof[i]; j < first_edge_dof[i+1]; j++)
		table[nr][cnt[nr]++] = j;
	    }
	  for (i = 0; i < nfa; i++)
	    {
	      int nr = ma.GetClusterRepFace (i);
	      for (j = first_face_dof[i]; j < first_face_dof[i+1]; j++)
 		table[nr][cnt[nr]++] = j;
	    }
	  if (!eliminate_internal)
	    for (i = 0; i < nel; i++)
	      {
		int nr = ma.GetClusterRepElement (i);
		for (j = first_inner_dof[i]; j < first_inner_dof[i+1]; j++)
		  table[nr][cnt[nr]++] = j;
	      }
	  break;
	}
      }
      
   /*(*testout) << "Hcurlho::createsmoothingblocks" << endl;
     (*testout) << "cnt = " << cnt << endl;
     (*testout) << "table = " << table << endl;	
   */
   
   return & table;
  }

  ARRAY<int> * 
  HCurlHighOrderFESpaceAlt :: CreateDirectSolverClusters (const Flags & precflags) const
  {
    int nv = ma.GetNV();
    int nd = GetNDof();
    int ne = ma.GetNE();
    int ned = ma.GetNEdges();

    ARRAY<int> ednums,edorient,fnums,forient;


    ARRAY<int> & clusters = *new ARRAY<int> (GetNDof());
    clusters = 0;


    int i, j, k;


    // All Vertical Edges in one Cluster for Hex and Prism + const_x*poly faces + const_y*polx faces (-> 2d Problems !) 
    

    //lo 
    for(i=0;i<ma.GetNEdges();i++)
      clusters[i]=3; 

    for (i = 0; i < ne; i++)
      {
// 	if (ma.GetElType(i) == ET_PYRAMID)
// 	  {
// 	    GetDofNrs(i,ednums);
// 	    for(j=0; j<ednums.Size(); j++) clusters[ednums[j]] = 3;
	    
// 	  }   

	if (ma.GetElType(i) == ET_PRISM)
	  {
	    ma.GetElEdges (i, ednums, edorient);
	    for (j = 0; j < 6; j++)  //horizontal Edges 
	      { 
		int first = first_edge_dof[ednums[j]];
		int next = first_edge_dof[ednums[j]+1];
		for (k = first; k < next; k++)
		  clusters[k] = 0; 
		clusters[ednums[j]]=0;
	      }


	    for (j = 6; j < 9; j++)  //vertical Edges 
	      { 
		int first = first_edge_dof[ednums[j]];
		int next = first_edge_dof[ednums[j]+1];
		for (k = first; k < next; k++)
		  clusters[k] = 3;      //verthoedge
		clusters[ednums[j]]=3; //ned
	      }
	    ma.GetElFaces (i, fnums, forient); 
	    
	    for (j=2; j<5 ; j++)  // vertical faces
	      {
		
		int first = first_face_dof[fnums[j]]; 
		int next = first_face_dof[fnums[j]+1]; 
		
		// TEST SZ eigentlich nurin eine richtung konst ausreichend
		for (k=first; k < next; k++) 
		  clusters[k]=0; 


		INT<2> p = order_face[fnums[j]]; 
		
		for(k=next-p[0]-p[1]; k<next; k++)
		  clusters[k] = 3;
		for(k=0; k<4; k++)
		  clusters[k] = 3;
		

		// for (k=first; k < next; k++) 
		//   clusters[k]=3; 
	
	      }
	     for (j=0; j<2 ; j++)  // horizontal faces
	      {
		
		int first = first_face_dof[fnums[j]]; 
		int next = first_face_dof[fnums[j]+1]; 
		
		for (k=first; k < next; k++) 
		  clusters[k]=0; 
	      }
 	  }

	else if (ma.GetElType(i) == ET_HEX)
	  {
	    ma.GetElEdges (i, ednums, edorient);
	   
	    for(j=0;j<8;j++) // horizontal edges
	       {	
		int first = first_edge_dof[ednums[j]];
		int next = first_edge_dof[ednums[j]+1];
		for (k = first; k < next; k++)
		  clusters[k] = 0;
		
		clusters[ednums[j]]=0; 

	      }
	    

	    for (j = 8; j < 12; j++)  // vertical edges 
	      {	
		int first = first_edge_dof[ednums[j]];
		int next = first_edge_dof[ednums[j]+1];
		for (k = first; k < next; k++)
		  clusters[k] = 3;
		clusters[ednums[j]]=0; 
	      }

	    ma.GetElFaces(i,fnums,forient); // vertical faces 
	      for(j=2;j<6;j++) 
	      {
	
		int first = first_face_dof[fnums[j]]; 
		int next = first_face_dof[fnums[j]+1]; 
		
		/*	for (k=first; k < next; k++) 
		  clusters[k]=0; 
		
		  INT<2> p = order_face[fnums[j]];
		for(k=2*(p[0]+1)*(p[0]+1);k<next;k++) //achtung
		  clusters[k]=3;  
		*/ 
		// TEST SZ  eigentlich in eine richtung konstante nur benoetigt
		for (k=first; k < next; k++) 
		  clusters[k]=3; 

	      }
	      for(j=0;j<2;j++)  //horizontal faces 
	      {
	
		int first = first_face_dof[fnums[j]]; 
		int next = first_face_dof[fnums[j]+1]; 
		
		for (k=first; k < next; k++) 
		  clusters[k]=0; 
	      } 
	  }
	
	for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
          clusters[k]=0; 
	
      }
    





    

//     //lo 
//     for(i=0;i<ma.GetNEdges();i++)
//       clusters[i]=0; 

//     for (i = 0; i < ne; i++)
//       {
// 	ma.GetElPNums(i,pnums); 
// 	if (ma.GetElType(i) == ET_PRISM)
// 	  {
// 	    ma.GetElEdges (i, ednums, edorient);
// 	     for (j = 0; j < 6; j++)  //horizontal Edges 
// 	      { 
// 		int first = first_edge_dof[ednums[j]];
// 		int next = first_edge_dof[ednums[j]+1];
// 		for (k = first; k < next; k++)
// 		  clusters[k] = 0;
// 		clusters[ednums[j]]=0; 
// 	      }


// 	    for (j = 6; j < 9; j++)  //vertical Edges 
// 	      { 
// 		int first = first_edge_dof[ednums[j]];
// 		int next = first_edge_dof[ednums[j]+1];
// 		for (k = first; k < next; k++)
// 		  clusters[k] = 3;      //verthoedge
// 		clusters[ednums[j]]=0; //ned
// 	      }
// 	    ma.GetElFaces (i, fnums, forient); 
	    
// 	    for (j=2; j<5 ; j++)  // vertical faces
// 	      {
		
// 		int first = first_face_dof[fnums[j]]; 
// 		int next = first_face_dof[fnums[j]+1]; 
		
// 		/* TEST SZ eigentlich nurin eine richtung konst ausreichend
// 		for (k=first; k < next; k++) 
// 		  clusters[k]=0; 

// 		int p = order_face[fnums[j]][0]; // achtung !!! 
// 		for(k=2*(p+1)*(p+1);k<next;k++)
// 		  clusters[k]=3;  
		
// 		*/

// 		for (k=first; k < next; k++) 
// 		  clusters[k]=3; 
	
// 	      }
// 	     for (j=0; j<2 ; j++)  // horizontal faces
// 	      {
		
// 		int first = first_face_dof[fnums[j]]; 
// 		int next = first_face_dof[fnums[j]+1]; 
		
// 		for (k=first; k < next; k++) 
// 		  clusters[k]=0; 
// 	      }
// 	  }

// 	else if (ma.GetElType(i) == ET_HEX)
// 	  {
// 	    ma.GetElEdges (i, ednums, edorient);
	   
// 	    for(j=0;j<8;j++) // horizontal edges
// 	       {	
// 		int first = first_edge_dof[ednums[j]];
// 		int next = first_edge_dof[ednums[j]+1];
// 		for (k = first; k < next; k++)
// 		  clusters[k] = 0;
		
// 		clusters[ednums[j]]=0; 

// 	      }
	    

// 	    for (j = 8; j < 12; j++)  // vertical edges 
// 	      {	
// 		int first = first_edge_dof[ednums[j]];
// 		int next = first_edge_dof[ednums[j]+1];
// 		for (k = first; k < next; k++)
// 		  clusters[k] = 3;
// 		clusters[ednums[j]]=0; 
// 	      }

// 	    ma.GetElFaces(i,fnums,forient); // vertical faces 
// 	      for(j=2;j<6;j++) 
// 	      {
	
// 		int first = first_face_dof[fnums[j]]; 
// 		int next = first_face_dof[fnums[j]+1]; 
		
// 		/*	for (k=first; k < next; k++) 
// 		  clusters[k]=0; 
		
// 		  INT<2> p = order_face[fnums[j]];
// 		for(k=2*(p[0]+1)*(p[0]+1);k<next;k++) //achtung
// 		  clusters[k]=3;  
// 		*/ 
// 		// TEST SZ  eigentlich in eine richtung konstante nur benoetigt
// 		for (k=first; k < next; k++) 
// 		  clusters[k]=3; 

// 	      }
// 	      for(j=0;j<2;j++)  //horizontal faces 
// 	      {
	
// 		int first = first_face_dof[fnums[j]]; 
// 		int next = first_face_dof[fnums[j]+1]; 
		
// 		for (k=first; k < next; k++) 
// 		  clusters[k]=0; 
// 	      } 
// 	  }
	
// 	for(k=first_inner_dof[i];k<first_inner_dof[i+1];k++) 
//           clusters[k]=0; 
	
//       }
    








    //  (*testout) << "direct clusters = " << endl << clusters << endl;
    
    for(i=0; directsolverclustered.Size() > 0 && i<ne; i++)
      {
	if(directsolverclustered[ma.GetElIndex(i)])
	  {
	    GetDofNrs(i,ednums);
	    for(k=0; k<ednums.Size(); k++)
	      {
		clusters[ednums[k]] = 4;
	      }
	  }
      }
    
    return &clusters;
    
  }
  


    
  
  /*  BitArray * HCurlHighOrderFESpaceAlt :: 
  CreateIntermediatePlanes (int type) const
  {
    int i;
    ARRAY<int> vnums;
 
    //bool has_cluster = 0;
     for (i = 0; i < ned; i++)
      {
	int pi1, pi2;
	ma.GetEdgePNums (i, pi1, pi2);
	
	if (ma.GetClusterRepVertex (pi1) ==
	    ma.GetClusterRepVertex (pi2))
	  has_cluster = 1;
      }

    if (!has_cluster) return 0;
    
   
    BitArray & ba = *new BitArray (GetNDof());
    ba.Clear();

    for (i = 0; i < ned; i++)
      {
	int pi1, pi2;
	ma.GetEdgePNums (i, pi1, pi2);
	
	if (ma.GetClusterRepVertex (pi1) ==
	    ma.GetClusterRepVertex (pi2))
	  {
	    ba.Set (i);
	    for (int l = first_edge_dof[i];
		 l < first_edge_dof[i+1]; l++)
	      ba.Set (l);
	  }
      }
    
    
    for (i = 0; i < nfa; i++)
      {
	ma.GetFacePNums (i, vnums);
	if (vnums.Size() == 4)
	  {
	    for (int l = first_face_dof[i];
		 l < first_face_dof[i+1]; l++)
	      ba.Set (l);
	  }	    
	  }
    return &ba;
    }*/


  SparseMatrix<double> * 
  HCurlHighOrderFESpaceAlt :: CreateGradient() const
  {
    Flags flags2 (flags);
    flags2.SetFlag("order", order+1);
    flags2.SetFlag("relorder", rel_order+1); 
    flags2.SetFlag("orderinner",uniform_order_inner+1); 
    flags2.SetFlag("orderface",uniform_order_face+1); 
    flags2.SetFlag("orderedge",uniform_order_edge+1); 
    flags2.SetFlag ("orderquad", -1);
    flags2.SetFlag ("ordertrig", -1);  
    flags2.SetFlag("variableorder",var_order); 
    
    int ndom = ma.GetNDomains();
    ARRAY<double> dom_order_h;
    dom_order_h.SetSize(ndom); 
    
    int i; 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_min[i][0]+1; 
    flags2.SetFlag("dom_order_min_x", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_min[i][1]+1; 
    flags2.SetFlag("dom_order_min_y", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_min[i][2]+1; 
    flags2.SetFlag("dom_order_min_z", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_max[i][0]+1; 
    flags2.SetFlag("dom_order_max_x", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_max[i][1]+1; 
    flags2.SetFlag("dom_order_max_y", dom_order_h); 
    for(i=0;i<ndom;i++) 
      dom_order_h[i] = dom_order_max[i][2]+1; 
    flags2.SetFlag("dom_order_max_z", dom_order_h); 
    
    H1HighOrderFESpace  fesh1(ma, flags2); 
    fesh1.Update();
     
    int ned = ma.GetNEdges(); 
    int nv  = ma.GetNV(); 
    int nel = ma.GetNE(); 
    int nfa = 0;
    if(ma.GetDimension()==3) nfa= ma.GetNFaces();  

    int dim     = fesh1.GetDimension();
    int dimcurl = GetDimension();

    ARRAY<int> dnums_h1l; 
    ARRAY<int> dnums_hcl;
    LocalHeap lh (1000000);
    
    // Matrix Graph for gradient matrix , start (end) position for assembling 
    ARRAY<int> cnts(ndof);
    cnts = 0; 

    *testout << "grad fine edges " << fine_edge << endl ; 
    *testout << "grad fine faces " << fine_edge << endl ; 
   
    for(int i=0; i<ned; i++)
      {
	if(fine_edge[i])
	  {
	    cnts[i] = 2;  // vertices-nedelec
	    int l = first_edge_dof[i]; 
	    for( int k = fesh1.GetFirstEdgeDof(i); k< fesh1.GetFirstEdgeDof(i+1); 
		 k++, l++)
	      cnts[l] = 1;
	  }
      }
    
    for(int i=0; i< nfa; i++) 
      {
	if(fine_face[i])
	  {
	    int l= first_face_dof[i]; 
	    for (int k = fesh1.GetFirstFaceDof(i); k<fesh1.GetFirstFaceDof(i+1); 
		 k++, l++) 
	      cnts[l] = 1;
	  }
      }
    
    for(int i=0; i<nel; i++)
      {
	int l= first_inner_dof[i]; 
	for (int k = fesh1.GetFirstElementDof(i); k<fesh1.GetFirstElementDof(i+1); 
	     k++, l++) 
	  cnts[l] = 1; 
      }
  
    //sparse matrix with above matrix graph cnts 
    SparseMatrix<double> & grad = *new SparseMatrix<double>(cnts);  
    
    // vertices - nedelec
    // ho-edges
    for (int i = 0; i < ned; i++) 
      {
	if(fine_edge[i])
	  { 
	    int p1,p2; 
	    ma.GetEdgePNums(i,p1,p2); 
	    grad.CreatePosition(i,p1); 
	    grad.CreatePosition(i,p2); 
	    if (p1 < p2) 
	      {	      
		grad(i,p1) = -1.;
		grad(i,p2) = 1.; 
	      }
	    else
	      {
		grad(i,p1) = 1.; 
		grad(i,p2) = -1.; 
	      }
	    
	    int l = first_edge_dof[i]; 
	    for( int k = fesh1.GetFirstEdgeDof(i); k< fesh1.GetFirstEdgeDof(i+1); 
		 k++, l++)
	      {
		grad.CreatePosition(l,k); 
		grad(l,k)=1.; 
	      }
	  }
      }
    
    for(int i=0; i< nfa; i++) 
      {
	if(fine_face[i])
	  {
	    int l= first_face_dof[i]; 
	    for (int k = fesh1.GetFirstFaceDof(i); k<fesh1.GetFirstFaceDof(i+1); 
		 k++, l++) 
	      {
		grad.CreatePosition(l,k); 
		grad(l,k)=1.;
	      }
	  }
      }
    
    for(int i=0; i<nel; i++)
      {
	int l= first_inner_dof[i]; 
	for (int k = fesh1.GetFirstElementDof(i); k<fesh1.GetFirstElementDof(i+1); 
	     k++, l++) 
	  {
	    grad.CreatePosition(l,k); 
	    grad(l,k)=1.;
	  }	
      }
    
   (*testout) << " Global grad " << grad << endl; 
    return &grad;
  }

  void HCurlHighOrderFESpaceAlt :: GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
  }

  void HCurlHighOrderFESpaceAlt :: GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(1);
    dnums[0] = ednr;
    int first = first_edge_dof[ednr];
    int next = first_edge_dof[ednr+1]; 
    for (int j = first; j <next; j++)
      dnums.Append (j);
  }

  void HCurlHighOrderFESpaceAlt :: GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_face_dof[fanr];
    int next = first_face_dof[fanr+1]; 
    for (int j = first; j <next; j++)
      dnums.Append (j);
  }

  void HCurlHighOrderFESpaceAlt :: GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    dnums.SetSize(0);
    int first = first_inner_dof[elnr];
    int next = first_inner_dof[elnr+1]; 
    for (int j = first; j <next; j++)
      dnums.Append (j);  
  }


    
  // register FESpaces
  namespace {
     
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hcurlho_alt", HCurlHighOrderFESpaceAlt::Create);
    }
    
    Init init;
  }
  
}


