#ifndef FILE_HDIVHOFESPACE
#define FILE_HDIVHOFESPACE



/*********************************************************************/
/* File:   hdivhofespace.hpp                                         */
/* Author: START 
/* Date:   Feb. 2003, update (SZ) Jan. 2007                          */
/*********************************************************************/

/**
   HDiv High Order Finite Element Space
*/

class HDivHighOrderFESpace : public FESpace
{
protected:
  // Level
  int level;
  // Number of Edges
  int ned;
  // Number of Faces
  int nfa;
  // Number of Elements
  int nel;
  // Number of Vertex
  int nv;
  // Number Dofs 
  int ndof;
  // order of curl-fields 
  int curl_order; 

  ARRAY<int> first_facet_dof;
  ARRAY<int> first_inner_dof;

  /// relative order to mesh-order
  int rel_order; 
  // curl-order relative to mesh order 
  int rel_curl_order; 
  // space is of variable order 
  bool var_order;
  // space is continuous/discontinuous 
  bool discont; 
    
  ARRAY<INT<3> > order_inner;
  ARRAY<INT<3> > order_inner_curl;
  ARRAY<INT<2> > order_facet; 
  ARRAY<bool> fine_facet; 
 
  ARRAY<int> ndlevel;
  int uniform_order_inner; 
  int uniform_order_facet; 

  bool print; 
     
public:

  HDivHighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags=false);
  ///
  virtual ~HDivHighOrderFESpace ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  void UpdateDofTables(); 

  virtual string GetClassName () const
  {
    return "HDivHighOrderFESpace";
  }

  ///
  virtual void Update();
  ///
  virtual int GetNDof () const;
  ///
  virtual int GetNDofLevel (int level) const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int selnr, LocalHeap & lh) const; // 2D: array =0.;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  ///
  virtual Table<int> * CreateSmoothingBlocks (const Flags & precflags) const;
  /// 
  virtual ARRAY<int> * CreateDirectSolverClusters (const Flags & precflags) const;

  /// 
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetFacetDofNrs(int fanr, ARRAY<int> & dnums) const 
  { 
    if (ma.GetDimension() == 2) GetEdgeDofNrs(fanr,dnums); 
    else if (ma.GetDimension() == 3) GetFaceDofNrs(fanr,dnums); 
  } 
  ///
  virtual void GetWireBasketDofNrs ( int elnr, ARRAY<int> & dnums ) const;
	     
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const; 
  
  /// 
  void GetFacetOrder (ARRAY<INT<2> > & of, ARRAY<bool> & ff) const 
  {of = order_facet; ff = fine_facet;};

  /// 
  int GetNElemDofs(int elnr) const 
  {
    if(discont) return(first_inner_dof[elnr+1] - first_inner_dof[elnr]); 
    else 
      { 
	ARRAY<int> dnums; 
	this->GetDofNrs(elnr,dnums); 
	return(dnums.Size()); 
      } 
  }

  int GetFirstInnerDof(int elnr) const { return(first_inner_dof[elnr]);}; 
  // virtual int LowOrderDof() const { if(discont) return(0); else return(1);} 


  virtual bool VarOrder() const { return var_order; } 
  virtual int GetRelOrder() const { return rel_order; } 

#ifdef PARALLEL
   virtual void UpdateParallelDofs_hoproc();
   virtual void UpdateParallelDofs_loproc();
#endif
};

#endif





