#ifndef FILE_FASTINTEGRATOR
#define FILE_FASTINTEGRATOR

/*********************************************************************/
/* File:   fastintegrator.hpp                                        */
/* Author: Start                                                     */
/* Date:   6. Jan. 2005                                              */
/*********************************************************************/

  
  template <int DIM>
  class FastLaplaceIntegrator : public BilinearFormIntegrator
  {
    CoefficientFunction * coef;
    BilinearFormIntegrator * orig;

    typedef DiffOpGradient<DIM> DIFFOP;
    enum { DIM_SPACE   = DIFFOP::DIM_SPACE };
    enum { DIM_ELEMENT = DIFFOP::DIM_ELEMENT };
    enum { DIM_DMAT    = DIFFOP::DIM_DMAT };
    // enum { DIM         = DIFFOP::DIM };

    DiagDMat<DIM> dmatop;
    
    typedef H1HighOrderTetFast FEL;
  public:
    FastLaplaceIntegrator (CoefficientFunction * acoef)
      : BilinearFormIntegrator (), coef(acoef), dmatop (acoef)
    { 
      orig = new LaplaceIntegrator<DIM> (acoef); 
      fast = 1;
    }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new FastLaplaceIntegrator (coeffs[0]);
    }

    virtual bool BoundaryForm () const { return 0; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const;


    virtual void
    AssembleElementMatrix (const FiniteElement & fel,
			   const ElementTransformation & eltrans, 
			   HashTable<INT<2>, double> & elmat,
			   LocalHeap & lh) const;




    virtual void
    AssembleElementMatrixDiag (const FiniteElement & fel,
			       const ElementTransformation & eltrans, 
			       FlatVector<double> & diag,
			       LocalHeap & lh) const;



    virtual void *  
    PrecomputeData (const FiniteElement & fel, 
		    const ElementTransformation & eltrans, 
		    LocalHeap & lh) const;

    

    virtual void
    ApplyElementMatrix (const FiniteElement & bfel, 
			const ElementTransformation & eltrans, 
			const FlatVector<double> & elx, 
			FlatVector<double> & ely,
			void * precomputed,
			LocalHeap & lh) const;


    virtual string Name () const { return "fast mass tet"; }
  };


#endif 
