#include <fem.hpp>
  
namespace ngfem
{
  using namespace ngfem;
  









  template <>
  void FastLaplaceIntegrator<3> ::
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    FlatMatrix<> mat1;
    if (checkfast)
      {
	orig -> AssembleElementMatrix (bfel, eltrans, mat1, lh); 
	// (*testout) << "original: " << mat1 << endl;
      }



    if (dynamic_cast<const H1HighOrderTetFast*> (&bfel))
      {
	const H1HighOrderTetFast & fel = static_cast<const H1HighOrderTetFast&> (bfel);
	int ndof = fel.GetNDof();
        
	elmat.AssignMemory (ndof, ndof, lh);
	elmat = 0;
	
	void * heapp = lh.GetPointer();

	int intorder = 2*fel.Order() + 6;
	if (common_integration_order >= 0)
	  intorder = common_integration_order;

	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, intorder);
	
	TensorProductIntegrationRule ir (fel, ir1d);
	

	FlatVector<Mat<3> > dmats(ir.GetNIP(), lh);
	
	void * heapp2 = lh.GetPointer();
	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    SpecificIntegrationPoint<3,3> sip(ir[i], eltrans, lh);
		
	    Mat<3> dmat, thextet, thextet_inv, jac_inv, trans;

	    double val = coef -> Evaluate (sip);
	    val *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
	    dmat = val * Id<3>();

	    jac_inv = sip.GetJacobianInverse();

	    ir.GetIP_DTet_DHex(i, thextet);
	    CalcInverse (thextet, thextet_inv);

	    trans = Trans (jac_inv) * Trans (thextet_inv);
	    dmats[i] = Trans (trans) * dmat * trans;

	    // (*testout) << ", dmats = " << dmats[i] << endl;
	    //	    if ( (i % ir1d.GetNIP()) == ir1d.GetNIP()-1) (*testout) << endl << endl;

	    lh.CleanUp (heapp2);
	  }
	



	int order = fel.Order();
	
	FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
	FlatMatrix<> dlegxbub(ir.GetNIPX(), order+1, lh);
	FlatArray<AutoDiff<1> > hvals (order+1, lh);
	FlatArray<AutoDiff<1> > hvals2 (order+1, lh);

	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    AutoDiff<1> x (ir.GetIPX(ix), 0);

	    AutoDiff<1> * hp = &hvals[2];
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);
	    
	    hvals[0] = x;
	    hvals[1] = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      hvals[j+2] *= x * (1-x);

	    for (int j = 0; j <= order; j++)
	      {
		legxbub(ix, j) = hvals[j].Value();
		dlegxbub(ix, j) = hvals[j].DValue(0);
	      }
	  }

	ARRAY<FlatMatrix<> > jaczbTs (order+1);
	ARRAY<FlatMatrix<> > djaczbTs (order+1);

	for (int k = 0; k <= order; k++)
	  {
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);
	    djaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);

	    if (k < 2)
	      {
		jaczbTs[k] = Trans (legxbub);
		djaczbTs[k] = Trans (dlegxbub);
	      }
	    else
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  AutoDiff<1> z (ir.GetIPZ(iz), 0);
		  AutoDiff<1> powzk = 1.0;
		  for (int l = 0; l < k; l++)
		    powzk *= (1-z);

		  AutoDiff<1> * hp = &hvals[1];
		  JacobiPolynomial (order-1, 2*z-1, 2*k-3, 1, hp);
		  for (int j = 1; j <= order; j++)
		    hvals[j] *= z * powzk;
		  hvals[0] = powzk;

		  for (int j = 0; j <= order; j++)
		    {
		      jaczbTs[k](j, iz) = hvals[j].Value();
		      djaczbTs[k](j, iz) = hvals[j].DValue(0);
		    }
		}
	  }
	


	int sort[4] = { 0, 1, 2, 3 };
	for (int i = 0; i < 4; i++)
	  for (int j = 0; j < 3; j++)
	    if (fel.GetVNums()[sort[j]] > fel.GetVNums()[sort[j+1]]) swap (sort[j], sort[j+1]);
    
	int isort[4];
	for (int i = 0; i < 4; i++) isort[sort[i]] = i;



	// precompute 2D matrices

	int ndof2d = (order+1)*(order+2)/2;
	int ndof1d = order+1;
	int ngrid2d = ir.GetNIPX()*ir.GetNIPY();


	ARRAY<FlatMatrix<> > mat2d(ir.GetNIPZ());

	ARRAY<FlatMatrix<> > mat1d(ir.GetNIPY());
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  mat1d[iy].AssignMemory (ndof1d, ndof1d, lh);

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  mat2d[iz].AssignMemory (ndof2d, ndof2d, lh);

	FlatMatrix<> mat1dy (order+1, lh);

	FlatVector<> coefy(ir.GetNIPY(), lh);
	FlatVector<> coefyb1(ir.GetNIPY(), lh);


	FlatArray<int> quad2trig((order+1)*(order+1), lh);
	quad2trig = -1;
	quad2trig[0] = 0;
	quad2trig[order+1] = 0;
	quad2trig[1] = 1;
	quad2trig[order+2] = 2;
	
	int pos = 3;
	for (int ky = 2; ky <= order; ky++)
	  {
	    quad2trig[ky] = pos++;
	    quad2trig[ky+order+1] = pos++;
	  }
	for (int ky = 0; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    quad2trig[kx*(order+1)+ky] = pos++;

	FlatVector<> coefx(ir.GetNIPX(), lh);
	FlatVector<> hvx(order+1, lh);



	elmat = 0.0;

	for (int diff1 = 0; diff1 < 3; diff1++)
	  for (int diff2 = 0; diff2 < 3; diff2++)
	    {
	      // (*testout) << "diff1 = " << diff1 << ", diff2 = " << diff2 << endl;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  mat2d[iz] = 0.0;

		  // precompute 1D matrices mat1d[iy]
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      int first_dmati = iz*ngrid2d+iy*ir.GetNIPX();

		      FlatMatrix<> base1( (diff1 == 0) ? dlegxbub : legxbub );
		      FlatMatrix<> base2( (diff2 == 0) ? dlegxbub : legxbub );
		      
		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			coefx(ix) = dmats[first_dmati+ix](diff1, diff2);

		      for (int i = 0; i < order+1; i++)
			for (int j = 0; j < order+1; j++)
			  {
			    double sum = 0;
			    for (int ix = 0; ix < ir.GetNIPX(); ix++)
			      sum += base1(ix, i) * coefx(ix) * base2(ix, j);
			    mat1d[iy](i,j) = sum;
			  }
		    }


		  
		  for (int kx1 = 0; kx1 <= order; kx1++)
		    for (int kx2 = 0; kx2 <= order; kx2++)
		      {
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  coefy(iy) = mat1d[iy](kx1,kx2);

			FlatMatrix<> base1( (diff1==1) ? djaczbTs[kx1] : jaczbTs[kx1]);
			FlatMatrix<> base2( (diff2==1) ? djaczbTs[kx2] : jaczbTs[kx2]);
			
			for (int i = 0; i < order+1; i++)
			  {
			    int mi = quad2trig[kx1*(order+1)+i];
			    if (mi != -1)
			      {
				for (int k = 0; k < ir.GetNIPY(); k++)
				  coefyb1(k) = base1(i,k) * coefy(k);
			  
				for (int j = 0; j < order+1; j++)
				  {
				    int mj = quad2trig[kx2*(order+1)+j];
				    if (mj != -1)
				      {
					double sum = 0;
					for (int k = 0; k < ir.GetNIPY(); k++)
					  sum += coefyb1(k) * base2(j,k);
					mat2d[iz](mi, mj) += sum;
				      }
				  }
			      }

			  }
		      }

		  if (iz == -1)
		    {
		      int nz = 0;
		      for (int hi = 0; hi < mat2d[iz].Height(); hi++)
			for (int hj = 0; hj < mat2d[iz].Width(); hj++)
			  if (sqr (mat2d[iz](hi,hj)) <= 1e-20 * fabs (mat2d[iz](hi,hi)*mat2d[iz](hj,hj)) ) nz++;
		      cout << "mat2d has " << double(nz) / (mat2d[iz].Height()*mat2d[iz].Width()) * 100 << " % zero entries ";
		      cout << "these are " << mat2d[iz].Width()-double(nz) / mat2d[iz].Height() << " nze elements per row" << endl;
		      
		      (*testout) << "mat2d has " << double(nz) / (mat2d[iz].Height()*mat2d[iz].Width()) * 100 << " % zero entries ";
		      (*testout) << "these are " << mat2d[iz].Width()-double(nz) / mat2d[iz].Height() << " nze elements per row" << endl;
		      // (*testout) << "mat2d = " << endl << mat2d[iz] << endl;
		    }
		}
      



	      // scaling with (1-z)^m
	
	      FlatArray<int> powerz(ndof2d, lh);
	      for (int k = 0; k < 3; k++)
		powerz[k] = 0;
	
	      pos = 3;
	      for (int k = 2; k <= order; k++)
		{
		  powerz[pos++] = k;
		  powerz[pos++] = k;
		}

	      for (int ky = 0; ky <= order; ky++)
		for (int kx = 2; kx <= order-ky; kx++)
		  powerz[pos++] = kx+ky;



	

	      // transform prism ---> tet

	      ARRAY<int> prism2tet((order+1)*ndof2d);
	      fel.GetPrismTetReordering (isort, prism2tet);
	

	      // build 3D - tensor product matrix:

	      FlatVector<> coefz(ir.GetNIPZ(), lh);
	      FlatVector<> coefzb1(ir.GetNIPZ(), lh);

	      for (int kxy = 0; kxy < ndof2d; kxy++)
		for (int kxy2 = 0; kxy2 < ndof2d; kxy2++)
		  {
		    // get coefficients in z-direction
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      coefz(iz) = mat2d[iz](kxy,kxy2);

		    FlatMatrix<> base1( (diff1==2) ? 
					djaczbTs[powerz[kxy]] : 
					jaczbTs[powerz[kxy]]);
		    FlatMatrix<> base2( (diff2==2) ? 
					djaczbTs[powerz[kxy2]] : 
					jaczbTs[powerz[kxy2]]);

		    for (int i = 0; i < order+1 - powerz[kxy]; i++)
		      {
			int mi = prism2tet[kxy*(order+1)+i];
			if (mi != -1)
			  {
			    for (int k = 0; k < ir.GetNIPZ(); k++)
			      coefzb1(k) = base1(i,k) * coefz(k);

			    for (int j = 0; j < order+1 - powerz[kxy2]; j++)
			      {
				int mj = prism2tet[kxy2*(order+1)+j];
				if (mj != -1)
				  {
				    double sum = 0;
				    for (int k = 0; k < ir.GetNIPZ(); k++)
				      sum += coefzb1(k) * base2(j,k);
				    elmat(mi, mj) += sum;
				  }
			      }
			  }
		      }
		  }
	    }
	
	/*
	int nz = 0;
	for (int hi = 0; hi < elmat.Height(); hi++)
	  for (int hj = 0; hj < elmat.Width(); hj++)
	    if (sqr (elmat(hi,hj)) <= 1e-20) nz++;
	cout << "laplace mat has " << double(nz) / (elmat.Height()*elmat.Width()) * 100 << " % zero entries" << endl;
	cout << "these are " << elmat.Width()-double(nz) / elmat.Height() << " nze elements per row" << endl;
	*/

	lh.CleanUp(heapp);

	if (checkfast)
	  {
	    double sum = 0;
	    for (int i = 0; i < mat1.Height()*mat1.Width(); i++)
	      sum += sqr (mat1(i)-elmat(i));
	    cout << "diff between std and fast: " << sqrt (sum) << endl;
	    (*testout) << "old mat " << endl << mat1 << endl;
	    (*testout) << "new mat " << endl << elmat << endl;
	    (*testout) << "diff mat " << endl << elmat-mat1 << endl;
	  }
      }

    else
      {
	orig -> AssembleElementMatrix (bfel, eltrans, elmat, lh); 	
	return;
      }
  }














  template <>
  void FastLaplaceIntegrator<3> ::
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 HashTable<INT<2>, double> & elmat,
			 LocalHeap & lh) const
  {
#ifdef ABC
    if (dynamic_cast<const H1HighOrderTetFast*> (&bfel))
      {
	const H1HighOrderTetFast & fel = static_cast<const H1HighOrderTetFast&> (bfel);
	int ndof = fel.GetNDof();
        
	void * heapp = lh.GetPointer();

	int intorder = 2*fel.Order() + 6;
	if (common_integration_order >= 0)
	  intorder = common_integration_order;

	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, intorder);
	
	TensorProductIntegrationRule ir (fel, ir1d);
	

	FlatVector<Mat<3> > dmats(ir.GetNIP(), lh);
	
	void * heapp2 = lh.GetPointer();
	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    SpecificIntegrationPoint<3,3> sip(ir[i], eltrans, lh);
		
	    Mat<3> dmat, thextet, thextet_inv, jac_inv, trans;

	    double val = coef -> Evaluate (sip);
	    val *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
	    dmat = val * Id<3>();

	    jac_inv = sip.GetJacobianInverse();

	    ir.GetIP_DTet_DHex(i, thextet);
	    CalcInverse (thextet, thextet_inv);

	    trans = Trans (jac_inv) * Trans (thextet_inv);
	    dmats[i] = Trans (trans) * dmat * trans;

	    // (*testout) << ", dmats = " << dmats[i] << endl;
	    //	    if ( (i % ir1d.GetNIP()) == ir1d.GetNIP()-1) (*testout) << endl << endl;

	    lh.CleanUp (heapp2);
	  }
	



	int order = fel.Order();
	
	FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
	FlatMatrix<> dlegxbub(ir.GetNIPX(), order+1, lh);
	FlatArray<AutoDiff<1> > hvals (order+1, lh);
	FlatArray<AutoDiff<1> > hvals2 (order+1, lh);

	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    AutoDiff<1> x (ir.GetIPX(ix), 0);

	    AutoDiff<1> * hp = &hvals[2];
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);
	    
	    hvals[0] = x;
	    hvals[1] = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      hvals[j+2] *= x * (1-x);

	    for (int j = 0; j <= order; j++)
	      {
		legxbub(ix, j) = hvals[j].Value();
		dlegxbub(ix, j) = hvals[j].DValue(0);
	      }
	  }

	ARRAY<FlatMatrix<> > jaczbTs (order+1);
	ARRAY<FlatMatrix<> > djaczbTs (order+1);

	for (int k = 0; k <= order; k++)
	  {
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);
	    djaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);

	    if (k < 2)
	      {
		jaczbTs[k] = Trans (legxbub);
		djaczbTs[k] = Trans (dlegxbub);
	      }
	    else
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  AutoDiff<1> z (ir.GetIPZ(iz), 0);
		  AutoDiff<1> powzk = 1.0;
		  for (int l = 0; l < k; l++)
		    powzk *= (1-z);

		  AutoDiff<1> * hp = &hvals[1];
		  JacobiPolynomial (order-1, 2*z-1, 2*k-3, 1, hp);
		  for (int j = 1; j <= order; j++)
		    hvals[j] *= z * powzk;
		  hvals[0] = powzk;

		  for (int j = 0; j <= order; j++)
		    {
		      jaczbTs[k](j, iz) = hvals[j].Value();
		      djaczbTs[k](j, iz) = hvals[j].DValue(0);
		    }
		}
	  }
	


	int sort[4] = { 0, 1, 2, 3 };
	for (int i = 0; i < 4; i++)
	  for (int j = 0; j < 3; j++)
	    if (fel.GetVNums()[sort[j]] > fel.GetVNums()[sort[j+1]]) swap (sort[j], sort[j+1]);
    
	int isort[4];
	for (int i = 0; i < 4; i++) isort[sort[i]] = i;



	// precompute 2D matrices

	int ndof2d = (order+1)*(order+2)/2;
	int ndof1d = order+1;
	int ngrid2d = ir.GetNIPX()*ir.GetNIPY();


	ARRAY<FlatMatrix<> > mat2d(ir.GetNIPZ());

	ARRAY<FlatMatrix<> > mat1d(ir.GetNIPY());
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  mat1d[iy].AssignMemory (ndof1d, ndof1d, lh);

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  mat2d[iz].AssignMemory (ndof2d, ndof2d, lh);

	FlatMatrix<> mat1dy (order+1, lh);

	FlatVector<> coefy(ir.GetNIPY(), lh);
	FlatVector<> coefyb1(ir.GetNIPY(), lh);


	FlatArray<int> quad2trig((order+1)*(order+1), lh);
	quad2trig = -1;
	quad2trig[0] = 0;
	quad2trig[order+1] = 0;
	quad2trig[1] = 1;
	quad2trig[order+2] = 2;
	
	int pos = 3;
	for (int ky = 2; ky <= order; ky++)
	  {
	    quad2trig[ky] = pos++;
	    quad2trig[ky+order+1] = pos++;
	  }
	for (int ky = 0; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    quad2trig[kx*(order+1)+ky] = pos++;

	FlatVector<> coefx(ir.GetNIPX(), lh);
	FlatVector<> hvx(order+1, lh);


	for (int diff1 = 0; diff1 < 3; diff1++)
	  for (int diff2 = 0; diff2 < 3; diff2++)
	    {
	      // (*testout) << "diff1 = " << diff1 << ", diff2 = " << diff2 << endl;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  mat2d[iz] = 0.0;

		  // precompute 1D matrices mat1d[iy]
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      int first_dmati = iz*ngrid2d+iy*ir.GetNIPX();

		      FlatMatrix<> base1( (diff1 == 0) ? dlegxbub : legxbub );
		      FlatMatrix<> base2( (diff2 == 0) ? dlegxbub : legxbub );
		      
		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			coefx(ix) = dmats[first_dmati+ix](diff1, diff2);

		      for (int i = 0; i < order+1; i++)
			for (int j = 0; j < order+1; j++)
			  {
			    double sum = 0;
			    for (int ix = 0; ix < ir.GetNIPX(); ix++)
			      sum += base1(ix, i) * coefx(ix) * base2(ix, j);
			    mat1d[iy](i,j) = sum;
			  }
		    }


		  
		  for (int kx1 = 0; kx1 <= order; kx1++)
		    for (int kx2 = 0; kx2 <= order; kx2++)
		      {
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  coefy(iy) = mat1d[iy](kx1,kx2);

			FlatMatrix<> base1( (diff1==1) ? djaczbTs[kx1] : jaczbTs[kx1]);
			FlatMatrix<> base2( (diff2==1) ? djaczbTs[kx2] : jaczbTs[kx2]);
			
			for (int i = 0; i < order+1; i++)
			  {
			    int mi = quad2trig[kx1*(order+1)+i];
			    if (mi != -1)
			      {
				for (int k = 0; k < ir.GetNIPY(); k++)
				  coefyb1(k) = base1(i,k) * coefy(k);
			  
				for (int j = 0; j < order+1; j++)
				  {
				    int mj = quad2trig[kx2*(order+1)+j];
				    if (mj != -1)
				      {
					double sum = 0;
					for (int k = 0; k < ir.GetNIPY(); k++)
					  sum += coefyb1(k) * base2(j,k);
					mat2d[iz](mi, mj) += sum;
				      }
				  }
			      }

			  }
		      }

		  if (iz == -1)
		    {
		      int nz = 0;
		      for (int hi = 0; hi < mat2d[iz].Height(); hi++)
			for (int hj = 0; hj < mat2d[iz].Width(); hj++)
			  if (sqr (mat2d[iz](hi,hj)) <= 1e-20 * fabs (mat2d[iz](hi,hi)*mat2d[iz](hj,hj)) ) nz++;
		      cout << "mat2d has " << double(nz) / (mat2d[iz].Height()*mat2d[iz].Width()) * 100 << " % zero entries ";
		      cout << "these are " << mat2d[iz].Width()-double(nz) / mat2d[iz].Height() << " nze elements per row" << endl;
		      
		      (*testout) << "mat2d has " << double(nz) / (mat2d[iz].Height()*mat2d[iz].Width()) * 100 << " % zero entries ";
		      (*testout) << "these are " << mat2d[iz].Width()-double(nz) / mat2d[iz].Height() << " nze elements per row" << endl;
		      // (*testout) << "mat2d = " << endl << mat2d[iz] << endl;
		    }
		}
      



	      // scaling with (1-z)^m
	
	      FlatArray<int> powerz(ndof2d, lh);
	      for (int k = 0; k < 3; k++)
		powerz[k] = 0;
	
	      pos = 3;
	      for (int k = 2; k <= order; k++)
		{
		  powerz[pos++] = k;
		  powerz[pos++] = k;
		}

	      for (int ky = 0; ky <= order; ky++)
		for (int kx = 2; kx <= order-ky; kx++)
		  powerz[pos++] = kx+ky;



	

	      // transform prism ---> tet

	      ARRAY<int> prism2tet((order+1)*ndof2d);
	      fel.GetPrismTetReordering (isort, prism2tet);
	

	      // build 3D - tensor product matrix:

	      FlatVector<> coefz(ir.GetNIPZ(), lh);
	      FlatVector<> coefzb1(ir.GetNIPZ(), lh);

	      for (int kxy = 0; kxy < ndof2d; kxy++)
		for (int kxy2 = 0; kxy2 < ndof2d; kxy2++)
		  {
		    // get coefficients in z-direction
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      coefz(iz) = mat2d[iz](kxy,kxy2);

		    FlatMatrix<> base1( (diff1==2) ? 
					djaczbTs[powerz[kxy]] : 
					jaczbTs[powerz[kxy]]);
		    FlatMatrix<> base2( (diff2==2) ? 
					djaczbTs[powerz[kxy2]] : 
					jaczbTs[powerz[kxy2]]);

		    for (int i = 0; i < order+1 - powerz[kxy]; i++)
		      {
			int mi = prism2tet[kxy*(order+1)+i];
			if (mi != -1)
			  {
			    for (int k = 0; k < ir.GetNIPZ(); k++)
			      coefzb1(k) = base1(i,k) * coefz(k);

			    for (int j = 0; j < order+1 - powerz[kxy2]; j++)
			      {
				int mj = prism2tet[kxy2*(order+1)+j];
				if (mj != -1)
				  {
				    double sum = 0;
				    for (int k = 0; k < ir.GetNIPZ(); k++)
				      sum += coefzb1(k) * base2(j,k);

				    if (fabs (sum) >= 1e-12 && mi >= mj)
				      {
					INT<2> i2;
					i2[0] = mi;
					i2[1] = mj;

					if (elmat.Used(i2)) sum += elmat.Get(i2);
					elmat.Set (i2, sum);
				      }
				    // elmat(mi, mj) += sum;
				  }
			      }
			  }
		      }
		  }
	    }
	
	lh.CleanUp(heapp);
      }
#endif
  }















  template <> 
  void FastLaplaceIntegrator<3> ::
  AssembleElementMatrixDiag (const FiniteElement & bfel,
			     const ElementTransformation & eltrans, 
			     FlatVector<double> & diag,
			     LocalHeap & lh) const
  {
    FlatVector<> diag1;
    if (checkfast)
      {
	orig -> AssembleElementMatrixDiag (bfel, eltrans, diag1, lh); 
	// (*testout) << "original: " << mat1 << endl;
      }



    if (dynamic_cast<const H1HighOrderTetFast*> (&bfel))
      {
	static int maintimer = NgProfiler::CreateTimer ("FastLaplace::AssembleDiag");
	static int blockxtimer = NgProfiler::CreateTimer ("FastLaplace::AssembleDiag, blockx");
	static int blockytimer = NgProfiler::CreateTimer ("FastLaplace::AssembleDiag, blocky");
	static int blockztimer = NgProfiler::CreateTimer ("FastLaplace::AssembleDiag, blockz");
	static int preparetimer = NgProfiler::CreateTimer ("FastLaplace::AssembleDiag, prepare");

	NgProfiler::StartTimer (maintimer);

	NgProfiler::StartTimer (preparetimer);

	const H1HighOrderTetFast & fel = static_cast<const H1HighOrderTetFast&> (bfel);
	int ndof = fel.GetNDof();
        
	diag.AssignMemory (ndof, lh);
	diag = 0;
	
	void * heapp = lh.GetPointer();

	int intorder = 2*fel.Order() + 6;
	if (common_integration_order >= 0)
	  intorder = common_integration_order;

	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, intorder);
	
	TensorProductIntegrationRule ir (fel, ir1d);
	

	FlatVector<Mat<3> > dmats(ir.GetNIP(), lh);
	
	void * heapp2 = lh.GetPointer();
	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    SpecificIntegrationPoint<3,3> sip(ir[i], eltrans, lh);
		
	    Mat<3> dmat, thextet, thextet_inv, jac_inv, trans;

	    double val = coef -> Evaluate (sip);
	    val *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
	    dmat = val * Id<3>();

	    jac_inv = sip.GetJacobianInverse();

	    ir.GetIP_DTet_DHex(i, thextet);
	    CalcInverse (thextet, thextet_inv);

	    trans = Trans (jac_inv) * Trans (thextet_inv);
	    dmats[i] = Trans (trans) * dmat * trans;

	    // (*testout) << ", dmats = " << dmats[i] << endl;
	    //	    if ( (i % ir1d.GetNIP()) == ir1d.GetNIP()-1) (*testout) << endl << endl;

	    lh.CleanUp (heapp2);
	  }
	



	int order = fel.Order();
	
	FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
	FlatMatrix<> dlegxbub(ir.GetNIPX(), order+1, lh);
	FlatArray<AutoDiff<1> > hvals (order+1, lh);
	FlatArray<AutoDiff<1> > hvals2 (order+1, lh);

	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    AutoDiff<1> x (ir.GetIPX(ix), 0);

	    AutoDiff<1> * hp = &hvals[2];
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);
	    
	    hvals[0] = x;
	    hvals[1] = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      hvals[j+2] *= x * (1-x);

	    for (int j = 0; j <= order; j++)
	      {
		legxbub(ix, j) = hvals[j].Value();
		dlegxbub(ix, j) = hvals[j].DValue(0);
	      }
	  }

	ARRAY<FlatMatrix<> > jaczbTs (order+1);
	ARRAY<FlatMatrix<> > djaczbTs (order+1);

	for (int k = 0; k <= order; k++)
	  {
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);
	    djaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);

	    if (k < 2)
	      {
		jaczbTs[k] = Trans (legxbub);
		djaczbTs[k] = Trans (dlegxbub);
	      }
	    else
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  AutoDiff<1> z (ir.GetIPZ(iz), 0);
		  AutoDiff<1> powzk = 1.0;
		  for (int l = 0; l < k; l++)
		    powzk *= (1-z);

		  AutoDiff<1> * hp = &hvals[1];
		  JacobiPolynomial (order-1, 2*z-1, 2*k-3, 1, hp);
		  for (int j = 1; j <= order; j++)
		    hvals[j] *= z * powzk;
		  hvals[0] = powzk;

		  for (int j = 0; j <= order; j++)
		    {
		      jaczbTs[k](j, iz) = hvals[j].Value();
		      djaczbTs[k](j, iz) = hvals[j].DValue(0);
		    }
		}
	  }
	


	int sort[4] = { 0, 1, 2, 3 };
	for (int i = 0; i < 4; i++)
	  for (int j = 0; j < 3; j++)
	    if (fel.GetVNums()[sort[j]] > fel.GetVNums()[sort[j+1]]) swap (sort[j], sort[j+1]);
    
	int isort[4];
	for (int i = 0; i < 4; i++) isort[sort[i]] = i;



	// precompute 2D matrices

	int ndof2d = (order+1)*(order+2)/2;
	int ndof1d = order+1;
	int ngrid2d = ir.GetNIPX()*ir.GetNIPY();


	ARRAY<FlatMatrix<> > mat2d(ir.GetNIPZ());

	ARRAY<FlatMatrix<> > mat1d(ir.GetNIPY());
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  mat1d[iy].AssignMemory (ndof1d, ndof1d, lh);

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  mat2d[iz].AssignMemory (ndof2d, ndof2d, lh);

	FlatMatrix<> mat1dy (order+1, lh);

	FlatVector<> coefy(ir.GetNIPY(), lh);
	FlatVector<> coefyb1(ir.GetNIPY(), lh);


	FlatArray<int> quad2trig((order+1)*(order+1), lh);
	quad2trig = -1;
	quad2trig[0] = 0;
	quad2trig[order+1] = 0;
	quad2trig[1] = 1;
	quad2trig[order+2] = 2;
	
	int pos = 3;
	for (int ky = 2; ky <= order; ky++)
	  {
	    quad2trig[ky] = pos++;
	    quad2trig[ky+order+1] = pos++;
	  }
	for (int ky = 0; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    quad2trig[kx*(order+1)+ky] = pos++;

	FlatVector<> coefx(ir.GetNIPX(), lh);
	FlatVector<> hvx(order+1, lh);

	NgProfiler::StopTimer (preparetimer);

	diag = 0.0;

	for (int diff1 = 0; diff1 < 3; diff1++)
	  for (int diff2 = 0; diff2 < 3; diff2++)
	    {
	      // (*testout) << "diff1 = " << diff1 << ", diff2 = " << diff2 << endl;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  NgProfiler::StartTimer (blockxtimer);


		  mat2d[iz] = 0.0;

		  // precompute 1D matrices mat1d[iy]
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		  
		      int first_dmati = iz*ngrid2d+iy*ir.GetNIPX();

		      FlatMatrix<> base1( (diff1 == 0) ? dlegxbub : legxbub );
		      FlatMatrix<> base2( (diff2 == 0) ? dlegxbub : legxbub );
		      
		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			coefx(ix) = dmats[first_dmati+ix](diff1, diff2);

		      /*
		      (*testout) << "coefx = " << endl;
		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			(*testout) << coefx(ix) / ir1d[ix].Weight() << endl;
		      */
		      /*
		      hvx = Trans (base1) * coefx;
		      int maxc = 0;
		      for (int i = 0; i < hvx.Size(); i++)
			if (fabs (hvx(i)) > 1e-10) maxc = i;
		      cout << "maxc = " << maxc << endl;
		      (*testout) << "b1x = " << endl << Trans (base1) * coefx << endl;
		      (*testout) << "b2x = " << endl << Trans (base1) * coefx << endl;
		      */


		      for (int i = 0; i < order+1; i++)
			for (int j = 0; j < order+1; j++)
			  {
			    double sum = 0;
			    for (int ix = 0; ix < ir.GetNIPX(); ix++)
			      sum += base1(ix, i) * coefx(ix) * base2(ix, j);
			    mat1d[iy](i,j) = sum;
			  }


		      /*
		      for (int i = 0; i < order+1; i++)
			for (int j = 0; j < order+1; j++)
			  if (fabs (mat1d[iy](i,j)) < 1e-10)
			    mat1d[iy](i,j) = 0.0;
		      // (*testout) << "diff1 = " << diff1 << ", diff2 = " << diff2 << "\n";
		      // (*testout) << "mat1d = " << mat1d[iy] << "\n";
		      */
		    }


		  NgProfiler::StopTimer (blockxtimer);
		  NgProfiler::StartTimer (blockytimer);

		  
		  for (int kx1 = 0; kx1 <= order; kx1++)
		    for (int kx2 = 0; kx2 <= order; kx2++)
		      if (kx1 == kx2 || kx1 <= 1 && kx2 <= 1)
		      {
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  coefy(iy) = mat1d[iy](kx1,kx2);

			FlatMatrix<> base1( (diff1==1) ? djaczbTs[kx1] : jaczbTs[kx1]);
			FlatMatrix<> base2( (diff2==1) ? djaczbTs[kx2] : jaczbTs[kx2]);

			
			for (int i = 0; i < order+1; i++)
			  {
			    int mi = quad2trig[kx1*(order+1)+i];
			    if (mi != -1)
			      {
				for (int k = 0; k < ir.GetNIPY(); k++)
				  coefyb1(k) = base1(i,k) * coefy(k);
			  
				for (int j = 0; j < order+1; j++)
				  {
				    int mj = quad2trig[kx2*(order+1)+j];
				    if (mj != -1)
				      {
					double sum = 0;
					for (int k = 0; k < ir.GetNIPY(); k++)
					  sum += coefyb1(k) * base2(j,k);
					mat2d[iz](mi, mj) += sum;
				      }
				  }
			      }

			  }

			/*			
			if (kx1 == kx2)
			  {
			    for (int i = 0; i < order+1; i++)
			      for (int j = 0; j < order+1; j++)
				{
				  double sum = 0;
				  for (int ix = 0; ix < ir.GetNIPX(); ix++)
				    sum += base1(i, ix) * coefy(ix) * base2(j, ix);
				  mat1dy(i,j) = sum;
				}
			    
			
			    for (int i = 0; i < order+1; i++)
			      for (int j = 0; j < order+1; j++)
				if (fabs (mat1dy(i,j)) < 1e-12)
				  mat1dy(i,j) = 0.0;
			    
			    *testout << "diff1 = " << diff1 << ", diff2 = " << diff2 << ", kx1 = " << kx1 << ", kx2 = " << kx2 << endl;
			    *testout << "mat1dy = " << endl << mat1dy << endl;
			  }
			*/
		      }

		  NgProfiler::StopTimer (blockytimer);
		}
      



	      // scaling with (1-z)^m
	
	      FlatArray<int> powerz(ndof2d, lh);
	      for (int k = 0; k < 3; k++)
		powerz[k] = 0;
	
	      pos = 3;
	      for (int k = 2; k <= order; k++)
		{
		  powerz[pos++] = k;
		  powerz[pos++] = k;
		}

	      for (int ky = 0; ky <= order; ky++)
		for (int kx = 2; kx <= order-ky; kx++)
		  powerz[pos++] = kx+ky;


	      NgProfiler::StartTimer (blockztimer);

	

	      // transform prism ---> tet

	      ARRAY<int> prism2tet((order+1)*ndof2d);
	      fel.GetPrismTetReordering (isort, prism2tet);
	

	      // build 3D - tensor product matrix:

	      FlatVector<> coefz(ir.GetNIPZ(), lh);
	      FlatVector<> coefzb1(ir.GetNIPZ(), lh);

	      for (int kxy = 0; kxy < ndof2d; kxy++)
		for (int kxy2 = 0; kxy2 < ndof2d; kxy2++)
		  if (kxy == kxy2 || powerz[kxy] == 0 && powerz[kxy2] == 0)
		  {
		    // get coefficients in z-direction
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      coefz(iz) = mat2d[iz](kxy,kxy2);

		    FlatMatrix<> base1( (diff1==2) ? 
					djaczbTs[powerz[kxy]] : 
					jaczbTs[powerz[kxy]]);
		    FlatMatrix<> base2( (diff2==2) ? 
					djaczbTs[powerz[kxy2]] : 
					jaczbTs[powerz[kxy2]]);

		    for (int i = 0; i < order+1 - powerz[kxy]; i++)
		      {
			int mi = prism2tet[kxy*(order+1)+i];
			if (mi != -1)
			  {
			    for (int k = 0; k < ir.GetNIPZ(); k++)
			      coefzb1(k) = base1(i,k) * coefz(k);

			    for (int j = 0; j < order+1 - powerz[kxy2]; j++)
			      {
				int mj = prism2tet[kxy2*(order+1)+j];
				if (mi == mj)
				  {
				    double sum = 0;
				    for (int k = 0; k < ir.GetNIPZ(); k++)
				      sum += coefzb1(k) * base2(j,k);
				    diag(mi) += sum;
				  }
			      }
			  }
		      }
		  }
	      NgProfiler::StopTimer (blockztimer);
	    }
	
	NgProfiler::StopTimer (maintimer);

	
	lh.CleanUp(heapp);

	if (checkfast)
	  {
	    double sum = 0;
	    for (int i = 0; i < diag1.Height(); i++)
	      sum += sqr (diag1(i)-diag(i));
	    cout << "diff between std and fast: " << sqrt (sum) << endl;
	    (*testout) << "old mat " << endl << diag1 << endl;
	    (*testout) << "new mat " << endl << diag << endl;
	    (*testout) << "diff mat " << endl << diag-diag1 << endl;
	  }
      }

    else
      {
	orig -> AssembleElementMatrixDiag (bfel, eltrans, diag, lh); 	
	return;
      }
  }




  template <> 
  void FastLaplaceIntegrator<2> ::
  AssembleElementMatrixDiag (const FiniteElement & bfel,
			     const ElementTransformation & eltrans, 
			     FlatVector<double> & diag,
			     LocalHeap & lh) const
  {
    orig -> AssembleElementMatrixDiag (bfel, eltrans, diag, lh);
  }




  template <>
  void FastLaplaceIntegrator<2> ::
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 HashTable<INT<2>, double> & elmat,
			 LocalHeap & lh) const
  {
    cout << "fastlap<2>::assemble hash not implemented" << endl;
  }





  template <>
  void FastLaplaceIntegrator<2> ::
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    FlatMatrix<> mat1;
    if (checkfast)
      {
	orig -> AssembleElementMatrix (bfel, eltrans, mat1, lh); 
	// (*testout) << "original: " << mat1 << endl;
      }


    if (dynamic_cast<const H1HighOrderTrig<>*> (&bfel))

      {
	// ****************************  Triangles *******************************

	const H1HighOrderTrig<> & fel = static_cast<const H1HighOrderTrig<>&> (bfel);
	int ndof = fel.GetNDof();
        
	elmat.AssignMemory (ndof, ndof, lh);
	elmat = 0;
	
	void * heapp = lh.GetPointer();

	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*fel.Order()+20);
	
	TensorProductIntegrationRule ir (fel, ir1d);
	
	FlatVector<Mat<2> > dmats(ir.GetNIP(), lh);
	
	void * heapp2 = lh.GetPointer();
	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    SpecificIntegrationPoint<2,2> sip(ir[i], eltrans, lh);
		
	    Mat<2> dmat, tquadtrig, tquadtrig_inv, jac_inv, trans;

	    double val = coef -> Evaluate (sip);
	    val *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
	    dmat = val * Id<2>();

	    jac_inv = sip.GetJacobianInverse();

	    ir.GetIP_DTrig_DQuad(i, tquadtrig);
	    CalcInverse (tquadtrig, tquadtrig_inv);

	    trans = Trans (jac_inv) * Trans (tquadtrig_inv);
	    dmats[i] = Trans (trans) * dmat * trans;

	    // (*testout) << ", dmats = " << dmats[i] << endl;
	    // if ( (i % ir1d.GetNIP()) == ir1d.GetNIP()-1) (*testout) << endl << endl;

	    lh.CleanUp (heapp2);
	  }


	int order = fel.Order();
	
	FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
	FlatMatrix<> dlegxbub(ir.GetNIPX(), order+1, lh);
	FlatArray<AutoDiff<1> > hvals (order+1, lh);
	FlatArray<AutoDiff<1> > hvals2 (order+1, lh);

	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    AutoDiff<1> x (ir.GetIPX(ix), 0);

	    AutoDiff<1> * hp = &hvals[2];
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);
	    
	    hvals[0] = x;
	    hvals[1] = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      hvals[j+2] *= x * (1-x);

	    /*
	    AutoDiff<1> * hp = &hvals[0];
	    IntegratedLegendrePolynomial (order, 2*x-1, hp);
	    hvals[0] = x;
	    hvals[1] = 1-x;
	    */

	    for (int j = 0; j <= order; j++)
	      {
		legxbub(ix, j) = hvals[j].Value();
		dlegxbub(ix, j) = hvals[j].DValue(0);
	      }
	  }

	ARRAY<FlatMatrix<> > jaczbTs (order+1);
	ARRAY<FlatMatrix<> > djaczbTs (order+1);

	for (int k = 0; k <= order; k++)
	  {
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);
	    djaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), lh);

	    if (k < 2)
	      {
		jaczbTs[k] = Trans (legxbub);
		djaczbTs[k] = Trans (dlegxbub);
	      }
	    else
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  AutoDiff<1> z (ir.GetIPZ(iz), 0);
		  AutoDiff<1> powzk = 1.0;
		  for (int l = 0; l < k; l++)
		    powzk *= (1-z);


		  AutoDiff<1> * hp = &hvals[1];
		  JacobiPolynomial (order-1, 2*z-1, 2*k-3, 1, hp);
		  for (int j = 1; j <= order; j++)
		    hvals[j] *= z * powzk;
		  hvals[0] = powzk;


		  /*
		  AutoDiff<1> * hp = &hvals[0];
		  JacobiPolynomial (order, 2*z-1, 2*k-2, -1, hp);
		  JacobiPolynomial (order, AutoDiff<1> (-1.0), 2*k-2, -1, hvals2);

		  for (int j = 0; j <= order; j++)
		    {
		      hvals[j] -= hvals2[j];
		      hvals[j] *= powzk;
		    }
		  hvals[0] = powzk;
		  */


		  for (int j = 0; j <= order; j++)
		    {
		      jaczbTs[k](j, iz) = hvals[j].Value();
		      djaczbTs[k](j, iz) = hvals[j].DValue(0);
		    }
		}
	  }
	


	int sort[3] = { 0, 1, 2 };
	for (int i = 0; i < 3; i++)
	  for (int j = 0; j < 2; j++)
	    if (fel.GetVNums()[sort[j]] > fel.GetVNums()[sort[j+1]]) swap (sort[j], sort[j+1]);
    
	int isort[3];
	for (int i = 0; i < 3; i++) isort[sort[i]] = i;





	// precompute 2D matrices

	int ndof1d = order+1;

	ARRAY<FlatMatrix<> > mat1d(ir.GetNIPY());
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  mat1d[iy].AssignMemory (ndof1d, ndof1d, lh);

	FlatMatrix<> mat1dy (order+1, lh);

	FlatVector<> coefy(ir.GetNIPY(), lh);
	FlatVector<> coefyb1(ir.GetNIPY(), lh);


	FlatArray<int> quad2trig((order+1)*(order+1), lh);
	quad2trig = -1;


	// vertex shapes
	for (int i = 0; i < 3; i++)
	  if (isort[i] == 0)
	    {  
	      quad2trig[0] = i;
	      quad2trig[order+1] = i;
	    }
	  else
	    { 
	      quad2trig[(isort[i]-1)*(order+1)+1] = i;
	    }
	
	/*	
	quad2trig[0] = 0;
	quad2trig[order+1] = 0;
	quad2trig[1] = 1;
	quad2trig[order+2] = 2;
	*/

	int ii = 3;
	const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
	for (int i = 0; i < 3; i++)
	  if (order >= 2)
	    {
	      int es = isort[edges[i][0]], ee = isort[edges[i][1]];	
	      if (ee < es) swap (ee, es);
	      
	      int p = order;
	      
	      switch (es)
		{
		case 0:  // edges 0-1, 0-2,
		  {
		    for (int kz = 0; kz <= p-2; kz++)
		      quad2trig[(ee-1)*(order+1)+kz+2] = ii++; 
		    break;
		  }
		  
		case 1:
		  {      // edges 1-2,
		    for (int kx = 0; kx <= p-2; kx++)
		      quad2trig[(kx+2)*(order+1)] = ii++;

		    break;
		  }
		}
	    }    
	
	/*
	int pos = 3;
	for (int ky = 2; ky <= order; ky++)
	  {
	    quad2trig[ky] = pos++;
	    quad2trig[ky+order+1] = pos++;
	  }
	for (int ky = 0; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    quad2trig[kx*(order+1)+ky] = pos++;
	*/

	for (int ky = 1; ky <= order; ky++)
	  for (int kx = 2; kx <= order-ky; kx++)
	    quad2trig[kx*(order+1)+ky] = ii++;


	FlatVector<> coefx(ir.GetNIPX(), lh);
	FlatVector<> hvx(order+1, lh);


	elmat = 0.0;

	for (int diff1 = 0; diff1 < 2; diff1++)
	  for (int diff2 = 0; diff2 < 2; diff2++)
	    {
	      // precompute 1D matrices mat1d[iy]
	      for (int iy = 0; iy < ir.GetNIPY(); iy++)
		{
		  int first_dmati = iy*ir.GetNIPX();
		  
		  FlatMatrix<> base1( (diff1 == 0) ? dlegxbub : legxbub );
		  FlatMatrix<> base2( (diff2 == 0) ? dlegxbub : legxbub );

		      
		  for (int ix = 0; ix < ir.GetNIPX(); ix++)
		    coefx(ix) = dmats[first_dmati+ix](diff1, diff2);
		  
		  for (int i = 0; i < order+1; i++)
		    for (int j = 0; j < order+1; j++)
		      {
			double sum = 0;
			for (int ix = 0; ix < ir.GetNIPX(); ix++)
			  sum += base1(ix, i) * coefx(ix) * base2(ix, j);
			mat1d[iy](i,j) = sum;
		      }

		  for (int i = 0; i < order+1; i++)
		    for (int j = 0; j < order+1; j++)
		      if (fabs (mat1d[iy](i,j)) < 1e-10)
			mat1d[iy](i,j) = 0.0;

		  /*
		  (*testout) << "coefx = " << endl;
		  for (int ix = 0; ix < ir.GetNIPX(); ix++)
		    (*testout) << coefx(ix) / ir1d[ix].Weight() << endl;
		  
		  (*testout) << "diff1 = " << diff1 << ", diff2 = " << diff2 << "\n";
		  (*testout) << "mat1d = " << mat1d[iy] << "\n";
		  */
		}
	      

		  
	      for (int kx1 = 0; kx1 <= order; kx1++)
		for (int kx2 = 0; kx2 <= order; kx2++)
		  {
		    for (int iy = 0; iy < ir.GetNIPY(); iy++)
		      coefy(iy) = mat1d[iy](kx1,kx2);
		    
		    FlatMatrix<> base1( (diff1==1) ? djaczbTs[kx1] : jaczbTs[kx1]);
		    FlatMatrix<> base2( (diff2==1) ? djaczbTs[kx2] : jaczbTs[kx2]);
		    
			
		    for (int i = 0; i < order+1; i++)
		      {
			int mi = quad2trig[kx1*(order+1)+i];
			if (mi != -1)
			  {
			    for (int k = 0; k < ir.GetNIPY(); k++)
			      coefyb1(k) = base1(i,k) * coefy(k);
			    
			    for (int j = 0; j < order+1; j++)
			      {
				int mj = quad2trig[kx2*(order+1)+j];
				if (mj != -1)
				  {
				    double sum = 0;
				    for (int k = 0; k < ir.GetNIPY(); k++)
				      sum += coefyb1(k) * base2(j,k);
				    elmat(mi, mj) += sum;
				  }
			      }
			  }
		      }

		    /*
		    if (fabs (double (kx1-kx2)) <= 2)
		      {
			for (int i = 0; i < order+1; i++)
			  for (int j = 0; j < order+1; j++)
			    {
			      double sum = 0;
			      for (int ix = 0; ix < ir.GetNIPX(); ix++)
				sum += base1(i, ix) * coefy(ix) * base2(j, ix);
				  mat1dy(i,j) = sum;
			    }
			
			
			for (int i = 0; i < order+1; i++)
			  for (int j = 0; j < order+1; j++)
			    if (fabs (mat1dy(i,j)) < 1e-12)
			      mat1dy(i,j) = 0.0;
			
			*testout << "diff1 = " << diff1 << ", diff2 = " << diff2 << ", kx1 = " << kx1 << ", kx2 = " << kx2 << endl;
			*testout << "mat1dy = " << endl << mat1dy << endl;
		      }
		    */
		  }
	    }
	
	
	/*
	int nz = 0;
	for (int hi = 0; hi < elmat.Height(); hi++)
	  for (int hj = 0; hj < elmat.Width(); hj++)
	    if (sqr (elmat(hi,hj)) <= 1e-20) nz++;
	cout << "laplace mat has " << double(nz) / (elmat.Height()*elmat.Width()) * 100 << " % zero entries" << endl;
	cout << "these are " << elmat.Width()-double(nz) / elmat.Height() << " nze elements per row" << endl;
	*/

	lh.CleanUp(heapp);

	if (checkfast)
	  {
	    double sum = 0;
	    for (int i = 0; i < mat1.Height()*mat1.Width(); i++)
	      sum += sqr (mat1(i)-elmat(i));
	    cout << "diff between std and fast: " << sqrt (sum) << endl;
	    (*testout) << "old mat " << endl << mat1 << endl;
	    (*testout) << "new mat " << endl << elmat << endl;
	    (*testout) << "diff mat " << endl << elmat-mat1 << endl;
	  }
      }

    else
      {
	orig -> AssembleElementMatrix (bfel, eltrans, elmat, lh); 	
	return;
      }
  }









  template <>
  void * FastLaplaceIntegrator<3> ::
  PrecomputeData (const FiniteElement & fel, 
		  const ElementTransformation & eltrans, 
		  LocalHeap & lh) const 
  {
    int order = 2*fel.Order() + 6;
    if (common_integration_order >= 0)
      order = common_integration_order;
    
    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, order);
    
    TensorProductIntegrationRule ir (fel, ir1d);
    
    if (const_coef)
      {
	Mat<DIM_DMAT> * dmats = new Mat<DIM_DMAT> [1];
	
	void * heapp = lh.GetPointer();
	for (int i = 0; i < 1; i++)
	  {
	    SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ir[i], eltrans, lh);
	    
	    Mat<DIM_DMAT> dmat;
	    Vec<DIM_DMAT> hv1, hv2;
	    
	    dmatop.GenerateMatrix (fel, sip, dmat, lh);
	    
	    for (int j = 0; j < DIM_DMAT; j++)
	      {
		hv1 = 0; 
		hv1(j) = 1;
		DIFFOP::Transform (sip, hv1);
		hv2 = dmat * hv1;
		DIFFOP::TransformT (sip, hv2);
		
		for (int k = 0; k < DIM_DMAT; k++)
	      dmats[i] (j,k) = hv2(k);
	      }
	    // (*testout) << "dmat = " << dmats[i] << endl;
	    dmats[i] *= fabs (sip.GetJacobiDet()); //  * sip.IP().Weight();
	    lh.CleanUp (heapp);
	  }    
    return dmats;
      }
    else
      {
	Mat<DIM_DMAT> * dmats = new Mat<DIM_DMAT> [ir.GetNIP()];
	
	void * heapp = lh.GetPointer();
	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ir[i], eltrans, lh);
	    
	    Mat<DIM_DMAT> dmat;
	    Vec<DIM_DMAT> hv1, hv2;
	    
	    dmatop.GenerateMatrix (fel, sip, dmat, lh);
	    
	    for (int j = 0; j < DIM_DMAT; j++)
	      {
		hv1 = 0; 
		hv1(j) = 1;
		DIFFOP::Transform (sip, hv1);
		hv2 = dmat * hv1;
		DIFFOP::TransformT (sip, hv2);
		
		for (int k = 0; k < DIM_DMAT; k++)
	      dmats[i] (j,k) = hv2(k);
	      }
	    // (*testout) << "dmat = " << dmats[i] << endl;
	    dmats[i] *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
	    lh.CleanUp (heapp);
	  }    
    return dmats;
      }

  }
  
  template <>
  void * FastLaplaceIntegrator<2> ::
  PrecomputeData (const FiniteElement & fel, 
		  const ElementTransformation & eltrans, 
		  LocalHeap & lh) const 
  { 
    return NULL; 
  }


  template <>
  void FastLaplaceIntegrator<3> ::
  ApplyElementMatrix (const FiniteElement & bfel, 
		      const ElementTransformation & eltrans, 
		      const FlatVector<double> & elx, 
		      FlatVector<double> & ely,
		      void * precomputed,
		      LocalHeap & lh) const
  {

    try
      {
	static int maintimer = NgProfiler::CreateTimer (string ("FastBDB::ApplyElementMatrix (fast)") + typeid(*this).name() );
	static int geomtimer = NgProfiler::CreateTimer (string ("FastBDB::ApplyElementMatrix, geometry evaluation") + typeid(*this).name() );
	static int trafotimer1 = NgProfiler::CreateTimer (string ("FastBDB::ApplyElementMatrix, apply grid") + typeid(*this).name() );
	static int trafotimer2 = NgProfiler::CreateTimer (string ("FastBDB::ApplyElementMatrix, apply grid trans") + typeid(*this).name() );

	NgProfiler::StartTimer (maintimer);


	const FEL & fel = dynamic_cast<const FEL&> (bfel);
	int ndof = fel.GetNDof();

	void * heapp = lh.GetPointer();

	if (fast)
	  {
	    int order = 2*fel.Order() + 2;

	    if (common_integration_order >= 0)
	      order = common_integration_order;

	    const IntegrationRule & ir1d =
	      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, order);
	    
	    TensorProductIntegrationRule ir (fel, ir1d);
	    
	    // ARRAY<Mat<DIM_DMAT> > dmats(ir.GetNIP());
	    ARRAY< Vec<DIM_DMAT> > hvgrid(ir.GetNIP());

	    NgProfiler::StartTimer (trafotimer1);

	    DIFFOP::ApplyGrid (fel, ir, elx, hvgrid, lh);

	    NgProfiler::StopTimer (trafotimer1);

	    NgProfiler::StartTimer (geomtimer);
	    
	    if (precomputed)
	      {
		// have dmats:
		if (!const_coef)
		  {
		    Mat<DIM_DMAT> * dmats = reinterpret_cast<Mat<DIM_DMAT>*> (precomputed);
		    Vec<DIM_DMAT> hv1;
		    for (int i = 0; i < ir.GetNIP(); i++)
		      {
			hv1 = hvgrid[i];
			hvgrid[i] = dmats[i] * hv1;
		      }
		  }
		else
		  {
		    Mat<DIM_DMAT> * dmats = reinterpret_cast<Mat<DIM_DMAT>*> (precomputed);
		    Vec<DIM_DMAT> hv1;
		    for (int i = 0; i < ir.GetNIP(); i++)
		      {
			hv1 = hvgrid[i];
			hvgrid[i] =  ir.GetWeight(i) * dmats[0] * hv1;
		      }
		  }

	      }
	    else
	      for (int i = 0; i < ir.GetNIP(); i++)
		{
		  SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ir[i], eltrans, lh);
		  
		  Vec<DIM_DMAT> hv1, hv2;
		  
		  hv1 = hvgrid[i];
		  DIFFOP::Transform (sip, hv1);
		  
		  dmatop.Apply (fel, sip, hv1, hv2, lh);
		  hv2 *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
		  
		  DIFFOP::TransformT (sip, hv2);		
		  hvgrid[i] = hv2;
		}

	    NgProfiler::StopTimer (geomtimer);

	    NgProfiler::StartTimer (trafotimer2);

	    DIFFOP::ApplyTransGrid (fel, ir, hvgrid, ely, lh);

	    NgProfiler::StopTimer (trafotimer2);


	    if (checkfast)
	      (*testout) << "fast, ely = " << endl << ely << endl;

	    lh.CleanUp (heapp);
	  }


	if (!fast || checkfast)
	  {
	    ely = 0;

	    Vec<DIM_DMAT,double> hv1;
	    Vec<DIM_DMAT,double> hv2;
	    
	    FlatVector<double> hely (ndof, lh);
	    
	    const IntegrationRule & ir = GetIntegrationRule (fel);

	    void * heapp = lh.GetPointer();
	    for (int i = 0; i < ir.GetNIP(); i++)
	      {
		SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip (ir[i], eltrans, lh);
		
		DIFFOP::Apply (fel, sip, elx, hv1, lh);
		dmatop.Apply (fel, sip, hv1, hv2, lh);
		DIFFOP::ApplyTrans (fel, sip, hv2, hely, lh);

		double fac = fabs (sip.GetJacobiDet()) * sip.IP().Weight();
		ely += fac * hely;
		lh.CleanUp (heapp);
	      }     

	    if (checkfast)
	      (*testout) << "std, ely = " << endl << ely << endl;
	  }
	NgProfiler::StopTimer (maintimer);
      }

    catch (Exception & e)
      {
	e.Append ("in AssembleElementMatrix, type = ");
	e.Append (typeid(*this).name());
	e.Append ("\n");
	throw;
      }
    catch (exception & e)
      {
	Exception e2(e.what());
	e2.Append ("\nin AssembleElementMatrix, type = ");
	e2.Append (typeid(*this).name());
	e2.Append ("\n");
	throw e2;
      }
  }


  template <>
  void FastLaplaceIntegrator<2> ::
  ApplyElementMatrix (const FiniteElement & bfel, 
		      const ElementTransformation & eltrans, 
		      const FlatVector<double> & elx, 
		      FlatVector<double> & ely,
		      void * precomputed,
		      LocalHeap & lh) const
  { ; }






  
  namespace
#ifdef MACOS
  fastlaplace_cpp
#endif
  {
    class Init
    { 
    public: 
      Init ();
    };        
    
    Init::Init()
    {
      GetIntegrators().AddBFIntegrator ("fastlaplace", 3, 1,
					FastLaplaceIntegrator<3>::Create);
      GetIntegrators().AddBFIntegrator ("fastlaplace", 2, 1,
					FastLaplaceIntegrator<2>::Create);
    }


    Init init;
    
  }
}


