#ifndef FILE_H1HOFE_FAST
#define FILE_H1HOFE_FAST

/*********************************************************************/
/* File:   h1hofast.hpp                                              */
/* Author: Start                                                     */
/* Date:   6. Jan. 2005                                              */
/*********************************************************************/




class H1HighOrderTrigFast : public H1HighOrderFiniteElement
{
  static Vector<> low_energy_vertex_shapes;
  static Vector<> low_energy_vertex_dshapes;
 
public:
  typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
  typedef TrigExtensionMin T_ORTHOPOL;

  H1HighOrderTrigFast (int aorder);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
  
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  virtual void EvaluateShapeGrid (const TensorProductIntegrationRule & ir,
				  const FlatVector<double> coefs,
				  FlatVector<double> gridvalues,
				  LocalHeap & lh) const;

  virtual void EvaluateShapeGridTrans (const TensorProductIntegrationRule & ir,
				       const FlatVector<double> gridvalues,
				       FlatVector<double> coefs,
				       LocalHeap & lh) const;


  virtual void EvaluateDShapeGrid (const TensorProductIntegrationRule & ir,
				   const FlatVector<double> coefs,
				   FlatMatrix<double> gridvalues,
				   LocalHeap & lh) const;

  virtual void EvaluateDShapeGridTrans (const TensorProductIntegrationRule & ir,
					const FlatMatrix<double> gridvalues,
					FlatVector<double> coefs,
					LocalHeap & lh) const;

  void GetQuadTrigReordering (FlatArray<int> & quad2trig) const;

private:
  
  template<typename Tx, typename Ty, typename TFA>  
  void T_CalcShape (Tx x, Ty y, TFA & shape) const; 

  
  static RecPol * polx;
  static ARRAY<RecPol*> polsy;

  static Matrix<> legxbub;
  static ARRAY<FlatMatrix<> > jacybs;

  static Matrix<> dlegxbub;
  static ARRAY<FlatMatrix<> > djacybs;
};





class H1HighOrderTetFast : public H1HighOrderFiniteElement
{
  // size ... NIP * 4 
  static Vector<> low_energy_vertex_shapes;
  static Vector<> low_energy_vertex_dshapes;

  /*
  static ARRAY<Matrix<>*> low_energy_edge_shapes;
  static ARRAY<Matrix<>*> low_energy_edge_shapes_des;
  static ARRAY<Matrix<>*> low_energy_edge_shapes_dee;
  */

public:
  typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
  typedef TrigExtensionMin T_ORTHOPOL;

  H1HighOrderTetFast (int aorder);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  virtual void EvaluateShapeGrid (const TensorProductIntegrationRule & ir,
				  const FlatVector<double> coefs,
				  FlatVector<double> gridvalues,
				  LocalHeap & lh) const;

  virtual void EvaluateShapeGridTrans (const TensorProductIntegrationRule & ir,
				       const FlatVector<double> gridvalues,
				       FlatVector<double> coefs,
				       LocalHeap & lh) const;

  virtual void EvaluateDShapeGrid (const TensorProductIntegrationRule & ir,
				   const FlatVector<double> coefs,
				   FlatMatrix<double> gridvalues,
				   LocalHeap & lh) const;

  virtual void EvaluateDShapeGridTrans (const TensorProductIntegrationRule & ir,
					const FlatMatrix<double> gridvalues,
					FlatVector<double> coefs,
					LocalHeap & lh) const;


  
  void GetPrismTetReordering (const int * isort,  ARRAY<int> & prism2tet) const;


private:
  
  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 
};


#endif
