/*********************************************************************/
/* File:   h1tet_fast.cpp                                            */
/* Author: Start                                                     */
/* Date:   30. Nov. 2004                                             */
/*********************************************************************/

 
#include <fem.hpp>
namespace ngfem
{
  using namespace ngfem;


  Vector<> H1HighOrderTetFast :: low_energy_vertex_shapes(0);
  Vector<> H1HighOrderTetFast :: low_energy_vertex_dshapes(0);

  
  H1HighOrderTetFast :: H1HighOrderTetFast (int aorder)
    : H1HighOrderFiniteElement(3, ET_TET)
  {
    order_inner = INT<3>(aorder,aorder,aorder);

    for (int i = 0; i < 6; i++)
      order_edge[i] = aorder;
    for (int i = 0; i < 4; i++)
      order_face[i] = INT<2>(aorder,aorder);
    for (int i = 0; i < 4; i++)
      order_vertex[i] = aorder;
    
    ComputeNDof();
  }


  void H1HighOrderTetFast :: ComputeNDof()
  {
    ndof = 4;
    if (augmented == 1) ndof *= 2;
    if (augmented == 2) ndof *= order_inner[0];

    for (int i = 0; i < 6; i++)
      ndof += order_edge[i] - 1;
    for (int i = 0; i < 4; i++)
      ndof += (order_face[i][0]-1)*(order_face[i][0]-2) / 2;

    ndof += (order_inner[0]-1)*(order_inner[0]-2)*(order_inner[0]-3) / 6;

    order = 1;
    for (int i = 0; i < 6; i++)
      if (order_edge[i] > order)
	order = order_edge[i];
    for (int i = 0; i < 4; i++)
      if (order_face[i][0] > order)
      order = order_face[i][0];
      
    if (order_inner[0] > order)
      order = order_inner[0];




    if (augmented == 1 && Integrator::GetCommonIntegrationOrder() != -1 &&
	low_energy_vertex_shapes.Size() == 0)
      {
	cout << "!!!!!!!!!  precompute low energy vertex and edge shapes" << endl;
	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, Integrator::GetCommonIntegrationOrder());
	cout << "common order = " << Integrator::GetCommonIntegrationOrder() << endl;


	int dummy_vnums[] = { 0, 1, 2, 3 };
	TensorProductIntegrationRule ir (ET_TET, dummy_vnums, ir1d);

	low_energy_vertex_shapes.SetSize (4*ir.GetNIP());
	low_energy_vertex_dshapes.SetSize (4*ir.GetNIP());

	for (int i = 0; i < ir.GetNIP(); i++)
	  {
	    IntegrationPoint ip = ir[i];

	    double lami[4] = { ip(0), ip(1), ip(2), 1-ip(0)-ip(1)-ip(2) };

	    for (int j = 0; j < 4; j++)
	      {
		AutoDiff<1> adlam (lami[j],0);

		AutoDiff<1> shape = T_VERTEXSHAPES::Calc (order_vertex[j], adlam);

		low_energy_vertex_shapes(4*i+j) = shape.Value();
		low_energy_vertex_dshapes(4*i+j) = shape.DValue(0);
	      }
	  }
      }
  }


    
  void H1HighOrderTetFast :: 
  GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);

    int base = 4;
    if (augmented == 1) base *= 2;
    if (augmented == 2) base *= order_inner[0];

    for (int i = 0; i < 6; i++)
      base += order_edge[i] - 1;
    for (int i = 0; i < 4; i++)
      base += (order_face[i][0]-1)*(order_face[i][0]-2) / 2;
     

    if(order_inner[0] > 2)
      {
	int ni = (order_inner[0]-1)*(order_inner[0]-2)*(order_inner[0]-3) / 6;
	for (int i = 0; i < ni; i++)
	  idofs.Append (base+i);
      }
  }

  



  void H1HighOrderTetFast :: CalcShape (const IntegrationPoint & ip, 
						FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderTetFast :: CalcDShape (const IntegrationPoint & ip, 
						 FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  }









#ifdef JACOBIxxx

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderTetFast :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx lami[4] = { x, y, z, 1-x-y-z };

    Tx lamisum[4] = { 0, 0, 0, 0 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
	if (vnums[j] > vnums[i]) lamisum[i] += lami[j];

    Tx lamiadd[4];
    for (int i = 0; i < 4; i++)
      lamiadd[i] = lami[i] + lamisum[i];

    ArrayMem<Tx, 20> polx(order+1), poly(order+1), polz(order+1); 
    ArrayMem<Tx, 20> power(order+1);

    // vertex shapes
    for (int i = 0; i < 4; i++)
      sds[i] = lami[i];
    int ii = 4; 

    if (augmented == 1)
      for (int i = 0; i < 4; i++)
	sds[ii++] = T_VERTEXSHAPES::Calc (order_vertex[i], lami[i]);
    else
      if (augmented == 2)
	for (int i = 0; i < 4; i++)
	  {
	    ArrayMem<Tx,20> pol(order+1);
	    LegendrePolynomial (order-1, -1+2*lami[i], pol);	
	    for (int j = 1; j < order; j++)
	      sds[ii++] = pol[j] * lami[i];
	  }



    // fast edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];

	  int nde = order_edge[i]-1;
	  ScaledJacobiPolynomial (order_edge[i]-2, lami[es]-lamisum[es], lamiadd[es], 1, 1, hp);
	  for (int j = 0; j < nde; j++)
	    hp[j] *= lami[ee]*lami[es];

	  ii += nde;
	}

    /*
    // low energy edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];
	  ii += T_ORTHOPOL::CalcTrigExt (order_edge[i], lami[ee]-lami[es], 
					 1-lami[es]-lami[ee], hp);

	}
    */




    // face dofs
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
     if (order_face[i][0] >= 3)
       {
	  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
	  
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
	  
	  int vop = 6 - fav[0] - fav[1] - fav[2];  	
	  
	  int n = order_face[i][0];
	  ArrayMem<Tx, 20> polx(n+1);
	  ArrayMem<Tx, 50> polsy((n+1) * (n+1));

	  ScaledJacobiPolynomial (n-3, lami[fav[1]]-lamisum[fav[1]], lamiadd[fav[1]], 1, 1, polx);
	  for (int j = 0; j <= n; j++)
	    {
	      Tx * hp = &polsy[j * (n+1)];
	      ScaledJacobiPolynomial (n, lami[fav[0]]-lamisum[fav[0]], lamiadd[fav[0]], 2*j-3, 1, hp);
	    }

	  Tx  bub = lami[fav[0]]*lami[fav[1]]*lami[fav[2]]; 
	  for (int j2 = 0; j2 <= n-3; j2++)
	    for (int i2 = 0; i2 <= n-3-j2; i2++)
	      sds[ii++] = bub * polx[i2] * polsy[(i2+2) * (n+1) + j2];
	}



    if (order_inner[0] >= 4)
      {
	Tx * psds = &sds[ii];
	int sort[4] = { 0, 1, 2, 3 };
	for (int i = 0; i < 4; i++)
	  for (int j = 0; j < 3; j++)
	    if (vnums[sort[j]] < vnums[sort[j+1]])
	      swap (sort[j], sort[j+1]);



	Tx lx = lami[sort[1]]-lami[sort[0]];
	Tx ly = lami[sort[2]];
	Tx lz = lami[sort[3]];

	int n = order_inner[0];

	ArrayMem<Tx, 20> polx(n+1); // , poly(n+1), polz(n+1);
	ArrayMem<Tx, 50> polsy((n+1) * (n+1));
	ArrayMem<Tx, 50> polsz((n+1) * (n+1));

	ScaledJacobiPolynomial (n-4, lx, 1-ly-lz, 1, 1, polx);
	// ScaledJacobiPolynomial (n-4, 2*ly - (1-lz) , (1-lz), n, 0, poly);
	//	JacobiPolynomial (n-4, 2*lz-1, n, 0, polz);

	for (int j = 0; j <= n; j++)
	  {
	    Tx * hp = &polsy[j * (n+1)];
	    ScaledJacobiPolynomial (n-4, 2*ly-(1-lz), (1-lz), 2*j-3 , 1, hp);
	  }
	for (int j = 0; j <= n; j++)
	  {
	    Tx * hp = &polsz[j * (n+1)];
	    JacobiPolynomial (n-4, 2*lz-1, 2*j-3, 1, hp);
	  }


	/*
	  ScaledJacobiPolynomial (n-4, lx, 1-ly-lz, 2, 2, polx);
	  ScaledJacobiPolynomial (n-4, 2*ly - (1-lz) , (1-lz), n, 0, poly);
	  JacobiPolynomial (n-4, 2*lz-1, n, 0, polz);
	*/

	Tx bub = lami[0]*lami[1]*lami[2]*lami[3];
	for (int i = 0; i <= n-4; i++)
	  for (int j = 0; j <= n-4-i; j++)
	    for (int k = 0; k <= n-4-i-j; k++)
	      sds[ii++] = bub * polx[k] * polsy[(k+2)*(n+1)+j] * polsz[(j+k+3)*(n+1)+i];

	/*
	  JS, Nov 24:
	  changed orientation of inner shapes (dep. on vertex numbers)
	  changed ordering in T_INNERSHAPES::Calc: inner loop is x-loop
	  swap x-direction
	*/
      }
  }
#endif


  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderTetFast :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx lami[4] = { x, y, z, 1-x-y-z };

    int sort[] = { 0, 1, 2, 3 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);

    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] < vnums[sort[3]]

    /*
      Tx lamisum[4] = { 0.0, 0.0, 0.0, 0.0 };
      for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
      if (vnums[j] > vnums[i]) lamisum[i] += lami[j];
    */
    Tx lamisum[4], sum = 0.0;
    for (int j = 3; j >= 0; j--)
      {
        lamisum[sort[j]] = sum;
        sum += lami[sort[j]];
      }

    ArrayMem<Tx, 80> pollam( 4 * (order+1) );
    Tx * poli[4];
    for (int i = 0; i < 4; i++) poli[i] = &pollam[i*(order+1)];
    
    // including lam_i - factor
    LegendrePolynomialMult (order-2, lami[sort[0]]-lamisum[sort[0]], lami[sort[0]], poli[sort[0]]);
    ScaledLegendrePolynomialMult (order-2, lami[sort[1]]-lamisum[sort[1]], lamisum[sort[0]], lami[sort[1]], poli[sort[1]]);
    ScaledLegendrePolynomialMult (order-2, lami[sort[2]]-lamisum[sort[2]], lamisum[sort[1]], lami[sort[2]], poli[sort[2]]);
    poli[sort[3]][0] = lami[sort[3]];
    

    // vertex shapes
    for (int i = 0; i < 4; i++)
      sds[i] = poli[i][0];   // lami[i]
    int ii = 4; 

    if (augmented == 1)
      for (int i = 0; i < 4; i++)
	sds[ii++] = T_VERTEXSHAPES::Calc (order_vertex[i], lami[i]);
    else
      if (augmented == 2)
	for (int i = 0; i < 4; i++)
	  {
	    ArrayMem<Tx,20> pol(order+1);
	    LegendrePolynomial (order-1, -1+2*lami[i], pol);	
	    for (int j = 1; j < order; j++)
              sds[ii++] = pol[j] * lami[i];
          }

    //  edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
          int es = edges[i][0], ee = edges[i][1];
          if (vnums[es] > vnums[ee]) swap (es, ee);
          
	  int nde = order_edge[i]-1;
          
          Tx * pol = poli[es];
          Tx bub = poli[ee][0];
	  for (int j = 0; j < nde; j++, ii++)
	    sds[ii] = bub * pol[j];
	}

    // face dofs
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
     if (order_face[i][0] >= 3)
       {
	  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
	  
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	

	  int n = order_face[i][0];

          Tx * polx = poli[fav[1]];
          Tx * poly = poli[fav[0]];
          Tx  bub = poli[fav[2]][0]; 

	  for (int j2 = 0; j2 <= n-3; j2++)
            {
              Tx hv = bub * poly[j2];
              for (int i2 = 0; i2 <= n-3-j2; i2++, ii++)
                sds[ii] = hv * polx[i2];
            }
	}

    if (order_inner[0] >= 4)
      {
	int n = order_inner[0];

        Tx * polx = poli[sort[0]];
        Tx * poly = poli[sort[1]];
        Tx * polz = poli[sort[2]];
        Tx bub = poli[sort[3]][0];

	for (int i = 0; i <= n-4; i++)
          {
            Tx hv1 = bub * polz[i];
            for (int j = 0; j <= n-4-i; j++)
              {
                Tx hv2 = hv1 * poly[j];
                for (int k = 0; k <= n-4-i-j; k++, ii++)
                  sds[ii] = hv2 * polx[k];
              }
          }
      }
  }










  void H1HighOrderTetFast ::
  GetPrismTetReordering (const int * isort,  ARRAY<int> & prism2tet) const
  {
    int ndof2d = (order+1)*(order+2)/2;
    prism2tet.SetSize((order+1)*ndof2d);
    
    prism2tet = -1;	

    // vertex shapes
    for (int i = 0; i < 4; i++)
      if (isort[i] == 0)
	{  
	  prism2tet[0] = i;
	  prism2tet[order+1] = i;
	  prism2tet[2*order+2] = i;
	}
      else
	{ 
	  prism2tet[(isort[i]-1)*(order+1)+1] = i;
	}
	

    int ii = 4;
    if (augmented == 1) ii = 8;
    if (augmented == 2) ii = 4*order;


    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order >= 2)
	{
	  int es = isort[edges[i][0]], ee = isort[edges[i][1]];	
	  if (ee < es) swap (ee, es);

	  int p = order;

	  switch (es)
	    {
	    case 0:  // edges 0-1, 0-2, 0-3
	      {
		for (int kz = 0; kz <= p-2; kz++)
		  prism2tet[(ee-1)*(order+1)+kz+2] = ii++; 
		break;
	      }

	    case 1:
	      {      // edges 1-2, 1-3
		for (int ky = 0; ky <= p-2; ky++)
		  prism2tet[(ee+1+2*ky)*(order+1)] = ii++;
		break;
	      }

	    case 2:
	      {   // edge 2-3
		int pos = 2*p+1;
		for (int kx = 0; kx <= p-2; kx++)
		  prism2tet[(pos++)*(order+1)] = ii++;
		break;
	      }
	    }
	}    
    

    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
      if (order >= 3)
	{
	  int fav[] = { isort[faces[i][0]], isort[faces[i][1]], isort[faces[i][2]] };

	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 
	  if(fav[1] > fav[2]) swap(fav[1],fav[2]);
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 	

	  int p = order;

	  switch (6-fav[0]-fav[1]-fav[2])
	    {
	    case 2:  // face (0,1,3)
	    case 3:  // face (0,1,2)
	      {
		for (int kz = 0; kz <= p-3; kz++)
		  for (int ky = 0; ky <= p-3-kz; ky++, ii++)
		    prism2tet[(fav[2]+1+2*ky)*(order+1)+kz+1] = ii;
		break;
	      }

	    case 1:  // face (0,2,3)
	      {
		for (int ky = 0; ky <= p-3; ky++)
		  {
		    int pos = 2*p+1;
		    for (int kx = 0; kx <= p-3-ky; kx++, ii++)
		      prism2tet[(pos++)*(order+1)+ky+1] = ii;
		  }
		break;
	      }


	    case 0:  // face (1,2,3)
	      {
		int pos = 3*p;
		for (int ky = 0; ky <= p-3; ky++)
		  for (int kx = 0; kx <= p-3-ky; kx++, ii++)
		    prism2tet[(pos++)*(order+1)] = ii;
		break;
	      }
	    default:
	      break;
	    }
	}

    // inner shapes
    int p = order;
    for (int kz = 0; kz <= p-4; kz++)
      {
	int pos = 3*order;
	for (int ky = 0; ky <= p-4-kz; ky++)
	  {
	    for (int kx = 0; kx <= p-4-ky-kz; kx++, ii++)
	      prism2tet[(pos+kx)*(order+1)+kz+1] = ii;
	    pos += p-2-ky;
	  }
      }	 
  }






  void H1HighOrderTetFast ::
  EvaluateShapeGrid (const TensorProductIntegrationRule & ir,
		     const FlatVector<double> coefs,
		     FlatVector<double> gridvalues,
		     LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();

    FlatMatrix<double> gridyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    FlatMatrix<double> gridz_polxy(ir.GetNIPZ(), (order+1)*(order+2)/2+1, lh);



    static int maintimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateShapeGrid");
    static int preparetimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateShapeGrid, prepare");
    static int transformtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateShapeGrid, transform");

    static int transformxtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateShapeGrid, transformx");
    static int transformytimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateShapeGrid, transformy");
    static int transformztimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateShapeGrid, transformz");


    NgProfiler::StartTimer (maintimer);


    static bool init = 0;
    
    static Matrix<> legxbub(ir.GetNIPX(), order+1);
    static Matrix<> legybub(ir.GetNIPY(), order+1);
    static Matrix<> legzbub(ir.GetNIPZ(), order+1);
    static Matrix<> powzm(ir.GetNIPZ(), order+1);
    static Matrix<> powym(ir.GetNIPZ(), order+1);
    
    static ARRAY<FlatMatrix<> > jaczbTs (order+1);
    static ARRAY<FlatMatrix<> > jacybTs (order+1);

    if (!init)
      {
	init = 1;


	NgProfiler::StartTimer (preparetimer);

	//FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    double x = ir.GetIPX(ix);
	    double * hp = &legxbub(ix, 2);
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);
	    
	    legxbub(ix, 0) = x;
	    legxbub(ix, 1) = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      legxbub(ix, j+2) *= x * (1-x);
	  }
	
	// FlatMatrix<> powzm(ir.GetNIPZ(), order+1, lh);
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    double fac = 1-ir.GetIPZ(iz);
	    double val = 1;
	    for (int j = 0; j <= order; j++, val *= fac)
	      powzm(iz, j) = val;
	  }
	
	// ARRAY<FlatMatrix<> > jaczbTs (order+1);
	FlatMatrix<> jaczb(ir.GetNIPZ(), order+1, lh);
	
	for (int k = 0; k <= order; k++)
	  {
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      {
		double * hp = &jaczb(iz, 1);
		JacobiPolynomial (order-1, 2*ir.GetIPZ(iz)-1, 2*k-3, 1, hp);
		for (int j = 1; j <= order; j++)
		  jaczb(iz, j) *= ir.GetIPZ(iz) * powzm(iz, k);
		jaczb(iz, 0) = powzm(iz, k);
	      }
	    
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), new double[(order+1)*ir.GetNIPZ()]);
	    jaczbTs[k] = Trans (jaczb);
	  }
	jaczbTs[0] = Trans (legxbub);
	jaczbTs[1] = Trans (legxbub);
	NgProfiler::StopTimer (preparetimer);
      }



    NgProfiler::StartTimer (transformtimer);


    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++) isort[sort[i]] = i;



    

    int ndof2d = (order+1)*(order+2)/2;

    
    // scaling with (1-z)^m
    
    FlatArray<int> powerz(ndof2d, lh);
    for (int k = 0; k < 3; k++)
      powerz[k] = 0;
    
    int pos = 3;
    for (int k = 2; k <= order; k++)
      {
	powerz[pos++] = k;
	powerz[pos++] = k;
      }
    
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	powerz[pos++] = kx+ky;
    


	

    // transform prism ---> tet
    
    ARRAY<int> prism2tet;
    GetPrismTetReordering (isort, prism2tet);
    


    // transform quad ---> trig
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    quad2trig = -1;
    quad2trig[0] = 0;
    quad2trig[order+1] = 0;
    quad2trig[1] = 1;
    quad2trig[order+2] = 2;
	
    pos = 3;
    for (int ky = 2; ky <= order; ky++)
      {
	quad2trig[ky] = pos++;
	quad2trig[ky+order+1] = pos++;
      }
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	quad2trig[kx*(order+1)+ky] = pos++;
    


    // transform z
    


    NgProfiler::StartTimer (transformztimer);    
    gridz_polxy = 0;
    /*
    for (int kxy = 0; kxy < ndof2d; kxy++)
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
    for (int kz = 0; kz <= order-powerz[kxy]; kz++)
    {
    int teti = prism2tet[kxy*(order+1)+kz];
    gridz_polxy(iz, kxy) += jaczbTs[powerz[kxy]] (kz, iz) * coefs(teti);
    }
    */
    for (int kxy = 0; kxy < ndof2d; kxy++)
      {
	FlatMatrix<> pol = jaczbTs[powerz[kxy]];
	for (int kz = 0; kz <= order-powerz[kxy]; kz++)
	  gridz_polxy.Col(kxy) += 
	    coefs (prism2tet[kxy*(order+1)+kz]) * pol.Row(kz);
      }


    NgProfiler::StopTimer (transformztimer);

    // transform y

    NgProfiler::StartTimer (transformytimer);
    /*
    gridyz_polx = 0.0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	{
	  for (int ky = 0; ky <= order; ky++)
	    for (int kx = 0; kx <= order; kx++)
	      if (quad2trig[kx*(order+1)+ky] != -1)
		gridyz_polx(ipyz, kx) += jaczbTs[kx](ky, iy) * gridz_polxy(iz, quad2trig[kx*(order+1)+ky]);
	}
    */
    /*
    gridyz_polx = 0.0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++, ipyz+=ir.GetNIPY() )
      for (int ky = 0; ky <= order; ky++)
	for (int kx = 0; kx <= order; kx++)
	  {
	    int q2t = quad2trig[kx*(order+1)+ky];
	    if (q2t != -1)
	      {
		gridyz_polx.Col(kx).Range (ipyz, ipyz+ir.GetNIPY())
		  += gridz_polxy(iz, q2t) * jaczbTs[kx].Row(ky);

		NgProfiler::AddFlops (transformytimer, ir.GetNIPY());
	      }
	  }
    */

    /*
    FlatVector<double> hvx(ir.GetNIPY(), lh);
    for (int kx = 0; kx <= order; kx++)
      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	{
	  hvx = 0.0;
	  FlatMatrix<> pol = jaczbTs[kx];
	  for (int ky = 0; ky <= order; ky++)
	    {
	      int q2t = quad2trig[kx*(order+1)+ky];
	      if (q2t != -1)
		{
		  hvx += gridz_polxy(iz, q2t) * pol.Row(ky);
		  NgProfiler::AddFlops (transformytimer, ir.GetNIPY());
		}
	    }
	  gridyz_polx.Col(kx).Range (iz*ir.GetNIPY(), (iz+1)*ir.GetNIPY()) = hvx;
	}
    */

    /*
    for (int kx = 0; kx <= order; kx++)
      {
	hvyz = 0;
	FlatMatrix<> pol = jaczbTs[kx];

	for (int ky = 0; ky <= order; ky++)
	  {
	    int q2t = quad2trig[kx*(order+1)+ky];
	    if (q2t != -1)
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    hvyz.Range (iz*ir.GetNIPY(), (iz+1)*ir.GetNIPY()) += gridz_polxy(iz, q2t) * pol.Row(ky);
		    NgProfiler::AddFlops (transformytimer, ir.GetNIPY());
		  }
	      }
	  }
	gridyz_polx.Col(kx) = hvyz;
      }
    */


    FlatVector<double> hvyz(ir.GetNIPY()*ir.GetNIPZ(), lh);
    FlatMatrix<double> hv_poly_z (ir.GetNIPZ(), order+1, lh);
    FlatMatrix<double> hm(ir.GetNIPY(), order+1, lh);

    for (int kx = 0; kx <= order; kx++)
      {
	FlatMatrix<> pol = jaczbTs[kx];
	int maxy = (kx <= 1) ? order : order-kx;

	FlatMatrix<double> hv_poly_z2 (ir.GetNIPZ(), maxy+1, &hv_poly_z(0,0));
	FlatMatrix<double> hv_yz2 (ir.GetNIPZ(), ir.GetNIPY(), &hvyz(0));
	// FlatMatrix<> pol2 (maxy+1, ir.GetNIPY(), &pol(0,0));
	FlatMatrix<> pol2 (ir.GetNIPY(), maxy+1, &hm(0,0));
	pol2 = Trans (pol);

	for (int ky = 0; ky <= maxy; ky++)
	  hv_poly_z2.Col(ky) = gridz_polxy.Col(quad2trig[kx*(order+1)+ky]);

	hv_yz2 = hv_poly_z2 * Trans (pol2);
	NgProfiler::AddFlops (transformytimer, (maxy+1)*ir.GetNIPY()*ir.GetNIPZ());

	gridyz_polx.Col(kx) = hvyz;
      }


    /*
    FlatVector<double> hvyz(ir.GetNIPY(), lh);
    FlatVector<double> hv_poly_z (order+1, lh);
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int kx = 0; kx <= order; kx++)
	{
	  FlatMatrix<> pol = jaczbTs[kx];
	  int maxy = (kx <= 1) ? order : order-kx;
	  
	  FlatVector<double> hv_poly_z2 (maxy+1, &hv_poly_z(0));
	  FlatMatrix<> pol2 (maxy+1, ir.GetNIPY(), &pol(0,0));
	  
	  for (int ky = 0; ky <= maxy; ky++)
	    hv_poly_z2(ky) = gridz_polxy(iz, quad2trig[kx*(order+1)+ky]);
	  
	  hvyz = pol2 * hv_poly_z2;
	  NgProfiler::AddFlops (transformytimer, (maxy+1)*ir.GetNIPY());
	  
	  gridyz_polx.Col(kx).Range (iz*ir.GetNIPY(), (iz+1)*ir.GetNIPY()) = hvyz;
	}
    */




    NgProfiler::StopTimer (transformytimer);
    

    // transform x
    NgProfiler::StartTimer (transformxtimer);

    /*
    gridvalues = 0;
    int ipnr = 0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  for (int kx = 0; kx <= order; kx++)
	     gridvalues(ipnr) += legxbub(ix, kx) * gridyz_polx(ipyz, kx);
    */

    FlatMatrix<> mat_gridvalues (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), &gridvalues[0]);
    mat_gridvalues = gridyz_polx * Trans (legxbub);


    NgProfiler::StopTimer (transformxtimer);
    NgProfiler::AddFlops (transformxtimer, gridyz_polx.Height()*legxbub.Height()*legxbub.Width());
    
    lh.CleanUp (heapp);

    NgProfiler::StopTimer (transformtimer);
    NgProfiler::StopTimer (maintimer);
  }












  void H1HighOrderTetFast :: 
  EvaluateShapeGridTrans (const TensorProductIntegrationRule & ir,
			  const FlatVector<double> gridvalues,
			  FlatVector<double> coefs,
			  LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();

    FlatMatrix<double> gridyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    FlatMatrix<double> gridz_polxy(ir.GetNIPZ(), (order+1)*(order+2)/2+1, lh);


    static bool init = 0;
    
    static Matrix<> legxbub(ir.GetNIPX(), order+1);
    static Matrix<> legybub(ir.GetNIPY(), order+1);
    static Matrix<> legzbub(ir.GetNIPZ(), order+1);
    static Matrix<> powzm(ir.GetNIPZ(), order+1);
    static Matrix<> powym(ir.GetNIPZ(), order+1);
    
    static ARRAY<FlatMatrix<> > jaczbTs (order+1);
    static ARRAY<FlatMatrix<> > jacybTs (order+1);

    if (!init)
      {
	init = 1;

	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    double x = ir.GetIPX(ix);
	    double * hp = &legxbub(ix, 2);
	    // LegendrePolynomial (order-2, 2*x-1, hp);
	    JacobiPolynomial (order-2, 2*x-1, 1, 1, hp);
	    
	    legxbub(ix, 0) = x;
	    legxbub(ix, 1) = 1-x;
	    for (int j = 0; j <= order-2; j++)
	      legxbub(ix, j+2) *= x * (1-x);
	  }
	
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    double y = ir.GetIPY(iy);
	    double * hp = &legybub(iy, 2);
	    // LegendrePolynomial (order-2, 2*y-1, hp);
	    JacobiPolynomial (order-2, 2*y-1, 1, 1, hp);
	    
	    for (int j = 2; j <= order; j++)
	      legybub(iy, j) *= y*(1-y);
	    legybub(iy,0) = y;
	    legybub(iy,1) = 1-y;
	  }
	
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    double z = ir.GetIPZ(iz);
	    double * hp = &legzbub(iz, 2);
	    // LegendrePolynomial (order-2, 2*z-1, hp);
	    JacobiPolynomial (order-2, 2*z-1, 1, 1, hp);
	    
	    for (int j = 2; j <= order; j++)
	      legzbub(iz, j) *= z*(1-z);
	    legzbub(iz,0) = z;
	    legzbub(iz,1) = 1-z;
	  }

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    double fac = 1-ir.GetIPZ(iz);
	    double val = 1;
	    for (int j = 0; j <= order; j++, val *= fac)
	      powzm(iz, j) = val;
	  }
	
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    double fac = 1-ir.GetIPY(iy);
	    double val = 1;
	    for (int j = 0; j <= order; j++, val *= fac)
	      powym(iy, j) = val;
	  }
    
	
	FlatMatrix<> jaczb(ir.GetNIPZ(), order+1, lh);
	FlatVector<> hvals(order+1, lh);
	
	for (int k = 0; k <= order; k++)
	  {
	    jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), new double[(order+1)*ir.GetNIPZ()]);
	    
	    if (k >= 2)
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    double * hp = &jaczb(iz, 1);
		    JacobiPolynomial (order-1, 2*ir.GetIPZ(iz)-1, 2*k-3, 1, hp);
		    for (int j = 1; j <= order; j++)
		      jaczb(iz, j) *= ir.GetIPZ(iz) * powzm(iz, k);

		    jaczb(iz, 0) = powzm(iz, k);
		  }
		jaczbTs[k] = Trans (jaczb);
	      }
	  }
	jaczbTs[0] = Trans (legzbub);
	jaczbTs[1] = Trans (legzbub);
	
	FlatMatrix<> jacyb(ir.GetNIPY(), order+1, lh);
	
	for (int k = 0; k <= order; k++)
	  {
	    jacybTs[k].AssignMemory (order+1, ir.GetNIPY(), new double[(order+1)*ir.GetNIPZ()]);
	    if (k >= 2)
	      {
		for (int iy = 0; iy < ir.GetNIPY(); iy++)
		  {
		    double * hp = &jacyb(iy, 1);
		    JacobiPolynomial (order-1, 2*ir.GetIPY(iy)-1, 2*k-3, 1, hp);
		    for (int j = 1; j <= order; j++)
		      jacyb(iy, j) *= ir.GetIPY(iy) * powym(iy, k);
		    
		    jacyb(iy, 0) = powym(iy, k);
		  }
		
		jacybTs[k] = Trans (jacyb);
	      }
	  }
	jacybTs[0] = Trans (legybub);
	jacybTs[1] = Trans (legybub);
      }



    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++) isort[sort[i]] = i;



    

    int ndof2d = (order+1)*(order+2)/2;

    
    // scaling with (1-z)^m
    
    FlatArray<int> powerz(ndof2d, lh);
    for (int k = 0; k < 3; k++)
      powerz[k] = 0;
    
    int pos = 3;
    for (int k = 2; k <= order; k++)
      {
	powerz[pos++] = k;
	powerz[pos++] = k;
      }
    
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	powerz[pos++] = kx+ky;
    


	

    // transform prism ---> tet
    
    ARRAY<int> prism2tet;
    GetPrismTetReordering (isort, prism2tet);
    


    // transform quad ---> trig
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    quad2trig = -1;
    quad2trig[0] = 0;
    quad2trig[order+1] = 0;
    quad2trig[1] = 1;
    quad2trig[order+2] = 2;
	
    pos = 3;
    for (int ky = 2; ky <= order; ky++)
      {
	quad2trig[ky] = pos++;
	quad2trig[ky+order+1] = pos++;
      }
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	quad2trig[kx*(order+1)+ky] = pos++;
    
    

    // transform x
    /*
    gridyz_polx = 0;
    
    int ipnr = 0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  for (int kx = 0; kx <= order; kx++)
	    gridyz_polx(ipyz, kx) += legxbub(ix, kx) * gridvalues(ipnr);
    */


    FlatMatrix<> hgv(ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), const_cast<double*> (&gridvalues(0)));
    gridyz_polx = hgv * legxbub;



    // transform y

    gridz_polxy = 0.0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int kx = 0; kx <= order; kx++)
	{
	  FlatMatrix<> pol = jacybTs[kx];
	  for (int ky = 0; ky <= order; ky++)
	    if (quad2trig[kx*(order+1)+ky] != -1)
	      {
		double sum = 0;
		for (int iy = 0, ipyz = iz*ir.GetNIPY(); iy < ir.GetNIPY(); iy++, ipyz++)
		  sum += pol(ky, iy) * gridyz_polx(ipyz, kx);
		gridz_polxy(iz, quad2trig[kx*(order+1)+ky]) += sum;
	      }
	}
    
    // transform z
    
    coefs = 0.0;
    for (int kxy = 0; kxy < ndof2d; kxy++)
      {
	FlatMatrix<> pol = jaczbTs[powerz[kxy]];
	for (int kz = 0; kz <= order; kz++)
	  if (prism2tet[kxy*(order+1)+kz] != -1)
	    {
	      double sum = 0;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		sum += pol(kz, iz) * gridz_polxy(iz, kxy);
	      coefs(prism2tet[kxy*(order+1)+kz]) += sum;
	    }
      }

    if (augmented == 1)
      {
	for (int i = 0; i < ir.GetNIP(); i++)
	  for (int j = 0; j < 4; j++)
	    coefs(4+j) += gridvalues(i) * low_energy_vertex_shapes(4*i+isort[j]);
      }

    lh.CleanUp (heapp);
  }
 













  void H1HighOrderTetFast ::
  EvaluateDShapeGrid (const TensorProductIntegrationRule & ir,
		      const FlatVector<double> coefs,
		      FlatMatrix<double> gridvalues_1,
		      LocalHeap & lh) const
  {
    static int maintimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid");
    static int preparetimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, prepare");
    static int transformtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transform");

    static int transformmaptimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transformmap");
    static int transformxtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transformx");
    static int transformytimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transformy");
    static int transformztimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transformz");
    static int augmentedtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transform augmented");
    static int tet2hextimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transform tet2hex");
    static int edgeshapestimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGrid, transform edges");



    NgProfiler::StartTimer (maintimer);
    NgProfiler::StartTimer (preparetimer);


    void * heapp = lh.GetPointer();

    FlatMatrixFixWidth<3> gridvalues (gridvalues_1);


    //    FlatMatrix<double> gridyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    //    FlatMatrix<double> griddyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    //    FlatMatrix<double> gridydz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    FlatMatrix<double> gridz_polxy(ir.GetNIPZ(), (order+1)*(order+2)/2+1, lh);
    FlatMatrix<double> griddz_polxy(ir.GetNIPZ(), (order+1)*(order+2)/2+1, lh);
    

    static bool init = 0;
    static Matrix<> legxbub(ir.GetNIPX(), order+1);
    static Matrix<> dlegxbub(ir.GetNIPX(), order+1);
    static ARRAY<FlatMatrix<> > jaczbs (order+1);
    static ARRAY<FlatMatrix<> > djaczbs (order+1);
    static ARRAY<FlatMatrix<> > jaczbs_inv (order+1);
    static ARRAY<double> inv_y (ir.GetNIPY());
    static ARRAY<double> inv_z (ir.GetNIPZ());

    if (!init)
      {
	init = 1;

	ArrayMem<AutoDiff<1>, 50> ada(order+1);

	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    AutoDiff<1> adx (ir.GetIPX(ix), 0);
	    AutoDiff<1> * hp = &ada[2];

	    JacobiPolynomial (order-2, 2*adx-1, 1, 1, hp);
	    for (int j = 2; j <= order; j++)
	      ada[j] *= adx * (1.0-adx);
	    ada[0] = adx;
	    ada[1] = 1.0-adx;

	    for (int j = 0; j <= order; j++)
	      {
		legxbub(ix, j) = ada[j].Value();
		dlegxbub(ix, j) = ada[j].DValue(0);
	      }
	  }

	ArrayMem<AutoDiff<1>, 30> powzk(ir.GetNIPZ());
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  powzk[iz] = 1.0;

	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  inv_y[iy] = 1.0 / (1-ir.GetIPY(iy));

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  inv_z[iz] = 1.0 / (1-ir.GetIPZ(iz));


	for (int k = 0; k <= order; k++)
	  {
	    int oo = (k < 2) ? order : order - k;

	    jaczbs[k].AssignMemory (ir.GetNIPZ(), oo+1, new double[(oo+1)*(ir.GetNIPZ())]);
	    djaczbs[k].AssignMemory (ir.GetNIPZ(), oo+1, new double[(oo+1)*(ir.GetNIPZ())]);
	    jaczbs_inv[k].AssignMemory (ir.GetNIPZ(), oo+1, new double[(oo+1)*(ir.GetNIPZ())]);
	
	    if (k < 2)
	      {
		jaczbs[k] = legxbub;
		djaczbs[k] = dlegxbub;
	      }
	    else
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    AutoDiff<1> adz (ir.GetIPZ(iz), 0);

		    AutoDiff<1> * hp = &ada[1];
		    JacobiPolynomial (order-1, 2*adz-1, 2*k-3, 1, hp);
		    
		    for (int j = 1; j <= order; j++)
		      ada[j] *= adz * powzk[iz];
		    ada[0] = powzk[iz];
		
		    for (int j = 0; j <= oo; j++)
		      {
			jaczbs[k](iz, j) = ada[j].Value();
			djaczbs[k](iz, j) = ada[j].DValue(0);
		      }
		  }
	      }

	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      for (int j = 0; j <= oo; j++)
		jaczbs_inv[k](iz,j) = jaczbs[k](iz,j)*inv_z[iz];
	
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      powzk[iz] *= 1 - AutoDiff<1> (ir.GetIPZ(iz), 0);
	  }


      }

    NgProfiler::StopTimer (preparetimer);

    NgProfiler::StartTimer (transformtimer);

    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3-i; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++) isort[sort[i]] = i;



    

    int ndof2d = (order+1)*(order+2)/2;

    
    // scaling with (1-z)^m
    
    FlatArray<int> powerz(ndof2d, lh);
    for (int k = 0; k < 3; k++)
      powerz[k] = 0;
    
    int pos = 3;
    for (int k = 2; k <= order; k++)
      {
	powerz[pos++] = k;
	powerz[pos++] = k;
      }
    
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	powerz[pos++] = kx+ky;
    


	

    // transform prism ---> tet
    
    ARRAY<int> prism2tet;
    GetPrismTetReordering (isort, prism2tet);
    


    // transform quad ---> trig
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    quad2trig = -1;
    quad2trig[0] = 0;
    quad2trig[order+1] = 0;
    quad2trig[1] = 1;
    quad2trig[order+2] = 2;
	
    pos = 3;
    for (int ky = 2; ky <= order; ky++)
      {
	quad2trig[ky] = pos++;
	quad2trig[ky+order+1] = pos++;
      }
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	quad2trig[kx*(order+1)+ky] = pos++;
    

    // transform z

    NgProfiler::StartTimer (transformztimer);
    
    FlatVector<double> z_values (order+1, lh);
    FlatVector<double> hvz (ir.GetNIPZ(), lh);
    FlatVector<double> hvdz (ir.GetNIPZ(), lh);

    gridz_polxy = 0;
    griddz_polxy = 0;
    for (int kxy = 0; kxy < ndof2d; kxy++)
      {
	int powz = powerz[kxy];
	for (int kz = 0; kz <= order-powz; kz++)
	  z_values[kz] = coefs(prism2tet[kxy*(order+1)+kz]);
	    
	hvz = jaczbs[powz] * z_values;
	hvdz = djaczbs[powz] * z_values;

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    gridz_polxy(iz, kxy) = hvz(iz) * inv_z[iz]; 
	    griddz_polxy(iz, kxy) = hvdz(iz);
	  }
      }

    NgProfiler::StopTimer (transformztimer);



    // transform y

    NgProfiler::StartTimer (transformytimer);

    FlatMatrix<> hvm4 (ir.GetNIPZ(), order+1, lh);
    FlatMatrix<> hvm5 (ir.GetNIPZ(), order+1, lh);
    
    FlatMatrix<> gridyz_polxT (order+1, ir.GetNIPY()*ir.GetNIPZ(), lh);
    FlatMatrix<> griddyz_polxT (order+1, ir.GetNIPY()*ir.GetNIPZ(), lh);
    FlatMatrix<> gridydz_polxT (order+1, ir.GetNIPY()*ir.GetNIPZ(), lh);

    for (int kx = 0; kx <= order; kx++)
      {
	int maxy = (kx <= 1) ? order : order-kx;

	FlatMatrix<> quad_values (ir.GetNIPZ(), maxy+1, &hvm4(0,0));
	FlatMatrix<> quad_valuesdz (ir.GetNIPZ(), maxy+1, &hvm5(0,0));
	
	for (int ky = 0; ky <= maxy; ky++)
	  {
	    int ind = quad2trig[kx*(order+1)+ky];
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      quad_values(iz, ky) = gridz_polxy(iz, ind);
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      quad_valuesdz(iz, ky) = griddz_polxy(iz, ind);
	  }

	FlatMatrix<> hvy1a (ir.GetNIPZ(), ir.GetNIPY(), &gridyz_polxT(kx,0));
	FlatMatrix<> hvy2a (ir.GetNIPZ(), ir.GetNIPY(), &griddyz_polxT(kx,0));
	FlatMatrix<> hvy3a (ir.GetNIPZ(), ir.GetNIPY(), &gridydz_polxT(kx,0));

#ifdef LAPACK
	LapackMultABt (quad_values, jaczbs_inv[kx], hvy1a);
	LapackMultABt (quad_values, djaczbs[kx],    hvy2a);
	LapackMultABt (quad_valuesdz, jaczbs[kx],   hvy3a);
#else
	
	hvy1a = quad_values * Trans (jaczbs_inv[kx]);
	hvy2a = quad_values * Trans (djaczbs[kx]);
	hvy3a = quad_valuesdz * Trans (jaczbs[kx]);
#endif

	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  for (int iy = 0; iy < ir.GetNIPY(); iy++)
	    hvy3a(iz,iy) += ir.GetIPY(iy) * hvy2a(iz,iy);

	NgProfiler::AddFlops (transformytimer, 3*ir.GetNIPZ()*(maxy+1)*ir.GetNIPY());
      }
    
    /*
    FlatVector<> hvy1 (ir.GetNIPY(), lh);
    FlatVector<> hvy2 (ir.GetNIPY(), lh);
    FlatVector<> hvy3 (ir.GetNIPY(), lh);
    
    FlatMatrix<> hvm4 (order+1, lh);
    FlatMatrix<> hvm5 (order+1, lh);
    
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int kx = 0; kx <= order; kx++)
	{
	  int maxy = (kx <= 1) ? order : order-kx;
	  
	  FlatVector<> quad_values (maxy+1, &hvm4(0));
	  FlatVector<> quad_valuesdz (maxy+1, &hvm5(0));
	
	  for (int ky = 0; ky <= maxy; ky++)
	    {
	      int ind = quad2trig[kx*(order+1)+ky];
	      quad_values(ky) = gridz_polxy(iz, ind);
	      quad_valuesdz(ky) = griddz_polxy(iz, ind);
	    }

	  hvy1 = jaczbs[kx] * quad_values;
	  hvy2 = djaczbs[kx] * quad_values;
	  hvy3 = jaczbs[kx] * quad_valuesdz;

	  for (int iy = 0, ipyz = iz*ir.GetNIPY(); iy < ir.GetNIPY(); iy++, ipyz++)
	    {
	      gridyz_polx(ipyz, kx) = hvy1(iy) * inv_y[iy];
	      griddyz_polx(ipyz, kx) = hvy2(iy);
	      gridydz_polx(ipyz, kx) = hvy3(iy) + ir.GetIPY(iy) * hvy2(iy);
	    }
	  

	NgProfiler::AddFlops (transformytimer, 3*(maxy+1)*ir.GetNIPY());
      }
    */

    NgProfiler::StopTimer (transformytimer);




    // transform x
    NgProfiler::StartTimer (transformxtimer);
 
    FlatMatrix<> gridvalues_dx (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), lh);
    FlatMatrix<> gridvalues_dy (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), lh);
    FlatMatrix<> gridvalues_dz (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), lh);
 
#ifdef LAPACK
    LapackMultAtBt (gridyz_polxT, dlegxbub, gridvalues_dx);
    LapackMultAtBt (griddyz_polxT, legxbub, gridvalues_dy);
    LapackMultAtBt (gridydz_polxT, legxbub, gridvalues_dz);
#else
    gridvalues_dx = Trans(gridyz_polxT)*Trans(dlegxbub);
    gridvalues_dy = Trans(griddyz_polxT)*Trans(legxbub);
    gridvalues_dz = Trans(gridydz_polxT)*Trans(legxbub);    
#endif

    NgProfiler::StopTimer (transformxtimer);
    NgProfiler::AddFlops (transformxtimer, 3*ir.GetNIPY()*ir.GetNIPZ()*ir.GetNIPX()*(order+1));


    NgProfiler::StartTimer (tet2hextimer);

    Vec<4> perm;
    perm(3) = 0;

    for (int iz = 0, ipyz = 0, ipnr = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  {
	    double x = ir.GetIPX(ix);

	    double sum1 = gridvalues_dx(ipyz, ix);
	    double sum2 = gridvalues_dy(ipyz, ix);
	    double sum3 = gridvalues_dz(ipyz, ix);

	    perm(2) = sum1;
	    perm(1) = x * sum1 + sum2;
	    perm(0) = x * sum1 + sum3;

	    for (int j = 0; j < 3; j++)
	      gridvalues(ipnr,j) = perm(isort[j]) - perm(isort[3]);
	  }

    NgProfiler::StopTimer (tet2hextimer);


    /*
      gridvalues = 0;
      int ipnr = 0;
      for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
      for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
      for (int kx = 0; kx <= order; kx++)
      {
      gridvalues(ipnr,0) += dlegxbub(ix, kx) * gridyz_polx(ipyz, kx);
      gridvalues(ipnr,1) += legxbub(ix, kx) * griddyz_polx(ipyz, kx);
      gridvalues(ipnr,2) += legxbub(ix, kx) * gridydz_polx(ipyz, kx);
      }


      for (int i = 0; i < ir.GetNIP(); i++)
      {
      Mat<3> thextet, thextet_inv;

      ir.GetIP_DTet_DHex(i, thextet);
      CalcInverse (thextet, thextet_inv);

      Vec<3> v1, v2;
      for (int j = 0; j < 3; j++)
      v1(j) = gridvalues (i, j);

      v2 =  Trans(thextet_inv) * v1;

      for (int j = 0; j < 3; j++)
      gridvalues(i,j) = v2(j);
      }
    */





    NgProfiler::StartTimer (augmentedtimer);

    if (augmented == 1)
      {
	Mat<4,3> lamis (0.0);
	for (int j = 0; j < 3; j++)
	  {
	    lamis(isort[j],j) = coefs(4+j);
	    lamis(isort[3],j) = -coefs(7);
	  }

	FlatMatrixFixWidth<4> vertex_shapes(ir.GetNIP(), &low_energy_vertex_dshapes(0));
	gridvalues += vertex_shapes * lamis;
      }

    NgProfiler::StopTimer (augmentedtimer);


    lh.CleanUp (heapp);

    NgProfiler::StopTimer (transformtimer);
    NgProfiler::StopTimer (maintimer);
  }










  void H1HighOrderTetFast ::
  EvaluateDShapeGridTrans (const TensorProductIntegrationRule & ir,
			   const FlatMatrix<double> gridvalues_1,
			   FlatVector<double> coefs,
			   LocalHeap & lh) const
  {
    static int maintimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT");
    static int preparetimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT, prepare");
    static int transformtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT, transform");

    static int transformxtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT, transformx");
    static int transformytimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT, transformy");
    static int transformztimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT, transformz");
    static int augmentedtimer = NgProfiler::CreateTimer ("H1HOtet::EvaluateDShapeGridT, transform augmented");


    FlatMatrixFixWidth<3> gridvalues (gridvalues_1.Height(), const_cast<double*> (&gridvalues_1(0,0)));


    NgProfiler::StartTimer (maintimer);
    NgProfiler::StartTimer (preparetimer);


    void * heapp = lh.GetPointer();

    FlatMatrix<double> gridyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    FlatMatrix<double> griddyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    FlatMatrix<double> gridydz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);
    FlatMatrix<double> gridz_polxy(ir.GetNIPZ(), (order+1)*(order+2)/2+1, lh);
    FlatMatrix<double> griddz_polxy(ir.GetNIPZ(), (order+1)*(order+2)/2+1, lh);
    

    static bool init = 0;
    static Matrix<> legxbub(ir.GetNIPX(), order+1);
    static Matrix<> dlegxbub(ir.GetNIPX(), order+1);
    static Matrix<> legxbubT(order+1, ir.GetNIPX());
    static Matrix<> dlegxbubT(order+1, ir.GetNIPX());
    static ARRAY<FlatMatrix<> > jaczbTs (order+1);
    static ARRAY<FlatMatrix<> > djaczbTs (order+1);



    if (!init)
      {
	init = 1;

    ArrayMem<AutoDiff<1>, 50> ada(order+1);

    for (int ix = 0; ix < ir.GetNIPX(); ix++)
      {
	AutoDiff<1> adx (ir.GetIPX(ix), 0);
	AutoDiff<1> * hp = &ada[2];

	JacobiPolynomial (order-2, 2*adx-1, 1, 1, hp);
	for (int j = 2; j <= order; j++)
	  ada[j] *= adx * (1.0-adx);
	ada[0] = adx;
	ada[1] = 1.0-adx;

	for (int j = 0; j <= order; j++)
	  {
	    legxbub(ix, j) = ada[j].Value();
	    dlegxbub(ix, j) = ada[j].DValue(0);
	    legxbubT(j, ix) = ada[j].Value();
	    dlegxbubT(j, ix) = ada[j].DValue(0);
	  }
      }

    ArrayMem<AutoDiff<1>, 30> powzk(ir.GetNIPZ());
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      powzk[iz] = 1.0;


    for (int k = 0; k <= order; k++)
      {
	jaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), new double[(order+1)*ir.GetNIPZ()]);
	djaczbTs[k].AssignMemory (order+1, ir.GetNIPZ(), new double[(order+1)*ir.GetNIPZ()]);

	if (k < 2)
	  {
	    jaczbTs[k] = legxbubT;
	    djaczbTs[k] = dlegxbubT;
	  }
	else
	  {
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      {
		AutoDiff<1> adz (ir.GetIPZ(iz), 0);

		AutoDiff<1> * hp = &ada[1];
		JacobiPolynomial (order-1, 2*adz-1, 2*k-3, 1, hp);

		for (int j = 1; j <= order; j++)
		  ada[j] *= adz * powzk[iz];
		ada[0] = powzk[iz];

		
		for (int j = 0; j <= order; j++)
		  {
		    jaczbTs[k](j, iz) = ada[j].Value();
		    djaczbTs[k](j, iz) = ada[j].DValue(0);
		  }
	      }
	  }
	
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  {
	    AutoDiff<1> adz (ir.GetIPZ(iz), 0);
	    powzk[iz] *= 1-adz;
	  }
      }
      }

    NgProfiler::StopTimer (preparetimer);
    NgProfiler::StartTimer (transformtimer);


    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++) isort[sort[i]] = i;



    

    int ndof2d = (order+1)*(order+2)/2;

    
    // scaling with (1-z)^m
    
    FlatArray<int> powerz(ndof2d, lh);
    for (int k = 0; k < 3; k++)
      powerz[k] = 0;
    
    int pos = 3;
    for (int k = 2; k <= order; k++)
      {
	powerz[pos++] = k;
	powerz[pos++] = k;
      }
    
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	powerz[pos++] = kx+ky;
    


	

    // transform prism ---> tet
    
    ARRAY<int> prism2tet;
    GetPrismTetReordering (isort, prism2tet);
    


    // transform quad ---> trig
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    quad2trig = -1;
    quad2trig[0] = 0;
    quad2trig[order+1] = 0;
    quad2trig[1] = 1;
    quad2trig[order+2] = 2;
	
    pos = 3;
    for (int ky = 2; ky <= order; ky++)
      {
	quad2trig[ky] = pos++;
	quad2trig[ky+order+1] = pos++;
      }
    for (int ky = 0; ky <= order; ky++)
      for (int kx = 2; kx <= order-ky; kx++)
	quad2trig[kx*(order+1)+ky] = pos++;
    


    NgProfiler::StartTimer (transformxtimer);






    // transform x
    /*    
    gridyz_polx = 0;
    griddyz_polx = 0;
    gridydz_polx = 0;

    int ipnr = 0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  {
	    double x = ir.GetIPX(ix);
	    double y = ir.GetIPY(iy);
	    double z = ir.GetIPZ(iz);
  
	    Vec<3> v1;
	    Vec<4> v2;

	    for (int j = 0; j < 3; j++)
	      v1(j) = gridvalues (ipnr, j);

	    for (int j = 0; j < 3; j++)
	      v2(isort[j]) = v1(j);
	    v2(isort[3]) = - (v1(0)+v1(1)+v1(2));

	    
	    v1(0) = x * v2(0) + x * v2(1) + v2(2);
	    v1(1) = y * v2(0) +     v2(1);
	    v1(2) =     v2(0);

	    v1(0) /= (1-y)*(1-z);
	    v1(1) /= 1-z;

	    for (int kx = 0; kx <= order; kx++)
	      gridyz_polx(ipyz, kx) += dlegxbub(ix, kx) * v1(0);
	    for (int kx = 0; kx <= order; kx++)
	      griddyz_polx(ipyz, kx) += legxbub(ix, kx) * v1(1);
	    for (int kx = 0; kx <= order; kx++)
	      gridydz_polx(ipyz, kx) += legxbub(ix, kx) * v1(2);
	  }
    */

    FlatMatrix<> gridvalues_dx (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), lh);
    FlatMatrix<> gridvalues_dy (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), lh);
    FlatMatrix<> gridvalues_dz (ir.GetNIPY()*ir.GetNIPZ(), ir.GetNIPX(), lh);

    int ipnr = 0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	{
	  double y = ir.GetIPY(iy);
	  double z = ir.GetIPZ(iz);
	  double fac1 = 1.0 / ((1-y)*(1-z));
	  double fac2 = 1.0 / (1-z);

	  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	    {
	      double x = ir.GetIPX(ix);
	      
	      Vec<3> v1;
	      Vec<4> v2;
	      
	      for (int j = 0; j < 3; j++)
		v1(j) = gridvalues (ipnr, j);
	      
	      for (int j = 0; j < 3; j++)
		v2(isort[j]) = v1(j);
	      v2(isort[3]) = - (v1(0)+v1(1)+v1(2));
	      
	      v1(0) = x * v2(0) + x * v2(1) + v2(2);
	      v1(1) = y * v2(0) +     v2(1);
	      v1(2) =     v2(0);
	      
	      gridvalues_dx(ipyz, ix) = fac1*v1(0);
	      gridvalues_dy(ipyz, ix) = fac2*v1(1);
	      gridvalues_dz(ipyz, ix) = v1(2);
	    }
	}

    FlatMatrix<double> hm(order+1, ir.GetNIPX(), &legxbubT(0,0));
    FlatMatrix<double> dhm(order+1, ir.GetNIPX(), &dlegxbubT(0,0));

#ifdef LAPACK
    LapackMultABt (gridvalues_dx, dhm, gridyz_polx);
    LapackMultABt (gridvalues_dy, hm, griddyz_polx);
    LapackMultABt (gridvalues_dz, hm, gridydz_polx);
#else
    gridyz_polx = gridvalues_dx * Trans(dhm);
    griddyz_polx = gridvalues_dy * Trans(hm);
    gridydz_polx = gridvalues_dz * Trans(hm);    
#endif

    /*
    int ipnr = 0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	{
	  double y = ir.GetIPY(iy);
	  double z = ir.GetIPZ(iz);
	  double fac1 = 1.0 / ((1-y)*(1-z));
	  double fac2 = 1.0 / (1-z);

	  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	    {
	      double x = ir.GetIPX(ix);
	      
	      Vec<3> v1;
	      Vec<4> v2;
	      
	      for (int j = 0; j < 3; j++)
		v1(j) = gridvalues (ipnr, j);
	      
	      for (int j = 0; j < 3; j++)
		v2(isort[j]) = v1(j);
	      v2(isort[3]) = - (v1(0)+v1(1)+v1(2));
	      
	      v1(0) = x * v2(0) + x * v2(1) + v2(2);
	      v1(1) = y * v2(0) +     v2(1);
	      v1(2) =     v2(0);
	      
	      gridvalues_row_dx(ix) = fac1*v1(0);
	      gridvalues_row_dy(ix) = fac2*v1(1);
	      gridvalues_row_dz(ix) = v1(2);
	    }

	  FlatMatrix<double> hm(order+1, ir.GetNIPX(), &legxbubT(0,0));
	  FlatMatrix<double> dhm(order+1, ir.GetNIPX(), &dlegxbubT(0,0));

	  FlatVector<double> gridyz_polx_vec(order+1, &gridyz_polx(ipyz, 0));
	  FlatVector<double> griddyz_polx_vec(order+1, &griddyz_polx(ipyz, 0));	  
	  FlatVector<double> gridydz_polx_vec(order+1, &gridydz_polx(ipyz, 0));	  

	  gridyz_polx_vec += dhm * gridvalues_row_dx;
	  griddyz_polx_vec += hm * gridvalues_row_dy;
	  gridydz_polx_vec += hm * gridvalues_row_dz;
	}
*/




    NgProfiler::StopTimer (transformxtimer);



    NgProfiler::StartTimer (transformytimer);

    // transform y

    gridz_polxy = 0.0;
    griddz_polxy = 0.0;
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int kx = 0; kx <= order; kx++)
	{
	  int ipyz = iz * ir.GetNIPY();
	  int maxy = (kx <= 1) ? order : order-kx;
	  
	  FlatMatrix<> pol = jaczbTs[kx];
	  FlatMatrix<> dpol = djaczbTs[kx];

	  for (int ky = 0; ky <= maxy; ky++)
	    {
	      double sum = 0.0;

	      for (int iy = 0; iy < ir.GetNIPY(); iy++)
		sum += pol(ky, iy) * gridyz_polx(ipyz+iy, kx);
	      for (int iy = 0; iy < ir.GetNIPY(); iy++)
		sum += dpol(ky, iy) * griddyz_polx(ipyz+iy, kx);

	      gridz_polxy(iz, quad2trig[kx*(order+1)+ky]) += sum;

	      sum = 0.0;
	      for (int iy = 0; iy < ir.GetNIPY(); iy++)
		sum += pol(ky, iy) * gridydz_polx(ipyz+iy, kx);
	      griddz_polxy(iz, quad2trig[kx*(order+1)+ky]) += sum;
	    }
	}



    /*
    gridz_polxy = 0.0;
    griddz_polxy = 0.0;
    for (int iz = 0, ipyz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	{
	  for (int ky = 0; ky <= order; ky++)
	    for (int kx = 0; kx <= order; kx++)
	      if (quad2trig[kx*(order+1)+ky] != -1)
		{
		  gridz_polxy(iz, quad2trig[kx*(order+1)+ky]) +=
		    jaczbTs[kx](ky, iy) * gridyz_polx(ipyz, kx) +
		    djaczbTs[kx](ky, iy) * griddyz_polx(ipyz, kx);
		    
		  griddz_polxy(iz, quad2trig[kx*(order+1)+ky]) +=
		    jaczbTs[kx](ky, iy) * gridydz_polx(ipyz, kx);
		  
		}
	}
    */
    NgProfiler::StopTimer (transformytimer);
    NgProfiler::StartTimer (transformztimer);

    // transform z

    coefs = 0;
    for (int kxy = 0; kxy < ndof2d; kxy++)
      {
	FlatMatrix<> pol = jaczbTs[powerz[kxy]];
	FlatMatrix<> dpol = djaczbTs[powerz[kxy]];
	
	for (int kz = 0; kz <= order-powerz[kxy]; kz++)
	  {
	    double sum = 0;
	    
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      sum += pol(kz, iz) * gridz_polxy(iz, kxy);
	    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	      sum += dpol (kz, iz) * griddz_polxy(iz, kxy);
	    
	    coefs(prism2tet[kxy*(order+1)+kz]) += sum;
	  }
      }


    NgProfiler::StopTimer (transformztimer);

    NgProfiler::StartTimer (augmentedtimer);

    if (augmented == 1)
      {
	Mat<4,3> lamis;
	FlatMatrixFixWidth<4> vertex_shapes(ir.GetNIP(), &low_energy_vertex_dshapes(0));
	
	lamis = Trans (vertex_shapes) * gridvalues;
	for (int j = 0; j < 3; j++)
	  {
	    coefs(4+j) += lamis(isort[j], j);
	    coefs(4+3) -= lamis(isort[3], j);
	  }


	/*
	for (int i = 0; i < ir.GetNIP(); i++)
	  for (int j = 0; j < 4; j++)
	    for (int k = 0; k < 3; k++)
	      coefs(4+j) += gridvalues(i,k) * low_energy_vertex_dshapes(4*i+isort[j]) * lami[j][k]; 
	*/
      }

    NgProfiler::StopTimer (augmentedtimer);


    lh.CleanUp (heapp);

    NgProfiler::StopTimer (transformtimer);
    NgProfiler::StopTimer (maintimer);


    /*
    (*testout) << "coefs, new : " << coefs << endl;

    Vector<> coefs2 (coefs.Size());
    NodalFiniteElement::EvaluateDShapeGridTrans (ir, gridvalues, coefs2, lh);
    
    (*testout) << "coefs, old : " << coefs2 << endl;
    (*testout) << "diff: " << (coefs-coefs2) << endl;
    cout << "|coefs_new - coefs_old| = " << L2Norm (coefs-coefs2) << endl;
    */
  }











 
}
