/*********************************************************************/
/* File:   h1trig_fast.cpp                                           */
/* Author: Start                                                     */
/* Date:   30. Nov. 2004                                             */
/*********************************************************************/

 
#include <fem.hpp>
namespace ngfem
{
  using namespace ngfem;


  Vector<> H1HighOrderTrigFast :: low_energy_vertex_shapes(0);
  Vector<> H1HighOrderTrigFast :: low_energy_vertex_dshapes(0);

  RecPol * H1HighOrderTrigFast :: polx = 0;
  ARRAY<RecPol*> H1HighOrderTrigFast :: polsy;

  Matrix<> H1HighOrderTrigFast :: legxbub;
  ARRAY<FlatMatrix<> > H1HighOrderTrigFast :: jacybs;

  Matrix<> H1HighOrderTrigFast :: dlegxbub;
  ARRAY<FlatMatrix<> > H1HighOrderTrigFast :: djacybs;



  H1HighOrderTrigFast :: H1HighOrderTrigFast(int aorder)
    : H1HighOrderFiniteElement(2, ET_TRIG)
  {
    order_inner =  INT<3> (aorder,aorder,aorder);
    for (int i = 0; i < 3; i++)
      order_edge[i] = aorder;
    for (int i = 0; i < 3; i++)
      order_vertex[i] = aorder;

    ComputeNDof();

    if (!polx)
      {
	polx = new RecPolJacobi (aorder, 1, 1);
	polx -> MultLinear (0.5, 0.5);   //  *= 0.5 (1+x)
	polx -> MultLinear (0.5, -0.5);  //  *= 0.5 (1-x)

	polsy.SetSize (aorder+1);

	polsy[0] = new RecPolJacobi (aorder, 1, 1);
	polsy[1] = new RecPolJacobi (aorder, 1, 1);

	for (int i = 2; i <= aorder; i++)
	  {
	    polsy[i] = new RecPolJacobi (aorder, 2*i-3, 1);
	    polsy[i] -> MultLinear (0.5, 0.5);
	  }


	const IntegrationRule & ir1d =
	  GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*order+2);
	TensorProductIntegrationRule ir (*this, ir1d);
	
	legxbub.SetSize(ir.GetNIPX(), order+1);
	jacybs.SetSize(order+1);
	dlegxbub.SetSize(ir.GetNIPX(), order+1);
	djacybs.SetSize(order+1);

	ARRAY<AutoDiff<1> > ad_pol (order+1);
	
	for (int ix = 0; ix < ir.GetNIPX(); ix++)
	  {
	    AutoDiff<1> ad_x (ir.GetIPX(ix), 0);
	    polx -> Evaluate (order, 2*ad_x-1, ad_pol);
	    ad_pol[0] = ad_x;
	    ad_pol[1] = 1-ad_x;

	    for (int j = 0; j <= order; j++)
	      {
		legxbub(ix, j) = ad_pol[j].Value();
		dlegxbub(ix, j) = ad_pol[j].DValue(0);
	      }
	  }

	for (int k = 0; k <= order; k++)
	  {
	    int maxy = (k <= 1) ? order : order-k;
	    jacybs[k].AssignMemory (ir.GetNIPY(), maxy+1, new double[(maxy+1)*ir.GetNIPY()]);
	    djacybs[k].AssignMemory (ir.GetNIPY(), maxy+1, new double[(maxy+1)*ir.GetNIPY()]);
	  }

	jacybs[0] = legxbub;
	jacybs[1] = legxbub;

	djacybs[0] = dlegxbub;
	djacybs[1] = dlegxbub;
	
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    AutoDiff<1> y (ir.GetIPY(iy),0);
	    AutoDiff<1> powy = (1-y)*(1-y);

	    for (int k = 2; k <= order; k++, powy *= 1-y)
	      {
		int maxy = (k <= 1) ? order : order-k;
		
		polsy[k] -> Evaluate (maxy, 2*y-1, ad_pol);
		ad_pol[0] = 1;
		
		for (int j = 0; j <= maxy; j++)
		  {
		    ad_pol[j] *= powy;
		    jacybs[k](iy,j) = ad_pol[j].Value();
		    djacybs[k](iy,j) = ad_pol[j].DValue(0);
		  }
	      }
	  }
      }
  }


  void H1HighOrderTrigFast :: ComputeNDof()
  {
    ndof = 3;

    for (int i = 0; i < 3; i++)
      ndof += order_edge[i] - 1;
    int oi = order_inner[0]; 
    ndof += (oi-1)*(oi-2) / 2;

    order = 1;
    for (int i = 0; i < 3; i++)
      if (order_edge[i] > order)
	order = order_edge[i];
    if (oi > order)
      order = oi;
  }



  void H1HighOrderTrigFast ::  
  GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize (0);

    int base = 3;

    for (int i = 0; i < 3; i++)
      base += order_edge[i] - 1;

    if(order_inner[0] > 2)
      {
	int ni = (order_inner[0]-1)*(order_inner[0]-2) / 2;
	for (int i = 0; i < ni; i++)
	  idofs.Append (base+i);
      }
  }


  void H1HighOrderTrigFast :: CalcShape (const IntegrationPoint & ip, 
					 FlatVector<> shape) const
  {
    T_CalcShape<double, double, FlatVector<> > (ip(0), ip(1), shape); 
  }

  void H1HighOrderTrigFast :: CalcDShape (const IntegrationPoint & ip, 
					  FlatMatrix<> dshape) const
  {
    AutoDiff<2> x(ip(0), 0);
    AutoDiff<2> y(ip(1), 1);

    ArrayMem<AutoDiff<2>,40> sds(ndof);
    T_CalcShape<AutoDiff<2>, AutoDiff<2>, FlatArray<AutoDiff<2> > >
      (x,y,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 2; j++)
	dshape(i, j) = sds[i].DValue(j);
  }



  template<typename Tx, typename Ty, typename TFA>  
  void  H1HighOrderTrigFast :: T_CalcShape (Tx x, Ty y, TFA & sds) const
  {
    Tx lami[3] = { x, y, 1-x-y };

    Tx lamisum[3] = { 0, 0, 0 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[j] > vnums[i]) lamisum[i] += lami[j];

    Tx lamiadd[3], lamisub[3];
    for (int i = 0; i < 3; i++)
      {
	lamiadd[i] = lami[i] + lamisum[i];
	lamisub[i] = lami[i] - lamisum[i];
      }

    for (int i = 0; i < 3; i++)
      sds[i] = lami[i];

    int ii = 3;

    // edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    for (int i = 0; i < 3; i++)
      if (order_edge[i] >= 2)
	{ 
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];

	  int nde = order_edge[i]-1;
	  ScaledJacobiPolynomial (order_edge[i]-2, lamisub[es], lamiadd[es], 1, 1, hp);
	  for (int j = 0; j < nde; j++)
	    hp[j] *= lami[ee]*lami[es];

	  ii += nde;
	}

    if (order_inner[0] >= 3)
      {
	int fav[3] = { 0, 1, 2 }; 
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	

	int n = order_inner[0];
	ArrayMem<Tx, 20> polx(n+1);
	ArrayMem<Tx, 50> polsy((n+1) * (n+1));

	ScaledJacobiPolynomial (n-3, lamisub[fav[1]], lamiadd[fav[1]], 1, 1, polx);
	
	for (int j = 0; j <= n; j++)
	  {
	    Tx * hp = &polsy[j * (n+1)];
	    ScaledJacobiPolynomial (n, lamisub[fav[0]], lamiadd[fav[0]], 2*j-3, 1, hp);
	  }

	Tx  bub = lami[fav[0]]*lami[fav[1]]*lami[fav[2]]; 
	for (int j2 = 0; j2 <= n-3; j2++)
	  for (int i2 = 0; i2 <= n-3-j2; i2++)
	    sds[ii++] = bub * polx[i2] * polsy[(i2+2) * (n+1) + j2];
      }
  }
  
  

  void H1HighOrderTrigFast ::
  GetQuadTrigReordering (FlatArray<int> & quad2trig) const
  {
    int sort[3] = { 0, 1, 2 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 2; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
    
    int isort[3];
    for (int i = 0; i < 3; i++) isort[sort[i]] = i;

    quad2trig = -1;	

    quad2trig[0] = sort[0];
    quad2trig[order+1] = sort[0];
    quad2trig[1] = sort[1];
    quad2trig[order+1+1] = sort[2];


    int ii = 3;

    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    for (int i = 0; i < 3; i++)
      if (order >= 2)
	{
	  int es = isort[edges[i][0]], ee = isort[edges[i][1]];	
	  if (ee < es) swap (ee, es);

	  int p = order;

	  switch (es)
	    {
	    case 0:  // edges 0-1, 0-2;
	      {
		for (int kz = 0; kz <= p-2; kz++)
		  quad2trig[(ee-1)*(order+1)+kz+2] = ii++; 
		break;
	      }

	    case 1:
	      {      // edges 1-2
		for (int ky = 2; ky <= p; ky++)
		  quad2trig[ky*(order+1)] = ii++;
		break;
	      }
	    }
	}    

    for (int ky = 0; ky <= order-3; ky++)
      for (int kx = 0; kx <= order-3-ky; kx++, ii++)
	quad2trig[(kx+2)*(order+1)+ky+1] = ii;
  }


  void H1HighOrderTrigFast :: 
  EvaluateShapeGrid (const TensorProductIntegrationRule & ir,
		     const FlatVector<double> coefs,
		     FlatVector<double> gridvalues,
		     LocalHeap & lh) const
  {
    HeapReset hr (lh);

    FlatMatrix<double> gridy_polx(ir.GetNIPY(), order+1, lh);
    
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    GetQuadTrigReordering (quad2trig);

    // transform y
    FlatVector<> hv(order+1, lh);
    for (int kx = 0; kx <= order; kx++)
      {
	int maxy = (kx <= 1) ? order : order-kx;
	for (int ky = 0; ky <= maxy; ky++)
	  hv(ky) = coefs(quad2trig[kx*(order+1)+ky]); 
	gridy_polx.Col(kx) = jacybs[kx] * hv.Range(0,maxy+1);
      }

    // transform x
    FlatMatrix<> mat_gridvalues (ir.GetNIPY(), ir.GetNIPX(), &gridvalues[0]);
    mat_gridvalues = gridy_polx * Trans (legxbub);
  }



  void H1HighOrderTrigFast :: 
  EvaluateShapeGridTrans (const TensorProductIntegrationRule & ir,
			  const FlatVector<double> gridvalues,
			  FlatVector<double> coefs,
			  LocalHeap & lh) const
  {
    HeapReset hr (lh);

    FlatMatrix<double> gridy_polx(ir.GetNIPY(), order+1, lh);
	
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    GetQuadTrigReordering (quad2trig);

    // transform x

    FlatMatrix<> mat_gridvalues (ir.GetNIPY(), ir.GetNIPX(), const_cast<double*> (&gridvalues[0]));
    gridy_polx = mat_gridvalues * legxbub;

    // transform y
    FlatVector<> hv(order+1, lh);
    coefs = 0.0;
    for (int kx = 0; kx <= order; kx++)
      {
	int maxy = (kx <= 1) ? order : order-kx;

	hv.Range (0, maxy+1) = Trans (jacybs[kx]) * gridy_polx.Col(kx);
	
	for (int ky = 0; ky <= maxy; ky++)
	  coefs(quad2trig[kx*(order+1)+ky]) += hv(ky);
      }
  }





  void H1HighOrderTrigFast :: 
  EvaluateDShapeGrid (const TensorProductIntegrationRule & ir,
		      const FlatVector<double> coefs,
		      FlatMatrix<double> gridvalues,
		      LocalHeap & lh) const
  {
    HeapReset hreset (lh);

    FlatMatrix<double> gridy_polx(ir.GetNIPY(), order+1, lh);
    FlatMatrix<double> griddy_polx(ir.GetNIPY(), order+1, lh);
    
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    GetQuadTrigReordering (quad2trig);

    // transform y
    FlatVector<> hv(order+1, lh);
    for (int kx = 0; kx <= order; kx++)
      {
	int maxy = (kx <= 1) ? order : order-kx;
	for (int ky = 0; ky <= maxy; ky++)
	  hv(ky) = coefs(quad2trig[kx*(order+1)+ky]); 

	gridy_polx.Col(kx) = jacybs[kx] * hv.Range(0,maxy+1);
	griddy_polx.Col(kx) = djacybs[kx] * hv.Range(0,maxy+1);
      }


    // transform x
    FlatMatrix<> gridvalues_dx (ir.GetNIPY(), ir.GetNIPX(), lh);
    FlatMatrix<> gridvalues_dy (ir.GetNIPY(), ir.GetNIPX(), lh);
    /*
    gridvalues_dx = gridy_polx * Trans (dlegxbub);
    gridvalues_dy = griddy_polx * Trans (legxbub);
    */

    LapackMultABt (gridy_polx, dlegxbub, gridvalues_dx);
    LapackMultABt (griddy_polx, legxbub, gridvalues_dy);

    /*
    for (int i = 0; i < ir.GetNIP(); i++)
      {
	Mat<2> thextet, thextet_inv;

	ir.GetIP_DTrig_DQuad(i, thextet);
	CalcInverse (thextet, thextet_inv);

	Vec<2> v1, v2;
	v1(0) = gridvalues_dx(i);
	v1(1) = gridvalues_dy(i);
	
	v2 =  Trans(thextet_inv) * v1;
	
	for (int j = 0; j < 2; j++)
	  gridvalues(i,j) = v2(j);
      }
    */



    int sort[3] = { 0, 1, 2 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 2; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);       
    int isort[3];
    for (int i = 0; i < 3; i++) isort[sort[i]] = i;

    Mat<2> m1, im1;
    Mat<3,2> dlami;
    dlami(0,0) = 0;
    dlami(0,1) = 1;
    dlami(1,0) = 1;
    dlami(1,1) = 0;
    dlami(2,0) = -1;
    dlami(2,1) = -1;
    
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
	m1(j,k) = dlami(isort[j],k);
    
    CalcInverse (m1, im1);

    Vec<2> v1;

    for (int iy = 0, i=0; iy < ir.GetNIPY(); iy++)
      {
	double inv = 1.0 / (1-ir.GetIPY(iy));
	for (int ix = 0; ix < ir.GetNIPX(); ix++, i++)
	  {
	    v1(0) = inv * gridvalues_dx(i);
	    v1(1) = ir.GetIPX(ix) * v1(0) + gridvalues_dy(i);

	    gridvalues.Row(i) = Trans (im1) * v1;
	  }
      }
  }




  void  H1HighOrderTrigFast :: 
  EvaluateDShapeGridTrans (const TensorProductIntegrationRule & ir,
			   const FlatMatrix<double> gridvalues,
			   FlatVector<double> coefs,
			   LocalHeap & lh) const
  {
    HeapReset hr (lh);

    // H1HighOrderFiniteElement :: EvaluateDShapeGridTrans (ir, gridvalues, coefs, lh);
    //   return;



    FlatMatrixFixWidth<2> gridvalues_fw(gridvalues);


    FlatMatrix<double> gridy_polx(ir.GetNIPY(), order+1, lh);
    FlatMatrix<double> griddy_polx(ir.GetNIPY(), order+1, lh);

    FlatMatrix<> gridvalues_dx (ir.GetNIPY(), ir.GetNIPX(), lh);
    FlatMatrix<> gridvalues_dy (ir.GetNIPY(), ir.GetNIPX(), lh);

    
    FlatArray<int> quad2trig((order+1)*(order+1), lh);
    GetQuadTrigReordering (quad2trig);


    /*
    for (int i = 0; i < ir.GetNIP(); i++)
      {
	Mat<2> thextet, thextet_inv;

	ir.GetIP_DTrig_DQuad(i, thextet);
	CalcInverse (thextet, thextet_inv);

	Vec<2> v1, v2;
	for (int j = 0; j < 2; j++)
	  v1(j) = gridvalues(i,j);

	v2 =  thextet_inv * v1;

	gridvalues_dx(i) = v2(0);
	gridvalues_dy(i) = v2(1);
      }
    */


    int sort[3] = { 0, 1, 2 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 2; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);       
    int isort[3];
    for (int i = 0; i < 3; i++) isort[sort[i]] = i;

    Mat<2> m1, im1;
    Mat<3,2> dlami;
    dlami(0,0) = 0;
    dlami(0,1) = 1;
    dlami(1,0) = 1;
    dlami(1,1) = 0;
    dlami(2,0) = -1;
    dlami(2,1) = -1;
    
    for (int j = 0; j < 2; j++)
      for (int k = 0; k < 2; k++)
	m1(j,k) = dlami(isort[j],k);
    
    CalcInverse (m1, im1);

    Vec<2> v1;

    for (int iy = 0, i=0; iy < ir.GetNIPY(); iy++)
      {
	double inv = 1.0 / (1-ir.GetIPY(iy));
	for (int ix = 0; ix < ir.GetNIPX(); ix++, i++)
	  {
	    v1 = im1 * gridvalues_fw.Row(i);

	    v1(0) += ir.GetIPX(ix) * v1(1);
	    v1(0) *= inv;

	    gridvalues_dx(i) = v1(0);
	    gridvalues_dy(i) = v1(1);
	  }
      }









    // transform x
    gridy_polx = gridvalues_dx * dlegxbub;
    griddy_polx = gridvalues_dy * legxbub;


    // transform y
    coefs = 0.0;
    FlatVector<> hv(order+1, lh);
    for (int kx = 0; kx <= order; kx++)
      {
	int maxy = (kx <= 1) ? order : order-kx;

	hv.Range(0, maxy+1) = 
	  Trans (jacybs[kx]) * gridy_polx.Col(kx);
	hv.Range(0, maxy+1) +=
	  Trans (djacybs[kx]) * griddy_polx.Col(kx);

	for (int ky = 0; ky <= maxy; ky++)
	  coefs(quad2trig[kx*(order+1)+ky]) += hv(ky);
      }
  }
}
