#include <solve.hpp>

#ifdef ASTRID
#include <helmholtz_single_exp.hpp>
#endif



namespace ngcomp
{
  template <class SCAL> class ElementByElementMatrix;

  template <class SCAL>
  ElementByElement_BilinearForm<SCAL> :: 
  ElementByElement_BilinearForm (const FESpace & afespace, const string & aname,
					    const Flags & flags)
    : S_BilinearForm<SCAL> (afespace, aname, flags)
  { ; }

  template <class SCAL>
  ElementByElement_BilinearForm<SCAL> :: ~ElementByElement_BilinearForm ()
  { ; }



  
  template <class SCAL>
  void ElementByElement_BilinearForm<SCAL> :: AllocateMatrix ()
  {
    cout << "alloc matrix" << endl;
    const FESpace & fespace = this->fespace;
    this->mats.Append (new ElementByElementMatrix<SCAL> (fespace.GetNDof()));

  }


  template<class SCAL>
  BaseVector * ElementByElement_BilinearForm<SCAL> :: CreateVector() const
  {
    return new VVector<SCAL> (this->fespace.GetNDof());
  }

  template<class SCAL>
  void ElementByElement_BilinearForm<SCAL> :: AddElementMatrix (const ARRAY<int> & dnums1,
						  const ARRAY<int> & dnums2,
						  const FlatMatrix<SCAL> & elmat,
						  bool inner_element, int elnr,
						  LocalHeap & lh)
  {
    /*
    (*testout) << "inner_element = " << inner_element << endl;
    (*testout) << "elnr = " << elnr << endl;
    (*testout) << "elmat = " << endl << elmat << endl;
    (*testout) << "dnums1 = " << endl << dnums1 << endl;
    */

    dynamic_cast<ElementByElementMatrix<SCAL>&> (this->GetMatrix()) . AddElementMatrix (dnums1, dnums2, elmat);

	 
  }




  template class ElementByElement_BilinearForm<double>;
  template class ElementByElement_BilinearForm<Complex>;



template <class SCAL>
void ElementByElementMatrix<SCAL> :: MultAdd (double s, const BaseVector & x, BaseVector & y) const
    {
      int maxs = 0;
      for (int i = 0; i < dnums.Size(); i++)
	maxs = max2 (maxs, dnums[i].Size());

      ArrayMem<SCAL, 100> mem1(maxs), mem2(maxs);
      
      FlatVector<SCAL> vx = dynamic_cast<const S_BaseVector<SCAL> & >(x).FVScal();
      FlatVector<SCAL> vy = dynamic_cast<S_BaseVector<SCAL> & >(y).FVScal();

      for (int i = 0; i < dnums.Size(); i++)
	{
	  FlatArray<int> di (dnums[i]);
	  FlatVector<SCAL> hv1(di.Size(), &mem1[0]);
	  FlatVector<SCAL> hv2(di.Size(), &mem2[0]);
	  
	  for (int j = 0; j < di.Size(); j++)
	    hv1(j) = vx (di[j]);

	  hv2 = elmats[i] * hv1;
	  // elmats[i].Mult (hv1, hv2);
	  hv2 *= s;

	  for (int j = 0; j < dnums[i].Size(); j++)
	    vy (di[j]) += hv2[j];
	}
    }

template <class SCAL>
BaseMatrix *  ElementByElementMatrix<SCAL> :: InverseMatrix ( BitArray * subset ) const
{
  ElementByElementMatrix<SCAL> * invmat = new ElementByElementMatrix<SCAL> (height);
  LocalHeap lh (100000);

  for ( int i = 0; i < dnums.Size(); i++ )
    {
      int nd = dnums[i] . Size();
      FlatMatrix<SCAL> mat(nd, nd, lh);
      ARRAY<int> dnumsarray(nd);
      for ( int j = 0; j < nd; j++ )
	dnumsarray[j] = dnums[i][j];
      mat = elmats[i];

      LapackInverse(mat);

      invmat -> AddElementMatrix(dnumsarray, dnumsarray, mat);
      lh.CleanUp();
    }

  return invmat;
}

template <class SCAL>
void ElementByElementMatrix<SCAL> :: AddElementMatrix (const ARRAY<int> & dnums1,
			   const ARRAY<int> & dnums2,
			   const FlatMatrix<SCAL> & elmat)
    {
      ARRAY<int> used;
      for (int i = 0; i < dnums1.Size(); i++)
	if (dnums1[i] >= 0) used.Append(i);

      int s = used.Size();

      FlatMatrix<SCAL> mat (s, new SCAL[s*s]);
      for (int i = 0; i < s; i++)
	for (int j = 0; j < s; j++)
	  mat(i,j) = elmat(used[i], used[j]);

      FlatArray<int> dn(s, new int[s]);
      for (int i = 0; i < s; i++)
	dn[i] = dnums1[used[i]];

      dnums.Append (dn);

      elmats.SetSize (elmats.Size()+1);
      elmats.Last().AssignMemory (s, s, &mat(0,0));
    }


  template <class SCAL>
  BaseBlockJacobiPrecond * ElementByElementMatrix<SCAL> :: 
  CreateBlockJacobiPrecond (Table<int> & blocks,
			    const BaseVector * constraint, int * paralleloptions) const
  { 
    return 0;//new helmholtz_exp_cpp::BlockJacobiPrecond_ElByEl<SCAL> (*this, blocks );
  }

  template class ElementByElementMatrix<double>;
  template class ElementByElementMatrix<Complex>;

  template <class SCAL>
  HO_BilinearForm<SCAL> :: HO_BilinearForm (const FESpace & afespace, const string & aname,
					    const Flags & flags)
    : S_BilinearForm<SCAL> (afespace, aname, flags)
  { ; }

  template <class SCAL>
  HO_BilinearForm<SCAL> :: ~HO_BilinearForm ()
  { ; }



  
  template <class SCAL>
  void HO_BilinearForm<SCAL> :: AllocateMatrix ()
  {
    cout << "alloc matrix" << endl;
    const FESpace & fespace = this->fespace;

    this->mats.Append (new ElementByElementMatrix<SCAL> (fespace.GetNDof()));


#ifdef ASTRID
	    const helmholtz_exp_cpp::HybridHelmholtzFESpace & hhspace = 
	      dynamic_cast<const helmholtz_exp_cpp::HybridHelmholtzFESpace &> ( this->fespace);
#endif
    cout << "generate inexact schur matrix" << endl;
    // generate inexact schur matrix

    MatrixGraph * graph = 0;

    if (this->ma.GetDimension() == 3)
      {
	int ne = this->ma.GetNE();
	int nfa = this->ma.GetNFaces();
	ARRAY<int> vnums, ednums, elnums, dnums;
	ARRAY<int> cnt(nfa+ne);
	cnt = 0;
	
	for (int i = 0; i < nfa; i++)
	  {
	    this->ma.GetFacePNums (i, vnums);

	    this->ma.GetFaceEdges (i, ednums);
	    this->ma.GetFaceElements (i, elnums);
	
	    // only for shells
	    if ( this->ma.GetElType(elnums[0]) == ET_PRISM 
		 && this->ma.GetElType(elnums[1]) == ET_PRISM )
	      if (vnums.Size() == 3) continue;

	    for (int j = 0; j < vnums.Size(); j++)
	      {
		fespace.GetVertexDofNrs (vnums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) cnt[i]++;
	      }

	    for (int j = 0; j < ednums.Size(); j++)
	      {
		fespace.GetEdgeDofNrs (ednums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) cnt[i]++;
	      }

	    fespace.GetFaceDofNrs (i, dnums);
	    for (int k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1) cnt[i]++;


	    for (int j = 0; j < elnums.Size(); j++)
	      {
#ifdef ASTRID
		ARRAY<int> elfaces;
		this->ma.GetElFaces(elnums[j], elfaces);
		int pos = elfaces.Pos(i);
		if ( hhspace . IsSlave(elnums[j], pos )) continue; 
#endif
		fespace.GetWireBasketDofNrs (elnums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) cnt[i]++;
	      }
	  }

	/*
	  for (int i = 0; i < ne; i++)
	  {
	  fespace.GetExternalDofNrs (i, dnums);
	  for (int j = 0; j < dnums.Size(); j++)
	  if (dnums[j] != -1)
	  cnt[nfa+i]++;
	  }
	*/

	Table<int> fa2dof(cnt);

	for (int i = 0; i < nfa; i++)
	  {
	    this->ma.GetFacePNums (i, vnums);

	    this->ma.GetFaceElements (i, elnums);

	    // only for shells
	    if ( this->ma.GetElType(elnums[0]) == ET_PRISM 
		 && this->ma.GetElType(elnums[1]) == ET_PRISM )
	      if (vnums.Size() == 3) continue;

	    this->ma.GetFaceEdges (i, ednums);
	    int ii = 0;
	    for (int j = 0; j < vnums.Size(); j++)
	      {
		fespace.GetVertexDofNrs (vnums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) fa2dof[i][ii++] = dnums[k];
	      }
	    for (int j = 0; j < ednums.Size(); j++)
	      {
		fespace.GetEdgeDofNrs (ednums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) fa2dof[i][ii++] = dnums[k];
	      }

	    fespace.GetFaceDofNrs (i, dnums);

	    for (int k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1) fa2dof[i][ii++] = dnums[k];


	    for (int j = 0; j < elnums.Size(); j++)
	      {
#ifdef ASTRID
		ARRAY<int> elfaces;
		this->ma.GetElFaces(elnums[j], elfaces);
		int pos = elfaces.Pos(i);
		if ( hhspace . IsSlave(elnums[j], pos )) continue; 
#endif
		fespace.GetWireBasketDofNrs (elnums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) fa2dof[i][ii++] = dnums[k];
	      }
	  }

	cout << "has graph table" << endl;
// 	(*testout) << "face graph table: " << endl << fa2dof << endl;
	/*    
	      for (int i = 0; i < ne; i++)
	      {
	      fespace.GetExternalDofNrs (i, dnums);
	      int ii = 0;
	      for (int j = 0; j < dnums.Size(); j++)
	      if (dnums[j] != -1)
	      fa2dof[nfa+i][ii++] = dnums[j];
	      }
	*/
	// (*testout) << "fa2dof = " << endl << fa2dof << endl;
    

	graph = new MatrixGraph (fespace.GetNDof(), fa2dof, 0);
      }
    
    else
      
      {
	int nel = this->ma.GetNE();
	int ned = this->ma.GetNEdges();
	int nsel = this->ma.GetNSE();
	ARRAY<int> vnums, ednums, elnums, dnums;
	ARRAY<int> cnt(2*ned);//+nsel);
	cnt = 0;

	for (int i = 0; i < ned; i++)
	  {
	    int vi[2];
	    this->ma.GetEdgePNums (i, vi[0], vi[1]);

	    for (int j = 0; j < 2; j++)
	      {
		fespace.GetVertexDofNrs (vi[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) cnt[i]++;
	      }

	    fespace.GetEdgeDofNrs (i, dnums);
	    for (int k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1) cnt[i]++;

	    this->ma.GetEdgeElements (i, elnums);
	    for (int j = 0; j < elnums.Size(); j++)
	      {
#ifdef ASTRID
		ARRAY<int> eledges;
		this->ma.GetElEdges(elnums[j], eledges);
		int pos = eledges.Pos(i);
		if ( hhspace . IsSlave(elnums[j], pos )) continue; 
#endif
		fespace.GetWireBasketDofNrs (elnums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) cnt[i]++;
	      }
	  }


	Table<int> fa2dof(cnt);

	for (int i = 0; i < ned; i++)
	  {
	    int ii = 0;
	    int vi[2];
	    this->ma.GetEdgePNums (i, vi[0], vi[1]);

	    for (int j = 0; j < 2; j++)
	      {
		fespace.GetVertexDofNrs (vi[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1)  fa2dof[i][ii++] = dnums[k];
	      }

	    fespace.GetEdgeDofNrs (i, dnums);
	    for (int k = 0; k < dnums.Size(); k++)
	      if (dnums[k] != -1) fa2dof[i][ii++] = dnums[k];

	    this->ma.GetEdgeElements (i, elnums);
	    for (int j = 0; j < elnums.Size(); j++)
	      {
#ifdef ASTRID
		ARRAY<int> eledges;
		this->ma.GetElEdges(elnums[j], eledges);
		int pos = eledges.Pos(i);
		if ( hhspace . IsSlave(elnums[j], pos )) continue; 
#endif
		fespace.GetWireBasketDofNrs (elnums[j], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) fa2dof[i][ii++] = dnums[k];
	      }
	  }


	cout << "has graph table" << endl;
// 	(*testout) << "face graph table: " << endl << fa2dof << endl;
	/*    
	      for (int i = 0; i < ne; i++)
	      {
	      fespace.GetExternalDofNrs (i, dnums);
	      int ii = 0;
	      for (int j = 0; j < dnums.Size(); j++)
	      if (dnums[j] != -1)
	      fa2dof[nfa+i][ii++] = dnums[j];
	      }
	*/
	// (*testout) << "fa2dof = " << endl << fa2dof << endl;
    

	graph = new MatrixGraph (fespace.GetNDof(), fa2dof, 0);
      }


    inexact_schur = new SparseMatrix<SCAL> (*graph, 1);
    delete graph;

    inexact_schur -> AsVector() = 0.0;

//     (*testout) << "matrix = " << endl << *inexact_schur << endl;
  }


  template<class SCAL>
  BaseVector * HO_BilinearForm<SCAL> :: CreateVector() const
  {
    return new VVector<SCAL> (this->fespace.GetNDof());
  }

  template<class SCAL>
  void HO_BilinearForm<SCAL> :: AddElementMatrix (const ARRAY<int> & dnums1,
						  const ARRAY<int> & dnums2,
						  const FlatMatrix<SCAL> & elmat,
						  bool inner_element, int elnr,
						  LocalHeap & lh)
  {
    /*
    (*testout) << "inner_element = " << inner_element << endl;
    (*testout) << "elnr = " << elnr << endl;
    (*testout) << "elmat = " << endl << elmat << endl;
    (*testout) << "dnums1 = " << endl << dnums1 << endl;
    */

    dynamic_cast<ElementByElementMatrix<SCAL>&> (this->GetMatrix()) . AddElementMatrix (dnums1, dnums2, elmat);

#ifdef ASTRID
    FlatMatrix<SCAL> elmat_is ( elmat.Height(), lh );
    elmat_is = 0;
#endif
#ifdef ASTRID
	    const helmholtz_exp_cpp::HybridHelmholtzFESpace & hhspace = 
	      dynamic_cast<const helmholtz_exp_cpp::HybridHelmholtzFESpace &> ( this->fespace);
#endif

    if (inner_element)
      {
	// build Schur complements for faces


	if (this->ma.GetDimension() == 3)
	  {

	    ARRAY<int> useddofs, usedi, unusedi;
	    ARRAY<int> vnums, ednums, fanums, dnums;	

	    this->ma.GetElPNums (elnr, vnums);
	    this->ma.GetElEdges (elnr, ednums);
	    this->ma.GetElFaces (elnr, fanums);

	    for (int i = 0; i < fanums.Size(); i++)
#ifdef ASTRID
	      if ( !hhspace.IsSlave(elnr, i) )
#endif
	      {
		useddofs.SetSize (0);
		usedi.SetSize (0);
		unusedi.SetSize (0);

		this->ma.GetFacePNums (fanums[i], vnums);

		// only for shells
		if ( this->ma.GetElType(elnr) == ET_PRISM )
		  if (vnums.Size() == 3) continue;
		
		// this->ma.GetElPNums (elnr, vnums);   // all vertices

		this->fespace.GetWireBasketDofNrs (elnr, useddofs);

		this->ma.GetFaceEdges (fanums[i], ednums);
	    
		int ii = 0;

		for (int j = 0; j < vnums.Size(); j++)
		  {
		    this->fespace.GetVertexDofNrs (vnums[j], dnums);
		    for (int k = 0; k < dnums.Size(); k++)
		      if (dnums[k] != -1) useddofs.Append (dnums[k]);
		  }
		for (int j = 0; j < ednums.Size(); j++)
		  {
		    this->fespace.GetEdgeDofNrs (ednums[j], dnums);
		    for (int k = 0; k < dnums.Size(); k++)
		      if (dnums[k] != -1) useddofs.Append (dnums[k]);
		  }
		this->fespace.GetFaceDofNrs (fanums[i], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) useddofs.Append (dnums[k]);

		// (*testout) << "useddofs = " << useddofs << endl;
	    
		for (int j = 0; j < dnums1.Size(); j++)
		  if (dnums1[j] != -1)
		    {
		      bool has = 0;
		      for (int k = 0; k < useddofs.Size(); k++)
			if (useddofs[k] == dnums1[j])
			  has = 1;
		  
		      if (has) usedi.Append (j);
		      else unusedi.Append (j);
		    }

		int su = usedi.Size();
		int sunu = unusedi.Size();

		FlatMatrix<SCAL> a(su, su, lh);
		FlatMatrix<SCAL> b(su, sunu, lh);
		FlatMatrix<SCAL> c(su, sunu, lh);
		FlatMatrix<SCAL> idc(su, sunu, lh);
		FlatMatrix<SCAL> d(sunu, sunu, lh);
			    
		for (int k = 0; k < su; k++)
		  for (int l = 0; l < su; l++)
		    a(k,l) = elmat(usedi[k], usedi[l]);
			    
		for (int k = 0; k < su; k++)
		  for (int l = 0; l < sunu; l++)
		    {
		      b(k,l) = elmat(usedi[k], unusedi[l]);
		      c(k,l) = elmat(unusedi[l], usedi[k]);
		    }
	    
		for (int k = 0; k < sunu; k++)
		  for (int l = 0; l < sunu; l++)
		    d(k,l) = elmat(unusedi[k], unusedi[l]);

		/*
		  (*testout) << "usedi = " << endl << usedi << endl;
		  (*testout) << "unusedi = " << endl << unusedi << endl;
		  (*testout) << "a = " << endl << a << endl;
		  (*testout) << "b = " << endl << b << endl;
		  (*testout) << "c = " << endl << c << endl;
		  (*testout) << "d = " << endl << d << endl;
		*/


#ifdef LAPACK
		if ( sunu > 0 )
		  {
		    LapackInverse (d);
		    LapackMultABt (c, d, idc);
		    LapackMultAddABt (b, idc, -1, a);
		  }
#else
		FlatMatrix<SCAL> invd(sunu, sunu, lh);
		CalcInverse (d, invd);
		d = invd;
		idc = c * Trans (invd);
		a -= b * Trans (idc);
#endif
	    
#ifdef ASTRID
		// testout schur complement for each face
		if ( this->printelmat )
		  (*testout) << "schur mat = " << endl << a << endl;
	  
		// calc this->elmat_ev for face schur complement
		if ( this->elmat_ev )
		  {
		    Vector<Complex> lami(a.Height());
		    Matrix<SCAL> evecs(a.Height());
		    FlatMatrix<SCAL> elmat_save(a.Height(), a.Width(), lh);
		    elmat_save = a;
#ifdef LAPACK
		    LapackEigenValues (elmat_save, lami, evecs);
		    (*testout) << "lam of face schurmat = " 
			       << endl << lami << endl << "evecs: " << endl << evecs << endl;
#endif
		    // CalcEigenSystem (a, lami, evecs);
		  }

		// find sum schur complement
		if ( this->printelmat || this->elmat_ev )
		  {
		    for (int k = 0; k < usedi.Size(); k++)
		      for (int l = 0; l < usedi.Size(); l++)
			elmat_is(usedi[k], usedi[l]) += a(k,l);
		  }
#endif		 
		for (int k = 0; k < usedi.Size(); k++)
		  for (int l = 0; l < usedi.Size(); l++)
		      (*inexact_schur)(dnums1[usedi[k]], dnums1[usedi[l]]) += a(k,l);

	      }
	  }
	else
	  {
	    ARRAY<int> useddofs, usedi, unusedi;
	    ARRAY<int> dnums, ednums;

	    this->ma.GetElEdges (elnr, ednums);


	    for (int i = 0; i < ednums.Size(); i++)
#ifdef ASTRID
	      if ( !hhspace.IsSlave(elnr, i) )
#endif
	      {
		useddofs.SetSize (0);
		usedi.SetSize (0);
		unusedi.SetSize (0);

		int vi[2];
		this->ma.GetEdgePNums (ednums[i], vi[0], vi[1]);
		this->fespace.GetWireBasketDofNrs (elnr, useddofs);


		int ii = 0;
		for (int j = 0; j < 2; j++)
		  {
		    this->fespace.GetVertexDofNrs (vi[j], dnums);
		    for (int k = 0; k < dnums.Size(); k++)
		      if (dnums[k] != -1) useddofs.Append (dnums[k]);
		  }
// 		*testout << "useddofs1 " << useddofs << endl;

		this->fespace.GetEdgeDofNrs (ednums[i], dnums);
		for (int k = 0; k < dnums.Size(); k++)
		  if (dnums[k] != -1) useddofs.Append (dnums[k]);
// 		*testout << "useddofs2 " << useddofs << endl;
	    
		for (int j = 0; j < dnums1.Size(); j++)
		  if (dnums1[j] != -1)
		    {
		      bool has = 0;
		      for (int k = 0; k < useddofs.Size(); k++)
			if (useddofs[k] == dnums1[j])
			  has = 1;
		  
		      if (has) usedi.Append (j);
		      else unusedi.Append (j);
		    }

		int su = usedi.Size();
		int sunu = unusedi.Size();

		FlatMatrix<SCAL> a(su, su, lh);
		FlatMatrix<SCAL> b(su, sunu, lh);
		FlatMatrix<SCAL> c(su, sunu, lh);
		FlatMatrix<SCAL> idc(su, sunu, lh);
		FlatMatrix<SCAL> d(sunu, sunu, lh);
			    
		for (int k = 0; k < su; k++)
		  for (int l = 0; l < su; l++)
		    a(k,l) = elmat(usedi[k], usedi[l]);
			    
		for (int k = 0; k < su; k++)
		  for (int l = 0; l < sunu; l++)
		    {
		      b(k,l) = elmat(usedi[k], unusedi[l]);
		      c(k,l) = elmat(unusedi[l], usedi[k]);
		    }
	    
		for (int k = 0; k < sunu; k++)
		  for (int l = 0; l < sunu; l++)
		    d(k,l) = elmat(unusedi[k], unusedi[l]);


		/*
		  (*testout) << "usedi = " << endl << usedi << endl;
		  (*testout) << "unusedi = " << endl << unusedi << endl;
		  (*testout) << "a = " << endl << a << endl;
		  (*testout) << "b = " << endl << b << endl;
		  (*testout) << "c = " << endl << c << endl;
		  (*testout) << "d = " << endl << d << endl;
		*/


#ifdef LAPACK
		if ( sunu > 0 )
		  {
		    LapackInverse (d);
		    LapackMultABt (c, d, idc);
		    LapackMultAddABt (b, idc, -1, a);
		  }
#else
		FlatMatrix<SCAL> invd(sunu, sunu, lh);
		CalcInverse (d, invd);
		d = invd;
		idc = c * Trans (invd);
		a -= b * Trans (idc);
#endif

#ifdef ASTRID
		// testout schur complement for each face
		if ( this->printelmat )
		  (*testout) << "schur mat = " << endl << a << endl;
	  
		// calc this->elmat_ev for face schur complement
		if ( this->elmat_ev )
		  {
		    Vector<Complex> lami(a.Height());
		    Matrix<SCAL> evecs(a.Height());
		    FlatMatrix<SCAL> elmat_save(a.Height(), a.Width(), lh);
		    elmat_save = a;
#ifdef LAPACK
		    LapackEigenValues (elmat_save, lami, evecs);
		    (*testout) << "lam of face schurmat = " 
			       << endl << lami << endl << "evecs: " << endl << evecs << endl;
#endif
		    // CalcEigenSystem (a, lami, evecs);
		  }

		// find sum schur complement
		if ( this->printelmat || this->elmat_ev )
		  {
		    for (int k = 0; k < usedi.Size(); k++)
		      for (int l = 0; l < usedi.Size(); l++)
			elmat_is(usedi[k], usedi[l]) += a(k,l);
		  }
#endif		 

		for (int k = 0; k < usedi.Size(); k++)
		  for (int l = 0; l < usedi.Size(); l++)
		    (*inexact_schur)(dnums1[usedi[k]], dnums1[usedi[l]]) += a(k,l);

	      }
	  }
      }
    else
      {
	for (int i = 0; i < dnums1.Size(); i++)
	  for (int j = 0; j < dnums2.Size(); j++)
	    if (dnums1[i] != -1 && dnums2[j] != -1)
	      (*inexact_schur)(dnums1[i], dnums2[j]) += elmat(i, j);
      }

#ifdef ASTRID
    // eigenvalues of element schur complement
    if ( this->elmat_ev )
      {
	Vector<Complex> lami(elmat_is.Height());
	Matrix<SCAL> evecs(elmat_is.Height());
	FlatMatrix<SCAL> elmat_save(elmat_is.Height(), elmat_is.Width(), lh);
	elmat_save = elmat_is;
#ifdef LAPACK
	LapackEigenValues (elmat_save, lami, evecs);
	(*testout) << "lam of element schurmat = " 
		   << endl << lami << endl << "evecs: " << endl << evecs << endl;
#endif
      }

    // testout sum schur matrix
    if ( this->printelmat )
      (*testout) << "element schur matrix = " << endl << elmat_is << endl;

#endif
		 
  }





  template class HO_BilinearForm<double>;
  template class HO_BilinearForm<Complex>;





  template <class SCAL>
  WireBasketPreconditioner<SCAL> ::
  WireBasketPreconditioner (const PDE * pde, const Flags & aflags, const string aname)
    : Preconditioner (pde, aflags, aname)
    {
      cout << "\ncreate wire-basket precond" << endl;
      bfa = dynamic_cast<const HO_BilinearForm<SCAL>*>(pde->GetBilinearForm (aflags.GetStringFlag ("bilinearform", NULL)));
      inversetype = flags.GetStringFlag("inverse", "sparsecholesky");
      smoothingtype = int(flags.GetNumFlag("blocktype", -1));
    }


  template <class SCAL>
  void WireBasketPreconditioner<SCAL> ::
  Update ()
    {
      cout << "update wirebasket" << endl;

      const SparseMatrix<SCAL> & mat = bfa -> InexactSchur();

      //      const SparseMatrix<SCAL> & mat = 
      //	dynamic_cast<const SparseMatrix<SCAL>& > (bfa -> GetMatrix());

//       (*testout) << "type = " << typeid(mat).name() << endl;
//       (*testout) << "mat = " << mat << endl;
      
      mat.SetInverseType(inversetype);
      if ( smoothingtype == -1 )
	{
	  pre = mat.InverseMatrix(); // const_cast<SparseMatrix<SCAL>*> (&mat); // 
	}
      else
	{
	  ARRAY<int> cnt(ma.GetNE());
	  const ElementByElementMatrix<SCAL> & elbyelmat =
	    dynamic_cast<const ElementByElementMatrix<SCAL>&> (bfa -> GetMatrix() );
// 	  const helmholtz_exp_cpp::HybridHelmholtzFESpace & hhspace = 
// 	    dynamic_cast<const helmholtz_exp_cpp::HybridHelmholtzFESpace & > 
// 	    (bfa -> GetFESpace() );
	  
	  Table<int> * blocktable = bfa->GetFESpace().CreateSmoothingBlocks(smoothingtype);
	  
	  pre = mat.  CreateBlockJacobiPrecond (*blocktable);
	}
      if (test) Test();
    }


  template class WireBasketPreconditioner<double>;
  template class WireBasketPreconditioner<Complex>;



    
  // register FESpaces
  namespace
#ifdef MACOS
  ho_biform_cpp
#endif
  {
   
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetPreconditionerClasses().AddPreconditioner ("wirebasket", WireBasketPreconditioner<double>::Create);
      GetPreconditionerClasses().AddPreconditioner ("wirebasket_complex", WireBasketPreconditioner<Complex>::Create);
    }
    
    Init init;
  }


}
