#include <comp.hpp>

#include <multigrid.hpp>
using namespace ngmg;

#include <solve.hpp>

// int perm[] = { 0, 1, 2};
// int perm[] = { 0, 2, 1 };
// int perm[] = { 1, 2, 0};
int perm[] = { 1, 0, 2 };
      // int perm[] = { 2, 0, 1};
      // int perm[] = { 2, 1, 0};


namespace ngcomp
{
  using namespace ngcomp;
  

  class RefElementASMMatrix : public BaseMatrix
  {
    const H1HighOrderFESpace & fes;
    const BilinearForm & bfa;
    const BilinearForm & bfac;

    Matrix<> inv_element;
    Matrix<> inv_face;
    Matrix<> inv_edge;
    double inv_augmented_vert;

    ARRAY<Matrix<>*> extensions;
    ARRAY<Matrix<>*> edge2face_extensions;
    ARRAY<Matrix<>*> edge2element_extensions;

    ARRAY<Vector<>*> vert2edge_extensions;
    ARRAY<Vector<>*> vert2face_extensions;

    
    BaseMatrix * coarse_inverse;
    int ncoarse;
    int p;

    ARRAY<double> fac_element, fac_face, fac_edge, fac_vert;
    /*
    ARRAY<double> n_elements_on_edge;
    ARRAY<double> n_elements_on_face;
    ARRAY<double> n_elements_on_vert;
    */

  public:
    RefElementASMMatrix (const H1HighOrderFESpace & afes,
			 const BilinearForm & abfa,
			 const BilinearForm & abfac)
      : fes(afes), bfa(abfa), bfac(abfac), 
	extensions(4), edge2face_extensions(3), edge2element_extensions(6),
	vert2edge_extensions(2), vert2face_extensions(3)

    {
      static int maintimer = NgProfiler::CreateTimer ("RefElementASM::Setup");
      NgProfiler::RegionTimer reg(maintimer);

      cout << "prepare RefElementASM Preconditioner" << endl;

      const MeshAccess & ma = fes.GetMeshAccess ();

      LocalHeap lh(300000000);
      
      p = fes.GetOrder();

      H1HighOrderTetFast tet(fes.GetOrder());
      ArrayMem<int,12> vnums; 
      for (int i = 0; i < 4; i++) vnums[i] = i;
      tet.SetVertexNumbers (vnums, lh);

      ConstantCoefficientFunction coef(1);
      ARRAY<CoefficientFunction*> coeffs(1);
      coeffs[0] = &coef;

      BilinearFormIntegrator & laplace = *GetIntegrators().CreateBFI("fastlaplace",3, coeffs);

      ElementTransformation eltrans;
      fes.GetMeshAccess().GetElementTransformation (0, eltrans, lh);
      
      FlatMatrix<double> elmat_lap;
      laplace.AssembleElementMatrix (tet, eltrans, elmat_lap, lh);

      Matrix<double> elmat_lap2(elmat_lap.Height());
      elmat_lap2 = elmat_lap;

      cout << "has element matrix" << endl;


      int augmented = 0; // fes.GetAugmented();

      int n = elmat_lap.Height();
      int n_vert = 4;
      if (augmented == 1) n_vert = 8;
      int n_edge = (p-1);
      int n_inner = (p-1)*(p-2)*(p-3) / 6;
      int n_face = (p-1)*(p-2)/2;

      int first_edge = n_vert;
      int first_face = first_edge+6*n_edge;
      int first_inner = first_face+4*n_face;


      Matrix<> element_inner(n_inner);
      inv_element.SetSize (n_inner);

      for (int i = 0; i < n_inner; i++)
	for (int j = 0; j < n_inner; j++)
	  element_inner(i,j) = elmat_lap(first_inner+i, first_inner+j);

      // CalcInverse (element_inner, inv_element);
      inv_element = element_inner;
      if (p >= 4)
	LapackInverse (inv_element);
      
      cout << "has inverse" << endl;

      int sort[4] = { 0, 1, 2, 3 };
      for (int i = 0; i < 4; i++)
	for (int j = 0; j < 3; j++)
	  if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);

      int isort[4];
      for (int i = 0; i < 4; i++) isort[sort[i]] = i;
      


      Matrix<> el_if(n_inner, n_face);
      Matrix<> schur_face(n_face);

      cout << "face extension " << flush;

      for (int fi = 0; fi < 4; fi++)
	{
	  extensions[fi] = new Matrix<> (n_inner, n_face);
	  int first_face_fi = first_face+sort[fi]*n_face;

	  for (int i = 0; i < n_inner; i++)
	    for (int j = 0; j < n_face; j++)
	      el_if(i, j) = elmat_lap(first_inner+i, first_face_fi+j);

	  LapackMultAB (inv_element, el_if, *extensions[fi]);
	  *extensions[fi] *= -1;

	  schur_face = Trans (el_if) * (*extensions[fi]);

	  for (int i = 0; i < n_face; i++)
	    for (int j = 0; j < n_face; j++)
	      schur_face(i, j) += elmat_lap(first_face_fi+i, first_face_fi+j);

	  cout << fi << " " << flush;
	}

      cout << endl;

      inv_face.SetSize (n_face);
      if (p >= 3)
	CalcInverse (schur_face, inv_face);







      const EDGE * edges = ElementTopology::GetEdges (ET_TET);

      // prepare edge2element
      Matrix<> el_ei(n_inner, n_edge);
      for (int ei = 0; ei < 6; ei++)
	{
	  int vn1 = isort[edges[ei][0]], vn2 = isort[edges[ei][1]];
	  int enr = ElementTopology::GetEdgeNr (ET_TET, vn1, vn2);

	  edge2element_extensions[enr] = new Matrix<> (n_inner, n_edge);
	  
	  for (int i = 0; i < n_inner; i++)
	    for (int j = 0; j < n_edge; j++)
	      el_ei(i, j) = elmat_lap(first_inner+i, first_edge+ei*n_edge+j);

	  *edge2element_extensions[enr] = (-inv_element) * el_ei;
	}


      cout << "edge to face extension" << endl;

      // precompute edge2face extensions

      BitArray used(n);
      used.Clear();
      int e2fsort[3];

      const FACE * faces = ElementTopology::GetFaces (ET_TET);
      // first face is 1,2,3

      int sortf[3] = { 1, 2, 3, };
      for (int i = 0; i < 3; i++)
	for (int j = 0; j < 2; j++)
	  if (vnums[sortf[j]] > vnums[sortf[j+1]]) swap (sortf[j], sortf[j+1]);
      

      for (int i = 0; i < n_face; i++)
	used.Set(first_face+i);

      int cnt = 0;
      for (int i = 0; i < 6; i++)
	if (edges[i][0] != 0 && edges[i][1] != 0)
	  {
	    for (int j = 0; j < n_edge; j++)
	      used.Set(first_edge+n_edge*i+j);

	    for (int j = 0; j < 3; j++)
	      if (edges[i][0] != sortf[j] && edges[i][1] != sortf[j])
		e2fsort[cnt++] = j;
	  }
	else
	  for (int j = 0; j < n_edge; j++)
	    elmat_lap(first_edge+n_edge*i+j,first_edge+n_edge*i+j) += 1e26;

      for (int i = 0; i < first_edge; i++)
	elmat_lap(i,i) += 1e26;

      Matrix<> faceedges(n_face+3*n_edge);
      CalcSchurComplement (elmat_lap, faceedges, used, lh);


      // (*testout) << "faceedge-schur = " << endl << faceedges << endl;

      Matrix<> el_fe(n_face, n_edge);
      Matrix<> inv_face(n_face);
      Matrix<> schur_edge(n_edge);

      for (int i = 0; i < n_face; i++)
	for (int j = 0; j < n_face; j++)
	  inv_face(i,j) = faceedges(3*n_edge+i, 3*n_edge+j);
      if (p >= 3)
	LapackInverse (inv_face);      


      for (int ei = 0; ei < 3; ei++)
	{
	  edge2face_extensions[e2fsort[ei]] = new Matrix<> (n_face, n_edge);
	  
	  for (int i = 0; i < n_face; i++)
	    for (int j = 0; j < n_edge; j++)
	      el_fe(i, j) = faceedges(3*n_edge+i, ei*n_edge+j);

	  *edge2face_extensions[e2fsort[ei]] = (-inv_face) * el_fe;
	  schur_edge = Trans (el_fe) * (*edge2face_extensions[e2fsort[ei]]);

	  for (int i = 0; i < n_edge; i++)
	    for (int j = 0; j < n_edge; j++)
	      schur_edge(i, j) += faceedges(ei*n_edge+i, ei*n_edge+j);
	  
	  cout << "has edge matrix " << ei << endl;
	  // (*testout) << "e2f-extension = " << endl << *edge2face_extensions[e2fsort[ei]] << endl;
	  // (*testout) << "schur_edge = " << schur_edge << endl;
	}
      
      inv_edge.SetSize (n_edge);
      CalcInverse (schur_edge, inv_edge);

      //      (*testout) << "inv_edge = " << endl << inv_edge << endl;





      cout << "vert 2 edge extension" << endl;
      
      for (int j = 0; j < 2; j++)
	{
	  int vei[2] = { edges[0][0], edges[0][1] };
	  if (vei[0] > vei[1]) swap (vei[0], vei[1]);

	  used.Clear();
	  used.Set (vei[j]);
	  for (int k = 0; k < n_edge; k++)
	    used.Set (first_edge+k);
	  
	  Matrix<> schur_edge_v(n_edge+1);
	  Matrix<> schur_edge_v1(n_edge);

	  elmat_lap = elmat_lap2;
	  
	  for (int k = 0; k < 4; k++)
	    if (k != vei[j])
	      elmat_lap(k,k) += 1e26;

	  for (int k = 0; k < 6; k++)
	    if (vei[j] != edges[k][0] && vei[j] != edges[k][1])
	      for (int l = 0; l < n_edge; l++)
		elmat_lap(first_edge+k*n_edge+l, first_edge+k*n_edge+l) += 1e26;
			  
	  for (int k = 0; k < 4; k++)
	    if (vei[j] != faces[k][0] && vei[j] != faces[k][1]  && vei[j] != faces[k][2])
	      for (int l = 0; l < n_face; l++)
		elmat_lap(first_face+k*n_face+l, first_face+k*n_face+l) += 1e26;
			  

	  CalcSchurComplement (elmat_lap, schur_edge_v, used, lh);
	  
	  for (int i = 0; i < n_edge; i++)
	    for (int j = 0; j < n_edge; j++)
	      schur_edge_v1(i,j) = schur_edge_v(i+1, j+1);
	  
	  LapackInverse (schur_edge_v1);      

	  Vector<> ext(n_edge);
	  for (int k = 0; k < n_edge; k++)
	    ext(k) = schur_edge_v(0, k+1);
	  
	  vert2edge_extensions[j] = new Vector<> (n_edge);
	  *vert2edge_extensions[j] = -schur_edge_v1 * ext;

	  inv_augmented_vert = 1.0 /
	    (schur_edge_v(0,0) + InnerProduct (ext, *vert2edge_extensions[j]));

	  // cout << "vert2edge = " << endl << *vert2edge_extensions[j] << endl;
	  // cout << "schur = " << inv_augmented_vert << endl;
	}

	    

      {
      cout << "vert 2 face extension" << endl;

      if (p >= 3)
      for (int i = 0; i < 3; i++)
	{
	  used.Clear();

	  int sort[3] = { 0, 1, 2 };
	  for (int j = 0; j < 3; j++)
	    for (int k = 0; k < 2; k++)
	      if (faces[0][sort[k]] > faces[0][sort[k+1]]) swap (sort[k], sort[k+1]);
      
	  int vi = faces[0][sort[i]];

	  used.Set (vi);
	  for (int j = 0; j < n_face; j++)
	    used.Set (first_face+j);
	  
	  Matrix<> schur_face_v(n_face+1);
	  Matrix<> schur_face_v1(n_face);

	  elmat_lap = elmat_lap2;


	  for (int k = 0; k < 4; k++)
	    if (k != vi)
	      elmat_lap(k,k) += 1e26;

	  for (int k = 0; k < 6; k++)
	    if (vi != edges[k][0] && vi != edges[k][1])
	      for (int l = 0; l < n_edge; l++)
		elmat_lap(first_edge+k*n_edge+l, first_edge+k*n_edge+l) += 1e26;

	  for (int k = 0; k < 4; k++)
	    if (vi != faces[k][0] && vi != faces[k][1]  && vi != faces[k][2])
	      for (int l = 0; l < n_face; l++)
		elmat_lap(first_face+k*n_face+l, first_face+k*n_face+l) += 1e26;


	  CalcSchurComplement (elmat_lap, schur_face_v, used, lh);

	  for (int j = 0; j < n_face; j++)
	    for (int k = 0; k < n_face; k++)
	      schur_face_v1(j,k) = schur_face_v(j+1, k+1);
      
	  LapackInverse (schur_face_v1);      

	  Vector<> ext(n_face);
	  for (int k = 0; k < n_face; k++)
	    ext(k) = schur_face_v(0, k+1);
	  
	  vert2face_extensions[i] = new Vector<> (n_face);
	  *vert2face_extensions[i] = -schur_face_v1 * ext;

	  // cout << "face ext = " << endl << *vert2face_extensions[i] << endl;
	}
      }
	    

      if (&bfac)
	{
	  coarse_inverse = 
	    dynamic_cast<const BaseSparseMatrix&> 
	    (bfac.GetMatrix())  . InverseMatrix();
	  
	  ncoarse = coarse_inverse->Height();
	}
      else
	{
	  coarse_inverse = 0;
	  ncoarse = 0;
	}
      // (*testout) << "coarse mat = " << endl << (*coarse_inverse) << endl;







      ARRAY<int> enums, fnums;

      int ne = ma.GetNE();
      int nv = ma.GetNV();
      int nfa = ma.GetNFaces();
      int ned = ma.GetNEdges();
      

      /*
      n_elements_on_face.SetSize (nfa);
      n_elements_on_face = 0;
      n_elements_on_edge.SetSize (ned);
      n_elements_on_edge = 0;
      n_elements_on_vert.SetSize (nv);
      n_elements_on_vert = 0;

      for (int i = 0; i < ne; i++)
	{
	  ma.GetElFaces (i, fnums);
	  double vol = ma.GetElementVolume (i);
	  for (int j = 0; j < fnums.Size(); j++)
	    n_elements_on_face[fnums[j]]++;
	  ma.GetElEdges (i, enums);

	  for (int j = 0; j < enums.Size(); j++)
	    n_elements_on_edge[enums[j]]++;
	  ma.GetElVertices (i, enums);
	  for (int j = 0; j < enums.Size(); j++)
	    n_elements_on_vert[enums[j]]++;
	}
      */

      fac_element.SetSize (ne);
      fac_face.SetSize (nfa);
      fac_edge.SetSize (ned);
      fac_vert.SetSize (nv);

      fac_vert = 0.0;
      fac_edge = 0.0;
      fac_face = 0.0;
      fac_element = 0.0;

      for (int i = 0; i < ne; i++)
	{
	  double fac = pow (ma.ElementVolume (i), 1.0/3.0);
	  fac_element[i] = fac; 

	  ma.GetElFaces (i, fnums);
	  for (int j = 0; j < fnums.Size(); j++)
	    fac_face[fnums[j]] += fac;

	  ma.GetElEdges (i, enums);
	  for (int j = 0; j < enums.Size(); j++)
	    fac_edge[enums[j]] += fac;

	  ma.GetElVertices (i, enums);
	  for (int j = 0; j < enums.Size(); j++)
	    fac_vert[enums[j]] += fac;
	}

      for (int i = 0; i < ne; i++)
	if (fac_element[i])
	  fac_element[i] = 1.0 / fac_element[i];
      for (int i = 0; i < nfa; i++)
	if (fac_face[i])
	  fac_face[i] = 1.0 / fac_face[i];
      for (int i = 0; i < ned; i++)
	if (fac_edge[i])
	  fac_edge[i] = 1.0 / fac_edge[i];
      for (int i = 0; i < nv; i++)
	if (fac_vert[i])
	  fac_vert[i] = 1.0 / fac_vert[i];
      /*
      fac_face = 0.0;
      fac_edge = 0.0;
      fac_vert = 0.0;
      */
    }


    virtual int VHeight() const { return fes.GetNDof(); }
    virtual int VWidth() const { return fes.GetNDof(); }

    virtual BaseVector * CreateVector () const
    {
      return new VVector<double> (fes.GetNDof());
    }




    virtual void MultAdd (double s, const BaseVector & x, BaseVector & y) const
    {
      static int maintimer = NgProfiler::CreateTimer ("RefElementASM::Mult");
      static int celltimer = NgProfiler::CreateTimer ("RefElementASM::Mult, cells");
      static int resttimer = NgProfiler::CreateTimer ("RefElementASM::Mult, Restriction");

      NgProfiler::RegionTimer reg(maintimer);

      FlatVector<double> vx = x.FVDouble();
      FlatVector<double> vy = y.FVDouble();


      const MeshAccess & ma = fes.GetMeshAccess();
      ARRAY<int> vnums, enums, fnums;

      int nv = ma.GetNV();
      int ne = ma.GetNE();
      int nfa = ma.GetNFaces();
      int ned = ma.GetNEdges();

      int nd_ed = inv_edge.Height();
      int nd_fa = inv_face.Height();
      int nd_el = inv_element.Height();



      ARRAY<double> hmem(nd_fa*10);


      NgProfiler::StartTimer (celltimer);
      NgProfiler::AddFlops (celltimer, ne * inv_element.Width()*inv_element.Width());
      
      FlatMatrix<double> elx (ne, inv_element.Width(), &vx[fes.GetFirstElementDof(0)]);
      FlatMatrix<double> ely (ne, inv_element.Width(), &vy[fes.GetFirstElementDof(0)]);
      
      LapackMultAddABt (elx, inv_element, s, ely);
      
      for (int i = 0; i < ne; i++)
	ely.Row(i) *= fac_element[i];

      NgProfiler::StopTimer(celltimer);


      
      Vector<> hv(vx.Size());
      Vector<> hv2(vx.Size());

      Vector<> augmented_vert(nv);
      Vector<> augmented_vert2(nv);

      augmented_vert = 0.0;
      augmented_vert2 = 0.0;

      hv = 0;
      
      // Restriction
      {
	NgProfiler::RegionTimer reg(resttimer);
	
	
	int firste = fes.GetFirstEdgeDof(0);
	int firstf = fes.GetFirstFaceDof(0);
	int firsti = fes.GetFirstElementDof(0);
	int h = vx.Size();
	
	hv = vx;
	

	for (int fi = 0; fi < 4; fi++)
	  for (int i1 = 0; i1 < ne; i1 += 10)
	    {
	      int bs = min (10, ne-i1);
	      
	      FlatMatrix<double> matx(bs, inv_element.Height(), &hv[fes.GetFirstElementDof(i1)]);
	      FlatMatrix<double> maty(bs, inv_face.Height(), &hmem[0]);
	      LapackMultAB (matx, *extensions[fi], maty);
	      
	      for (int i2 = 0; i2 < bs; i2++)
		{
		  int i = i1+i2;
		  ma.GetElVertices(i, vnums);
		  ma.GetElFaces(i, fnums);
		  
		  int sort[4] = { 0, 1, 2, 3 };
		  for (int i = 0; i < 4; i++)
		    for (int j = 0; j < 3; j++)
		      if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
		  
		  int first_f = fes.GetFirstFaceDof(fnums[sort[fi]]);
		  for (int j = 0; j < inv_face.Height(); j++)
		    hv(first_f+j) += maty(i2, j);
		}
	    }
      }
      
      static int rest2timer = NgProfiler::CreateTimer ("RefElementASM::Mult, Restriction, 2");
      {
	NgProfiler::RegionTimer reg(rest2timer);
	
	// edge 2 element restriction
	for (int i = 0; i < ne; i++)
	  {
	    ma.GetElVertices(i, vnums);
	    ma.GetElEdges(i, enums);
	    
	    int first = fes.GetFirstElementDof(i);
	    int next = fes.GetFirstElementDof(i+1);
	    
	    FlatVector<double> elx  (next-first, &hv[first]);
	    
	    int sort[4] = { 0, 1, 2, 3 };
	    for (int i = 0; i < 4; i++)
	      for (int j = 0; j < 3; j++)
		if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	    
	    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
	    
	    for (int ei = 0; ei < 6; ei++)
	      {
		int v1 = sort[edges[ei][0]], v2 = sort[edges[ei][1]];
		int inv_ei = ElementTopology::GetEdgeNr (ET_TET, v1, v2);
		
		int first_e = fes.GetFirstEdgeDof(enums[inv_ei]);
		int next_e = fes.GetFirstEdgeDof(enums[inv_ei]+1);
		
		FlatVector<double> edge_x  (next_e-first_e, &hv[first_e]);
		edge_x += Trans (*edge2element_extensions[ei]) * elx;
	      }
	  }
	








      augmented_vert = 0.0;

      // vertex 2 face restriction
      if (p >= 3)
      for (int i = 0; i < nfa; i++)
	{
	  ma.GetFacePNums(i, vnums);
	  int sort[3] = { 0, 1, 2 };
	  for (int j = 0; j < 3; j++)
	    for (int k = 0; k < 2; k++)
	      if (vnums[sort[k]] > vnums[sort[k+1]]) swap (sort[k], sort[k+1]);

	  int first = fes.GetFirstFaceDof(i);
	  int next = fes.GetFirstFaceDof(i+1);
	  FlatVector<double> elx  (next-first, &hv[first]);
	  
	  for (int j = 0; j < 3; j++)
	    augmented_vert(vnums[sort[j]]) += InnerProduct (*vert2face_extensions[j], elx);
	}


      // vertex 2 edge restriction

      for (int i = 0; i < ned; i++)
	{
	  int vi[2];
	  ma.GetEdgePNums(i, vi[0], vi[1]);
	  int sort[2] = { 0, 1 };
	  if (vi[0] > vi[1]) swap (sort[0], sort[1]);

	  int first = fes.GetFirstEdgeDof(i);
	  int next = fes.GetFirstEdgeDof(i+1);
	  FlatVector<double> elx  (next-first, &hv[first]);
	  
	  for (int j = 0; j < 2; j++)
	    augmented_vert(vi[sort[j]]) += InnerProduct (*vert2edge_extensions[j], elx);
	}








	// edge 2 face restriction
	for (int i = 0; i < nfa; i++)
	  {
	    ma.GetFacePNums(i, vnums);
	    ma.GetFaceEdges(i, enums);
	    
	    int first = fes.GetFirstFaceDof(i);
	    int next = fes.GetFirstFaceDof(i+1);
	    FlatVector<double> elx  (next-first, &hv[first]);
	    
	    for (int j = 0; j < 3; j++)
	      for (int k = 0; k < 2; k++)
		if (vnums[k] > vnums[k+1]) swap (vnums[k], vnums[k+1]);
	    
	    for (int ei = 0; ei < 3; ei++)
	      {
		int ev1, ev2;
		ma.GetEdgePNums (enums[ei], ev1, ev2);
		
		int first_e = fes.GetFirstEdgeDof(enums[ei]);
		int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
		FlatVector<double> edge_x  (next_e-first_e, &hv[first_e]);
		
		for (int j = 0; j < 3; j++)
		  if (ev1 != vnums[j] && ev2 != vnums[j])
		    edge_x += Trans (*edge2face_extensions[j]) * elx;
	      }
	  }
      }





      for (int i = 0; i < nv; i++)
	augmented_vert(i) += hv(i);





      for (int i = 0; i < nv; i++)
	if (fes.IsDirichletVertex (i))
	  {
	    hv(i) = 0;
	    augmented_vert(i) = 0;
	  }

      for (int i = 0; i < ned; i++)
	if (fes.IsDirichletEdge (i))
	  {
	    int first = fes.GetFirstEdgeDof (i);
	    int next = fes.GetFirstEdgeDof (i+1);
	    for (int j = first; j < next; j++)
	      hv(j) = 0;
	  }

      for (int i = 0; i < nfa; i++)
	if (fes.IsDirichletFace (i))
	  {
	    int first = fes.GetFirstFaceDof (i);
	    int next = fes.GetFirstFaceDof (i+1);
	    for (int j = first; j < next; j++)
	      hv(j) = 0;
	  }





      for (int i = 0; i < nv; i++)
	augmented_vert2(i) = 0.5 * inv_augmented_vert * fac_vert[i] * augmented_vert(i);
      // augmented_vert2(i) = 0.5 * inv_augmented_vert / n_elements_on_vert[i] * augmented_vert(i);

      hv2 = 0;

      FlatMatrix<double> elx_fa (nfa, nd_fa, &hv[fes.GetFirstFaceDof(0)]);
      FlatMatrix<double> ely_fa (nfa, nd_fa, &hv2[fes.GetFirstFaceDof(0)]);
      LapackMultAddABt (elx_fa, inv_face, 1, ely_fa);

      for (int i = 0; i < nfa; i++)
	ely_fa.Row(i) *= fac_face[i];
      /*
	{
	  // double fac = 1.0 / n_elements_on_face[i]; 
	  double fac = fac_face[i];
	  for (int j = 0; j < nd_fa; j++)
	    ely_fa(i,j) *= fac;
	}
      */
      FlatMatrix<double> elx_ed (ned, nd_ed, &hv[fes.GetFirstEdgeDof(0)]);
      FlatMatrix<double> ely_ed (ned, nd_ed, &hv2[fes.GetFirstEdgeDof(0)]);
      LapackMultAddABt (elx_ed, inv_edge, 1, ely_ed);

      for (int i = 0; i < ned; i++)
	{
	  // double fac = 1.0 / n_elements_on_edge[i]; 
	  double fac = fac_edge[i];
	  for (int j = 0; j < nd_ed; j++)
	    ely_ed(i,j) *= fac;
	}


      for (int i = 0; i < nv; i++)
	hv2(i) += augmented_vert2(i);

      




      // edge 2 face extension
      for (int i = 0; i < nfa; i++)
	{
	  ma.GetFacePNums(i, vnums);
	  ma.GetFaceEdges(i, enums);

	  int first = fes.GetFirstFaceDof(i);
	  int next = fes.GetFirstFaceDof(i+1);
	  FlatVector<double> elx  (next-first, &hv2[first]);

	  for (int j = 0; j < 3; j++)
	    for (int k = 0; k < 2; k++)
	      if (vnums[k] > vnums[k+1]) swap (vnums[k], vnums[k+1]);
	  
	  for (int ei = 0; ei < 3; ei++)
	    {
	      int ev1, ev2;
	      ma.GetEdgePNums (enums[ei], ev1, ev2);

	      
	      int first_e = fes.GetFirstEdgeDof(enums[ei]);
	      int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
	      FlatVector<double> edge_x  (next_e-first_e, &hv2[first_e]);

	      for (int j = 0; j < 3; j++)
		if (ev1 != vnums[j] && ev2 != vnums[j])
		  elx += *edge2face_extensions[j] * edge_x;
	    }
	}






      // vertex 2 edge extension
      for (int i = 0; i < ned; i++)
	{
	  int vi[2];
	  ma.GetEdgePNums(i, vi[0], vi[1]);
	  int sort[2] = { 0, 1 };
	  if (vi[0] > vi[1]) swap (sort[0], sort[1]);

	  int first = fes.GetFirstEdgeDof(i);
	  int next = fes.GetFirstEdgeDof(i+1);
	  FlatVector<double> elx  (next-first, &hv2[first]);
	  
	  for (int j = 0; j < 2; j++)
	    elx += augmented_vert2(vi[sort[j]]) * *vert2edge_extensions[j];
	}

      // vert 2 face
      if (p >= 3)
      for (int i = 0; i < nfa; i++)
	{
	  ma.GetFacePNums(i, vnums);
	  int sort[3] = { 0, 1, 2 };
	  for (int j = 0; j < 3; j++)
	    for (int k = 0; k < 2; k++)
	      if (vnums[sort[k]] > vnums[sort[k+1]]) swap (sort[k], sort[k+1]);

	  int first = fes.GetFirstFaceDof(i);
	  int next = fes.GetFirstFaceDof(i+1);
	  FlatVector<double> elx  (next-first, &hv2[first]);
	  
	  for (int j = 0; j < 3; j++)
	    elx += augmented_vert2(vnums[sort[j]]) * *vert2face_extensions[j];
	}      







      // edge 2 element extension
      for (int i = 0; i < ne; i++)
	{
	  ma.GetElVertices(i, vnums);
	  ma.GetElEdges(i, enums);

	  int first = fes.GetFirstElementDof(i);
	  int next = fes.GetFirstElementDof(i+1);

	  FlatVector<double> elx  (next-first, &hv2[first]);


	  int sort[4] = { 0, 1, 2, 3 };
	  for (int i = 0; i < 4; i++)
	    for (int j = 0; j < 3; j++)
	      if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	  
	  int isort[4];
	  for (int i = 0; i < 4; i++) isort[sort[i]] = i;
	  
	  
	  const EDGE * edges = ElementTopology::GetEdges (ET_TET);
	  for (int ei = 0; ei < 6; ei++)
	    {
	      int first_e = fes.GetFirstEdgeDof(enums[ei]);
	      int next_e = fes.GetFirstEdgeDof(enums[ei]+1);
	      FlatVector<double> edge_x  (next_e-first_e, &hv2[first_e]);

	      int v1 = isort[edges[ei][0]], v2 = isort[edges[ei][1]];
	      int enr = ElementTopology::GetEdgeNr (ET_TET, v1, v2);

	      elx += *edge2element_extensions[enr]  * edge_x;
	    }
	}




      // face 2 element extension
      for (int fi = 0; fi < 4; fi++)
	for (int i1 = 0; i1 < ne; i1 += 10)
	  {
	    int bs = min (10, ne-i1);

	    FlatMatrix<double> matx(bs, inv_element.Height(), &hv2[fes.GetFirstElementDof(i1)]);
	    FlatMatrix<double> maty(bs, inv_face.Height(), &hmem[0]);
	    
	    for (int i2 = 0; i2 < bs; i2++)
	      {
		ma.GetElVertices(i1+i2, vnums);
		ma.GetElFaces(i1+i2, fnums);
	    
		int sort[4] = { 0, 1, 2, 3 };
		for (int i = 0; i < 4; i++)
		  for (int j = 0; j < 3; j++)
		    if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
		
		int first_f = fes.GetFirstFaceDof(fnums[sort[fi]]);
		for (int j = 0; j < inv_face.Height(); j++)
		  maty(i2, j) = hv2(first_f+j);
	      }

	    LapackMultAddABt (maty, *extensions[fi], 1, matx);
	  }


      vy += s * hv2;



      //      for (int i = 0; i < firste; i++)
      //	vy(i) += s * diag(i) * vx(i);


      if (coarse_inverse)
	{
	  VFlatVector<> coarse_x (ncoarse, &hv2(0));
	  VFlatVector<> coarse_hv (ncoarse, &hv(0));
	  VFlatVector<> coarse_y (ncoarse, &vy(0));
	  
	  hv2 = vx;
	  
	  for (int i = 0; i < nv; i++)
	    if (fes.IsDirichletVertex (i))
	      hv2(i) = 0;
	  
	  coarse_hv = (*coarse_inverse) * coarse_x;
	  
	  for (int i = 0; i < nv; i++)
	    if (fes.IsDirichletVertex (i))
	      hv(i) = 0;
	  
	  coarse_y += s * coarse_hv;
	}

      // (*testout) << "vy = " << endl << vy << endl;
    }


    virtual ostream & Print (ostream & ost) const
    {
      ost << "RefElementASMMatrix:" << endl;
      return ost;
    }
  };













  /* ************************************************************************** */
  /* ************************************************************************** */
  /* ************************************************************************** */
  /*                                                                            */
  /*                                                                            */
  /*                2D                                                          */
  /*                                                                            */
  /*                                                                            */
  /* ************************************************************************** */
  /* ************************************************************************** */
  /* ************************************************************************** */
  /* ************************************************************************** */
  /* ************************************************************************** */





  class RefElementASMMatrix2d : public BaseMatrix
  {
    const H1HighOrderFESpace & fes;
    const BilinearForm & bfa;
    const BilinearForm & bfac;

    Matrix<> inv_element;
    Matrix<> inv_edge;
    double inv_augmented_vert;

    ARRAY<Matrix<>*> extensions;
    ARRAY<Vector<>*> vert2edge_extensions;
    ARRAY<Vector<>*> vert_extensions;
    
    BaseMatrix * coarse_inverse;
    int ncoarse;
    int p;

    ARRAY<double> fac_element, fac_edge, fac_vert;

  public:
    RefElementASMMatrix2d (const H1HighOrderFESpace & afes,
			   const BilinearForm & abfa,
			   const BilinearForm & abfac)
      : fes(afes), bfa(abfa), bfac(abfac), 
	extensions(3), 
	vert2edge_extensions(2), vert_extensions(3)

    {
      cout << "prepare RefElementASM Preconditioner" << endl;

      const MeshAccess & ma = fes.GetMeshAccess ();

      LocalHeap lh(100000000);
      
      p = fes.GetOrder();

      H1HighOrderTrigFast trig(p);
      ArrayMem<int,12> vnums; 

      ma.GetElVertices (0, vnums);
      trig.SetVertexNumbers (vnums, lh);

      ConstantCoefficientFunction coef(1);
      ARRAY<CoefficientFunction*> coeffs(1);
      coeffs[0] = &coef;

      BilinearFormIntegrator & laplace = *GetIntegrators().CreateBFI("laplace",2, coeffs);
      laplace.SetFastIntegration (1);

      ElementTransformation eltrans;
      fes.GetMeshAccess().GetElementTransformation (0, eltrans, lh);
      
      FlatMatrix<double> elmat_lap;
      laplace.AssembleElementMatrix (trig, eltrans, elmat_lap, lh);

      Matrix<double> elmat_lap2(elmat_lap.Height());
      elmat_lap2 = elmat_lap;

      cout << "has element matrix" << endl;


      int n = elmat_lap.Height();
      int n_vert = 3;
      int n_edge = (p-1);
      int n_inner = (p-1)*(p-2) / 2;

      int first_edge = n_vert;
      int first_inner = first_edge+3*n_edge;

      Matrix<> element_inner(n_inner);
      inv_element.SetSize (n_inner);

      for (int i = 0; i < n_inner; i++)
	for (int j = 0; j < n_inner; j++)
	  element_inner(i,j) = elmat_lap(first_inner+i, first_inner+j);

      inv_element = element_inner;
      if (p >= 3)
	LapackInverse (inv_element);
      
      cout << "has inverse" << endl;


      int sort[3] = { 0, 1, 2 };
      for (int i = 0; i < 3; i++)
	for (int j = 0; j < 2; j++)
	  if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
      int perm[] = { 1, 0, 2 };

      Matrix<> el_ie(n_inner, n_edge);
      Matrix<> schur_edge(n_edge);

      cout << "edge extension " << flush;

      for (int edgenr = 0; edgenr < 3; edgenr++)
	{
	  extensions[edgenr] = new Matrix<> (n_inner, n_edge);
	  int first_edgei = first_edge+perm[sort[edgenr]]*n_edge;

	  for (int i = 0; i < n_inner; i++)
	    for (int j = 0; j < n_edge; j++)
	      el_ie(i, j) = elmat_lap(first_inner+i, first_edgei+j);

	  LapackMultAB (inv_element, el_ie, *extensions[edgenr]);
	  *extensions[edgenr] *= -1;

	  (*testout) << "face ext = " << endl << *extensions[edgenr] << endl;
	  cout << edgenr << " " << flush;



	  if (edgenr == 2)
	    {
	      schur_edge = Trans (el_ie) * (*extensions[edgenr]);
	      
	      for (int i = 0; i < n_edge; i++)
		for (int j = 0; j < n_edge; j++)
		  schur_edge(i, j) += elmat_lap(first_edgei+i, first_edgei+j);
	      
	      inv_edge.SetSize (n_edge);
	      if (p >= 3)
		CalcInverse (schur_edge, inv_edge);
	    }    
	}

      cout << endl;






      if (&bfac)
	{
	  coarse_inverse = 
	    dynamic_cast<const BaseSparseMatrix&> 
	    (bfac.GetMatrix())  . InverseMatrix();
	  
	  ncoarse = coarse_inverse->Height();
	}
      else
	{
	  coarse_inverse = 0;
	  ncoarse = 0;
	}




      ARRAY<int> enums, fnums;

      int ne = ma.GetNE();
      int nv = ma.GetNV();
      int nfa = ma.GetNFaces();
      int ned = ma.GetNEdges();
      

      fac_element.SetSize (ne);
      fac_edge.SetSize (ned);
      fac_vert.SetSize (nv);

      fac_vert = 0.0;
      fac_edge = 0.0;
      fac_element = 0.0;

      for (int i = 0; i < ne; i++)
	{
	  double h = pow (ma.ElementVolume (i), 1.0/2.0);  // element size
	  double fac = 1;   // in 3D .. h
	  fac_element[i] = fac; 

	  ma.GetElEdges (i, enums);
	  for (int j = 0; j < enums.Size(); j++)
	    fac_edge[enums[j]] += fac;

	  ma.GetElVertices (i, enums);
	  for (int j = 0; j < enums.Size(); j++)
	    fac_vert[enums[j]] += fac;
	}


      for (int i = 0; i < ne; i++)
	if (fac_element[i])
	  fac_element[i] = 1.0 / fac_element[i];
      for (int i = 0; i < ned; i++)
	if (fac_edge[i])
	  fac_edge[i] = 1.0 / fac_edge[i];
      for (int i = 0; i < nv; i++)
	if (fac_vert[i])
	  fac_vert[i] = 1.0 / fac_vert[i];

    }


    virtual int VHeight() const { return fes.GetNDof(); }
    virtual int VWidth() const { return fes.GetNDof(); }

    virtual BaseVector * CreateVector () const
    {
      return new VVector<double> (fes.GetNDof());
    }




#undef TEXTBOOK
#ifdef TEXTBOOK

    virtual void MultAdd (double s, const BaseVector & x, BaseVector & y) const
    {
      // y = y + s M x


      FlatVector<double> vx = x.FVDouble();
      FlatVector<double> vy = y.FVDouble();


      const MeshAccess & ma = fes.GetMeshAccess();
      ARRAY<int> vnums, enums;

      int nv = ma.GetNV();
      int ne = ma.GetNE();
      int ned = ma.GetNEdges();

      int nd_ed = inv_edge.Height();
      int nd_el = inv_element.Height();


      
      Vector<> hv(vx.Size());
      Vector<> hv2(vx.Size());
      
      // Restriction
	
      hv = vx;

      for (int fi = 0; fi < 3; fi++)
	for (int i = 0; i < ne; i++)
	  {
	    ma.GetElVertices(i, vnums);
	    ma.GetElEdges(i, enums);
	    swap (enums[0], enums[1]);
	    
	    int sort[3] = { 0, 1, 2 };
	    for (int k = 0; k < 3; k++)
	      for (int j = 0; j < 2; j++)
		if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	    
	    int ednr = enums[sort[fi]];
	    
	    FlatVector<> elx = hv.Range (fes.GetFirstElementDof(i),
					 fes.GetFirstElementDof(i+1));

	    FlatVector<> edy = hv.Range (fes.GetFirstEdgeDof(ednr),
					 fes.GetFirstEdgeDof(ednr+1));
	    
	    edy += Trans (*extensions[fi]) * elx;
	  }
      
      
      hv2 = 0;

      for (int i = 0; i < ne; i++)
	{
	  FlatVector<> elx = hv.Range (fes.GetFirstElementDof(i),
				       fes.GetFirstElementDof(i+1));
	  FlatVector<> ely = hv2.Range (fes.GetFirstElementDof(i),
					fes.GetFirstElementDof(i+1));

	  ely += fac_element[i] * (inv_element * elx);
	}


      
      for (int i = 0; i < ned; i++)
	{
	  FlatVector<> elx = hv.Range (fes.GetFirstEdgeDof(i),
				       fes.GetFirstEdgeDof(i+1));
	  FlatVector<> ely = hv2.Range (fes.GetFirstEdgeDof(i),
					fes.GetFirstEdgeDof(i+1));
	  
	  ely += fac_edge[i] * (inv_edge * elx);
	}


      // edge 2 element extension

      for (int fi = 0; fi < 3; fi++)
	for (int i = 0; i < ne; i++)
	  {
	    ma.GetElVertices(i, vnums);
	    ma.GetElEdges(i, enums);

	    swap (enums[0], enums[1]);
	    int sort[3] = { 0, 1, 2 };
	    for (int k = 0; k < 3; k++)
	      for (int j = 0; j < 2; j++)
		if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	    
	    int ednr = enums[sort[fi]];


	    FlatVector<double> ely = hv2.Range (fes.GetFirstElementDof(i),
						fes.GetFirstElementDof(i+1));

	    FlatVector<double> edx = hv2.Range (fes.GetFirstEdgeDof(ednr),
						fes.GetFirstEdgeDof(ednr+1));

	    ely +=  *extensions[fi] * edx;
	  }


      vy += s * hv2;



      if (coarse_inverse)
	{
	  VFlatVector<> coarse_x (ncoarse, &hv2(0));
	  VFlatVector<> coarse_hv (ncoarse, &hv(0));
	  VFlatVector<> coarse_y (ncoarse, &vy(0));
	  
	  hv2 = vx;
	  
	  /*
	  for (int i = 0; i < nv; i++)
	    if (fes.IsDirichletVertex (i))
	      hv2(i) = 0;
	  */

	  coarse_hv = (*coarse_inverse) * coarse_x;
	  /*	  
	  for (int i = 0; i < nv; i++)
	    if (fes.IsDirichletVertex (i))
	      hv(i) = 0;
	  */
	  coarse_y += s * coarse_hv;
	}
    }

#endif







#define OPTIMIZED
#ifdef OPTIMIZED
    virtual void MultAdd (double s, const BaseVector & x, BaseVector & y) const
    {
      // y = y + s M x


      FlatVector<double> vx = x.FVDouble();
      FlatVector<double> vy = y.FVDouble();


      const MeshAccess & ma = fes.GetMeshAccess();
      ARRAY<int> vnums, enums;

      int nv = ma.GetNV();
      int ne = ma.GetNE();
      int ned = ma.GetNEdges();

      int nd_ed = inv_edge.Height();
      int nd_el = inv_element.Height();


      ARRAY<double> hmem(nd_ed*10);

      FlatMatrix<double> elx (ne, nd_el, &vx[fes.GetFirstElementDof(0)]);
      FlatMatrix<double> ely (ne, nd_el, &vy[fes.GetFirstElementDof(0)]);
      
      LapackMultAddABt (elx, inv_element, s, ely);
      
      for (int i = 0; i < ne; i++)
	ely.Row(i) *= fac_element[i];

      
      Vector<> hv(vx.Size());
      Vector<> hv2(vx.Size());

      Vector<> augmented_vert(nv);
      Vector<> augmented_vert2(nv);

      augmented_vert = 0.0;
      augmented_vert2 = 0.0;

      hv = vx;


      
      // Restriction
	
      hv = vx;
      
      for (int fi = 0; fi < 3; fi++)
	for (int i = 0; i < ne; i++)
	  {
	    FlatVector<> elx = hv.Range (fes.GetFirstElementDof(i),
					 fes.GetFirstElementDof(i+1));
	    
	    ma.GetElVertices(i, vnums);
	    ma.GetElEdges(i, enums);
	    swap (enums[0], enums[1]);
	    
	    int sort[3] = { 0, 1, 2 };
	    for (int k = 0; k < 3; k++)
	      for (int j = 0; j < 2; j++)
		if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);
	    
	    int ednr = enums[sort[fi]];
	    
	    FlatVector<> edy = hv.Range (fes.GetFirstEdgeDof(ednr),
					 fes.GetFirstEdgeDof(ednr+1));
	    
	    edy += Trans (*extensions[fi]) * elx;
	  }

      hv2 = 0;


      FlatMatrix<double> elx_ed (ned, nd_ed, &hv[fes.GetFirstEdgeDof(0)]);
      FlatMatrix<double> ely_ed (ned, nd_ed, &hv2[fes.GetFirstEdgeDof(0)]);
      LapackMultAddABt (elx_ed, inv_edge, 1, ely_ed);

      for (int i = 0; i < ned; i++)
	ely_ed.Row(i) *= fac_edge[i];


      // edge 2 element extension

      for (int fi = 0; fi < 3; fi++)
	for (int i1 = 0; i1 < ne; i1 += 10)
	  {
	    int bs = min (10, ne-i1);

	    FlatMatrix<double> matx(bs, nd_el, &hv2[fes.GetFirstElementDof(i1)]);
	    FlatMatrix<double> maty(bs, nd_ed, &hmem[0]);
	    
	    for (int i2 = 0; i2 < bs; i2++)
	      {
		ma.GetElVertices(i1+i2, vnums);
		ma.GetElEdges(i1+i2, enums);
		swap (enums[0], enums[1]);

		int sort[3] = { 0, 1, 2 };
		for (int i = 0; i < 3; i++)
		  for (int j = 0; j < 2; j++)
		    if (vnums[sort[j]] > vnums[sort[j+1]]) swap (sort[j], sort[j+1]);

		int first_e = fes.GetFirstEdgeDof(enums[sort[fi]]);
		for (int j = 0; j < nd_ed; j++)
		  maty(i2, j) = hv2(first_e+j);
	      }

	    LapackMultAddABt (maty, *extensions[fi], 1, matx);
	  }

      vy += s * hv2;







      if (coarse_inverse)
	{
	  VFlatVector<> coarse_x (ncoarse, &hv2(0));
	  VFlatVector<> coarse_hv (ncoarse, &hv(0));
	  VFlatVector<> coarse_y (ncoarse, &vy(0));
	  
	  hv2 = vx;
	  
	  /*
	  for (int i = 0; i < nv; i++)
	    if (fes.IsDirichletVertex (i))
	      hv2(i) = 0;
	  */

	  coarse_hv = (*coarse_inverse) * coarse_x;
	  /*	  
	  for (int i = 0; i < nv; i++)
	    if (fes.IsDirichletVertex (i))
	      hv(i) = 0;
	  */
	  coarse_y += s * coarse_hv;
	}
    }
#endif














    virtual ostream & Print (ostream & ost) const
    {
      ost << "RefElementASMMatrix2d:" << endl;
      return ost;
    }
  };




































  
  
  
  class RefElementPreconditioner : public Preconditioner
  {
  private:
    const BilinearForm * bfa;
    const BilinearForm * bfac;
    const H1HighOrderFESpace * fes;
    BaseMatrix * inverse;
  public:

    RefElementPreconditioner (const PDE & pde, const Flags & flags)
      : Preconditioner(&pde,flags)
    {
      bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
      bfac = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformc", ""), 1);
      fes = dynamic_cast<const H1HighOrderFESpace*> (pde.GetFESpace (flags.GetStringFlag ("fespace", "")));
      inverse = NULL;
    }
    
    virtual ~RefElementPreconditioner()
    {
      delete inverse;
    }
    
    static Preconditioner * Create (const PDE & pde, const Flags & flags)
    {
      return new RefElementPreconditioner (pde, flags);
    }

    virtual void Update ()
    {
      cout << "called update" << endl;
      delete inverse;
      if (bfa->GetMeshAccess().GetDimension() == 3)
	inverse = new RefElementASMMatrix (*fes, *bfa, *bfac);
      else
	inverse = new RefElementASMMatrix2d (*fes, *bfa, *bfac);

      if (test) Test();
    }

    virtual const BaseMatrix & GetMatrix() const
    {
      return *inverse;
    }

    virtual const BaseMatrix & GetAMatrix() const
    {
      return bfa->GetMatrix(); 
    }
    
    virtual const char * ClassName() const
    { 
      return "ReferenceElement Preconditioner"; 
    }
  };


  



  
  // register FESpaces
  namespace
#ifdef MACOS
  refelement_asm_cpp
#endif
 {
    
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetPreconditionerClasses().AddPreconditioner ("refelementasm", RefElementPreconditioner::Create);
    }
    
    Init init;
  }
  
}
