/*********************************************************************/
/* File:   h1hofe.cpp                                                */
/* Author: Start                                                      */
/* Date:   6. Feb. 2003                                              */
/*********************************************************************/

 
#include <fem.hpp>
namespace ngfem
{
  using namespace ngfem;


#ifdef FAST


  template <class T_ORTHOPOL>
  void H1HighOrderTet<T_ORTHOPOL> :: CalcShape (const IntegrationPoint & ip, 
						FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  
  template <class T_ORTHOPOL>
  void H1HighOrderTet<T_ORTHOPOL> :: CalcDShape (const IntegrationPoint & ip, 
						 FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  }


  template<typename T_ORTHOPOL> 
  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderTet<T_ORTHOPOL> :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx lami[4] = { x, y, z, 1-x-y-z };

    Tx lamisum[4] = { 0, 0, 0, 0 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
	if (vnums[j] > vnums[i]) lamisum[i] += lami[j];

    Tx lamiadd[4];
    for (int i = 0; i < 4; i++)
      lamiadd[i] = lami[i] + lamisum[i];

    ArrayMem<Tx, 20> polx(order+1), poly(order+1), polz(order+1); 
    ArrayMem<Tx, 20> power(order+1);

    // vertex shapes
    for (int i = 0; i < 4; i++)
      sds[i] = lami[i];
    int ii = 4; 

    if (augmented == 1)
      for (int i = 0; i < 4; i++)
	sds[ii++] = T_VERTEXSHAPES::Calc (order, lami[i]);
    else
      if (augmented == 2)
	for (int i = 0; i < 4; i++)
	  {
	    ArrayMem<Tx,20> pol(order+1);
	    LegendrePolynomial (order-1, -1+2*lami[i], pol);	
	    for (int j = 1; j < order; j++)
	      sds[ii++] = pol[j] * lami[i];
	  }
 
    // edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];

	  int nde = order_edge[i]-1;
	  ScaledLegendrePolynomial (order_edge[i]-2, lami[es]-lamisum[es], lamiadd[es], hp);
	  for (int j = 0; j < nde; j++)
	    hp[j] *= 4*lami[ee]*lami[es];

	  Tx fac = 1;
	  for (int j = nde-1; j >= 0; j--)
	    {
	      hp[j] *= fac;
	      fac *= lamiadd[es];
	    }
	  ii += nde;
	}



    // face dofs
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
      if (order_face[i] >= 3)
	{
	  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
	  
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
	  
	  int vop = 6 - fav[0] - fav[1] - fav[2];  	
	  
	  Tx lx = lami[fav[2]]-lami[fav[1]];
	  Tx ly = lami[fav[0]];
	  Tx lz = lami[vop];

	  int n = order_face[i];
	  ArrayMem<Tx, 20> polx(n+1), poly(n+1);
    
	  ScaledLegendrePolynomial (n-3, lami[fav[1]]-lamisum[fav[1]], lamiadd[fav[1]], polx);
	  // ScaledLegendrePolynomial (n-3, lami[fav[0]]-lamisum[fav[0]], lamiadd[fav[0]], poly);
	  ScaledJacobiPolynomial (n-3, lami[fav[0]]-lamisum[fav[0]], lamiadd[fav[0]], order_face[i], 0, poly);


	  Tx  bub = (1-lx-ly-lz) * (1+lx-ly-lz)*ly; 
	  /*
	  int cntsm = 0;
	  for (int j = 0; j < 4; j++)
	    if (vnums[j] < vnums[vop]) cntsm++;

	  if (cntsm == 1)   // alternative for 1 face p^3 possible ?
	    {
	      Tx fac = 1;
	      for (int j = 0; j < order+1; j++)
		{
		  power[j] = fac;
		  fac *= lamiadd[vop];    
		}
	    }
	  else
	    {
	      Tx fac = 1;
	      for (int j = 0; j < order+1; j++)
		{
		  power[j] = fac;
		  fac *= lamiadd[fav[1]];    
		}
	    }
	  */


	  Tx fac = 1;
	  for (int j = 0; j < order+1; j++)
	    {
	      power[j] = fac;
	      fac *= lamiadd[fav[1]];    
	    }

	  /*
	  // alternative
	  Tx fac = 1;
	  for (int j = 0; j < order+1; j++)
	    {
	      power[j] = fac;
	      fac *= lamisum[fav[0]];    
	    }
	  */
	  for (int j2 = 0; j2 <= n-3; j2++)
	    for (int i2 = 0; i2 <= n-3-j2; i2++)
	      sds[ii++] = bub * polx[i2] * poly[j2] * power[n-3-i2-j2];
	}


    if (order_inner >= 4)
      {
	Tx * psds = &sds[ii];
	int sort[4] = { 0, 1, 2, 3 };
	for (int i = 0; i < 4; i++)
	  for (int j = 0; j < 3; j++)
	    if (vnums[sort[j]] < vnums[sort[j+1]])
	      swap (sort[j], sort[j+1]);



	Tx lx = lami[sort[1]]-lami[sort[0]];
	Tx ly = lami[sort[2]];
	Tx lz = lami[sort[3]];

	int n = order_inner;

	ArrayMem<Tx, 20> polx(n+1), poly(n+1), polz(n+1);


	ScaledLegendrePolynomial (n-4, lx, 1-ly-lz, polx);
	ScaledJacobiPolynomial (n-4, 2*ly - (1-lz) , (1-lz), n, 0, poly);
	JacobiPolynomial (n-4, 2*lz-1, n, 0, polz);

	/*
	ScaledJacobiPolynomial (n-4, lx, 1-ly-lz, 2, 2, polx);
	ScaledJacobiPolynomial (n-4, 2*ly - (1-lz) , (1-lz), n, 0, poly);
	JacobiPolynomial (n-4, 2*lz-1, n, 0, polz);
	*/

	Tx bub = (1-lx-ly-lz) * (1+lx-ly-lz) * ly * lz ; 
	
	Tx fac = 1;
	for (int j = 0; j < order+1; j++)
	  {
	    power[j] = fac;
	    fac *= 1-lz;
	  }

	for (int i = 0; i <= n-4; i++)
	  for (int j = 0; j <= n-4-i; j++)
	    for (int k = 0; k <= n-4-i-j; k++)
	      sds[ii++] = bub * polx[k] * poly[j] * polz[i]; //  * power[n-4-i-j-k];
	

	// ii += T_INNERSHAPES::Calc (order_inner, lami[sort[1]]-lami[sort[0]], lami[sort[2]], lami[sort[3]], psds);

	/*
	  JS, Nov 24:
	  changed orientation of inner shapes (dep. on vertex numbers)
	  changed ordering in T_INNERSHAPES::Calc: inner loop is x-loop
	  swap x-direction
	 */
      }
  }

#endif





#ifdef VERSION1


  template <class T_ORTHOPOL>
  void H1HighOrderTet<T_ORTHOPOL> :: 
  EvaluateShapeGrid (const TensorProductIntegrationRule & ir,
		     const FlatVector<double> coefs,
		     FlatVector<double> gridvalues,
		     LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();


    FlatVector<double> shapes(GetNDof(), lh);
    FlatVector<double> polx(order+1, lh), poly(order+1, lh), polz(order+1, lh); 
    FlatVector<double> sumx(order+1, lh);
    FlatMatrix<> sumyx(order+1, order+1, lh);

    FlatVector<double> sumup(ir.GetNIPX(), lh);
    FlatVector<double> powy(ir.GetNIPY(), lh);
    FlatVector<double> powz(ir.GetNIPZ(), lh);


    FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
    FlatMatrix<> legxbubT(order+1, ir.GetNIPX(), lh);

    for (int ix = 0; ix < ir.GetNIPX(); ix++)
      {
	double * hp = &legxbub(ix, 0);
	LegendrePolynomial (order, 2*ir.GetIPX(ix)-1, hp);
	for (int j = 0; j <= order; j++)
	  legxbub(ix, j) *= 4 * ir.GetIPX(ix) * (1-ir.GetIPX(ix));
      }
    legxbubT = Trans (legxbub);


    FlatVector<> bubx(ir.GetNIPX(), lh);
    for (int ix = 0; ix < ir.GetNIPX(); ix++)
      {
	double x = ir.GetIPX(ix);
	bubx(ix) = 4 * x*(1-x);
      }

    FlatMatrix<> powzm(ir.GetNIPZ(), order+1, lh);
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      {
	double fac = 1-ir.GetIPZ(iz);
	double val = 1;
	for (int j = 0; j <= order; j++, val *= fac)
	  powzm(iz, j) = val;
      }


    shapes = 0;
    int first_inner = 0;
    int first_face = 0, next_face;

    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]])
	  swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++)
      isort[sort[i]] = i;

    int ipnr = 0;


    // vertex shapes
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  {
	    /*
	    IntegrationPoint ip = ir.GetIP(ix, iy, iz);
	    double x = ip(0);
	    double y = ip(1);
	    double z = ip(2);
	    */

	    double xi = ir.GetIPX(ix);
	    double eta = ir.GetIPY(iy);
	    double zeta = ir.GetIPZ(iz);

	    double lami[4] = { zeta, eta * (1-zeta), xi * (1-eta)*(1-zeta), (1-xi)*(1-eta)*(1-zeta) };

	    double x = lami[isort[0]];
	    double y = lami[isort[1]];
	    double z = lami[isort[2]];

	    gridvalues(ipnr) = coefs(0) * x + coefs(1) * y + coefs(2) * z + coefs(3) * (1-x-y-z);
	  }
    
    
    // edge dofs
    
    int ii = 4;
    
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = isort[edges[i][0]];
	  int ee = isort[edges[i][1]];
	  if (ee < es) swap (ee, es);

	  for (int iy = 0; iy < ir.GetNIPY(); iy++)
	    powy(iy) = pow (1-ir.GetIPY(iy), order_edge[i]-1);
	  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	    powz(iz) = powzm(iz, order_edge[i]-1);

	  sumup = 0.0;
	  ipnr = 0;

	  switch (es)
	    {
	    case 0:
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPZ(iz), shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      sumup(iz) += coefs(ii+hi) * shapes(hi);
		  }
		
		switch (ee)
		  {
		  case 1:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  {
			    double fac = sumup(iz) * ir.GetIPY(iy);
			    for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			      gridvalues(ipnr) += fac;
			  }
		      break;
		    }
		  case 2:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  {
			    double fac = sumup(iz) * (1-ir.GetIPY(iy));
			    for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			      gridvalues(ipnr) += fac * ir.GetIPX(ix);
			  }
			      
		      break;
		    }
		  case 3:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  {
			    double fac = sumup(iz) * (1-ir.GetIPY(iy));
			    for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			      gridvalues(ipnr) += fac * (1-ir.GetIPX(ix));
			  }
		      break;
		    }
		    
		  }


		break;
	      }
	    case 1:
	      {
		for (int iy = 0; iy < ir.GetNIPY(); iy++)
		  {
		    T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPY(iy), shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      sumup(iy) += coefs(ii+hi) * shapes(hi);
		  }

		if (ee == 2)
		  {
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      for (int iy = 0; iy < ir.GetNIPY(); iy++)
			{
			  double fac = sumup(iy) * (1-ir.GetIPZ(iz)) * powz(iz); 
			  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			    gridvalues(ipnr) += fac * ir.GetIPX(ix);
			}
		  }
		else
		  {
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      for (int iy = 0; iy < ir.GetNIPY(); iy++)
			{
			  double fac = sumup(iy) *(1-ir.GetIPZ(iz)) * powz(iz);
			  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			    gridvalues(ipnr) += fac * (1-ir.GetIPX(ix));
			}
		  }

		break;
	      }
	    case 2:
	      {
		for (int ix = 0; ix < ir.GetNIPX(); ix++)
		  {
		    T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPX(ix), shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      sumup(ix) += coefs(ii+hi) * shapes(hi);
		  }

		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double fac = (1-ir.GetIPY(iy))*(1-ir.GetIPZ(iz)) * powz(iz)*powy(iy);
		      for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			gridvalues(ipnr) += sumup(ix) * fac;
		    }

		break;
	      }
	    }

	  ii += order_edge[i]-1;
	}


    
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    
    for (int i = 0; i < 4; i++)
      if (order_face[i] >= 3)
	{
	  int fav[3];
	  for (int j = 0; j < 3; j++)
	    fav[j] = isort[faces[i][j]];
	  
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 
	  if(fav[1] > fav[2]) swap(fav[1],fav[2]);
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 	
	  
	  int vop = 6 - fav[0] - fav[1] - fav[2];  	

	  int p = order_face[i]-3;


	  if (fav[1] == 1)
	    {
	      ipnr = 0;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  double z = ir.GetIPZ(iz);
		  
		  JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);
		  double fac = 1;
		  for (int j = p; j >= 0; j--, fac *= 1-z)
		    poly[j] *= fac;

		  sumx = 0.0;
		  int cnt = ii;
		  for (int j2 = 0; j2 <= p; j2++)
		    for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
		      sumx(i2) += coefs(cnt) * poly[j2];

		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double y = ir.GetIPY(iy);
		      sumx *= 4*y*z*(1-z)*(1-z)*(1-y);

		      LegendrePolynomial (p, 2*y-1, polx);

		      double sum = 0;
		      for (int i2 = 0; i2 <= p; i2++)
			sum += sumx(i2) * polx[i2];

		      if (vop == 2)
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  gridvalues(ipnr) += sum * (1-ir.GetIPX(ix));
		      else
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  gridvalues(ipnr) += sum * ir.GetIPX(ix);
		    }
		}

	    }
	  else
	    {


	      if (fav[0] == 0)
		{
		  // fav[0]=0, fav[1] = 2, fav[2]=3

		  /*
		  ipnr = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    for (int iy = 0; iy < ir.GetNIPY(); iy++)
		      {
			double y = ir.GetIPY(iy);
			double z = ir.GetIPZ(iz);

			JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);

			double fac = z*(1-z)*(1-z)*(1-y)*(1-y);
			for (int j = p; j >= 0; j--, fac *= (1-y)*(1-z))
			  poly[j] *= fac;

			sumx = 0;
			for (int i2 = 0; i2 <= p; i2++)
			  {
			    double val = 0;
			    for (int ix = 0; ix < ir.GetNIPX(); ix++)
			      val += gridvalues(ipnr+ix) * legxbubT(i2, ix);
			    sumx(i2) = val;
			  }
			ipnr += ir.GetNIPX();



			int ii = first_face;
			int cnt = 0;
			for (int j2 = 0; j2 <= p; j2++)
			  for (int i2 = 0; i2 <= p-j2; i2++)
			    {
			      coefs(ii+cnt) += sumx[i2] * poly[j2];
			      cnt++;
			    }
			first_inner = ii + cnt;
		      }
		  */		  
		  

		  // alternative shape functions for face 1
		  ipnr = 0;
		  FlatVector<> powery(order+1, lh);

		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    {
		      double z = ir.GetIPZ(iz);
		      JacobiPolynomial (order_face[i]-3, 2*z-1, order_face[i], 0, poly);

		      double fac = z*(1-z)*(1-z);
		      for (int j = order_face[i]-3; j >= 0; j--, fac *= (1-z))
			poly[j] *= fac;

		      for (int iy = 0; iy < ir.GetNIPY(); iy++)
			{
			  double y = ir.GetIPY(iy);
			  
			  fac = (1-y)*(1-y);
			  for (int j = 0; j <= order; j++, fac *= (1-y))
			    powery[j] = fac;
			  
			  sumx = 0.0;
			  int cnt = 0;
			  for (int j2 = 0; j2 <= p; j2++)
			    for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
			      sumx(i2) += coefs(ii+cnt)*powery(i2) * poly[j2];
			  
			  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			    {
			      double sum = 0;
			      for (int i2 = 0; i2 <= p; i2++)
				sum += sumx(i2) * legxbub(ix, i2);
			      gridvalues(ipnr) += sum;
			    }
			}
		    }
		}
	      else
		{
		  // fav[0]=1, fav[1] = 2, fav[2]=3

		  FlatMatrix<double> sumxy(ir.GetNIPX(), lh);
		  ipnr = 0;
		  sumxy = 0.0;

		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double y = ir.GetIPY(iy);
			
		      JacobiPolynomial (p, 2*y-1, order_face[i], 0, poly);

		      double fac = 1;
		      for (int j = p; j >= 0; j--, fac *= (1-y))
			poly[j] *= fac;

		      sumx = 0.0;
		      
		      int cnt = 0;
		      for (int j2 = 0; j2 <= p; j2++)
			for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
			  sumx(i2) += coefs(ii+cnt) * poly[j2];

		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			{
			  double sum = 0;
			  for (int i2 = 0; i2 <= p; i2++)
			    sum += sumx(i2) * legxbub(ix, i2); // polx[i2];
			  sumxy(iy,ix) += sum;
			}
		    }

		  ipnr = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    for (int iy = 0; iy < ir.GetNIPY(); iy++)
		      {
			double y = ir.GetIPY(iy);
			double z = ir.GetIPZ(iz);
			double bubyz = y*(1-y)*(1-y) * powzm(iz, p+3);
			
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  gridvalues(ipnr) += sumxy(iy,ix) * bubyz; 
		      }
		}
	    }

	  ii += (order_face[i]-2)*(order_face[i]-1)/2;
	}	    
    


    // inner dofs
    ipnr = 0;
    if (order_inner >= 4)
      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	{
	  double z = ir.GetIPZ(iz);
	  JacobiPolynomial (order_inner-4, 2*z-1, order_inner, 0, polz);
	  for (int k = 0; k < polz.Size(); k++) 
	    polz[k] *= z;
	  
	  sumyx = 0.0;
	  
	  int hi = ii;
	  for (int i = 0; i <= order_inner-4; i++)
	    {
	      double fac = polz[i];
	      for (int j = 0; j <= order_inner-4-i; j++)
		for (int k = 0; k <= order_inner-4-i-j; k++)
		  sumyx(j,k) += fac * coefs(hi++);
	    }


	  for (int iy = 0; iy < ir.GetNIPY(); iy++)
	    {
	      double y = ir.GetIPY(iy);
	      JacobiPolynomial (order_inner-4, 2*y-1, order_inner, 0, poly); 
	      
	      double fac = y*(1-z);
	      for (int k = 0; k < poly.Size(); k++, fac *= 1-z) 
		poly[k] *= fac;
	      
	      sumx = 0.0;
	      
	      for (int j = 0; j <= order-4; j++)
		for (int k = 0; k <= order-4-j; k++)
		  sumx(k) += sumyx(j, k) * poly[j];

	      fac = (1-y)*(1-y)*(1-z)*(1-z);
	      for (int k = 0; k <= order_inner-4; k++, fac *= (1-y)*(1-z))
		sumx(k) *= fac;
	      
	      
	      for (int ix = 0; ix < ir.GetNIPX(); ix++)
		{
		  double val = 0;
		  for (int k = 0; k <= order_inner-4; k++) 
		    val += sumx(k) * legxbub(ix, k); 
		  gridvalues(ipnr++) += val;
		}
	    }
	}
    


    lh.CleanUp (heapp);
  }
#endif








  template <class T_ORTHOPOL>
  void H1HighOrderTet<T_ORTHOPOL> :: 
  EvaluateShapeGrid (const TensorProductIntegrationRule & ir,
		     const FlatVector<double> coefs,
		     FlatVector<double> gridvalues,
		     LocalHeap & lh) const
  {
    void * heapp = lh.GetPointer();

    FlatMatrix<double> gridyz_polx(ir.GetNIPY()*ir.GetNIPZ(),order+1, lh);


    FlatVector<double> shapes(GetNDof(), lh);
    FlatVector<double> poly(order+1, lh), polz(order+1, lh); 
    FlatVector<double> sumx(order+1, lh);
    FlatMatrix<> sumyx(order+1, order+1, lh);
    FlatVector<double> sumup(ir.GetNIPX(), lh);
    FlatMatrix<double> gridy_polx(ir.GetNIPY(), order+1, lh);

    FlatMatrix<> legy(ir.GetNIPY(), order+1, lh);
    for (int iy = 0; iy < ir.GetNIPY(); iy++)
      {
	double * hp = &legy(iy, 0);
	LegendrePolynomial (order, 2*ir.GetIPY(iy)-1, hp);
      }

    FlatMatrix<> powzm(ir.GetNIPZ(), order+1, lh);
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      {
	double fac = 1-ir.GetIPZ(iz);
	double val = 1;
	for (int j = 0; j <= order; j++, val *= fac)
	  powzm(iz, j) = val;
      }

    FlatMatrix<> powym(ir.GetNIPZ(), order+1, lh);
    for (int iy = 0; iy < ir.GetNIPY(); iy++)
      {
	double fac = 1-ir.GetIPY(iy);
	double val = 1;
	for (int j = 0; j <= order; j++, val *= fac)
	  powym(iy, j) = val;
      }


    shapes = 0;

    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]])
	  swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++)
      isort[sort[i]] = i;

    int ipnr = 0;
    int ipyz;
    gridyz_polx = 0.0;

    for (int i = 0; i < 4; i++)
      {
	ipyz = 0;
	for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	  for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	    {
	      double y = ir.GetIPY(iy);
	      double z = ir.GetIPZ(iz);
	      // double lamiyz[4] = { z, y * (1-z), 1, 1 };
	      
	      switch (isort[i])
		{
		case 0: 
		  gridyz_polx(ipyz,0) += z * coefs(i);
		  break;
		case 1:
		  gridyz_polx(ipyz,0) += y*(1-z) * coefs(i);
		  break;
		case 2:
		  gridyz_polx(ipyz,0) += 0.5 * coefs(i) * (1-y)*(1-z);
		  gridyz_polx(ipyz,1) += 0.5 * coefs(i);
		  break;
		case 3:
		  gridyz_polx(ipyz,0) += 0.5 * coefs(i) * (1-y)*(1-z);
		  gridyz_polx(ipyz,1) -= 0.5 * coefs(i);
		  break;
		}
	    }
      }
      

    
    // edge dofs
    
    int ii = 4;
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = isort[edges[i][0]];
	  int ee = isort[edges[i][1]];
	  if (ee < es) swap (ee, es);

	  sumup = 0.0;
	  ipnr = 0;
	  ipyz = 0;
	  int p = order_edge[i]-2;

	  switch (es)
	    {
	    case 0:
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    double z = ir.GetIPZ(iz);
		    LegendrePolynomial (p, 2*z-1, shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      sumup(iz) += coefs(ii+hi) * shapes(hi)*4*z;
		  }
		
		switch (ee)
		  {
		  case 1:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			  gridyz_polx(ipyz, 0) += sumup(iz) * ir.GetIPY(iy) * (1-ir.GetIPZ(iz));
		      break;
		    }
		  case 2:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			  {
			    gridyz_polx(ipyz, 0) += 0.5 * sumup(iz) * (1-ir.GetIPY(iy)) * (1-ir.GetIPZ(iz));
			    gridyz_polx(ipyz, 1) += 0.5 * sumup(iz);
			  }
		      break;
		    }
		  case 3:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			  {
			    gridyz_polx(ipyz, 0) += 0.5 * sumup(iz) * (1-ir.GetIPY(iy)) * (1-ir.GetIPZ(iz));
			    gridyz_polx(ipyz, 1) -= 0.5 * sumup(iz);
			  }
		      break;
		    }
		    
		  }

		break;
	      }
	    case 1:
	      {
		for (int iy = 0; iy < ir.GetNIPY(); iy++)
		  for (int hi = 0; hi <= p; hi++)
		    sumup(iy) += coefs(ii+hi) * 4*ir.GetIPY(iy)* legy(iy, hi); 

		if (ee == 2)
		  {
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			{
			  gridyz_polx(ipyz, 0) += 0.5 * sumup(iy)*powzm(iz,p+1) * (1-ir.GetIPY(iy))*(1-ir.GetIPZ(iz));
			  gridyz_polx(ipyz, 1) += 0.5 * sumup(iy)*powzm(iz,p+1);
			}
		  }
		else
		  {
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			{
			  gridyz_polx(ipyz, 0) +=  0.5 * sumup(iy) * powzm(iz, p+1) * (1-ir.GetIPY(iy))*(1-ir.GetIPZ(iz) );
			  gridyz_polx(ipyz, 1) += -0.5 * sumup(iy) * powzm(iz, p+1); 
			}
		  }

		break;
	      }
	    case 2:
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
		    for (int hi = 0; hi <= p; hi++)
		      gridyz_polx(ipyz, hi+2) += coefs(ii+hi) * powzm(iz, p-hi) * powym(iy, p-hi);
		break;
	      }
	    }

	  ii += order_edge[i]-1;
	}





    
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    
    for (int i = 0; i < 4; i++)
      if (order_face[i] >= 3)
	{
	  int fav[3];
	  for (int j = 0; j < 3; j++)
	    fav[j] = isort[faces[i][j]];
	  
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 
	  if(fav[1] > fav[2]) swap(fav[1],fav[2]);
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 	
	  
	  int vop = 6 - fav[0] - fav[1] - fav[2];  	
	  int p = order_face[i]-3;

	  if (fav[1] == 1)
	    {
	      // fav[0]=0, fav[1] = 1, fav[2]= 2 or 3

	      ipnr = 0;
	      ipyz = 0;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  double z = ir.GetIPZ(iz);
		  JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);

		  double fac = z*(1-z);
		  for (int j = p; j >= 0; j--, fac *= 1-z) poly[j] *= fac;

		  sumx = 0.0;
		  int cnt = ii;
		  for (int j2 = 0; j2 <= p; j2++)
		    for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
		      sumx(i2) += coefs(cnt) * poly[j2];

		  for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
		    {
		      double y = ir.GetIPY(iy);
		      double sum = 0;
		      for (int i2 = 0; i2 <= p; i2++)
			sum += sumx(i2) * legy(iy, i2) * 4*y;

		      if (vop == 2)
			{
			  gridyz_polx(ipyz,0) +=  0.5 * sum * (1-y)*(1-z);
			  gridyz_polx(ipyz,1) += -0.5 * sum;
			}
		      else
			{
			  gridyz_polx(ipyz,0) += 0.5 * sum * (1-y)*(1-z);
			  gridyz_polx(ipyz,1) += 0.5 * sum;
			}
		    }
		}

	    }
	  else
	    {

	      if (fav[0] == 0)
		{
		  // fav[0]=0, fav[1] = 2, fav[2]=3

		  // alternative shape functions for face 1
		  ipyz = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    {
		      double z = ir.GetIPZ(iz);
		      JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);
		      poly *= z;
		      sumx = 0.0;


		      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			{
			  int cnt = ii;
			  for (int j2 = 0; j2 <= p; j2++)
			    for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
			      gridyz_polx(ipyz, i2+2) += coefs(cnt)*poly[j2] * powzm(iz, p-i2-j2)*powym(iy, p-j2-i2);
			}
		    }

		  /*
		  // alternative shape functions for face 1
		  ipyz = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    {
		      double z = ir.GetIPZ(iz);
		      JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);
		      poly *= z;
		      sumx = 0.0;
		      int cnt = ii;
		      for (int j2 = 0; j2 <= p; j2++)
			for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
			  sumx(i2) += coefs(cnt) * poly[j2] * powzm(iz, p-j2-i2);

		      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
			for (int i2 = 0; i2 <= p; i2++)
			  gridyz_polx(ipyz, i2+2) += sumx(i2);
		    }
		  */

		}
	      else
		{
		  // fav[0]=1, fav[1] = 2, fav[2]=3

		  // FlatMatrix<double> gridy_polx(ir.GetNIPY(), p+1, lh);
		  gridy_polx = 0.0;

		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double y = ir.GetIPY(iy);
			
		      JacobiPolynomial (p, 2*y-1, order_face[i], 0, poly);
		      poly *= y;

		      int cnt = ii;
		      for (int j2 = 0; j2 <= p; j2++)
			for (int i2 = 0; i2 <= p-j2; i2++, cnt++)
			  gridy_polx(iy, i2) += coefs(cnt) * poly[j2] * powym(iy, p-j2-i2);
		    }
		  
		  ipyz = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
		      for (int i2 = 0; i2 <= p; i2++)
			gridyz_polx(ipyz, i2+2) += gridy_polx(iy, i2) * powzm(iz,p+1-i2);
		}
	    }

	  ii += (order_face[i]-2)*(order_face[i]-1)/2;
	}	    
    


    // inner dofs
    ipnr = 0;
    ipyz = 0;
    if (order_inner >= 4)
      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	{
	  double z = ir.GetIPZ(iz);
	  JacobiPolynomial (order_inner-4, 2*z-1, order_inner, 0, polz);
	  polz *= z;
	  
	  sumyx = 0.0;
	  
	  int hi = ii;
	  for (int i = 0; i <= order_inner-4; i++)
	    {
	      double fac = polz[i];
	      for (int j = 0; j <= order_inner-4-i; j++)
		for (int k = 0; k <= order_inner-4-i-j; k++)
		  sumyx(j,k) += fac * coefs(hi++);
	    }

	  for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	    {
	      double y = ir.GetIPY(iy);
	      JacobiPolynomial (order_inner-4, 2*y-1, order_inner, 0, poly); 
	      
	      double fac = y*(1-z);
	      for (int k = 0; k < poly.Size(); k++, fac *= 1-z) 
		poly[k] *= fac;
	      
	      sumx = 0.0;
	      
	      for (int j = 0; j <= order-4; j++)
		for (int k = 0; k <= order-4-j; k++)
		  sumx(k) += sumyx(j, k) * poly[j];

	      for (int k = 0; k <= order_inner-4; k++)
		gridyz_polx(ipyz, k+2) += sumx(k);
	    }
	}









    FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
    for (int ix = 0; ix < ir.GetNIPX(); ix++)
      {
	double x = ir.GetIPX(ix);
	double * hp = &legxbub(ix, 2);
	LegendrePolynomial (order-2, 2*x-1, hp);

	legxbub(ix, 0) = 1;
	legxbub(ix, 1) = 2*x-1;
	for (int j = 0; j <= order-2; j++)
	  legxbub(ix, j+2) *= 4 * x * (1-x);
      }


    /*
    ipnr = 0;
    ipyz = 0;
    gridvalues = 0;
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	{
	  for (int j = 1; j <= order; j++)
	    gridyz_polx(ipyz, j) *= powym(iy,j)*powzm(iz,j);

	  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	    {
	      double sum = 0;
	      for (int j = 0; j <= order; j++)
		sum += gridyz_polx(ipyz, j) * legxbub(ix, j);
	      gridvalues(ipnr) += sum;	      
	    }
	}
    */


    ipnr = 0;
    ipyz = 0;
    gridvalues = 0;
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++, ipyz++)
	for (int j = 1; j <= order; j++)
	  gridyz_polx(ipyz, j) *= powym(iy,j)*powzm(iz,j);

    FlatMatrix<> gvm(ir.GetNIPX()*ir.GetNIPY(), ir.GetNIPX(), &gridvalues(0));
    gvm = gridyz_polx * Trans (legxbub);
    

    lh.CleanUp (heapp);
  }



















  template <class T_ORTHOPOL>
  void H1HighOrderTet<T_ORTHOPOL> :: 
  EvaluateShapeGridTrans (const TensorProductIntegrationRule & ir,
			  const FlatVector<double> gridvalues,
			  FlatVector<double> coefs,
			  LocalHeap & lh) const
  {
    //    cout << "nip = " << gridvalues.Size() << ", ndof = " << coefs.Size() << endl;
    void * heapp = lh.GetPointer();

    coefs = 0.0;

    FlatVector<double> shapes(GetNDof(), lh);
    FlatVector<double> polx(order+1, lh), poly(order+1, lh), polz(order+1, lh); 
    FlatVector<double> sumx(order+1, lh);
    FlatMatrix<> sumyx(order+1, order+1, lh);
    

    FlatMatrix<> legxbub(ir.GetNIPX(), order+1, lh);
    FlatMatrix<> legxbubT(order+1, ir.GetNIPX(), lh);
    for (int ix = 0; ix < ir.GetNIPX(); ix++)
      {
	double * hp = &legxbub(ix, 0);
	LegendrePolynomial (order, 2*ir.GetIPX(ix)-1, hp);
	for (int j = 0; j <= order; j++)
	  legxbub(ix, j) *= 4 * ir.GetIPX(ix) * (1-ir.GetIPX(ix));
      }
    legxbubT = Trans (legxbub);


    shapes = 0;
    int first_inner = 0;
    int first_face = 0, next_face;

    int sort[4] = { 0, 1, 2, 3 };
    for (int i = 0; i < 4; i++)
      for (int j = 0; j < 3; j++)
	if (vnums[sort[j]] > vnums[sort[j+1]])
	  swap (sort[j], sort[j+1]);
    
    int isort[4];
    for (int i = 0; i < 4; i++)
      isort[sort[i]] = i;

    int ipnr = 0;

    const EDGE * edges = ElementTopology::GetEdges (ET_TET);

    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  {
	    /*
	    IntegrationPoint ip = ir.GetIP(ix, iy, iz);
	    double x = ip(0);
	    double y = ip(1);
	    double z = ip(2);
	    */
	    double xi = ir.GetIPX(ix);
	    double eta = ir.GetIPY(iy);
	    double zeta = ir.GetIPZ(iz);

	    double lami[4] = { zeta, eta * (1-zeta), xi * (1-eta)*(1-zeta), (1-xi)*(1-eta)*(1-zeta) };

	    double x = lami[isort[0]];
	    double y = lami[isort[1]];
	    double z = lami[isort[2]];


	    double val = gridvalues(ipnr);

	    coefs(0) += val * x;
	    coefs(1) += val * y;
	    coefs(2) += val * z;
	    coefs(3) += val * (1-x-y-z);
	  }
    



    // edge dofs
    FlatVector<double> sumup(ir.GetNIPX(), lh);
    FlatVector<double> powy(ir.GetNIPY(), lh);
    FlatVector<double> powz(ir.GetNIPZ(), lh);

    int ii = 4;

    /*
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = isort[edges[i][0]];
	  int ee = isort[edges[i][1]];
	  if (ee < es) swap (ee, es);

	  for (int iy = 0; iy < ir.GetNIPY(); iy++)
	    powy(iy) = pow (1-ir.GetIPY(iy), order_edge[i]-1);
	  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	    powz(iz) = pow (1-ir.GetIPZ(iz), order_edge[i]-1);


	  sumup = 0.0;

	  ipnr = 0;
	  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	    for (int iy = 0; iy < ir.GetNIPY(); iy++)
	      for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
		{
		  int ia [] = { iz, iy, ix };
		  double lamis_sum_pow[4] = { 1, powz(iz), powz(iz)*powy(iy), -999 };
		  double lamis_dir[4] = { 1-2*ir.GetIPZ(iz), 1-2*ir.GetIPY(iy), 1-2*ir.GetIPX(ix), 0 };
		  
		  double lamedge[4][4] =
		    { { 0, 0, 0, 1e25 }, 
		      { ir.GetIPY(iy), 0, 0, 0 },
		      { ir.GetIPX(ix) * (1-ir.GetIPY(iy)), ir.GetIPX(ix) * (1-ir.GetIPZ(iz)), 0 },
		      { (1-ir.GetIPX(ix))*(1-ir.GetIPY(iy)), (1-ir.GetIPX(ix))*(1-ir.GetIPZ(iz)), (1-ir.GetIPY(iy))*(1-ir.GetIPZ(iz)) } };
		  
		  sumup(ia[es]) += lamedge[ee][es] * gridvalues(ipnr) * lamis_sum_pow[es];
		}

	  switch (es)
	    {
	    case 0:
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPZ(iz), shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      coefs(ii+hi) += sumup(iz) * shapes(hi);
		  }
		break;
	      }
	    case 1:
	      {
		for (int iy = 0; iy < ir.GetNIPY(); iy++)
		  {
		    T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPY(iy), shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      coefs(ii+hi) += sumup(iy) * shapes(hi);
		  }
		break;
	      }
	    case 2:
	      {
		for (int ix = 0; ix < ir.GetNIPX(); ix++)
		  {
		    T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPX(ix), shapes);
		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      coefs(ii+hi) += sumup(ix) * shapes(hi);
		  }
		break;
	      }
	    }

	  ii += order_edge[i]-1;
	}
    */



   for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = isort[edges[i][0]];
	  int ee = isort[edges[i][1]];
	  if (ee < es) swap (ee, es);

	  for (int iy = 0; iy < ir.GetNIPY(); iy++)
	    powy(iy) = pow (1-ir.GetIPY(iy), order_edge[i]-1);
	  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	    powz(iz) = pow (1-ir.GetIPZ(iz), order_edge[i]-1);


	  sumup = 0.0;

	  ipnr = 0;


	  switch (es)
	    {
	    case 0:
	      {
		switch (ee)
		  {
		  case 1:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			    sumup(iz) += ir.GetIPY(iy) * gridvalues(ipnr);
		      break;
		    }
		  case 2:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			    sumup(iz) += ir.GetIPX(ix)*(1-ir.GetIPY(iy)) * gridvalues(ipnr);
		      break;
		    }
		  case 3:
		    {
		      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
			for (int iy = 0; iy < ir.GetNIPY(); iy++)
			  for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			    sumup(iz) += (1-ir.GetIPX(ix))*(1-ir.GetIPY(iy)) * gridvalues(ipnr);
		      break;
		    }
		    
		  }

		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  {
		    // T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPZ(iz), shapes);

		    double z = ir.GetIPZ(iz);
		    LegendrePolynomial (order_edge[i]-2, 2*z-1, shapes);

		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      coefs(ii+hi) += sumup(iz) * shapes(hi)*4*z*(1-z);
		  }
		break;
	      }
	    case 1:
	      {
		if (ee == 2)
		  {
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      for (int iy = 0; iy < ir.GetNIPY(); iy++)
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  sumup(iy) += ir.GetIPX(ix) * (1-ir.GetIPZ(iz)) * gridvalues(ipnr) * powz(iz); 
		  }
		else
		  {
		    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		      for (int iy = 0; iy < ir.GetNIPY(); iy++)
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  sumup(iy) += (1-ir.GetIPX(ix))*(1-ir.GetIPZ(iz)) * gridvalues(ipnr) * powz(iz);
		  }

		for (int iy = 0; iy < ir.GetNIPY(); iy++)
		  {
		    // T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPY(iy), shapes);

		    double y = ir.GetIPY(iy);
		    LegendrePolynomial (order_edge[i]-2, 2*y-1, shapes);


		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      coefs(ii+hi) += sumup(iy) * shapes(hi)*4*y*(1-y);
		  }
		break;
	      }
	    case 2:
	      {
		for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double fac = (1-ir.GetIPY(iy))*(1-ir.GetIPZ(iz)) * powz(iz)*powy(iy);
		      for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			sumup(ix) += fac * gridvalues(ipnr);
		    }

		for (int ix = 0; ix < ir.GetNIPX(); ix++)
		  {
		    // T_ORTHOPOL::Calc (order_edge[i], 1-2*ir.GetIPX(ix), shapes);

		    double x = ir.GetIPX(ix);
		    LegendrePolynomial (order_edge[i]-2, 2*x-1, shapes);

		    for (int hi = 0; hi < order_edge[i]-1; hi++)
		      coefs(ii+hi) += sumup(ix) * shapes(hi)*4*x*(1-x);
		  }
		break;
	      }
	    }

	  ii += order_edge[i]-1;
	}







    
    first_face = ii;


    /*
    // face shapes
    const int faces[4][3] = 
      { { 1, 2, 3 },
	{ 0, 2, 3 },
	{ 0, 1, 3 },
	{ 0, 1, 2 } };

    shapes = 0.0;
    ipnr = 0;
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      for (int iy = 0; iy < ir.GetNIPY(); iy++)
	for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	  {
	    double zeta = ir.GetIPZ(iz);
	    double eta = ir.GetIPY(iy)*(1-zeta);
	    double xi = ir.GetIPX(ix)*(1-eta-zeta);

	    double lamis[4] = { zeta, eta, xi, 1-xi-eta-zeta };

	    int ii = first_face;
	    
	    // face dofs
	    for (int i = 0; i < 4; i++)
	      if (order_face[i] >= 3)
		{
		  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
		  
		  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
		  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
		  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
		  
		  int vop = 6 - fav[0] - fav[1] - fav[2];  	
		  
		  double * hp = &shapes(ii);
		  ii += T_FACESHAPES::Calc (order_face[i], 
					    lamis[isort[fav[2]]]-lamis[isort[fav[1]]],
					    lamis[isort[fav[0]]], lamis[isort[vop]],  hp);
		}

	    first_inner = ii;
	    coefs += gridvalues(ipnr) * shapes;
	  }	    
    */





    /*
    first_inner = first_face;
    
    for (int i = 0; i < 4; i++)
      if (order_face[i] >= 3)
	{
	  int fav[3];
	  for (int j = 0; j < 3; j++)
	    fav[j] = isort[faces[i][j]];
	  
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 
	  if(fav[1] > fav[2]) swap(fav[1],fav[2]);
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 	
	  
	  int vop = 6 - fav[0] - fav[1] - fav[2];  	

	  
	  ipnr = 0;
	  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
	    for (int iy = 0; iy < ir.GetNIPY(); iy++)
	      {
		double y = ir.GetIPY(iy);
		double z = ir.GetIPZ(iz);

		double lamis_add0[] = { 1, 1-z, (1-y)*(1-z) };

		double local_y = (fav[0] == 0) ? 2*z-1 : 2*y-1;
		JacobiPolynomial (order_face[i]-3, local_y, order_face[i], 0, poly);

		double fac = 1;
		for (int j = 0; j <= poly.Size(); j++, fac *= lamis_add0[fav[0]])
		  poly[j] *= fac;

		fac = 1;
		for (int j = order_face[i]-3; j >= 0; j--, fac *= lamis_add0[fav[1]])
		  poly[j] *= fac;

		double bubyz;
		switch (vop)
		  {
		  case 0: bubyz =  y  *(1-z)*(1-z)*(1-z)*(1-y)*(1-y); break;
		  case 1: bubyz =    z*(1-z)*(1-z)      *(1-y)*(1-y); break;
		  case 2: bubyz =  y*z*(1-z)*(1-z)      *(1-y)      ; break;
		  case 3: bubyz =  y*z*(1-z)*(1-z)      *(1-y)      ; break;
		  }

		poly *= bubyz;

		sumx = 0;
		for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
		  {
		    double x = ir.GetIPX(ix);
		    double local_x = (fav[1] == 1) ? 2*y-1 : 2*x-1;

		    LegendrePolynomial (order_face[i]-3, local_x, polx);

		    double bubx;
		    switch (vop)
		      {
		      case 0: bubx = x*(1-x); break;
		      case 1: bubx = x*(1-x); break;
		      case 2: bubx =   (1-x); break;
		      case 3: bubx = x      ; break;
		      }

		    double val = gridvalues(ipnr) * 4 * bubx; 

		    for (int i2 = 0; i2 <= order_face[i]-3; i2++)
		      sumx(i2) += val * polx(i2);
		  }

		int ii = first_face;
		int cnt = 0;
		for (int j2 = 0; j2 <= order_face[i]-3; j2++)
		  for (int i2 = 0; i2 <= order_face[i]-3-j2; i2++)
		    {
		      coefs(ii+cnt) += sumx[i2] * poly[j2];
		      cnt++;
		    }
		first_inner = ii + cnt;
	      }

	  first_face = first_inner;
	}	    
    */











    first_inner = first_face;
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    
    for (int i = 0; i < 4; i++)
      if (order_face[i] >= 3)
	{
	  int fav[3];
	  for (int j = 0; j < 3; j++)
	    fav[j] = isort[faces[i][j]];
	  
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 
	  if(fav[1] > fav[2]) swap(fav[1],fav[2]);
	  if(fav[0] > fav[1]) swap(fav[0],fav[1]); 	
	  
	  int vop = 6 - fav[0] - fav[1] - fav[2];  	

	  int p = order_face[i]-3;

	  if (fav[1] == 1)
	    {
	      ipnr = 0;
	      for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		{
		  double z = ir.GetIPZ(iz);
		  
		  JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);
		  double fac = 1;
		  for (int j = p; j >= 0; j--, fac *= 1-z)
		    poly[j] *= fac;
		  
		  sumx = 0.0;
		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double y = ir.GetIPY(iy);
		      double bubyz = y*z*(1-z)*(1-z)*(1-y);
		      
		      LegendrePolynomial (p, 2*y-1, polx);
		      
		      double sum = 0;

		      if (vop == 2)
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  sum += gridvalues(ipnr) * 4 * (1-ir.GetIPX(ix));
		      else
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  sum += gridvalues(ipnr) * 4 * ir.GetIPX(ix);
			
		      for (int j = 0; j <= p; j++)
			sumx(j) += bubyz * sum * polx(j);
		    }


		  int ii = first_face;
		  int cnt = 0;
		  for (int j2 = 0; j2 <= p; j2++)
		    for (int i2 = 0; i2 <= p-j2; i2++)
		      {
			coefs(ii+cnt) += sumx[i2] * poly[j2];
			cnt++;
		      }
		  first_inner = ii + cnt;
		}


	    }
	  else
	    {


	      if (fav[0] == 0)
		{
		  // fav[0]=0, fav[1] = 2, fav[2]=3


		  ipnr = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    for (int iy = 0; iy < ir.GetNIPY(); iy++)
		      {
			double y = ir.GetIPY(iy);
			double z = ir.GetIPZ(iz);

			JacobiPolynomial (p, 2*z-1, order_face[i], 0, poly);

			double fac = z*(1-z)*(1-z)*(1-y)*(1-y);
			for (int j = p; j >= 0; j--, fac *= (1-y)*(1-z))
			  poly[j] *= fac;

			sumx = 0;
			for (int i2 = 0; i2 <= p; i2++)
			  {
			    double val = 0;
			    for (int ix = 0; ix < ir.GetNIPX(); ix++)
			      val += gridvalues(ipnr+ix) * legxbubT(i2, ix);
			    sumx(i2) = val;
			  }
			ipnr += ir.GetNIPX();



			int ii = first_face;
			int cnt = 0;
			for (int j2 = 0; j2 <= p; j2++)
			  for (int i2 = 0; i2 <= p-j2; i2++)
			    {
			      coefs(ii+cnt) += sumx[i2] * poly[j2];
			      cnt++;
			    }
			first_inner = ii + cnt;
		      }
		  
		  /*

		  // alternative shape functions for face 1
		  ipnr = 0;
		  FlatVector<> powery(order+1, lh);
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    {
		      double z = ir.GetIPZ(iz);
		      
		      JacobiPolynomial (order_face[i]-3, 2*z-1, order_face[i], 0, poly);

		      double fac = z*(1-z)*(1-z);
		      for (int j = order_face[i]-3; j >= 0; j--, fac *= (1-z))
			poly[j] *= fac;


		      for (int iy = 0; iy < ir.GetNIPY(); iy++)
			{
			  double y = ir.GetIPY(iy);
			  
			  fac = (1-y)*(1-y);
			  for (int j = 0; j <= order; j++, fac *= (1-y))
			    powery[j] = fac;
			  
			  sumx = 0;
			  for (int i2 = 0; i2 <= order_face[i]-3; i2++)
			    for (int ix = 0; ix < ir.GetNIPX(); ix++)
			      sumx(i2) += gridvalues(ipnr+ix) * legxbubT(i2, ix);
			  ipnr += ir.GetNIPX();

			  int ii = first_face;
			  int cnt = 0;
			  for (int j2 = 0; j2 <= order_face[i]-3; j2++)
			    for (int i2 = 0; i2 <= order_face[i]-3-j2; i2++)
			      {
				coefs(ii+cnt) += sumx[i2] * powery(i2) * poly[j2];
				cnt++;
			      }
			  first_inner = ii + cnt;
			}
		    }
		  */

		}
	      else
		{
		  // fav[0]=1, fav[1] = 2, fav[2]=3
		  /*
		  ipnr = 0;
		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    for (int iy = 0; iy < ir.GetNIPY(); iy++)
		      {
			double y = ir.GetIPY(iy);
			double z = ir.GetIPZ(iz);
			
			JacobiPolynomial (order_face[i]-3, 2*y-1, order_face[i], 0, poly);

			double fac = 1;
			for (int j = order_face[i]-3; j >= 0; j--, fac *= (1-y))
			  poly[j] *= fac;

			double bubyz = y  *(1-z)*(1-z)*(1-z)*(1-y)*(1-y);
			poly *= bubyz * pow (1-z, order_face[i]-3);
			
			sumx = 0;
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  {
			    double x = ir.GetIPX(ix);
			    LegendrePolynomial (order_face[i]-3, 2*x-1, polx);
			    
			    double val = gridvalues(ipnr) * 4 * x*(1-x);
			    for (int i2 = 0; i2 <= order_face[i]-3; i2++)
			      sumx(i2) += val * polx(i2);
			  }
			
			int ii = first_face;
			int cnt = 0;
			for (int j2 = 0; j2 <= order_face[i]-3; j2++)
			  for (int i2 = 0; i2 <= order_face[i]-3-j2; i2++)
			    {
			      coefs(ii+cnt) += sumx[i2] * poly[j2];
			      cnt++;
			    }
			first_inner = ii + cnt;
		      }
		  */

		  FlatMatrix<double> sumxy(ir.GetNIPX(), lh);
		  ipnr = 0;
		  sumxy = 0.0;

		  for (int iz = 0; iz < ir.GetNIPZ(); iz++)
		    for (int iy = 0; iy < ir.GetNIPY(); iy++)
		      {
			double y = ir.GetIPY(iy);
			double z = ir.GetIPZ(iz);
			double bubyz = y*(1-y)*(1-y) * pow (1-z, p+3);
			
			for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
			  {
			    double x = ir.GetIPX(ix);
			    sumxy(iy,ix) += bubyz * gridvalues(ipnr) * 4 * x*(1-x);
			  }
		      }

		  for (int iy = 0; iy < ir.GetNIPY(); iy++)
		    {
		      double y = ir.GetIPY(iy);
			
		      JacobiPolynomial (p, 2*y-1, order_face[i], 0, poly);

		      double fac = 1;
		      for (int j = p; j >= 0; j--, fac *= (1-y))
			poly[j] *= fac;

		      sumx = 0;
		      for (int ix = 0; ix < ir.GetNIPX(); ix++)
			{
			  LegendrePolynomial (order_face[i]-3, 2*ir.GetIPX(ix)-1, polx);
			  for (int i2 = 0; i2 <= order_face[i]-3; i2++)
			    sumx(i2) += sumxy(iy,ix) * polx(i2);
			}
			
		      int ii = first_face;
		      int cnt = 0;
		      for (int j2 = 0; j2 <= p; j2++)
			for (int i2 = 0; i2 <= p-j2; i2++)
			  {
			    coefs(ii+cnt) += sumx[i2] * poly[j2];
			    cnt++;
			  }
		      first_inner = ii + cnt;
		    }
		}
	    }
	  first_face = first_inner;
	}	    








    
    // first_inner = next_face;


    /*
    // inner dofs
    ipnr = 0;
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      {
	double zeta = ir.GetIPZ(iz);
	LegendrePolynomial (order_inner-4, 2*zeta-1, polz);
	for (int k = 0; k < polz.Size(); k++) polz[k] *= zeta;

	sumyx = 0.0;

	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    double eta = ir.GetIPY(iy)*(1-zeta);
	    ScaledLegendrePolynomial (order_inner-4, 2*eta - (1-zeta) , (1-zeta),poly); 
	    for (int k = 0; k < poly.Size(); k++) poly[k] *= eta;

	    sumx = 0.0;
	    for (int ix = 0; ix < ir.GetNIPX(); ix++, ipnr++)
	      {
		double xi = -(2*ir.GetIPX(ix)-1)*(1-eta-zeta);
		ScaledLegendrePolynomial (order_inner-4, xi, 1-eta-zeta, polx);

		double val = gridvalues(ipnr) * (1-xi-eta-zeta) * (1+xi-eta-zeta);
		for (int k = 0; k <= order_inner-4; k++)
		  sumx(k) += val * polx[k];
	      }

	    for (int j = 0; j <= order-4; j++)
	      for (int k = 0; k <= order-4; k++)
		sumyx(j, k) += sumx(k) * poly[j];
	  }
	
	int ii = first_inner;
	for (int i = 0; i <= order_inner-4; i++)
	  for (int j = 0; j <= order_inner-4-i; j++)
	    for (int k = 0; k <= order_inner-4-i-j; k++)
	      coefs(ii++) += sumyx(j,k) * polz[i];

      }
    */



    // inner dofs
    ipnr = 0;
    for (int iz = 0; iz < ir.GetNIPZ(); iz++)
      {
	double z = ir.GetIPZ(iz);
	JacobiPolynomial (order_inner-4, 2*z-1, order_inner, 0, polz);
	for (int k = 0; k < polz.Size(); k++) 
	  polz[k] *= z;

	sumyx = 0.0;
	for (int iy = 0; iy < ir.GetNIPY(); iy++)
	  {
	    double y = ir.GetIPY(iy);
	    double eta = y * (1-z);
	    JacobiPolynomial (order_inner-4, 2*y-1, order_inner, 0, poly); 

	    double fac = y*(1-z);
	    for (int k = 0; k < poly.Size(); k++, fac *= 1-z) 
	      poly[k] *= fac;


	    sumx = 0.0;
	    for (int k = 0; k <= order_inner-4; k++) 
	      {
		double val = 0;
		for (int ix = 0; ix < ir.GetNIPX(); ix++)
		  val += gridvalues(ipnr+ix) * legxbubT(k, ix); 
		sumx(k) = val;
	      }
	    ipnr += ir.GetNIPX();
	    

	    fac = (1-y)*(1-y)*(1-z)*(1-z);
	    for (int k = 0; k <= order_inner-4; k++, fac *= (1-y)*(1-z))
	      sumx(k) *= fac;

	    for (int j = 0; j <= order-4; j++)
	      for (int k = 0; k <= order-4; k++)
		sumyx(j, k) += sumx(k) * poly[j];
	  }
	
	int ii = first_inner;
	for (int i = 0; i <= order_inner-4; i++)
	  for (int j = 0; j <= order_inner-4-i; j++)
	    for (int k = 0; k <= order_inner-4-i-j; k++)
	      coefs(ii++) += sumyx(j,k) * polz[i];
      }



    lh.CleanUp (heapp);
  }
  






























  
  template class H1HighOrderTet<IntegratedLegendreMonomialExt>;

  /*
  template class H1HighOrderTet<TrigExtensionMonomial>;
  template class H1HighOrderTet<TrigExtensionOptimal>;
  template class H1HighOrderTet<TrigExtensionMin>;
  */
}
