#ifndef FILE_H1HOFETP
#define FILE_H1HOFETP

/*********************************************************************/
/* File:   h1hofetp.hpp                                              */
/* Author: Start                                                     */
/* Date:   11 June 2007                                              */
/*********************************************************************/

/**
  High order finite elements for H^1  in full Tensor Product basis
*/


#define JACOBI



class IntegrationRuleTP
{
  const IntegrationRule *irx, *iry, *irz;
  int sort[8];

  ArrayMem<Vec<3>, 100> xi;
  ArrayMem<double, 100> weight;
  ArrayMem<Vec<3>, 100> x;
  ArrayMem<Mat<3,3>, 100> dxdxi;
  ArrayMem<Mat<3,3>, 100> dxdxi_duffy;
  
public:
  IntegrationRuleTP (const FiniteElement & afel, 
                     const ElementTransformation & eltrans,
                     int order, LocalHeap & lh);


  const IntegrationRule & GetIRX() const { return *irx; }
  const IntegrationRule & GetIRY() const { return *iry; }
  const IntegrationRule & GetIRZ() const { return *irz; }

  int GetNIP() const { return xi.Size(); }
  double GetWeight (int i) const { return weight[i]; }
  const Vec<3> GetXi(int i) const { return xi[i]; }
  const Vec<3> GetPoint(int i) const { return x[i]; }
  const Mat<3,3> & GetJacobian(int i) const { return dxdxi[i]; }
  const Mat<3,3> & GetDuffyJacobian(int i) const { return dxdxi_duffy[i]; }
};







class H1HighOrderTrigTP : public H1HighOrderTrig<>
{
public:
  H1HighOrderTrigTP (int aorder)
    : H1HighOrderTrig<> (aorder) {;}
  
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename TFA>  
  void T_CalcShape (Tx x, Ty y, TFA & shape) const; 

  void GetTrig2TensorMapping(FlatArray<int[3]> trig2tensor) const; 

  int GetTensor2SegmMapping(FlatArray<int[3] > trig2tensor, FlatArray<int> tensor2segm) const; 
};


class H1HighOrderQuadTP : public H1HighOrderQuad<>
{

public:
  H1HighOrderQuadTP (int aorder)
    : H1HighOrderQuad<> (aorder) { ; }

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename TFA>  
  void T_CalcShape (Tx x, Ty y, TFA & shape) const; 

  void GetQuad2TensorMapping (FlatArray<int[2]> quad2tensor, FlatArray<double> fac) const;
};











class H1HighOrderTetTP : public H1HighOrderTet<>
{
  int sort[4];

  FlatArray<int[4]> tet2tensor;
  FlatArray<int[3]> trig2tensor;
  FlatArray<int[2]> segm2tensor;
  FlatArray<int> map3dto2d;
  FlatArray<int> map2dto1d;

  FlatArray<int> map2dto3d;
  FlatArray<int> pmap2dto3d;
  FlatArray<int> map1dto2d;
  FlatArray<int> pmap1dto2d;

public:
  H1HighOrderTetTP (int aorder) : H1HighOrderTet<> (aorder) { ; }

  /// builds tables
  virtual void SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh);

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  int Sort (int i) const { return sort[i]; }

  FlatArray<int[4]> GetTet2TensorMapping () const { return tet2tensor; }
  FlatArray<int[3]> GetTrig2TensorMapping () const { return trig2tensor; }
  FlatArray<int[2]> GetSegm2TensorMapping () const { return segm2tensor; }

  FlatArray<int> Get3dTo2dMapping () const { return map3dto2d; }
  FlatArray<int> Get2dTo1dMapping () const { return map2dto1d; }
  int GetNDof2d () const { return trig2tensor.Size(); }
  int GetNDof1d () const { return segm2tensor.Size(); }

  FlatArray<int> Get2dTo3dMapping () const { return map2dto3d; }
  FlatArray<int> Get2dTo3dMappingPointer () const { return pmap2dto3d; }
  FlatArray<int> Get1dTo2dMapping () const { return map1dto2d; }
  FlatArray<int> Get1dTo2dMappingPointer () const { return pmap1dto2d; }


  // shape[i] = facx[i] * facy[3t2[i]] * facz[2t1[3t2[i]]]

  template <typename Tx, typename Tres>
  void CalcXFactor (Tx x,  Tres & facx, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max (2, order);
    FlatMatrix<Tx> polsx(order+1, horder, lh);

    Tx hv = 1;
    for (int i = 0; i <= order; i++)
      {
#ifdef JACOBI
        JacobiPolynomialMult (horder-2, 2*x-1, max(0,2*i-2), 0, x*hv, polsx.Row(i).Addr(1));
#else
        LegendrePolynomialMult (horder-2, 2*x-1, x*hv, polsx.Row(i).Addr(1));
#endif
        polsx(i,0) = hv;
        hv *= 1-x;
      }

    for (int jj = 0; jj < ndof; jj++)
      facx[jj] = polsx(tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3],
                       tet2tensor[jj][0]); 
  }

  template <typename Ty, typename Tres>
  void CalcYFactor (Ty y, Tres & facy, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max (2, order);
    FlatMatrix<Ty> polsy(order+1, horder, lh);

    Ty hv = 1.0;
    for (int i = 0; i <= order; i++)
      {
#ifdef JACOBI
        JacobiPolynomialMult (horder-2, 2*y-1, max(0,2*i-2), 0, y*hv, polsy.Row(i).Addr(1));
#else
        LegendrePolynomialMult (horder-2, 2*y-1, y*hv, polsy.Row(i).Addr(1));
#endif
        polsy(i,0) = hv;
        hv *= 1-y;
      }

    for (int jj = 0; jj < trig2tensor.Size(); jj++)
      facy[jj] = polsy(trig2tensor[jj][1]+trig2tensor[jj][2],
                       trig2tensor[jj][0]); 
  }

  template <typename Tz, typename Tres>
  void CalcZFactor (Tz z, Tres & facz, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max (order, 2);
    FlatArray<Tz> polz(horder, lh);
    Vec<2, Tz> polzz;
    
    LegendrePolynomialMult (horder-2, 2*z-1, z, polz.Addr(1));
    polz[0] = 1;

    polzz(0) = 1;
    polzz(1) = (1-z); 

    for (int j1d = 0; j1d < segm2tensor.Size(); j1d++)
      facz[j1d] = polz[segm2tensor[j1d][0]] * polzz(segm2tensor[j1d][1]);
  }
  
};












class H1HighOrderPrismTP : public H1HighOrderPrism<>
{
  int sort[6];

  FlatArray<int[4]> prism2tensor;
  FlatArray<int[3]> trig2tensor;   // y-z quad
  FlatArray<int[2]> segm2tensor;   // z
  FlatArray<int> map3dto2d;
  FlatArray<int> map2dto1d;
  FlatArray<double> factor;


  FlatArray<int> map2dto3d;
  FlatArray<int> pmap2dto3d;
  FlatArray<int> map1dto2d;
  FlatArray<int> pmap1dto2d;

public:
  H1HighOrderPrismTP (int aorder)
    : H1HighOrderPrism<> (aorder) { ; }

  virtual void SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh);

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  // void GetPrism2TensorMapping (FlatArray<int[4]> prism2tensor, FlatArray<double> fac) const;
  int Sort (int i) const { return sort[i]; }

  FlatArray<int[4]> GetPrism2TensorMapping () const { return prism2tensor; }
  FlatArray<int[3]> GetQuad2TensorMapping () const { return trig2tensor; }
  FlatArray<int[2]> GetSegm2TensorMapping () const { return segm2tensor; }
  FlatArray<int> Get3dTo2dMapping () const { return map3dto2d; }
  FlatArray<int> Get2dTo1dMapping () const { return map2dto1d; }
  FlatArray<double> GetFactor () const { return factor; }

  int GetNDof2d () const { return trig2tensor.Size(); }
  // int GetNDof1d () const { return order+1; }
  int GetNDof1d () const { return segm2tensor.Size(); }


  FlatArray<int> Get2dTo3dMapping () const { return map2dto3d; }
  FlatArray<int> Get2dTo3dMappingPointer () const { return pmap2dto3d; }
  FlatArray<int> Get1dTo2dMapping () const { return map1dto2d; }
  FlatArray<int> Get1dTo2dMappingPointer () const { return pmap1dto2d; }



  // shape[i] = facx[i] * facy[3t2[i]] * facz[2t1[3t2[i]]]

  /*
  template <typename Tx, typename Tres>
  void CalcXFactor (Tx x, Tres & facx, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max(2, order);
    FlatArray<Tx> powx(order+1, lh);
    FlatMatrix<Tx> polsx(order+1, horder, lh);

    for (int i = 0; i <= order; i++)
      {
        Tx * hp = &polsx(i, 1);
#ifdef JACOBI
        JacobiPolynomial (horder-2, 2*x-1, max(0,2*i-2), 0, hp);
#else
        LegendrePolynomial (horder-2, 2*x-1, hp);
#endif
        for (int j = 1; j < horder; j++) polsx(i,j) *= x;
        polsx(i,0) = 1.0;
      }

    Tx hv = 1.0;
    for (int i = 0; i <= order; i++, hv *= (1-x))
      powx[i] = hv;
    
    for (int jj = 0; jj < ndof; jj++)
      {
        int power = prism2tensor[jj][1]+prism2tensor[jj][2];
        facx[jj] = factor[jj] * polsx(power, prism2tensor[jj][0]) * powx[power];
      }
  }



  template <typename Ty, typename Tres>
  void CalcYFactor (Ty y, Tres & facy, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max(2, order);
    FlatVector<Ty> poly(horder, lh);
    Vec<2,Ty> polyy;

    Ty * hp = &poly(1);
    LegendrePolynomial (horder-2, 2*y-1, hp);
    for (int j = 1; j < horder; j++) poly(j) *= y;
    poly(0) = 1.0;

    polyy(0) = 1;
    polyy(1) = 1-y;
    
    int nd2d = trig2tensor.Size();
    for (int jj = 0; jj < nd2d; jj++)
      facy[jj] = poly(trig2tensor[jj][0]) * polyy(trig2tensor[jj][1]);
  }


  template <typename Tz, typename Tres>
  void CalcZFactor (Tz z, Tres & facz, LocalHeap & lh) const
  {
    HeapReset hr(lh);
    FlatArray<Tz> polz(order+1, lh);

    LegendrePolynomialMult (order-2, 2*z-1, z*(1-z), polz);
    for (int i = 2; i <= order; i++)
      facz[i] = polz[i-2];

    facz[0] = 1-z;
    facz[1] = z;
  }
  */


  template <typename Tz, typename Tres>
  void CalcXFactor (Tz z, Tres & facz, LocalHeap & lh) const
  {
    HeapReset hr(lh);
    FlatArray<Tz> polz(order+1, lh);

    LegendrePolynomialMult (order-2, 2*z-1, z*(1-z), polz.Addr(2));
    polz[0] = 1-z;
    polz[1] = z;

    for (int jj = 0; jj < ndof; jj++)
      {
        facz[jj] = factor[jj] * polz[prism2tensor[jj][3]];
      }

    // *testout << "xfactor = " << endl << facz << endl;
  }



  template <typename Tx, typename Tres>
  void CalcYFactor (Tx x, Tres & facx, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max(2, order);
    FlatArray<Tx> powx(order+1, lh);
    FlatMatrix<Tx> polsx(order+1, horder, lh);

    for (int i = 0; i <= order; i++)
      {
        Tx * hp = &polsx(i, 1);
#ifdef JACOBI
        JacobiPolynomial (horder-2, 2*x-1, max(0,2*i-2), 0, hp);
#else
        LegendrePolynomial (horder-2, 2*x-1, hp);
#endif
        for (int j = 1; j < horder; j++) polsx(i,j) *= x;
        polsx(i,0) = 1.0;
      }

    Tx hv = 1.0;
    for (int i = 0; i <= order; i++, hv *= (1-x))
      powx[i] = hv;
    
    int nd2d = trig2tensor.Size();
    for (int jj = 0; jj < nd2d; jj++)
      {
        int power = trig2tensor[jj][1]+trig2tensor[jj][2];
        facx[jj] = polsx(power, trig2tensor[jj][0]) * powx[power];
      }

    // *testout << "yfactor = " << endl << facx << endl;
  }



  template <typename Ty, typename Tres>
  void CalcZFactor (Ty y, Tres & facy, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    int horder = max(2, order);
    FlatVector<Ty> poly(horder, lh);
    Vec<2,Ty> polyy;

    Ty * hp = &poly(1);
    LegendrePolynomial (horder-2, 2*y-1, hp);
    for (int j = 1; j < horder; j++) poly(j) *= y;
    poly(0) = 1.0;

    polyy(0) = 1;
    polyy(1) = 1-y;
    
    int nd1d = segm2tensor.Size();
    for (int jj = 0; jj < nd1d; jj++)
      facy[jj] = poly(segm2tensor[jj][0]) * polyy(segm2tensor[jj][1]);

    // *testout << "zfactor = " << endl << facy << endl;
  }

};



class H1HighOrderHexTP : public H1HighOrderHex<>
{

  FlatArray<int[3]> hex2tensor;
  FlatArray<int[2]> quad2tensor;   // y-z quad
  FlatArray<int[1]> segm2tensor;   // z
  FlatArray<int> map3dto2d;
  FlatArray<int> map2dto1d;
  FlatArray<double> factor;


  FlatArray<int> map2dto3d;
  FlatArray<int> pmap2dto3d;
  FlatArray<int> map1dto2d;
  FlatArray<int> pmap1dto2d;

public:
  H1HighOrderHexTP (int aorder)
    : H1HighOrderHex<> (aorder) { ; }

  virtual void SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh);

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  // void GetHex2TensorMapping (FlatArray<int[3]> hex2tensor, FlatArray<double> fac) const;


  FlatArray<int[3]> GetHex2TensorMapping () const { return hex2tensor; }
  FlatArray<int[2]> GetQuad2TensorMapping () const { return quad2tensor; }
  FlatArray<int[1]> GetSegm2TensorMapping () const { return segm2tensor; }
  FlatArray<int> Get3dTo2dMapping () const { return map3dto2d; }
  FlatArray<int> Get2dTo1dMapping () const { return map2dto1d; }
  FlatArray<double> GetFactor () const { return factor; }

  int GetNDof2d () const { return sqr (order+1); }
  int GetNDof1d () const { return order+1; }


  FlatArray<int> Get2dTo3dMapping () const { return map2dto3d; }
  FlatArray<int> Get2dTo3dMappingPointer () const { return pmap2dto3d; }
  FlatArray<int> Get1dTo2dMapping () const { return map1dto2d; }
  FlatArray<int> Get1dTo2dMappingPointer () const { return pmap1dto2d; }


  // shape[i] = facx[i] * facy[3t2[i]] * facz[2t1[3t2[i]]]

  template <typename Tx, typename Tres>
  void CalcXFactor (Tx x, Tres & facx, LocalHeap & lh) const
  {
    HeapReset hr(lh);
    FlatVector<Tx> polx(order+1, lh);

    LegendrePolynomialMult (order-2, 2*x-1, x*(1-x), polx.Addr(2));
    polx[0] = 1-x;
    polx[1] = x;

    for (int jj = 0; jj < ndof; jj++)
      facx[jj] = factor[jj] * polx(hex2tensor[jj][0]);
  }

  template <typename Ty, typename Tres>
  void CalcYFactor (Ty y, Tres & facy, LocalHeap & lh) const
  {
    HeapReset hr(lh);
    FlatVector<Ty> poly(order+1, lh);

    LegendrePolynomialMult (order-2, 2*y-1, y*(1-y), poly.Addr(2));
    poly[0] = 1-y;
    poly[1] = y;

    int nd2d = (order+1)*(order+1);
    for (int jj = 0; jj < nd2d; jj++)
      facy[jj] = poly(jj/(order+1));
  }


  template <typename Tz, typename Tres>
  void CalcZFactor (Tz z, Tres & facz, LocalHeap & lh) const
  {
    HeapReset hr(lh);

    LegendrePolynomialMult (order-2, 2*z-1, z*(1-z), facz.Range(2,order+1));

    facz[0] = 1-z;
    facz[1] = z;
  }


};











#endif



