#ifndef FILE_DDH1HOFESPACE
#define FILE_DDH1HOFESPACE


class DDH1HighOrderFESpace : public FESpace
{
private:
  
  // Level
  int level;

  // Number of Edges
  int ned;
  // Number of Faces
  int nfa;
  // Number of Elements
  int nel;
  // Number of Vertex
  int nv;

  ARRAY<int> first_vert_dof;
  ARRAY<int> first_edge_dof;
  ARRAY<int> first_face_dof;
  ARRAY<int> first_element_dof;

  /// relative order to mesh-order
  int rel_order; 
  bool var_order; 
  

  ARRAY<int> order_edge;
  ARRAY<INT<2> > order_face;
  ARRAY<INT<3> > order_inner;
  ARRAY<int> order_avertex; 
  ARRAY<bool> fine_edge; 
  ARRAY<bool> fine_face; 

  int ndof;
  int uniform_order_inner;
  int uniform_order_face;
  int uniform_order_edge;
  int uniform_order_quad;
  int uniform_order_trig;
  ARRAY<INT<3> > dom_order_min; 
  ARRAY<INT<3> > dom_order_max;
  int smoother; 
  
  bool minext, optext;
  bool fast_pfem;
  bool plate;  // thin plate: prism trig dofs are internal
  bool print; 

  BitArray *usedvert, *usededge, *usedface, *usedel;

public:

  DDH1HighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags=false);
  ///
  virtual ~DDH1HighOrderFESpace ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const
  {
    return "DDH1HighOrderFESpace";
  }

  ///
  virtual void Update();
  ///
  virtual void PrintReport (ostream & ost);

  ///
  virtual int GetNDof () const;
  ///
  virtual int GetNDofLevel (int alevel) const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;

  virtual void GetWireBasketDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;

  ///
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  
  virtual Table<int> * CreateSmoothingBlocks (const Flags & precflags) const; 
  ///
  virtual ARRAY<int> * CreateDirectSolverClusters (int type = 0) const;

  int GetFirstVertexDof(int i) const {return first_vert_dof[i]; } ;
  ///
  int GetFirstFaceDof(int i) const {return(first_face_dof[i]);} ;  
  ///
  int GetFirstEdgeDof(int i) const {return(first_edge_dof[i]);} ; 
  ///
  int GetFirstElementDof(int i) const {return(first_element_dof[i]);} ; 

  void UpdateDofTables ();
  void SetEdgeOrder (int enr, int eo) { order_edge[enr] = eo; }
  void SetFaceOrder (int fnr, int fo) { order_face[fnr] = INT<2> (fo, fo); }
  void SetFaceOrder (int fnr, int ox, int oy) { order_face[fnr] = INT<2> (ox, oy); }
  void SetElementOrder (int elnr, int elo) 
  { order_inner[elnr] = INT<3> (elo, elo, elo); }
  void SetElementOrder (int elnr, int ox, int oy, int oz) 
  { order_inner[elnr] = INT<3> (ox, oy, oz); }

  /// get relative (to mesh) order of finite elements
  virtual int GetRelOrder() const { return rel_order; }
  virtual bool VarOrder() const { return var_order; }

  bool UsedVert( int i ) const { return usedvert->Test(i); }
  bool UsedEdge ( int ed ) const { return usededge->Test(ed); }
  bool UsedFace ( int fa ) const { return usedface->Test(fa); }
  bool UsedEl ( int el ) const { return usedel->Test(el); }

  void SetUsedVert ( int vert, bool isused = true ) 
  { 
    isused ? usedvert->Set(vert) : usedvert->Clear(vert); 
  }
  void SetUsedEdge ( int vert, bool isused = true ) 
  { 
    isused ? usededge->Set(vert) : usededge->Clear(vert); 
  }
  void SetUsedFace ( int vert, bool isused = true ) 
  { 
    isused ? usedface->Set(vert) : usedface->Clear(vert); 
  }
  void SetUsedEl ( int vert, bool isused = true ) 
  { 
    usedel ? usedel->Set(vert) : usedel->Clear(vert); 
  }

  virtual MatrixGraph * GetGraph (int level, bool symmetric);

#ifdef PARALLEL
   virtual void UpdateParallelDofs_hoproc();
   virtual void UpdateParallelDofs_loproc();
#endif

  int GetVertexDof( const int vert ) const
  {
    if ( first_vert_dof[vert+1] - first_vert_dof[vert] )
      return first_vert_dof[vert];
    else
      return -1;
  }

  int GetNVertexDofs() const
  { return first_vert_dof.Last(); }
};

#endif

