class FETIL2SurfaceHighOrderFESpace : public L2SurfaceHighOrderFESpace
{
protected:
  
  int ndomains;

public:

  FETIL2SurfaceHighOrderFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags=false);
  ///
  virtual ~FETIL2SurfaceHighOrderFESpace ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const
  {
    return "FETIL2SurfaceHighOrderFESpace";
  }

  ///
  virtual void Update();
  /// 
  //virtual void UpdateDofTables();
  ///
  virtual int GetNDof () const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  
  virtual Table<int> * CreateSmoothingBlocks ( int type = 0) const;

  /// 
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;

  virtual bool VarOrder() const { return var_order; } 
  virtual int GetRelOrder() const { return rel_order; }   


};



class FETILagrangeFESpace : public FESpace
{
protected:
  
  int ndof;
  int ndomains;
  ARRAY<INT<2> > domains;

  ARRAY<int> first_vertex_dof;
  ARRAY<int> first_edge_dof;
  ARRAY<int> first_face_dof;
public:

  FETILagrangeFESpace (const MeshAccess & ama, const Flags & flags, bool parseflags=false);
  ///
  virtual ~FETILagrangeFESpace ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const
  {
    return "FETILagrangeFESpace";
  }

  ///
  virtual void Update();
  /// 
  //virtual void UpdateDofTables();
  ///
  virtual int GetNDof () const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  
  /// 
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;

  INT<2> GetDofDomains (int dof)
  { return domains[dof];}
};



class FETICompoundFESpace : public CompoundFESpace
{
private:
  int ndomains;

public:
  FETICompoundFESpace (const MeshAccess & ama, 		   
		       const ARRAY<const FESpace*> & aspaces,
		       const Flags & aflags)
    : CompoundFESpace (ama, aspaces, aflags)
  {
    ndomains = ma.GetNDomains();
  }
  
  virtual ~FETICompoundFESpace () { ; }
  
  static FESpace *   Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const 
  { 
    return "FETICompoundFESpace"; 
  }

  int GetNDomains () const { return ndomains; }


};



class L2FETICompoundFESpace : public CompoundFESpace
{
private:
  int ndomains;

public:
  L2FETICompoundFESpace (const MeshAccess & ama, 		   
		       const ARRAY<const FESpace*> & aspaces,
		       const Flags & aflags)
    : CompoundFESpace (ama, aspaces, aflags)
  {
    ndomains = ma.GetNDomains();
  }
  
  virtual ~L2FETICompoundFESpace () { ; }
  
  static FESpace *   Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const 
  { 
    return "L2FETICompoundFESpace"; 
  }

  int GetNDomains () const { return ndomains; }


};


template<int D>
class FETIMassIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction * coeff;

  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 0 };

public:
  FETIMassIntegrator (CoefficientFunction * acoeff)
    : BilinearFormIntegrator()
  { 
    coeff = acoeff;
  }
  
  virtual ~FETIMassIntegrator () { ; }
  
  virtual bool BoundaryForm () const 
  { return 0; }
  
  virtual int DimFlux () const { return 1; }
  
  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const;


  virtual void
  CalcFlux (const FiniteElement & bfel,
	    const BaseSpecificIntegrationPoint & bsip,
	    const FlatVector<double> & elx, 
	    FlatVector<double> & flux,
	    bool applyd,
	    LocalHeap & lh) const;


  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new FETIMassIntegrator (coeffs[0]); 
  }
  
};

