class FETIHighOrderFiniteElement : public NodalFiniteElement
{
public:
  
  int vnums[8]; //he: make oriented FETI elements => can be used for facets

  INT<3> order_inner; 

  FETIHighOrderFiniteElement (int dim, ELEMENT_TYPE aeltype);

  void SetVertexNumbers (FlatArray<int> & avnums);

  void SetOrder (int o);
  
  void SetOrderInner (INT<3> oi);

  virtual void ComputeNDof () = 0; 

  // const int * GetVNums() const { return vnums; }

  virtual void GetInternalDofs (ARRAY<int> & idofs) const; 
  
private:
  
  virtual const ARRAY<IPData> & GetIPData () const
  {
    throw Exception ("GetIPData not available for FETIHighOrderFE");
  }
};

/**
  High order 1D finite element
*/
class FETIHighOrderSegm : public FETIHighOrderFiniteElement
{
public:
  FETIHighOrderSegm (int aorder=0);
  virtual void ComputeNDof();

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;
};


/**
  High order triangular finite element
*/
class FETIHighOrderTrig : public FETIHighOrderFiniteElement
{
public:
  FETIHighOrderTrig (int aorder=0);
  virtual void ComputeNDof();

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;
};


/**
  High order quadrilateral finite element
*/
class FETIHighOrderQuad : public FETIHighOrderFiniteElement
{
  //typedef TrigShapesInnerLegendre T_INNERSHAPES;
  // typedef TrigShapesInnerJacobi T_INNERSHAPES;
public:
  FETIHighOrderQuad (int aorder=0);
  virtual void ComputeNDof();

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  /// compute gradient of shape 
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const; 
};
