#include <solve.hpp>
#include <feti.hpp>
#include <sstream>


namespace ngsolve
{
  using namespace ngsolve;
  // using namespace ngbla;
  // using namespace ngla;
  using namespace std;


  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: CalcQ()
  {

    int ndof_lambda = uspace-> GetNDof();
    int nv = ma.GetNV();
    int order = uspace -> GetOrder();
    matQ.SetSize ( ndof_lambda );

    if ( identityQ )
      {
	matQ = 1.0;
	return;
      }


    for ( int vert = 0; vert < nv; vert++ )
      {
	ARRAY<int> dofs,elems;
	uspace->GetVertexDofNrs(vert, dofs);

	if ( !dofs.Size() ) continue;

	FlatArray<int> vert_domains = (*vert_doms)[vert];

	double H = 1.0;
	double h;

	ma.GetVertexElements(vert, elems);

	if ( D == 2 )
	  h = sqrt(ma.ElementVolume (elems[0]));
	else
	  h = pow(ma.ElementVolume (elems[0]), 1./3.);

	// in facet
	if ( vert_domains.Size() == 2 )
	  {
	    matQ(dofs[0]) = min2 ( (*laplacecoeff)[vert_domains[0]], (*laplacecoeff)[vert_domains[1]] );
	    matQ(dofs[0]) *= (1 + order*log(H / h)) * h;
	    if ( D == 3 ) matQ(dofs[0]) *= h;
	  }
	else
	  {
	    int ii = 0;
	    for ( int i1 = 0; i1 < vert_domains.Size(); i1++ )
	      for ( int i2 = i1+1; i2 < vert_domains.Size(); i2++)
		{
		  matQ(dofs[ii]) = min2 ( (*laplacecoeff)[vert_domains[i1]], (*laplacecoeff)[vert_domains[i2]] );
		  if ( D == 3 ) matQ(dofs[ii]) *= h;
		  ii++;
		}
	  }

      }


    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	ARRAY<int> dofs,elems;
	uspace->GetEdgeDofNrs(edge, dofs);

	if ( !dofs.Size() ) continue;

	FlatArray<int> edge_domains = (*edge_doms)[edge];

	double H = 1.0;
	double h;

	ma.GetEdgeElements(edge, elems);

	if ( D == 2 )
	  h = sqrt(ma.ElementVolume (elems[0]));
	else
	  h = pow(ma.ElementVolume (elems[0]), 1./3.);

	// compute local order from number of dofs --> better method in fespace...
	int ndom = edge_domains.Size();
	int local_order = int(2.0 * dofs.Size() / ndom / (ndom-1) + 1.0);
	// in facet
	if ( edge_domains.Size() == 2 )
	  for ( int i = 0; i < dofs.Size(); i++ )
	    {
	      matQ(dofs[i]) = min2 ( (*laplacecoeff)[edge_domains[0]], (*laplacecoeff)[edge_domains[1]] );
	      matQ(dofs[i]) *= (1 + local_order * ( H / h)) * h;
	      if ( D == 3 ) matQ(dofs[0]) *= h;
	    }
	else
	  {
	    int ii = 0;
	    for ( int i1 = 0; i1 < edge_domains.Size(); i1++ )
	      for ( int i2 = i1+1; i2 < edge_domains.Size(); i2++)
		for ( int i = 0; i < local_order-1; i++ )
		  {
		    matQ(dofs[ii]) = min2 ( (*laplacecoeff)[edge_domains[i1]], (*laplacecoeff)[edge_domains[i2]] );
		    if ( D == 3 ) matQ(dofs[ii]) *= h;
		    ii++;
		  }
	  }

      }

    if ( D == 3 )
      for ( int face = 0; face < ma.GetNFaces(); face++ )
      {
	ARRAY<int> dofs,elems;
	uspace->GetFaceDofNrs(face, dofs);

	if ( !dofs.Size() ) continue;

	FlatArray<int> face_domains = (*face_doms)[face];

	double H = 1.0;
	double h;

	ma.GetFaceElements(face, elems);

	if ( D == 2 )
	  h = sqrt(ma.ElementVolume (elems[0]));
	else
	  h = pow(ma.ElementVolume (elems[0]), 1./3.);

	int local_order = int( 3./2. + sqrt( 9./4. - 2. + 2*dofs.Size() ) + 0.01 );
	// in facet -- face is always in facet
	if ( face_domains.Size() == 2 )
	  for ( int i = 0; i < dofs.Size(); i++ )
	    {
	      matQ(dofs[i]) = min2 ( (*laplacecoeff)[face_domains[0]], (*laplacecoeff)[face_domains[1]] );
	      matQ(dofs[i]) *= (1 + log(H / h)) * h;
	      if ( D == 3 ) matQ(dofs[0]) *= h;
	    }
// 	else
// 	  {
// 	    int ii = 0;
// 	    for ( int i1 = 0; i1 < face_domains.Size(); i1++ )
// 	      for ( int i2 = i1+1; i2 < face_domains.Size(); i2++)
// 		{
// 		  matQ(dofs[ii]) = min2 ( (*laplacecoeff)[face_domains[i1]], (*laplacecoeff)[face_domains[i2]] );
// 		  if ( D == 3 ) matQ(dofs[ii]) *= h;
// 		  ii++;
// 		}
// 	  }

      }

  }


  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  ApplyQ ( const BaseVector & lambda, BaseVector & mu )
  {
    int base = compoundspace->GetStorageStart(ndomains);
    int nd = compoundspace->GetNDof();
    int nd_lambda = uspace -> GetNDof();
    const S_BaseVector<SCALAR> & slambda = dynamic_cast<const S_BaseVector<SCALAR> & > (lambda);
    S_BaseVector<SCALAR> & smu = dynamic_cast<S_BaseVector<SCALAR> & > (mu);

    for ( int i = base; i < nd; i++ )
      {
	smu.FVScal()(i) = matQ(i - base) * slambda.FVScal()(i);
      }
  }


  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  CalcInvGtQG ()
  {
    matGtQG.SetSize(ndomains);
    matGtQG = SCALAR(0.0);

    Vector<SCALAR> ei(ndomains);
    BaseVector & vechelp = (*gfu->GetVector().CreateVector());

    for ( int i = 0; i < ndomains; i++ )
      {
	ei = 0;
	ei(i) = 1.0;
	if ( isfloating.Test(i) ) 
	  {
	    ApplyG(ei, vechelp);
	    ApplyQ(vechelp, vechelp);
	    ApplyGTrans(vechelp, ei);
	    for ( int j = 0; j < ndomains; j++ )
	      {
		if ( isfloating.Test(j) ) 
		  matGtQG(i,j) = ei(j);
	      }
	  }
	else
	  matGtQG(i,i) = 1.0;

      }

    invGtQG.SetSize(ndomains);
    CalcInverse (matGtQG , invGtQG);
    
    delete &vechelp;

  }


  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  Precond ( BaseVector & vecu, BaseVector & vecv)
  {
    ApplyBTrans ( vecu , vecu );
    ApplyS ( vecu, vecv );
    ApplyB ( vecv, vecv );
//     ApplyBarBTrans ( vecu , vecu );
//     ApplyS ( vecu, vecv );
//     ApplyBarB ( vecv, vecv );
  }


  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  ApplyOp ( BaseVector & vecu, BaseVector & vecv)
  {
    ApplyBTrans ( vecu, vecu );
    ApplyInvS ( vecu, vecv );
    ApplyB ( vecv, vecv );
  }


  // B : u \mapsto lambda
  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  ApplyB ( const BaseVector & vecu, BaseVector & vecv, bool bar)
  {
    int nsel = ma.GetNSE();
    int nv = ma.GetNV();
    int dom1, dom2;

    ARRAY<int> dofsv1, dofsv2, dofslambda;
    int basev1, basev2, baselambda;
    int nd_v1, nd_v2, nd_lambda, nd;
    
    ARRAY<int> vert_domains, elems;
    
    const FlatVector<SCALAR> vecus = dynamic_cast<const S_BaseVector<SCALAR> & > (vecu) . FVScal();
    FlatVector<SCALAR> vecvs = dynamic_cast<S_BaseVector<SCALAR> & > (vecv) .FVScal();

    nd = compoundspace->GetNDof();
    baselambda = compoundspace->GetStorageStart(ndomains);

    vecvs.Range(baselambda, nd ) = 0.0;

    for ( int vert = 0; vert < nv; vert++ )
      {
	uspace->GetVertexDofNrs (vert, dofslambda );
	nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	FlatArray<int> vert_domains = (*vert_doms)[vert];

	/* coefficient ... 1 if ApplyB, c_j / sum c_i if ApplyBarB */
	SCALAR coeff = bar ? 0.0 : 1.0;
	for ( int i = 0; bar && i < vert_domains.Size(); i++ )
	  coeff += (*laplacecoeff)[vert_domains[i]] ;
	coeff = 1.0 / coeff;

	int ii = 0;
	for ( int i_dom1 = 0; i_dom1 < vert_domains.Size(); i_dom1++ )
	  for ( int i_dom2 = i_dom1+1; i_dom2 < vert_domains.Size(); i_dom2++ )
	    {

	      int dom1 = vert_domains[i_dom1];
	      int dom2 = vert_domains[i_dom2];
	      
	      const DDH1HighOrderFESpace * vspace1 = 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom1]);
	      const DDH1HighOrderFESpace * vspace2= 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom2]);
	      
	      vspace1->GetVertexDofNrs(vert, dofsv1);
	      vspace2->GetVertexDofNrs(vert, dofsv2);

	      basev1 = compoundspace->GetStorageStart(dom1);
	      basev2 = compoundspace->GetStorageStart(dom2);
	      SCALAR v1value = vecus(basev1 + dofsv1[0]);
	      SCALAR v2value = vecus(basev2 + dofsv2[0]);
	      SCALAR diffuvalue = bar ? 
		(*laplacecoeff)[dom2] * coeff * v1value - (*laplacecoeff)[dom1] * coeff * v2value :
		v1value - v2value;

	      vecvs(baselambda + dofslambda[ii]) = diffuvalue;
	      ii++;
	    }
      }


    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	uspace->GetEdgeDofNrs (edge, dofslambda );
	nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	FlatArray<int> edge_domains = (*edge_doms)[edge];

	/* coefficient ... 1 if ApplyB, c_j / sum c_i if ApplyBarB */
	SCALAR coeff = bar ? 0.0 : 1.0;
	for ( int i = 0; bar && i < edge_domains.Size(); i++ )
	  coeff += (*laplacecoeff)[edge_domains[i]] ;
	coeff = 1.0 / coeff;

	int ii = 0;
	for ( int i_dom1 = 0; i_dom1 < edge_domains.Size(); i_dom1++ )
	  for ( int i_dom2 = i_dom1+1; i_dom2 < edge_domains.Size(); i_dom2++ )
	    {

	      int dom1 = edge_domains[i_dom1];
	      int dom2 = edge_domains[i_dom2];
	      
	      const DDH1HighOrderFESpace * vspace1 = 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom1]);
	      const DDH1HighOrderFESpace * vspace2= 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom2]);
	      
	      vspace1->GetEdgeDofNrs(edge, dofsv1);
	      vspace2->GetEdgeDofNrs(edge, dofsv2);

	      basev1 = compoundspace->GetStorageStart(dom1);
	      basev2 = compoundspace->GetStorageStart(dom2);
	      for ( int dof = 0; dof < dofsv1.Size(); dof++ )
		{
		  SCALAR v1value = vecus(basev1 + dofsv1[dof]);
		  SCALAR v2value = vecus(basev2 + dofsv2[dof]);
		  SCALAR diffuvalue = bar ? 
		    (*laplacecoeff)[dom2] * coeff * v1value - (*laplacecoeff)[dom1] * coeff * v2value :
		    v1value - v2value;

		  vecvs(baselambda + dofslambda[ii]) = diffuvalue;
		  ii++;
		}
	    }
      }


    if ( D == 3 )
      for ( int face = 0; face < ma.GetNFaces(); face++ )
	{
	  uspace->GetFaceDofNrs (face, dofslambda );
	  nd_lambda = dofslambda.Size();
	  if ( !nd_lambda ) continue;

	  FlatArray<int> face_domains = (*face_doms)[face];

	  /* coefficient ... 1 if ApplyB, c_j / sum c_i if ApplyBarB */
	  SCALAR coeff = bar ? 0.0 : 1.0;
	  for ( int i = 0; bar && i < face_domains.Size(); i++ )
	    coeff += (*laplacecoeff)[face_domains[i]] ;
	  coeff = 1.0 / coeff;
	  
	  int ii = 0;
	  for ( int i_dom1 = 0; i_dom1 < face_domains.Size(); i_dom1++ )
	    for ( int i_dom2 = i_dom1+1; i_dom2 < face_domains.Size(); i_dom2++ )
	      {

		int dom1 = face_domains[i_dom1];
		int dom2 = face_domains[i_dom2];
	      
		const DDH1HighOrderFESpace * vspace1 = 
		  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom1]);
		const DDH1HighOrderFESpace * vspace2= 
		  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom2]);
	      
		vspace1->GetFaceDofNrs(face, dofsv1);
		vspace2->GetFaceDofNrs(face, dofsv2);

		basev1 = compoundspace->GetStorageStart(dom1);
		basev2 = compoundspace->GetStorageStart(dom2);
		for ( int dof = 0; dof < dofsv1.Size(); dof++ )
		  {
		    SCALAR v1value = vecus(basev1 + dofsv1[dof]);
		    SCALAR v2value = vecus(basev2 + dofsv2[dof]);
		    SCALAR diffuvalue = bar ? 
		      (*laplacecoeff)[dom2] * coeff * v1value - (*laplacecoeff)[dom1] * coeff * v2value :
		      v1value - v2value;

		    vecvs(baselambda + dofslambda[ii]) = diffuvalue;
		    ii++;
		  }
	      }
	}

  }





  template <int D, class SCALAR>
  void  NumProcFETIBVP<D, SCALAR> :: 
  ApplyBTrans ( const BaseVector & vecu, BaseVector & vecv, bool bar)
  {

    int nsel = ma.GetNSE();
    int nv = ma.GetNV();
    int dom1, dom2;

    ARRAY<int> dofsv1, dofsv2, dofslambda;
    int basev1, basev2, baselambda;
    int nd_v1, nd_v2, nd_lambda, nd;
    
    ARRAY<int> vert_domains, elems;
    
    const FlatVector<SCALAR> vecus = dynamic_cast<const S_BaseVector<SCALAR> & > (vecu) . FVScal();
    FlatVector<SCALAR> vecvs = dynamic_cast<S_BaseVector<SCALAR> & > (vecv) .FVScal();

    nd = compoundspace->GetNDof();
    baselambda = compoundspace->GetStorageStart(ndomains);

    vecvs.Range(0, baselambda ) = 0.0;

    for ( int vert = 0; vert < nv; vert++ )
      {
	uspace->GetVertexDofNrs (vert, dofslambda );
	nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	FlatArray<int> vert_domains = (*vert_doms)[vert];

	int ii = 0;
	/* coefficient ... 1 if ApplyB, c_j / sum c_i if ApplyBarB */
	SCALAR coeff = bar ? 0.0 : 1.0;
	for ( int i = 0; bar && i < vert_domains.Size(); i++ )
	  coeff += (*laplacecoeff)[vert_domains[i]] ;
	coeff = 1.0 / coeff;

	for ( int i_dom1 = 0; i_dom1 < vert_domains.Size(); i_dom1++ )
	  for ( int i_dom2 = i_dom1+1; i_dom2 < vert_domains.Size(); i_dom2++ )
	    {

	      int dom1 = vert_domains[i_dom1];
	      int dom2 = vert_domains[i_dom2];
	      
	      const DDH1HighOrderFESpace * vspace1 = 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom1]);
	      const DDH1HighOrderFESpace * vspace2= 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom2]);
	      
	      vspace1->GetVertexDofNrs(vert, dofsv1);
	      vspace2->GetVertexDofNrs(vert, dofsv2);

	      basev1 = compoundspace->GetStorageStart(dom1);
	      basev2 = compoundspace->GetStorageStart(dom2);


	      SCALAR lambdavalue = vecus(baselambda + dofslambda[ii]);

	      vecvs(basev1 + dofsv1[0]) += bar ? (*laplacecoeff)[dom2] * coeff * lambdavalue : lambdavalue;
	      vecvs(basev2 + dofsv2[0]) -= bar ? (*laplacecoeff)[dom1] * coeff * lambdavalue : lambdavalue;

	      ii++;
	    }
      }

    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	uspace->GetEdgeDofNrs (edge, dofslambda );
	nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	FlatArray<int> edge_domains = (*edge_doms)[edge];

	/* coefficient ... 1 if ApplyB, c_j / sum c_i if ApplyBarB */
	SCALAR coeff = bar ? 0.0 : 1.0;
	for ( int i = 0; bar && i < edge_domains.Size(); i++ )
	  coeff += (*laplacecoeff)[edge_domains[i]] ;
	coeff = 1.0 / coeff;
 
	int ii = 0;
	for ( int i_dom1 = 0; i_dom1 < edge_domains.Size(); i_dom1++ )
	  for ( int i_dom2 = i_dom1+1; i_dom2 < edge_domains.Size(); i_dom2++ )
	    {

	      int dom1 = edge_domains[i_dom1];
	      int dom2 = edge_domains[i_dom2];
	      
	      const DDH1HighOrderFESpace * vspace1 = 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom1]);
	      const DDH1HighOrderFESpace * vspace2= 
		dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom2]);
	      
	      vspace1->GetEdgeDofNrs(edge, dofsv1);
	      vspace2->GetEdgeDofNrs(edge, dofsv2);

	      basev1 = compoundspace->GetStorageStart(dom1);
	      basev2 = compoundspace->GetStorageStart(dom2);
	      for ( int dof = 0; dof < dofsv1.Size(); dof++ )
		{

		  SCALAR lambdavalue = vecus(baselambda + dofslambda[ii]);
		  
		  vecvs(basev1 + dofsv1[dof]) += bar ? (*laplacecoeff)[dom2] * coeff * lambdavalue : lambdavalue;
		  vecvs(basev2 + dofsv2[dof]) -= bar ? (*laplacecoeff)[dom1] * coeff * lambdavalue : lambdavalue;
		  ii++;
		}
	    }
      }

    if ( D == 3 )
      for ( int face = 0; face < ma.GetNFaces(); face++ )
	{
	  uspace->GetFaceDofNrs (face, dofslambda );
	  nd_lambda = dofslambda.Size();
	  if ( !nd_lambda ) continue;

	  FlatArray<int> face_domains = (*face_doms)[face];

	  /* coefficient ... 1 if ApplyB, c_j / sum c_i if ApplyBarB */
	  SCALAR coeff = bar ? 0.0 : 1.0;
	  for ( int i = 0; bar && i < face_domains.Size(); i++ )
	    coeff += (*laplacecoeff)[face_domains[i]] ;
	  coeff = 1.0 / coeff;

	  int ii = 0;
	  for ( int i_dom1 = 0; i_dom1 < face_domains.Size(); i_dom1++ )
	    for ( int i_dom2 = i_dom1+1; i_dom2 < face_domains.Size(); i_dom2++ )
	      {

		int dom1 = face_domains[i_dom1];
		int dom2 = face_domains[i_dom2];
	      
		const DDH1HighOrderFESpace * vspace1 = 
		  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom1]);
		const DDH1HighOrderFESpace * vspace2= 
		  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom2]);
	      
		vspace1->GetFaceDofNrs(face, dofsv1);
		vspace2->GetFaceDofNrs(face, dofsv2);

		basev1 = compoundspace->GetStorageStart(dom1);
		basev2 = compoundspace->GetStorageStart(dom2);
		for ( int dof = 0; dof < dofsv1.Size(); dof++ )
		  {
		    SCALAR lambdavalue = vecus(baselambda + dofslambda[ii]);
		    
		    vecvs(basev1 + dofsv1[dof]) += bar ? (*laplacecoeff)[dom2] * coeff * lambdavalue : lambdavalue;
		    vecvs(basev2 + dofsv2[dof]) -= bar ? (*laplacecoeff)[dom1] * coeff * lambdavalue : lambdavalue;
		    ii++;
		  }
	      }
	}

  }  


//   template <int D, class SCALAR>
//   void  NumProcFETIBVP<D, SCALAR> :: 


  // v = S u
  template <int D, class SCALAR>
  void  NumProcFETIBVP<D, SCALAR> :: ApplyS ( const  BaseVector & u, BaseVector & v )
  {
    if ( &u == &v ) 
      { 
	cerr << "ApplyS(u, v) called with u == v !!!!!!!!!!!! " << endl;
	exit(0);
      }
    BaseVector & help =  (*v.CreateVector());
    help = 0.0;
    v = 0.0;

    for ( int i = 0; i < ndomains; i++ )
      {
	int ndof = (*compoundspace)[i]->GetNDof();
	int basedomi = compoundspace->GetStorageStart(i);
	int basenextdom = basedomi + ndof;

	auto_ptr<BaseVector> helpi = help.Range(basedomi, basenextdom);
	auto_ptr<BaseVector> ui = u.Range(basedomi, basenextdom);
	auto_ptr<BaseVector> vi = v.Range(basedomi, basenextdom);

	FlatVector<SCALAR> helpis = dynamic_cast<S_BaseVector<SCALAR> & > (*helpi) . FVScal();
	FlatVector<SCALAR> vis = dynamic_cast<S_BaseVector<SCALAR> & > (*vi) . FVScal();

	const SparseMatrix<SCALAR> & mata = dynamic_cast<const SparseMatrix<SCALAR> &> (bfa[i]->GetMatrix() );

	// ui = (0, ui_gamma )
	// vi = ( B^T ui_gamma, C ui_gamma )
	*vi = mata * (*ui ) ;

	// helpi = (invA B^T ui_gamma, 0 )
	// vi = (0, C ui_gamma )

	*helpi = (*invAinner[i]) * (*vi); // (*help2i);
	for ( int index = 0; index < ndof; index++ )
	  if ( innerdofs[i]->Test(index) ) vis(index) = 0.0;
	  else helpis(index) = 0.0;

	// helpi = (0, B invA B^T ui_gamma )
	// vi = (0, (C - B invA B^T) ui_gamma )
	*vi -= mata * (*helpi) ; 

	for ( int index = 0; index < ndof; index++ )
	  if ( innerdofs[i]->Test(index) ) vis(index) = 0.0;
      }
    int ndof = (*compoundspace)[ndomains]->GetNDof();
    int basedomi = compoundspace->GetStorageStart(ndomains);
    int basenextdom = basedomi + ndof;

 
    delete & help;

  }


  // v = S u
  template <int D, class SCALAR>
  void  NumProcFETIBVP<D, SCALAR> :: ApplyInvS ( const  BaseVector & u, BaseVector & v )
  {
    if ( &u == &v ) 
      { 
	cerr << "ApplyInvS(u, v) called with u == v !!!!!!!!!!!! " << endl;
	exit(0);
      }

    for ( int i = 0; i < ndomains; i++ )
      {
	int ndof = (*compoundspace)[i]->GetNDof();
	int basedomi = compoundspace->GetStorageStart(i);
	int basenextdom = basedomi + (*compoundspace)[i]->GetNDof();

	*v.Range(basedomi, basenextdom) = (*invA[i]) * (*u.Range(basedomi, basenextdom)) ;
      }

  }

  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  ApplyR ( const Vector<SCALAR> & alpha, BaseVector & vecu)
  {
    FlatVector<SCALAR> vecus = dynamic_cast<S_BaseVector<SCALAR> & > (vecu) . FVScal();

    vecus.Range(0, compoundspace->GetStorageStart(ndomains) ) = 0.0;

    for ( int i = 0;  i < ndomains; i++ )
      {
	if ( !isfloating.Test(i) ) continue;

	int base = compoundspace -> GetStorageStart(i);
	ARRAY<int> dnums;

	const DDH1HighOrderFESpace * vspace = 
	  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[i]);

	for ( int vert = 0; vert < ma.GetNV(); vert++ )
	  {
	    vspace -> GetVertexDofNrs ( vert, dnums );

	    if ( !dnums.Size() ) continue;

	    if ( innerdofs[i]->Test(dnums[0]) || !nondcdofs[i]->Test(dnums[0]) ) continue;

	    for ( int ii = 0; ii < dnums.Size(); ii++ )
	      vecus(base + dnums[ii]) = alpha(i);

	  }


      }
  }

  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: ApplyRTrans ( const  BaseVector & vecu, Vector<SCALAR> & alpha)
  {
    alpha = 0;
    const FlatVector<SCALAR> vecus = dynamic_cast<const S_BaseVector<SCALAR> & > (vecu) . FVScal();

    for ( int i = 0;  i < ndomains; i++ )
      {
	if ( !isfloating.Test(i) ) continue;

	int base = compoundspace -> GetStorageStart(i);
	ARRAY<int> dnums;

	const DDH1HighOrderFESpace * vspace = 
	  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[i]);

	for ( int vert = 0; vert < ma.GetNV(); vert++ )
	  {
	    vspace -> GetVertexDofNrs ( vert, dnums );

	    if ( !dnums.Size() ) continue;

	    if ( innerdofs[i]->Test(dnums[0]) || !nondcdofs[i]->Test(dnums[0]) ) continue;

	    for ( int ii = 0; ii < dnums.Size(); ii++ )
	      alpha(i) += vecus(base + dnums[ii]);

	  }


      }

//     bool allfloating = true;
//     for ( int i = 0;  i < ndomains; i++ )
//       if ( !isfloating.Test(i) ) allfloating = false;

//     if ( allfloating )
//       {
// 	double mv = 0;
// 	for ( int i = 0; i < ndomains; i++ )
// 	  mv += alpha(i);
// 	mv /= ndomains;
// 	for ( int i = 0; i < ndomains; i++ )
// 	  alpha(i) -= mv;
//       }
  }


  /*
    Nf = f_G - A_GI invA_II f_I
  */
  template <int D, class SCALAR>
  void NumProcFETIBVP<D, SCALAR> :: 
  NewtonPotential ( const BaseVector & vecf, BaseVector & vecNf )
  {
    if ( &vecNf == &vecf )
      {
	cerr << " NewtonPotential ( f, Nf ) called with f == Nf!!!!!" << endl;
	exit(0);
      }

    BaseVector & help = (*vecNf.CreateVector());
    help = 0.0;
    vecNf = 0.0;

    for ( int i = 0; i < ndomains; i++ )
      {
	int ndof = (*compoundspace)[i]->GetNDof();
	int basedomi = compoundspace->GetStorageStart(i);
	int basenextdom = basedomi + ndof;

	auto_ptr<BaseVector> helpi = help.Range(basedomi, basenextdom);
	auto_ptr<BaseVector> ui = vecf.Range(basedomi, basenextdom);
	auto_ptr<BaseVector> vi = vecNf.Range(basedomi, basenextdom);

	FlatVector<SCALAR> vis = dynamic_cast<S_BaseVector<SCALAR> & > (*vi) . FVScal();
	FlatVector<SCALAR> helpis = dynamic_cast<S_BaseVector<SCALAR> & > (*helpi) . FVScal();

	const SparseMatrix<SCALAR> & mata = dynamic_cast<const SparseMatrix<SCALAR> &> (bfa[i]->GetMatrix() );

	// helpi = (invA ui_inner, 0 )
	// vi = (0, ui_gamma )
	(*vi) = (*ui);
	*helpi = (*invAinner[i]) * (*ui); // (*help2i);
	for ( int index = 0; index < ndof; index++ )
	  if ( innerdofs[i]->Test(index) ) vis(index) = 0.0;
	  else helpis(index) = 0.0;

	// helpi = (0, B invA B^T ui_gamma )
	// vi = (0, (C - B invA B^T) ui_gamma )
	*vi -= mata * (*helpi) ; // (*help2i);
	for ( int index = 0; index < ndof; index++ )
	  if ( innerdofs[i]->Test(index) ) vis(index) = 0.0;

      }
    int ndof = (*compoundspace)[ndomains]->GetNDof();
    int basedomi = compoundspace->GetStorageStart(ndomains);
    int basenextdom = basedomi + ndof;

 
    delete & help;

  }


  template class NumProcFETIBVP<2,double>;
  template class NumProcFETIBVP<3,double>;
  template class NumProcFETIBVP<2,Complex>;
  template class NumProcFETIBVP<3,Complex>;



}
