#include <solve.hpp>
#include <feti.hpp>
#include <sstream>


namespace ngsolve
{
  using namespace ngsolve;
  // using namespace ngbla;
  // using namespace ngla;
  using namespace std;

  string IntToString ( int number )
  {
    std::ostringstream oss;
    
    // Works just like cout
    oss<< number;
    
    // Return the underlying string
    return oss.str();
  }


  /* *************************** Numproc FETIBVP ********************** */


  ///

  template <int D, class SCAL>
  NumProcFETIBVP<D, SCAL> :: NumProcFETIBVP (PDE & apde, const Flags & flags)
    : NumProc (apde),
      matQ(0),
      matGtQG(0),
      invGtQG(0)
  {
    compoundspace = dynamic_cast<FETICompoundFESpace * > 
      ( pde.GetFESpace ( flags.GetStringFlag ( "fetifespace", "") ) );
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
    prec = flags.GetNumFlag ("prec", 1e-12);
    tau = flags.GetNumFlag ("tau", 1);
    
    ndomains = ma.GetNDomains();
    
    print = flags.GetDefineFlag ("print");
    noprecond = flags.GetDefineFlag ("noprecond");

    bfa.SetSize(ndomains);
    bfareg.SetSize(ndomains);
    lff.SetSize(ndomains);
//     lfreg.SetSize(ndomains);
    //     pre.SetSize(ndomains);

    invA.SetSize(ndomains);
    invAinner.SetSize(ndomains);

    identityQ = flags.GetDefineFlag("noQ") ;
    innerdofs.SetSize(ndomains);
    nondcdofs.SetSize(ndomains);

    SymbolTable<FESpace*> & pdefespaces = pde.GetSpaceTable();

    for ( int i = 0; i < ndomains; i++ )
      {
	string name = "v" + IntToString(i);

	(*compoundspace)[i]->SetName (name);
	pdefespaces.Set(name, (*compoundspace)[i]);
      }

    string uspacename = "u";
    (*compoundspace)[ndomains]->SetName(uspacename);
    pdefespaces.Set(uspacename, (*compoundspace)[ndomains]);
    uspace = dynamic_cast<FETILagrangeFESpace * > ( (*compoundspace)[ndomains]);

    // Bilinear forms A_i

    for ( int i = 0; i < ndomains; i++ )
      {
	Flags bfaflags;
	string name = string("bfa") + IntToString(i) ;
	string fespacename = "v" + IntToString(i);

	masscoeff = 0;
	robincoeff = 0;
	
	FESpace * space = pde.GetFESpace( fespacename );

	bfaflags.SetFlag ( "symmetric" );
	bfaflags.SetFlag ( "fespace", fespacename.c_str() );

	pde.AddBilinearForm ( name, bfaflags);

	// laplaceintegrator
	    
	string integrator_name = "laplace";
	const ngfem::Integrators::IntegratorInfo * info =
	  ngfem::GetIntegrators() 
	  . GetBFI(integrator_name, 
		   pde.GetMeshAccess().GetDimension());
	
	if (info)
	  {
	    ARRAY<CoefficientFunction*> coeffs(1);
	    string coeffname = flags.GetStringFlag("laplacecoefficient","");
	    if (coeffname == "" )
	      {
		cout << "missing laplacecoefficient" << endl;
		return;
	      }

	    coeffs[0] = pde.GetCoefficientFunction (coeffname , 1);
		

	    laplacecoeff = dynamic_cast<DomainConstantCoefficientFunction * > ( coeffs[0] );
	    if ( !laplacecoeff ) exit(0);

	    cout << "laplacecoeff " << flags.GetStringFlag("laplacecoefficient","") << endl;
	    ngfem::BilinearFormIntegrator * integrator = 
	      dynamic_cast<ngfem::BilinearFormIntegrator*> (info->creator(coeffs));
	    integrator -> SetName (integrator_name);
	    pde.AddBilinearFormIntegrator (name, integrator);

	  }

	// massintegrator
	    
	integrator_name = "mass";
	const ngfem::Integrators::IntegratorInfo * infomass =
	  ngfem::GetIntegrators() 
	  . GetBFI(integrator_name, 
		   pde.GetMeshAccess().GetDimension());

	string masscoeffname = flags.GetStringFlag("masscoefficient","");
	if (infomass && masscoeffname != "" )
	  {
	    ARRAY<CoefficientFunction*> coeffs(1);
	    coeffs[0] = pde.GetCoefficientFunction (masscoeffname, 1);
		
	    masscoeff = pde.GetCoefficientFunction (masscoeffname, 1);

	    ngfem::BilinearFormIntegrator * integrator = 
	      dynamic_cast<ngfem::BilinearFormIntegrator*> (infomass->creator(coeffs));
	    integrator -> SetName (integrator_name);
	    cout << "integrator_name " << integrator_name << endl;
	    cout << integrator -> Name() << endl;
	    pde.AddBilinearFormIntegrator (name, integrator);
	  }

		   
	integrator_name = "robin";
	const ngfem::Integrators::IntegratorInfo * inforobin =
	  ngfem::GetIntegrators()
	  . GetBFI(integrator_name, pde.GetMeshAccess().GetDimension());

	string robincoeffname = flags.GetStringFlag("robincoefficient","");
	if ( inforobin && robincoeffname != "" )
	  {
	    ARRAY<CoefficientFunction*> coeffs(1);
	    coeffs[0] = pde.GetCoefficientFunction ( robincoeffname, 1);

	    robincoeff = coeffs[0];
		
	    ngfem::BilinearFormIntegrator * integrator = 
	      dynamic_cast<ngfem::BilinearFormIntegrator*> (inforobin->creator(coeffs));
	    integrator -> SetName (integrator_name);

	    pde.AddBilinearFormIntegrator (name, integrator);

	  }

		   
	integrator_name = "robin";
	const ngfem::Integrators::IntegratorInfo * inforobinimag =
	  ngfem::GetIntegrators()
	  . GetBFI(integrator_name, pde.GetMeshAccess().GetDimension());

	robincoeffname = flags.GetStringFlag("robincoefficientimag","");
	if ( inforobinimag && robincoeffname != "" )
	  {
	    ARRAY<CoefficientFunction*> coeffs(1);
	    coeffs[0] = pde.GetCoefficientFunction ( robincoeffname, 1);

	    robincoeff = coeffs[0];
		
	    ngfem::BilinearFormIntegrator * integrator = 
	      dynamic_cast<ngfem::BilinearFormIntegrator*> (inforobinimag->creator(coeffs));
	    integrator -> SetName (integrator_name);

	    integrator = new ComplexBilinearFormIntegrator
	      (*integrator, Complex(0,1));

	    pde.AddBilinearFormIntegrator (name, integrator);

	  }

	bfa[i] = pde.GetBilinearForm(name); // CreateBilinearForm (space, name, bfaflags);

      }




    // Linear forms f_i

    for ( int i = 0; i < ndomains; i++ )
      {
	Flags lffflags;

	// FESpace * space = compoundfespace[i];
	string name = "lff" + IntToString(i) ;
	string fespacename = "v" + IntToString(i);

	lffflags.SetFlag( "fespace", fespacename.c_str() );
 
	pde.AddLinearForm( name, lffflags); // CreateLinearForm (space, name, lffflags);

	// sourceintegrator
	    
	string integrator_name = "source";
	const ngfem::Integrators::IntegratorInfo * info =
	  ngfem::GetIntegrators() 
	  . GetLFI(integrator_name, 
		   pde.GetMeshAccess().GetDimension());
	
	if (info && flags.GetStringFlag("sourcecoefficient", "") != "" )
	  {
	    ARRAY<CoefficientFunction*> coeffs(1);
	    coeffs[0] = pde.GetCoefficientFunction (flags.GetStringFlag("sourcecoefficient",""), 1);
		
	    ngfem::LinearFormIntegrator * integrator = 
	      dynamic_cast<ngfem::LinearFormIntegrator*> (info->creator(coeffs));
	    integrator -> SetName (integrator_name);
	    pde.AddLinearFormIntegrator ( name, integrator );	    
	  }

	string integrator_name_neumann = "neumann";
	const ngfem::Integrators::IntegratorInfo * infoneumann =
	  ngfem::GetIntegrators() 
	  . GetLFI(integrator_name_neumann, 
		   pde.GetMeshAccess().GetDimension());
	
	if (infoneumann && flags.GetStringFlag("neumanncoefficient","") != "" )
	  {
	    ARRAY<CoefficientFunction*> coeffs(1);
	    coeffs[0] = pde.GetCoefficientFunction (flags.GetStringFlag("neumanncoefficient",""), 1);
		
	    ngfem::LinearFormIntegrator * integrator = 
	      dynamic_cast<ngfem::LinearFormIntegrator*> (infoneumann->creator(coeffs));
	    integrator -> SetName (integrator_name_neumann);
	    pde.AddLinearFormIntegrator ( name, integrator );	    
	  }
	lff[i] = pde.GetLinearForm( name );	   

 
      }

    // Linear forms f_reg for regularization

//     for ( int i = 0; i < ndomains; i++ )
//       {
// 	Flags lffflags;

// 	// FESpace * space = compoundfespace[i];
// 	string name = "lfreg" + IntToString(i) ;
// 	string fespacename = "v" + IntToString(i);

// 	lffflags.SetFlag( "fespace", fespacename.c_str() );
 
// 	pde.AddLinearForm( name, lffflags); // CreateLinearForm (space, name, lffflags);

// 	// sourceintegrator
	    
// 	string integrator_name = "neumann";
// 	const ngfem::Integrators::IntegratorInfo * info =
// 	  ngfem::GetIntegrators() 
// 	  . GetLFI(integrator_name, 
// 		   pde.GetMeshAccess().GetDimension());
	
// 	if (info)
// 	  {
// 	    ARRAY<CoefficientFunction*> coeffs(1);
// 	    ARRAY<double> ones(ndomains);
// 	    ones = 1.0;
// 	    coeffs[0] = new DomainConstantCoefficientFunction (ones);
// 	    // de.GetCoefficientFunction (flags.GetStringFlag("sourcecoefficient",""), 1);
		
// 	    ngfem::LinearFormIntegrator * integrator = 
// 	      dynamic_cast<ngfem::LinearFormIntegrator*> (info->creator(coeffs));
// 	    integrator -> SetName (integrator_name);
// 	    pde.AddLinearFormIntegrator ( name, integrator );	    
// 	  }
// 	lfreg[i] = pde.GetLinearForm( name );	   

 
//       }

    //     // precondtioners 

    //     for ( int i = 0; i < ndomains; i++ )
    //       {
    // 	Flags preflags;

    // 	string name = "pre" + IntToString(i);
    // 	string bfaname = "bfa" + IntToString(i);
    // 	preflags.SetFlag("inversetype", flags.GetStringFlag( "inversetype", "sparsecholesky"));
    // 	preflags.SetFlag("bilinearform",bfaname.c_str());
    // 	if ( flags.GetDefineFlag("test")) preflags.SetFlag("test");
    // 	preflags.SetFlag("type", "direct");
    // 	pre[i] = pde.AddPreconditioner ( name, preflags );
    //       }



    isfloating.SetSize(ndomains);
    isfloating.Clear();

  }

  template <int D, class SCAL>
  NumProcFETIBVP<D, SCAL> :: ~NumProcFETIBVP()
  {
    for ( int i = 0; i < ndomains; i++ )
      {
	delete innerdofs[i];
	delete nondcdofs[i];
      }
    ;
  }

  template <int D, class SCAL>
  void NumProcFETIBVP<D, SCAL> :: PrintDoc (ostream & ost)
  {
    ;
  }


  template <int D, class SCAL>
  bool NumProcFETIBVP<D, SCAL> :: IsFloatingDomain ( const int dom ) const
  {
    bool floating = true;
  
    if ( masscoeff ) return false;

    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {

	int base = compoundspace -> GetStorageStart(dom);
	ARRAY<int> dofs;
	const DDH1HighOrderFESpace * vspace = 
	  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom]);

	vspace->GetVertexDofNrs(vert, dofs);

	//       cout << "noncdcdofs[" << dom << "] = " << *nondcdofs[dom] << endl;

	if ( dofs.Size() )
	  if ( !nondcdofs[dom]->Test(dofs[0]) )
	    {
	      // 	    cout << "found proof for floating subdomain " << dom << ": Dirichlet dof " << dofs[0] << endl;

	      floating = false;
	      break;
	    }
      
      }

    return floating;

  }


  template <int D, class SCAL>
  void NumProcFETIBVP<D, SCAL> :: Do(LocalHeap & lh)
  {

    if ( ma.GetDimension() != D )
      {
	cout << "Dimensions of mesh and numproc do not match, please use fetibvp" 
	     << D << "d!" << endl;
	return;
      }

    static int solvetimer = NgProfiler::CreateTimer ("Equation solving");
    NgProfiler::RegionTimer reg (solvetimer);

    for ( int i = 0; i < ndomains; i++)
      {
	innerdofs[i] = new BitArray ( (*compoundspace)[i]->GetNDof() ) ;
	nondcdofs[i] = new BitArray ( (*compoundspace)[i]->GetNDof() ) ;
	int ndof = (*compoundspace)[i]->GetNDof();
	int basedomi = compoundspace->GetStorageStart(i);
	int basenextdom = basedomi + ndof;

	innerdofs[i]->Set();
	nondcdofs[i]->Set();

	for ( int sel = 0; sel < ma.GetNSE(); sel++ )
	  {
	    // if ( ma.GetSElIndex(sel) == 0 ) continue;

	    int dom1, dom2;
	    // GetSElNeighDoms --> outer domain nr 0, inners start with 1, --> domnr--
	    ma.GetSElNeighbouringDomains(sel, dom1, dom2);
	    dom1--; dom2--;

	    if ( dom1 != i && dom2 != i ) continue;
	    ARRAY<int> seldofs;
	    (*compoundspace)[i]->GetSDofNrs(sel, seldofs);
	    for ( int j = 0; j < seldofs.Size(); j++ )
	      if ( seldofs[j] >= 0 )
		{
		  if ( ma.GetSElIndex(sel) == 0 )
		    nondcdofs[i]->Clear(seldofs[j]);
		  
		  innerdofs[i]->Clear(seldofs[j]);
		}
	  }


	if ( IsFloatingDomain(i) )
	  isfloating.Set(i);
	
	// 	cout << "update bilinear-form " << bfa[i]->GetName() << endl;
		    
	// 	bfa[i]->Assemble(lh);

	lh.CleanUp();

	// 	cout << "Update linear-form " << lff[i]->GetName() << endl;
	// 	lff[i]->Assemble(lh);
	lh.CleanUp();


	BaseSparseMatrix & matai = dynamic_cast<BaseSparseMatrix&> (bfa[i]->GetMatrix());
	SparseMatrix<SCAL> & mataireg = ( IsFloatingDomain(i) ) ? 
	  dynamic_cast<SparseMatrix<SCAL> &> (*matai.CreateMatrix() ) :
	  dynamic_cast<SparseMatrix<SCAL> &> (matai);

	if ( isfloating.Test(i) )
	  {
	    cout << "floating domain " << i << endl;
	    mataireg.AsVector() = matai.AsVector();

// 	    BaseVector & vecreg = lfreg[i] -> GetVector();

	    int ndof_v = (*compoundspace)[i]->GetNDof();

	    const DDH1HighOrderFESpace & vspace = dynamic_cast<const DDH1HighOrderFESpace &> (*(*compoundspace)[i]);

	    int ndof_lo = vspace.GetNVertexDofs();
	    for ( int dofi = 0; dofi < ndof_lo; dofi++ )
	      for ( int dofj = 0; dofj <= dofi; dofj++ )
		{
// 		  double val = vecreg.FVDouble()(dofi) * vecreg.FVDouble()(dofj);
// 		  if ( abs( val) > 1e-10 )
		  if ( !innerdofs[i]->Test(dofj) )
		    if ( !innerdofs[i]->Test(dofi) )
		      mataireg(dofi, dofj) +=  1;
		}
	  }
 	matai . SetInverseType ( SPARSECHOLESKY );
	invAinner[i] = matai.InverseMatrix(innerdofs[i]);

 	mataireg . SetInverseType ( SPARSECHOLESKY );
	invA[i] = mataireg.InverseMatrix(nondcdofs[i]);
      }



    int nd = compoundspace->GetNDof();
    int baselambda = compoundspace->GetStorageStart(ndomains);

    /* Tables containing neighbouring domains for verts, edges, faces */
    CreateDomainArrays();

    CalcQ();
    CalcInvGtQG ();


    /* CG METHOD */

    BaseVector & vecu = (gfu->GetVector());
    BaseVector & vecf = (*vecu.CreateVector());

    BaseVector & vech = (*vecu.CreateVector());

    BaseVector & d = (*vecf.CreateVector());
    BaseVector & w = (*vecf.CreateVector());
    BaseVector & s = (*vecf.CreateVector());
    
    Vector<SCAL> vece(ndomains), vechelpe(ndomains), alpha(ndomains), beta(ndomains);

    int n = 0;
    SCAL al, be, wd, wdn, kss;
    double err;

    for ( int i = 0; i < ndomains; i++ )
      {
	int basedomi = compoundspace->GetStorageStart(i);
	int basenextdom = basedomi + (*compoundspace)[i]->GetNDof();
	*(vecf.Range(basedomi, basenextdom)) =  (lff[i]->GetVector());
      }

    int baselagrange = compoundspace->GetStorageStart(ndomains);
    int ndof = compoundspace->GetNDof();

    // startwert
    // u_0 = Q G (GQG)^(-1) R^T N f
    // 1. Vector vech = f_gamma - A_gi invAinner f_i
    // 2. Vector vece = ApplyRTrans ( vecNf, vece );
    NewtonPotential ( vecf, vech );
    ApplyRTrans ( vech, vece );
    vechelpe = invGtQG * vece;
    ApplyG ( vechelpe, vecu );
    ApplyQ ( vecu, vecu );

    // solve equation for schur complement
    // B S^-1 B^T vecu = B S^-1 vecf
    // residual d = B S^-1 vecf - Op (vecu)

    ApplyInvS ( vecf, vech );
    ApplyB ( vech, vecf );

    d = vecf;
    ApplyOp(vecu, w);
    d -= w;
    cout << "d " << L2Norm(d) << endl;
    // Projection for floating subdomains
    ApplyPTrans (d, d);

    cout << "d after P " << L2Norm(*d.Range(baselagrange, ndof)) << endl;
    //	  w = (*c) * d;
    if ( noprecond )
      w = d;
    else 
      {
	Precond ( d, w );
      }
    ApplyP ( w, w);
    // w = d;
    cout << "d after Precond " << L2Norm(*d.Range(baselagrange, ndof)) << endl;
    cout << "w after Precond " << L2Norm(*w.Range(baselagrange, ndof)) << endl;

    s = w;

    wdn = S_InnerProduct<SCAL> (*w.Range(baselagrange, ndof) ,*d.Range(baselagrange, ndof) );

    cout << "0 " << sqrt(fabs(wdn)) << endl;
    if (wdn == 0.0) wdn = 1;	

    err = prec * prec * fabs (wdn);
	
    double lwstart = log(fabs(wdn));
    double lerr = log(err);
	

    while ( n++ < maxsteps  && fabs(wdn) > err )
      {
	//	    w = (*a) * s;
	ApplyOp ( s, w );
	ApplyPTrans ( w, w);

	wd = wdn;

	kss = S_InnerProduct<SCAL> (*s.Range(baselagrange, ndof), *w.Range(baselagrange, ndof));
	if (kss == 0.0) break;
	
	al = wd / kss;
	
	vecu += al * s;
	d -= al * w;

	//      w = (*c) * d;
	if ( noprecond )
	  w = d;
	else
	  Precond( d, w );
	ApplyP ( w, w);

	wdn = S_InnerProduct<SCAL> (*d.Range(baselagrange, ndof), *w.Range(baselagrange, ndof));
	
	be = wdn / wd;
	
	s *= be;
	s += w;

	cout << n << " ****************************** " << sqrt (fabs (wdn)) << endl;
	
      } 
	
    // calculate solution
    // u =  S^-1 ( f - B^T d )
    ApplyBTrans ( vecu, d );

    ApplyInvS ( d, w );

    vecu = vech - w;
    vech = vecu;

//     cout << "isfloating ... " << endl << isfloating << endl;
    ApplyB ( vecu, s );

    ApplyQ (s, w);
    ApplyGTrans ( w, alpha );
    beta = invGtQG * alpha;

//     cout << "BETA " << endl << alpha << endl << endl << beta << endl;
    vecu = vech;
    for ( int dom = 0; dom < ndomains; dom++ )
      {
	int base = compoundspace->GetStorageStart(dom);
	const DDH1HighOrderFESpace * vspace = 
	  dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[dom]);
	for ( int vert = 0; vert < ma.GetNV(); vert++ )
	  {
	    ARRAY<int> dofs;
	    vspace -> GetVertexDofNrs ( vert, dofs );
	    if ( dofs.Size() )
	      dynamic_cast<S_BaseVector<SCAL> & > (vecu).FVScal()(base + dofs[0]) -= beta(dom);
	  }
      }

    delete &d;
    delete &w;
    delete &s;

    delete & vecf;
    delete & vech;

    delete vert_doms;
    delete edge_doms;
    if ( D == 3 )
      delete face_doms;
    *testout << "Solution vector " << endl << vecu << endl;

    lh.CleanUp ();
  }


    /* Tables containing neighbouring domains for verts, edges, faces */
  template <int D, class SCAL>
  void NumProcFETIBVP<D, SCAL> ::   CreateDomainArrays()
  {
    ARRAY<int> vert_domains, edge_domains, face_domains, elems, cnt;
    
    cnt.SetSize ( ma.GetNV() );
    cnt = 0;

    int nd = compoundspace->GetNDof();
    int baselambda = compoundspace->GetStorageStart(ndomains);

    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {
	ARRAY<int> dofslambda;
	uspace->GetVertexDofNrs (vert, dofslambda );
	int nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	ma.GetVertexElements(vert, elems);

	vert_domains.SetSize(0);
	for ( int i = 0; i < elems.Size(); i++ )
	  {
	    int index = ma.GetElIndex (elems[i]);
	    if ( !vert_domains.Contains(index) )
	      vert_domains.Append(index);

	  }
	cnt[vert] = vert_domains.Size();
      }

    vert_doms = new Table<int> (cnt);

    for ( int vert = 0; vert < ma.GetNV(); vert++ )
      {
	ARRAY<int> dofslambda;
	uspace->GetVertexDofNrs (vert, dofslambda );
	int nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	ma.GetVertexElements(vert, elems);

	vert_domains.SetSize(0);
	for ( int i = 0; i < elems.Size(); i++ )
	  {
	    int index = ma.GetElIndex (elems[i]);
	    if ( !vert_domains.Contains(index) )
	      vert_domains.Append(index);

	  }
	BubbleSort(vert_domains);
	for ( int i = 0; i < vert_domains.Size(); i++ )
	  (*vert_doms)[vert][i] = vert_domains[i];
	// 	(*vert_doms)[vert] = vert_domains;
      }

    cnt.SetSize(ma.GetNEdges());
    cnt = 0;
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	ARRAY<int> dofslambda;
	uspace->GetEdgeDofNrs ( edge, dofslambda );
	int nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	ma.GetEdgeElements(edge, elems);

	edge_domains.SetSize(0);
	for ( int i = 0; i < elems.Size(); i++ )
	  {
	    int index = ma.GetElIndex (elems[i]);
	    if ( !edge_domains.Contains(index) )
	      edge_domains.Append(index);
	  }
	cnt[edge] += edge_domains.Size();
      }

    edge_doms = new Table<int> (cnt);
    for ( int edge = 0; edge < ma.GetNEdges(); edge++ )
      {
	ARRAY<int> dofslambda;
	uspace->GetEdgeDofNrs ( edge, dofslambda );
	int nd_lambda = dofslambda.Size();
	if ( !nd_lambda ) continue;

	ma.GetEdgeElements(edge, elems);

	edge_domains.SetSize(0);
	for ( int i = 0; i < elems.Size(); i++ )
	  {
	    int index = ma.GetElIndex (elems[i]);
	    if ( !edge_domains.Contains(index) )
	      edge_domains.Append(index);
	  }
	BubbleSort ( edge_domains );
	for ( int i = 0; i < edge_domains.Size(); i++ )
	  (*edge_doms)[edge][i] = edge_domains[i];
      }

    if ( D == 3 )
      {
	cnt.SetSize(ma.GetNFaces());
	cnt = 0;
	for ( int face = 0; face < ma.GetNFaces(); face++ )
	  {
	    ARRAY<int> dofslambda;
	    uspace->GetFaceDofNrs ( face, dofslambda );
	    int nd_lambda = dofslambda.Size();
	    if ( !nd_lambda ) continue;

	    ma.GetFaceElements(face, elems);

	    face_domains.SetSize(0);
	    for ( int i = 0; i < elems.Size(); i++ )
	      {
		int index = ma.GetElIndex (elems[i]);
		if ( !face_domains.Contains(index) )
		  face_domains.Append(index);
	      }
	    cnt[face] += face_domains.Size();
	  }

	face_doms = new Table<int> (cnt);
	for ( int face = 0; face < ma.GetNFaces(); face++ )
	  {
	    ARRAY<int> dofslambda;
	    uspace->GetFaceDofNrs ( face, dofslambda );
	    int nd_lambda = dofslambda.Size();
	    if ( !nd_lambda ) continue;

	    ma.GetFaceElements(face, elems);

	    face_domains.SetSize(0);
	    for ( int i = 0; i < elems.Size(); i++ )
	      {
		int index = ma.GetElIndex (elems[i]);
		if ( !face_domains.Contains(index) )
		  face_domains.Append(index);
	      }
	    BubbleSort ( face_domains );
	    for ( int i = 0; i < face_domains.Size(); i++ )
	      (*face_doms)[face][i] = face_domains[i];
	  }
      }
}
  template class NumProcFETIBVP<2, double>;
  template class NumProcFETIBVP<3, double>;
  template class NumProcFETIBVP<2, Complex>;
  template class NumProcFETIBVP<3, Complex>;



  namespace
#ifdef MACOS
  fetibvpl2_cpp
#endif
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("fetibvp2d", NumProcFETIBVP<2,double>::Create, NumProcFETIBVP<2,double>::PrintDoc);
      GetNumProcs().AddNumProc ("fetibvp3d", NumProcFETIBVP<3,double>::Create, NumProcFETIBVP<3,double>::PrintDoc);
      GetNumProcs().AddNumProc ("fetibvp2dc", NumProcFETIBVP<2,Complex>::Create, NumProcFETIBVP<2,Complex>::PrintDoc);
      GetNumProcs().AddNumProc ("fetibvp3dc", NumProcFETIBVP<3,Complex>::Create, NumProcFETIBVP<3,Complex>::PrintDoc);
    }
    
    
    Init init;
    
  }
  


}



// 	for ( int i = 0; i < ndomains; i++)
// 	  {
	    
// 	    int ndof = (*compoundspace)[i]->GetNDof();
// 	    int basedomi = compoundspace->GetStorageStart(i);
// 	    int basenextdom = basedomi + ndof;
	    
// 	    auto_ptr<BaseVector> ui = w.Range(basedomi, basenextdom);
// 	    BitArray innerdofs(ndof);
// 	    innerdofs.Set();
	    
// 	    for ( int sel = 0; sel < ma.GetNSE(); sel++ )
// 	      {
// 		if ( ma.GetSElIndex(sel) == 0 ) continue;
// 		int dom1, dom2;
// 		ma.GetSElNeighbouringDomains(sel, dom1, dom2);
// 		dom1--; dom2--;
		
// 		if ( dom1 != i && dom2 != i ) continue;
// 		ARRAY<int> seldofs;
// 		(*compoundspace)[i]->GetSDofNrs(sel, seldofs);
// 		for ( int j = 0; j < seldofs.Size(); j++ )
// 		  if ( seldofs[j] >= 0 )
// 		    innerdofs.Clear(seldofs[j]);
// 	      }
// 	    for ( int index = 0; index < ndof; index++ )
// 	      if ( innerdofs.Test(index) && ui->FVDouble()(index) ) cout << "!!! " << ui->FVDouble()(index) << endl;
// 	  }
