#ifndef FILE_NUMPROCFETI_
#define FILE_NUMPROCFETI_


string IntToString ( int number );

template < int D, class SCALAR >
  class NumProcFETIBVP : public NumProc
  {
  protected:
    ///
    ARRAY<BilinearForm *> bfa, bfareg;
    ///
    ARRAY<LinearForm *> lff;
    ///
    ARRAY<LinearForm *> lfreg;
    ///
    GridFunction * gfu;
    ///
    BilinearForm * bfs;
    ///
    LinearForm * lfh;
    ///
    FETICompoundFESpace * compoundspace;
    ///
    FETILagrangeFESpace * uspace;
    ///
    int ndomains;
//     ///
//     ARRAY<Preconditioner * > pre;
    ///
    ARRAY<BaseMatrix * > invAinner, invA;
    ///
    int maxsteps;
    ///
    double prec;
    ///
    double tau;
    ///
    bool print;
    ///
    bool noprecond, identityQ;
    ///
    ARRAY<BitArray*> innerdofs, nondcdofs;

    Table<int> * vert_doms, *edge_doms, *face_doms;

    DomainConstantCoefficientFunction * laplacecoeff;

    CoefficientFunction * masscoeff, * robincoeff;

    Vector<SCALAR> matQ;

    Matrix<SCALAR> matGtQG;

    Matrix<SCALAR> invGtQG;

    BitArray isfloating;

  public:
    ///
    NumProcFETIBVP (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcFETIBVP();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcFETIBVP (pde, flags);
    }


    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "FETI BVP";
    }

    virtual void PrintReport (ostream & ost)
    {
      ;
    }

    ///
    static void PrintDoc (ostream & ost);

    void ApplyB ( const BaseVector & u, BaseVector & v, bool bar = 0 );

    void ApplyBTrans ( const BaseVector & u, BaseVector & v, bool bar = 0);

    inline void ApplyBarB ( const BaseVector & u, BaseVector & v )
    {
      ApplyB ( u, v, true );
    }

    inline void ApplyBarBTrans ( const BaseVector & u, BaseVector & v )
    {
      ApplyBTrans ( u, v, true );
    }

    void ApplyS ( const BaseVector & u,  BaseVector & v);

    void ApplyInvS ( const BaseVector & u,  BaseVector & v);

    void Precond ( BaseVector & lambda, BaseVector & mu);

    void ApplyOp ( BaseVector & lambda, BaseVector & mu);

//     void ApplyR ( const Vector<SCALARAR> & alpha, BaseVector & vecu);

//     template < class SCALARAR >
//     void ApplyRTrans ( const  BaseVector & vecu, Vector<SCALARAR> & alpha);

    void ApplyR ( const Vector<SCALAR> & alpha, BaseVector & vecu);
//     {
//       FlatVector<SCALAR> vecus = dynamic_cast<S_BaseVector<SCALAR> & > (vecu) . FVScal();
      
//       vecus.Range(0, compoundspace->GetStorageStart(ndomains) ) = 0.0;
      
//       for ( int i = 0;  i < ndomains; i++ )
// 	{
// 	  if ( !isfloating.Test(i) ) continue;
	  
// 	  int base = compoundspace -> GetStorageStart(i);
// 	  ARRAY<int> dnums;
	  
// 	  const DDH1HighOrderFESpace * vspace = 
// 	    dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[i]);
	  
// 	  for ( int vert = 0; vert < ma.GetNV(); vert++ )
// 	    {
// 	      vspace -> GetVertexDofNrs ( vert, dnums );
	      
// 	      if ( !dnums.Size() ) continue;
	      
// 	      if ( innerdofs[i]->Test(dnums[0]) || !nondcdofs[i]->Test(dnums[0]) ) continue;
	      
// 	      for ( int ii = 0; ii < dnums.Size(); ii++ )
// 		vecus(base + dnums[ii]) = alpha(i);
	      
// 	    }
	  
	  
// 	}
//     }

    void ApplyRTrans ( const  BaseVector & vecu, Vector<SCALAR> & alpha);
//     {
//       alpha = 0;
//       const FlatVector<SCALAR> vecus = dynamic_cast<const S_BaseVector<SCALAR> & > (vecu) . FVScal();
      
//       for ( int i = 0;  i < ndomains; i++ )
// 	{
// 	  if ( !isfloating.Test(i) ) continue;
	  
// 	  int base = compoundspace -> GetStorageStart(i);
// 	  ARRAY<int> dnums;
	  
// 	  const DDH1HighOrderFESpace * vspace = 
// 	    dynamic_cast<const DDH1HighOrderFESpace*> ((*compoundspace)[i]);
	  
// 	  for ( int vert = 0; vert < ma.GetNV(); vert++ )
// 	    {
// 	      vspace -> GetVertexDofNrs ( vert, dnums );
	      
// 	      if ( !dnums.Size() ) continue;
	      
// 	      if ( innerdofs[i]->Test(dnums[0]) || !nondcdofs[i]->Test(dnums[0]) ) continue;
	      
// 	      for ( int ii = 0; ii < dnums.Size(); ii++ )
// 		alpha(i) += vecus(base + dnums[ii]);
	      
// 	    }
	  
	  
// 	}
      
//       //     bool allfloating = true;
//       //     for ( int i = 0;  i < ndomains; i++ )
//       //       if ( !isfloating.Test(i) ) allfloating = false;
      
//       //     if ( allfloating )
//       //       {
//       // 	double mv = 0;
//       // 	for ( int i = 0; i < ndomains; i++ )
//       // 	  mv += alpha(i);
//       // 	mv /= ndomains;
//       // 	for ( int i = 0; i < ndomains; i++ )
//       // 	  alpha(i) -= mv;
//       //       }
//     }
    
    bool IsFloatingDomain ( const int dom ) const;

    inline void ApplyG ( const Vector<SCALAR> & alpha, BaseVector & lambda)
    {
      ApplyR ( alpha, lambda);
      ApplyB ( lambda, lambda);
    }

    inline void ApplyGTrans ( const BaseVector & lambda, Vector<SCALAR> & alpha)
    {
      BaseVector & help = *lambda.CreateVector();
      ApplyBTrans ( lambda, help);
      ApplyRTrans ( help, alpha);
    }

    void CalcQ();

    void ApplyQ ( const BaseVector & lambda, BaseVector & mu);

    void CalcInvGtQG ();

    void ApplyP ( const BaseVector & lambda, BaseVector & mu )
    {
      BaseVector & lambda2 = *lambda.CreateVector();
      Vector<SCALAR> alpha(ndomains), beta(ndomains);

      ApplyGTrans ( lambda, alpha);
      beta = invGtQG * alpha;

      ApplyG ( beta, lambda2 );
      ApplyQ ( lambda2, lambda2 );

      mu = lambda - lambda2;


      delete & lambda2;

    }

    void ApplyPTrans ( const BaseVector & lambda, BaseVector & mu )
    {
      BaseVector & lambda2 = *lambda.CreateVector();
      Vector<SCALAR> alpha(ndomains), beta(ndomains);

      ApplyQ ( lambda, lambda2 );
      ApplyGTrans ( lambda2, alpha);
      beta = invGtQG * alpha;

      ApplyG ( beta, lambda2 );

      mu = lambda - lambda2;


      delete & lambda2;
    }

    void NewtonPotential ( const BaseVector & vecf, BaseVector & vecNf );

    void CreateDomainArrays();
  };

#endif
