#ifndef FILE_FETIDP_MATRIX_HPP
#define FILE_FETIDP_MATRIX_HPP



class BaseMatrix;
template<class TM, 
	 class TV_ROW, 
	 class TV_COL>
class SparseMatrix;

template <class SCAL>
class FETIDP_Matrix : public BaseMatrix
{
  int height;
  int width;
  BitArray * primaldofs;
  BitArray * dualdofs;
  BitArray * lagrangedofs;
  BitArray * invsdd_dofs;

  SparseMatrix<SCAL> * mats;
  SparseMatrix<SCAL> * matapp;
  SparseMatrix<SCAL> * matapd;
  SparseMatrix<SCAL> * matadd;
  BaseMatrix * invsdd;
  SparseMatrix<SCAL> * matb;
  SparseMatrix<SCAL> * matbarb;

  BaseMatrix * invspp;
//   BaseMatrix * invsdd;

  INVERSETYPE inversetype;


public:

  FETIDP_Matrix ();

  FETIDP_Matrix (const MatrixGraph & agraph_matspp, 
		 const MatrixGraph & agraph_matapd, 
		 const MatrixGraph & agraph_matadd, 
		 bool stealgraph, 
		 INVERSETYPE ainversetype=SPARSECHOLESKY);

  ~FETIDP_Matrix ();

  virtual int VHeight () const
  {
    return height;
  }

  virtual int VWidth () const
  { return width; }

  int Height () const 
  { return height; }

  int Width () const
  { return width; }

  
  virtual ostream & Print (ostream & ost) const;
 
  /// creates matrix of same type
  virtual BaseMatrix * CreateMatrix () const;
  /// creates a compativle vector, size = width
  virtual BaseVector * CreateRowVector () const;
  /// creates a compativle vector, size = height
  virtual BaseVector * CreateColVector () const;
  /// creates a fitting vector (for square matrices)
  virtual BaseVector * CreateVector () const;

  /// y = matrix * x. Multadd should be implemented, instead
  virtual void Mult (const BaseVector & x, BaseVector & y) const;
  /// y += s matrix * x
  virtual void MultAdd (double s, const BaseVector & x, BaseVector & y) const;
  /// y += s matrix * x
  virtual void MultAdd (Complex s, const BaseVector & x, BaseVector & y) const;
  
  /// y += s Trans(matrix) * x
  virtual void MultTransAdd (double s, const BaseVector & x, BaseVector & y) const;
  /// y += s Trans(matrix) * x
  virtual void MultTransAdd (Complex s, const BaseVector & x, BaseVector & y) const;

  const BitArray & PrimalDofs() const { return *primaldofs; }
  const BitArray & DualDofs() const { return *dualdofs; }
  const BitArray & LagrangeDofs() const { return *lagrangedofs; }
  BitArray & PrimalDofs()  { return *primaldofs; }
  BitArray & DualDofs()  { return *dualdofs; }
  BitArray & LagrangeDofs()  { return *lagrangedofs; }
  
  BaseMatrix & GetAddMat () { return *matadd; }
  BaseMatrix & GetAppMat () { return *matapp; }
  BaseMatrix & GetSMat () { return *mats; }
  BaseMatrix & GetInvSpp () { return *invspp; }
  BaseMatrix & GetInvSdd () { return *invsdd; }
  BaseMatrix & GetBMat () { return *matb; }
  BaseMatrix & GetBarBMat () { if ( matbarb ) return *matbarb; else return *matb; }
  BaseMatrix & GetApdMat () { return *matapd; }

  const BaseMatrix & GetAddMat () const { return *matadd; }
  const BaseMatrix & GetAppMat () const { return *matapp; }
  const BaseMatrix & GetSMat () const { return *mats; }
  const BaseMatrix & GetInvSpp () const { return *invspp; }
  const BaseMatrix & GetInvSdd () const { return *invsdd; }
  const BaseMatrix & GetBMat () const { return *matb; }
  const BaseMatrix & GetApdMat () const { return *matapd; }
  const BaseMatrix & GetBarBMat () const 
  { if ( matbarb ) return *matbarb; else return *matb; }
  virtual BaseVector & AsVector() 
  {
    return *new VVector<double> (1);
  }

  void CalcInverses ( bool eliminate_internal) ;

  void SetBMat ( SparseMatrix<SCAL> * abmat )
  { matb = abmat; }

  void SetBarBMat ( SparseMatrix<SCAL> * abarbmat )
  { matbarb = abarbmat; }

};

template <class SCAL>
class FETIDP_PrecMatrix : public BaseMatrix
{
  const FETIDP_Matrix<SCAL> & fetidpmat;

public:
  FETIDP_PrecMatrix ( const FETIDP_Matrix<SCAL> & afetidpmat );

  ~FETIDP_PrecMatrix ();

  virtual int VHeight () const
  {
    return fetidpmat.VHeight();
  }

  virtual int VWidth () const
  { return fetidpmat.VWidth(); }

  int Height () const 
  { return fetidpmat.Height(); }

  int Width () const
  { return fetidpmat.Width(); }

  
  virtual ostream & Print (ostream & ost) const;
 
  /// creates matrix of same type
  virtual BaseMatrix * CreateMatrix () const;
  /// creates a compativle vector, size = width
  virtual BaseVector * CreateRowVector () const;
  /// creates a compativle vector, size = height
  virtual BaseVector * CreateColVector () const;
  /// creates a fitting vector (for square matrices)
  virtual BaseVector * CreateVector () const;

  /// y = matrix * x. Multadd should be implemented, instead
  virtual void Mult (const BaseVector & x, BaseVector & y) const;
  /// y += s matrix * x
  virtual void MultAdd (double s, const BaseVector & x, BaseVector & y) const;
  /// y += s matrix * x
  virtual void MultAdd (Complex s, const BaseVector & x, BaseVector & y) const;
  
  /// y += s Trans(matrix) * x
  virtual void MultTransAdd (double s, const BaseVector & x, BaseVector & y) const;
  /// y += s Trans(matrix) * x
  virtual void MultTransAdd (Complex s, const BaseVector & x, BaseVector & y) const;

};

#endif
