#ifndef FILE_FETI_DP_BILINEARFORM
#define FILE_FETI_DP_BILINEARFORM



template<class SCAL>
class FETIDP_BilinearForm : public S_BilinearForm<SCAL>
{
  CoefficientFunction * feticoeff;
  bool deletefeticoeff;
  INVERSETYPE inversetype;
public:
  FETIDP_BilinearForm (const FESpace & afespace, const string & aname,
		   const Flags & flags);
  virtual ~FETIDP_BilinearForm ();
  
  virtual void AllocateMatrix ();
  virtual BaseVector * CreateVector() const;
  
  virtual void AddElementMatrix (const ARRAY<int> & dnums1,
				 const ARRAY<int> & dnums2,
				 const FlatMatrix<SCAL> & elmat,
				 bool inner_element, int elnr,
				 LocalHeap & lh);  
  

  void ComputeSMatrix ( int elnr, LocalHeap & lh );

  virtual void DoAssemble (LocalHeap & lh);
  ///
  void ModifyRightHandSide ();

  void ComputePrimalDual ( BaseVector & vecu, LocalHeap & lh ) const ;

  virtual void ComputeInternal (BaseVector & u, LocalHeap & lh) const;

  BaseMatrix * CreateBMat () const;

  BaseMatrix * CreateBarBMat ( LocalHeap & lh) const;


#ifdef PARALLEL
  virtual void AllocateConsistentMatrix ()
  { 
    ; 
  }

  virtual void BuildConsistentMatrix (LocalHeap & lh) 
  { ; }
#endif

};




  template <class SCAL>
  class FETIDP_Preconditioner : public Preconditioner
  {
    const FETIDP_BilinearForm<SCAL> * bfa;
    FETIDP_PrecMatrix<SCAL> * precmat;

  public:
    FETIDP_Preconditioner (const PDE * pde, const Flags & aflags, 
			   const string aname = "hhprecond")
      : Preconditioner (pde, aflags, aname),
	precmat(0)
    {
      cout << "create helmholtz precond" << endl;
      bfa = dynamic_cast<const FETIDP_BilinearForm<SCAL>*>
	(pde->GetBilinearForm (aflags.GetStringFlag ("bilinearform", NULL)));

    }

    virtual ~FETIDP_Preconditioner() 
    { delete precmat; }

    static Preconditioner * Create (const PDE & pde, const Flags & flags)
    {
      return new FETIDP_Preconditioner<SCAL> (&pde, flags);
    }

    virtual void Update ();

    virtual const BaseMatrix & GetAMatrix() const
    {
      return bfa->GetMatrix();
    }

    virtual const BaseMatrix & GetMatrix() const
    {
      return *precmat;
    }
  };












#endif
