#include <fem.hpp>
#include <feti.hpp>
#include <la.hpp>


namespace ngfem
{
  using namespace ngfem;


  void FETIDP_CompoundFiniteElement :: ComputePrimalDualDofs ( int version )
  {
    primaldofs.SetSize(0);
    dualdofs.SetSize(0);

    const H1HighOrderFiniteElement& fel_h1 = dynamic_cast<const H1HighOrderFiniteElement&> 
      ( (*this)[0] );

    int nv = ElementTopology :: GetNVertices ( eltype );
    int ned = ElementTopology :: GetNEdges ( eltype);
    int nfa = (dimspace <= 2 ) ? 0 : ElementTopology::GetNFaces ( eltype );

    switch ( eltype )
      {
      case ET_TRIG:
	nv = 3; ned = 3; nfa = 0; break;
      case ET_SEGM:
	nv = 2; ned = 0; nfa = 0; break;
      case ET_QUAD:
	nv = 4; ned = 4; nfa = 0; break;
      case ET_TET:
	nv = 4; ned = 6; nfa = 4; break;
      case ET_HEX:
	nv = 8; ned = 12; nfa = 6; break;
      case ET_PYRAMID:
	nv = 5; ned = 7; nfa = 5; break;
      case ET_PRISM:
	nv = 6; ned = 9; nfa = 5; break;
      }

    int ii = 0;
    switch (version)
      {
      case 1:
	for ( int i = 0; i < nv; i++ )
	  primaldofs.Append(ii++);

	for ( int i = 0; i < ned; i++ )
	  {
	    for ( int j = 0; j < fel_h1.order_edge[i]+1; j++ )
	      dualdofs.Append(ii++);
	  }
	for ( int i = 0; i < nfa; i++ )
	  {
	    ELEMENT_TYPE facettype = ElementTopology::GetFacetType(eltype, i);
	    if ( facettype == ET_TRIG )
	      for ( int j = 0; j <= fel_h1.order_face[i][0]; j++ )
		for ( int k = j; k <= fel_h1.order_face[i][0]; k++ )
		  dualdofs.Append(ii++);
	  }

	break;

      case 2:
	for ( int i = 0; i < nv; i++ )
	  primaldofs.Append(ii++);

	for ( int i = 0; i < ned; i++ )
	  {
	    primaldofs.Append(ii++);
	    for ( int j = 1; j < fel_h1.order_edge[i]+1; j++ )
	      dualdofs.Append(ii++);
	  }
	for ( int i = 0; i < nfa; i++ )
	  {
	    ELEMENT_TYPE facettype = ElementTopology::GetFacetType(eltype, i);
	    if ( facettype == ET_TRIG )
	      for ( int j = 0; j <= fel_h1.order_face[i][0]; j++ )
		for ( int k = j; k <= fel_h1.order_face[i][0]; k++ )
		  dualdofs.Append(ii++);
	  }

      }

  }

}
