
class FETIDP_CompoundFiniteElement : public CompoundFiniteElement
{
 public:
  ARRAY<int> primaldofs;
  ARRAY<int> dualdofs;

  FETIDP_CompoundFiniteElement (ARRAY<const FiniteElement*> & afea)
    : CompoundFiniteElement ( afea )
  {
    primaldofs.SetSize(0);
    dualdofs.SetSize(0);
  }
  
  ~FETIDP_CompoundFiniteElement ()
  {
    ;
  }

  void ComputePrimalDualDofs ( int version );

};


/* Finite Element for Lagrange Parameters: has no shape, only dofs */
class LagrangeFiniteElement : public FiniteElement
{
private:
  ARRAY<INT<2> > elnums;
  ARRAY<int> dualdof;
  int firstvertdof[9];
  int firstedgedof[13];
  int firstfacedof[7];

public :
 
  LagrangeFiniteElement ( int adimspace, ELEMENT_TYPE aeltype, int andof, int aorder) 
    : FiniteElement ( adimspace, aeltype, andof, aorder )
  {
    elnums.SetSize(andof);
    dualdof.SetSize(andof);
  }


  virtual string ClassName(void) const {return "LagrangeFiniteElement";}

  virtual ~LagrangeFiniteElement () { ; }

  // no internal dofs
  virtual void GetInternalDofs (ARRAY<int> & idofs) const
  {
    idofs.SetSize(0);
  }

  const int & FirstVertexDof ( int locvert ) const
  { return firstvertdof[locvert]; }
  const int & FirstEdgeDof ( int locedge ) const
  { return firstedgedof[locedge]; }
  const int & FirstFaceDof ( int locedge ) const
  { return firstfacedof[locedge]; }

   int & FirstVertexDof ( int locvert ) 
  { return firstvertdof[locvert]; }
   int & FirstEdgeDof ( int locedge ) 
  { return firstedgedof[locedge]; }
   int & FirstFaceDof ( int locedge ) 
  { return firstfacedof[locedge]; }

  int DofOrientation ( int dofnr, int myelnum ) const
  { 
    int orient = 0;
    if ( myelnum == elnums[dofnr][0] ) orient = 1;
    else if ( myelnum == elnums[dofnr][1] ) orient = -1;
    return orient;
  }

  void SetElNums ( int dof, INT<2> dofelnums )
  { elnums[dof] = dofelnums; }

  INT<2> ElNums ( int dof ) const
  { return elnums[dof]; }

  void SetDualDof ( int lagdof, int adualdof )
  { dualdof[lagdof] = adualdof; }

  int DualDof ( int lagdof )
  { return dualdof[lagdof]; }

//   void SetDofOrientation ( int dofnr, bool orient = true )
//   { 
//     if ( orient ) orientation.Set(dofnr);
//     else orientation.Clear(dofnr);
//   }


//   bool VertexOrientation ( int locvert ) const
//   { return vertexorientation.Test(locvert); }
//   bool EdgeOrientation ( int locedge ) const
//   { return edgeorientation.Test(locedge); }
//   bool FaceOrientation ( int locface ) const
//   { return faceorientation.Test(locface); }

//   void SetVertexOrientation (int i,  bool orient = true ) 
//   { 
//     if ( orient ) vertexorientation.Set(i);
//     else vertexorientation.Clear(i);
//   }
//   void SetEdgeOrientation (int i,  bool orient = true ) 
//   { 
//     if ( orient ) edgeorientation.Set(i);
//     else edgeorientation.Clear(i);
//   }
//   void SetFaceOrientation (int i,  bool orient = true ) 
//   { 
//     if ( orient ) faceorientation.Set(i);
//     else faceorientation.Clear(i);
//   }

};



