#ifndef FILE_FETI_DP_SPACE
#define FILE_FETI_DP_SPACE



class FETIDP_PrimalSpace
{
protected:
  BitArray * dualdofs;
  BitArray * primaldofs;

public:
  int npvertdofs;
  int npedgedofs;
  int npfacedofs;

  FETIDP_PrimalSpace (const Flags & flags);
//   {
//     ;
//   }

  ~FETIDP_PrimalSpace ();
//   {
//     ;
//   }

  virtual void GetPrimalDualDofNrs ( int elnr, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualDofNrs ( int elnr, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualVertexDofNrs ( int elnr, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualEdgeDofNrs ( int elnr, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualFaceDofNrs ( int elnr, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualVertexDofNrs ( int elnr, int locvnum, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualEdgeDofNrs ( int elnr, int locednum, ARRAY<int> & dnums ) const = 0;
  virtual void GetDualFaceDofNrs ( int elnr, int locfanum, ARRAY<int> & dnums ) const = 0;

  virtual int GetNDualDofs ( int elnr ) const = 0;
  virtual int GetNDualEdgeDofs ( int ednum ) const = 0;
  virtual int GetNDualFaceDofs ( int ednum ) const = 0;
  virtual int GetNDualVertexDofs ( int ednum ) const = 0;

  virtual bool IsDualDof ( int dof ) const 
  { return dualdofs->Test(dof); }

  virtual bool IsPrimalDof ( int dof ) const 
  { return primaldofs->Test(dof); }

  virtual void SetDualDof ( int dof, bool isdual  = 1 ) 
  { 
    if (isdual) 
      {
	dualdofs->Set(dof);
	primaldofs->Clear(dof);
      }
    else
      {
	dualdofs->Clear(dof);
	primaldofs->Set(dof);
      }
  }

};



class FETIDP_H1HOSpace : public H1HighOrderFESpace, public FETIDP_PrimalSpace
{
  // number of vertex/edge/face-dofs in primal space
  ARRAY<int> first_vertex_dof;
  bool refinep;
  Table<int> * firstsdualvertexdof;
  Table<int> * firstsdualedgedof;
  ARRAY<int> * firstsdualfacedof;
public:
  FETIDP_H1HOSpace (const MeshAccess & ama, const Flags & flags, bool parseflags = false);
  
  virtual ~FETIDP_H1HOSpace () 
  { 
    if ( firstsdualvertexdof ) delete firstsdualvertexdof; 
    if ( firstsdualedgedof ) delete firstsdualedgedof;
    if ( firstsdualfacedof ) delete firstsdualfacedof; 
  }
    
  static FESpace * Create (const MeshAccess & ma, const Flags & flags)  
  {
    return new FETIDP_H1HOSpace(ma, flags, true);
  }


  virtual string GetClassName () const
  {
    return "FETIDP_H1HOSpace";
  }

  ///
  virtual void Update();
  void UpdateDofTables ();
  /// 
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetPrimalDualDofNrs ( int elnr, ARRAY<int> & dnums ) const;
  virtual void GetDualDofNrs ( int elnr, ARRAY<int> & dnums ) const;
  virtual void GetDualVertexDofNrs ( int elnr, ARRAY<int> & dnums ) const;
  virtual void GetDualEdgeDofNrs ( int elnr, ARRAY<int> & dnums ) const;
  virtual void GetDualFaceDofNrs ( int elnr, ARRAY<int> & dnums ) const;
  virtual void GetDualVertexDofNrs ( int elnr, int locvnum, ARRAY<int> & dnums ) const;
  virtual void GetDualEdgeDofNrs ( int elnr, int locednum, ARRAY<int> & dnums ) const;
  virtual void GetDualFaceDofNrs ( int elnr, int locfanum, ARRAY<int> & dnums ) const;

//   virtual int GetNDualDofs ( int elnr ) const;
//   virtual int GetNDualEdgeDofs ( int ednum ) const;
//   virtual int GetNDualFaceDofs ( int ednum ) const;
//   virtual int GetNDualVertexDofs ( int ednum ) const;
  virtual inline int GetNDualVertexDofs ( int vnum ) const
  {
    return 1-npvertdofs;
  }

  virtual inline int GetNDualEdgeDofs ( int ednum ) const
  {
    return order_edge[ednum]-1-npedgedofs;
  }

  virtual inline int GetNDualFaceDofs ( int fanum ) const 
  {
    INT<2> p = order_face[fanum];
    int ndof_face;
    ARRAY<int> fapnums;
    ma.GetFacePNums (fanum, fapnums);
    switch(fapnums.Size())
      {
      case 3: 
	if(p[0]>2)
	  return (p[0]-1)*(p[0]-2)/2-npfacedofs;
	break;
      case 4:
	if(p[0] > 1 && p[1]>1)
	  return (p[0]-1)*(p[1]-1)-npfacedofs;
	break; 
      }
    return 0;
  }

  virtual inline int GetNDualDofs ( int elnr ) const
  {
    int ndualdofs = 0;

    int i, j;
    ARRAY<int> ednums, fanums, fapnums;
    ma.GetElEdges (elnr, ednums);
    ma.GetElFaces(elnr, fanums);
    int nvert = ElementTopology :: GetNVertices ( ma.GetElType(elnr) );

    if ( !npvertdofs )
    for ( i = 0; i < nvert; i++ )
      ndualdofs ++;

    for (i = 0; i < ednums.Size(); i++)
      {
	ndualdofs += order_edge[ednums[i]]-1-npedgedofs;
      }

    for (i = 0; i < fanums.Size(); i++)
      {
	INT<2> p = order_face[fanums[i]];
	int ndof_face;
	ELEMENT_TYPE facettype = ElementTopology :: GetFacetType ( ma.GetElType(elnr), i );
	switch(facettype)
	  {
	  case ET_TRIG: 
	    if(p[0]>2)
	      ndualdofs += (p[0]-1)*(p[0]-2)/2-npfacedofs;
	    break;
	  case ET_QUAD:
	    if(p[0] > 1 && p[1]>1)
	      ndualdofs += (p[0]-1)*(p[1]-1)-npfacedofs;
	    break; 
	  }
      }

  }
};


// lagrange space for feti-dp
// ordering of dofs:
// 
// element 0   element 1   dofs
//             element 2   dofs
// element 1   element 2   dofs

class FETIDP_LagrangeSpace : public FESpace
{
  int ndof;
  ARRAY<int> first_vertex_dof;
  ARRAY<int> first_edge_dof;
  ARRAY<int> first_face_dof;
  ARRAY<INT<2> > dualdofs;
public:

  FETIDP_LagrangeSpace (const MeshAccess & ama, const Flags & flags, bool parseflags=false);
  ///
  virtual ~FETIDP_LagrangeSpace ()
  {
    ;
  }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  { return new FETIDP_LagrangeSpace ( ma, flags, false); }

  virtual string GetClassName () const
  {
    return "FETIDP_LagrangeSpace";
  }

  ///
  virtual void Update(const FETIDP_PrimalSpace & primalspace);
  /// 
  //virtual void UpdateDofTables();
  ///
  virtual int GetNDof () const; //  { return ndof; }
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  
  /// 
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;

  INT<2> DualDofs ( int lagdof ) const { return dualdofs[lagdof]; }

  void SetDualDofs ( int lagdof, INT<2> adualdofs ) { dualdofs[lagdof] = adualdofs; }

  Table<int> * CreateVertexElementsTable () const;
  Table<int> * CreateEdgeElementsTable () const;
  Table<int> * CreateFaceElementsTable () const;
};


class FETIDP_CompoundSpace : public CompoundFESpace
{
  FETIDP_PrimalSpace * primalspace;

  public:
    FETIDP_CompoundSpace (const MeshAccess & ama, 		   
                            const ARRAY<const FESpace*> & aspaces,
                            const Flags & flags);

    virtual ~FETIDP_CompoundSpace () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);
//     {
//       ARRAY<const FESpace*> spaces(2);
      
//       Flags primalflags(flags);
//       Flags dualflags(flags);

//       int order = int (flags.GetNumFlag ("order", 1));

//       primalflags.SetFlag ("order", order);
//       primalflags.SetFlag ("discontinuous");

//       string type = flags.GetStringFlag ( "space", "");
//       if ( type == "h1ho" )
// 	spaces[0] = new FETIDP_H1HOSpace (ma, primalflags);

//       const FETIDP_H1HOSpace * aprimalspace = 
// 	dynamic_cast<const FETIDP_H1HOSpace*> (spaces[0]);

//       spaces[1] = new FETIDP_LagrangeSpace ( ma, dualflags );

//       FETIDP_CompoundSpace * fes = new FETIDP_CompoundSpace (ma, spaces, flags);
      
//       fes->PrimalSpace () = *aprimalspace;
//       return fes;
//     }
    
    virtual string GetClassName () const { return "FETIDP_CompoundSpace"; }

  virtual void Update ();
    // direct solver clusters:

    // 0 ...... no clusters
    // 1 ...... low order facets
    // >= 2 ... divide sphere into n layers

    virtual ARRAY<int> * CreateDirectSolverClusters (const Flags & precflags) const
  {
    ;
  }
      
    // #define VER1
    Table<int> * CreateSmoothingBlocks ( const Flags & precflags) const
    {
      ;
    }

  void GetPDLDofNrs ( int elnr, ARRAY<int> & dnums ) const;

  virtual MatrixGraph * GetGraph (int level, bool symmetric);

  virtual MatrixGraph * GetGraph_S (bool symmetric);

  void SetPrimalSpace ( FETIDP_PrimalSpace * aprimalspace ) 
  { primalspace = aprimalspace; }

  FETIDP_PrimalSpace & PrimalSpace () 
  { return *primalspace; }

  const FETIDP_PrimalSpace & PrimalSpace () const
  { return *primalspace; }

};

#endif
