#include <solve.hpp>
#include <comp.hpp>

using namespace ngsolve;

#include "generalcontact.hpp"

using namespace contact;

namespace contact
{
  using namespace contact;
 Vec<3> Cross ( const Vec<3> veca, const  Vec<3> vecb)
 {
   Vec<3> vecc;
   vecc(0) = veca(1)*vecb(2) - veca(2)*vecb(1);
   vecc(1) = veca(2)*vecb(0) - veca(0)*vecb(2);
   vecc(2) = veca(0)*vecb(1) - veca(1)*vecb(0);
   return vecc;
 }
}



ContactInfo ::  ContactInfo( const MeshAccess & ma, const FESpace & afespace, const ARRAY<bool> & acontactbound, const ARRAY<bool> & amasterbound, const double & tmax)
{
  nsel = ma.GetNSE();
  ndof = ma.GetNV();
  fespace = &afespace;

  ARRAY<double> gap(ndof);
  for(int i=0; i<ndof; i++)
    gap[i] = tmax;

  MatchingSEl.SetSize(ndof);
  for(int i=0; i<ndof; i++)
    {
	MatchingSEl[i] = -1;
    }


  contactbound.SetSize ( acontactbound.Size() );
  masterbound.SetSize ( amasterbound.Size() );
  for(int index=0; index < acontactbound.Size(); index++)
    {
      contactbound[index] = acontactbound[index];
      masterbound[index] = amasterbound[index];
    }

  Vec<3> point0, point1, point2, pointx, n;
  Vec<3> p20, p21, p22;
  Vec<3> lambda;
  double t;
  ARRAY<int> vnums(3), vnums2(3);
  int domain1, domain2, dummy;

  MinSel.SetSize(ma.GetNV());
  for(int i=0; i<ndof; i++)
    MinSel[i] = nsel + 2;
  /*
  for(int sel=0; sel<nsel; sel++)
    {
      if(IsMasterBound(ma.GetSElIndex(sel)))
	{
	  ma.GetSElVertices (sel, vnums);
	  for(int i=0; i<3; i++)
	  MinSel[vnums[i]] = min2(MinSel[vnums[i]], sel);
	}
    }
  */

  for(int sel=0; sel<nsel; sel++)
    {
      if(IsMasterBound(ma.GetSElIndex(sel)))
	{
	  ma.GetSElVertices (sel, vnums);
	  ma.GetPoint(vnums[0], point0);
	  ma.GetPoint(vnums[1], point1);
	  ma.GetPoint(vnums[2], point2);
	  n = Cross(point1 - point0, point2 - point0);
	  n = 1.0 / L2Norm(n) * n;
	  
	  ma.GetSElNeighbouringDomains(sel, domain1, dummy);
	  for(int i=0; i<3; i++)
	    {
	      ma.GetPoint(vnums[i], pointx) ;
	      
	      for(int sel2=0; sel2<nsel; sel2++)
		{
		  if ( !IsContactBound (ma.GetSElIndex(sel2) ) ) continue;
		  if ( IsMasterBound (ma.GetSElIndex(sel2) ) ) continue;

		  ma.GetSElNeighbouringDomains(sel2, domain2, dummy);
		  if(domain1 != domain2 && ma.GetSElIndex(sel) != ma.GetSElIndex(sel2) )
		    {
		      ma.GetSElVertices (sel2, vnums2);
		      ma.GetPoint(vnums2[0], p20);
		      ma.GetPoint(vnums2[1], p21);
		      ma.GetPoint(vnums2[2], p22);
		      
		      CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		      if(lambda(0) >=0 && lambda(0) <=1 && lambda(1) >=0 && lambda(1) <=1 && lambda(2) >=0 && lambda(2) <=1 )
			{ 
			  if ( t < gap[vnums[i]] && abs(t) < tmax )
			    {
			      MatchingSEl[vnums[i]] = sel2;
			      gap[vnums[i]] = t;
			      MinSel[vnums[i]] = sel;
			    }
			}
		    }
		}
	    }
	  
	}
     
    }

  const IntegrationRule & ir = 
    GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);
  matchingip.SetSize(nsel);
  for(int i=0; i<matchingip.Size(); i++)
    {
      matchingip[i].SetSize(ir.GetNIP());
      for(int j=0; j<matchingip[i].Size(); j++)
	{
	  matchingip[i][j] = -1;
	}
    }


  for(int sel=0; sel<nsel; sel++)
    {
      if(IsMasterBound(ma.GetSElIndex(sel)) )
	{
	  ma.GetSElVertices (sel, vnums);
	  ma.GetPoint(vnums[0], point0);
	  ma.GetPoint(vnums[1], point1);
	  ma.GetPoint(vnums[2], point2);
	  n = Cross(point1 - point0, point2 - point0);
	  n = 1.0 / L2Norm(n) * n;
	  
	  for(int i=0; i<matchingip[sel].Size(); i++)
	    {
	      pointx = ir[i](0) * point0 + ir[i](1) * point1  + (1 - ir[i](0) - ir[i](1) ) *  point2;
	      
	      for(int sel2=0; sel2<nsel; sel2++)
		{
		  if(!IsMasterBound(ma.GetSElIndex(sel2) ) && IsContactBound(ma.GetSElIndex(sel2)) )
		    {
		      ma.GetSElVertices (sel2, vnums2);
		      ma.GetPoint(vnums2[0], p20);
		      ma.GetPoint(vnums2[1], p21);
		      ma.GetPoint(vnums2[2], p22);
		      
		      CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		      if(lambda(0) >=0 && lambda(0) <=1 && lambda(1) >=0 && lambda(1) <=1 && lambda(2) >=0 && lambda(2) <=1  && abs(t) < tmax)
			{
			  matchingip[sel][i] = sel2;
			}
		    }
		}
	    }
	}
    }

}





namespace contact
{
  using namespace contact;

void CalcProjectionN ( const Vec<3> & xx, const Vec<3> & nn, const Vec<3> & aa, const Vec<3> & bb, const Vec<3> & cc, Vec<3> & lambda, double & t)
{


  Vec<3> ne = Cross( aa - bb, aa - cc);

  double nenn = 0, na=0;
  lambda = -1.0;
  for(int i=0; i<3; i++)
    {
      nenn += ne(i) * nn(i);
      na += ne(i) * ( aa(i) - xx(i));
    }
  if (nenn == 0)
    {
      t = -1; lambda = -1; return;
    }
  t = 1.0 * na / nenn;
  Vec<3> ss = xx + t * nn;
  if(fabs(bb(1) - aa(1)) > 1e-10 && fabs( (cc(0) - aa(0))*(bb(1) - aa(1)) - (cc(1) - aa(1))*(bb(0) - aa(0)) ) > 1e-10 )
    {
      lambda(2) = 1.0 * ( (bb(1) - aa(1))*(ss(0) - aa(0)) + (bb(0) - aa(0))*(aa(1) - ss(1)) ) / 
	( (cc(0) - aa(0))*(bb(1) - aa(1)) - (cc(1) - aa(1))*(bb(0) - aa(0)) );
      
      lambda(1) = 1.0 * ( ss(1) - aa(1) - (cc(1) - aa(1)) * lambda(2)) / (bb(1) - aa(1));
      
      lambda(0) = 1.0 - lambda(1) - lambda(2);
    }
 
  else if ( fabs( (bb(1)-aa(1))*(cc(2)-aa(2)) - (bb(2)-aa(2))*(cc(1)-aa(1)) ) > 1e-10 && fabs(cc(2) - aa(2) ) > 1e-10 )
    {
      lambda(1) = 1.0 * ( (ss(1)-aa(1))*(cc(2)-aa(2)) - (ss(2)-aa(2))*(cc(1)-aa(1)) ) / 
	( (bb(1)-aa(1))*(cc(2)-aa(2)) - (bb(2)-aa(2))*(cc(1)-aa(1)) );

      lambda(2) = 1.0 * (ss(2) - aa(2) - lambda(1) * (bb(2) - aa(2)) ) / (cc(2) - aa(2) );

      lambda(0) = 1.0 - lambda(2) - lambda(1);
    }
}

}




  void NumProcCont :: Assembleg ( const MeshAccess & ma, const ContactInfo & cinfo, 
		   const CompoundFESpace & compound_fespace, Vector<double > & vecg)
  {
    int nsel = ma.GetNSE();
    int ndof_g = cinfo.GetNDof();
    //    int ncela = cmesh.GetNCEa(), ncelb = cmesh.GetNCEb();

    vecg.SetSize(ndof_g);
    vecg = 10.0;


    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);

    int fac =-1;
    int domain1, domain2, dummy;
    for(int vert=0; vert<ndof_g; vert++)
      {
	int sel = cinfo.GetMinSEl(vert);
	if(sel < nsel)
	  if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	    {
	      ma.GetSElVertices (sel, vnums);
	      ma.GetPoint(vnums[0], point0);
	      ma.GetPoint(vnums[1], point1);
	      ma.GetPoint(vnums[2], point2);
	      
	      n = Cross(point1 - point0, point2 - point0);
	      n = 1.0 / L2Norm(n) * n;
	      ma.GetSElNeighbouringDomains(sel, domain1, dummy);
	      ma.GetPoint(vert, pointx);
	      int sel2 = cinfo.GetMatchingSEl(vert);
		if(sel2 >= 0 && sel2 < nsel)
		  {
		    ma.GetSElVertices (sel2, vnums2);
		    ma.GetPoint(vnums2[0], p20);
		    ma.GetPoint(vnums2[1], p21);
		    ma.GetPoint(vnums2[2], p22);

		    
		    CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		    if(abs(t) > 0) vecg(vert) = t; 
		    
		  }
	    }
      }


  }




  void NumProcCont :: ApplyCTranspose( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
			const Vector<double > & w, const Vector<Vec<3> > & veclambda, BaseVector & rhs,
			LocalHeap & lh)
  {
   
    //rhs.FVDouble() = 1.0;
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    rhs.FVDouble() = 0.0;

    
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);

 
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*compound_fespace[1]);

    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> point0m, point1m, point2m;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3), vnumsm(3);

    // int sigma_nn w_n ds
    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsContactBound(ma.GetSElIndex(sel)))
	  {
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		n = Cross( point1 - point0, point2 - point0);
		double astriddet = L2Norm(n);
		n = 1.0 / L2Norm(n) * n;
		
		
		const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
		
		const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
		
		ARRAY<int> dofs(fel_sigma.GetNDof());
		
		ElementTransformation eltrans;
		ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		
		sigmaspace.GetSDofNrs (sel, dofs);
		
		FlatMatrixFixHeight<1> sigma(fel_sigma.GetNDof(), lh);
		
		SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );
		
		double det = fabs (sip.GetJacobiDet());
		DiffOpBoundIdHDivSym<3> :: GenerateMatrix(fel_sigma, sip, sigma, lh);

		double wvalue = w(vnums[2]) * (1-ir[l](0) - ir[l](1)) + w(vnums[0]) * ir[l](0) + w(vnums[1]) * ir[l](1);
		for(int i=0; i<fel_sigma.GetNDof(); i++)
		  {
		    rhs.FVDouble()(dofs[i]) -= 2* sigma(0,i) * (wvalue)  * astriddet *
		      ir[l].Weight();
		  }

	      }
	  }
      }

    // int u_tau lambda ds

    // integration on master boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		n = Cross( point1 - point0, point2 - point0);
		double astriddet = L2Norm(n);
		n = 1.0 / L2Norm(n) * n;
		
		
		const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
		
		const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
		
		ARRAY<int> dofsu(fel_u.GetNDof());
		
		ElementTransformation eltrans;
		ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		uspace.GetSDofNrs (sel, dofsu);
		
		FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		
		SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );
		
		double det = fabs (sip.GetJacobiDet());

		DiffOpIdBoundaryEdge<3> :: GenerateMatrix (fel_u, sip, u, lh);

		
		Vec<3> lambdavalue = ir[l](0) * veclambda(vnums[0]) + ir[l](1) * veclambda(vnums[1]) + (1 - ir[l](0) - ir[l](1)) * veclambda(vnums[2]);  

		for(int i=0; i<fel_u.GetNDof(); i++)
		  {
		    rhs.FVDouble()(dofsu[i] + ndof_sigma_total) -= 2 * (u(0,i)*lambdavalue(0) + u(1,i)*lambdavalue(1) + u(2,i)*lambdavalue(2)) 
		      * astriddet * ir[l].Weight();
		  }

	      }
	  }
      }


    // integration on slave boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsContactBound(ma.GetSElIndex(sel)) && !cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    for(int l=0; l<ir.GetNIP(); l++)
	      {

		int selm = cinfo.GetMatchingIP(sel,l);
		if(selm >= 0 && selm < nsel)
		  {
		    // points on slave boundary
		    ma.GetSElVertices (sel, vnums);
		    ma.GetPoint(vnums[0], point0);
		    ma.GetPoint(vnums[1], point1);
		    ma.GetPoint(vnums[2], point2);
		    n = Cross( point1 - point0, point2 - point0);
		    double astriddet = L2Norm(n);
		    n = 1.0 / L2Norm(n) * n;
		
		    pointx = ir[l](0) * point0 + ir[l](1) * point1 + (1 - ir[l](0) - ir[l](1)) * point2;

		    // slave element
		    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
		    
		    const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
		
		    ARRAY<int> dofsu(fel_u.GetNDof());
		    
		    ElementTransformation eltrans;
		    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		    
		    uspace.GetSDofNrs (sel, dofsu);
		    
		    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		    
		    SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );
		    
		    double det = fabs (sip.GetJacobiDet());
		    
		    DiffOpIdBoundaryEdge<3> :: GenerateMatrix (fel_u, sip, u, lh);

		    // get barycentric coordinates of point on master boundary
		    ma.GetSElVertices(selm, vnumsm);
		    ma.GetPoint(vnumsm[0], point0m);
		    ma.GetPoint(vnumsm[1], point1m);
		    ma.GetPoint(vnumsm[2], point2m);

		    CalcProjectionN (pointx, n, point0m, point1m, point2m, lambda, t);

		    Vec<3> lambdavalue = lambda(0) * veclambda(vnumsm[0]) + lambda(1) * veclambda(vnumsm[1]) + lambda(2) * veclambda(vnumsm[2]);  

		    for(int i=0; i<fel_u.GetNDof(); i++)
		      {
			rhs.FVDouble()(dofsu[i] + ndof_sigma_total) += 2 * (u(0,i)*lambdavalue(0) + u(1,i)*lambdavalue(1) + u(2,i)*lambdavalue(2)) 
			  * astriddet * ir[l].Weight();
		      }
		    
		  }
	      }
	  }
      }
    lh.CleanUp();
  }

  void NumProcCont :: ApplyCn( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
			       const BaseVector & rhs, Vector<double > & w, LocalHeap & lh)
  {
   
    //rhs.FVDouble() = 1.0;
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    w = 0.0;
    
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);

 
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*compound_fespace[1]);

    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);

    for(int sel=0; sel<nsel; sel++)
      {
	if( cinfo.IsContactBound ( ma.GetSElIndex(sel) ) )
	for(int l=0; l<ir.GetNIP(); l++)
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    n = Cross( point1 - point0, point2 - point0);
	    double astriddet = L2Norm(n);
	    n = 1.0 / L2Norm(n) * n;
	    

	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
	    
	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
	    
	    ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		
	    sigmaspace.GetSDofNrs (sel, dofs);
	    uspace.GetSDofNrs (sel, dofsu);

	    Vector<double> sigma(fel_sigma.GetNDof());
	    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		
	    SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );

	    double det = fabs (sip.GetJacobiDet());
	    //	    DiffOpBoundIdHDivSym<3> :: GenerateMatrix(fel_sigma, sip, sigma, lh);
	    fel_sigma.CalcShape (sip.IP(), sigma);
	    sigma *= 1.0 / sqr(astriddet);

	    DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);
	    double sigmavalue = 0.0;
	    for(int i=0; i<fel_sigma.GetNDof(); i++)
	      sigmavalue += sigma(i) * rhs.FVDouble()(dofs[i]);

	    w(vnums[2]) -= sigmavalue * (1 - ir[l](0) - ir[l](1)) * astriddet  * ir[l].Weight();
	    w(vnums[0]) -= sigmavalue * (ir[l](0)) * astriddet * ir[l].Weight();
	    w(vnums[1]) -= sigmavalue * (ir[l](1)) * astriddet * ir[l].Weight();
	      
   


	  }
	   
      }
    lh.CleanUp();
  }


  void NumProcCont :: ApplyCTau( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
			       const BaseVector & rhs, Vector<Vec<3> > & veclambda, LocalHeap & lh)
  {
   
  
    int ndof_sigma_total = compound_fespace[0]->GetNDof();
    int nsel = ma.GetNSE();
    veclambda = 0.0;
    
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);

 
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*compound_fespace[1]);

    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> point0m, point1m, point2m;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnumsm(3);
   

    // integration on master boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if( cinfo.IsMasterBound ( ma.GetSElIndex(sel) ) )
	for(int l=0; l<ir.GetNIP(); l++)
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    n = Cross( point1 - point0, point2 - point0);
	    double astriddet = L2Norm(n);
	    n = 1.0 / L2Norm(n) * n;
	    

	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
	    
	    const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
	    
	    ARRAY<int> dofsu(fel_u.GetNDof());

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
	    uspace.GetSDofNrs (sel, dofsu);

	    //Vector<double> sigma(fel_sigma.GetNDof());
	    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		
	    SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );

	    double det = fabs (sip.GetJacobiDet());
	    
	    DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);
	    Vec<3> uvalue = 0.0;
	    for(int i=0; i<fel_u.GetNDof(); i++)
	      {
		uvalue(0) += u(0,i) * rhs.FVDouble()(dofsu[i] + ndof_sigma_total);
		uvalue(1) += u(1,i) * rhs.FVDouble()(dofsu[i] + ndof_sigma_total);
		uvalue(2) += u(2,i) * rhs.FVDouble()(dofsu[i] + ndof_sigma_total);
	      }
	    veclambda(vnums[2]) -= (1 - ir[l](0) - ir[l](1)) * astriddet  * ir[l].Weight() * uvalue;
	    veclambda(vnums[0]) -= (ir[l](0)) * astriddet * ir[l].Weight() * uvalue;
	    veclambda(vnums[1]) -= (ir[l](1)) * astriddet * ir[l].Weight() * uvalue;
	  }
	   
      }

    // integration on slave boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if( cinfo.IsContactBound(ma.GetSElIndex(sel)) && !cinfo.IsMasterBound ( ma.GetSElIndex(sel) ) )
	  for(int l=0; l<ir.GetNIP(); l++)
	    {

	      int selm = cinfo.GetMatchingIP(sel,l);
	      if(selm >= 0 && selm < nsel)
		{
		  // vnums, points on slave
		  ma.GetSElVertices (sel, vnums);
		  ma.GetPoint(vnums[0], point0);
		  ma.GetPoint(vnums[1], point1);
		  ma.GetPoint(vnums[2], point2);
		  n = Cross( point1 - point0, point2 - point0);
		  double astriddet = L2Norm(n);
		  n = 1.0 / L2Norm(n) * n;
	    
		  pointx = ir[l](0) * point0 + ir[l](1) * point1 + (1 - ir[l](0) - ir[l](1)) * point2;
		  // slave element
		  const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
	    
		  const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
		
		  ARRAY<int> dofsu(fel_u.GetNDof());

		  ElementTransformation eltrans;
		  ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		  uspace.GetSDofNrs (sel, dofsu);

		  //Vector<double> sigma(fel_sigma.GetNDof());
		  FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		
		  SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );
		
		  double det = fabs (sip.GetJacobiDet());
		
		  DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);
		  Vec<3> uvalue = 0.0;
		  for(int i=0; i<fel_u.GetNDof(); i++)
		    {
		      uvalue(0) += u(0,i) * rhs.FVDouble()(dofsu[i] + ndof_sigma_total);
		      uvalue(1) += u(1,i) * rhs.FVDouble()(dofsu[i] + ndof_sigma_total);
		      uvalue(2) += u(2,i) * rhs.FVDouble()(dofsu[i] + ndof_sigma_total);
		    }
		
		  // master element: find barycentric coordinates of matching point

		  ma.GetSElVertices(selm, vnumsm);
		  ma.GetPoint(vnumsm[0], point0m);
		  ma.GetPoint(vnumsm[1], point1m);
		  ma.GetPoint(vnumsm[2], point2m);

		  CalcProjectionN (pointx, n, point0m, point1m, point2m, lambda, t);

		  veclambda(vnumsm[2]) += (lambda(2)) * astriddet * ir[l].Weight() * uvalue;
		  veclambda(vnumsm[0]) += (lambda(0)) * astriddet * ir[l].Weight() * uvalue;
		  veclambda(vnumsm[1]) += (lambda(1)) * astriddet * ir[l].Weight() * uvalue;
		}
	    }
      }
    lh.CleanUp();
  }



  void NumProcCont :: ApplyBI( const MeshAccess & ma, const ContactInfo & cinfo,
			       const Vector<double> & w, Vector<double > & p)
  {
    int nsel = ma.GetNSE();
    int ndof_g = cinfo.GetNDof();
    
    Vec<3> point0, point1, point2, pointx, n, n2;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);
    p = 0.0;
    for(int vert=0; vert<ndof_g; vert++)
      {
	int sel = cinfo.GetMinSEl(vert);
	if(sel < nsel)
	  if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	    {
	      ma.GetSElVertices (sel, vnums);
	      ma.GetPoint(vnums[0], point0);
	      ma.GetPoint(vnums[1], point1);
	      ma.GetPoint(vnums[2], point2);
	      n = Cross(point1 - point0, point2 - point0);
	      n = 1.0 / L2Norm(n) * n;

	   
	      ma.GetPoint(vert, pointx);
	      int sel2 = cinfo.GetMatchingSEl(vert);
	      
	      if(sel2 >= 0 && sel2 < nsel)
		{
		  ma.GetSElVertices (sel2, vnums2);
		  ma.GetPoint(vnums2[0], p20);
		  ma.GetPoint(vnums2[1], p21);
		  ma.GetPoint(vnums2[2], p22);
		    
		  CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		  double wvalue = 0.0;
		  for(int i=0; i<3; i++)
		    wvalue += w(vnums2[i]) * lambda(i);
		  //w(vnums2[2]) * abs(1 - lambda(0) - lambda(1)) + w(vnums2[0]) * abs(lambda(0)) + w(vnums2[1]) * abs(lambda(1));
		  p(vert) = w(vert) + wvalue;
	
		}
	    }
      }  
    
  }


void NumProcCont :: ApplyBITrans( const MeshAccess & ma, const ContactInfo & cinfo, 
			       const Vector<double > & p, Vector<double> & w)
  {
    
    int nsel = ma.GetNSE();
    int ndof_g = cinfo.GetNDof();

    Vec<3> point0, point1, point2, pointx, n, n2;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);

    w = 0.0;
    
    for(int vert=0; vert<ndof_g; vert++)
      {
	int sel = cinfo.GetMinSEl(vert);
	  if(sel < nsel)
	    if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		n = Cross(point1 - point0, point2 - point0);
		n = 1.0 / L2Norm(n) * n;
		ma.GetPoint(vert, pointx);
		int sel2 = cinfo.GetMatchingSEl(vert);
	      
		if(sel2 >= 0 && sel2 < nsel)
		  {
		    ma.GetSElVertices (sel2, vnums2);
		    ma.GetPoint(vnums2[0], p20);
		    ma.GetPoint(vnums2[1], p21);
		    ma.GetPoint(vnums2[2], p22);
		    
		    CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		    w(vert) = p(vert);
		    w(vnums2[2]) += p(vert)  * (1-lambda(0)-lambda(1));
		    w(vnums2[0]) += p(vert) * lambda(0);
		    w(vnums2[1]) += p(vert) * lambda(1);
		  }
	      }
      }
    
    
  }
 

void NumProcCont :: FindPhiBeta(const MeshAccess & ma, const ContactInfo & cinfo, const Vector<double> & vecg, const Vector<double> & vecw, 
				const Vector<double> & vecr, Vector<double> & phi, Vector<double> & beta)
{
  int nsel = ma.GetNSE();
  int ndof_g = cinfo.GetNDof();
  
  Vector<double> lambda(ndof_g), p(ndof_g), q(ndof_g), rhs(ndof_g), r(ndof_g);
  Vector<double> v(ndof_g);
  double np = 1.0/ndof_g;
  v = 0.0;
  p = 0.0;
  q = 0.0; lambda = 0.0;
  rhs = 0.0;
  r=0;
  ApplyBI( ma, cinfo, vecr, rhs);
   
  ApplyBI( ma, cinfo, vecw, v);
   
  for(int i=0; i<vecg.Size(); i++)
    {if(v(i) >= vecg(i) -1e-10 ) rhs(i) = rhs(i);
      else rhs(i) = 0.0;}
  r = - rhs;
  p = r;
  double alpha_cg, alpha, alpha_f;
  bool flag = false;
  int step = 0;

  while( L2Norm(p) > 1e-16 && step < 1000)
    {
      step ++;
      ApplyBITrans(ma, cinfo, p, v);
      ApplyBI(ma, cinfo, v, q);
      double psp = InnerProduct(q, p);
      alpha_cg = InnerProduct(r, p);
      alpha_cg = alpha_cg / psp;
      lambda = lambda - alpha_cg * p;
       
      r = r - alpha_cg * q;
      double gamma = InnerProduct(r, q) / psp;
      p = r - gamma * p;
    }

  ApplyBITrans (ma, cinfo, lambda, v);
  phi =   vecr - v;
  cout  << "inner cg " <<step << endl;
  ApplyBI(ma, cinfo, phi, r);
  cout << "l2r " << L2Norm(r ) << endl;

  v = 0.0; p = 0.0; q = 0.0; lambda = 0.0;
  ApplyBI( ma, cinfo, vecr, rhs);
   
  ApplyBI( ma, cinfo, vecw, v);
 
  for(int i=0; i<vecg.Size(); i++)
    {
      if(v(i) >= vecg(i) - 1e-10) 
	rhs(i) = max2(rhs(i),0.0);
      else 
	rhs(i) = rhs(i);
    }
  r = - rhs;
  p = r;
  step=0;
  while( L2Norm(p) > 1e-16 && step < 1000)
    {
      step ++;
      ApplyBITrans(ma, cinfo, p, v);
      ApplyBI(ma, cinfo, v, q);
      double psp = InnerProduct(q, p);
      alpha_cg = InnerProduct(r, p);
      alpha_cg = alpha_cg / psp;
      lambda = lambda - alpha_cg * p;
       
      r = r - alpha_cg * q;
      double gamma = InnerProduct(r, q) / psp;
      p = r - gamma * p;
    }

  ApplyBITrans (ma, cinfo, lambda, v);
  beta = vecr - v;

}


void NumProcCont :: FindPhiBetaTau(const MeshAccess & ma, const ContactInfo & cinfo, 
				   const Vector<double > & sigmann, const Vector<Vec<3> > & veclambda, 
				   const Vector<Vec<3> > & vecr, Vector<Vec<3> > & phi, Vector<Vec<3> > & beta)
{
   int nsel = ma.GetNSE();
   int ndof_g = cinfo.GetNDof();
   Vec<3> n, lambdaneu;

   if (coeff_f == 0)
     {
       phi = 0.0;
       beta = 0.0;
     }
   else
     {
       ARRAY<int> vnums(3), vnums2(3);
       /*
	 for(int sel=0; sel<nsel; sel++)
	 {
	 if( cinfo.IsMasterBound ( ma.GetSElIndex(sel) ) )
	 for(int l=0; l<3; l++)
	 {
	 ma.GetSElVertices (sel, vnums);
	 
	 double normlambda = L2Norm( veclambda(vnums[l]) );
	 if(normlambda >= coeff_f * sigmann(sel)(l) - 1e-10 && normlambda > 0)
	 {
	 n = 1.0 / L2Norm(veclambda(vnums[l])) * veclambda(vnums[l]);
	 double lambdan = InnerProduct ( n, vecr(vnums[l]) );
		phi(vnums[l]) = 0.0;
		beta(vnums[l]) = vecr(vnums[l]) - max2(0.0, lambdan) * n ;
		}
		else
	      {
	      phi(vnums[l]) = vecr(vnums[l]);
	      beta(vnums[l]) = 0.0;
	      }
	      }
	      }
       */
       for(int vert=0; vert<ndof_g; vert++)
	 {
	   double normlambda = L2Norm(veclambda(vert));
	   if(sigmann(vert) == 0)
	     {
	       phi(vert) = 0;
	       beta(vert) = 0;
	     }
	   else if(normlambda >= coeff_f * sigmann(vert) - 1e-6 && normlambda > 0)
	     {
	       lambdaneu = veclambda(vert) -  tautau * vecr(vert);
	       double t = coeff_f * abs(sigmann(vert))/ L2Norm(lambdaneu);
	       lambdaneu = t * lambdaneu;
	       beta(vert) = 1.0 / tautau * (veclambda(vert) - lambdaneu);
	       phi(vert) = 0.0;
	     }
	   else
	     {
 	       phi(vert) = vecr(vert);
 	       beta(vert) = 0.0;
// 	       phi(vert) = 0.0;
// 	       beta(vert) = vecr(vert);

	     }
	 }
     }
}




 void NumProcCont :: ProjectDisp2 (const MeshAccess & ma, const ContactInfo & cinfo, const Vector<double > & vecg, Vector<double> & w)
 {
   
   int nsel = ma.GetNSE();
   int ndof_g = cinfo.GetNDof();
   Vector<double> lambda(ndof_g), p(ndof_g), q(ndof_g), rhs(ndof_g), r(ndof_g), phi(ndof_g), beta(ndof_g);
   Vector<double> v(ndof_g);   Vector<double> w2(ndof_g);

   w2 = w;
   v = 0.0;
   p = 0.0;
   q = 0.0; lambda = 0.0;
   rhs = 0.0;
   r=0; phi=0; beta=0;

   ApplyBI( ma, cinfo, w, rhs);
   for(int i=0; i<ndof_g; i++)
     rhs(i) = -vecg(i) + rhs(i);

   r = - rhs;
   phi = 0;
   p = phi;
   double alpha_cg, alpha, alpha_f;
   bool flag = false;
   for(int i=0; i<ndof_g; i++)
     beta(i) = min2(r(i),0.0);
   int step = 0;
   

   while( L2Norm(phi+beta) > 1e-16 && step < 200)
     {
       
       step ++;
       if(L2Norm(beta) < L2Norm(phi) )
	 {
	   
	   ApplyBITrans(ma, cinfo, p, v);
	   ApplyBI(ma, cinfo, v, q);
	   double psp = InnerProduct(q, p);
	   alpha_cg = InnerProduct(r, p);
	   alpha_cg = alpha_cg / psp;
	   alpha_f = alpha_cg;
	   lambda = lambda - alpha_cg * p;
	   
	   for(int j=0; j<ndof_g; j++)
	     {
	       if(lambda(j) < 0){lambda(j) = 0; flag = true;}
	     }

	   if(flag)
	     {
	       ApplyBITrans(ma, cinfo, lambda, v);
	       ApplyBI(ma, cinfo, v, r);
	       r = r - rhs;
	       for(int j=0; j<ndof_g; j++)
		 {
		   if(lambda(j) > 0) {phi(j) = r(j); beta(j) = 0;}
		   else {phi(j) = 0; beta(j) = min2(r(j), 0.0);}
		 }
	       p = phi;
	     }
	   
	   else
	     {
	       r = r - alpha_cg * q;
	       
	       for(int j=0; j<ndof_g; j++)
		 {
		   if(lambda(j) > 0) {phi(j) = r(j); beta(j) = 0;}
		   else {phi(j) = 0; beta(j) = min2(r(j), 0.0);}
		 }
	       
	       double gamma = InnerProduct(phi,q) / psp;
	       
	       //p = phi - gamma * p;
	       
	       p *= -gamma;
	       p += phi;
	       
	     }
	   
	 }
       
       else
	 {
	   
	   lambda -= taup * beta;
	   
// 	   (*testout) << "damits geht " << endl;
	   ApplyBITrans(ma, cinfo, lambda, v);
	   ApplyBI(ma, cinfo, v, r);
	   
	   r -=  rhs;
	   
	   for(int j=0; j<ndof_g; j++)
	     {
	       if(lambda(j) > 0) {phi(j) = r(j); beta(j) = 0;}
	       else {lambda(j) = 0; phi(j) = 0; beta(j) = min2(r(j), 0.0);}
	     }
	   
	   p = phi;
	   
	 }

     }
   
   ApplyBITrans (ma, cinfo, lambda, v);
  
   w = w - v;
   ARRAY<int> vnums(3);
   
   (*testout) << "project disp " << step << endl;
       ApplyBI(ma,cinfo,w,v);
     for(int sel=0; sel<nsel; sel++)
       if(ma.GetSElIndex(sel) == cbound1)
	 {
	   ma.GetSElVertices(sel, vnums);
	   if(vecg(vnums[0]) < v(vnums[0]))
	     {
	       (*testout) << "w " << sel << "  " << v(vnums[0]) << "  " << v(vnums[1]) << "  " << v(vnums[2]) << endl;
	       (*testout) << "g " << sel << "  " << vecg(vnums[0]) << "  " << vecg(vnums[1]) << "  " << vecg(vnums[2]) << endl;
	     }
	 }
     cout << "aendert " << L2Norm(w2 - w)<< endl;

   
 }

void NumProcCont :: ProjectDisp ( const MeshAccess & ma, const ContactInfo & cinfo,  const Vector<double > & vecg, Vector<double > & w)
  {
    
    int nsel = ma.GetNSE();
    int ndof_g = cinfo.GetNDof();
    /*
    Vector<double > wneu(ndof_g);
    wneu = w;
    Vec<3> point0, point1, point2, pointx, n, n2;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);
    Vector<int> flag(nsel);
    flag = 0;
    int fac =-1;
    int domain1, domain2, dummy;
    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    n = Cross(point1 - point0, point2 - point0);
	    n = 1.0 / L2Norm(n) * n;
	    ma.GetSElNeighbouringDomains(sel, domain1, dummy);
	    for(int i=0; i<3; i++)
	      {
		ma.GetPoint(vnums[i], pointx);
		int sel2 = cinfo.GetMatchingSEl(sel, i);
	      
		if(sel2 >= 0 && sel2 < nsel)
		  {
		    ma.GetSElVertices (sel2, vnums2);
		    ma.GetPoint(vnums2[0], p20);
		    ma.GetPoint(vnums2[1], p21);
		    ma.GetPoint(vnums2[2], p22);
		    
		    CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		    double wvalue = w(vnums2[2]) * (1 - lambda(0) - lambda(1)) + w(vnums2[0]) * lambda(0) + w(vnums2[1]) * lambda(1);
		    double gvalue = vecg(vnums2[2]) * (1 - lambda(0) - lambda(1)) + vecg(vnums2[0]) * lambda(0) + vecg(vnums2[1]) *lambda(1);
		    double vg = 0.5 * (vecg(vnums[i]) + gvalue);
		    if(w(vnums[i]) + wvalue > vecg(vnums[i]))
		      {
			wneu(vnums[i]) = 0.5 * (vecg(sel*3+i) + w(3*sel+i) - wvalue);
			wneu(vnums2[0]) = 0.5 * (vecg(sel*3+i) - w(3*sel+i) + wvalue);
			wneu(3*sel2+1) = 0.5* (vecg(sel*3+i) - w(3*sel+i) + wvalue);
			wneu(3*sel2+2) = 0.5 *(vecg(sel*3+i) - w(3*sel+i) + wvalue);
		      
		      }
		    
		  }
	      }
	  }
      }
    for(int vert=0; vert < ndof_g; vert++)
       
      w(vert) = wneu(vert);
    */
  }




void NumProcCont :: CalcSigmann( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
				 const BaseVector & vecu, Vector<double > & sigmann, LocalHeap & lh)

{
  
  int ndof_sigma_total = compound_fespace[0]->GetNDof();
  int nsel = ma.GetNSE();
  sigmann.SetSize(cinfo.GetNDof());
  sigmann = 10000.0;
    
  const IntegrationRule & ir = 
    GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);

 
  const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
  const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*compound_fespace[1]);

  Vec<3> point0, point1, point2, pointx, n;
  Vec<3> p20, p21, p22;
  Vec<3> lambda;
  double t;
  ARRAY<int> vnums(3), vnums2(3);
  
  for(int sel=0; sel<nsel; sel++)
    {
      if( cinfo.IsMasterBound ( ma.GetSElIndex(sel) ) )
	for(int l=0; l<3; l++)
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    n = Cross( point1 - point0, point2 - point0);
	    double astriddet = L2Norm(n);
	    n = 1.0 / L2Norm(n) * n;
	    

	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
	    
	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
	    
	    ARRAY<int> dofs(fel_sigma.GetNDof()), dofsu(fel_u.GetNDof());

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		
	    sigmaspace.GetSDofNrs (sel, dofs);
	    uspace.GetSDofNrs (sel, dofsu);

	    FlatMatrixFixHeight<1> sigma(fel_sigma.GetNDof(), lh);
		
	    Vec<2> vertex = 0.0;
	    if( l == 0) vertex(0) = 1.0;
	    else if (l == 1) vertex(1) = 1.0;

	    IntegrationPoint ip(vertex(0), vertex(1));
	    SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );

	    double det = fabs (sip.GetJacobiDet());
	    
	    DiffOpBoundIdHDivSym<3> :: GenerateMatrix(fel_sigma, sip, sigma, lh);

	    double sigmavalue = 0.0;
	    for(int i=0; i<fel_sigma.GetNDof(); i++)
	      {
		sigmavalue += sigma(0,i) * vecu.FVDouble()(dofs[i]);
	      }
	    sigmann(vnums[l]) = min(abs(sigmavalue), abs(sigmann(vnums[l])) );
	    if(abs(sigmann(vnums[l])) < 1e-2) sigmann(vnums[l]) = 0.0;


	  }
	   
    }
  
  lh.CleanUp();
    
}



void NumProcCont :: SolveGMRES ( const Vec<5> & alpha, const Vec<5> & beta, Vec<5> & y) 
{
  Vec<5> gamma = alpha, beta2 = beta;
  Vec<5> rhs = 0;
  rhs(0) = beta(0);
  double m;

  for(int i=0; i<4; i++)
    {
      m = beta(i+1) / gamma(i);
      gamma(i+1) = alpha(i+1) - m * beta(i+1);
      rhs(i+1) = - m * rhs(i);
    }
  y(4) = rhs(4) / gamma(4);
  for(int i=3; i>=0; i--)
    {
      y(i) = (rhs(i) - alpha(i+1) * y(i+1)) / gamma(i);
    }

  return;
}



bool NumProcCont :: ProjectStress ( const ContactInfo & cinfo, const MeshAccess & ma, const double coeff_f, const Vector<double > & sigmann, 
				    Vector<Vec<3> > & lambda)
  {

    bool flag = false;
    int ndof_c = lambda.Size();
    int nsel = ma.GetNSE();
    ARRAY<int> vnums(3);
    /*
    for(int sel=0; sel<nsel; sel++)
      if(cinfo.IsMasterBound(ma.GetSElIndex(sel)) )
	{
	  ma.GetSElVertices (sel, vnums);
	  
	  for(int i=0; i<3; i++)
	    {
	      double normlambda = L2Norm(lambda(vnums[i]));
	      
	      if(normlambda > abs(coeff_f * sigmann(sel)(i)) + 1e-10 )
		{
		  double t = abs(coeff_f * sigmann(sel)(i)) / normlambda;
		  lambda(vnums[i]) = t * lambda(vnums[i]);
		  cout << "rel. change lambda " << vnums[i] << "   " << t << endl;
		  flag = true;
		}
	    }
	}
    */
    for(int vert=0; vert<ndof_c; vert++)
      {
	double normlambda = L2Norm(lambda(vert));

	if(normlambda > abs(coeff_f * sigmann(vert)) + 1e-5)
	  {
	    double t = abs(coeff_f * sigmann(vert)) / normlambda;
	    lambda(vert) = t * lambda(vert);
	    if( t > 0.1 && t < 1) cout << "rel. change lambda " << vert << "    " << t << endl;
	    flag = true;
	  }
      }
    return flag;
  }
    
/*
void NumProcCont :: CalcSigmann (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
				    const BaseVector & vecu, Vector<double> & sigmann, LocalHeap & lh)
{

    int nsel = ma.GetNSE();
    int ncel = cmesh.GetNCE();
    int ncvert = cmesh.GetNDof();
    //    int ncela = cmesh.GetNCEa(), ncelb = cmesh.GetNCEb();

    Vec<3> point_0, point_a, point_b;
    cmesh.GetDomain(point_0, point_a, point_b);
    Vec<3> normal = Cross(point_0 - point_a, point_0 - point_b);
    double length = sqrt( sqr(normal(0)) + sqr(normal(1)) + sqr(normal(2)));
    normal = 1.0 / length * normal;

    ARRAY<int> vnums(3);
    Vec<3> point0, point1, point2, lambda;
    double t;
    int fac =-1;

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);

    sigmann = 0.0;
    
    for(int cvert=0; cvert<ncvert; cvert++)
      {
	Vec<3> xx = cmesh.GetPoint(cvert);
	int sel = cmesh.GetMatchingSEl(cvert, 0);
	if(sel >= 0)
	  {
	    ma.GetSElVertices (sel, vnums);
	    ma.GetPoint(vnums[0], point0);
	    ma.GetPoint(vnums[1], point1);
	    ma.GetPoint(vnums[2], point2);
	    CalcProjectionN(xx, normal, point0, point1, point2, lambda, t);
	    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));

	    const HDivSymNormalFiniteElement & fel_sigma = dynamic_cast< const HDivSymNormalFiniteElement & > (fel[0]);
	    ARRAY<int> dofs(fel_sigma.GetNDof());

	    sigmaspace.GetSDofNrs (sel, dofs);

	    ElementTransformation eltrans;
	    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
	    IntegrationPoint ip(lambda(0), lambda(1));

	    SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );
	    
	    double det = fabs (sip.GetJacobiDet());
	    Vector<double> sigma(fel_sigma.GetNDof());

	    fel_sigma.CalcShape(sip.IP(), sigma);
	    sigma *= 1.0 / sqr(det);
	    for(int i=0; i<fel_sigma.GetNDof(); i++)
	      sigmann(cvert) += sigma(i) * vecu.FVDouble()(dofs[i]);
	  }
	else 
	  sigmann(cvert) = 0;
      }
}








*/



  void NumProcCont :: Do(LocalHeap & lh)
  {
    cout << "solve contact problem" << endl;
    lh.CleanUp();
    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();
    BaseVector & vecv = *vecu.CreateVector();

    const CompoundFESpace  &   compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HCurlHighOrderFESpaceA & vspace = dynamic_cast<const HCurlHighOrderFESpaceA & > ( *compound_fespace[1]);
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);

    MeshAccess  ma =  bfa -> GetFESpace(). GetMeshAccess();

    int ndof, nd_u, nd_sigma;
    int nsel = ma.GetNSE();

    ofstream outfilephiw("/home/astrid/netgen/ngsolve/astrid/freegrad.dat"), outfilephilambda("/home/astrid/netgen/ngsolve/astrid/philambda.dat"),
      outfilebetaw("/home/astrid/netgen/ngsolve/astrid/choppedgrad.dat"), outfilebetalambda("/home/astrid/netgen/ngsolve/astrid/betalambda.dat");
    ofstream outfileres("/home/astrid/netgen/ngsolve/astrid/residual.dat"); 
    ofstream outfilestep("/home/astrid/netgen/ngsolve/astrid/step.dat"); 
    ndof = compound_fespace.GetNDof();
    nd_u = compound_fespace[1]->GetNDof();
    nd_sigma = compound_fespace[0]->GetNDof();

    ///
    int maxsteps = 3;
    ///
    double prec = 1e-5;
    ///
    bool print=false;
    ///
    enum SOLVER { CG, GMRES, QMR, NCG };
    ///
    SOLVER solver = CG;

    // ********* in short **************
    // make contact mesh

    // find gap function g

    // solve variational inequality by uzawa type method

    // given w_0, lambda_0

    // while ( error > eps or n < nmax or whatever)

    //   right hand side
    //   rhs = f - (apply_cn(w), apply_ctau(lambda) )

    //   find (sigma, u) by solving system
    //   mata * vecu = rhs

    //   update w, lambda in direction of steepest descent/ascent
    //   w = w - tau * (apply_cn(sigma))
    //   lambda = lambda + tau * apply_ctau(u)
   
    //   project lambda, w onto admissible sets

    //   w = max ( w, g )

    //   lambda = max(min( lambda, 1), -1);

    // end;

    // ************ details ******************

    // make contact mesh
    // int ncel = contactelements;
    int intorder = 2;
    ARRAY<bool> contactbound(15), masterbound(15); //contactbound.SetSize(5);
    for(int i=0; i<15; i++)
      {
	contactbound[i] = 0; masterbound[i] = 0;
      }

    contactbound[cbound1] = 1;
    contactbound[cbound2] = 1;
    masterbound[mbound] = 1;

    ContactInfo cinfo (ma, compound_fespace, contactbound, masterbound, tmax);
    int nd_c = cinfo.GetNDof();
    // find gap function g
    Vector<double > vecg (ma.GetNV());
    Assembleg ( ma, cinfo, compound_fespace, vecg);
    cout << endl << nd_c << endl;
    (*testout) << "gggggg" << vecg << endl;

    // init lambda_0, w_0
    Vector<Vec<3>  > veclambda(nd_c),  lambdaold(nd_c); Vector<double  > vecw(nd_c), vecwnew(nd_c), wtest(nd_c), wzeros(nd_c), sr(nd_c);
    Vector<double > cbeta(nd_c), r(nd_c), phi(nd_c), beta(nd_c), p(nd_c), sp(nd_c), vech(nd_c);
    Vector<double > vecsigmann(nd_c);
    Vector<Vec<3> > rtau(nd_c), phitau(nd_c), betatau(nd_c), ptau(nd_c), sptau(nd_c), vechtau(nd_c), lambdazeros(nd_c), srtau(nd_c);

    vecsigmann = 0.0;
    vecw = 0.0;
    veclambda = 0.0;
    cbeta = 0.0;
    vechtau = 0.0;
    ARRAY<int> vnums(3);

    // solve by Uzawa type method
    BaseVector & rhs = *vecf.CreateVector();

    const BaseMatrix * premat = NULL;
    if (pre)  premat = &(pre->GetMatrix()); 
     KrylovSpaceSolver * invmat;

	switch (solver)
	  {
	  case CG:
	    cout << "cg solve for real problem" << endl;
	    invmat = new CGSolver<double>(mata, *premat);
	    break;
	  case QMR:
	    invmat = new QMRSolver<double>(mata, *premat);
	    break;
	  case GMRES:
	    invmat = new GMRESSolver<double>(mata, *premat);
	    break;
	  }

    invmat->SetMaxSteps (maxsteps);
    invmat->SetPrecision (prec);
    invmat->SetPrintRates ();
    invmat->SetInitialize (0);
    invmat->SetStatusHandler(ma);

//     clock_t starttime, endtime, soltime;
//     starttime = clock();

    int uzawasteps = 0;
    double elambda = 1.0, ew = 1.0;
    double eps = 1e-9;
    double epsw=0, epslambda =0;
   
    
    /*
    
    
    ApplyCTranspose( ma, cinfo, compound_fespace, vecw, veclambda, vecv, lh);
    invmat->Mult(vecv, vecu);
    
    ApplyCn(ma, cinfo, compound_fespace, vecu, r, lh);
     ApplyCTau(ma, cinfo, compound_fespace, vecu, rtau, lh);
    // vecw = 0, also r = -f
    invmat->Mult( vecf, vecv);

    ApplyCn(ma, cinfo, compound_fespace, vecv, vech, lh);
    ApplyCTau(ma, cinfo, compound_fespace, vecv, vechtau, lh);
    r = vech - r;
    rtau = vechtau - rtau;

    Matrix<double> pgmres(nd_c,5);
    Matrix<Vec<3> > ptaugmres(nd_c, 5);
    Vec<5> ca, cb, y;
    pgmres = 0.0;
    ptaugmres = 0.0;


    vecsigmann = 0.0;
    FindPhiBeta(ma, cinfo, vecg, vecw, r, phi, beta);
    FindPhiBetaTau(ma, cinfo, vecsigmann, veclambda, rtau, phitau, betatau);
   
    p = phi;
     ptau = phitau;
     ptau = 0;
//     cout << "l2betatau " << L2Norm(betatau) << endl;
    int step = 0;
    double l2phi, l2beta, alpha_cg = 1;
    l2phi = L2Norm(phi);
    l2beta = L2Norm(beta);
    ApplyBI (ma, cinfo, beta, cbeta);
    ApplyBI (ma, cinfo, vecw, wtest);
    cout << l2phi << "  " << l2beta << endl;
     l2phi = 0; l2beta = 1;

     while( l2phi + l2beta > 3e-4 && step < numuzawasteps && alpha_cg > 0)// && alpha_cg *( L2Norm(p) + L2Norm(ptau)) > 1e-10)// && abs(alpha_cg) > 1e-10)
      {
	if(step%10 == 0)
	  {
	    CalcSigmann( ma, cinfo, compound_fespace, vecu, vecsigmann, lh );
	    FindPhiBeta(ma, cinfo, vecg, vecw, r, phi, beta);
	    FindPhiBetaTau(ma, cinfo, vecsigmann, veclambda, rtau, phitau, betatau);
	    p = phi; ptau = phitau; alpha_cg = 1;
	  }
	cout << L2Norm(vecsigmann) << endl;
	if(l2phi > l2beta)// && alpha_cg )//* L2Norm(p) > 1e-17)
	  {
	    cout << "cg step " << endl;
	    outfilestep << step << "  cg step: ";
	    bool flag = false;
	    int igmres = 0;
	    for(int i=0; i<nd_c; i++)
	      {
		pgmres(igmres,i) = phi(i);
		ptaugmres(igmres, i) = phitau(i);
	      }

	    p = phi; ptau = phitau;
	    while(igmres < 4)
	      {
		ApplyCTranspose( ma, cinfo, compound_fespace, p, ptau, vecv, lh);
		invmat->Mult(vecv, vecv);
		ApplyCn(ma, cinfo, compound_fespace, vecv, sp, lh);
		ApplyCTau( ma, cinfo, compound_fespace, vecv, sptau, lh);

		ca(igmres) = InnerProduct(p, sp) + InnerProduct(ptau, sptau);
		p = sp - ca(igmres) * p;
		ptau = sptau - ca(igmres) * ptau;
		if(igmres > 0)
		  {
		    for(int i=0; i<nd_c; i++)
		      {
			p(i) -= cb(igmres-1) * pgmres(igmres-1,i);
			ptau(i) -= cb(igmres-1) * ptaugmres(igmres-1,i);
		      }
		  }
		cb(igmres+1) = sqrt(sqr(L2Norm(p)) + sqr(L2Norm(ptau)) );
		p *= 1.0 / cb(igmres+1);
		ptau *= 1.0 / cb(igmres+1);
		for(int i=0; i<nd_c; i++)
		  {
		    pgmres(igmres+1, i) = p(i);
		    ptaugmres(igmres+1,i) = ptau(i);
		  }
		igmres++;
	      }


	    SolveGMRES(ca, cb, y);
	    for(int igmres=0; igmres<4; igmres++)
	      {
		for(int i=0; i<nd_c; i++)
		  {
		    vecw(i) += pgmres(igmres,i) * y(igmres);
		    veclambda(i) += y(igmres) * ptaugmres(igmres,i);
		  }
	      }
	    ProjectDisp2(ma, cinfo, vecg,  vecw);
	    flag = ProjectStress (cinfo, ma, coeff_f, vecsigmann, veclambda);
	    ApplyCTranspose( ma, cinfo, compound_fespace, vecw, veclambda, vecv, lh);
	    invmat->Mult(vecv, vecv);
	    
	    ApplyCn(ma, cinfo, compound_fespace, vecv, r, lh);
 	    ApplyCTau(ma, cinfo, compound_fespace, vecv, rtau, lh);

	    r = vech - r;
 	    rtau = vechtau - rtau;
	    
	    FindPhiBeta(ma, cinfo, vecg, vecw, r, phi, beta);
 	    FindPhiBetaTau(ma, cinfo, vecsigmann, veclambda, rtau, phitau, betatau);
		
	    p = phi;	  
 	    ptau = phitau;
	    cout << L2Norm(phi) << "    " << L2Norm(phitau) << "    " << L2Norm(beta) << "        " << L2Norm(betatau) << endl;
	  }
	else
	  {
	    cout << "projection step " << endl;
	    outfilestep << step << "  projection step " << endl;

	    vecw += taun * beta;
 	    veclambda -= tautau * rtau;
	    //ProjectDisp(ma, cinfo, vecg, vecw);
	    ProjectDisp2(ma, cinfo, vecg,  vecw);
	    bool flag; flag = ProjectStress (cinfo, ma, coeff_f, vecsigmann, veclambda);
	    ApplyCTranspose( ma, cinfo, compound_fespace, vecw, veclambda, vecv, lh);
	    invmat->Mult(vecv, vecv);
	    
	    ApplyCn(ma, cinfo, compound_fespace, vecv, r, lh);
 	    ApplyCTau(ma, cinfo, compound_fespace, vecv, rtau, lh);

	    r = vech - r;
 	    rtau = vechtau - rtau;
	    
	    FindPhiBeta(ma, cinfo, vecg, vecw, r, phi, beta);
 	    FindPhiBetaTau(ma, cinfo, vecsigmann, veclambda, rtau, phitau, betatau);
		
	    p = phi;	  
 	    ptau = phitau;
	    cout << L2Norm(phi) << "    " << L2Norm(phitau) << "    " << L2Norm(beta) << "        " << L2Norm(betatau) << endl;

	  }
	step ++;
	outfilephilambda << step << ", " << L2Norm(phitau) << ", " << endl;
	outfilephiw << step << ", " << L2Norm(phi) << ", " << endl;
	outfilebetalambda << step << ", " << L2Norm(betatau)<<" , " << endl;
	outfilebetaw << step << ", " <<L2Norm(beta)<<", " << endl;
	outfileres << step << ", " << L2Norm(beta) + L2Norm(phi)<<", " << endl;
	cout << step << endl;
	l2phi = L2Norm(phi) + L2Norm(phitau);
	l2beta = L2Norm(beta) + L2Norm(betatau);

      }
     outfilephilambda << step << ", " << L2Norm(phitau) << endl;
     outfilephiw << step << ", " << L2Norm(phi) << endl;
     outfilebetalambda << step << ", " << L2Norm(betatau) << endl;
     outfilebetaw << step << ", " << L2Norm(beta) << endl;
     ApplyCTranspose(ma, cinfo, compound_fespace, vecw, veclambda, rhs, lh);
     rhs.FVDouble() = vecf.FVDouble() - rhs.FVDouble();
     invmat->Mult (rhs, vecu);

     // CalcSigmann( ma, cinfo, compound_fespace, vecu, vecsigmann, lh );
     

  
    */
 
     


     rtau = 0.0; r=0; vecv.FVDouble() = 1.0;
     while( uzawasteps < numuzawasteps && L2Norm(vecv.FVDouble()) > 1e-5)
     {
     wtest = vecw;
     lambdaold = veclambda;

     //   right hand side
     //   rhs = f - (apply_cn(w), apply_ctau(lambda) )
     rhs.FVDouble() = 0.0;
     ApplyCTranspose( ma, cinfo, compound_fespace, vecw, veclambda, rhs, lh);
     rhs.FVDouble() = vecf.FVDouble() - rhs.FVDouble();
     //   find (sigma, u) by solving system
     //   mata * vecu = rhs
     mata.Mult ( vecu, vecv);

     vecv.FVDouble() = vecv.FVDouble() - rhs.FVDouble();

     invmat->Mult (rhs, vecu);

     //     //   update w, lambda in direction of steepest descent/ascent
     //     //   w = w - tau * (apply_cn(sigma))
     //     //   lambda = lambda + tau * apply_ctau(u)
	

     ApplyCn (ma, cinfo, compound_fespace, vecu, r,  lh);
     ApplyCTau (ma, cinfo, compound_fespace, vecu, rtau, lh);


     for(int i = 0; i < nd_c; i++)
     {
       vecw(i) += taun*r(i);
       veclambda(i) -= tautau*rtau(i);
     }

    

     //     //   project lambda, w onto admissible set
	
     // // 	ProjectDisp (ma, cmesh, compound_fespace, vecu, vecg, vecw, lh);
     ProjectDisp2 ( ma, cinfo, vecg, vecw);


     ApplyBI(ma, cinfo, vecw, wtest);
     for(int sel=0; sel<nsel; sel++)
       if(ma.GetSElIndex(sel) == cbound1)
	 {
	   ma.GetSElVertices(sel, vnums);
	   if(vecg(vnums[0]) < 5)
	     {
	       (*testout) << "w " << sel << "  " << wtest(vnums[0]) << "  " << wtest(vnums[1]) << "  " <<wtest(vnums[2]) << endl;
	       (*testout) << "g " << sel << "  " << vecg(vnums[0]) << "  " << vecg(vnums[1]) << "  " << vecg(vnums[2]) << endl;
	     }
	 }

     CalcSigmann( ma, cinfo, compound_fespace, vecu, vecsigmann, lh );
     (*testout) << L2Norm(veclambda) << "  ";
     bool flag = ProjectStress (cinfo, ma, coeff_f, vecsigmann, veclambda);
     (*testout) << L2Norm(veclambda) << "  " <<flag << endl;
     Ng_Redraw();
     uzawasteps ++;
     //cout << ew / epsw  << "         " << elambda / epslambda  << endl;
     epslambda=0; elambda=1;
     cout << "step " << uzawasteps << ":   " << L2Norm(vecv.FVDouble() ) << endl;
     (*testout) << "step " << uzawasteps << ":   " << L2Norm(vecv.FVDouble() ) << endl;
     }
       
     

     ApplyBI(ma, cinfo, vecw, wtest);
     for(int sel=0; sel<nsel; sel++)
       if(ma.GetSElIndex(sel) == cbound1)
	 {
	   ma.GetSElVertices(sel, vnums);
	   if(vecg(vnums[0]) < 5)
	     {
	       (*testout) << "w " << sel << "  " << wtest(vnums[0]) << "  " << wtest(vnums[1]) << "  " <<wtest(vnums[2]) << endl;
	       (*testout) << "g " << sel << "  " << vecg(vnums[0]) << "  " << vecg(vnums[1]) << "  " << vecg(vnums[2]) << endl;
	     }
	 }
    
     
     /*
       cout << "lambda_n = " << endl;
       for(int i=0; i<veclambdax.Size(); i++)
       cout << veclambdax(i) << "   " << veclambday(i) <<  "  " << veclambdaz(i) << " sigma " << vecsigmann(i) <<  endl;
       cout << "numberofsteps " << uzawasteps << endl;
       for(int ia =0; ia < ncela+1; ia++){
       for(int ib =0; ib < ncelb+1; ib++)
       cout << vecsigmann(ia + ib*(ncela+1)) << "  ";
       cout << endl;}
     */
     delete invmat;
     
  }












// declare the numproc 'contact' 

namespace
{
  class Init
  { 
  public: 
    Init ();
  };
    
  Init::Init()
  {
    GetNumProcs().AddNumProc ("generalcontact", NumProcCont::Create, NumProcCont::PrintDoc);
  }
    
  Init init;
}
  
