#ifndef CONTACTGENERAL
#define CONTACTGENERAL

namespace contact
{
  void CalcProjectionN ( const Vec<3> & xx, const Vec<3> & nn, const Vec<3> & aa, const Vec<3> & bb, const Vec<3> & cc, Vec<3> & lambda, double & t);
  
  Vec<3> Cross ( const Vec<3> veca, const  Vec<3> vecb);
}

class ContactInfo
{
protected: 

  int ndof;
  int nsel;
  ARRAY<bool> contactbound;
  const FESpace * fespace;
  ARRAY<int> MatchingSEl;
  ARRAY<bool> masterbound;
  int ncel;
  ARRAY<int> MinSel;

  int intorder;
  ARRAY<ARRAY<int> > matchingip;

public:



  ContactInfo( const MeshAccess & ma, const FESpace & afespace, const ARRAY<bool> & acontactbound, const ARRAY<bool> & amasterbound, const double & tmax);


  ~ContactInfo()
  { 
   ;
  }


  const int GetNCE() const
  {
    return ncel;
  }


  const int GetNDof() const
  {
    return ndof;
  }

  const FESpace* GetFESpace() const
  {
    return fespace;
  }

  //  const double EvaluateContactFunction(const Vector<double> & vecf, const Vec<3> & x) const;

  const int  GetMatchingSEl ( const int & sel) const
  {
    int sel2 = MatchingSEl[sel];
    return sel2;
  }

  const int GetMinSEl (const int & vert) const
  {
    return MinSel[vert];
  }


  const bool IsContactBound(const int index) const
  {
    return contactbound[index];
  }
  const bool IsMasterBound(const int index) const
  {
    return masterbound[index];
  }
  /*
  void GetContactElement(const Vec<3> & ipx, int & cel) const
  {
    int cela = int ( ncel_a * 1.0 * (ipx(0) - point_0(0)) / (point_a(0) - point_0(0)));
    int celb = int ( ncel_b * 1.0 * (ipx(1) - point_0(1)) / (point_b(1) - point_0(1)));
    if(cela == ncel_a) cela --;
    if(celb == ncel_b) celb --;

    cel = cela + celb*ncel_a;

  }
  */
  const int GetMatchingIP(const int sel2, const int ip) const
  {
    return matchingip[sel2][ip];
  }

};





class NumProcCont : public NumProc
{
protected:
  // bilinear-form for the Hellinger Reissner system of elasticity
  BilinearForm * bfa;
  // linear-form providing the right hand side, contains (0, f)
  LinearForm * lff;
  // solution vector, containes (sigma, u)
  GridFunction * gfu;

  //contact surfaces
  int cbound1, cbound2;
  int mbound;

  Preconditioner * pre;

  double taun, tautau, taup;
  int numuzawasteps;
  int contactelements;
  double coeff_f;
  int ncela, ncelb;
  double tmax;
public:
  
    
  NumProcCont (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {

    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", "a"));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", "f"));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", "u"));

    cbound1 = int(flags.GetNumFlag ( "cbound1" , 1.0));
    cbound2 = int(flags.GetNumFlag ( "cbound2" , 2.0));
    mbound = int (flags.GetNumFlag ( "masterbound", cbound1 ) );

    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""), 1);

    taun = flags.GetNumFlag ( "tau", 1e-5);
    tautau = taun;
    taup = flags.GetNumFlag ("taup", 1e-5);

    taun = flags.GetNumFlag ( "taun", taun);
    tautau = flags.GetNumFlag ( "tautau", taun);

    numuzawasteps = int(flags.GetNumFlag( "usteps", 10));
    contactelements = int(flags.GetNumFlag( "cels", 10));

    coeff_f = flags.GetNumFlag( "friction", 1.0);

    ncela = int( flags.GetNumFlag( "ncela", 10));
    ncelb = int( flags.GetNumFlag( "ncelb", 10));
    tmax = flags.GetNumFlag( "gap", 0.1);
  }


  virtual ~NumProcCont() 
  { ; }


  // creates an solver object
  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcCont (pde, flags);
  }



  void Assembleg ( const MeshAccess & ma, const ContactInfo & cinfo, 
		   const CompoundFESpace & compound_fespace, Vector<double > & vecg);

  void  ApplyCTranspose( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
			const Vector<double> & w, const Vector<Vec<3> > & lambda, BaseVector & rhs,
			 LocalHeap & lh);

  void ApplyCn( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
			       const BaseVector & rhs, Vector<double > & w, LocalHeap & lh);

  void ProjectDisp ( const MeshAccess & ma, const ContactInfo & cinfo,  const Vector<double> & vecg, Vector<double > & w);

  void ApplyCTau( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
		  const BaseVector & rhs, Vector<Vec<3> > & veclambda, LocalHeap & lh);
  void  CalcSigmann( const MeshAccess & ma, const ContactInfo & cinfo, const CompoundFESpace & compound_fespace,
		     const BaseVector & vecu, Vector<double > & sigmann, LocalHeap & lh);


  bool ProjectStress ( const ContactInfo & cinfo, const MeshAccess & ma, const double coeff_f, const Vector<double > & sigmann, 
		       Vector<Vec<3> > & lambda);

  void ProjectDisp2 (const MeshAccess & ma, const ContactInfo & cinfo, const Vector<double  > & vecg, Vector<double> & w);
  

  void  ApplyBITrans( const MeshAccess & ma, const ContactInfo & cinfo, 
		      const Vector<double > & p, Vector<double> & w);
  void ApplyBI( const MeshAccess & ma, const ContactInfo & cinfo,
		const Vector<double> & w, Vector<double> & p);


  void FindPhiBeta(const MeshAccess & ma, const ContactInfo & cinfo, const Vector<double> & vecg, const Vector<double> & vecw, 
		   const Vector<double> & r, Vector<double> & phi, Vector<double> & beta);

  void FindPhiBetaTau(const MeshAccess & ma, const ContactInfo & cinfo, 
		      const Vector<double > & sigmann, const Vector<Vec<3> > & veclambda, 
		      const Vector<Vec<3> > & vecr, Vector<Vec<3> > & phi, Vector<Vec<3> > & beta);

  void  SolveGMRES ( const Vec<5> & alpha, const Vec<5> & beta, Vec<5> & y) ;


  /*
  void ApplyCTranspose( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
			const Vector<double > & w0, const Vector<double> & w1, 
			const Vector<double> & lambdax, const Vector<double> & lambday, const Vector<double> & lambdaz, BaseVector & rhs,
			LocalHeap & lh);




  void ApplyCn ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		 const BaseVector & vecu, Vector<double> & updatew0, Vector<double> & updatew1, LocalHeap & lh );



  void ApplyCtau ( const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		   const BaseVector & vecu, Vector<double> & updatelambdax, Vector<double> & updatelambday, Vector<double> & updatelambdaz, LocalHeap & lh );


  void ProjectDisp ( const Vector<double > & vecg, Vector<double> & w);


  void ProjectStress ( const double coeff_f, const Vector<double> & sigmann, 
		       Vector<double > & lambdax, Vector<double > & lambday, Vector<double > & lambdaz );


  void CalcSigmann (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		    const BaseVector & vecu, Vector<double> & sigmann, LocalHeap & lh);
  */
  virtual void Do(LocalHeap & lh);

  /*
  void ProjectDisp (const MeshAccess & ma, const ContactMesh3d & cmesh, const CompoundFESpace & compound_fespace,
		    const BaseVector & vecu, const Vector<double> & vecg, Vector<double > & w0, Vector<double> & w1, LocalHeap & lh);

  */
  virtual string GetClassName () const
  {
    return " Solver for Contact Problems ";
  }

  virtual void PrintReport (ostream & ost)
  {
    ost << GetClassName() << endl
	<< "Bilinear-form A = " << bfa->GetName() << endl
	<< "Linear-form     = " << lff->GetName() << endl
	<< "Gridfunction    = " << gfu->GetName() << endl;
  }

  ///
  static void PrintDoc (ostream & ost)
  {
    /*    ost << 
      "\n\nNumproc Parabolic:\n" \
      "------------------\n" \
      "Solves a parabolic partial differential equation by an implicite Euler method\n\n" \
      "Required flags:\n" 
      "-bilinearforma=<bfname>\n" 
      "    bilinear-form providing the stiffness matrix\n" \
      "-bilinearformm=<bfname>\n" 
      "    bilinear-form providing the mass matrix\n" \
      "-linearform=<lfname>\n" \
      "    linear-form providing the right hand side\n" \
      "-gridfunction=<gfname>\n" \
      "    grid-function to store the solution vector\n" 
      "-dt=<value>\n"
      "    time step\n"
      "-tend=<value>\n"
      "    total time\n"
	<< endl;
    */
  }
};









#endif
