#ifndef FILE_HCURLHOFE_A
#define FILE_HCURLHOFE_A  

// zur Numerischen Berechnung von curl(phi) bei TETS: 
// #define NUMCURLTET 
// #define NUMCURLPRISM 
//  #define NUMCURLHEX
#define NUMCURLPYR_A

/*********************************************************************/
/* File:   hcurlhofe.hpp                                             */
/* Author: Sabine Zaglmayr                                           */
/* Date:   20. Maerz 2003                                            */
/*********************************************************************/
   
/**
   High order H(curl) finite element
 */




template <int D>
class HCurlHighOrderFiniteElementA : public HCurlFiniteElementD<D> 
{
 
public:
  //  enum { DIM = D };

protected:
  // int ndof; 
  int vnums[8]; 
  INT<3> order_inner;
  INT<2> order_face[6];
  int order_edge[12];
  //int ndof_edge; 
  int ned; // number of edges in element  
  int nv; // number of vertices in element 
  int nf; // number of faces in element
  int usegrad_face[6]; 
  int usegrad_cell; 
  int usegrad_edge[12]; 
  int order_vertex[8];   // for augmented
  int augmented; 
 

  int alternative;

public:
  ///
  HCurlHighOrderFiniteElementA (ELEMENT_TYPE aeltype);

  void SetVertexNumbers (FlatArray<int> & avnums);
  void SetOrderInner (int oi);
  void SetOrderInner (INT<3> oi);
  void SetOrderFace (FlatArray<int> & of);
  void SetOrderFace (FlatArray<INT<2> > & of); 
  void SetOrderEdge (FlatArray<int> & oen);
  void SetUsegradEdge(FlatArray<int> & uge); 
  void SetUsegradFace(FlatArray<int> & ugf); 
  void SetUsegradCell(int ugc); 
  void SetOrderVertex (FlatArray<int> & ov);
  void SetAugmented (int aa);
  const   int GetAlternative () const;

  virtual void ComputeNDof () = 0;
  virtual void ComputeSNDof() = 0;

  void PrintInfo() const;
};
 
/// 
template <class T_ORTHOPOL> 
class HCurlHighOrderSegmA:  public HCurlHighOrderFiniteElementA<1>
{
private: 
   typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
 
public:
  
  HCurlHighOrderSegmA (int aorder);
  HCurlHighOrderSegmA (int aorder, int aalternative);
  virtual void ComputeNDof();
  virtual void ComputeSNDof();
 
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<1> shape) const;
  
};

///
template <class T_ORTHOPOL> 
class HCurlHighOrderTrigA : public HCurlHighOrderFiniteElementA<2>
{
private:
  typedef TrigShapesInnerLegendre T_INNERSHAPES; 
  typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
public:

  HCurlHighOrderTrigA (int aorder);
  HCurlHighOrderTrigA (int aorder, int aalternative);
  virtual void ComputeNDof();
  virtual void ComputeSNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
 
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<2> shape) const;
 
  /// compute Curl of shape 
  virtual void CalcCurlShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<1> curlshape) const; 
 
};


///
template <class T_ORTHOPOL>
class HCurlHighOrderQuadA : public HCurlHighOrderFiniteElementA<2>
{
private: 
   typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
public:
  HCurlHighOrderQuadA (int aorder);
  HCurlHighOrderQuadA (int aorder, int aalternative);
  virtual void ComputeNDof();
  virtual void ComputeSNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
 
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<2> shape) const;

   /// compute Curl of shape 
  virtual void CalcCurlShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<1> curlshape) const; 
  
};


///
template <class T_ORTHOPOL> 
class HCurlHighOrderTetA : public HCurlHighOrderFiniteElementA<3>
{
private:
  typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
  typedef TetShapesFaceLegendre T_FACESHAPES; 
  typedef TetShapesInnerLegendre T_INNERSHAPES; 
public:
  HCurlHighOrderTetA (int aorder);
  HCurlHighOrderTetA (int aorder, int aalternative);

  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const;
 #ifndef NUMCURLTET
  ///compute Curl of shape 
  virtual void CalcCurlShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<3> shape) const;
#endif 


  void ComputeSNDof() {;}
};

///
template <class T_ORTHOPOL>
class HCurlHighOrderHexA : public HCurlHighOrderFiniteElementA<3>
{
private: 
  typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
public:
  HCurlHighOrderHexA (int aorder);
  HCurlHighOrderHexA (int aorder, int aalternative);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
 
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const;
#ifndef NUMCURLHEX
  /// compute Curl of shape 
  virtual void CalcCurlShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<3> curlshape) const;
#endif

  void ComputeSNDof() {;}
};

///
template <class T_ORTHOPOL> 
class HCurlHighOrderPrismA : public HCurlHighOrderFiniteElementA<3>
{
private:
  typedef TrigShapesInnerLegendre T_TRIGFACESHAPES;
  typedef VertexExtensionOptimal<3> T_VERTEXSHAPES;
public:
  HCurlHighOrderPrismA (int aorder);
  HCurlHighOrderPrismA (int aorder, int aalternative);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const;

/// compute curl of shape
#ifndef NUMCURLPRISM
  virtual void CalcCurlShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<3> shape) const;
#endif
 
  void ComputeSNDof() {;}
};

///
template <class T_ORTHOPOL>
class HCurlHighOrderPyrA : public HCurlHighOrderFiniteElementA<3>
{
  typedef TrigShapesInnerLegendre T_TRIGFACESHAPES; 

public:

  HCurlHighOrderPyrA (int aorder);
  HCurlHighOrderPyrA (int aorder, int aalternative);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
 
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const;

  /// compute Curl of shape 
#ifndef NUMCURLPYR_A
  virtual void CalcCurlShape (const IntegrationPoint & ip, 
			      FlatMatrixFixWidth<3> curlshape) const;
#endif 
  void ComputeSNDof() {;}
};



#endif

