#include <comp.hpp>
#include <fem.hpp>  

#ifdef PARALLEL
#include <parallelngs.hpp>
#endif

namespace ngcomp
{
  using namespace ngcomp; 

#ifdef PARALLEL
  using namespace ngparallel;
#endif

  HDivHighOrderFESpaceA ::  
  HDivHighOrderFESpaceA (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : HDivHighOrderFESpace (ama, flags, parseflags)
  {
    ;
  }

  const FiniteElement & HDivHighOrderFESpaceA :: GetFE (int elnr, LocalHeap & lh) const
  {
    FiniteElement * fe;
    
    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;
    // typedef TrigExtensionMonomial T_ORTHOPOL;
    

    
    switch (ma.GetElType(elnr))
      {
      case ET_TET:
	{ 
	  fe = new (lh.Alloc (sizeof(HDivHighOrderTet<T_ORTHOPOL>)))  HDivHighOrderTetA<T_ORTHOPOL> (order);
	  break;
	}
	/*
      case ET_PYRAMID:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderPyramidSZ<T_ORTHOPOL>)))  HDivHighOrderPyramidSZ<T_ORTHOPOL> (order);
	  break;
	}
	*/
      case ET_PRISM:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderPrism<T_ORTHOPOL>)))  HDivHighOrderPrism<T_ORTHOPOL> (order);
	  break;
	}
      case ET_HEX:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderHex<T_ORTHOPOL>)))  HDivHighOrderHex<T_ORTHOPOL> (order);
	  break;
	}
      case ET_TRIG:
	{ 
	  fe = new (lh.Alloc (sizeof(HDivHighOrderTrig<T_ORTHOPOL>)))  HDivHighOrderTrigA<T_ORTHOPOL> (order);
	  break;
	}
      case ET_QUAD:
	{
	  fe = new (lh.Alloc (sizeof(HDivHighOrderQuad<T_ORTHOPOL>)))  HDivHighOrderQuad<T_ORTHOPOL> (order);
	  break;
	}
      default:
	fe = 0; 
      }
  
    if (!fe)
      {
	stringstream str;
	str << "HDivHighOrderFESpace " << GetClassName() 
	    << ", undefined eltype " 
	    << ElementTopology::GetElementName(ma.GetElType(elnr))
	    << ", order = " << order << endl;
	throw Exception (str.str());
      }
    
    ARRAY<int> vnums;
    ma.GetElVertices(elnr, vnums);
    if ( ma.GetDimension() == 2)
      {
	
	HDivHighOrderFiniteElement<2> * hofe =
	  dynamic_cast<HDivHighOrderFiniteElement<2>*> (fe);
	ArrayMem<int, 12> ednums, order_ed;
	
	ma.GetElEdges(elnr, ednums);
	order_ed.SetSize (ednums.Size());
	
	for (int j = 0; j < ednums.Size(); j++)
	  order_ed[j] = order_facet[ednums[j]][0];
      
	hofe -> SetVertexNumbers (vnums);
	hofe -> SetOrderEdge (order_ed);
	
	// #ifndef NEW_HDIVFE // old version 
	//  hofe -> SetOrderInner (order_inner[elnr][0]);
	// hofe -> SetDiscontinuous(discont); 
	// #else
        // new anisotropic FE
	hofe -> SetOrderInner (order_inner[elnr]); 
	//hofe -> SetOrderInnerCurl (order_inner_curl[elnr]);
	hofe -> SetDiscontinuous(discont); 

	hofe -> ComputeNDof();
      }
      else // dim=3
      {
	HDivHighOrderFiniteElement<3> * hofe =
	  dynamic_cast<HDivHighOrderFiniteElement<3>*> (fe);
	
	ArrayMem<int, 6> fanums; 
	ArrayMem<INT<2>, 6> order_fa;
	ma.GetElFaces(elnr, fanums);
	order_fa.SetSize (fanums.Size());
	
	for (int j = 0; j < fanums.Size(); j++)
	  order_fa[j] = order_facet[fanums[j]];
	
	hofe -> SetVertexNumbers (vnums);
	
	// #ifndef NEW_HDIVFE // old  version 
	// ArrayMem<int,6> order_fa2;
	// order_fa2.SetSize (fanums.Size()); 
	// for (int j = 0; j < fanums.Size(); j++)
	//   order_fa2[j] = order_facet[fanums[j]][0];
	// hofe -> SetOrderFace (order_fa2);
	// hofe -> SetOrderInner (order_inner[elnr][0]);
        // hofe -> SetDiscontinuous(discont); 
	//#else  // anisotropic FE
	hofe -> SetOrderFace (order_fa);
	hofe -> SetOrderInner (order_inner[elnr]);
	//hofe -> SetOrderInnerCurl (order_inner_curl[elnr]); // under construction
	hofe -> SetDiscontinuous(discont); 
	hofe -> ComputeNDof();
      }
    return *fe;
  }

  // register FESpaces
  namespace
#ifdef MACOS
  hdivhofespace_a_cpp
#endif
  {

    class Init
    {
    public:
      Init ();
    };

    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("hdivho_a", HDivHighOrderFESpaceA::Create);
    }

    Init init;
  }

}
