#ifndef HDIVSURFACE
#define HDIVSURFACE


class HDivSurfaceHighOrderFESpace : public FESpace
{
private:
  
  // Number of Edges
  int ned;
  // Number of Elements
  int nel;
  // Number of Vertex
  int nv;

  ARRAY<int> first_edge_dof;
  ARRAY<int> first_inner_dof;



  ARRAY<int> order_edge;
  ARRAY<int> order_inner;

  int ndof;

public:

  HDivSurfaceHighOrderFESpace (const MeshAccess & ama, const Flags & flags);
  ///
  virtual ~HDivSurfaceHighOrderFESpace ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const
  {
    return "HDivSurfaceHighOrderFESpace";
  }  

  void  Update();

  const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
 
  const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
 

  int GetNDof () const;

  void GetSDofNrs (int elnr, ARRAY<int> & dnums) const;

  void GetDofNrs (int elnr, ARRAY<int> & dnums) const;

};


class HDivSurfaceFiniteElement : public FiniteElement
{
protected:
  int  vnums[12];
  int ninner;

  int order;
  int order_inner;
  int order_edge[12];

public:
  HDivSurfaceFiniteElement (ELEMENT_TYPE etype);
  virtual ~HDivSurfaceFiniteElement();

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const    = 0;


  void SetVertexNumbers (FlatArray<int> & avnums);
  void SetOrderEdge(FlatArray<int> & oe);
  void SetOrderFace (FlatArray<int> & of);
  void SetOrderInner (int oi);

  virtual void ComputeNDof () = 0;

};


class HDivSurfaceTrig : public HDivSurfaceFiniteElement
{
public:
  HDivSurfaceTrig (int aorder);


  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const   ;
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;


  virtual void ComputeNDof ();
};

class HDivSurfaceQuad : public HDivSurfaceFiniteElement
{
public:
  HDivSurfaceQuad (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const   ;

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void ComputeNDof ();
};







class DiffOpIdHDivSurface : public DiffOp<DiffOpIdHDivSurface>
{
public:
  enum { D = 3};
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D-1 };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 0 };

  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSurfaceFiniteElement & fel = 
      dynamic_cast<const HDivSurfaceFiniteElement&> (bfel);
    
    int nd = fel.GetNDof();

    FlatMatrixFixWidth<3> shape(nd, lh);
    fel.CalcShape (sip.IP(), shape);

     double det =  (sip.GetJacobiDet());
     double inv_det = 1.0 / det;

       mat =  (1.0/det) * (sip.GetJacobian() * Trans(shape));
  }
  /*
  template <typename FEL, typename SIP, class TVX, class TVY>
  static void Apply (const FEL & fel, const SIP & sip,
		     const TVX & x, TVY & y,
		     LocalHeap & lh)
  {

    y = (1.0/sip.GetJacobiDet())*(Trans (fel.GetShape (sip.IP(), lh)) * x);

  }

  template <typename FEL, typename SIP, class TVX, class TVY>
  static void ApplyTrans (const FEL & fel, const SIP & sip,
			  const TVX & x, TVY & y,
			  LocalHeap & lh)
  {
    y = fel.GetShape (sip.IP(), lh)*((1.0/sip.GetJacobiDet())* x);
   
  }
  */
};


class NeumannHDivSIntegrator
  : public T_BIntegrator<DiffOpIdHDivSurface, DVec<1>, HDivSurfaceFiniteElement>
{
public:
  ///
  NeumannHDivSIntegrator (CoefficientFunction * coeff)
    : T_BIntegrator<DiffOpIdHDivSurface, DVec<1>, HDivSurfaceFiniteElement> (DVec<1> (coeff))
  { ; }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new NeumannHDivSIntegrator (coeffs[0]);
  }

  ///
  virtual bool BoundaryForm () const { return 1; }
  ///
  virtual string Name () const { return "NeumannHDivSurface"; }
};


/// integrator for $\int_\Gamma \sigma_n \tau_n \, ds$

class RobinHDivSIntegrator
  : public T_BDBIntegrator<DiffOpIdHDivSurface, DiagDMat<1>, HDivSurfaceFiniteElement >
{
public:
  RobinHDivSIntegrator (CoefficientFunction * coeff)
    : T_BDBIntegrator<DiffOpIdHDivSurface, DiagDMat<1>, HDivSurfaceFiniteElement > (DiagDMat<1> (coeff))
  { ; }


  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new RobinHDivSIntegrator (coeffs[0]);

  }

  virtual bool BoundaryForm () const { return 1; }
  virtual string Name () const { return "RobinHDivSurface"; }
};
#endif
