#ifndef FILE_HDIVSYMHOFESPACE
#define FILE_HDIVSYMHOFESPACE


class HDivSymHighOrderFESpace : public FESpace
{
private:
  int nfa;
  int nel;
//   int order_facet;
  int orderz;

  int uniform_order_facet, uniform_order_inner;
  ARRAY<INT<2> > order_facet;
  ARRAY<INT<3> > order_inner;

  ARRAY<int> first_face_dof;
  ARRAY<int> first_element_dof;
  int ndof;
  ARRAY<int> ndof_level;
  int noncontinuous;

  ARRAY<bool> fine_face; 
  bool complex_element;

public:

  HDivSymHighOrderFESpace (const MeshAccess & ama, const Flags & aflags,  bool parseflags=false);

  virtual ~HDivSymHighOrderFESpace ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    return new HDivSymHighOrderFESpace (ma, flags);
  }

  virtual string GetClassName () const
  {
    return "HDivSymHighOrderFESpace";
  }


  virtual void Update();

  virtual int GetNDof () const
  {
    return ndof;
  }

  virtual int GetNDofLevel (int level) const
  {
    return ndof_level[level];
  }

  int GetFirstFaceDof (const int f) const { return first_face_dof[f]; }
  int GetFirstCellDof (const int c) const { return first_element_dof[c]; }


  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;

  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetInternalDofs (int elnr, ARRAY<int> & dnums ) const;

  virtual void GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs (elnr, dnums); }
   virtual const ngmg::Prolongation * GetProlongation () const
   { return 0; }

#ifdef PARALLEL
   virtual void UpdateParallelDofs_hoproc();
   virtual void UpdateParallelDofs_loproc();
#endif

};






#endif
