#include <solve.hpp>
#include <comp.hpp>
#include <fem.hpp>

using namespace ngsolve;
using namespace ngcomp;
using namespace ngfem;

 #ifdef PARALLEL
 #include <parallelngs.hpp>
 using namespace ngparallel;
 #endif


#include "generalcontact.hpp"
#include "hybridcontact.hpp"

using namespace contact;


//  Vec<3> Cross ( const Vec<3> & veca, const  Vec<3> & vecb)
//  {
//    Vec<3> vecc;
//    vecc(0) = veca(1)*vecb(2) - veca(2)*vecb(1);
//    vecc(1) = veca(2)*vecb(0) - veca(0)*vecb(2);
//    vecc(2) = veca(0)*vecb(1) - veca(1)*vecb(0);
//  }


  double NumProcContHybrid :: TangL2Product( const MeshAccess & ma, const ContactInfo & cinfo,
					     const Vector<Vec<3> > & veca, const Vector<Vec<3> > & vecb)
  {
    double abtau = 0;
    int ndof_g = cinfo.GetNDof();
    ARRAY<int> vnums(4);
    Vec<3> point0, point1, point2, normal;
    int nsel = ma.GetNSE();

    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);


    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		normal = Cross( point1 - point0, point2 - point0);
		double det = L2Norm(normal);
		normal = 1.0 / L2Norm(normal) * normal;
		
		
		Vec<3> vecai = ir[l](0) * veca(vnums[0]) + 
		  ir[l](1) * veca(vnums[1]) + 
		  (1 - ir[l](0) - ir[l](1)) * veca(vnums[2]);  
		Vec<3> vecbi = ir[l](0) * vecb(vnums[0]) + 
		  ir[l](1) * vecb(vnums[1]) + 
		  (1 - ir[l](0) - ir[l](1)) * vecb(vnums[2]);  
		double vecan = InnerProduct(normal, vecai);
		double vecbn = InnerProduct(normal, vecbi);

		vecai -= vecan * normal;
		vecbi -= vecbn * normal;

		abtau += InnerProduct(vecai, vecbi);// *  ir[l].Weight();// * det;

	      }
	  }
      }




    return abtau;
  }


  double NumProcContHybrid :: NormalL2Product( const MeshAccess & ma, const ContactInfo & cinfo,
					       const Vector<double> & veca, const Vector<double> & vecb)
  {
    double abtau = 0;
    int ndof_g = cinfo.GetNDof();
    ARRAY<int> vnums(4);
    Vec<3> point0, point1, point2, normal;
    int nsel = ma.GetNSE();

    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);


    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		normal = Cross( point1 - point0, point2 - point0);
		double det = L2Norm(normal);
		normal = 1.0 / L2Norm(normal) * normal;
		
		
		double vecai = ir[l](0) * veca(vnums[0]) + 
		  ir[l](1) * veca(vnums[1]) + 
		  (1 - ir[l](0) - ir[l](1)) * veca(vnums[2]);  
		double vecbi = ir[l](0) * vecb(vnums[0]) + 
		  ir[l](1) * vecb(vnums[1]) + 
		  (1 - ir[l](0) - ir[l](1)) * vecb(vnums[2]);  

		abtau += vecai * vecbi *  ir[l].Weight() * det;

	      }
	  }
      }




    return abtau;
  }




  void NumProcContHybrid :: Assembleg ( const ContactInfo & cinfo, 
					Vector<double > & vecg)
  {
    const MeshAccess & ma =  bfa -> GetFESpace(). GetMeshAccess();
    const CompoundFESpace  &   compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    int nsel = ma.GetNSE();
    int ndof_g = cinfo.GetNDof();

    vecg.SetSize(ndof_g);
    Vector<double> gap(ndof_g);
    gap = 10.0;


    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);

    int fac =-1;
    int domain1, domain2, dummy;
    for(int vert=0; vert<ndof_g; vert++)
      {
	int sel = cinfo.GetMinSEl(vert);
	if(sel < nsel)
	  if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	    {
	      ma.GetSElVertices (sel, vnums);
	      ma.GetPoint(vnums[0], point0);
	      ma.GetPoint(vnums[1], point1);
	      ma.GetPoint(vnums[2], point2);
	      
	      n = Cross(point1 - point0, point2 - point0);
	      n = 1.0 / L2Norm(n) * n;
	      ma.GetSElNeighbouringDomains(sel, domain1, dummy);
	      ma.GetPoint(vert, pointx);
	      int sel2 = cinfo.GetMatchingSEl(vert);
	      if(sel2 >= 0 && sel2 < nsel)
		{
		  ma.GetSElVertices (sel2, vnums2);
		  ma.GetPoint(vnums2[0], p20);
		  ma.GetPoint(vnums2[1], p21);
		  ma.GetPoint(vnums2[2], p22);

		    
		  CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		  if(abs(t) > 0)gap(vert) = min2( gap(vert), t );
		    
		}
	    }
      }

    *testout << "GAP " << gap << endl;
    // calculate vecg = int_gammac g * phi_lambda ds
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);

    vecg = 0.0;
    for(int sel =0; sel < nsel; sel ++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		n = Cross( point1 - point0, point2 - point0);
		double area = L2Norm(n);
		n = 1.0 / L2Norm(n) * n;

		lambda(0) = ir[l](0);
		lambda(1) = ir[l](1);
		lambda(2) = 1 - lambda(0) - lambda(1);

		double gvalue = 0;
		for(int i=0; i<3; i++)
		  gvalue += lambda(i) * gap(vnums[i]);
		for(int i=0; i<3; i++)
		  {
		    vecg(vnums[i]) += gvalue * lambda(i) * ir[l].Weight() * area;
		  }
	      }
	  }

      }

  }




  void NumProcContHybrid :: ApplyCTranspose(  const ContactInfo & cinfo, 
					      const HDivSurfaceHighOrderFESpace & lambdatspace,
					      const Vector<double > & veclambdan, const Vector<double> & veclambda, BaseVector & rhs,
					      LocalHeap & lh)
  {
    const CompoundFESpace  &   compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*compound_fespace[1]);
    const FESpaceHybrid & pspace = dynamic_cast<const FESpaceHybrid & > (*compound_fespace[2]);

    const MeshAccess & ma = dynamic_cast<const MeshAccess & > (uspace.GetMeshAccess() );
   
    int nd_sigma_tot = compound_fespace[0]->GetNDof();
    int nd_u_tot = compound_fespace[1]->GetNDof();
    int nsel = ma.GetNSE();
    rhs.FVDouble() = 0.0;
    void * heapp = lh.GetPointer();

    
    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);


    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> point0s, point1s, point2s;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3), vnumss(3);

    // int u_tau lambda ds + int p_n lambda_n
    
    // integration on master boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    lh.CleanUp();
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		n = Cross( point1 - point0, point2 - point0);
		double astriddet = L2Norm(n);
		n = 1.0 / L2Norm(n) * n;
		
		
		const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
		
		const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);

		const HybridNormalFiniteElement & fel_p = dynamic_cast< const HybridNormalFiniteElement & > (fel[2]);
		
		const HDivSurfaceFiniteElement & fel_lambdat = dynamic_cast< const HDivSurfaceFiniteElement & > (lambdatspace.GetSFE(sel, lh));
		
		ARRAY<int> dofsu (fel_u.GetNDof());
		ARRAY<int> dofsp (fel_p.GetNDof());
		ARRAY<int> dofslambdat (fel_lambdat.GetNDof());

		ElementTransformation eltrans;
		ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		
		uspace.GetSDofNrs (sel, dofsu);
		pspace.GetSDofNrs (sel, dofsp);
		lambdatspace.GetSDofNrs (sel, dofslambdat);

		FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		FlatMatrixFixHeight<1> p(fel_p.GetNDof(), lh);
		FlatMatrixFixHeight<3> lambdat(fel_lambdat.GetNDof(), lh);

		SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );
		
		double det = fabs (sip.GetJacobiDet());

		DiffOpIdBoundaryEdge<3> :: GenerateMatrix (fel_u, sip, u, lh);

		DiffOpBoundIdHybrid<3> :: GenerateMatrix (fel_p, sip, p, lh);
		
		DiffOpIdHDivSurface :: GenerateMatrix (fel_lambdat, sip, lambdat, lh);

		Vec<3> lambdatau = 0;
		for(int i=0; i<fel_lambdat.GetNDof(); i++)
		  for(int j=0; j<3; j++)
		    {
		      lambdatau(j) += veclambda(dofslambdat[i]) * lambdat(j,i);
		    }

		double lambdan = ir[l](0) * veclambdan(vnums[0]) + 
		  ir[l](1) * veclambdan(vnums[1]) + 
		  (1 - ir[l](0) - ir[l](1)) * veclambdan(vnums[2]);
		for(int i=0; i<fel_u.GetNDof(); i++)
		  {
		    rhs.FVDouble()(dofsu[i] + nd_sigma_tot) -= (u(0,i)*lambdatau(0) + u(1,i)*lambdatau(1) + u(2,i)*lambdatau(2)) 
		      * det * ir[l].Weight();
		  }
		for(int i=0; i<fel_p.GetNDof(); i++)
		  {
		    rhs.FVDouble()(dofsp[i] + nd_sigma_tot + nd_u_tot) += p(0,i) * lambdan * det * ir[l].Weight();
		  }
	      }
	  }
      }

   
    // integration on slave boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	  {
	    lh.CleanUp();
	    for(int l=0; l<ir.GetNIP(); l++)
	      {

		int sels = cinfo.GetMatchingIP(sel,l);
		if(sels >= 0 && sels < nsel)
		  {
		    // points on slave boundary
		    ma.GetSElVertices (sel, vnums);
		    ma.GetPoint(vnums[0], point0);
		    ma.GetPoint(vnums[1], point1);
		    ma.GetPoint(vnums[2], point2);
		    n = Cross( point1 - point0, point2 - point0);
		    double astriddet = L2Norm(n);
		    n = 1.0 / L2Norm(n) * n;
		
		    pointx = ir[l](0) * point0 + ir[l](1) * point1 + (1 - ir[l](0) - ir[l](1)) * point2;

		    // slave element
		    const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sels, lh));
		    
		    const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
		
		    const HybridNormalFiniteElement & fel_p = dynamic_cast< const HybridNormalFiniteElement & > (fel[2]);

		    const HDivSurfaceFiniteElement & fel_lambdat = 
		      dynamic_cast< const HDivSurfaceFiniteElement & > (lambdatspace.GetSFE(sel, lh));

		    ARRAY<int> dofsu(fel_u.GetNDof());
		    ARRAY<int> dofsp (fel_p.GetNDof());
		    ARRAY<int> dofslambdat (fel_lambdat.GetNDof());

		    // eltrans on master, eltranss for slave
		    ElementTransformation eltrans;
		    ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		    ElementTransformation eltranss;
		    ma.GetSurfaceElementTransformation ( sels, eltranss, lh);
		    
		    uspace.GetSDofNrs (sels, dofsu);
		    pspace.GetSDofNrs (sels, dofsp);
		    lambdatspace.GetSDofNrs (sel, dofslambdat);

		    FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		    FlatMatrixFixHeight<1> p(fel_p.GetNDof(), lh);
		    FlatMatrixFixHeight<3> lambdat(fel_lambdat.GetNDof(), lh);

		    SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );

		    // get barycentric coordinates of point on slave boundary
		    ma.GetSElVertices(sels, vnumss);
		    ma.GetPoint(vnumss[0], point0s);
		    ma.GetPoint(vnumss[1], point1s);
		    ma.GetPoint(vnumss[2], point2s);

		    CalcProjectionN (pointx, n, point0s, point1s, point2s, lambda, t);
		    IntegrationPoint lam ( lambda(0), lambda(1) );

		    SpecificIntegrationPoint<2,3> sips( lam, eltranss, lh );
		    
		    double det = fabs (sip.GetJacobiDet());
		    
		    DiffOpIdBoundaryEdge<3> :: GenerateMatrix (fel_u, sips, u, lh);

		    DiffOpBoundIdHybrid<3> :: GenerateMatrix (fel_p, sips, p, lh);

		    DiffOpIdHDivSurface :: GenerateMatrix (fel_lambdat, sip, lambdat, lh);

		    Vec<3> lambdatau = 0;
		    for(int i=0; i<fel_lambdat.GetNDof(); i++)
		      for(int j=0; j<3; j++)
			{
			  lambdatau(j) += veclambda(dofslambdat[i]) * lambdat(j,i);
			}

		    double lambdan = ir[l](0) * veclambdan(vnums[0]) + 
		      ir[l](1) * veclambdan(vnums[1]) + 
		      (1 - ir[l](0) - ir[l](1)) * veclambdan(vnums[2]);

		    for(int i=0; i<fel_u.GetNDof(); i++)
		      {
			rhs.FVDouble()(dofsu[i] + nd_sigma_tot) += 
			  (u(0,i)*lambdatau(0) + u(1,i)*lambdatau(1) + u(2,i)*lambdatau(2)) 
			  * det * ir[l].Weight();
		      }
		    for(int i=0; i<fel_p.GetNDof(); i++)
		      {
			rhs.FVDouble()(dofsp[i] + nd_sigma_tot + nd_u_tot) +=  
			   lambdan * p(0,i) * det * ir[l].Weight();

		      }
		  }

	      }
	  }
      }
    
    lh.CleanUp(heapp);
  }



  void NumProcContHybrid :: ApplyC(  const ContactInfo & cinfo, 
				     const HDivSurfaceHighOrderFESpace & lambdatspace,
				     const BaseVector & rhs, Vector<double> & veclambdan, Vector<double> & veclambda, LocalHeap & lh)
  {
    const CompoundFESpace  &   compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const HCurlHighOrderFESpaceA & uspace = dynamic_cast<const HCurlHighOrderFESpaceA & > (*compound_fespace[1]);
    const FESpaceHybrid & pspace = dynamic_cast<const FESpaceHybrid & > (*compound_fespace[2]);

    const MeshAccess & ma = dynamic_cast<const MeshAccess & > (uspace.GetMeshAccess() );

    int nd_sigma_tot = compound_fespace[0]->GetNDof();
    int nd_u_tot = compound_fespace[1]-> GetNDof();
    int nsel = ma.GetNSE();
    veclambda = 0.0;
    veclambdan = 0.0;
    void * heapp = lh.GetPointer();

    const IntegrationRule & ir = 
      GetIntegrationRules().SelectIntegrationRule(ET_TRIG, 8);

 


    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> point0s, point1s, point2s;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnumss(3);
   

    // integration on master boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if( cinfo.IsMasterBound ( ma.GetSElIndex(sel) ) )
	  {

	    lh.CleanUp();
	    for(int l=0; l<ir.GetNIP(); l++)
	      {
		ma.GetSElVertices (sel, vnums);
		ma.GetPoint(vnums[0], point0);
		ma.GetPoint(vnums[1], point1);
		ma.GetPoint(vnums[2], point2);
		n = Cross( point1 - point0, point2 - point0);
		double astriddet = L2Norm(n);
		n = 1.0 / L2Norm(n) * n;


		const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
	    
		const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);
	    
		const HybridNormalFiniteElement & fel_p = dynamic_cast< const HybridNormalFiniteElement & > (fel[2]);

		const HDivSurfaceFiniteElement & fel_lambdat = dynamic_cast < const HDivSurfaceFiniteElement & > (lambdatspace.GetSFE(sel, lh));

		ARRAY<int> dofsu(fel_u.GetNDof());
		ARRAY<int> dofsp(fel_p.GetNDof());
		ARRAY<int> dofslambdat (fel_lambdat.GetNDof());

		ElementTransformation eltrans;
		ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
		uspace.GetSDofNrs (sel, dofsu);
		pspace.GetSDofNrs (sel, dofsp);
		lambdatspace.GetSDofNrs (sel, dofslambdat);

		//Vector<double> sigma(fel_sigma.GetNDof());
		FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		FlatMatrixFixHeight<1> p(fel_p.GetNDof(), lh);
		FlatMatrixFixHeight<3> lambdat(fel_lambdat.GetNDof(), lh);

		SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );

		double det = fabs (sip.GetJacobiDet());
	    
		DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sip, u, lh);

		DiffOpBoundIdHybrid<3> :: GenerateMatrix (fel_p, sip, p, lh);

		DiffOpIdHDivSurface :: GenerateMatrix (fel_lambdat, sip, lambdat, lh);

		Vec<3> uvalue = 0.0;
		double pvalue = 0.0;

		for(int i=0; i<fel_u.GetNDof(); i++)
		  {
		    uvalue(0) += u(0,i) * rhs.FVDouble()(dofsu[i] + nd_sigma_tot);
		    uvalue(1) += u(1,i) * rhs.FVDouble()(dofsu[i] + nd_sigma_tot);
		    uvalue(2) += u(2,i) * rhs.FVDouble()(dofsu[i] + nd_sigma_tot);
		  }
		for(int i=0; i<fel_p.GetNDof(); i++)
		  {
		    pvalue += p(0,i) * rhs.FVDouble()(dofsp[i] + nd_sigma_tot + nd_u_tot);
		  }

		for(int i=0; i<fel_lambdat.GetNDof(); i++)
		  for(int j=0; j<3; j++)
		    {
		      veclambda(dofslambdat[i]) -= lambdat(j,i) * uvalue(j) * det * ir[l].Weight();
		    }
		veclambdan(vnums[2]) += (1 - ir[l](0) - ir[l](1)) * det  * ir[l].Weight() * pvalue;
		veclambdan(vnums[0]) += (ir[l](0)) * det * ir[l].Weight() * pvalue;
		veclambdan(vnums[1]) += (ir[l](1)) * det * ir[l].Weight() * pvalue;
	      }
	  }
      }

    // integration on slave boundary
    for(int sel=0; sel<nsel; sel++)
      {
	if( cinfo.IsMasterBound ( ma.GetSElIndex(sel) ) )
	  for(int l=0; l<ir.GetNIP(); l++)
	    {

	      lh.CleanUp();
	      int sels = cinfo.GetMatchingIP(sel,l);
	      if(sels >= 0 && sels < nsel)
		{
		  // vnums, points on master
		  ma.GetSElVertices (sel, vnums);
		  ma.GetPoint(vnums[0], point0);
		  ma.GetPoint(vnums[1], point1);
		  ma.GetPoint(vnums[2], point2);

		  n = Cross( point1 - point0, point2 - point0);
		  double astriddet = L2Norm(n);
		  n = 1.0 / L2Norm(n) * n;
	    
		  pointx = ir[l](0) * point0 + ir[l](1) * point1 + (1 - ir[l](0) - ir[l](1)) * point2;
		  // slave element
		  const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sels, lh));
	    
		  const HCurlHighOrderFiniteElementA<2> & fel_u = dynamic_cast< const HCurlHighOrderFiniteElementA<2> & > (fel[1]);

		  const HybridNormalFiniteElement & fel_p = dynamic_cast< const HybridNormalFiniteElement & > (fel[2]);
		  
		  // master element
		  const HDivSurfaceFiniteElement & fel_lambdat = dynamic_cast < const HDivSurfaceFiniteElement & > (lambdatspace.GetSFE(sel, lh));

		  ARRAY<int> dofsu(fel_u.GetNDof());
		  ARRAY<int> dofsp(fel_p.GetNDof());
		  ARRAY<int> dofslambdat(fel_lambdat.GetNDof());		

		  ElementTransformation eltrans, eltranss;
		  ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		  ma.GetSurfaceElementTransformation ( sels, eltranss, lh);

		  uspace.GetSDofNrs (sels, dofsu);
		  pspace.GetSDofNrs (sels, dofsp);
		  lambdatspace.GetSDofNrs (sel, dofslambdat);

		  //Vector<double> sigma(fel_sigma.GetNDof());
		  FlatMatrixFixHeight<3> u(fel_u.GetNDof(), lh);
		  FlatMatrixFixHeight<1> p(fel_p.GetNDof(), lh);
		  FlatMatrixFixHeight<3> lambdat(fel_lambdat.GetNDof(), lh);

		  // master element: use integration rule
		  SpecificIntegrationPoint<2,3> sip( ir[l], eltrans, lh );
		
		  double det = fabs (sip.GetJacobiDet());

		  // slave element: find barycentric coordinates of matching point

		  ma.GetSElVertices(sels, vnumss);
		  ma.GetPoint(vnumss[0], point0s);
		  ma.GetPoint(vnumss[1], point1s);
		  ma.GetPoint(vnumss[2], point2s);

		  CalcProjectionN (pointx, n, point0s, point1s, point2s, lambda, t);
		  IntegrationPoint lam ( lambda(0), lambda(1) );
		  SpecificIntegrationPoint<2,3> sips (lam, eltranss, lh);

		  DiffOpIdBoundaryEdge<3>::GenerateMatrix (fel_u, sips, u, lh);

		  DiffOpBoundIdHybrid<3> :: GenerateMatrix (fel_p, sips, p, lh);

		  DiffOpIdHDivSurface :: GenerateMatrix (fel_lambdat, sip, lambdat, lh);

		  Vec<3> uvalue = 0.0;
		  double pvalue = 0.0;

		  for(int i=0; i<fel_u.GetNDof(); i++)
		    {
		      uvalue(0) += u(0,i) * rhs.FVDouble()(dofsu[i] + nd_sigma_tot);
		      uvalue(1) += u(1,i) * rhs.FVDouble()(dofsu[i] + nd_sigma_tot);
		      uvalue(2) += u(2,i) * rhs.FVDouble()(dofsu[i] + nd_sigma_tot);
		    }
		  for(int i=0; i<fel_p.GetNDof(); i++)
		    {
		      pvalue += p(0,i) * rhs.FVDouble()(dofsp[i] + nd_sigma_tot + nd_u_tot);
		    }


		  for(int i=0; i<fel_lambdat.GetNDof(); i++)
		    for(int j=0; j<3; j++)
		      {
			veclambda(dofslambdat[i]) +=  lambdat(j,i) * uvalue(j) * det * ir[l].Weight();
		      }


		  veclambdan(vnums[2]) += (1 - ir[l](0) - ir[l](1)) * det * ir[l].Weight() * pvalue;
		  veclambdan(vnums[0]) += (ir[l](0)) * det * ir[l].Weight() * pvalue;
		  veclambdan(vnums[1]) += (ir[l](1)) * det * ir[l].Weight() * pvalue;

		}
	    }
      }


    lh.CleanUp(heapp);
  }





  bool NumProcContHybrid :: ProjectStress ( const ContactInfo & cinfo, const HDivSurfaceHighOrderFESpace & lambdatspace,
					    const double coeff_f, 
					    const Vector<double> & vecsigmann, Vector<double > & veclambdan, Vector<double> & veclambda,
					    LocalHeap & lh)
  {

    const MeshAccess & ma =  bfa -> GetFESpace(). GetMeshAccess();

    bool flag = false;
    int ndof_c = veclambdan.Size();
    int nsel = ma.GetNSE();
    double t;

    void * heapp = lh.GetPointer();


    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> point0m, point1m, point2m;
    Vec<3> lambda;
    ARRAY<int> vnums(3), vnumsm(3);

    
    for(int vert=0; vert<ndof_c; vert++)
      {
       	if(veclambdan(vert) < 0)
	  {
	    veclambdan(vert) = 0;
	    flag = true;
	  }
	/*
	  double normlambda = L2Norm(veclambda(vert));

	  if(normlambda > abs(coeff_f * vecsigmann(vert)) )
	  {
	  double t = abs(coeff_f * vecsigmann(vert)) / normlambda;
	  veclambda(vert) = t * veclambda(vert);
	  flag = true;
	  }
	*/
      }

    if (coeff_f == 0) return flag;

    for(int sel =0; sel < nsel; sel++)
      if( cinfo.IsMasterBound( ma.GetSElIndex(sel)) )
	{
	  ma.GetSElVertices (sel, vnums);
	  ma.GetPoint(vnums[0], point0);
	  ma.GetPoint(vnums[1], point1);
	  ma.GetPoint(vnums[2], point2);
	  n = Cross( point1 - point0, point2 - point0);
	  double astriddet = L2Norm(n);
	  n = 1.0 / L2Norm(n) * n;

	  const HDivSurfaceFiniteElement & fel_lambdat = dynamic_cast < const HDivSurfaceFiniteElement & > (lambdatspace.GetSFE(sel, lh));
	
	  ARRAY<int> dofslambdat (fel_lambdat.GetNDof());
	    
	  ElementTransformation eltrans;
	  ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
	  lambdatspace.GetSDofNrs (sel, dofslambdat);
	
	  FlatMatrixFixHeight<3> lambdat(fel_lambdat.GetNDof(), lh);

	  IntegrationPoint ip(1.0 / 3.0, 1.0 / 3.0);

	  SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );

	  double det = fabs (sip.GetJacobiDet());
	    
	  DiffOpIdHDivSurface :: GenerateMatrix (fel_lambdat, sip, lambdat, lh);
	  double fsigmann =  abs ( vecsigmann(vnums[0]) + vecsigmann(vnums[1]) + vecsigmann(vnums[2]) ) / 3.0;
	  fsigmann *= coeff_f;
	
	  Vec<3> lambdatvalue = 0.0;
	  for(int i=0; i<fel_lambdat.GetNDof(); i++)
	    for(int j=0; j<3; j++)
	      {
		lambdatvalue(j) += veclambda(dofslambdat[i]) * lambdat(j,i);
	      }
	  double normlambda = L2Norm(lambdatvalue);
	  if(normlambda > fsigmann )
	    {
	      t = fsigmann / normlambda;
	      for(int i=0; i<fel_lambdat.GetNDof(); i++)
		veclambda(dofslambdat[i]) *= t;
	      flag = true;
	      *testout << "old sigma " << lambdatvalue << ",  projected " << t << endl;
	    }
	
	}
    lh.CleanUp(heapp);
    return flag;
  }
    




  void NumProcContHybrid :: CalcURel ( const ContactInfo & cinfo, 
				       const BaseVector & vecu, Vector<double > & vecurel, LocalHeap & lh)
  {
    const CompoundFESpace  &   compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const FESpaceHybrid & pspace = dynamic_cast<const FESpaceHybrid & > (*compound_fespace[2]);
  
    const MeshAccess & ma = dynamic_cast<const MeshAccess & > (sigmaspace.GetMeshAccess() );
  
    void * heapp = lh.GetPointer();

    int ndof_c = cinfo.GetNDof();
    int nsel = ma.GetNSE();
    Vec<3> point0, point1, point2, pointx, n, n2;
    Vec<3> p20, p21, p22;
    Vec<3> lambda;
    double t;
    ARRAY<int> vnums(3), vnums2(3);
    vecurel = 0.0;

    int nd_sigma_tot = compound_fespace[0]->GetNDof();
    int nd_u_tot = compound_fespace[1]-> GetNDof();
  
  


    for(int vert=0; vert<ndof_c; vert++)
      {
	int sel = cinfo.GetMinSEl(vert);
	if(sel < nsel)
	  if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
	    {
	      // vertices of surface element
	      ma.GetSElVertices (sel, vnums);
	      ma.GetPoint(vnums[0], point0);
	      ma.GetPoint(vnums[1], point1);
	      ma.GetPoint(vnums[2], point2);
	      n = Cross(point1 - point0, point2 - point0);
	      n = 1.0 / L2Norm(n) * n;
	      // local coords of vertex vert
	      ma.GetPoint(vert, pointx);
	      CalcProjectionN (pointx, n, point0, point1, point2, lambda, t);

	      const CompoundFiniteElement & fel = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
	    	    
	      const HybridNormalFiniteElement & fel_p = dynamic_cast< const HybridNormalFiniteElement & > (fel[2]);
	    
	      ARRAY<int> dofsp(fel_p.GetNDof());
	    
	    
	      ElementTransformation eltrans;
	      ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
	    
	      pspace.GetSDofNrs (sel, dofsp);
	    
	      FlatMatrixFixHeight<1> p(fel_p.GetNDof(), lh);
	    
	      IntegrationPoint ip(lambda(0), lambda(1));
	      SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );
	    
	      double det = fabs (sip.GetJacobiDet());
	    	    
	      DiffOpBoundIdHybrid<3> :: GenerateMatrix (fel_p, sip, p, lh);
	    
	      for(int i=0; i<fel_p.GetNDof(); i++)
		{
		  vecurel(vert) += vecu.FVDouble()(nd_sigma_tot + nd_u_tot + dofsp[i]) * p(i);
		}
	    
	      int sel2 = cinfo.GetMatchingSEl(vert);
	    
	      if(sel2 >= 0 && sel2 < nsel)
		{
		  ma.GetSElVertices (sel2, vnums2);
		  ma.GetPoint(vnums2[0], p20);
		  ma.GetPoint(vnums2[1], p21);
		  ma.GetPoint(vnums2[2], p22);
		
		  CalcProjectionN (pointx, n, p20, p21, p22, lambda, t);
		
		  const CompoundFiniteElement & fel2 = dynamic_cast< const CompoundFiniteElement & > (compound_fespace.GetSFE(sel, lh));
				
		  const HybridNormalFiniteElement & fel_p2 = dynamic_cast< const HybridNormalFiniteElement & > (fel[2]);

		  ARRAY<int> dofsp2(fel_p2.GetNDof());
				
		  ElementTransformation eltrans2;
		  ma.GetSurfaceElementTransformation ( sel2, eltrans2, lh);
		  
		  pspace.GetSDofNrs (sel2, dofsp2);
		  
		  FlatVector<double> p2(fel_p.GetNDof(), lh);
		
		  IntegrationPoint ip2(lambda(0), lambda(1));
		  SpecificIntegrationPoint<2,3> sip2( ip2, eltrans, lh );
	      
		  double det = fabs (sip2.GetJacobiDet());

		  for(int i=0; i<fel_p2.GetNDof(); i++)
		    {
		      vecurel(vert) += vecu.FVDouble()(nd_sigma_tot + nd_u_tot + dofsp2[i]) * p2(i);
		    }	
		}
	    }
      }  
    lh.CleanUp(heapp);
  }


  void NumProcContHybrid :: FindPhiBeta(const Vector<double> & veclambdan, 
					const Vector<double> & r, Vector<double> & phi, Vector<double> & beta)
  {
  
    int nd_c = veclambdan.Size();
    for (int i=0; i<nd_c; i++)
      {
	if(veclambdan(i) <= 0)
	  {
	    phi(i) = 0.0;
	    beta(i) = max2(r(i), 0.0);
	  }
	else
	  {
	    phi(i) = r(i);
	    beta(i) = 0.0;
	  }
      }

  }

  void NumProcContHybrid :: FindPhiBetaTau( const ContactInfo & cinfo, const HDivSurfaceHighOrderFESpace & lambdatspace,
					    const double & coeff_f,
					    const Vector<double> & veclambdan, const Vector<double> & veclambda,
					    const Vector<double> & rtau, Vector<double> & phitau, Vector<double> & betatau, LocalHeap & lh)
  {

    /*
      for(int vert=0; vert<nd_c; vert++)
      {
      int sel = cinfo.GetMinSEl(vert);
      if(sel < nsel)
      if(cinfo.IsMasterBound(ma.GetSElIndex(sel)))
      {
      ma.GetSElVertices (sel, vnums);
      ma.GetPoint(vnums[0], point0);
      ma.GetPoint(vnums[1], point1);
      ma.GetPoint(vnums[2], point2);
	      
      nn = Cross(point1 - point0, point2 - point0);
      double det = L2Norm(nn);
      nn = 1.0 / L2Norm(nn) * nn;


      double normlambda = L2Norm(veclambda(vert));

      Vec<3> rtaun = rtau(vert) - InnerProduct(rtau(vert) , nn) * nn;

      if(coeff_f * veclambdan(vert) == 0)
      {
      betatau(vert) = 0.0;
      phitau(vert) = 0.0;
      }
      else if(abs(coeff_f * veclambdan(vert) ) <= normlambda)
      {
      Vec<3> normal = 1.0 / normlambda * veclambda(vert);
      double t = InnerProduct(rtaun, normal);
      betatau(vert) = rtaun - min2(t,0.0) * normal;
      phitau(vert) = 0.0;
      }
      else
      {
      betatau(vert) = 0.0;
      phitau(vert) = rtaun;
      }
      }
      }
    */
    /*
      for(int vert = 0; vert < nd_c; vert ++)
      {
      double normlambda = L2Norm(veclambda(vert));

      if(coeff_f * veclambdan(vert) == 0)
      {
      betatau(vert) = 0.0;
      phitau(vert) = 0.0;
      }
      else if(abs(coeff_f * veclambdan(vert) ) <= normlambda)
      {
      Vec<3> normal = 1.0 / normlambda * veclambda(vert);
      double t = InnerProduct(rtau(vert), normal);
      betatau(vert) = rtau(vert) - min2(t,0.0) * normal;
      phitau(vert) = 0.0;
      }
      else
      {
      betatau(vert) = 0.0;
      phitau(vert) = rtau(vert);
      }

      }
    */

    const MeshAccess & ma =  bfa -> GetFESpace(). GetMeshAccess();

    int ndof_c = veclambdan.Size();
    int nsel = ma.GetNSE();
    double t;

    void * heapp = lh.GetPointer();


    Vec<3> point0, point1, point2, pointx, n;
    Vec<3> point0m, point1m, point2m;
    Vec<3> lambda;
    ARRAY<int> vnums(3), vnumsm(3);
    BitArray iszero(veclambda.Size());
    iszero.Clear();
    
    for(int sel =0; sel < nsel; sel++)
      if( cinfo.IsMasterBound( ma.GetSElIndex( sel ) ) )
	{
	  ma.GetSElVertices (sel, vnums);
	  ma.GetPoint(vnums[0], point0);
	  ma.GetPoint(vnums[1], point1);
	  ma.GetPoint(vnums[2], point2);
	  n = Cross( point1 - point0, point2 - point0);
	  double astriddet = L2Norm(n);
	  n = 1.0 / L2Norm(n) * n;

	  const HDivSurfaceFiniteElement & fel_lambdat = dynamic_cast < const HDivSurfaceFiniteElement & > (lambdatspace.GetSFE(sel, lh));
	
	  ARRAY<int> dofslambdat (fel_lambdat.GetNDof());
	    
	  ElementTransformation eltrans;
	  ma.GetSurfaceElementTransformation ( sel, eltrans, lh);
		
	  lambdatspace.GetSDofNrs (sel, dofslambdat);
	
	  FlatMatrixFixHeight<3> lambdat(fel_lambdat.GetNDof(), lh);

	  IntegrationPoint ip(1.0 / 3.0, 1.0 / 3.0);

	  SpecificIntegrationPoint<2,3> sip( ip, eltrans, lh );

	  double det = fabs (sip.GetJacobiDet());
	    
	  DiffOpIdHDivSurface :: GenerateMatrix (fel_lambdat, sip, lambdat, lh);
	  double fsigmann =  abs ( veclambdan(vnums[0]) + veclambdan(vnums[1]) + veclambdan(vnums[2]) ) / 3.0;
	  fsigmann *= coeff_f;

	  Vec<3> lambdatvalue = 0.0, rtauvalue = 0.0;
	  for(int i=0; i<fel_lambdat.GetNDof(); i++)
	    for(int j=0; j<3; j++)
	      {
		lambdatvalue(j) += veclambda(dofslambdat[i]) * lambdat(j,i);
		rtauvalue(j) += rtau(dofslambdat[i]) * lambdat(j,i);
	      }
	  double normlambda = L2Norm(lambdatvalue);
	  if(fsigmann  == 0)
	    {
	      for(int i=0; i<fel_lambdat.GetNDof(); i++)
		{
		  phitau(dofslambdat[i]) = 0.0;
		  betatau(dofslambdat[i]) = 0.0;
		  iszero.Set(dofslambdat[i]);
		}
	    }
	  else if ( fsigmann <= normlambda)
	    {
	      for(int i=0; i<fel_lambdat.GetNDof(); i++)
		if(!iszero.Test(dofslambdat[i]) )
		  {
		    int d = dofslambdat[i];
		    t = InnerProduct( rtauvalue, lambdatvalue) / sqr(normlambda);

		    phitau(dofslambdat[i]) = 0.0;
		    betatau(d) = rtau(d) - min2(t, 0.0) * veclambda(d);
		  }

	    }
	  else
	    {
	      for(int i=0; i<fel_lambdat.GetNDof(); i++)
		if(!iszero.Test(dofslambdat[i]) )
		  {
		    phitau(dofslambdat[i]) = rtau(dofslambdat[i]);
		    betatau(dofslambdat[i]) = 0.0;
		  }
	    }

	}

    lh.CleanUp(heapp);

  }



  void NumProcContHybrid :: Do(LocalHeap & lh)
  {
    cout << "solve contact problem" << endl;
    lh.CleanUp();
    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();
    BaseVector & vecv = *vecu.CreateVector();

    const CompoundFESpace  &   compound_fespace =   dynamic_cast<const CompoundFESpace & > (bfa -> GetFESpace()) ;

    const HCurlHighOrderFESpaceA & vspace = dynamic_cast<const HCurlHighOrderFESpaceA & > ( *compound_fespace[1]);
    const HDivSymHighOrderFESpace & sigmaspace = dynamic_cast<const HDivSymHighOrderFESpace & > (*compound_fespace[0]);
    const FESpaceHybrid & pspace = dynamic_cast<const FESpaceHybrid & > (*compound_fespace[2]);

    const MeshAccess & ma =  bfa -> GetFESpace(). GetMeshAccess();

   
    int nsel = ma.GetNSE();

    ofstream outfilefreegrad("/home/astrid/netgen/ngsolve/astrid/freegrad.dat"), 
      outfilephilambda("/home/astrid/netgen/ngsolve/astrid/philambda.dat"),
      outfilechoppedgrad("/home/astrid/netgen/ngsolve/astrid/choppedgrad.dat"), 
      outfilebetalambda("/home/astrid/netgen/ngsolve/astrid/betalambda.dat");
    ofstream outfileres("/home/astrid/netgen/ngsolve/astrid/residual.dat"); 
    ofstream outfilestep("/home/astrid/netgen/ngsolve/astrid/step.dat"); 

    int ndof = compound_fespace.GetNDof();
    int nd_u = compound_fespace[1]->GetNDof();
    int nd_sigma = compound_fespace[0]->GetNDof();
    int nd_p = compound_fespace[2]->GetNDof();

    Flags lambdaflags, gflambdatflags;
    lambdaflags.SetFlag("order", vspace.GetOrder());
 
    HDivSurfaceHighOrderFESpace lambdatspace ( ma, lambdaflags);
    lambdatspace.Update();

    cout  << "dofs lambda  " << lambdatspace.GetNDof() << endl;
    cout << "order lambda  " << lambdatspace.GetOrder() << endl;
    int nd_tau = lambdatspace.GetNDof();

    Vector<double> veclambdat ( lambdatspace.GetNDof() );

    //
    int maxsteps = 3;
    //
    double prec = 1e-5;
    //
    bool print=false;
    //
    enum SOLVER { CG, GMRES, QMR, NCG };
    //
    SOLVER solver = CG;

    double h = taup;

  

    int intorder = 2;
    ARRAY<bool> contactbound(15), masterbound(15); 
    for(int i=0; i<15; i++)
      {
	contactbound[i] = 0; masterbound[i] = 0;
      }

    contactbound[cbound1] = 1;
    contactbound[cbound2] = 1;
    masterbound[mbound] = 1;

    ContactInfo cinfo (ma, compound_fespace, contactbound, masterbound, tmax);
    int nd_c = cinfo.GetNDof();
    // find gap function g
    Vector<double > vecg (ma.GetNV());
    Assembleg ( cinfo, vecg);


    // init lambda_0, w_0
    Vector<double  > veclambda(nd_tau); 
    Vector<double  > veclambdan(nd_c);
    Vector<double > vecsigmann(nd_c), r(nd_c), phi(nd_c), beta(nd_c), p(nd_c), sp(nd_c), vech(nd_c);
    Vector<double> rtau(nd_tau), phitau(nd_tau), betatau(nd_tau), ptau(nd_tau), sptau(nd_tau), vechtau(nd_tau);


    veclambdan = 0.0;
    veclambda = 0.0;
    vechtau = 0.0;
    vech = 0.0;
    vecsigmann = 0.0;
    ARRAY<int> vnums(3);

    // solve by Uzawa type method
    BaseVector & rhs = *vecf.CreateVector();

    const BaseMatrix * premat = NULL;
    if (pre)  premat = &(pre->GetMatrix()); 
    KrylovSpaceSolver * invmat;
    
    switch (solver)
      {
      case CG:
	cout << "cg solve for real problem" << endl;
	invmat = new CGSolver<double>(mata, *premat);
	break;
      case QMR:
	invmat = new QMRSolver<double>(mata, *premat);
	break;
      case GMRES:
	invmat = new GMRESSolver<double>(mata, *premat);
	break;
      }
    
    invmat->SetMaxSteps (maxsteps);
    invmat->SetPrecision (prec);
    invmat->SetPrintRates ();
    invmat->SetInitialize (0);
    invmat->SetStatusHandler(ma);

    //     clock_t starttime, endtime, soltime;
    //     starttime = clock();

    int uzawasteps = 0;
    double elambda = 1.0, ew = 1.0;
    double eps = 1e-9;
    double epsw=0, epslambda =0;
   
    
    
    
    ApplyCTranspose( cinfo, lambdatspace, veclambdan, veclambda, vecv, lh);
    
    invmat->Mult(vecv, vecu);
    ApplyC( cinfo, lambdatspace, vecu, r, rtau, lh);
    // vecw = 0, also r = -f
    
    invmat->Mult( vecf, vecv);

    ApplyC( cinfo, lambdatspace, vecv, vech, vechtau, lh);
    vech -= vecg;
    
    r = vech - r;
    rtau = vechtau - rtau;
    
    FindPhiBeta(veclambdan, r, phi, beta);
    FindPhiBetaTau( cinfo, lambdatspace, coeff_f, veclambdan, veclambda, rtau, phitau, betatau, lh);
    p = phi;
    ptau = phitau;
    int step = 0;
    double l2phi, l2beta, alpha_cg = -10;
    
    l2phi = sqrt( InnerProduct(phi, phi) + InnerProduct( phitau, phitau) );
    l2beta = sqrt( InnerProduct(beta, beta) + InnerProduct( betatau, betatau) );
    
    cout << L2Norm(phi) << "    " << L2Norm(phitau) << "    " 
	 << L2Norm(beta) << "   " << L2Norm(betatau) << endl;
    
    while( l2phi + l2beta > 1e-10 && step < numuzawasteps  &&  abs(alpha_cg) > 1e-10 )
    {
//       if(step % 10 == 0 && step < 100)
// 	{
// 	  vecsigmann = veclambdan;
// 	  FindPhiBeta(veclambdan, r, phi, beta);
// 	  FindPhiBetaTau(cinfo, lambdatspace, coeff_f, vecsigmann, veclambda, rtau, phitau, betatau, lh);
// 	  p = phi; ptau = phitau;
// 	}

// achtung -- NO FRICTION
      phitau = 0; betatau = 0; veclambda = 0;

      l2phi = sqrt( InnerProduct(phi, phi) + h * h * InnerProduct( phitau, phitau) );
      l2beta = sqrt( InnerProduct(beta, beta) + h * h * InnerProduct( betatau, betatau) );
      if(l2phi > l2beta)// && alpha_cg )//* L2Norm(p) > 1e-17)
	{
	  cout << "cg step " << endl;
	  outfilestep << step << "  cg step: ";
	  bool flag = false;
	  ApplyCTranspose( cinfo, lambdatspace, p, ptau, vecv, lh);
	  invmat->Mult(vecv, vecv);
	  ApplyC( cinfo,  lambdatspace, vecv, sp, sptau, lh);
	  
	  
	  double pspn = InnerProduct( sp,p );
	  double psptau = InnerProduct (sptau, ptau);
	  double psp = pspn + h * h * psptau;
	  alpha_cg = ( InnerProduct( r,p ) + h * h * InnerProduct( rtau, ptau) ) / psp;
	  (*testout) << "alpha_cg = " << alpha_cg << "  n " << InnerProduct( r,p ) 
		     << "   tang " << InnerProduct( rtau, ptau) << endl
		     << "p S p " << pspn << "    ptau S ptau " << psptau << endl;;
	  
	  cout << "alpha_cg " << alpha_cg << endl;

	  veclambdan += alpha_cg * p;
	  veclambda += h * alpha_cg * ptau;
	  
	  
	  double alpha2 =  100;
	  
	  
	  flag = 
	    ProjectStress(cinfo,  lambdatspace,  coeff_f, vecsigmann, veclambdan, veclambda, lh);

	  if (flag)
	    {
	      cout << "projected cg step " << endl;
	      outfilestep << "  projected" << endl;
	      ApplyCTranspose( cinfo,  lambdatspace, veclambdan, veclambda, vecv, lh);
	      invmat->Mult(vecv, vecu);
	      ApplyC( cinfo, lambdatspace, vecu, r, rtau, lh);
	      r = vech - r;
	      rtau = vechtau - rtau;
	      FindPhiBeta(veclambdan, r, phi, beta);
	      FindPhiBetaTau( cinfo, lambdatspace, coeff_f, vecsigmann, veclambda, rtau, phitau, betatau, lh);
	      p = phi;
	      ptau = phitau;
	      
	      veclambdan += taun * phi;
	      veclambda += tautau * phitau;
	      
	      flag = ProjectStress (cinfo, lambdatspace, coeff_f, vecsigmann, veclambdan, veclambda, lh);
	      ApplyCTranspose(  cinfo,  lambdatspace, veclambdan, veclambda, vecv, lh);
	      invmat->Mult(vecv, vecv);
	      
	      ApplyC( cinfo,  lambdatspace, vecv, r, rtau, lh);
	      r = vech - r;
	      rtau = vechtau - rtau;
	
	      
	      FindPhiBeta(veclambdan, r, phi, beta);
	      FindPhiBetaTau( cinfo, lambdatspace, coeff_f, vecsigmann, veclambda, rtau, phitau, betatau, lh);
	      
	      p = phi;	  
	      ptau = phitau;
	    }
	  else
	    {
	      cout << " cg cg step" << endl;
	      outfilestep << "  cg" << endl;
	      
	      r -= alpha_cg * sp;
	      rtau -= h * alpha_cg * sptau;
	      
	      FindPhiBeta(veclambdan, r, phi, beta);
	      FindPhiBetaTau( cinfo, lambdatspace, coeff_f, vecsigmann, veclambda, rtau, phitau, betatau, lh);
	      
	      double gamma = (InnerProduct(phi,sp) + h * h * InnerProduct( phitau, sptau)) / psp;
	      p = phi - gamma * p;
	      ptau = phitau + h * gamma * ptau;
	    }
	}
      else
	{
	  cout << "projection step " << endl;
	  outfilestep << step << "  projection step " << endl;
	  
	  veclambdan += taun * beta;
	  veclambda += tautau * betatau;
	  
	  bool flag = ProjectStress (cinfo, lambdatspace, coeff_f, vecsigmann, veclambdan, veclambda, lh);
	  ApplyCTranspose(  cinfo,  lambdatspace, veclambdan, veclambda, vecv, lh);
	  invmat->Mult(vecv, vecv);
	  
	  ApplyC( cinfo,  lambdatspace, vecv, r, rtau, lh);
	  r = vech - r;
	  rtau = vechtau - rtau;
	  
	  
	  FindPhiBeta(veclambdan, r, phi, beta);
	  FindPhiBetaTau( cinfo, lambdatspace, coeff_f, vecsigmann, veclambda, rtau, phitau, betatau, lh);
	  
	  p = phi;	  
	  ptau = phitau;
	  cout << "  n " << InnerProduct(r,p) << "  " << 
	    InnerProduct ( r,p ) << "   tang " << InnerProduct( rtau, ptau) << endl;
	  
	}
      step ++;
      l2phi = sqrt( InnerProduct(phi, phi) + h * h * InnerProduct( phitau, phitau) );
      l2beta = sqrt( InnerProduct(beta, beta) + h * h * InnerProduct( betatau, betatau) );
      
      
      cout << sqrt( InnerProduct(phi, phi)) << "    " << 
	sqrt(InnerProduct( phitau, phitau) ) << "    " << 
	sqrt( InnerProduct(beta, beta)) << "    " << 
	sqrt(InnerProduct( betatau, betatau) ) << endl;
      

      
      cout << step << endl;
      outfilefreegrad << step << ", " <<l2phi + l2beta << ", " << endl;
      outfilechoppedgrad << step << ", " <<l2beta<<" , " << endl;
      
      
    }
    
     outfilephilambda << step << ", " << L2Norm(phitau) << endl;
    outfilefreegrad << step << ", " << L2Norm(phi) + h*L2Norm(phitau) << endl;
    outfilebetalambda << step << ", " << L2Norm(betatau) << endl;
    outfilechoppedgrad << step << ", " << L2Norm(beta) + h*L2Norm(betatau) << endl;
    ApplyCTranspose( cinfo,  lambdatspace, veclambdan, veclambda, rhs, lh);

    rhs.FVDouble() = vecf.FVDouble() - rhs.FVDouble();
    invmat->Mult (rhs, vecu);
    
    
    
  
    /*

	     
    rtau = 0.0; r=0; vecv.FVDouble() = 0.01;
    veclambdan = 0;  veclambda = 0.0; 
    vecu.FVDouble() = 0.0;
    
    while( uzawasteps <  numuzawasteps && L2Norm(vecv.FVDouble()) > 1e-8)
      {
	
       
	//   right hand side
	//   rhs = f - (apply_cn(w), apply_ctau(lambda) )
	rhs.FVDouble() = 0.0;
	ApplyCTranspose(  cinfo, lambdatspace, veclambdan, veclambda, rhs, lh);
	rhs.FVDouble() = vecf.FVDouble() - rhs.FVDouble();
	
	//   find (sigma, u) by solving system
	//   mata * vecu = rhs
	mata.Mult ( vecu, vecv);

	vecv.FVDouble() = vecv.FVDouble() - rhs.FVDouble();

	invmat->Mult (rhs, vecu);


	//     //   update w, lambda in direction of steepest descent/ascent
	//     //   w = w - tau * (apply_cn(sigma))
	//     //   lambda = lambda + tau * apply_ctau(u)
	

	ApplyC ( cinfo, lambdatspace, vecu, r, rtau, lh);

	r -= vecg;
	//(*testout) << "r " << r << endl << endl << "lambda " << veclambdan << endl;

	veclambdan += taun * r;
	veclambda += tautau * rtau;

       
	
	//     //   project lambda, w onto admissible set
	for(int i=0; i<nd_c; i++)
	  vecsigmann(i) = abs(veclambdan(i));

	//*testout << "VECLAMBDA " << veclambdan << endl;
	bool flag = ProjectStress (cinfo,  lambdatspace, coeff_f, vecsigmann, veclambdan, veclambda, lh);
	//	CalcURel(cinfo, ma, compound_fespace, vecu, r, lh);
	
      
	double error = L2Norm(vecv);
	for(int i=0; i<nd_c; i++)
	  {
	    if(r(i) > 0)
	      error += abs(r(i));
	  }
	
// 	for(int i=0; i<veclambda.Size(); i++)
// 	  error += abs(rtau(i));
	
	Ng_Redraw();
	uzawasteps ++;
	//cout << ew / epsw  << "         " << elambda / epslambda  << endl;
	epslambda=0; elambda=1;
	
	cout << "step " << uzawasteps << ":   " << error << endl;
	(*testout) << "step " << uzawasteps << ":   " << error << endl;
      }
    
    
    */
 
    bfa -> ComputeInternal (vecu, lh);
    Ng_Redraw();
    
    *testout << "VECLAMBDAN "  << veclambdan << endl;
    
    /*
      cout << "lambda_n = " << endl;
      for(int i=0; i<veclambdax.Size(); i++)
      cout << veclambdax(i) << "   " << veclambday(i) <<  "  " << veclambdaz(i) << " sigma " << vecsigmann(i) <<  endl;
      cout << "numberofsteps " << uzawasteps << endl;
      for(int ia =0; ia < ncela+1; ia++){
      for(int ib =0; ib < ncelb+1; ib++)
      cout << vecsigmann(ia + ib*(ncela+1)) << "  ";
      cout << endl;}
    */
    delete invmat;
    lh.CleanUp();
  }












  // declare the numproc 'contact' 

  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("hybridcontact", NumProcContHybrid::Create, NumProcContHybrid::PrintDoc);
    }
    
    Init init;
  }
  




