



class HybridFiniteElement : public FiniteElement
{
protected:
  int vnums[8];
  int firstfacedof[8];
  INT<2> order_facet[6];

  int nfacets;
public:
  HybridFiniteElement (ELEMENT_TYPE etype);
  virtual ~HybridFiniteElement();

  void SetVertexNumbers (FlatArray<int> & avnums)
  {
    for (int i = 0; i < avnums.Size(); i++)
      vnums[i] = avnums[i];
  }

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const = 0;

  virtual void GetFDofNrs(const int & face, ARRAY<int> & fdofs) const = 0;


  virtual void GetOrderFacet(ARRAY<INT<2> > & aorder_facet) const 
  {
    aorder_facet.SetSize(nfacets);
    for ( int i = 0; i < nfacets; i++ )
      aorder_facet[i] = order_facet[i];
  }

  virtual void SetOrderFacet( const ARRAY<INT<2> > & aorder_facet )
  {
    for ( int i = 0; i < aorder_facet.Size(); i++ )
      order_facet[i] = aorder_facet[i];
  }

  virtual void ComputeNDof() = 0;


};


class HybridNormalFiniteElement : public FiniteElement
{
protected:
  int vnums[4];
  INT<2> order_inner;
public:
  HybridNormalFiniteElement (ELEMENT_TYPE etype);
  virtual ~HybridNormalFiniteElement();

  void SetVertexNumbers (FlatArray<int> & avnums)
  {
    for (int i = 0; i < avnums.Size(); i++)
      vnums[i] = avnums[i];
  }

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const = 0;

  virtual void ComputeNDof() = 0;

  virtual void SetOrderInner ( const INT<2> & aorder_inner )
  {
    order_inner = aorder_inner;
  }

};

class HybridNormalSegm : public HybridNormalFiniteElement
{
public:
  HybridNormalSegm (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  virtual void ComputeNDof();
};


class HybridNormalTrig : public HybridNormalFiniteElement
{
public:
  HybridNormalTrig (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  virtual void ComputeNDof();
};


class HybridNormalQuad : public HybridNormalFiniteElement
{
public:
  HybridNormalQuad (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  virtual void ComputeNDof();
};


class HybridTrig : public HybridFiniteElement
{
public:
  HybridTrig (int aorder);


  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void GetFDofNrs(const int & face, ARRAY<int> & fdofs) const;

  virtual void ComputeNDof();
};

class HybridQuad : public HybridFiniteElement
{
public:
  HybridQuad (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
  virtual void GetFDofNrs(const int & face, ARRAY<int> & fdofs) const;

  virtual void ComputeNDof();
};




class HybridTet : public HybridFiniteElement
{
public:
  HybridTet (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;
  virtual void GetFDofNrs(const int & face, ARRAY<int> & fdofs) const;

  virtual void ComputeNDof();
};


class HybridHex : public HybridFiniteElement
{
public:
  HybridHex ( int aorder);

  void GetInternalDofs (ARRAY<int> & idofs) const;

  void CalcShape (const IntegrationPoint & ip, 
		  FlatVector<> shape) const;  
  virtual void GetFDofNrs(const int & face, ARRAY<int> & fdofs) const;
  virtual void ComputeNDof();
};


class HybridPrism : public HybridFiniteElement
{
public:
  HybridPrism ( int aorder);

  void GetInternalDofs (ARRAY<int> & idofs) const;

  void CalcShape (const IntegrationPoint & ip, 
		  FlatVector<> shape) const;  
  virtual void GetFDofNrs(const int & face, ARRAY<int> & fdofs) const;
  virtual void ComputeNDof();
};





/******** Integrators ********************/



template <int D>
class DiffOpBoundIdHybrid : public DiffOp<DiffOpBoundIdHybrid<D> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D-1 };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HybridNormalFiniteElement & fel = 
      dynamic_cast<const HybridNormalFiniteElement&> (bfel);
    
    int nd = fel.GetNDof();

    FlatVector<> shape(nd, lh);
    fel.CalcShape (sip.IP(), shape);
    
     double det = fabs (sip.GetJacobiDet());

     // scaling for lagrange parameter such that it fits sigma_nn
     // transformed_shape = det * shape

     shape *= det;
     mat =  Trans (shape);

  }
};




template<int D>
class DiffOpEpsHCurlA : public DiffOp<DiffOpEpsHCurlA<D> >
{ };


template<>
class DiffOpEpsHCurlA<3> : public DiffOp<DiffOpEpsHCurlA<3> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 3 };
  enum { DIM_ELEMENT = 3 };
  enum { DIM_DMAT = 6 };
  enum { DIFFORDER = 1 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {

    const HCurlFiniteElementD<3> & fel = 
      dynamic_cast<const HCurlFiniteElementD<3>&> (bfel);

    const int D = 3;
    int nd = fel.GetNDof();
    FlatMatrixFixWidth<3> u_ref(nd, lh);
    FlatMatrixFixWidth<3> u(nd, lh);


    Mat<3> jac = sip.GetJacobian();
    Mat<3> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());
    fel.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;


    FlatMatrixFixWidth<D> u_refr(nd, lh);
    FlatMatrixFixWidth<D> ur(nd, lh);
    FlatMatrixFixWidth<D> ul(nd, lh);
    FlatMatrixFixWidth<D> dudx(nd, lh);
    FlatMatrixFixWidth<D> dudy(nd, lh);
    FlatMatrixFixWidth<D> dudz(nd, lh);
    FlatMatrixFixHeight<6> epsmat(nd, lh), aepsmat(nd, lh);
	

    Mat<6> ainv;

    for (int k = 0; k < D; k++)
      {
	IntegrationPoint ipr(sip.IP());
	IntegrationPoint ipl(sip.IP());
	ipr(k) += 1e-6;
	ipl(k) -= 1e-6;
	
	SpecificIntegrationPoint<D,D> sipr (ipr,sip.GetTransformation(), lh);
	SpecificIntegrationPoint<D,D> sipl (ipl,sip.GetTransformation(), lh);
	
	fel.CalcShape (ipr, u_refr);
	ur = u_refr * sipr.GetJacobianInverse();
	
	fel.CalcShape (ipl, u_refr);
	ul = u_refr * sipl.GetJacobianInverse();
	
	ur -= ul;
	ur /= (2e-6);
	
	switch (k)
	  {
	  case 0: dudx = ur; break;
	  case 1: dudy = ur; break;
	  case 2: dudz = ur; break;
	  }
      }
    
    
    // nonconforming 
    
    
    for (int i = 0; i < nd; i++)
      {
	Mat<3,3> gradur, gradu;
	for (int j = 0; j < 3; j++)
	  {
	    gradur(j,0) = dudx(i,j);
	    gradur(j,1) = dudy(i,j);
	    gradur(j,2) = dudz(i,j);
	  }
	gradu =  gradur * inv_jac;
	
	mat(0,i) = gradu(0,0);
	mat(1,i) = gradu(1,1);
	mat(2,i) = gradu(2,2);
	mat(3,i) = (gradu(0,1)+gradu(1,0)) * 0.5 * 2;
	mat(4,i) = (gradu(0,2)+gradu(2,0)) * 0.5 * 2;
	mat(5,i) = (gradu(1,2)+gradu(2,1)) * 0.5 * 2;
	
	    
	    
      }
    


 
  }
};





