/**********************
 FESpace for Lagrange parameters
****************************/


class FESpaceHybrid : public FESpace
{
private:
  int nfa;
  int nel;
  ARRAY<INT<2> > order_facet;

  ARRAY<int> first_face_dof;
  ARRAY<int> first_element_dof;
  int ndof;
  ARRAY<int> ndof_level;
  bool noncontinuous;

  int orderz;

  ARRAY<bool> fine_face; 

public:

  FESpaceHybrid (const MeshAccess & ama, const Flags & aflags, bool parseflags=false);

  virtual ~FESpaceHybrid ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FESpaceHybrid (ma, flags);
  }

  virtual string GetClassName () const
  {
    return "FESpace for Hybridization";
  }


  virtual void Update();


  virtual int GetNDof () const
  {
    return ndof;
  }

  virtual int GetNDofLevel (int level) const
  {
    return ndof_level[level];
  }

  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;

  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  
  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
  
  int GetFirstFaceDof (const int f) const { return first_face_dof[f]; }
  int GetFirstCellDof (const int c) const { return first_element_dof[c]; }


  virtual void GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const;

  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;

  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }

  bool GetFineFace ( const int i ) const
  { return fine_face[i]; };

  virtual const ngmg::Prolongation * GetProlongation () const
  { return 0; }

  virtual void PrintReport (ostream & ost);

#ifdef PARALLEL
   virtual void UpdateParallelDofs_hoproc();
   virtual void UpdateParallelDofs_loproc();
#endif

};



