#include "../../include/solve.hpp"

namespace ngsolve
{
  using namespace ngsolve;
  // using namespace ngbla;
  // using namespace ngla;






  /* *************************** Numproc Printmesh ********************** */


  ///
  class NumProcPrintMesh : public NumProc
  {
  protected:
    ///
    char outputfile[50];
  public:
    ///
    NumProcPrintMesh (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcPrintMesh();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcPrintMesh (pde, flags);
    }


    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "Boundary Value Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl;
    }

    ///
    static void PrintDoc (ostream & ost);
  };



  NumProcPrintMesh :: NumProcPrintMesh (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    strcpy (outputfile,  flags.GetStringFlag("file", "out.gmv"));
  }

  NumProcPrintMesh :: ~NumProcPrintMesh()
  {
    ;
  }

  void NumProcPrintMesh :: PrintDoc (ostream & ost)
  {
    ost << 
      "\n\nNumproc PrintMesh:\n" \
      "------------\n" \
	<< endl;
  }


  void NumProcPrintMesh :: Do(LocalHeap & lh)
  {

    const MeshAccess & ma = pde . GetMeshAccess();

    int D =  ma.GetDimension() ;

    double point[3]
;
    cout << ma.GetNE() << endl;

    ofstream outfile ( outputfile );

    outfile << "gmvinput ascii" << endl << endl 
	    << "comments " << endl 
	    << "  File generated by Netgen/NgSolve" << endl 
	    << "endcomm" << endl << endl << endl;

    outfile << "nodev " << ma.GetNV() << endl << endl;

    for ( int i = 0; i < ma.GetNV(); i++)
      {
	ma.GetPoint(i, point);
	for (int d= 0; d < 3; d++)
	  outfile << point[d] << " ";
	outfile << endl; 
      }

    
    outfile << endl << endl << endl;

    outfile << "cells " << ma.GetNE() << endl;

    for ( int el = 0; el < ma.GetNE(); el ++)
      {

	ELEMENT_TYPE  eltype = ma.GetElType ( el );
	int nv = ma.GetElNV(el);
	ARRAY<int> pnums;
	ma.GetElPNums ( el, pnums );

	if ( eltype == ET_TET )
	  outfile << "tet " ;
	else if (eltype == ET_TRIG)
	  outfile << "tri " ;
	else if (eltype == ET_QUAD)
	  outfile << "quad ";

	outfile << nv << "  ";
	for(int vert=0; vert<nv; vert++)
	  outfile << pnums[vert] << " ";
	outfile << endl;
      }


    outfile << endl << endl << "surface ";
    outfile << ma.GetNSE() << endl << endl;

    for ( int sel = 0; sel < ma.GetNSE() ; sel ++)
      {
	ARRAY<int> pnums;
	ma.GetSElPNums ( sel, pnums );
	int nv = pnums.Size();
	outfile << nv << "   ";

	for ( int vert = 0; vert < nv; vert ++)
	  outfile << pnums[vert] << "   ";
	outfile << endl;


      }


    outfile << endl << endl 
	    << "surfflag" << endl
	    << "     boundarytype " << ma.GetNBoundaries () << endl << "     ";
    for ( int i = 0; i < ma.GetNBoundaries(); i++)
      outfile << i << "  ";

    outfile << endl << "     ";
    for(int sel = 0; sel < ma.GetNSE(); sel++)
      {
	outfile << ma.GetSElIndex(sel) << " ";
      }


    outfile << endl << endl << "     " <<
      "surfacetype " << 1 << endl << "     ";

    if ( ma.GetDimension() == 2 )
      outfile << "segment" << endl << "     ";
    else 
      outfile << "     trig" << endl << "     ";
    for ( int sel = 0; sel < ma.GetNSE(); sel++)
      outfile << "1 " ;

    outfile << endl << "endsflag" << endl << endl;

    outfile << "endgmv" << endl;
  }
















  


  namespace
#ifdef MACOS
  bvp_cpp
#endif
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("printmesh", NumProcPrintMesh::Create, NumProcPrintMesh::PrintDoc);
    }
    
    
    Init init;
    
  }
  


}
