########################################################################################
# solve poisson equation
#    - eps Delta u = f in Omega,  u=g on Gamma0
# by means of a mixed-hybrid method 
#   (see the hybrid demo in ngsolve/tutorial for details)
#
# compare RT and MariniPietra elements  
########################################################################################


geometry = square.in2d
mesh = square.vol

# geometry = cube.geo
# mesh = cube.vol

define constant heapsize = 10000000
define constant b1 = 1.1
define constant b2 = 1
define constant b3 = 0
define constant eps = 1e-2
define constant d = 2

## boundary terms ########################################
define coefficient crob
# 1e12, 1e12, 0, 0, 0, 0,
1e12, 1e12, 1e12, 1e12, 1e12, 1e12,

define coefficient cneu
# -1e12, 1e12, 0, 0, 0, 0, 
0, 0, 0, 0, 0, 0, 


## some coefficients #####################################
define coefficient one
1, 1, 1, 1, 1, 1,

define coefficient d
(-d/eps), 

define coefficient b1
(b1/eps), 

define coefficient b2
(b2/eps), 

define coefficient b3
(b3/eps), 


define coefficient uex
( x * y * ( 1 - exp((y-1)/eps) ) * ( 1 - exp((x-1)/eps) )),

define coefficient minusf
( -1/eps*((x-y-2*x*y) * exp((y-1)/eps) + (x+y+2*x*y) + (y-x-2*x*y)*exp((x-1)/eps) + exp((x+y-2)/eps) * (2*x*y-x-y)) ),

# 

###########################################################
## Marini Pietra Hybrid Formulierung  
define fespace v1 -type=mphybrid -eliminate_internal
# define fespace v1 -type=hybrid -order=1 -eliminate_internal

define gridfunction u1 -fespace=v1

## linearform : (-f, v), <g,mu>
define linearform f1 -fespace=v1
neumann cneu -comp=3
source minusf -comp=2

numproc mporient mp -b1=b1 -b2=b2 -fespace=v1

## bilinearform: laplace and robin
define bilinearform a1 -fespace=v1 -eliminate_internal -linearform=f1 #-printelmat
expfitlaplace one b1 b2 #b3
# hybridlaplace one
mass d -comp=2
robin crob -comp=3

## elementary solver
numproc bvp np1 -bilinearform=a1 -linearform=f1 -gridfunction=u1 -solver=direct


# visualization
define bilinearform drawsigma1 -fespace=v1 -nonassemble
masshdiv one -comp=1
define bilinearform drawu1 -fespace=v1 -nonassemble
mass one -comp=2
define bilinearform drawlambda1 -fespace=v1 -nonassemble
mass one -comp=3

numproc drawflux df11 -bilinearform=drawsigma1 -solution=u1 -label=sigma1
numproc drawflux df12 -bilinearform=drawu1 -solution=u1 -label=u1
numproc drawflux df13 -bilinearform=drawlambda1 -solution=u1 -label=lambda1


# automatic visualizaiton for 2D
numproc visualization vis1 -scalarfunction=u1 -subdivision=0





###########################################################
## Raviart Thomas Hybrid Formulierung  
define fespace v2 -type=hybrid -order=0 -eliminate_internal

define gridfunction u2 -fespace=v2

## linearform : (-f, v), <g,mu>
define linearform f2 -fespace=v2
neumann cneu -comp=3
source minusf -comp=2

## bilinearform: laplace and robin
define bilinearform a2 -fespace=v2 -eliminate_internal -linearform=f2 
hybridlaplace one
hybridconvection b1 b2 one
mass d -comp=2
robin crob -comp=3

## elementary solver
numproc bvp np2 -bilinearform=a2 -linearform=f2 -gridfunction=u2 -solver=direct


# visualization
define bilinearform drawsigma2 -fespace=v2 -nonassemble
masshdiv one -comp=1
define bilinearform drawu2 -fespace=v2 -nonassemble
mass one -comp=2
define bilinearform drawlambda2 -fespace=v2 -nonassemble
mass one -comp=3

numproc drawflux df11 -bilinearform=drawsigma2 -solution=u2 -label=sigma2
numproc drawflux df12 -bilinearform=drawu2 -solution=u2 -label=u2
numproc drawflux df13 -bilinearform=drawlambda2 -solution=u2 -label=lambda2









## exact solution and error ############################################
define fespace vex -type=h1 -order=4 -eliminate_internal

define gridfunction uex -fespace=vex 

define linearform fex -fespace=vex
source uex


define bilinearform aex -fespace=vex  -linearform=fex -eliminate_internal
mass one

define preconditioner cex -type=direct -bilinearform=aex #-inverse=pardiso
numproc bvp npex -bilinearform=aex -linearform=fex -gridfunction=uex -preconditioner=cex

define fespace verr -type=l2
define gridfunction err1 -fespace=verr
define gridfunction err2 -fespace=verr
# define gridfunction err -fespace=verr

define bilinearform massex -fespace=vex -eliminate_internal
mass one

# define bilinearform massh1 -fespace=vh1 -eliminate_internal
# mass one

# numproc difference diff1 -bilinearform1=massex -solution1=uex -bilinearform2=drawu1 -solution2=u1 -diff=err1  
# numproc difference diff2 -bilinearform1=massex -solution1=uex -bilinearform2=drawu2 -solution2=u2 -diff=err2  
# numproc difference npdiffh1 -bilinearform1=drawu -solution1=u -bilinearform2=massh1 -solution2=uh1 -diff=err


# numproc markelements npmark -error=err -factor=0.1 -minlevel=0

