#ifndef MARINI_PIETRA_HPP___
#define MARINI_PIETRA_HPP___

#include <fem.hpp>

namespace ngfem {

  // dummy surface element
  // space works only in hybrid setting, hence no boundary elements needed.
  // surface element only required for compound fe surface element
  template <int D>
  class MariniPietraNormalFiniteElement : public HDivNormalFiniteElement<D>
  {
    protected:
      mutable int fmax;
    public:
      MariniPietraNormalFiniteElement (ELEMENT_TYPE aeltype) 
      : HDivNormalFiniteElement<D>(aeltype, 0, -1) {};
      
      virtual void CalcShape (const IntegrationPoint & ip, FlatVector<> shape) const {};

  };

  // Base Volume Element 
  template <int D>
  class MariniPietraFiniteElement : public HDivFiniteElement<D>
  {
    protected:
      mutable int fmax;
    public:
      //
      MariniPietraFiniteElement (ELEMENT_TYPE aeltype);
      void SetOrientation(int afmax) const { fmax=afmax; } // the integrator may work on that!!
  };



  // Triangle
  class MariniPietraTrig : public MariniPietraFiniteElement<2>
  {
    public:
      MariniPietraTrig ();
  
      virtual void ComputeNDof() {};
  
      virtual void GetInternalDofs (ARRAY<int> & idofs) const;

      virtual void CalcShape (const IntegrationPoint & ip,  FlatMatrixFixWidth<2> shape) const;
  
      virtual void CalcDivShape (const IntegrationPoint & ip, FlatVector<> shape) const;
  };

  // Tetrahedron
  class MariniPietraTet : public MariniPietraFiniteElement<3>
  {
    public:
      MariniPietraTet ();
  
      virtual void ComputeNDof() {};
  
      virtual void GetInternalDofs (ARRAY<int> & idofs) const;

      virtual void CalcShape (const IntegrationPoint & ip,  FlatMatrixFixWidth<3> shape) const;
  
      virtual void CalcDivShape (const IntegrationPoint & ip, FlatVector<> shape) const;
  };

} // namespace


#endif



