// hybrid demo with marini pietra elements for hdiv

#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

#include "marinipietrafe.hpp"
#include "marinipietrafespace.hpp"

namespace marini_pietra_hybrid
{

  using namespace ngfem;
  using namespace ngsolve;
  using namespace ngstd;

 
  /* *********************************************************** */
  /* ********************** HybridFESpace ********************** */
  /* *********************************************************** */
  class HybridFESpace : public CompoundFESpace
  {
    public:
      HybridFESpace (const MeshAccess & ama, const ARRAY<const FESpace*> & aspaces)
      : CompoundFESpace (ama, aspaces)
      { }

      virtual ~HybridFESpace () { ; }

      static FESpace * Create (const MeshAccess & ma, const Flags & flags)
      {
        ARRAY<const FESpace*> spaces(3);

        Flags hdivflags(flags), l2flags(flags), facetflags(flags);

        int order = 0; 

        l2flags.SetFlag ("order", 0.0);
        facetflags.SetFlag("order", 0.0);

        // a discontinuous HDiv space
        spaces[0] = new MariniPietraFESpace (ma, hdivflags);
         
        spaces[1] = new L2HighOrderFESpace (ma, l2flags);    
        
        // a space for functions defined on element interfaces 
        spaces[2] = new FacetFESpace (ma, facetflags);        

        HybridFESpace * fes = new HybridFESpace (ma, spaces);
        return fes;
      }
    
      virtual string GetClassName () const { return "MariniPietraHybridFESpace"; }
  };

  
  class Init
  { 
    public: 
      Init ();
  };
   
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("mphybrid", HybridFESpace::Create);
  }

  Init init;

} // namespace

