########################################################################################
# solve pde
#    - Delta u + 10 u = 1 in Omega,  u=g on Gamma0
# by means of a mixed-hybrid method 
#   (see the hybrid demo in ngsolve/tutorial for details)
#
# compare RT and MariniPietra elements  
########################################################################################


# geometry = square.in2d
# mesh = square.vol

geometry = cube.geo
mesh = cube.vol

define constant heapsize = 10000000

## boundary terms ########################################
define coefficient crob
1e12, 1e12, 1e12, 1e12, 1e12, 1e12,

define coefficient cneu
0, 0, 0, 0, 0, 0, 


## some coefficients #####################################
define coefficient minusone 
-1,-1,-1,-1,-1,-1,

define coefficient one
1,1,1,1,1,1,

define coefficient minusten 
-10,-10,-10,-10,-10,-10,

define coefficient ten
10,10,10,10,10,10,


###########################################################
## Marini Pietra Hybrid Formulierung  
# define fespace v1 -type=mphybrid -eliminate_internal
define fespace v1 -type=hybrid -eliminate_internal

define gridfunction u1 -fespace=v1

## linearform : (-f, v), <g,mu>
define linearform f1 -fespace=v1
neumann cneu -comp=3
source minusone -comp=2

## bilinearform: laplace and robin
define bilinearform a1 -fespace=v1 -eliminate_internal -linearform=f1 -printelmat
hybridlaplace one
mass minusten -comp=2  
robin crob -comp=3

## elementary solver
numproc bvp np1 -bilinearform=a1 -linearform=f1 -gridfunction=u1 -solver=direct


# visualization
define bilinearform drawsigma1 -fespace=v1 -nonassemble
masshdiv one -comp=1
define bilinearform drawu1 -fespace=v1 -nonassemble
mass one -comp=2
define bilinearform drawlambda1 -fespace=v1 -nonassemble
robin one -comp=3

numproc drawflux df11 -bilinearform=drawsigma1 -solution=u1 -label=sigma1
numproc drawflux df12 -bilinearform=drawu1 -solution=u1 -label=u1
numproc drawflux df13 -bilinearform=drawlambda1 -solution=u1 -label=lambda1


## automatic visualizaiton for 2D
# numproc visualization vis1 -scalarfunction=u1 -subdivision=0





###########################################################
## Raviart Thomas Hybrid Formulierung  
define fespace v2 -type=hybrid -order=0 -eliminate_internal

define gridfunction u2 -fespace=v2

## linearform : (-f, v), <g,mu>
define linearform f2 -fespace=v2
neumann cneu -comp=3
source minusone -comp=2

## bilinearform: laplace and robin
define bilinearform a2 -fespace=v2 -eliminate_internal -linearform=f2 
hybridlaplace one
mass minusten -comp=2
robin crob -comp=3

## elementary solver
numproc bvp np2 -bilinearform=a2 -linearform=f2 -gridfunction=u2 -solver=direct


# visualization
define bilinearform drawsigma2 -fespace=v2 -nonassemble
masshdiv one -comp=1
define bilinearform drawu2 -fespace=v2 -nonassemble
mass one -comp=2
define bilinearform drawlambda2 -fespace=v2 -nonassemble
robin one -comp=3

numproc drawflux df11 -bilinearform=drawsigma2 -solution=u2 -label=sigma2
numproc drawflux df12 -bilinearform=drawu2 -solution=u2 -label=u2
numproc drawflux df13 -bilinearform=drawlambda2 -solution=u2 -label=lambda2








## exact solution and error ############################################
define fespace vex -type=h1 -order=4 

define gridfunction uex -fespace=vex 

define linearform fex -fespace=vex
neumann cneu
source one 

define bilinearform aex -fespace=vex 
laplace one
mass ten
robin crob 

define preconditioner cex -type=direct -bilinearform=aex 
numproc bvp npex -bilinearform=aex -linearform=fex -gridfunction=uex -preconditioner=cex

define fespace verr -type=l2
define gridfunction err1 -fespace=verr
define gridfunction err2 -fespace=verr
# define gridfunction err -fespace=verr



define bilinearform massex -fespace=vex 
mass one

# define bilinearform massh1 -fespace=vh1 -eliminate_internal
# mass one

numproc difference diff1 -bilinearform1=massex -solution1=uex -bilinearform2=drawu1 -solution2=u1 -diff=err1  
numproc difference diff2 -bilinearform1=massex -solution1=uex -bilinearform2=drawu2 -solution2=u2 -diff=err2  
# numproc difference npdiffh1 -bilinearform1=drawu -solution1=u -bilinearform2=massh1 -solution2=uh1 -diff=err


# numproc markelements npmark -error=err -factor=0.1 -minlevel=0

