#include <solve.hpp>

using namespace std;
using namespace ngbla;
using namespace ngla;



namespace ngsolve
{
  using namespace ngsolve;
  ///
  class NumProcABBCurrents : public NumProc
  {
  protected:
    ///
    LinearForm * lf;
    ///
    int v1, v2;
    ///
    double current,phase;
    ///
    int surfindex; 

  public:
    ///
    NumProcABBCurrents (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcABBCurrents();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcABBCurrents (pde, flags);
    }

    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "ABB-Currents";
    }

    virtual void PrintReport (ostream & ost)
    {
      ;
    }
  };



  NumProcABBCurrents :: NumProcABBCurrents (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    lf = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    v1 = int(flags.GetNumFlag ("v1", 0));
    v2 = int(flags.GetNumFlag ("v2", 0));
    current = flags.GetNumFlag("current",1.); 
    phase = flags.GetNumFlag("phase",0.); 
    surfindex = int(flags.GetNumFlag("nocurrsurf",-1)); 
  }

  NumProcABBCurrents :: ~NumProcABBCurrents()
  {
    ;
  }


  void NumProcABBCurrents :: Do(LocalHeap & lh)
  {
    cout << "assemble ABB Currents" << endl;

    // FlatVector<double> fv = lf->GetVector().FVDouble();
    FlatVector<Complex> fv = lf->GetVector().FVComplex();

    cout << "v1 = " << v1 << endl;
    cout << "v2 = " << v2 << endl;
    
    int nv = ma.GetNV();
    
    ARRAY<int> dist(nv);
    ARRAY<int> from_vert(nv);
    ARRAY<int> edgenr(nv);

    dist = 99999;

    int i,j; 
    int nse = ma.GetNSE(); 
    
  
    if(surfindex > -1)     
      for(i=0; i<nse; i++)
	// if(ma.GetSElIndex(i) == surfindex)
	  {
	    
	    ARRAY<int> vnums; 
	    ma.GetSElVertices(i,vnums);
	    
	    for(j=0;j<vnums.Size();j++)
	      {
		
		dist[vnums[j]]=200000;
	      }
	  }
	
    
    dist[v1] = 0;	
    dist[v2] = 99999;
    
    Complex fac = Complex(current * cos(phase/180.*M_PI), current * sin(phase/180.*M_PI)); 
 
    bool changed;
    do
      {
	changed = 0;

	for ( i = 0; i < ma.GetNEdges(); i++)
	  {
	    int ev1, ev2;
	    ma.GetEdgePNums (i, ev1, ev2);
	    if(dist[ev1] != 200000 && dist[ev2] != 200000)
	      {
		int minval = min2 (dist[ev1], dist[ev2]);
		if (dist[ev1] > minval+1)
		  {
		    changed = 1;
		    dist[ev1] = minval+1;
		    from_vert[ev1] = ev2;
		    edgenr[ev1] = i;
		  }
		
		if (dist[ev2] > minval+1)
		  {
		    changed = 1;
		    dist[ev2] = minval+1;
		    from_vert[ev2] = ev1;
		    edgenr[ev2] = i;
		  }
	      }
	  }
	cout << "dist[v2] = " << dist[v2] << endl;
      }
    while (changed);

    int start, end;
    end = v2;
    do
      {
	start = from_vert[end];
	int enr = edgenr[end];
	cout << "edge " << enr << " verts = " << start << " - " << end << endl;
	end = start;

	fv(enr) += (end > start) ? fac : -fac;
      }
    while (start != v1); 
    
  }



 class NumProcABBRhsH1Pot : public NumProc
  {
  protected:
    ///
    LinearForm * lf;
    ///
    BilinearForm * bfme; 
    ///
    GridFunction * gfphi; 
   
  public:
    ///
    NumProcABBRhsH1Pot (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcABBRhsH1Pot();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcABBRhsH1Pot (pde, flags);
    }

    ///
    virtual void Do();
    ///
    virtual string GetClassName () const
    {
      return "ABB-RHS-Potential";
    }

    virtual void PrintReport (ostream & ost)
    {
      ;
    }
  };



  NumProcABBRhsH1Pot :: NumProcABBRhsH1Pot (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    lf = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    bfme = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformme",""));
    gfphi = pde.GetGridFunction(flags.GetStringFlag("h1pot","")); 
        
  }

  NumProcABBRhsH1Pot :: ~NumProcABBRhsH1Pot()
  {
    ;
  }


  void NumProcABBRhsH1Pot :: Do()
  {
    
    BaseVector & fv = lf->GetVector();

    const BaseMatrix & matme = bfme->GetMatrix();
    const BaseVector & phi = gfphi->GetVector();
    
    
    SparseMatrix<double> * grad = 0;
  
    if (dynamic_cast<const HCurlHighOrderFESpace*> (&bfme->GetFESpace()))
      grad = dynamic_cast<const HCurlHighOrderFESpace*> (&bfme->GetFESpace())->CreateGradient();

    if (dynamic_cast<const NedelecFESpace*> (&bfme->GetFESpace()))
      grad = dynamic_cast<const NedelecFESpace*> (&bfme->GetFESpace())->CreateGradient();
    
    Real2ComplexMatrix<double,Complex> cgrad (grad); 

    BaseVector &  hv = *fv.CreateVector();
    hv = cgrad*phi; 
    fv = matme * hv;  
     

  }


  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("abbcurrents", NumProcABBCurrents::Create);
      GetNumProcs().AddNumProc ("abbcurrentsh1pot", NumProcABBRhsH1Pot::Create);
    }

    
    Init init;
  }

}
