#include <solve.hpp>
using namespace ngsolve;
extern ngsolve::PDE * pde;



class FESpaceHCurlH1 : public CompoundFESpace
{
public:
  FESpaceHCurlH1 (const MeshAccess & ama, 		   
		  const ARRAY<const FESpace*> & aspaces,
		  const Flags & flags)
    : CompoundFESpace (ama, aspaces, flags)
  { ; }
  
  virtual ~FESpaceHCurlH1 () { ; }
  
  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    cout << "Create H(curl)-H1 FE Space" << endl;

    ARRAY<const FESpace*> spaces(2);

    int order = int (flags.GetNumFlag ("order", 2));

    Flags uflags;
    uflags.SetFlag ("order", order);
    spaces[0] = new H1HighOrderFESpace (ma, uflags);
    
    uflags.SetFlag ("order", order-1);
    spaces[1] = new HCurlHighOrderFESpace (ma, uflags);
    
    FESpaceHCurlH1 * fes = new FESpaceHCurlH1 (ma, spaces, flags);
    return fes;
  }
  
  virtual string GetClassName () const { return "HCurl-H1-FESpace"; }
};



class FESpaceHCurlH1H1 : public CompoundFESpace
{
public:
  FESpaceHCurlH1H1 (const MeshAccess & ama, 		   
		  const ARRAY<const FESpace*> & aspaces,
		  const Flags & flags)
    : CompoundFESpace (ama, aspaces, flags)
  { ; }
  
  virtual ~FESpaceHCurlH1H1 () { ; }
  
  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    cout << "Create H(curl)-H1 FE Space" << endl;

    ARRAY<const FESpace*> spaces(3);

    int order = int (flags.GetNumFlag ("order", 2));

    Flags uflags;
    uflags.SetFlag ("order", order);
    spaces[0] = new H1HighOrderFESpace (ma, uflags);
    spaces[2] = new H1HighOrderFESpace (ma, uflags);
    
    uflags.SetFlag ("order", order-1);
    spaces[1] = new HCurlHighOrderFESpace (ma, uflags);

    
    FESpaceHCurlH1H1 * fes = new FESpaceHCurlH1H1 (ma, spaces, flags);
    return fes;
  }
  
  virtual string GetClassName () const { return "HCurl-H1-H1-FESpace"; }
};







class DiffOpCoupledMass : public DiffOp<DiffOpCoupledMass>
{
public:
  enum { DIM = 1 };          // just one copy of the spaces
  enum { DIM_SPACE = 2 };    // 2D space
  enum { DIM_ELEMENT = 2 };  // 2D elements (in contrast to 1D boundary elements)
  enum { DIM_DMAT = 2 };     // D-matrix is 5x5
  enum { DIFFORDER = 0 };    // minimal differential order (to determine integration order)
  

  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    typedef typename SIP::TSCAL TSCAL;
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);

    // a scalar H1 element
    const NodalFiniteElement & fel_h1 = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_hcurl = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_h1    = fel_h1.GetNDof();
    int nd_hcurl = fel_hcurl.GetNDof();

    // transformation of derivatives from reference element to general element:
    FlatMatrix<TSCAL> grad_h1(2, nd_h1, lh);
    grad_h1 = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_h1.GetDShape(sip.IP(), lh));

    FlatMatrix<TSCAL> shape_hcurl(2, nd_hcurl, lh);
    shape_hcurl = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_hcurl.GetShape(sip.IP(), lh));

    mat = TSCAL(0.0);
    for (int j = 0; j < nd_h1; j++)
      {
	mat(0,j) = grad_h1(0,j);
	mat(1,j) = grad_h1(1,j);
      }

    for (int j = 0; j < nd_hcurl; j++)
      {
	mat(0, nd_h1+j) = shape_hcurl(0,j);
	mat(1, nd_h1+j) = shape_hcurl(1,j);
      }
  }
};

class CoupledMassIntegrator 
  : public T_BDBIntegrator<DiffOpCoupledMass, DiagDMat<2>, FiniteElement>
{
public:
  CoupledMassIntegrator (CoefficientFunction * coeff)
    :  T_BDBIntegrator<DiffOpCoupledMass, DiagDMat<2>, FiniteElement>
  (DiagDMat<2> (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new CoupledMassIntegrator (coeffs[0]);
  }
  
  virtual string Name () const { return "CoupledMass"; }
};


class PML_CoupledMassIntegrator 
  : public PML_BDBIntegrator<DiffOpCoupledMass, DiagDMat<2>, FiniteElement>
{
public:
  PML_CoupledMassIntegrator (CoefficientFunction * coeff)
    :  PML_BDBIntegrator<DiffOpCoupledMass, DiagDMat<2>, FiniteElement>
  (DiagDMat<2> (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new PML_CoupledMassIntegrator (coeffs[0]);
  }
  
  virtual string Name () const { return "CoupledMass"; }
};



















class DiffOpChannelConstraints : public DiffOp<DiffOpChannelConstraints>
{
  // v_l, v_tau, lam, grad lam
public:
  enum { DIM = 1 };          // just one copy of the spaces
  enum { DIM_SPACE = 2 };    // 2D space
  enum { DIM_ELEMENT = 2 };  // 2D elements (in contrast to 1D boundary elements)
  enum { DIM_DMAT = 6 };     // D-matrix is 6x6
  enum { DIFFORDER = 0 };    // minimal differential order (to determine integration order)
  

  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);

    // a scalar H1 element
    const NodalFiniteElement & fel_h1 = 
      dynamic_cast<const NodalFiniteElement&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_hcurl = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);

    const NodalFiniteElement & fel_lam = 
      dynamic_cast<const NodalFiniteElement&> (cfel[2]);
    
    int nd_h1    = fel_h1.GetNDof();
    int nd_hcurl = fel_hcurl.GetNDof();
    int nd_lam    = fel_h1.GetNDof();

    // transformation of derivatives from reference element to general element:
    FlatVector<> shape_h1(nd_h1, lh);
    fel_h1.CalcShape (sip.IP(), shape_h1);
    FlatMatrix<> grad_h1(2, nd_h1, lh);
    grad_h1 = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_h1.GetDShape(sip.IP(), lh));

    FlatMatrix<> shape_hcurl(2, nd_hcurl, lh);
    shape_hcurl = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_hcurl.GetShape(sip.IP(), lh));


    FlatVector<> shape_lam(nd_lam, lh);
    fel_lam.CalcShape (sip.IP(), shape_lam);
    FlatMatrix<> grad_lam(2, nd_lam, lh);
    grad_lam = Trans (sip.GetJacobianInverse ()) * 
      Trans (fel_lam.GetDShape(sip.IP(), lh));

    mat = 0.0;

    for (int j = 0; j < nd_h1; j++)
      mat(0,j) = shape_h1(j);

    for (int j = 0; j < nd_hcurl; j++)
      {
	mat(1, nd_h1+j) = shape_hcurl(0,j);
	mat(2, nd_h1+j) = shape_hcurl(1,j);
      }

    for (int j = 0; j < nd_lam; j++)
      {
	mat(3, nd_h1+nd_hcurl+j) = shape_lam(j);
	mat(4, nd_h1+nd_hcurl+j) = grad_lam(0,j);
	mat(5, nd_h1+nd_hcurl+j) = grad_lam(1,j);
      }
  }
};



class ChannelConstraintsDMat : public DMatOp<ChannelConstraintsDMat>
{
  CoefficientFunction * muinv;
  CoefficientFunction * omegaeps;
public:

  enum { DIM_DMAT = 6 };

  ChannelConstraintsDMat (CoefficientFunction * amuinv,
			  CoefficientFunction * aomegaeps)
    : muinv(amuinv), omegaeps(aomegaeps) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    // (nu  grud u, grad v)
    double mu = muinv -> Evaluate (sip);
    double eps = omegaeps -> Evaluate (sip);

    // (v_l, lam)_muinv
    mat(0,3) = mat(3,0) = eps;
    // (v_t, grad lam) 
    mat(1,4) = mat(4,1) = mu;
    mat(2,5) = mat(5,2) = mu;
  }  
};


class ChannelConstraintsIntegrator 
  : public T_BDBIntegrator<DiffOpChannelConstraints, ChannelConstraintsDMat, FiniteElement>
{
public:
  ChannelConstraintsIntegrator (CoefficientFunction * coeff1,
				CoefficientFunction * coeff2)
    :  T_BDBIntegrator<DiffOpChannelConstraints, ChannelConstraintsDMat, FiniteElement>
  (ChannelConstraintsDMat (coeff1, coeff2))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new ChannelConstraintsIntegrator (coeffs[0], coeffs[1]);
  }
  
  virtual string Name () const { return "ChannelConstraings"; }
};













namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("channel", FESpaceHCurlH1::Create);
    GetFESpaceClasses().AddFESpace ("channel2", FESpaceHCurlH1H1::Create);

    GetIntegrators().AddBFIntegrator ("coupledmass", 2, 1,
				      CoupledMassIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("PML_coupledmass", 2, 1,
				      PML_CoupledMassIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("channelconstraints", 2, 2,
				      ChannelConstraintsIntegrator::Create);
  }
  
  Init init;
}
