#include <solve.hpp>
// #include <nginterface.h>


#include "../lapack/LapackGEP.hpp"

using namespace std;
using namespace ngbla;
using namespace ngla;

extern "C" void jdqz_ (Complex * alpha, Complex * beta, Complex * eivec,
		       char * wanted, int * n, Complex * target, double * eps,
		       int * kmax, int * jmax, int * jmin, int * method, int * m, int * l,
		       int * mxmv, int * maxstep, double * lock, int * order, int * testspace,
		       Complex * work, int * lwork);

extern "C" void amul_ (int & n, Complex * x, Complex * y);
extern "C" void bmul_ (int & n, Complex * x, Complex * y);
extern "C" void precon_ (int & n, Complex * x);

static BaseMatrix * globmata;
static BaseMatrix * globmatb;
static BaseMatrix * globmatainv;

void amul_ (int & n, Complex * x, Complex * y)
{
  VFlatVector<Complex> vx(n, x);
  VFlatVector<Complex> vy(n, y);
  vy = (*globmata) * vx;

  /*
  for (int i = 0; i < n; i++)
    y[i] = double(i+1) * x[i];
  */
}

void bmul_ (int & n, Complex * x, Complex * y)
{
  VFlatVector<Complex> vx(n, x);
  VFlatVector<Complex> vy(n, y);
  vy = (*globmatb) * vx;
  /*
  for (int i = 0; i < n; i++)
    y[i] = (1.0 / (i+1.0)) * x[i];
  */
}

void precon_ (int & n, Complex * x)
{
  VFlatVector<Complex> vx(n, x);
  VVector<Complex> vy(n);

  //  cout << "pre, |x| = " << L2Norm (vx) << endl;
  /*
  for (int i = 0; i < n; i++)
    x[i] *= (i+1.0) / ((i+1)*(i+1)-31);
  */
  //  cout << "after pre, |x| = " << L2Norm (vx) << endl;
  //  q(i) = i*q(i)/(i*i-31)

  vy = (*globmatainv) * vx;
  vx = 1.0 * vy;
  //  std::cout << "pre" << endl;
}


namespace ngsolve
{
  using namespace ngsolve;
  ///
  class NumProcEVPCoarse : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    ///
    BilinearForm * bfm;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    ///
    int maxsteps, nr, num, maxnewton;
    ///
    double prec;
    ///
    bool print;
  public:
    ///
    NumProcEVPCoarse (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcEVPCoarse();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcEVPCoarse (pde, flags);
    }

    ///
    virtual void Do();
    ///
    virtual string GetClassName () const
    {
      return " Eigenvalue Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form A = " << bfa->GetName() << endl
	  << "Bilinear-form M = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl;
    }
  };



  NumProcEVPCoarse :: NumProcEVPCoarse (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""));
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
    maxnewton = int(flags.GetNumFlag ("maxnewton", 0));
    nr = int(flags.GetNumFlag ("nr", 0));
    num = int(flags.GetNumFlag ("num", 0));
  }

  NumProcEVPCoarse :: ~NumProcEVPCoarse()
  {
    ;
  }


  void NumProcEVPCoarse :: Do()
  {
    cout << "solve  evp" << endl;

    if (!bfa->GetFESpace().IsComplex())
      {
	int level = ma.GetNLevels()-1;
	cout << "level = " << level;
	
	if (level <= 0) return;
	
	const BilinearForm * bfac = 
	  bfa->HasLowOrderBilinearForm() ?  
	  &bfa->GetLowOrderBilinearForm() : bfa;

	const BilinearForm * bfmc = 
	  bfm->HasLowOrderBilinearForm() ? 
	  &bfm->GetLowOrderBilinearForm() : bfm;

	const SparseMatrix<double> & mata = 
	  dynamic_cast<const SparseMatrix<double>&> (bfa->GetMatrix());
	const SparseMatrix<double> & matac = 
	  dynamic_cast<const SparseMatrix<double>&> (bfac->GetMatrix(0));
	const SparseMatrix<double> & matm = 
	  dynamic_cast<const SparseMatrix<double>&> (bfm->GetMatrix());
	const SparseMatrix<double> & matmc = 
	  dynamic_cast<const SparseMatrix<double>&> (bfmc->GetMatrix(0));
	
	BaseVector & vecu = gfu->GetVector();
	
	const BaseMatrix & matpre = pre->GetMatrix();
	
	// int n = matac.Height();
	int nc = matac.Height();
	// nc = 0;
	
	int i, j, k, l;
	
	BaseVector & d = *vecu.CreateVector();
	BaseVector & w = *vecu.CreateVector();
	BaseVector & au = *vecu.CreateVector();
	BaseVector & mu = *vecu.CreateVector();
	
	BaseVector & auc = *matac.CreateVector();
	BaseVector & muc = *matac.CreateVector();
	
	Vector<double> lami(nc+2*num);
	
	
	vecu.SetRandom();
	
	const Prolongation & prol = *gfu->GetFESpace().GetProlongation();
	
	cout.precision(16);
	
	ARRAY<BaseVector*> vecui(2*num);
	ARRAY<BaseVector*> vecunew(num);
	
	
	for (i = 0; i < num; i++)
	  {
	    stringstream strname;
	    strname << "gfu" << i;
	    string name = strname.str();
	    
	    Flags flags;
	    GridFunction * gf = CreateGridFunction (&gfu->GetFESpace(), name, flags);
	    gf->Update();
	    gf->Visualize (name);
	    
	    vecui[i] = &gf->GetVector(); // vecu.CreateVector();
	    vecui[i] -> SetRandom();
	  }
	
	for (i = num; i < 2*num; i++)
	  {
	    vecui[i] = vecu.CreateVector();
	    vecui[i] -> SetRandom();
	  }
	
	for (i = 0; i < num; i++)
	  {
	    vecunew[i] = vecu.CreateVector();
	  }
	
	Matrix<double> ha(nc+2*num, nc+2*num);
	Matrix<double> hm(nc+2*num, nc+2*num);


	/*
	FlatVector<> fu = vecu.FVDouble();
	if (nc > 0)
	  {
	    Matrix<> matac2(nc);
	    Matrix<> matmc2(nc);
	    for (int i = 0; i < nc; i++)
	      {
		fu = 0;
		fu(i) = 1;
		for (l = 1; l <= level; l++)
		  prol.ProlongateInline (l, vecu);
		au = mata * vecu;
		mu = matm * vecu;
		
		for (int j = 0; j < nc; j++)
		  {
		    fu = 0;
		    fu(j) = 1;
		    for (l = 1; l <= level; l++)
		      prol.ProlongateInline (l, vecu);
		    
		    matac2(i,j) = InnerProduct (au, vecu);
		    matmc2(i,j) = InnerProduct (mu, vecu);
		  }
	      }
	    (*testout) << "matac = " << endl << matac << endl;
	    (*testout) << "matac2 = " << endl << matac2 << endl;
	    (*testout) << "matmc = " << endl << matmc << endl;
	    (*testout) << "matmc2 = " << endl << matmc2 << endl;

	    for (int i = 0; i < nc; i++)
	      for (int j = 0; j < nc; j++)
		{
		  matac2(i,j) -= matac(i,j);
		  matmc2(i,j) -= matmc(i,j);
		}

	    (*testout) << "diffa = " << endl << matac2 << endl;
	    (*testout) << "diffm = " << endl << matmc2 << endl;

	  }
	*/


	
	// large space 
	for (int it = 0; it < maxsteps; it++)
	  {
	    cout << "it = " << it << " " << endl;
	    
	    // coarse grid matrix
	    for (i = 0; i < nc; i++)
	      for (j = 0; j < nc; j++)
		{
		  ha(i,j) = matac(i,j);
		  hm(i,j) = matmc(i,j);
		}
	    
	    
	    for (int ei = 0; ei < num; ei++)
	      {
		au = mata * *vecui[ei];
		mu = matm * *vecui[ei];
		
		double lam = 
		  InnerProduct (*vecui[ei], au) /
		  InnerProduct (*vecui[ei], mu);
		
		d = au - lam * mu;
		w = matpre * d;
		double norm = InnerProduct (w, mu=matm*w);
		*vecui[num+ei] = (1.0 / sqrt(norm)) * w;
	      }
	    
	    
	    for (int ei = 0; ei < 2*num; ei++)
	      {
		au = mata * *vecui[ei];
		mu = matm * *vecui[ei];
		
		for (int ej  = 0; ej < 2*num; ej++)
		  {
		    ha(nc+ei, nc+ej) = InnerProduct (au, *vecui[ej]);
		    hm(nc+ei, nc+ej) = InnerProduct (mu, *vecui[ej]);
		  }
		
		for (l = level; l >= 1; l--)
		  {
		    prol.RestrictInline (l, au);
		    prol.RestrictInline (l, mu);
		  }
		
		FlatVector<double> fau = au.FVDouble();
		FlatVector<double> fmu = mu.FVDouble();
		
		for (i = 0; i < nc; i++)
		  {
		    ha(i, nc+ei) = ha(nc+ei, i) = fau(i);
		    hm(i, nc+ei) = hm(nc+ei, i) = fmu(i);
		  }
	      }
	    
	    (*testout) << "ha = " << endl << ha << endl;
	    (*testout) << "hm = " << endl << hm << endl;
	    
	    // cout << "lapack ..." << flush;
	    LapackGHEPEPairs (nc+2*num, &ha(0,0), &hm(0,0), &lami(0));
	    // cout << " done " << flush;
	    
	    for (int ei = 0; ei < num; ei++)
	      cout << "lam(" << ei << ") = " << lami(ei) << endl;
	    
	    for (int ei = 0; ei < num; ei++)
	      {
		FlatVector<double> fau = au.FVDouble();
		for (i = 0; i < nc; i++)
		  fau(i) = ha(ei,i);
		
		if (nc > 0)
		  for (l = 1; l <= level; l++)
		    prol.ProlongateInline (l, au);
		else
		  au = 0;
		
		for (i = 0; i < 2*num; i++)
		  au += ha(ei,nc+i) * *vecui[i];
		
		*vecunew[ei] = au;
	      }
	    
	    for (int ei = 0; ei < num; ei++)
	      *vecui[ei] = *vecunew[ei];



	    /*
	    // coarse space, u+w

	    Matrix<double> ha1(nc+num, nc+num);
	    Matrix<double> hm1(nc+num, nc+num);
	    Matrix<double> ha2(2*num, 2*num);
	    Matrix<double> hm2(2*num, 2*num);

	    for (int it = 0; it < maxsteps; it++)
	    {
	    cout << "it = " << it << " ";

	    // coarse grid matrix
	    for (i = 0; i < nc; i++)
	    for (j = 0; j < nc; j++)
	    {
	    ha1(i,j) = matac(i,j);
	    hm1(i,j) = matmc(i,j);
	    }

	    for (int ei = 0; ei < num; ei++)
	    {
	    au = mata * *vecui[ei];
	    mu = matm * *vecui[ei];
	    
	    for (int ej  = 0; ej < num; ej++)
	    {
	    ha1(nc+ei, nc+ej) = InnerProduct (au, *vecui[ej]);
	    hm1(nc+ei, nc+ej) = InnerProduct (mu, *vecui[ej]);
	    }
	    
	    for (l = level; l >= 1; l--)
	    {
	    prol.RestrictInline (l, au);
	    prol.RestrictInline (l, mu);
	    }

	    FlatVector<double> fau = au.FVDouble();
	    FlatVector<double> fmu = mu.FVDouble();

	    for (i = 0; i < nc; i++)
	    {
	    ha1(i, nc+ei) = ha1(nc+ei, i) = fau(i);
	    hm1(i, nc+ei) = hm1(nc+ei, i) = fmu(i);
	    }
	    }

	    cout << "lapack ..." << flush;
	    LapackGHEPEPairs (nc+num, &ha1(0,0), &hm1(0,0), &lami(0));
	    cout << " done " << flush;
	
	    for (int ei = 0; ei < num; ei++)
	    cout << "lam(" << ei << ") = " << lami(ei) << endl;
	
	    for (int ei = 0; ei < num; ei++)
	    {
	    FlatVector<double> fau = au.FVDouble();
	    for (i = 0; i < nc; i++)
	    fau(i) = ha1(ei,i);

	    for (l = 1; l <= level; l++)
	    prol.ProlongateInline (l, au);

	    for (i = 0; i < num; i++)
	    au += ha1(ei,nc+i) * *vecui[i];

	    *vecunew[ei] = au;
	    }

	    for (int ei = 0; ei < num; ei++)
	    *vecui[ei] = *vecunew[ei];



	    // u+w
	    for (int ei = 0; ei < num; ei++)
	    {
	    au = mata * *vecui[ei];
	    mu = matm * *vecui[ei];

	    double lam = 
	    InnerProduct (*vecui[ei], au) /
	    InnerProduct (*vecui[ei], mu);
	      
	    d = au - lam * mu;
	    w = matpre * d;
	    *vecui[num+ei] = 1.0 / sqrt (InnerProduct (d,w)) * w;
	    }


	    for (int ei = 0; ei < 2*num; ei++)
	    {
	    au = mata * *vecui[ei];
	    mu = matm * *vecui[ei];
	    
	    for (int ej  = 0; ej < 2*num; ej++)
	    {
	    ha2(ei, ej) = InnerProduct (au, *vecui[ej]);
	    hm2(ei, ej) = InnerProduct (mu, *vecui[ej]);
	    }
	    }

	    (*testout) << "ha2 = " << endl << ha2 
	    << "hm2 = " << endl << hm2 << endl;
	    cout << "lapack ..." << flush;
	    LapackGHEPEPairs (2*num, &ha2(0,0), &hm2(0,0), &lami(0));
	    cout << " done " << flush;
	
	    for (int ei = 0; ei < num; ei++)
	    cout << "lam(" << ei << ") = " << lami(ei) << endl;
	
	    for (int ei = 0; ei < num; ei++)
	    {
	    FlatVector<double> fau = au.FVDouble();
	    for (i = 0; i < nc; i++)
	    fau(i) = ha2(ei,i);

	    au = 0;
	    for (i = 0; i < 2*num; i++)
	    au += ha2(ei,i) * *vecui[i];

	    *vecunew[ei] = au;
	    }

	    for (int ei = 0; ei < num; ei++)
	    *vecui[ei] = *vecunew[ei];
	    }
	    */




	    /*
	      Matrix<double> ha2(2,2);
	      Matrix<double> hm2(2,2);

	      for (int it = 0; it < maxsteps; it++)
	      {
	      cout << "it = " << it << " ";
	      au = mata * vecu;
	      mu = matm * vecu;

	      double uau = InnerProduct (au, vecu);
	      double umu = InnerProduct (mu, vecu);
	
	      double lam = uau / umu;

	      cout << "lam = " << lam << endl;
	      (*testout) << "lam = " << lam << endl;
	      (*testout) << ", uau = " << uau << ", umu = " << umu;

	      d = au - lam * mu;
	      w = matpre * d;

	      ha2(0,0) = uau;
	      hm2(0,0) = umu;

	      au = mata * w;
	      mu = matm * w;

	      ha2(1,0) = ha2(0,1) = InnerProduct (au, vecu);
	      hm2(1,0) = hm2(0,1) = InnerProduct (mu, vecu);
	      ha2(1,1) = InnerProduct (au, w);
	      hm2(1,1) = InnerProduct (mu, w);

	      (*testout) << ", waw = " << ha2(1,1) << ", wmw = " << hm2(1,1) << endl;

	      if (hm2(1,1) < 1e-18) break;

	      cout << "lapack ..." << flush;
	      LapackGHEPEPairs (2, &ha2(0,0), &hm2(0,0), &lami(0));
	      cout << " done " << flush;
	
	      vecu *= ha2(0,0);
	      vecu += ha2(0,1) * w;
	      }
	    */
	  }
      }
    else
      {
	int level = ma.GetNLevels()-1;
	const BilinearForm * bfac = &bfa->GetLowOrderBilinearForm();
	if (!bfac)
	  {
	    throw Exception ("EVP<Complex> needs high-order/low-order bilinear-forms");
	  }
	const BilinearForm * bfmc = &bfm->GetLowOrderBilinearForm();
	if (!bfmc)
	  {
	    throw Exception ("EVP<Complex> needs high-order/low-order bilinear-forms");
	  }

	const SparseMatrix<Complex> & mata = 
	  dynamic_cast<const SparseMatrix<Complex>&> (bfa->GetMatrix());
	const SparseMatrix<Complex> & matac = 
	  dynamic_cast<const SparseMatrix<Complex>&> (bfac->GetMatrix());
	const SparseMatrix<Complex> & matm = 
	  dynamic_cast<const SparseMatrix<Complex>&> (bfm->GetMatrix());
	const SparseMatrix<Complex> & matmc = 
	  dynamic_cast<const SparseMatrix<Complex>&> (bfmc->GetMatrix());
	
	
	const BaseMatrix & matpre = pre->GetMatrix();
	
	int nc = matac.Height();
	int n = mata.Height();

	cout << "n = " << n << ", nc = " << nc << endl;

	int i, j, k, l;
	
	BaseVector & vecu = *mata.CreateVector();
	BaseVector & d = *vecu.CreateVector();
	BaseVector & w = *vecu.CreateVector();
	BaseVector & au = *vecu.CreateVector();
	BaseVector & mu = *vecu.CreateVector();
	
	BaseVector & auc = *matac.CreateVector();
	BaseVector & muc = *matac.CreateVector();
	
	const Prolongation & prol = *bfa->GetFESpace().GetProlongation();
	
	cout.precision(16);
	
	ARRAY<BaseVector*> vecui(2*num);
	ARRAY<BaseVector*> vecunew(num);
	
	for (i = 0; i < 2*num; i++)
	  vecui[i] = &gfu->GetVector(i);
	    
	for (i = 0; i < num; i++)
	  vecunew[i] = vecu.CreateVector();

	maxsteps = 1;

	for (int it = 0; it < 1000; it++)
	  {
	    cout << "it = " << it << " " << endl;

	    int dimevp;
	    switch (it)
	      {
	      case 0: dimevp = nc; break;
	      case 1: dimevp = nc+num; break;
	      default: dimevp = nc+2*num; break;
	      }
	    
	    Matrix<Complex> ha(dimevp);
	    Matrix<Complex> hm(dimevp);
	    Matrix<Complex> evecs(dimevp);
	    Vector<Complex> lami(dimevp);

	    // coarse grid matrix
	    for (i = 0; i < nc; i++)
	      for (j = 0; j < nc; j++)
		{
		  ha(i,j) = matac(i,j);
		  hm(i,j) = matmc(i,j);
		}

	    if (it >= 1)
	      for (int ei = 0; ei < num; ei++)
		{
		  au = mata * *vecui[ei];
		  mu = matm * *vecui[ei];
		  
		  Complex lam = 
		    S_InnerProduct<Complex> (*vecui[ei], au) /
		    S_InnerProduct<Complex> (*vecui[ei], mu);
		  
		  (*testout) << "lam = " << lam << endl;
		  (*testout) << "au = " << endl << au << endl;
		  d = lam * mu;
		  (*testout) << "lam muu = " << d << endl;
		  
		  d = au - lam * mu;
		  w = matpre * d;
		  
		  double norm = abs (S_InnerProduct<Complex> (w, mu=matm*w));
		  *vecui[num+ei] = (1.0 / sqrt(norm)) * w;
		  
		  (*testout) << "d = " << d << endl;
		  (*testout) << "w = " << w << endl;
		  prol.RestrictInline (level, d);
		  (*testout) << "dc = " << d << endl;
		}
	    

	    if (it >= maxsteps) break;

	    int base, numadd;
	    switch (it)
	      {
	      case 0: base = 0; numadd = 0; break;
	      case 1: base = num; numadd = num; break;
	      default: base = 0; numadd = 2*num; break;
	      }

	    for (int ei = 0; ei < numadd; ei++)
	      {
		au = mata * *vecui[ei+base];
		mu = matm * *vecui[ei+base];
		
		for (int ej  = 0; ej < numadd; ej++)
		  {
		    ha(nc+ei, nc+ej) = S_InnerProduct<Complex> (au, *vecui[ej+base]);
		    hm(nc+ei, nc+ej) = S_InnerProduct<Complex> (mu, *vecui[ej+base]);
		  }
		
		FlatVector<Complex> fau = au.FVComplex();
		FlatVector<Complex> fmu = mu.FVComplex();
		
		for (i = 0; i < nc; i++)
		  {
		    ha(i, nc+ei) = ha(nc+ei, i) = fau(i);
		    hm(i, nc+ei) = hm(nc+ei, i) = fmu(i);
		  }
	      }
		
	    // (*testout) << "ha = " << endl << ha << endl;
	    // (*testout) << "hm = " << endl << hm << endl;
		
	    cout << "lapack ..." << flush;
	    // LapackGHEPEPairs (nc+2*num, &ha(0,0), &hm(0,0), &lami(0));
	    // LaEigNSSolve (mata.Height(), &mata(0,0), &matm(0,0), &lami(0), 1, &evecs(0,0), 0, 'N');
	    
	    LaEigNSSolve (dimevp, &ha(0,0), &hm(0,0), &lami(0), 1, &evecs(0,0), 0, 'N');
	    (*testout) << "lami = " << endl << lami << endl;
		
	    cout << " done " << flush;

	    ofstream out ("eigenvalues.dat");
	    for (int ei = 0; ei < lami.Size(); ei++)
	      out << ei << " " << lami(ei).real() << " " << lami(ei).imag() << endl;

	    for (int ei = 0; ei < num; ei++)
	      cout << "lam(" << ei << ") = " << lami(ei) << endl;


	    for (int ei = 0; ei < num; ei++)
	      {
		FlatVector<Complex> fau = au.FVComplex();
		fau = 0.0;
		for (i = 0; i < nc; i++)
		  fau(i) = evecs(ei, i);
		    
		for (i = 0; i < numadd; i++)
		  au += evecs(ei,nc+i) * *vecui[i+base];
		    
		*vecunew[ei] = au;
	      }

	    for (int ei = 0; ei < num; ei++)
	      *vecui[ei] = *vecunew[ei];
	  }



	cout << "start with Newton iteration" << endl;


	for (int evnum = 0; evnum < num; evnum++)
	  {
	    
	    Complex lam;
	    vecu = *vecui[evnum];

	    for (int i = 1; i <= maxnewton; i++)
	      {
		au = mata * vecu;
		mu = matm * vecu;
		
		lam = S_InnerProduct<Complex> (au, vecu) / S_InnerProduct<Complex> (mu, vecu);
		
		cout << "lam = " << lam << "    ";
		
		SparseMatrix<Complex> & hmat = 
		  dynamic_cast<SparseMatrix<Complex>&> (pde.GetBilinearForm("hmat")->GetMatrix());

		hmat.AsVector() = mata.AsVector() - lam * matm.AsVector();
		BaseMatrix & inv = *hmat.InverseMatrix();
		
		d = au - lam * mu;
		cout << "res = " << L2Norm(d) << endl;
		w = inv * mu;
	
		vecu = (1.0/L2Norm(w)) * w;
	      }

	    /*
	    for (int i = 1; i <= 5; i++)
	      {
		au = mata * vecu;
		mu = matm * vecu;
		
		lam = S_InnerProduct<Complex> (au, vecu) / S_InnerProduct<Complex> (mu, vecu);
		
		cout << "lam = " << lam << "    ";
		
		SparseMatrix<Complex> & hmat = 
		  dynamic_cast<SparseMatrix<Complex>&> (pde.GetBilinearForm("hmat")->GetMatrix());

		hmat.AsVector() = mata.AsVector() - lam * matm.AsVector();
		BaseMatrix & inv = *hmat.InverseMatrix();
		
		d = au - lam * mu;
		cout << "res = " << L2Norm(d) << endl;
		w = inv * d;
	
		Complex s = S_InnerProduct<Complex> (d, w) / 
		  (0.5 * S_InnerProduct<Complex> (vecu, mu) + S_InnerProduct<Complex> (w, mu));

		d -= s * mu;
		w = inv * d;

		vecu -= w;
	      }
	    */

	    *vecui[evnum] = vecu;
	  }
      }

  }



#ifdef abc

  ///
  class NumProcEVPJDQZ : public NumProc
  {
  protected:
    BilinearForm * bfa;
    BilinearForm * bfm;

    GridFunction * gfu;

  public:

    NumProcEVPJDQZ (PDE & apde, const Flags & flags);
    virtual ~NumProcEVPJDQZ();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcEVPJDQZ (pde, flags);
    }

    virtual void Do();

    virtual string GetClassName () const
    {
      return " Eigenvalue Problem (JDQZ)";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form A = " << bfa->GetName() << endl
	  << "Bilinear-form M = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl;
    }
  };


  NumProcEVPJDQZ :: NumProcEVPJDQZ (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
  }

  NumProcEVPJDQZ :: ~NumProcEVPJDQZ()
  {
    ;
  }


  void NumProcEVPJDQZ :: Do()
  {
    cout << "solve evp with JDQZ" << endl;

    int dim = bfa->GetFESpace().GetDimension();
    int size = bfa->GetMatrix().Height();
    bool iscomplex = bfa->GetFESpace().IsComplex();
    
    if (!iscomplex)
      {
	;
      }
    else
      {
	int n = dim*size;
	int kmax = 5;
	int jmax = 20;
	int jmin = 10;
	char wanted = 'v';
	Complex target = Complex (100, -20);
	double eps = 1e-6;
	int method = 2;
	int m = 30;
	int l = 2;
	int mxmv = 100;
	int maxstep = 1000;
	double lock = 1e-6;
	int order = 0;
	int testspace = 3;
	int lwork = 1000000;
	Vector<Complex> zwork(lwork);

	Vector<Complex> alpha(n), beta(n);
	Matrix<Complex> eivec(kmax, n);

	globmata = &bfa->GetMatrix();
	globmatb = &bfm->GetMatrix();
	globmatainv = dynamic_cast<BaseSparseMatrix&> (bfa->GetMatrix()).InverseMatrix();

	jdqz_ (&alpha(0), &beta(0), &eivec(0,0), &wanted, &n, 
	       &target, &eps, &kmax, &jmax, &jmin, &method, &m, &l,
	       &mxmv, &maxstep, &lock, &order, &testspace, &zwork(0), &lwork);

	cout << "evps = " << kmax << endl;
	for (int i = 0; i < kmax; i++)
	  {
	    cout << "lam(" << i << ") = " << alpha(i) << " / " << beta(i) << " = " 
		 << alpha(i) / beta(i) << endl;
	  }
	return;


	VVector<Complex> hv(dim*size), hva(dim*size), hvm(dim*size);
	Matrix<Complex> mata(dim*size), matm(dim*size);
	
	for (int i = 0; i < dim*size; i++)
	  {
	    hv = 0.0;
	    hv(i) = 1.0;
	    hva = bfa->GetMatrix() * hv;
	    hvm = bfm->GetMatrix() * hv;

	    for (int j = 0; j < dim*size; j++)
	      {
		mata(j,i) = hva(j);
		matm(j,i) = hvm(j);
	      }
	  }

	cout << "matrix copied" << endl;

	Vector<Complex> lami(dim*size);
	Matrix<Complex> evecs(dim*size);

	LaEigNSSolve (mata.Height(), &mata(0,0), &matm(0,0), &lami(0), 1, &evecs(0,0), 0, 'N');

	cout << "eigensystem done" << endl;


	// sort by angle
	BitArray used(n);
	ARRAY<int> reorder(n);
	used.Clear();
	for (int i = 0; i < n; i++)
	  {
	    int mini = -1;
	    double val = 0;
	    for (int j = 0; j < n; j++)
	      if (!used[j])
		{
		  if (mini < 0 ||
		      lami(j).imag() / lami(j).real() > val)
		    {
		      mini = j;
		      val = lami(j).imag() / lami(j).real();
		    }
		}
	    reorder[i] = mini;
	    used.Set(mini);
	  }

	cout << "lami = " << endl << lami << endl;
	cout << "reorder = " << reorder << endl;
	ofstream out ("eigenvalues.dat");
	for (int i = 0; i < lami.Size(); i++)
	  out << i << " " << lami(reorder[i]).real() << " " << lami(reorder[i]).imag() << endl;
	
	/*
	(*testout) << "mata = " << bfa->GetMatrix() << endl
		   << " = " << endl << mata << endl;
	*/

	for (int i = 0; i < min2 (gfu->GetMultiDim(), mata.Height()); i++)
	  {
	    FlatVector<Complex> vu = gfu->GetVector(i).FVComplex();
	    for (int j = 0; j < dim*size; j++)
	      vu(j) = evecs(reorder[i],j);
	  }
      }
    
  }
#endif


  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("evp", NumProcEVPCoarse::Create);
      // GetNumProcs().AddNumProc ("evpjdqz", NumProcEVPJDQZ::Create);
    }

    
    Init init;
  }

}
