#include "../include/solve.hpp"

namespace ngsolve
{
  using namespace ngsolve;
  // using namespace ngbla;
  // using namespace ngla;











  class NumProcGoalErrorEstimator : public NumProc
  {
  private:
    BilinearForm * bfa;
    GridFunction * gfu;
    GridFunction * gfz;
    GridFunction * gferr;
    GridFunction * gferr1;
    GridFunction * gferr2;
    GridFunction * gferr3;
    GridFunction * gferr4;

  public:
    NumProcGoalErrorEstimator (PDE & apde, const Flags & flags);
    virtual ~NumProcGoalErrorEstimator();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcGoalErrorEstimator (pde, flags);
    }

    virtual void Do(LocalHeap & lh);

    virtual string GetClassName () const
    {
      return "Goal-driven Error Estimator";
    }
  };



  NumProcGoalErrorEstimator :: NumProcGoalErrorEstimator (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("primalsolution", ""));
    gfz = pde.GetGridFunction (flags.GetStringFlag ("dualsolution", ""));
    gferr = pde.GetGridFunction (flags.GetStringFlag ("error", ""));
    gferr1 = pde.GetGridFunction (flags.GetStringFlag ("error1", ""));
    gferr2 = pde.GetGridFunction (flags.GetStringFlag ("error2", ""));
    gferr3 = pde.GetGridFunction (flags.GetStringFlag ("error3", ""));
    gferr4 = pde.GetGridFunction (flags.GetStringFlag ("error4", ""));
  }

  NumProcGoalErrorEstimator :: ~NumProcGoalErrorEstimator()
  {
    ;
  }
  
  void NumProcGoalErrorEstimator :: Do(LocalHeap & lh)
  {
    cout << "Goal error-estimator" << endl;

    BilinearFormIntegrator & bfi_curl = *bfa->GetIntegrator(0);
    BilinearFormIntegrator & bfi_id = *bfa->GetIntegrator(1);

    /*
    Flags fesflags;
    fesflags.SetFlag ("order", 1);
    fesflags.SetFlag ("dim", 3);
    if (bfa->GetFESpace().IsComplex())
      fesflags.SetFlag ("complex");

    H1HighOrderFESpace & fesflux = *new H1HighOrderFESpace (ma, fesflags);
    */


    Flags fesflags;
    fesflags.SetFlag ("order", 1);
    if (bfa->GetFESpace().IsComplex())
      fesflags.SetFlag ("complex");

    HCurlHighOrderFESpace & fesflux = *new HCurlHighOrderFESpace (ma, fesflags);





    fesflux.Update();


    Flags flags;
    //    flags.SetFlag ("novisual");
    GridFunction * smooth_curlu = CreateGridFunction (&fesflux, "smooth_curlu", flags);
    smooth_curlu->Update();
    GridFunction * smooth_u = CreateGridFunction (&fesflux, "smooth_u", flags);
    smooth_u->Update();
    GridFunction * smooth_curlz = CreateGridFunction (&fesflux, "smooth_curlz", flags);
    smooth_curlz->Update();
    GridFunction * smooth_z = CreateGridFunction (&fesflux, "smooth_z", flags);
    smooth_z->Update();




    FlatVector<double> err1 = 
      dynamic_cast<T_BaseVector<double>&> (gferr1->GetVector()).FV();
    err1 = 0;

    FlatVector<double> err2 = 
      dynamic_cast<T_BaseVector<double>&> (gferr2->GetVector()).FV();
    err2 = 0;

    FlatVector<double> err3 = 
      dynamic_cast<T_BaseVector<double>&> (gferr3->GetVector()).FV();
    err3 = 0;

    FlatVector<double> err4 = 
      dynamic_cast<T_BaseVector<double>&> (gferr4->GetVector()).FV();
    err4 = 0;

    FlatVector<double> err = 
      dynamic_cast<T_BaseVector<double>&> (gferr->GetVector()).FV();
    err = 0;

    double sum = 0;
    double sum1 = 0;
    double sum2 = 0;
    

    int ndom = ma.GetNDomains();
    for (int k = 0; k < ndom; k++)
      {
	err1 = 0;
	err2 = 0;
	err3 = 0;
	err4 = 0;

	if (!bfa->GetFESpace().IsComplex())
	  {
	    // computed smoothed curl u
	    CalcFluxProject (ma, 
			     dynamic_cast<const S_GridFunction<double>&> (*gfu), 
			     dynamic_cast<S_GridFunction<double>&> (*smooth_curlu), 
			     bfi_curl,   // computes curl u
			     1, k, lh);
	
	    CalcError (ma, 
		       dynamic_cast<const S_GridFunction<double>&> (*gfu), 
		       dynamic_cast<const S_GridFunction<double>&> (*smooth_curlu), 
		       bfi_curl,
		       err1, k, lh);



	    // computed smoothed curl z
	    CalcFluxProject (ma, 
			     dynamic_cast<const S_GridFunction<double>&> (*gfz), 
			     dynamic_cast<S_GridFunction<double>&> (*smooth_curlz), 
			     bfi_curl,   // computes curl u
			     1, k, lh);
	
	    CalcError (ma, 
		       dynamic_cast<const S_GridFunction<double>&> (*gfz), 
		       dynamic_cast<const S_GridFunction<double>&> (*smooth_curlz), 
		       bfi_curl,
		       err2, k, lh);



	    /*
	    // computed smoothed u
	    CalcFluxProject (ma, 
			     dynamic_cast<const S_GridFunction<double>&> (*gfu), 
			     dynamic_cast<S_GridFunction<double>&> (*smooth_u), 
			     bfi_id,   // computes curl u
			     1, k, lh);
	
	    CalcError (ma, 
		       dynamic_cast<const S_GridFunction<double>&> (*gfu), 
		       dynamic_cast<const S_GridFunction<double>&> (*smooth_u), 
		       bfi_id,
		       err3, k, lh);
	    */

	    /*
	    // computed smoothed curl u
	    CalcFluxProject (ma, 
			     dynamic_cast<const S_GridFunction<double>&> (*gfz), 
			     dynamic_cast<S_GridFunction<double>&> (*smooth_z), 
			     bfi_id,   // computes curl u
			     1, k, lh);
	
	    CalcError (ma, 
		       dynamic_cast<const S_GridFunction<double>&> (*gfz), 
		       dynamic_cast<const S_GridFunction<double>&> (*smooth_z), 
		       bfi_id,
		       err4, k, lh);
	    */
	  }


	for (int i = 0; i < err.Size(); i++)
	  {
	    err(i) += sqrt (err1(i)*err2(i)) + sqrt(err3(i)*err4(i));
	    sum +=  sqrt (err1(i)*err2(i)) + sqrt(err3(i)*err4(i));
	    sum1 += sqrt (err1(i)*err2(i));
	    sum2 += sqrt (err3(i)*err4(i));
	  }
      }
    cout << " estimated error = " << sum << ", e1 = " << sum1 << ", e2 = " << sum2 << endl;

    static ofstream out("goalee.out");


    double fu =
      InnerProduct (pde.GetLinearForm ("b")->GetVector(), gfu->GetVector());

    out << gfu->GetFESpace().GetNDof() 
	<< " \t"  << fabs (fu - 0.0089822827254)    // 0.0054613529)  // 0.0105806306
	<< " \t" << sum 
	<< " \t" << sum1 
	<< " \t" << sum2 << endl;
  }




  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("goaldrivenerrorestimator", NumProcGoalErrorEstimator::Create, NumProcGoalErrorEstimator::PrintDoc);
    }

    Init init;
      
  }
}
