#include <solve.hpp>
// #include <nginterface.h>


namespace ngsolve
{
  using namespace ngsolve;



  ///
  class NumProcMaxwellEVP : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    /// 
    BilinearForm * bfm;
    ///
    GridFunction * gfu;
    ///
    Preconditioner * pre;
    ///
    Preconditioner * prepot;
    // 
    string filename1, filename2;



    ///
    int maxsteps;
    int proj_steps;
    int nev; 
    ///
    double prec;
    ///
    bool print;
    bool lshape; 
  public:
    ///
    NumProcMaxwellEVP (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcMaxwellEVP();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcMaxwellEVP (pde, flags);
    }

    ///
    virtual void Do(LocalHeap & lh);
    ///
    virtual string GetClassName () const
    {
      return "Maxwell Eigenvalue Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form A = " << bfa->GetName() << endl
	  << "Bilinear-form M = " << bfm->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl
	  << "Preconditioner = " << ((pre) ? pre->ClassName() : "None") << endl;
    }
  };



  NumProcMaxwellEVP :: NumProcMaxwellEVP (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearforma", ""));
    bfm = pde.GetBilinearForm (flags.GetStringFlag ("bilinearformm", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    pre = pde.GetPreconditioner (flags.GetStringFlag ("preconditioner", ""));
    prepot = pde.GetPreconditioner (flags.GetStringFlag ("preconditionerpot", ""));
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
    proj_steps = int(flags.GetNumFlag ("projsteps", 5));
    nev =  gfu->GetMultiDim(); 
    lshape = flags.GetDefineFlag("lshape"); 
    filename1 = flags.GetStringFlag ("filename","eval.out");
    filename2 = flags.GetStringFlag ("filename2","eval2.out");
    
  }

  NumProcMaxwellEVP :: ~NumProcMaxwellEVP()
  {
    ;
  }

  void NumProcMaxwellEVP :: Do(LocalHeap & lh)
  {
    cout << "solve Maxwell evp" << endl;

    const BaseMatrix & mata = bfa->GetMatrix();
    const BaseMatrix & matm = bfm->GetMatrix();
    BaseVector & vu = gfu->GetVector();
    
    const BaseMatrix & matpre = pre->GetMatrix();
    const BaseMatrix & matprepot = prepot->GetMatrix();
    
    SparseMatrix<double> * grad;
    clock_t start = clock(); 

    cout << " vor grad " << endl; 
    if (dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace()))
      grad = dynamic_cast<const HCurlHighOrderFESpace*> (&gfu->GetFESpace())->CreateGradient();

    if (dynamic_cast<const NedelecFESpace*> (&gfu->GetFESpace()))
      grad = dynamic_cast<const NedelecFESpace*> (&gfu->GetFESpace())->CreateGradient();

    double thistime = (clock() - start)/double(CLOCKS_PER_SEC); 
    (*testout) << " Time for computing grad     " << thistime << endl;
    (cout) << " Time for computing grad     " << thistime << endl;
    start = clock(); 

    BaseVector & vw = *mata.CreateVector();
    BaseVector & vd = *mata.CreateVector();
    BaseVector & vv = *mata.CreateVector();
    BaseVector & vau = *mata.CreateVector();
    BaseVector & vmu = *mata.CreateVector();

    BaseVector & vwpot = *matprepot.CreateVector();
    BaseVector & vdpot = *matprepot.CreateVector();

    double lam;
    double lam_old = 0.; 

    //vv = 1;
    
    ofstream evout (filename1.c_str());
    ofstream evout2 (filename2.c_str());
    double solev[10]; 
    solev[0]= 0.963972384472E+01;
    solev[1]= 0.113452262252E+02;
    solev[2]= 0.134036357679E+02; 
    solev[3]= 0.151972519265E+02; 
    solev[4]= 0.195093282458E+02;
    solev[5]= 0.197392088022E+02;
    solev[6]= 0.197392088022E+02;
    solev[7]= 0.197392088022E+02;
    solev[9]= 0.212590837990E+02; 
    
    double compev[nev]; 
    double sum_old=100; 
    
    for(int m=0;m<nev;m++)
      {
	vw.SetRandom(); 
       
	gfu->GetVector(m)=matpre*vw; 
      }
    cout.precision(12);
    (*testout).precision(12); 
    
    //vu = mata * vv;
    
    Vector<double> Lami(nev); 
    Lami=1.; 
    int i;

    Matrix<double> asmall(nev); 
    Matrix<double> msmall(nev); 
    
    ARRAY<BaseVector*> u1(nev); 
    for(int m=0;m<nev; m++)
      u1[m]=mata.CreateVector();
    
    for (i = 1; i <= maxsteps; i++) // maxsteps 
      {
	cout << "i = " << i << flush;
	
	for(int m=0;m<nev;m++)
	  {
	    BaseVector & vu = gfu->GetVector(m); 
	    vmu = matm * gfu->GetVector(m); 	  		
	    vau = mata * gfu->GetVector(m);
	    
	    vd = vau - Lami(m) * vmu;
	    vw = matpre * vd;
	    vu -= vw;
	    
	    *testout << "m = " << m 
		     << " before: " << sqrt(InnerProduct(vmu = matm*vu,vu)); 

	    // projection
	    for (int j = 1; j <= proj_steps; j++)
	      {
		vd = matm * vu;
		vdpot = Transpose (*grad) * vd;
		vwpot = matprepot * vdpot;
		vw = (*grad) * vwpot; 
		vu -= vw;
	      }

	    *testout << " after: " << sqrt(InnerProduct(vmu = matm*vu,vu)) << endl;
	  }
	// X^T A X y = lami X^t M X y 
	for(int j=0;j<nev;j++)
	  {
	    vmu = matm * gfu->GetVector(j); 	  		
	    vau = mata * gfu->GetVector(j);
	    for(int k=0;k<nev;k++) 
	      {
		asmall(j,k) = InnerProduct(vau,gfu->GetVector(k)); 
		msmall(j,k) = InnerProduct(vmu,gfu->GetVector(k)); 
	      }
	  }

	LapackGHEPEPairs(nev, &asmall(0,0), &msmall(0,0), &Lami(0));

	for(int m=0;m<nev;m++)
	  *u1[m]=gfu->GetVector(m);
	
	for(int m=0;m<nev;m++)
	  {
	    BaseVector & vu = gfu->GetVector(m); 
	    vu =0.; 
	    for(int k=0;k<nev;k++) 
	      vu += asmall(m,k) * (*u1[k]);   
	  }

	
	if(lshape) 
	  {
	    double sum=0.;
	    cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
	    for (int m=0;m<3; m++)
	      {
		cout << "lshape-error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
		
		sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
	      }	
	   
	    if(fabs(sum-sum_old)/fabs(sum)/3. < 1.e-2) i=maxsteps+1;  
	    sum_old=sum;
	    cout << i <<  " sum " << sum_old/3.<< endl; 
	  }
	else
	  cout << "cube-err: " << Lami(0)-2 << endl;
	  
	  
	// if(fabs(lam_old-lam) < fabs(lam)*1.E-12 || i == maxsteps) break;
	//	if(lshape && (fabs(lam_old-lam)/fabs(lam) < 1.e-4*fabs(solev[0]-lam_old)/fabs(solev[0])))  break; 
	// lam_old = lam; 
	

	
	
	//double y = sqrt(InnerProduct(vmu = matm*vu,vu)); 
	// vu /= sqrt(msmall(m,m));
	    

      }

    if(lshape) 
      {
	double sum=0.;
	cout << " ndofs " << gfu->GetFESpace().GetNDof() << endl; 
	for (int m=0;m<3; m++)
	  {
	    cout << "lshape-error[m] = "<< fabs((solev[m] - Lami(m))/solev[m]) << "\t" << flush;
	    
	    evout <<  gfu->GetFESpace().GetNDof() << "\t" << fabs((solev[m] - Lami(m))/solev[m]) << endl; 
	    sum +=  fabs((solev[m] - Lami(m))/solev[m]); 
	  }	
	    
	 evout2 << gfu->GetFESpace().GetNDof()  << "\t" << sum/3. << endl; 	       
	 cout << "ndofs " << gfu->GetFESpace().GetNDof()  << "\t sum/nev " << sum/3 << endl; 	       
	

      }
    

    (*testout) << " ********************************************************* " << endl; 


  }


  namespace {
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetNumProcs().AddNumProc ("maxwellevp", NumProcMaxwellEVP::Create);
    }

    
    Init init;
  }
}
