/*

Mixed elasticity

*/


#include <solve.hpp>


using namespace ngfem;
using namespace ngsolve;


/* ********************* FESpace **************************** */




#include "mixedmech.hpp"








FESpaceHDivSym :: FESpaceHDivSym (const MeshAccess & ama, const Flags & aflags)
  : FESpace (ama, aflags)
{
  cout << "**********************+ HDivSym, order = " << order << endl;
  cout << "flags = " << aflags << endl;

  order_facet = int (aflags.GetNumFlag ("facetorder", order));  
  
<<<<<<< mixedmech.cpp
  
  noncontinuous = aflags.GetDefineFlag ("noncontinuous");

=======

  noncontinuous = aflags.GetDefineFlag ("noncontinuous");
  
>>>>>>> 1.8

  if (ma.GetDimension() == 2)
    {
      if (noncontinuous)
	segm = new HDivSymNormalSegm (-1);
      else
	segm = new HDivSymNormalSegm (order_facet);

      trig = new HDivSymTrig (order, order_facet);
      quad = new HDivSymQuad (order, order_facet);
    }
  else
    {
      if (noncontinuous)
	trig = new HDivSymNormalTrig (-1);
      else
	trig = new HDivSymNormalTrig (order_facet);

      quad = new HDivSymNormalQuad (order_facet);
      prism = new HDivSymPrism (order, order_facet);
      tet = new HDivSymTet (order, order_facet);
      hex = new HDivSymHex (order, order_facet);
    }
}

FESpaceHDivSym :: ~FESpaceHDivSym ()
{
  ;
}


void FESpaceHDivSym :: Update()
{
  nel = ma.GetNE();
  ndof = 0;

  if (ma.GetDimension() == 2)
    {
      nfa = ma.GetNEdges();
      first_face_dof.SetSize(nfa+1);
      for (int i = 0; i < nfa; i++)
	{
	  first_face_dof[i] = ndof;
	  ndof += order_facet+1 ; //  + 1;
	}
      first_face_dof[nfa] = ndof;    
    }
  else
    {
      nfa = ma.GetNFaces();
      first_face_dof.SetSize(nfa+1);
      ARRAY<int> pnums;

      for (int i = 0; i < nfa; i++)
	{
	  first_face_dof[i] = ndof;

	  ma.GetFacePNums(i, pnums);

	  if (pnums.Size() == 3)
	    ndof += (order_facet+1)*(order_facet+2)/2;
	  else
	    ndof += sqr (order_facet+1);
	}
      first_face_dof[nfa] = ndof;    
    }

  first_element_dof.SetSize(nel+1);
  for (int i = 0; i < nel; i++)
    {
      first_element_dof[i] = ndof;
      switch (ma.GetElType(i))
	{
	case ET_TRIG:
	  ndof += dynamic_cast<HDivSymFiniteElement<2>*> (trig) -> GetNInner();
	  break;
	case ET_QUAD:
	  ndof += dynamic_cast<HDivSymFiniteElement<2>*> (quad) -> GetNInner();
	  break;
	case ET_PRISM:
	  ndof += dynamic_cast<HDivSymFiniteElement<3>*> (prism) -> GetNInner();
	  break;
	case ET_TET:
	  ndof += dynamic_cast<HDivSymFiniteElement<3>*> (tet) -> GetNInner();
	  break;
	case ET_HEX:
	  ndof += dynamic_cast<HDivSymFiniteElement<3>*> (hex) -> GetNInner();
	  break;
	}
    }

  first_element_dof[nel] = ndof;    


  if (noncontinuous)
    {
      first_face_dof = 0;
      ndof = 0;

      for (int i = 0; i < nel; i++)
	{
	  first_element_dof[i] = ndof;
	  switch (ma.GetElType(i))
	    {
	    case ET_TRIG:
	      ndof += trig -> GetNDof();
	      break;
	    case ET_TET:
	      ndof += tet -> GetNDof();
	      break;
	    }
	}
      first_element_dof[nel] = ndof;    
    }

  *testout << "HDiv-Sym" << endl;
  *testout << "first face dofs = " << endl << first_face_dof << endl;
  *testout << "first element dofs = " << endl << first_element_dof << endl;
}


const FiniteElement & FESpaceHDivSym :: GetFE (int elnr, LocalHeap & lh) const
{
  HDivSymFiniteElement<2> * fe2d = 0;
  HDivSymFiniteElement<3> * fe3d = 0;

  if (ma.GetElType(elnr) == ET_TRIG)
    fe2d = dynamic_cast<HDivSymFiniteElement<2>*> (trig);
  else if (ma.GetElType(elnr) == ET_QUAD)
    fe2d = dynamic_cast<HDivSymFiniteElement<2>*> (quad);
  else if (ma.GetElType(elnr) == ET_TET)
    fe3d = dynamic_cast<HDivSymFiniteElement<3>*> (tet);
  else if (ma.GetElType(elnr) == ET_PRISM)
    fe3d = dynamic_cast<HDivSymFiniteElement<3>*> (prism);
  else if (ma.GetElType(elnr) == ET_HEX)
    fe3d = dynamic_cast<HDivSymFiniteElement<3>*> (hex);


  ArrayMem<int,12> vnums; 
  ma.GetElVertices(elnr, vnums);
  
  if (fe2d)
    {
      fe2d -> SetVertexNumbers (vnums);
      return *fe2d;
    }
  if (fe3d)
    {
      fe3d -> SetVertexNumbers (vnums);
      return *fe3d;
    }
  
}

const FiniteElement & FESpaceHDivSym :: GetSFE (int elnr, LocalHeap & lh) const
{
  HDivSymNormalFiniteElement * fe;

  if (ma.GetSElType(elnr) == ET_SEGM)
    fe = dynamic_cast<HDivSymNormalFiniteElement*> (segm);
  else if (ma.GetSElType(elnr) == ET_TRIG)
    fe = dynamic_cast<HDivSymNormalFiniteElement*> (trig);
  else if (ma.GetSElType(elnr) == ET_QUAD)
    fe = dynamic_cast<HDivSymNormalFiniteElement*> (quad);


  ArrayMem<int,12> vnums; 
  ma.GetSElVertices(elnr, vnums);
  fe -> SetVertexNumbers (vnums);

  return *fe;
}



void FESpaceHDivSym :: GetDofNrs (int elnr, ARRAY<int> & dnums) const
{   
  dnums.SetSize(0);

  ArrayMem<int,12> vnums, ednums, fanums; 
  int i, j;
  int first, next; 

  if (ma.GetDimension() == 2)
    ma.GetElEdges (elnr, ednums);
  else
    ma.GetElFaces (elnr, ednums);

  for (i = 0; i < ednums.Size(); i++)
    {
      first = first_face_dof[ednums[i]];
      next = first_face_dof[ednums[i]+1];
      for (j = first; j < next; j++)
	dnums.Append (j);
    }

  first = first_element_dof[elnr];
  next = first_element_dof[elnr+1];
  for (int j = first; j < next; j++)
    dnums.Append (j);
}



  void FESpaceHDivSym :: GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const
  {
    if (!eliminate_internal) 
      {
	GetDofNrs (elnr, dnums);
	return;
      }

    dnums.SetSize(0);
    
    ArrayMem<int,12> vnums, ednums, fanums; 
    int i, j;
    int first, next; 
    
    if (ma.GetDimension() == 2)
      ma.GetElEdges (elnr, ednums);
    else
      ma.GetElFaces (elnr, ednums);
    
    for (i = 0; i < ednums.Size(); i++)
      {
	first = first_face_dof[ednums[i]];
	next = first_face_dof[ednums[i]+1];
	for (j = first; j < next; j++)
	  dnums.Append (j);
      }
  }
 









void FESpaceHDivSym :: GetSDofNrs (int elnr, ARRAY<int> & dnums) const
{
  dnums.SetSize (0);

    
  ArrayMem<int,12> vnums, ednums, fanums; 
  int i, j;
  int first, next; 

  if (ma.GetDimension() == 2)
    ma.GetSElEdges (elnr, ednums);
  else
    {
      ednums.SetSize(1);
      ednums[0] = ma.GetSElFace (elnr);
    }

    
  for (i = 0; i < ednums.Size(); i++)
    {
      first = first_face_dof[ednums[i]];
      next = first_face_dof[ednums[i]+1];
      for (j = first; j < next; j++)
	dnums.Append (j);
    }
}


























class FESpaceMixedMech : public CompoundFESpace
{
public:
  FESpaceMixedMech (const MeshAccess & ama, 		   
		    const ARRAY<const FESpace*> & aspaces,
		    const Flags & aflags)
    : CompoundFESpace (ama, aspaces, aflags)
  {
    ;
  }

  virtual ~FESpaceMixedMech () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    ARRAY<const FESpace*> spaces(2);

    cout << "Create MixedMech FE Space" << endl;
    
    int order = int (flags.GetNumFlag ("order", 1.0));

    Flags sigmaflags;
    sigmaflags.SetFlag ("order", flags.GetNumFlag ("order", 1.0));
    sigmaflags.SetFlag ("facetorder", flags.GetNumFlag ("facetorder", order));
    if (flags.GetDefineFlag ("eliminate_internal")) sigmaflags.SetFlag ("eliminate_internal");
    spaces[0] = new FESpaceHDivSym (ma, sigmaflags);
    
    Flags uflags;
    int curlorder = int (flags.GetNumFlag ("curlorder", order));
    uflags.SetFlag ("order", curlorder);
    uflags.SetFlag ("orderinner", flags.GetNumFlag ("orderinner", curlorder));
    uflags.SetFlag ("orderface", flags.GetNumFlag ("orderface", curlorder));
    uflags.SetFlag ("orderedge", flags.GetNumFlag ("orderedge", curlorder));


    ARRAY<double> maxz(ma.GetNDomains());
    maxz = 1;
    uflags.SetFlag ("dom_order_min_z", maxz);
    uflags.SetFlag ("dom_order_max_z", maxz);
    uflags.SetFlag ("variableorder");

    if (flags.GetDefineFlag ("eliminate_internal")) uflags.SetFlag ("eliminate_internal");
    if (flags.GetDefineFlag ("alt"))
      spaces[1] = new HCurlHighOrderFESpaceAlt (ma, uflags);
    else
      spaces[1] = new HCurlHighOrderFESpace (ma, uflags);

    
    FESpaceMixedMech * fes = new FESpaceMixedMech (ma, spaces, flags);
    return fes;
  }
    
  virtual string GetClassName () const { return "MixedMech-FESpace"; }
};





















/* ******************** Integrator ************************** */










template <int D>
class HDivSymMassIntegrator 
  : public T_BDBIntegrator<DiffOpIdHDivSym<D>, DiagDMat<DiffOpIdHDivSym<D>::DIM_DMAT>, FiniteElement>
{
public:
  HDivSymMassIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpIdHDivSym<D>, DiagDMat<DiffOpIdHDivSym<D>::DIM_DMAT>, FiniteElement>
  (DiagDMat<DiffOpIdHDivSym<D>::DIM_DMAT> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymMassIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-Mass"; }
};


template <int D> class HDivSymSourceIntegrator { ; };



template <>
class HDivSymSourceIntegrator<2> 
  : public T_BIntegrator<DiffOpIdHDivSym<2>, DVec<3>, FiniteElement>
{
public:
  HDivSymSourceIntegrator (CoefficientFunction * coeff1,
			   CoefficientFunction * coeff2,
			   CoefficientFunction * coeff3)
    :  T_BIntegrator<DiffOpIdHDivSym<2>, DVec<3>, FiniteElement>
  (DVec<3> (coeff1, coeff2, coeff3))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymSourceIntegrator (coeffs[0], coeffs[1], coeffs[2]);
  }

  virtual string Name () const { return "HDivSym-Source"; }
};



template <>
class HDivSymSourceIntegrator<3> 
  : public T_BIntegrator<DiffOpIdHDivSym<3>, DVec<6>, FiniteElement>
{
public:
  HDivSymSourceIntegrator (CoefficientFunction * coeff1,
			   CoefficientFunction * coeff2,
			   CoefficientFunction * coeff3,
			   CoefficientFunction * coeff4,
			   CoefficientFunction * coeff5,
			   CoefficientFunction * coeff6)
    :  T_BIntegrator<DiffOpIdHDivSym<3>, DVec<6>, FiniteElement>
  (DVec<6> (coeff1, coeff2, coeff3, coeff4, coeff5, coeff6)) 
  { ; }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymSourceIntegrator (coeffs[0], coeffs[1], coeffs[2],
					coeffs[3], coeffs[4], coeffs[5]);
  }

  virtual string Name () const { return "HDivSym-Source"; }
};




template <int D>
class HDivSymRobinIntegrator 
  : public T_BDBIntegrator<DiffOpBoundIdHDivSym<D>, DiagDMat<1>, FiniteElement>
{
public:
  HDivSymRobinIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpBoundIdHDivSym<D>, DiagDMat<1>, FiniteElement>
  (DiagDMat<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymRobinIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-Robin"; }
  virtual bool BoundaryForm () const { return 1; }
};



template <int D>
class HDivSymNeumannIntegrator 
  : public T_BIntegrator<DiffOpBoundIdHDivSym<D>, DVec<1>, FiniteElement>
{
public:
  HDivSymNeumannIntegrator (CoefficientFunction * coef)
    :  T_BIntegrator<DiffOpBoundIdHDivSym<D>, DVec<1>, FiniteElement>
  (DVec<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymNeumannIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-Neumann"; }
  virtual bool BoundaryForm () const { return 1; }
};







template <int D>
class HDivSymDivDivIntegrator
  : public T_BDBIntegrator<DiffOpDivHDivSym<D>, DiagDMat<D>, FiniteElement>
{
public:
  HDivSymDivDivIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpDivHDivSym<D>, DiagDMat<D>, FiniteElement>
  (DiagDMat<D> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivDivIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-DivDiv"; }
};


class HDivSymDivIntegrator 
  : public T_BIntegrator<DiffOpDivHDivSym<2>, DVec<2>, FiniteElement>
{
public:
  HDivSymDivIntegrator (CoefficientFunction * coeff1,
			CoefficientFunction * coeff2)
    :  T_BIntegrator<DiffOpDivHDivSym<2>, DVec<2>, FiniteElement>
  (DVec<2> (coeff1, coeff2))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivIntegrator (coeffs[0], coeffs[1]);
  }

  virtual string Name () const { return "HDivSym-Div"; }
};









template <int D>
class HDivSymDivDivDivDivIntegrator
  : public T_BDBIntegrator<DiffOpDivDivHDivSym<D>, DiagDMat<1>, FiniteElement>
{
public:
  HDivSymDivDivDivDivIntegrator (CoefficientFunction * coef)
    :  T_BDBIntegrator<DiffOpDivDivHDivSym<D>, DiagDMat<1>, FiniteElement>
  (DiagDMat<1> (coef))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new HDivSymDivDivDivDivIntegrator (coeffs[0]);
  }

  virtual string Name () const { return "HDivSym-DivDivDivDiv"; }
};









template <int D>
class DiffOpMixedMech { };


template <>
class DiffOpMixedMech<2> : public DiffOp<DiffOpMixedMech<2> >
{

  // (stress, div stress, u)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 7 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<2> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    mat = 0.0;

    Mat<2> jac = sip.GetJacobian();
    Mat<2> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());

    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

    FlatMatrix<> div_sigma_bmat(2, nd_sigma, lh);
    DiffOpDivHDivSym<2>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);

    for (int i = 0; i < nd_sigma; i++)
      {
	mat(0,i) = sigma_bmat(0,i);
	mat(1,i) = sigma_bmat(1,i);
	mat(2,i) = sigma_bmat(2,i);
	mat(3,i) = div_sigma_bmat(0,i);
	mat(4,i) = div_sigma_bmat(1,i);
      }

    FlatMatrix<> u_ref(nd_u, 2, lh);
    FlatMatrixFixWidth<2> u(nd_u, lh);

    fel_u.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;

    for (int i = 0; i < nd_u; i++)
      {
	mat(5,i+nd_sigma) = u(i,0);
	mat(6,i+nd_sigma) = u(i,1);
      }
  }
};





template <>
class DiffOpMixedMech<3> : public DiffOp<DiffOpMixedMech<3> >
{

  // (stress, div stress, u)

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 3 };
  enum { DIM_ELEMENT = 3 };
  enum { DIM_DMAT = 12 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<3> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (cfel[0]);
    const HCurlFiniteElementD<3> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    mat = 0.0;


    Mat<3> jac = sip.GetJacobian();
    Mat<3> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());



    FlatMatrix<> sigma_bmat(6, nd_sigma, lh);
    DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

    FlatMatrix<> div_sigma_bmat(3, nd_sigma, lh);
    DiffOpDivHDivSym<3>::GenerateMatrix (fel_sigma, sip, div_sigma_bmat, lh);

    for (int i = 0; i < nd_sigma; i++)
      {
	for (int j = 0; j < 6; j++)
	  mat(j,i) = sigma_bmat(j,i);
	for (int j = 0; j < 3; j++)
	  mat(6+j,i) = div_sigma_bmat(j,i);
      }


    FlatMatrix<> u_ref(nd_u, 3, lh);
    FlatMatrixFixWidth<3> u(nd_u, lh);

    fel_u.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;

    for (int i = 0; i < nd_u; i++)
      {
	mat(9,i+nd_sigma) = u(i,0);
	mat(10,i+nd_sigma) = u(i,1);
	mat(11,i+nd_sigma) = u(i,2);
      }
  }
};






















template <int D> class DMatMixedMech { };

template <>
class DMatMixedMech<2> : public DMatOp<DMatMixedMech<2> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
public:

  enum { DIM_DMAT = 7 };

  DMatMixedMech (CoefficientFunction * ae,
		 CoefficientFunction * anu) : e(ae), nu(anu) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);


    mat(0,0) = mat(1,1) = 1/ei * 1;
    mat(0,1) = mat(1,0) = -1/ei * nui;
    mat(2,2) = 2/ei * (1+nui);

    mat(3,5) = mat(5,3) = 1;
    mat(4,6) = mat(6,4) = 1;
    
    //    mat(5,5) = -1e-6;
    //    mat(6,6) = -1e-6;
  }  
};


template <>
class DMatMixedMech<3> : public DMatOp<DMatMixedMech<3> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
public:

  enum { DIM_DMAT = 12 };

  DMatMixedMech (CoefficientFunction * ae,
		 CoefficientFunction * anu) : e(ae), nu(anu) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);


    mat(0,0) = mat(1,1) = mat(2,2) = 1/ei;
    mat(0,1) = mat(1,0) = 
      mat(0,2) = mat(2,0) = 
      mat(2,1) = mat(1,2) = -nui / ei;

    mat(3,3) = 2 / ei * (1+nui);
    mat(4,4) = 2 / ei * (1+nui);
    mat(5,5) = 2 / ei * (1+nui);

    
    mat(6,9) = mat(9,6) = 1;
    mat(7,10) = mat(10,7) = 1;
    mat(8,11) = mat(11,8) = 1;
  }  
};








template <int D>
class MixedMechIntegrator 
  : public T_BDBIntegrator<DiffOpMixedMech<D>, DMatMixedMech<D>, FiniteElement>
{
public:
  ///
  MixedMechIntegrator (CoefficientFunction * ae,
		       CoefficientFunction * anu)
    :  T_BDBIntegrator<DiffOpMixedMech<D>, DMatMixedMech<D>, FiniteElement>
  (DMatMixedMech<D> (ae, anu))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechIntegrator (coeffs[0], coeffs[1]);
  }
  
  ///
  virtual string Name () const { return "MixedMech"; }
};




















template <int D>
class DiffOpMixedMechStab : public DiffOp<DiffOpMixedMechStab<D> >
{
public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = D*(D+1) };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<D> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<D>&> (cfel[0]);
    const HCurlFiniteElementD<D> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<D>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    mat = 0.0;

    Mat<D> jac = sip.GetJacobian();
    Mat<D> inv_jac = sip.GetJacobianInverse();
    double det = fabs (sip.GetJacobiDet());


    FlatMatrix<> sigma_bmat(DIM_DMAT/2, nd_sigma, lh);
    // DiffOpIdHDivSym<D>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

    FlatMatrixFixWidth<DIM_DMAT/2> shape_sigma(nd_sigma, lh);
    fel_sigma.CalcShape (sip.IP(), shape_sigma);

    for (int i = 0; i < nd_sigma; i++)
      {
	Mat<3> sigma_ref, hm, sigma;
	sigma_ref(0,0) = shape_sigma(i, 0);
	sigma_ref(1,1) = shape_sigma(i, 1);
	sigma_ref(2,2) = shape_sigma(i, 2);
	sigma_ref(0,1) = sigma_ref(1,0) = shape_sigma(i, 3);
	sigma_ref(0,2) = sigma_ref(2,0) = shape_sigma(i, 4);
	sigma_ref(2,1) = sigma_ref(1,2) = shape_sigma(i, 5);

	hm = jac * sigma_ref;
	sigma = hm * Trans(jac);
	sigma *= (1.0 / sqr (det));

	sigma_bmat(0,i) = sigma(0,0);
	sigma_bmat(1,i) = sigma(1,1);
	sigma_bmat(2,i) = sigma(2,2);
	sigma_bmat(3,i) = sigma(0,1);
	sigma_bmat(4,i) = sigma(0,2);
	sigma_bmat(5,i) = sigma(2,1);
      }









    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < DIM_DMAT/2; j++)
	mat(j,i) = sigma_bmat(j,i);

    FlatMatrixFixWidth<D> u_ref(nd_u, lh);
    FlatMatrixFixWidth<D> ur(nd_u, lh);
    FlatMatrixFixWidth<D> ul(nd_u, lh);
    FlatMatrixFixWidth<D> dudx(nd_u, lh);
    FlatMatrixFixWidth<D> dudy(nd_u, lh);
    FlatMatrixFixWidth<D> dudz(nd_u, lh);
	
    for (int k = 0; k < D; k++)
      {
	IntegrationPoint ipr(sip.IP());
	IntegrationPoint ipl(sip.IP());
	ipr(k) += 1e-6;
	ipl(k) -= 1e-6;

	SpecificIntegrationPoint<D,D> sipr (ipr,sip.GetTransformation(), lh);
	SpecificIntegrationPoint<D,D> sipl (ipl,sip.GetTransformation(), lh);

	fel_u.CalcShape (ipr, u_ref);
	ur = u_ref * sipr.GetJacobianInverse();

	fel_u.CalcShape (ipl, u_ref);
	ul = u_ref * sipl.GetJacobianInverse();

	ur -= ul;
	ur /= (2e-6);

	switch (k)
	  {
	  case 0: dudx = ur; break;
	  case 1: dudy = ur; break;
	  case 2: dudz = ur; break;
	  }
      }


    if (D == 2)
      for (int i = 0; i < nd_u; i++)
	{
	  Mat<2,2> gradur, gradu;
	  gradur(0,0) = dudx(i,0);
	  gradur(0,1) = dudy(i,0);
	  gradur(1,0) = dudx(i,1);
	  gradur(1,1) = dudy(i,1);
	  
	  gradu = gradur * inv_jac;
	  mat(3,nd_sigma+i) = gradu(0,0);
	  mat(4,nd_sigma+i) = gradu(1,1);
	  mat(5,nd_sigma+i) = gradu(0,1)+gradu(1,0);
	}
    else
      for (int i = 0; i < nd_u; i++)
	{
	  Mat<3,3> gradur, gradu;
	  for (int j = 0; j < 3; j++)
	    {
	      gradur(j,0) = dudx(i,j);
	      gradur(j,1) = dudy(i,j);
	      gradur(j,2) = dudz(i,j);
	    }
	  gradu = gradur * inv_jac;

	  mat(6,nd_sigma+i) = gradu(0,0);
	  mat(7,nd_sigma+i) = gradu(1,1);
	  mat(8,nd_sigma+i) = gradu(2,2);
	  mat(9,nd_sigma+i) = gradu(0,1)+gradu(1,0);
	  mat(10,nd_sigma+i) = gradu(0,2)+gradu(2,0);
	  mat(11,nd_sigma+i) = gradu(1,2)+gradu(2,1);
	}

    // (*testout) << "bmat = " << endl << mat << endl;

    /*

    FlatMatrix<> u_refr(nd_u, 2, lh);
    FlatMatrix<> u_refl(nd_u, 2, lh);
    FlatMatrix<> dudx(nd_u, 2, lh);
    FlatMatrix<> dudy(nd_u, 2, lh);
	
    for (int k = 0; k < 2; k++)
      {
	IntegrationPoint ipr(sip.IP());
	IntegrationPoint ipl(sip.IP());
	ipr(k) += 1e-6;
	ipl(k) -= 1e-6;

	fel_u.CalcShape (ipr, u_refr);
	fel_u.CalcShape (ipl, u_refl);

	u_refr -= u_refl;
	u_refr /= (2e-6);

	if (k == 0)
	  dudx = u_refr;
	else
	  dudy = u_refr;
      }

    for (int i = 0; i < nd_u; i++)
      {
	Mat<2,2> gradur, gradu;
	gradur(0,0) = dudx(i,0);
	gradur(0,1) = dudy(i,0);
	gradur(1,0) = dudx(i,1);
	gradur(1,1) = dudy(i,1);
	
	gradu = Trans(inv_jac) * gradur * inv_jac;
	mat(3,nd_sigma+i) = gradu(0,0);
	mat(4,nd_sigma+i) = gradu(1,1);
	mat(5,nd_sigma+i) = gradu(0,1)+gradu(1,0);
      }
    */


    /*
    FlatMatrix<> u(nd_u, 2, lh);
    fel_u.CalcShape (sip.IP(), u_ref);
    u = u_ref * inv_jac;

    for (int i = 0; i < nd_u; i++)
      {
	mat(4,i+nd_sigma) = u(i,0);
	mat(5,i+nd_sigma) = u(i,0);
	mat(6,i+nd_sigma) = u(i,1);
      }
    */
  }
};


template <int D>
class DMatMixedMechStab { };

template <>
class DMatMixedMechStab<2> : public DMatOp<DMatMixedMechStab<2> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
  CoefficientFunction * alpha;
public:

  enum { DIM_DMAT = 6 };

  DMatMixedMechStab (CoefficientFunction * ae,
		     CoefficientFunction * anu,
		     CoefficientFunction * aalpha) : e(ae), nu(anu), alpha(aalpha) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);
    double alphai = alpha->Evaluate (sip);
    
    mat(0,0) = mat(1,1) = 1/ei * 1;
    mat(0,1) = mat(1,0) = -1/ei * nui;
    mat(2,2) = 2/ei * (1+nui);

    double fac = ei / (1-nui*nui);
    mat(3,3) = mat(4,4) = fac;
    mat(3,4) = mat(4,3) = fac*nui;
    mat(5,5) = fac*(1-nui)/2;

    for (int i = 0; i < 3; i++)
      {
	mat(i,3+i) = -1;
	mat(3+i,i) = -1;
      }

    mat *= -alphai;
  }  
};


template <>
class DMatMixedMechStab<3> : public DMatOp<DMatMixedMechStab<3> >
{
  CoefficientFunction * e;
  CoefficientFunction * nu;
  CoefficientFunction * alpha;
public:

  enum { DIM_DMAT = 12 };

  DMatMixedMechStab (CoefficientFunction * ae,
		     CoefficientFunction * anu,
		     CoefficientFunction * aalpha) : e(ae), nu(anu), alpha(aalpha) { ; }
  
  template <typename FEL, typename SIP, typename MAT>
  void GenerateMatrix (const FEL & fel, const SIP & sip,
		       MAT & mat, LocalHeap & lh) const
  {
    mat = 0;

    double ei = e->Evaluate (sip);
    double nui = nu->Evaluate (sip);
    double alphai = alpha->Evaluate (sip);
    
    mat(0,0) = mat(1,1) = mat(2,2) = 1/ei;
    mat(0,1) = mat(1,0) =
      mat(0,2) = mat(2,0) =
      mat(1,2) = mat(2,1) = -nui/ei;

    mat(3,3) = mat(4,4) = mat(5,5) = 2/ei * (1+nui);


    mat(6,6) = mat(7,7) = mat(8,8) = ei * (1-nui) / ( (1+nui)*(1-2*nui) );

    for (int i = 6; i < 9; i++)
      for (int j = 6; j < i; j++)
	mat(i,j) = mat(j,i) = ei * nui  / ( (1+nui)*(1-2*nui) );

    for (int i = 9; i < 12; i++)
      mat(i,i) = ei/ (2*(1+nui));

    for (int i = 0; i < 6; i++)
      {
	mat(i,6+i) = -1;
	mat(6+i,i) = -1;
      }

    mat *= -alphai;
  }  
};














template <int D>
class MixedMechStabIntegrator 
  : public T_BDBIntegrator<DiffOpMixedMechStab<D>, DMatMixedMechStab<D>, FiniteElement>
{
public:
  ///
  MixedMechStabIntegrator (CoefficientFunction * ae,
			   CoefficientFunction * anu,
			   CoefficientFunction * aalpha)
    :  T_BDBIntegrator<DiffOpMixedMechStab<D>, DMatMixedMechStab<D>, FiniteElement>
  (DMatMixedMechStab<D> (ae, anu, aalpha))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechStabIntegrator<D> (coeffs[0], coeffs[1], coeffs[2]);
  }
  
  ///
  virtual string Name () const { return "MixedMechStab"; }
};

































class MixedMechJumpsIntegrator : public BilinearFormIntegrator
{
public:
  MixedMechJumpsIntegrator () 
    : BilinearFormIntegrator()
  { ; }

  virtual ~MixedMechJumpsIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  /*
  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // \int sigma_nt u_t \ds

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();
    int nd = bfel.GetNDof();

    // FlatMatrix<> sigma_ref(nd_sigma, 3, lh);
    FlatMatrix<> u_ref(nd_u, 2, lh);

    // FlatMatrix<> sigma(nd_sigma, 3, lh);
    FlatMatrix<> u(nd_u, 2, lh);

    FlatMatrix<> mat_u(1, nd, lh);
    FlatMatrix<> mat_sigma(1, nd, lh);

    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    
    mat_u = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
    int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4+2*fel_sigma.Order());


    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = fabs (sip.GetJacobiDet());

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);


	    DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
	    for (int i = 0; i < nd_sigma; i++)	    
	      mat_sigma(0,i) = 
		normal(0) * tau(0) * sigma_bmat(0,i) +
		normal(1) * tau(1) * sigma_bmat(1,i) +
		(normal(0) * tau(1) + normal(1) * tau(0)) * sigma_bmat(2,i);

	    
	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;
	    
	    for (int i = 0; i < nd_u; i++)
	      mat_u(0,i+nd_sigma) = u(i,0)*tau(0) + u(i,1)*tau(1);
	    
	    mat_u *= -len * ir1d[l].Weight();
	    elmat += Trans(mat_u) * mat_sigma;
	    elmat += Trans(mat_sigma) * mat_u;
	  }
      }
  }
*/




  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // \int sigma_nt u_t \ds

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<2> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();
    int nd = bfel.GetNDof();

    FlatMatrix<> u_ref(nd_u, 2, lh);
    FlatMatrix<> u(nd_u, 2, lh);

    FlatVector<> mat_u(nd_u, lh);
    FlatVector<> mat_sigma(nd_sigma, lh);

    FlatMatrix<> mat_u_sigma(nd_u, nd_sigma, lh);
    
    FlatMatrix<> sigma_bmat(3, nd_sigma, lh);
    
    mat_u = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
    int ned = (fel_sigma.ElementType() == ET_TRIG) ? 3: 4;

    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 4+2*fel_sigma.Order());

    mat_u_sigma = 0.0;
    for (int k = 0; k < ned; k++)
      {
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	void * heapp = lh.GetPointer();

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);
	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = fabs (sip.GetJacobiDet());

	    tau = jac * tau_ref;
	    double len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);

	    DiffOpIdHDivSym<2>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	    for (int i = 0; i < nd_sigma; i++)	    
	      mat_sigma(i) = 
		normal(0) * tau(0) * sigma_bmat(0,i) +
		normal(1) * tau(1) * sigma_bmat(1,i) +
		(normal(0) * tau(1) + normal(1) * tau(0)) * sigma_bmat(2,i);

	    
	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;
	    
	    for (int i = 0; i < nd_u; i++)
	      mat_u(i) = u(i,0)*tau(0) + u(i,1)*tau(1);
	    
	    mat_u *= -len * ir1d[l].Weight();
	    mat_u_sigma += mat_u * Trans (mat_sigma);
	    lh.CleanUp (heapp);
	  }
      }

    for (int i = 0; i < nd_u; i++)
      for (int j = 0; j < nd_sigma; j++)
	{
	  elmat(nd_sigma+i, j) += mat_u_sigma(i,j);
	  elmat(j, nd_sigma+i) += mat_u_sigma(i,j);
	}
  }













  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechJumpsIntegrator (); 
  }

};











class MixedMechJumpsIntegrator3d : public BilinearFormIntegrator
{
  MixedMechStabIntegrator<3> stab;
public:
  MixedMechJumpsIntegrator3d () 
    : BilinearFormIntegrator(), stab (new ConstantCoefficientFunction (1),
				      new ConstantCoefficientFunction (0),
				      new ConstantCoefficientFunction (1) )
				      
  { ; }

  virtual ~MixedMechJumpsIntegrator3d () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    // \int sigma_nt u_t \ds

    double ei = 1;
    double nui = 0;

    int ndof = bfel.GetNDof();

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<3> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (cfel[0]);
    const HCurlFiniteElementD<3> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    FlatMatrixFixWidth<3> u_ref(nd_u, lh);
    FlatMatrixFixWidth<3> u(nd_u, lh);

    FlatMatrixFixHeight<3> mat_u(nd_u, lh);
    FlatMatrixFixHeight<3> mat_u2(nd_u, lh);
    FlatMatrixFixHeight<3> mat_sigma(nd_sigma, lh);
    FlatMatrixFixHeight<3> mat_sigmau(nd_u, lh);
    FlatMatrix<> hmat(nd_sigma, nd_u, lh);
    FlatMatrix<> hmat2(nd_sigma, nd_u, lh);
    FlatMatrix<> hmatu(nd_u, nd_u, lh);

    FlatMatrix<> hmate(ndof, ndof, lh);

    FlatVector<> sigmae(nd_sigma, lh), phie(nd_u, lh);
    FlatVector<> sumsigmae(nd_sigma, lh);
    FlatVector<> compound_vec(ndof, lh);
    FlatVector<> sum_compound_vec(ndof, lh);

    FlatMatrixFixHeight<6> sigma_bmat(nd_sigma, lh);
    FlatMatrixFixHeight<6> strain_bmat(nd_u, lh);

    FlatMatrixFixHeight<12> compound_bmat(ndof, lh);

    mat_u = 0;
    mat_u2 = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const FACE * faces = ElementTopology::GetFaces (fel_sigma.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
    int nf = ElementTopology::GetNFaces (fel_sigma.ElementType());
    int ned = ElementTopology::GetNEdges (fel_sigma.ElementType());


    int order_sigma = fel_sigma.Order();
    int order_u = fel_u.Order();


    const IntegrationRule & ir1d = SelectIntegrationRule ( ET_SEGM, 6 );

    FlatArray<int> ncdofs (nd_u, lh);
    ncdofs = 0;

    switch (nd_u)
      {
      case 12: 
	for (int i = 0; i < 6; i++)
	  ncdofs[6+i] = 1;
	break;
      case 30:
	for (int i = 0; i < 6; i++)
	  {
	    ncdofs[6+2*i] = 1;
	    ncdofs[7+2*i] = 1;
	  }
	for (int i = 0; i < 4; i++)
	  ncdofs[18+3*i] = 1;
	break;
      default:
	cerr << "unknown displacement element" << endl;
      }

    
    hmat = 0;
    hmat2 = 0;
    mat_sigmau = 0.0;
    mat_u2 = 0.0;

    for (int k = 0; k < nf; k++)
      {
	bool trig = ( faces[k][3] == -1 );

	const IntegrationRule & ir2d =
	  GetIntegrationRules().SelectIntegrationRule ( trig ? ET_TRIG : ET_QUAD, order_sigma+order_u+2);

	Vec<3> pstart, vx, vy;

	for (int j = 0; j < 3; j++)
	  {
	    pstart(j) = points[faces[k][0]][j];
	    vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	    vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	  }

	Vec<3> taux, tauy, normal;

	for (int l = 0; l < ir2d.GetNIP(); l++)
	  {
	    Vec<3> p3d = pstart + ir2d[l](0) * vx + ir2d[l](1) * vy;

	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
	    Mat<3> jac = sip.GetJacobian();
	    Mat<3> inv_jac = sip.GetJacobianInverse();

	    taux = jac * vx;
	    tauy = jac * vy;

	    normal(0) = taux(1) * tauy(2) - taux(2) * tauy(1);
	    normal(1) = taux(2) * tauy(0) - taux(0) * tauy(2);
	    normal(2) = taux(0) * tauy(1) - taux(1) * tauy(0);
	    
	    DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);

	    // sigma n
	    for (int i = 0; i < nd_sigma; i++)	    
	      {
		mat_sigma(0,i) = normal(0) * sigma_bmat(0,i) + normal(1) * sigma_bmat(3,i) + normal(2) * sigma_bmat(4,i);
		mat_sigma(1,i) = normal(0) * sigma_bmat(3,i) + normal(1) * sigma_bmat(1,i) + normal(2) * sigma_bmat(5,i);
		mat_sigma(2,i) = normal(0) * sigma_bmat(4,i) + normal(1) * sigma_bmat(5,i) + normal(2) * sigma_bmat(2,i);
	      }

	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;


	    double inv_area2 = 1.0 / L2Norm2(normal);
	    for (int i = 0; i < nd_u; i++)
	      {
		// n * u
		double un = 0;
		for (int j = 0; j < 3; j++)
		  un += normal(j) * u(i,j);
		un *= inv_area2;


		for (int j = 0; j < 3; j++)
		  mat_u(j,i) = u(i,j) - normal(j) * un;

		/*
		if (i >= 6) 
		  un = 0;
		*/
		
		// if (i < 6)
		  for (int j = 0; j < 3; j++)
		    mat_u2(j,i) = u(i,j) - normal(j) * un;
	      }

	    mat_u *= ir2d[l].Weight();
	    hmat += Trans(mat_sigma) * mat_u;
	    mat_u2 *= ir2d[l].Weight();
	    hmat2 += Trans(mat_sigma) * mat_u2;
	  }
      }

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;
    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < nd_u; j++)
	{
	  elmat(i,j+nd_sigma) = hmat(i,j);
	  elmat(j+nd_sigma,i) = hmat2(i,j);
	}




    /*
    // for consistency of non-conforming displacements, non-symmetric matrix

    hmat = 0;
    hmatu = 0;
    for (int k = 0; k < nf; k++)
      {
	bool trig = ( faces[k][3] == -1 );

	const IntegrationRule & ir2d =
	  GetIntegrationRules().SelectIntegrationRule ( trig ? ET_TRIG : ET_QUAD, order_sigma+order_u+2);

	Vec<3> pstart, vx, vy;

	for (int j = 0; j < 3; j++)
	  {
	    pstart(j) = points[faces[k][0]][j];
	    vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	    vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	  }

	Vec<3> taux, tauy, normal;

	for (int l = 0; l < ir2d.GetNIP(); l++)
	  {
	    Vec<3> p3d = pstart + ir2d[l](0) * vx + ir2d[l](1) * vy;

	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
	    Mat<3> jac = sip.GetJacobian();
	    Mat<3> inv_jac = sip.GetJacobianInverse();
	    // double det = fabs (sip.GetJacobiDet());

	    taux = jac * vx;
	    tauy = jac * vy;

	    normal(0) = taux(1) * tauy(2) - taux(2) * tauy(1);
	    normal(1) = taux(2) * tauy(0) - taux(0) * tauy(2);
	    normal(2) = taux(0) * tauy(1) - taux(1) * tauy(0);

	    
	    // DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
	    DiffOpMixedMechStab<3>::GenerateMatrix (cfel, sip, compound_bmat, lh);

	    for (int i = 0; i < nd_u; i++)	    
	      {
		mat_sigmau(0,i) = 
		  normal(0) * compound_bmat(6,i+nd_sigma) + 
		  0.5 * normal(1) * compound_bmat(9,i+nd_sigma) + 
		  0.5 * normal(2) * compound_bmat(10,i+nd_sigma);
		mat_sigmau(1,i) = 
		  0.5 * normal(0) * compound_bmat(9,i+nd_sigma) + 
		  normal(1) * compound_bmat(7,i+nd_sigma) + 
		  0.5 * normal(2) * compound_bmat(11,i+nd_sigma);
		mat_sigmau(2,i) = 
		  0.5 * normal(0) * compound_bmat(10,i+nd_sigma) + 
		  0.5 * normal(1) * compound_bmat(11,i+nd_sigma) + 
		  normal(2) * compound_bmat(8,i+nd_sigma);
	      }

	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;

	    for (int i = 0; i < nd_u; i++)
	      if (ncdofs[i] == 0)
		u.Row(i) = 0.0;

	    double inv_area2 = 1.0 / L2Norm2(normal);
	    for (int i = 0; i < nd_u; i++)
	      {
		// n * u
		double un = 0;
		for (int j = 0; j < 3; j++)
		  un += normal(j) * u(i,j);
		un *= inv_area2;

		// un *= 0.3;

		for (int j = 0; j < 3; j++)
		  mat_u(j,i) = u(i,j) - normal(j) * un;
	      }

	    mat_u *= ir2d[l].Weight();
	    hmatu += Trans(mat_sigmau) * mat_u;
	  }
      }

    (*testout) << "correction matrix = " << endl << hmatu << endl;
    for (int i = 0; i < nd_u; i++)
      for (int j = 0; j < nd_u; j++)
	elmat(j+nd_sigma,i+nd_sigma) -= hmatu(i,j);	
    */


    // alternative correction matrix:
    // \int A^{-1} eps(u): eps(v) - \int_partial  (A^{-1} eps(u))_nn v_n
    
    hmatu = 0;

    const IntegrationRule & ir3d =
      GetIntegrationRules().SelectIntegrationRule ( ET_TET, 2*order_u+1);

    for (int l = 0; l < ir3d.GetNIP(); l++)
      {
	SpecificIntegrationPoint<3,3> sip (ir3d[l], eltrans, lh);
	    
	Mat<3> jac = sip.GetJacobian();
	Mat<3> inv_jac = sip.GetJacobianInverse();
	double det = fabs (sip.GetJacobiDet());

	DiffOpMixedMechStab<3>::GenerateMatrix (cfel, sip, compound_bmat, lh);
	
	for (int i = 0; i < nd_u; i++)	    
	  for (int j = 0; j < 6; j++)
	    strain_bmat(j,i) = compound_bmat(6+j,i+nd_sigma);

	//	(*testout) << "strain = " << endl << strain_bmat << endl;

	Mat<6> dmat;
	dmat = 0.0;
	dmat(0,0) = dmat(1,1) = dmat(2,2) = ei * (1-nui) / ( (1+nui)*(1-2*nui) );
	
	for (int i = 0; i < 3; i++)
	  for (int j = 0; j < i; j++)
	    dmat(i,j) = dmat(j,i) = ei * nui  / ( (1+nui)*(1-2*nui) );

	for (int i = 3; i < 6; i++)
	  dmat(i,i) = ei/ (2*(1+nui));
	
	hmatu += ir3d[l].Weight() * det * (Trans (strain_bmat) * dmat * strain_bmat);
      }
    (*testout) << "eps(u):eps(v) = " << hmatu << endl;


    for (int i = 0; i < nd_u; i++)
      for (int j = 0; j < nd_u; j++)
	elmat(j+nd_sigma,i+nd_sigma) += hmatu(i,j);	

    // (A eps(u)]_n u_n
    hmatu = 0;
    for (int k = 0; k < nf; k++)
      {
	bool trig = ( faces[k][3] == -1 );

	const IntegrationRule & ir2d =
	  GetIntegrationRules().SelectIntegrationRule ( trig ? ET_TRIG : ET_QUAD, order_sigma+order_u+2);

	Vec<3> pstart, vx, vy;

	for (int j = 0; j < 3; j++)
	  {
	    pstart(j) = points[faces[k][0]][j];
	    vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	    vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	  }

	Vec<3> taux, tauy, normal;

	for (int l = 0; l < ir2d.GetNIP(); l++)
	  {
	    Vec<3> p3d = pstart + ir2d[l](0) * vx + ir2d[l](1) * vy;

	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
	    Mat<3> jac = sip.GetJacobian();
	    Mat<3> inv_jac = sip.GetJacobianInverse();
	    // double det = fabs (sip.GetJacobiDet());

	    taux = jac * vx;
	    tauy = jac * vy;

	    normal(0) = taux(1) * tauy(2) - taux(2) * tauy(1);
	    normal(1) = taux(2) * tauy(0) - taux(0) * tauy(2);
	    normal(2) = taux(0) * tauy(1) - taux(1) * tauy(0);
	    
	    // DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
	    DiffOpMixedMechStab<3>::GenerateMatrix (cfel, sip, compound_bmat, lh);

	    for (int i = 0; i < nd_u; i++)	    
	      {
		mat_sigmau(0,i) = 
		  normal(0) * compound_bmat(6,i+nd_sigma) + 
		  0.5 * normal(1) * compound_bmat(9,i+nd_sigma) + 
		  0.5 * normal(2) * compound_bmat(10,i+nd_sigma);
		mat_sigmau(1,i) = 
		  0.5 * normal(0) * compound_bmat(9,i+nd_sigma) + 
		  normal(1) * compound_bmat(7,i+nd_sigma) + 
		  0.5 * normal(2) * compound_bmat(11,i+nd_sigma);
		mat_sigmau(2,i) = 
		  0.5 * normal(0) * compound_bmat(10,i+nd_sigma) + 
		  0.5 * normal(1) * compound_bmat(11,i+nd_sigma) + 
		  normal(2) * compound_bmat(8,i+nd_sigma);
	      }

	    fel_u.CalcShape (ip, u_ref);
	    u = u_ref * inv_jac;

	    for (int i = 0; i < nd_u; i++)
	      if (ncdofs[i] == 0)
		u.Row(i) = 0.0;

	    double inv_area2 = 1.0 / L2Norm2(normal);
	    for (int i = 0; i < nd_u; i++)
	      {
		// n * u
		double un = 0;
		for (int j = 0; j < 3; j++)
		  un += normal(j) * u(i,j);
		un *= inv_area2;

		for (int j = 0; j < 3; j++)
		  mat_u(j,i) = normal(j) * un;
	      }

	    mat_u *= ir2d[l].Weight();
	    hmatu += Trans(mat_sigmau) * mat_u;
	  }
      }
    //    (*testout) << "normal components: = " << hmatu << endl;

    for (int i = 0; i < nd_u; i++)
      for (int j = 0; j < nd_u; j++)
	elmat(j+nd_sigma,i+nd_sigma) += hmatu(i,j);	






    /*
    // face penalty
    FlatVector<> sumsigma(nd_sigma, lh);
    FlatMatrix<> matsigma(nd_sigma, lh);
    FlatMatrix<> matidsigma(6, nd_sigma, lh);
    matsigma = 0.0;

    for (int k = 0; k < nf; k++)
      {

	bool trig = ( faces[k][3] == -1 );
	
	const IntegrationRule & ir2d =
	  GetIntegrationRules().SelectIntegrationRule ( trig ? ET_TRIG : ET_QUAD, 8);

	for (int j = 0; j < 3; j++)
	  {
	    int pi1 = faces[k][j];
	    int pi2 = faces[k][(j+1)%3];
	    int pi3 = faces[k][(j+2)%3];

	    Vec<3> tau;

	    *testout << "pi1,2,3 = " << pi1 << ", " << pi2 << ", " << pi3 << endl;

	    Vec<3> p0, v1, v2;
	    for (int l = 0; l < 3; l++)
	      {
		p0(l) = points[pi1][l];
		v1(l) = points[pi2][l]-points[pi1][l];
		v2(l) = points[pi3][l]-points[pi1][l];
	      }

	    sumsigma = 0;
	    for (int i = 0; i < ir2d.GetNIP(); i++)
	      {
		Vec<3> p3d = p0 + ir2d[i](0) * v1 + ir2d[i](1) * v2;

		IntegrationPoint ip (p3d(0), p3d(1), p3d(2));
		SpecificIntegrationPoint<3,3> sip(ip, eltrans, lh);

		double lami[4] = { ip(0), ip(1), ip(2), 1-ip(0)-ip(1)-ip(2) };
		double dlami[4][3] = 
		  { { 1, 0, 0 },
		    { 0, 1, 0 },
		    { 0, 0, 1 },
		    { -1, -1, -1 } };

		Vec<3> utaur, utau, putau;
		for (int l = 0; l < 3; l++)
		  utaur(l) = lami[pi1] * dlami[pi2][l] + lami[pi2] * dlami[pi1][l];
		
		utau = Trans (sip.GetJacobianInverse ()) * utaur;
		tau = sip.GetJacobian() * v1;
		tau /= L2Norm (tau);
		// putau = InnerProduct (tau, utau) * tau;
		putau = utau;

		Vec<3> tv1, tv2, n;
		tv1 = sip.GetJacobian() * v1;
		tv2 = sip.GetJacobian() * v2;
		n = Cross (tv1, tv2);

		DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, matidsigma, lh);
		Vec<6> utaun;
		utaun(0) = putau(0) * n(0);
		utaun(1) = putau(1) * n(1);
		utaun(2) = putau(2) * n(2);
		utaun(3) = putau(0) * n(1) + putau(1) * n(0);
		utaun(4) = putau(0) * n(2) + putau(2) * n(0);
		utaun(5) = putau(1) * n(2) + putau(2) * n(1);

		sumsigma += ir2d[i].Weight() * Trans (matidsigma) * utaun;
	      }

	    // (*testout) << "sumsigma = " << endl << sumsigma << endl;
	    matsigma += sumsigma * Trans (sumsigma);
	  }
      }
    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < nd_sigma; j++)
	elmat(i,j) += 1e8 * matsigma(i,j);
    */





    /*
    // consistent edge stabilization 
    hmate = 0;

    for (int k = 0; k < ned; k++)
      {
	const IntegrationRule & ir1d = SelectIntegrationRule ( ET_SEGM, 3 );
	Vec<3> p1, p2, v, tau;
	for (int j = 0; j < 3; j++)
	  {
	    p1(j) = points[edges[k][0]][j];
	    p2(j) = points[edges[k][1]][j];
	  }
	v = p2 - p1;

	sum_compound_vec = 0.0;
	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<3> p3d = p1 + ir1d[l](0) * v;
	    
	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);

	    Mat<3> jac = sip.GetJacobian();
	    tau = jac * v;

	    // sigma - D eps
	    DiffOpMixedMechStab<3>::GenerateMatrix (bfel, sip, compound_bmat, lh);
	    
	    // (*testout) << "comp - bamt = " << compound_bmat << endl;

	    Vec<6> tautau, dtautau;
	    Vec<12> taudiff;
	    Mat<6> d_inv;
	    d_inv = 0.0;
	    d_inv(0,0) = d_inv(1,1) = d_inv(2,2) = 1/ei;
	    d_inv(0,1) = d_inv(1,0) =
	      d_inv(0,2) = d_inv(2,0) =
	      d_inv(1,2) = d_inv(2,1) = -nui/ei;
	    
	    d_inv(3,3) = d_inv(4,4) = d_inv(5,5) = 2/ei * (1+nui);

	    for (int j = 0; j < 3; j++)
	      tautau(j) = tau(j) * tau(j);
	    tautau(3) = tau(0) * tau(1);
	    tautau(4) = tau(0) * tau(2);
	    tautau(5) = tau(1) * tau(2);
	      
	    dtautau = d_inv * tautau;
	    for (int j = 0; j < 6; j++)
	      {
		taudiff(j) = dtautau(j);
		taudiff(6+j) = -tautau(j);
	      }

	    // (*testout) << "taudiff = " << endl << taudiff << endl;

	    compound_vec = Trans (compound_bmat) * taudiff;

	    double lam = 2 * ir1d[l](0) - 1;
	    double fac = lam;
	    sum_compound_vec += (ir1d[l].Weight()) * compound_vec;
	  }
	// (*testout) << "compound_vec = " << compound_vec << endl;
	hmate += sum_compound_vec * Trans (sum_compound_vec);
      }

    (*testout) << "hmate = " << endl << hmate << endl;
    for (int i = nd_sigma; i < ndof; i++)
      for (int j = 0; j < ndof; j++)
	elmat(i,j) -= 1e4 * hmate(i,j);
    */


    /*
    stab.AssembleElementMatrix (bfel, eltrans, hmate, lh);
    for (int i = nd_sigma; i < ndof; i++)
      for (int j = 0; j < ndof; j++)
	elmat(i,j) -= 1e4 * hmate(i,j);
    */

    
    // (*testout) << "hmate = " << endl << hmate << endl;
    /*
    (*testout) << "hmat2 = " << endl << hmat2 << endl;
    (*testout) << "sum = " << endl << (hmat+hmat2) << endl;

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;
    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < nd_u; j++)
	elmat(j+nd_sigma,i) = elmat(i,j+nd_sigma) = hmat(i,j); //  + 1 * hmat2(i,j);
    */


    /*
    // penalize  (sigma n1, n2)
    FlatMatrix<> hmate2 (nd_sigma, nd_sigma, lh);
    hmate2 = 0;
    for (int k = 0; k < ned; k++)
      {
	FlatMatrix<> matidsigma(6, nd_sigma, lh);
	FlatVector<> sumsigma(nd_sigma, lh);
	sumsigma = 0.0;
	FlatVector<> sumsigma2(nd_sigma, lh);
	sumsigma2 = 0.0;

	const IntegrationRule & ir1d = SelectIntegrationRule ( ET_SEGM, 5 );
	Vec<3> p1, p2, v, tau;
	for (int j = 0; j < 3; j++)
	  {
	    p1(j) = points[edges[k][0]][j];
	    p2(j) = points[edges[k][1]][j];
	  }
	v = p2 - p1;

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<3> p3d = p1 + ir1d[l](0) * v;
	    
	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);

	    Mat<3> jac = sip.GetJacobian();
	    tau = jac * v;

	    double len = L2Norm (tau);
	    tau /= len;

	    Vec<3> n1, n2;
	    if (fabs (tau(0)) > fabs(tau(1)))
	      n1 = Vec<3> (-tau(2), 0, tau(0));
	    else
	      n1 = Vec<3> (0, -tau(2), tau(1));
	    n1 /= L2Norm (n1);

	    n2 = Cross (n1, tau);

	    DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, matidsigma, lh);

	    Vec<6> n1n2;
	    n1n2(0) = n1(0) * n2(0);
	    n1n2(1) = n1(1) * n2(1);
	    n1n2(2) = n1(2) * n2(2);
	    n1n2(3) = n1(0) * n2(1) + n2(0) * n1(1);
	    n1n2(4) = n1(0) * n2(2) + n2(0) * n1(2);
	    n1n2(5) = n1(1) * n2(2) + n2(2) * n1(1);

	    double lam = 2 * ir1d[l](0) - 1;
	    double fac = lam * ir1d[l].Weight();
	    sumsigma += fac * Trans (matidsigma) * n1n2;

	    n1n2(0) = n1(0) * n1(0) - n2(0) * n2(0);
	    n1n2(1) = n1(1) * n1(1) - n2(1) * n2(1);
	    n1n2(2) = n1(2) * n1(2) - n2(2) * n2(2);
	    n1n2(3) = 2*n1(0)*n1(1) - 2*n2(0)*n2(1);
	    n1n2(4) = 2*n1(0)*n1(2) - 2*n2(0)*n2(2);
	    n1n2(5) = 2*n1(1)*n1(2) - 2*n2(1)*n2(2);

	    sumsigma2 += fac * Trans (matidsigma) * n1n2;
	  }

	(*testout) << "sumsigma = " << endl << sumsigma << endl;

	hmate2 += sumsigma * Trans (sumsigma);
	hmate2 += sumsigma2 * Trans (sumsigma2);
      }

    (*testout) << "hmate2 = " << endl << hmate2 << endl;

    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < nd_sigma; j++)
	elmat(i,j) += 1e8 * hmate2(i,j);

    // (*testout) << "hmate = " << endl << hmate << endl;

//    (*testout) << "hmat2 = " << endl << hmat2 << endl;
//    (*testout) << "sum = " << endl << (hmat+hmat2) << endl;

    */


    /*
    // face boundary integrals, subtract dominant terms
    hmat2 = 0.0;
    for (int k = 0; k < nf; k++)
      {
	bool trig = ( faces[k][3] == -1 );
	int ned = trig ? 3 : 4;


	Vec<3> pstart, vx, vy, estart, ve, taue;
	Vec<3> taux, tauy, normal, normalf;

	for (int j = 0; j < 3; j++)
	  {
	    pstart(j) = points[faces[k][0]][j];
	    vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	    vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	  }


	for (int ke = 0; ke < ned; ke++)
	  {
	    // (*testout) << "face " << k << " edge " << ke << endl;

	    int vi1 = faces[k][ke];
	    int vi2 = faces[k][(ke+1)%ned];

	    for (int j = 0; j < 3; j++)
	      {
		estart(j) = points[vi1][j];
		ve(j) = points[vi2][j] - points[vi1][j];
	      }

	    for (int l = 0; l < ir1d.GetNIP(); l++)
	      {
		Vec<3> p3d = estart + ir1d[l](0) * ve;

		IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
		SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
		Mat<3> jac = sip.GetJacobian();
		Mat<3> inv_jac = sip.GetJacobianInverse();

		taux = jac * vx;
		tauy = jac * vy;
		taue = jac * ve;

		normal = Cross (taux, tauy);
		normalf = Cross (taue, normal);

		normal /= L2Norm(normal);
		normalf /= L2Norm (normalf);
	    
		DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
		
		// sigma n,nf
		for (int i = 0; i < nd_sigma; i++)	    
		  {
		    sigmae(i) = 
		      normalf(0) * (normal(0) * sigma_bmat(0,i) + normal(1) * sigma_bmat(3,i) + normal(2) * sigma_bmat(4,i))
		      +normalf(1)* (normal(0) * sigma_bmat(3,i) + normal(1) * sigma_bmat(1,i) + normal(2) * sigma_bmat(5,i))
		      +normalf(2)* (normal(0) * sigma_bmat(4,i) + normal(1) * sigma_bmat(5,i) + normal(2) * sigma_bmat(2,i));
		  }
		

		// calc phiE
		phie = 0;
		int ii = 6;
		const EDGE * edges = ElementTopology::GetEdges (ET_TET);
		double lami[4] = { p3d(0), p3d(1), p3d(2), 1-p3d(0)-p3d(1)-p3d(2) };
		for (int i = 0; i < 6; i++)
		  {
		    int es = edges[i][0], ee = edges[i][1];
		    // if (vnums[es] > vnums[ee]) swap (es, ee);
		    double * hp = &phie(ii);
		    ii += IntegratedLegendreMonomialExt::CalcTrigExt (2, lami[ee]-lami[es], 
								      1-lami[es]-lami[ee], hp);
		  }
		
		phie *= ir1d[l].Weight() * L2Norm(taue);
		hmat2 -= sigmae * Trans (phie);
	      }
	  }
      }
    //    (*testout) << "elmat, 1 = " << endl << elmat << endl;

    (*testout) << "corr matrix 2 = " << endl << hmat2 << endl;
    */
    /*
    for (int i = 0; i < nd_sigma; i++)
      for (int j = 0; j < nd_u; j++)
	{
	  elmat(j+nd_sigma, i) += hmat2(i,j);
	  // elmat(i, j+nd_sigma) += hmat2(i,j);
	}
    */
    // (*testout) << "elmat, 2 = " << endl << elmat << endl;







    /*
    // closed face boundary integrals
    // (\int_\partial face \sigma_n,n_f  )^2
    mat_sigma = 0.0;
    for (int k = 0; k < nf; k++)
      {
	bool trig = ( faces[k][3] == -1 );
	int ned = trig ? 3 : 4;

	Vec<3> pstart, vx, vy, estart, ve, taue;
	Vec<3> taux, tauy, normal, normalf;

	for (int j = 0; j < 3; j++)
	  {
	    pstart(j) = points[faces[k][0]][j];
	    vx(j) = points[faces[k][1]][j] - points[faces[k][0]][j];
	    vy(j) = points[faces[k][ trig?2:3 ]][j] - points[faces[k][0]][j];
	  }

	sumsigmae = 0.0;

	for (int ke = 0; ke < ned; ke++)
	  {
	    // (*testout) << "face " << k << " edge " << ke << endl;

	    int vi1 = faces[k][ke];
	    int vi2 = faces[k][(ke+1)%ned];

	    for (int j = 0; j < 3; j++)
	      {
		estart(j) = points[vi1][j];
		ve(j) = points[vi2][j] - points[vi1][j];
	      }

	    for (int l = 0; l < ir1d.GetNIP(); l++)
	      {
		Vec<3> p3d = estart + ir1d[l](0) * ve;

		IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
		SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);
	    
		Mat<3> jac = sip.GetJacobian();
		Mat<3> inv_jac = sip.GetJacobianInverse();

		taux = jac * vx;
		tauy = jac * vy;
		taue = jac * ve;

		normal = Cross (taux, tauy);
		normalf = Cross (taue, normal);

		normal /= L2Norm(normal);
		normalf /= L2Norm (normalf);
	    
		DiffOpIdHDivSym<3>::GenerateMatrix (fel_sigma, sip, sigma_bmat, lh);
		
		// sigma n,nf
		for (int i = 0; i < nd_sigma; i++)	    
		  {
		    sigmae(i) = 
		      normalf(0) * (normal(0) * sigma_bmat(0,i) + normal(1) * sigma_bmat(3,i) + normal(2) * sigma_bmat(4,i))
		      +normalf(1)* (normal(0) * sigma_bmat(3,i) + normal(1) * sigma_bmat(1,i) + normal(2) * sigma_bmat(5,i))
		      +normalf(2)* (normal(0) * sigma_bmat(4,i) + normal(1) * sigma_bmat(5,i) + normal(2) * sigma_bmat(2,i));
		  }
		
		sumsigmae += ir1d[l].Weight() * L2Norm(taue) * sigmae;
	      }
	  }
	for (int i = 0; i < nd_sigma; i++)
	  for (int j = 0; j < nd_sigma; j++)
	    elmat(i, j) += 1e6 * sumsigmae(i) * sumsigmae(j);
      }
    */
  }







  virtual void 
  ApplyElementMatrix (const FiniteElement & bfel, 
		      const ElementTransformation & eltrans, 
		      const FlatVector<double> & elx, 
		      FlatVector<double> & ely,
		      void * precomputed,
		      LocalHeap & lh) const
  {
    double ei = 1;
    double nui = 0;

    int ndof = bfel.GetNDof();

    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement<3> & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (cfel[0]);
    const HCurlFiniteElementD<3> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<3>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();

    FlatMatrixFixWidth<3> u_ref(nd_u, lh);
    FlatMatrixFixWidth<3> u(nd_u, lh);

    FlatMatrixFixHeight<3> mat_u(nd_u, lh);
    FlatMatrixFixHeight<3> mat_sigma(nd_sigma, lh);
    FlatMatrix<> hmat(nd_sigma, nd_u, lh);
    FlatMatrix<> hmat2(nd_sigma, nd_u, lh);

    FlatVector<> sigmae(nd_sigma, lh), phie(nd_u, lh);
    FlatVector<> compound_vec(ndof, lh);

    FlatMatrixFixHeight<6> sigma_bmat(nd_sigma, lh);

    FlatMatrixFixHeight<12> compound_bmat(ndof, lh);

    mat_u = 0;
    mat_sigma = 0;

    const POINT3D * points = ElementTopology::GetVertices (fel_sigma.ElementType());
    const FACE * faces = ElementTopology::GetFaces (fel_sigma.ElementType());
    const EDGE * edges = ElementTopology::GetEdges (fel_sigma.ElementType());
    int nf = ElementTopology::GetNFaces (fel_sigma.ElementType());
    int ned = ElementTopology::GetNFaces (fel_sigma.ElementType());
    hmat = 0;
    




    // edge penalty integrals
    const IntegrationRule & ir1d = SelectIntegrationRule ( ET_SEGM, 0 );
    for (int k = 0; k < ned; k++)
      {
	Vec<3> p1, p2, v, tau;

	for (int j = 0; j < 3; j++)
	  {
	    p1(j) = points[edges[k][0]][j];
	    p2(j) = points[edges[k][1]][j];
	  }
	v = p2 - p1;

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    Vec<3> p3d = p1 + ir1d[l](0) * v;
	    
	    IntegrationPoint ip( p3d(0), p3d(1), p3d(2) );
	    SpecificIntegrationPoint<3,3> sip (ip, eltrans, lh);

	    Mat<3> jac = sip.GetJacobian();
	    tau = jac * v;
	    
	    DiffOpMixedMechStab<3>::GenerateMatrix (bfel, sip, compound_bmat, lh);
	    
	    Vec<6> tautau, dtautau;
	    Vec<12> taudiff;
	    Mat<6> d_inv;
	    d_inv = 0.0;
	    d_inv(0,0) = d_inv(1,1) = d_inv(2,2) = 1/ei;
	    d_inv(0,1) = d_inv(1,0) =
	      d_inv(0,2) = d_inv(2,0) =
	      d_inv(1,2) = d_inv(2,1) = -nui/ei;
	    
	    d_inv(3,3) = d_inv(4,4) = d_inv(5,5) = 2/ei * (1+nui);

	    for (int j = 0; j < 3; j++)
	      tautau(j) = tau(j) * tau(j);
	    tautau(3) = tau(0) * tau(1);
	    tautau(4) = tau(0) * tau(2);
	    tautau(5) = tau(1) * tau(2);
	      
	    dtautau = d_inv * tautau;
	    for (int j = 0; j < 6; j++)
	      {
		taudiff(j) = dtautau(j);
		taudiff(6+j) = -tautau(j);
	      }

	    (*testout) << "taudiff = " << endl << taudiff << endl;

	    compound_vec = Trans (compound_bmat) * taudiff;

	    Vec<12> sigeps;
	    sigeps = compound_bmat * elx;
	    (*testout) << "sigma - eps = " << sigeps << endl;
	    
	    (*testout) << "diff = " << InnerProduct (taudiff, sigeps);

	    (*testout) << "compound_vec = " << compound_vec << endl;

	    (*testout) << "in one step = " << InnerProduct (compound_vec, elx) << endl;

	  }
      }
    
  }







  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechJumpsIntegrator3d (); 
  }

};

































/*

class MixedMechStabIntegrator : public BilinearFormIntegrator
{
  CoefficientFunction * coef;
public:
  MixedMechStabIntegrator (CoefficientFunction * acoef) 
    : BilinearFormIntegrator()
  { 
    coef  = acoef;
  }

  virtual ~MixedMechStabIntegrator () { ; }

  virtual bool BoundaryForm () const 
  { return 0; }


  virtual void
  AssembleElementMatrix (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatMatrix<double> & elmat,
			 LocalHeap & lh) const
  {
    int ndof = bfel.GetNDof();

    elmat.AssignMemory (ndof, ndof, lh);
    elmat = 0;

    // (\int sigma_nt q_t \ds)^2

    
    const CompoundFiniteElement & cfel = 
      dynamic_cast<const CompoundFiniteElement&> (bfel);
    
    const HDivSymFiniteElement & fel_sigma = 
      dynamic_cast<const HDivSymFiniteElement&> (cfel[0]);
    const HCurlFiniteElementD<2> & fel_u = 
      dynamic_cast<const HCurlFiniteElementD<2>&> (cfel[1]);
    
    int nd_sigma = fel_sigma.GetNDof();
    int nd_u = fel_u.GetNDof();


    FlatMatrix<> sigma_ref(nd_sigma, 3, lh);
    FlatMatrix<> sigma(nd_sigma, 3, lh);

    FlatVector<> vec(nd_sigma+nd_u, lh);

    vec = 0;

    const POINT3D * points = ElementTopology::GetVertices (ET_TRIG);
    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    
    const IntegrationRule & ir1d =
      GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 7);

    double cval = 1;
    double len = 1;

    for (int k = 0; k < 3; k++)
      {
	vec = 0;
	Vec<2> ps, pe, tau_ref, sig_n, tau, normal;
	ps(0) = points[edges[k][0]][0];
	ps(1) = points[edges[k][0]][1];
	pe(0) = points[edges[k][1]][0];
	pe(1) = points[edges[k][1]][1];

	tau_ref = pe-ps;

	for (int l = 0; l < ir1d.GetNIP(); l++)
	  {
	    // IntegrationPoint ip( 0.5 * (ps(0)+pe(0)), 0.5 * (ps(1)+pe(1)));
	    Vec<2> p2d = ps + ir1d[l](0) * (pe-ps);
	    IntegrationPoint ip( p2d(0), p2d(1) );
	    
	    SpecificIntegrationPoint<2,2> sip (ip, eltrans, lh);

	    cval = coef -> Evaluate (sip);

	    
	    Mat<2> jac = sip.GetJacobian();
	    Mat<2> inv_jac = sip.GetJacobianInverse();
	    double det = fabs (sip.GetJacobiDet());

	    tau = jac * tau_ref;
	    len = L2Norm(tau);
	    
	    tau *= (1.0/len);
	    normal(0) = tau(1);
	    normal(1) = -tau(0);
	    
	    fel_sigma.CalcShape (ip, sigma_ref);
	    for (int i = 0; i < nd_sigma; i++)
	      {
		Mat<2> sig_ref, sig;
		sig_ref(0,0) = sigma_ref(i, 0);
		sig_ref(1,1) = sigma_ref(i, 1);
		sig_ref(0,1) = sig_ref(1,0) = sigma_ref(i, 2);
		
		sig = (jac) * sig_ref * Trans(jac);
		sig *= (1.0 / sqr (det));
		sig_n = sig * normal;
		vec(i) += ir1d[l].Weight() * (2*ir1d[l](0)-1) 
		  * InnerProduct (tau, sig_n) ;
	      }
	  }
	elmat += (cval) * vec * Trans(vec);
      }
  }

  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new MixedMechStabIntegrator (coeffs[0]); 
  }

};



*/


























  

class NumProcMixedTest : public NumProc
{
public: 
  ///
  NumProcMixedTest (PDE & apde, const Flags & flags)
    : NumProc (apde)
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcMixedTest (pde, flags);
  }

  virtual void Do(LocalHeap & lh)
  {
    FlatMatrix<> elmat1, elmat2;
    
    MixedMechIntegrator<3> mmi(new ConstantCoefficientFunction(1),
			       new ConstantCoefficientFunction(0));

    MixedMechJumpsIntegrator3d mmj;

    
    const FiniteElement & fel = pde.GetFESpace("v")->GetFE (0, lh);

    ElementTransformation eltrans;
    ma.GetElementTransformation (0, eltrans, lh);

    mmi.AssembleElementMatrix (fel, eltrans, elmat1, lh);
    mmj.AssembleElementMatrix (fel, eltrans, elmat2, lh);
    
    (*testout) << "mmi.mat = " << endl << elmat1 << endl;
    (*testout) << "mmj.mat = " << endl << elmat2 << endl;
    (*testout) << "diff = " << endl << (elmat1-elmat2) << endl;
    (*testout) << "sum  = " << endl << (elmat1+elmat2) << endl;



    BaseVector & vecu = pde.GetGridFunction ("u")->GetVector();
    BaseVector & vecuid = pde.GetGridFunction ("uid")->GetVector();
    const BaseMatrix & mata = pde.GetBilinearForm ("a")->GetMatrix();
    const BaseMatrix & matana = pde.GetBilinearForm ("ana")->GetMatrix();
    const BaseMatrix & matapen = pde.GetBilinearForm ("apenalty")->GetMatrix();
    const BaseVector & vecf = pde.GetLinearForm ("f")->GetVector();
    const CompoundFESpace & fes = 
      dynamic_cast<const CompoundFESpace&> (*pde.GetFESpace("v"));

    BaseVector & hv = *vecu.CreateVector();
    BaseVector & hv2 = *vecu.CreateVector();

    (*testout) << "u = " << endl << vecu << endl;
    (*testout) << "f = " << endl << vecf << endl;
    hv = mata * vecu;
    (*testout) << "A u = " << endl << hv << endl;
    hv -= vecf;
    hv *= -1;
    (*testout) << "res = " << endl << hv << endl;


    (*testout) << "uid = " << endl << vecuid << endl;
    (*testout) << "f = " << endl << vecf << endl;
    hv = mata * vecuid;
    (*testout) << "A uid = " << endl << hv << endl;
    hv -= vecf;
    hv *= -1;
    (*testout) << "res uid = " << endl << hv << endl;
    (*testout) << "|| res uid || = " << L2Norm (hv) << endl;

    hv -= matapen * vecu;
    (*testout) << "res m penalty = " << endl << hv << endl;

    hv = vecu - vecuid;
    (*testout) << "diff = " << endl << hv << endl;
    (*testout) << "|| diff || = " << L2Norm (hv) << endl;


    FlatVector<> fvu = vecu.FVDouble();
    FlatVector<> fvuid = vecuid.FVDouble();
    FlatVector<> fvhv = hv.FVDouble();

    const BaseSparseMatrix & spmata = 
      dynamic_cast<const BaseSparseMatrix &> (mata);

    int ndof = fes.GetNDof(); 
    int nds = fes.GetStorageEnd(0);
    int ndu = ndof-nds;
    int ned = ma.GetNEdges();

    cout << "nds = " << nds << ", ndu =" << ndu << ", ndof = " << ndof << endl;
    cout << "ned = " << ned;

    BitArray sigdofs(ndof);
    BitArray uhodofs(ndof);
    sigdofs.Clear();
    for (int i = 0; i < nds; i++)
      sigdofs.Set(i);

    uhodofs.Clear();
    for (int i = nds+ned; i < ndof; i++)
      uhodofs.Set(i);

    sigdofs.Or (uhodofs);

    BaseMatrix & invsig = *spmata.InverseMatrix (&sigdofs);

    hv = 0;
    for (int i = nds; i < nds+ned; i++)
      fvhv(i) = fvuid(i);

    hv2 = vecf - mata * hv;
    vecu = invsig * hv2;
    
    for (int i = nds; i < nds+ned; i++)
      fvu(i) = fvuid(i);


    hv = vecf - mata * vecuid;

    double res1 = L2Norm (fvhv.Range (0, nds));
    double res2 = L2Norm (fvhv.Range (nds, nds+ned));
    double res3 = L2Norm (fvhv.Range (nds+ned, ndof));

    cout << "res1 = " << res1 << endl;
    cout << "res2 = " << res2 << endl;
    cout << "res3 = " << res3 << endl;

    (*testout) << "apply edgejumps" << endl;
    hv = matana * vecuid;



    /*
    BaseMatrix & invsig = *spmata.InverseMatrix (&sigdofs);
    BaseMatrix & invuho = *spmata.InverseMatrix (&uhodofs);

    hv = 0;
    for (int i = 0; i < nds+ned; i++)
      fvhv(i) = fvuid(i);

    hv2 = vecf - mata * hv;
    vecu = invuho * hv2;
    
    for (int i = 0; i < nds+ned; i++)
      fvu(i) = fvuid(i);
    */
  }
};
  





namespace  {
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetFESpaceClasses().AddFESpace ("hdivsymnn", FESpaceHDivSym::Create);
    GetFESpaceClasses().AddFESpace ("mixedmech", FESpaceMixedMech::Create);


    GetIntegrators().AddBFIntegrator ("mixedmech", 2, 2,
				      MixedMechIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("mixedmech", 3, 2,
				      MixedMechIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("mixedmechjumps", 2, 0,
				      MixedMechJumpsIntegrator::Create);
    GetIntegrators().AddBFIntegrator ("mixedmechjumps", 3, 0,
				      MixedMechJumpsIntegrator3d::Create);

    GetIntegrators().AddBFIntegrator ("mixedmechstab", 2, 3,
				      MixedMechStabIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("mixedmechstab", 3, 3,
				      MixedMechStabIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("masshdivsym", 2, 1,
				      HDivSymMassIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("masshdivsym", 3, 1,
				      HDivSymMassIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("robinhdivsym", 2, 1,
				      HDivSymRobinIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("robinhdivsym", 3, 1,
				      HDivSymRobinIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("divdivhdivsym", 2, 1,
				      HDivSymDivDivIntegrator<2>::Create);
    GetIntegrators().AddBFIntegrator ("divdivhdivsym", 3, 1,
				      HDivSymDivDivIntegrator<3>::Create);

    GetIntegrators().AddBFIntegrator ("divdivdivdivhdivsym", 2, 1,
				      HDivSymDivDivDivDivIntegrator<2>::Create);


    GetIntegrators().AddLFIntegrator ("sourcehdivsym", 2, 3,
				      HDivSymSourceIntegrator<2>::Create);
    GetIntegrators().AddLFIntegrator ("neumannhdivsym", 2, 1,
				      HDivSymNeumannIntegrator<2>::Create);
    GetIntegrators().AddLFIntegrator ("sourcehdivsym", 3, 6,
				      HDivSymSourceIntegrator<3>::Create);
    GetIntegrators().AddLFIntegrator ("neumannhdivsym", 3, 1,
				      HDivSymNeumannIntegrator<3>::Create);
    GetIntegrators().AddLFIntegrator ("divhdivsym", 2, 2,
				      HDivSymDivIntegrator::Create);


    GetNumProcs().AddNumProc ("mixedtest", NumProcMixedTest::Create);
  }
  
  Init init;
}
