

template <int DIM>
class HDivSymFiniteElement : public FiniteElement
{
protected:
  int vnums[8];
  int order_facet;
  int order_inner;
  int ninner;
  enum { DIM_STRESS = DIM * (DIM+1) / 2 };
public:
  HDivSymFiniteElement (ELEMENT_TYPE etype);
  virtual ~HDivSymFiniteElement();

  void SetVertexNumbers (FlatArray<int> & avnums);

  int GetNInner() const { return ninner; }

  // sigma_xx, sigma_yy, sigma_xy
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<DIM_STRESS> shape) const = 0;
  
  virtual void CalcDivShape (const IntegrationPoint & ip, 
			     FlatMatrixFixWidth<DIM> divshape) const;

  virtual void CalcDivDivShape (const IntegrationPoint & ip, 
				FlatVector<> ddshape) const;
};
  


class HDivSymTrig : public HDivSymFiniteElement<2>
{
public:
  HDivSymTrig (int aorder, int aorder_facet);

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const;  // n * 3 
};


class HDivSymQuad : public HDivSymFiniteElement<2>
{
public:
  HDivSymQuad (int aorder, int aorder_facet);

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<3> shape) const;

  virtual void CalcDivShape (const IntegrationPoint & ip, 
			     FlatMatrixFixWidth<2> divshape) const;
};







class HDivSymTet : public HDivSymFiniteElement<3>
{
public:
  HDivSymTet (int aorder, int aorderfacet);

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<6> shape) const;
};










// [P^2(x,y) x P^2(z)]^6  =   108
// sigma_nn = P^1   bot/top  ->  -2 * 3
//                  quad-faces -> -3 * 5
//                            == 87
// div sigma in [P^0]^3       -15

// quad-face ... 4    -> 12
// trig-face ... 3    -> 6
//  n-inner = 54

// inner,  sigma(x,y)-tensor  Trig-inner tensor P2(z)  -> 5 x 3 = 15
// inner,  sigma_xz, sigma_yz   voll P2                -> 2 x 18 = 36
// inner,  sigma_zz, P2 * z(1-z)                                    6
//                                                              = 57
// sigma_xx, sigma_yy, sigma_zz, sigma_xy, sigma_xz, sigma_yz
class HDivSymPrism : public HDivSymFiniteElement<3>
{
  int orderz;
public:
  HDivSymPrism (int aorder, int aorderfacet);

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<6> shape) const;
};


class HDivSymHex : public HDivSymFiniteElement<3>
{
public:
  HDivSymHex (int aorder, int aorderfacet);

  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatMatrixFixWidth<6> shape) const;
};










class HDivSymNormalFiniteElement : public FiniteElement
{
protected:
  int vnums[4];
public:
  void SetVertexNumbers (FlatArray<int> & avnums);
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const = 0;
};


class HDivSymNormalSegm : public HDivSymNormalFiniteElement
{
public:
  HDivSymNormalSegm (int aorder);
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
};



class HDivSymNormalTrig : public HDivSymNormalFiniteElement
{
public:
  HDivSymNormalTrig (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
};

class HDivSymNormalQuad : public HDivSymNormalFiniteElement
{
public:
  HDivSymNormalQuad (int aorder);

  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
};
























class FESpaceHDivSym : public FESpace
{
private:
  int nfa;
  int nel;
  int order_facet;

  ARRAY<int> first_face_dof;
  ARRAY<int> first_element_dof;
  int ndof;
  bool noncontinuous;
public:

  FESpaceHDivSym (const MeshAccess & ama, const Flags & aflags);

  virtual ~FESpaceHDivSym ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FESpaceHDivSym (ma, flags);
  }

  virtual string GetClassName () const
  {
    return "FESpaceHDivSym";
  }


  virtual void Update();

  virtual int GetNDof () const
  {
    return ndof;
  }


  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;

  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;

  virtual void GetWireBasketDofNrs (int elnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); }
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const
  { GetDofNrs (elnr, dnums); }
};












template<int D>
class DiffOpIdHDivSym : public DiffOp<DiffOpIdHDivSym<D> >
{ };


template<>
class DiffOpIdHDivSym<2> : public DiffOp<DiffOpIdHDivSym<2> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 3 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSymFiniteElement<2> & fel = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (bfel);
    
    int nd = fel.GetNDof();

    Mat<2> jac = sip.GetJacobian();
    double det = fabs (sip.GetJacobiDet());

    FlatMatrix<> shape(nd, 3, lh);
    fel.CalcShape (sip.IP(), shape);


    Mat<3> trans;
    for (int i = 0; i < 3; i++)
      {
	Mat<2> sigma_ref, hm, sigma;
	sigma_ref = 0.0;
	switch (i)
	  {
	  case 0: sigma_ref(0,0) = 1.0; break;
	  case 1: sigma_ref(1,1) = 1.0; break;
	  case 2: sigma_ref(0,1) = sigma_ref(1,0) = 1.0; break;
	  }

	hm = jac * sigma_ref;
	sigma = hm * Trans(jac);
	sigma *= (1.0 / sqr (det));

	trans(0,i) = sigma(0,0);
	trans(1,i) = sigma(1,1);
	trans(2,i) = sigma(0,1);
      }

    mat = trans * Trans (shape);
  }
};



template<>
class DiffOpIdHDivSym<3> : public DiffOp<DiffOpIdHDivSym<3> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 3 };
  enum { DIM_ELEMENT = 3 };
  enum { DIM_DMAT = 6 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSymFiniteElement<3> & fel = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (bfel);
    
    int nd = fel.GetNDof();

    FlatMatrixFixWidth<6> shape(nd, lh);
    fel.CalcShape (sip.IP(), shape);

    Mat<3> jac = sip.GetJacobian();
    double det = fabs (sip.GetJacobiDet());

    for (int i = 0; i < nd; i++)
      {
	Mat<3> sigma_ref, hm, sigma;
	sigma_ref(0,0) = shape(i, 0);
	sigma_ref(1,1) = shape(i, 1);
	sigma_ref(2,2) = shape(i, 2);
	sigma_ref(0,1) = sigma_ref(1,0) = shape(i, 3);
	sigma_ref(0,2) = sigma_ref(2,0) = shape(i, 4);
	sigma_ref(2,1) = sigma_ref(1,2) = shape(i, 5);

	hm = jac * sigma_ref;
	sigma = hm * Trans(jac);
	sigma *= (1.0 / sqr (det));

	mat(0,i) = sigma(0,0);
	mat(1,i) = sigma(1,1);
	mat(2,i) = sigma(2,2);
	mat(3,i) = sigma(0,1);
	mat(4,i) = sigma(0,2);
	mat(5,i) = sigma(2,1);
      }
  }
};














template <int D>
class DiffOpBoundIdHDivSym : public DiffOp<DiffOpBoundIdHDivSym<D> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D-1 };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSymNormalFiniteElement & fel = 
      dynamic_cast<const HDivSymNormalFiniteElement&> (bfel);
    
    int nd = fel.GetNDof();

    FlatVector<> shape(nd, lh);
    fel.CalcShape (sip.IP(), shape);
    
    double det = fabs (sip.GetJacobiDet());
    mat = (1.0/sqr(det)) * Trans (shape);
  }
};










template <int D> class DiffOpDivHDivSym { };


template <> 
class DiffOpDivHDivSym<2> : public DiffOp<DiffOpDivHDivSym<2> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 2 };
  enum { DIM_ELEMENT = 2 };
  enum { DIM_DMAT = 2 };
  enum { DIFFORDER = 0 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSymFiniteElement<2> & fel = 
      dynamic_cast<const HDivSymFiniteElement<2>&> (bfel);
    
    int nd = fel.GetNDof();

    FlatMatrixFixWidth<2> div_shape(nd, lh);
    fel.CalcDivShape (sip.IP(), div_shape);

    FlatMatrixFixWidth<3> shape(nd, lh);
    fel.CalcShape (sip.IP(), shape);
    
    Mat<2> jac = sip.GetJacobian();
    double det = fabs (sip.GetJacobiDet());
    Mat<2> sjac = (1.0/(det*det)) * jac;


    Mat<2> hesse[2];
    sip.CalcHesse (hesse[0], hesse[1]);
    
    Mat<2,2,AutoDiff<2> > fad;
    for (int i = 0; i < 2; i++)
      for (int j = 0; j < 2; j++)
	{
	  fad(i,j).Value() = jac(i,j);
	  for (int k = 0; k < 2; k++)
	    fad(i,j).DValue(k) = hesse[i](j,k);
	}

    AutoDiff<2> ad_det = Det (fad);

    if (ad_det.Value() < 0.0)
      {
	cout << "neg det" << endl;
	ad_det *= -1;
      }    

    AutoDiff<2> iad_det = 1.0 / ad_det;
    fad *= iad_det;
    
    mat = (sjac) * Trans (div_shape);

    for (int i = 0; i < nd; i++)
      {
	Mat<2> sigma_ref;
	sigma_ref(0,0) = shape(i, 0);
	sigma_ref(1,1) = shape(i, 1);
	sigma_ref(0,1) = sigma_ref(1,0) = shape(i, 2);

	Vec<2> hv2;
	hv2 = 0.0;
	for (int j = 0; j < 2; j++)
	  for (int k = 0; k < 2; k++)
	    for (int l = 0; l < 2; l++)
	      hv2(k) += fad(k,l).DValue(j) * sigma_ref(l,j);

	hv2 *= iad_det.Value();

        mat(0,i) += hv2(0);
	mat(1,i) += hv2(1);
      }
  }
};


template <> 
class DiffOpDivHDivSym<3> : public DiffOp<DiffOpDivHDivSym<3> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = 3 };
  enum { DIM_ELEMENT = 3 };
  enum { DIM_DMAT = 3 };
  enum { DIFFORDER = 1 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSymFiniteElement<3> & fel = 
      dynamic_cast<const HDivSymFiniteElement<3>&> (bfel);
    
    int nd = fel.GetNDof();

    FlatMatrixFixWidth<3> div_shape(nd, lh);
    fel.CalcDivShape (sip.IP(), div_shape);

    FlatMatrixFixWidth<6> shape(nd, lh);
    fel.CalcShape (sip.IP(), shape);


    Mat<3> jac = sip.GetJacobian();
    double det = fabs (sip.GetJacobiDet());
    Mat<3> sjac = (1.0/(det*det)) * jac;

    mat = (sjac) * Trans (div_shape);








    Mat<3> hesse[3];
    sip.CalcHesse (hesse[0], hesse[1], hesse[2]);
    
    Mat<3,3,AutoDiff<3> > fad;
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
	{
	  fad(i,j).Value() = jac(i,j);
	  for (int k = 0; k < 3; k++)
	    fad(i,j).DValue(k) = hesse[i](j,k);
	}

    AutoDiff<3> ad_det = Det (fad);

    if (ad_det.Value() < 0.0)
      {
	// cout << "neg det" << endl;
	// ad_det *= -1;
      }    

    AutoDiff<3> iad_det = 1.0 / ad_det;
    fad *= iad_det;
    
    mat = (sjac) * Trans (div_shape);

    for (int i = 0; i < nd; i++)
      {
	Mat<3> sigma_ref;
	sigma_ref(0,0) = shape(i, 0);
	sigma_ref(1,1) = shape(i, 1);
	sigma_ref(2,2) = shape(i, 2);
	sigma_ref(0,1) = sigma_ref(1,0) = shape(i, 3);
	sigma_ref(0,2) = sigma_ref(2,0) = shape(i, 4);
	sigma_ref(1,2) = sigma_ref(2,1) = shape(i, 5);

	Vec<3> hv2;
	hv2 = 0.0;
	for (int j = 0; j < 3; j++)
	  for (int k = 0; k < 3; k++)
	    for (int l = 0; l < 3; l++)
	      hv2(k) += fad(k,l).DValue(j) * sigma_ref(l,j);

	hv2 *= iad_det.Value();

        mat(0,i) += hv2(0);
	mat(1,i) += hv2(1);
	mat(2,i) += hv2(2);
      }


  }
};
















template <int D>
class DiffOpDivDivHDivSym : public DiffOp<DiffOpDivDivHDivSym<D> >
{

public:
  enum { DIM = 1 };
  enum { DIM_SPACE = D };
  enum { DIM_ELEMENT = D };
  enum { DIM_DMAT = 1 };
  enum { DIFFORDER = 1 };
  
  template <typename FEL, typename SIP, typename MAT>
  static void GenerateMatrix (const FEL & bfel, const SIP & sip,
			      MAT & mat, LocalHeap & lh)
  {
    const HDivSymFiniteElement<D> & fel = 
      dynamic_cast<const HDivSymFiniteElement<D>&> (bfel);
    
    int nd = fel.GetNDof();

    FlatVector<> ddshape(nd, lh);
    fel.CalcDivDivShape (sip.IP(), ddshape);

    mat = Trans (ddshape);
  }
};






