#include <solve.hpp>


namespace ngsolve
{
  

  template <int DIM>
  class Cut_Ieq_Integrator : public BilinearFormIntegrator
  {

    double wave_k, rad;
    Matrix<Complex> mat_m, mat_k;
    double Cd;

    LaplaceBoundaryIntegrator<DIM> lap_int;
    RobinIntegrator<DIM> mass_int;

  public:
    
    Cut_Ieq_Integrator (CoefficientFunction * acoef_k,
			CoefficientFunction * acoef_rad);
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new Cut_Ieq_Integrator<DIM> (coeffs[0], coeffs[1]);
    }

    
    virtual string Name () const { return "Cut_Ieq integrator (Thorsten)"; }    
    
    
    virtual bool BoundaryForm () const { return 1; }

    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & locheap) const
    { ; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<Complex> & elmat,
			   LocalHeap & locheap) const;
  };

  template <int DIM>
  class Axis_Ode_Integrator : public BilinearFormIntegrator
  {

    double wave_k, rad;
    Complex test_point_z;
    Matrix<Complex> mat_m, mat_k;

    LaplaceBoundaryIntegrator<DIM> lap_int;
    RobinIntegrator<DIM> mass_int;

  public:
    
    Axis_Ode_Integrator (CoefficientFunction * acoef_k,
			CoefficientFunction * acoef_rad);
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new Axis_Ode_Integrator<DIM> (coeffs[0], coeffs[1]);
    }

    
    virtual string Name () const { return "Axis_Ode integrator (Thorsten)"; }    
    
    
    virtual bool BoundaryForm () const { return 1; }

    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & locheap) const
    { ; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<Complex> & elmat,
			   LocalHeap & locheap) const;
  };






  template <int DIM>
  class RealAxis_IE_Laplace_Integrator : public BilinearFormIntegrator
  {
    double rad, lam;
    Matrix<Complex> mat_m, mat_k;

    LaplaceBoundaryIntegrator<DIM> lap_int;
    RobinIntegrator<DIM> mass_int;

  public:
    
    RealAxis_IE_Laplace_Integrator (CoefficientFunction * acoef_lam,
				 CoefficientFunction * acoef_rad);
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new RealAxis_IE_Laplace_Integrator<DIM> (coeffs[0], coeffs[1]);
    }

    
    virtual string Name () const { return "RealAxis_IE_Laplace Integrator (Thorsten+Joachim)"; }    
    
    
    virtual bool BoundaryForm () const { return 1; }

    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & locheap) const
    { ; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<Complex> & elmat,
			   LocalHeap & locheap) const;
  };









  template <int DIM>
  class RealAxis_ODE_Laplace_Integrator : public BilinearFormIntegrator
  {
    double rad, lam, rho;
    Matrix<Complex> mat_m, mat_k;

    LaplaceBoundaryIntegrator<DIM> lap_int;
    RobinIntegrator<DIM> mass_int;

  public:
    
    RealAxis_ODE_Laplace_Integrator (CoefficientFunction * acoef_lam,
				     CoefficientFunction * acoef_rho,
				     CoefficientFunction * acoef_rad);
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new RealAxis_ODE_Laplace_Integrator<DIM> (coeffs[0], coeffs[1], coeffs[2]);
    }

    
    virtual string Name () const { return "RealAxis_ODE_Laplace Integrator (Thorsten+Joachim)"; }    
    
    
    virtual bool BoundaryForm () const { return 1; }

    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & locheap) const
    { ; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<Complex> & elmat,
			   LocalHeap & locheap) const;
  };








  class Hardy_Integrator2D : public BilinearFormIntegrator
  {
    double rad, fac_laplace, wave_k;
    Complex s0;
    Matrix<Complex> mat_m, mat_k, mat_mkappa;

    LaplaceBoundaryIntegrator<2> lap_int;
    RobinIntegrator<2> mass_int;

  public:
    
    Hardy_Integrator2D (CoefficientFunction * acoef_k,
			CoefficientFunction * acoef_faclaplace,
			CoefficientFunction * acoef_rad,
			CoefficientFunction * acoef_s0r,
			CoefficientFunction * acoef_s0i);
  
    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new Hardy_Integrator2D (coeffs[0], coeffs[1], coeffs[2], coeffs[3], coeffs[4]);
    }

    
    virtual string Name () const { return "Hardy Space Method, symmetric"; }
    
    
    virtual bool BoundaryForm () const { return 1; }

    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & locheap) const
    { ; }
    
    virtual void
    AssembleElementMatrix (const FiniteElement & fel, 
			   const ElementTransformation & eltrans, 
			   FlatMatrix<Complex> & elmat,
			   LocalHeap & locheap) const;

    
  private:
    void SetupDiffOp(Matrix<Complex> &Diffop);
  };











  template <int DIM>
  Cut_Ieq_Integrator<DIM> ::
  Cut_Ieq_Integrator (CoefficientFunction * acoef_k,
		      CoefficientFunction * acoef_rad)
    : lap_int(new ConstantCoefficientFunction (1)),
      mass_int (new ConstantCoefficientFunction (1))
  {
    // lap_int.SetIntegrationOrder (10);
    // mass_int.SetIntegrationOrder (10);

    int i, j, k, l;

    wave_k = acoef_k -> EvaluateConst();
    rad = acoef_rad -> EvaluateConst();

    Cd = (DIM-1)*(3-DIM)/4.0;


    double ma[3][3] =
      {
	{ (88-7*sqrt(6.))/360, (296-169*sqrt(6.))/1800, (-2+3*sqrt(6.))/225 },
	{ (296+169*sqrt(6.))/1800, (88+7*sqrt(6.))/360, (-2-3*sqrt(6.))/225 },
	{ (16-sqrt(6.))/36, (16+sqrt(6.))/36, 1.0/9 }
      };
    double vb[3] = {(16-sqrt(6.))/36, (16+sqrt(6.))/36, 1./9 };
    double vc[3] = { (4-sqrt(6.))/10, (4+sqrt(6.))/10, 1 };
    int rk_stages = 3;

    /*
    double ma[1][1] = {{ 0.5 } };
    double vb[1] = { 1.0 };
    double vc[1] = { 0.5 };
    int rk_stages = 1;
    */

    FlatMatrix<double> rk_mata(rk_stages, rk_stages,&ma[0][0]);
    FlatVector<double> rk_vecb(rk_stages, &vb[0]), rk_vecc(rk_stages, &vc[0]);

    int n_pts = 4;
    Vector<double> pts(n_pts+1);

    for (j = 0; j <= n_pts; j++)
      pts(j) = n_pts * double(j*j) / double(n_pts*n_pts);
    
    int n = n_pts * rk_stages;

    Vector<double> int_pts(n), int_weights(n);

    for (j = 1; j <= n_pts; j++)
      for (k = 0; k < rk_stages; k++)
	{
	  int_pts((j-1)*rk_stages+k) = pts(j-1) + rk_vecc(k) * (pts(j)-pts(j-1));
	  int_weights((j-1)*rk_stages+k) = (pts(j)-pts(j-1))*rk_vecb(k);
	}


    mat_m.SetSize(n+2);
    mat_m = Complex(0.0);
    mat_k.SetSize(n+2);
    mat_k = Complex(0.0);

    for (i = 0; i < n_pts; i++)
      { 
	for (k = 0; k < rk_stages; k++)
	  for (l = 0; l < rk_stages; l++)
	    {
	      double t = int_pts(i*rk_stages+k);
	      double tau = int_pts(i*rk_stages+l);

	      mat_k(i*rk_stages+k, i*rk_stages+l) = 
		rk_mata(k,l) * (pts(i+1)-pts(i)) * 
		(t - tau) * exp (-rad*(t-tau)) / (t*(t-2.*Complex(0,1)*wave_k));

	      mat_m(i*rk_stages+k, i*rk_stages+l) = 
		Cd * mat_k(i*rk_stages+k, i*rk_stages+l);
	    }

	for (j = 0; j < i; j++)
	  {
	    for (k = 0; k < rk_stages; k++)
	      for (l = 0; l < rk_stages; l++)
		{
		  double t = int_pts(i*rk_stages+k);
		  double tau = int_pts(j*rk_stages+l);

		  mat_k(i*rk_stages+k, j*rk_stages+l) = 
		    int_weights (j*rk_stages+l) *
		    (t - tau) * exp (-rad*(t-tau)) / (t*(t-2.*Complex(0,1)*wave_k));

		  mat_m(i*rk_stages+k, j*rk_stages+l) = 
		    Cd * mat_k(i*rk_stages+k, j*rk_stages+l);
		}
	  }
      }

    for (i = 0; i < n; i++)
      mat_m(i,i) += 1.0;


    // u_\infty term
    for (i = 0; i < n; i++)
      {
	mat_k(i, n) =
	  exp (-rad*int_pts(i)) / (int_pts(i)-2.0*Complex(0,1)*wave_k);

	mat_m(i, n) = Cd * mat_k(i, n);
      }


    // representation formula at coupling boundary
    for (i = 0; i < n; i++)
      mat_m(n, i) = int_weights(i);
    mat_m(n, n) = 1.0;
    mat_m(n, n+1) =
      -pow (rad, (DIM-1)/2.0) * exp (-Complex(0,1) * wave_k * rad);


    // derivative of representation formula at coupling boundary
    Complex help = pow (rad, -(DIM-1)/2.0)*exp (Complex(0,1)*wave_k*rad);
    for (i = 0; i < n; i++)
      mat_m(n+1,i) = -help * (Complex(0,1)*wave_k-int_pts(i)) * int_weights(i);
    mat_m(n+1,n) = -help * Complex(0,1) * wave_k;
    mat_m(n+1,n+1) = (DIM-1)/ (2.0 * rad); 
    

    (*testout) << "k = " << wave_k << ", rad = " << rad << endl;
    (*testout) << "mat_m = " << endl << mat_m << endl;
    (*testout) << "mat_k = " << endl << mat_k << endl;

    {
    Matrix<Complex> sum(n+2,n+2), a(n+1,n+1), ainv(n+1,n+1), 
      b1(n+1,1), b2(1,n+1), c(1,1);
    ofstream out("schur.dat");

    //    for (double t = 0.1; t <= 1000; t *= 1.1)
    for (double t = 0; t <= 30; t++)
      {
	sum = (-t*t) * mat_k + mat_m;

	for (i = 0; i < n+1; i++)
	  for (j = 0; j < n+1; j++)
	    a(i,j) = sum(i,j);

	for (i = 0; i < n+1; i++)
	  {
	    b1(i,0) = sum(i,n+1);
	    b2(0,i) = sum(n+1,i);
	  }

	CalcInverse (a, ainv);
	c = b2 * ainv * b1;
	Complex s = sum(n+1,n+1) - c(0,0); 
	
	out << t << " " << s.real() << " " << s.imag() << endl;
	
	/*
	(*testout) << "t = " << t << endl
		   << "(psi, u_infty) = " << endl << (ainv * b1) << endl;
	*/
      }
    }
    
    // eliminate u_infty
    for (i = 0; i < n+2; i++)
      {
	swap (mat_m(i,n), mat_m(i,n+1));
	swap (mat_k(i,n), mat_k(i,n+1));
      }
    for (i = 0; i < n+2; i++)
      {
	swap (mat_m(n,i), mat_m(n+1,i));
	swap (mat_k(n,i), mat_k(n+1,i));
      }

    Matrix<Complex> mat_m2(n+1), mat_k2(n+1);
    Complex inv = 1.0/mat_m(n+1,n+1);

    for (i = 0; i < n+1; i++)
      for (j = 0; j < n+1; j++)
	{
	  mat_m2(i,j) = mat_m(i,j) - mat_m(n+1,j) * mat_m(i,n+1) * inv;
	  mat_k2(i,j) = mat_k(i,j) - mat_m(n+1,j) * mat_k(i,n+1) * inv;
	}
    
    mat_m.SetSize(n+1);
    mat_k.SetSize(n+1);
    mat_m = mat_m2;
    mat_k = mat_k2;
    (*testout) << "mat_m, now = " << endl << mat_m << endl;
    (*testout) << "mat_k, now = " << endl << mat_k << endl;
    cout << "************** YOU NEED inf_dof = " << mat_m.Height()-1 << " !!!" << endl;
  }

  template <int DIM>
  void Cut_Ieq_Integrator<DIM> ::
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & locheap) const
  {
    FlatMatrix<double> lap_mat, mass_mat;

    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);

    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;

    elmat.AssignMemory (n*m, n*m, locheap);

    (*testout) << "assemble elmat, n = " << n << ", m = " << m << endl;

    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = -mat_k(i,j)*lap_mat(k,l)+mat_m(i,j)*mass_mat(k,l);
  }



  /* ************************** Axis-ODE ************************************** */

  template <int DIM>
  Axis_Ode_Integrator<DIM> ::
  Axis_Ode_Integrator (CoefficientFunction * acoef_k,
		       CoefficientFunction * acoef_rad)
    : lap_int(new ConstantCoefficientFunction (1)),
      mass_int (new ConstantCoefficientFunction (1))
  {
    int i, j, k, l;

    wave_k = acoef_k -> EvaluateConst();
    rad = acoef_rad -> EvaluateConst();
    test_point_z = Complex(0,1);
    


    double ma[3][3] =
      {
	{ (88-7*sqrt(6.))/360, (296-169*sqrt(6.))/1800, (-2+3*sqrt(6.))/225 },
	{ (296+169*sqrt(6.))/1800, (88+7*sqrt(6.))/360, (-2-3*sqrt(6.))/225 },
	{ (16-sqrt(6.))/36, (16+sqrt(6.))/36, 1.0/9 }
      };
    double vb[3] = {(16-sqrt(6.))/36, (16+sqrt(6.))/36, 1./9 };
    double vc[3] = { (4-sqrt(6.))/10, (4+sqrt(6.))/10, 1 };
    
    int rk_stages = 3;

    /*
    double ma[1][1] = { { 1. } };
    double vb[1] = { 1. };
    double vc[1] = { 1. };
    int rk_stages = 1;
    */
    FlatMatrix<double> rk_mata(rk_stages, rk_stages,&ma[0][0]);
    FlatVector<double> rk_vecb(rk_stages, &vb[0]), rk_vecc(rk_stages, &vc[0]);

    int n_blocks = 6;
    Vector<double> pts(n_blocks+1);

    for (j = 0; j <= n_blocks; j++)
      pts(j) = sinh(4.0-0.25*j);
    // pts(j) = sinh(1.0-j);

    double R = pts(0), S = pts(n_blocks);
    
    int blocklen = 2*rk_stages,
      n = n_blocks * blocklen;
    
    Vector<double> int_pts(n/2), int_weights(n/2);

    for (j = 1; j <= n_blocks; j++)
      for (k = 0; k < rk_stages; k++)
	{
	  int_pts((j-1)*blocklen/2+k) = pts(j-1) + rk_vecc(k) * (pts(j)-pts(j-1));
	  int_weights((j-1)*blocklen/2+k) = (pts(j)-pts(j-1))*rk_vecb(k);
	}

    mat_m.SetSize(n+4);
    mat_m = Complex(0.0);
    mat_k.SetSize(n+4);
    mat_k = Complex(0.0);

    int diri_index = n+3,
      neum_index = n+2,
      polecond_index = n+2;

    mat_m(0,0) = mat_m(1,1) = 1.;
    mat_m(0,diri_index) = - 1/R;
    mat_m(0,neum_index) = - 1/(R*R);
    mat_m(1,diri_index) = - 1/(rad*R) + 1./(rad*rad*R*R);
    mat_m(1,neum_index) = - 1/(rad*R*R);
    
    int offset;
    double stepsize,
      RK,
      current_s;

    for (i=0; i<n_blocks;i++)
      {
	offset = i*blocklen+2;
	stepsize = pts(i+1)-pts(i);
	for (k=0; k<rk_stages;k++)
	  {
	    current_s = int_pts(i*rk_stages+k); 

	    for (l=0; l<rk_stages;l++)
	      {
		double s_l = int_pts(i*rk_stages+l);

		RK = rk_mata(k,l)*stepsize;
		mat_m(offset+2*k,offset+2*l) =RK*(rad+(DIM-3)*s_l/(s_l*s_l+wave_k*wave_k));
		mat_m(offset+2*k+1,offset+2*l) = -RK;
		mat_m(offset+2*k,offset+2*l+1) = 0.;
		mat_m(offset+2*k+1,offset+2*l+1) = RK*rad;
		
		mat_k(offset+2*k,offset+2*l+1) = RK/(s_l*s_l+wave_k*wave_k);


		mat_m(offset+2*k,diri_index) -=
		  RK*(rad*s_l+DIM-2)/(s_l*s_l+wave_k*wave_k);
		mat_m(offset+2*k,neum_index) -=
		  RK*rad/(s_l*s_l+wave_k*wave_k);
	      }	      
	    
	    mat_m(offset+2*k,offset-2) = 1.; 
	    mat_m(offset+2*k+1,offset-1) = 1.; 
	    mat_m(offset+2*k,offset+2*k) -= 1.;
	    mat_m(offset+2*k+1,offset+2*k+1) -= 1.;
	  }
      }

    // pole condition (integral over real axis)
    for (i=0; i<n_blocks;i++)
      for (l=0; l<rk_stages;l++)
	{
	  mat_m(polecond_index,2+i*blocklen+2*l) 
	    = int_weights(i*rk_stages+l)/(int_pts(i*rk_stages+l)-test_point_z);
	  mat_m(polecond_index,neum_index)
	    = 1./(test_point_z*R) - 1./(test_point_z*S) 
	    + (1./(test_point_z*test_point_z))*log((R-test_point_z)*S/(R*(S-test_point_z)));
	  mat_m(polecond_index,diri_index)
	    = (1./test_point_z)*log((R-test_point_z)*S/(R*(S-test_point_z)));
	}
    
    mat_m(diri_index, neum_index) = -1;



    (*testout) << "k = " << wave_k << ", rad = " << rad << endl;
    (*testout) << "mat_m = " << endl << mat_m << endl;
    (*testout) << "mat_k = " << endl << mat_k << endl;

    Matrix<Complex> sum(n+4,n+4), a(n+3,n+3), ainv(n+3,n+3), 
      b1(n+3,1), b2(1,n+3), c(1,1);
    ofstream out("schur.dat");

    //    for (double t = 0.1; t <= 1000; t *= 1.1)
    for (double t = 0; t <= 30; t++)
      {
	sum = (-t*t) * mat_k + mat_m;

	for (i = 0; i < n+3; i++)
	  for (j = 0; j < n+3; j++)
	    a(i,j) = sum(i,j);

	for (i = 0; i < n+3; i++)
	  {
	    b1(i,0) = sum(i,n+3);
	    b2(0,i) = sum(n+3,i);
	  }

	CalcInverse (a, ainv);
	c = b2 * ainv * b1;
	Complex s = sum(n+3,n+3) - c(0,0); 
	
	out << t << " " << s.real() << " " << s.imag() << endl;

	(*testout) << "t = " << t << endl
		   << "(psi, u_infty) = " << endl << (ainv * b1) << endl;
      }


  }


  template <int DIM>
  void Axis_Ode_Integrator<DIM> ::
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & locheap) const
  {
    FlatMatrix<double> lap_mat, mass_mat;

    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);


    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;

    elmat.AssignMemory (n*m, n*m, locheap);

    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = -mat_k(i,j)*lap_mat(k,l)+mat_m(i,j)*mass_mat(k,l);
  }


  







  /* ******************************* Real-Axis Laplace Integrator ******************** */



  template <int DIM>
  RealAxis_IE_Laplace_Integrator<DIM> ::
  RealAxis_IE_Laplace_Integrator (CoefficientFunction * acoef_lam,
				  CoefficientFunction * acoef_rad)
    : lap_int(new ConstantCoefficientFunction (1)),
      mass_int (new ConstantCoefficientFunction (1))
  {
    int i, j, k, l;
    
    Complex z(0,1);

    lam = acoef_lam -> EvaluateConst();
    rad = acoef_rad -> EvaluateConst();

    double ma[3][3] =
      {
	{ (88-7*sqrt(6.))/360, (296-169*sqrt(6.))/1800, (-2+3*sqrt(6.))/225 },
	{ (296+169*sqrt(6.))/1800, (88+7*sqrt(6.))/360, (-2-3*sqrt(6.))/225 },
	{ (16-sqrt(6.))/36, (16+sqrt(6.))/36, 1.0/9 }
      };
    double vb[3] = {(16-sqrt(6.))/36, (16+sqrt(6.))/36, 1./9 };
    double vc[3] = { (4-sqrt(6.))/10, (4+sqrt(6.))/10, 1 };
    int rk_stages = 3;

    /*
    double ma[1][1] = {{ 0.5 } };
    double vb[1] = { 1.0 };
    double vc[1] = { 0.5 };
    int rk_stages = 1;
    */

    FlatMatrix<double> rk_mata(rk_stages, rk_stages,&ma[0][0]);
    FlatVector<double> rk_vecb(rk_stages, &vb[0]), rk_vecc(rk_stages, &vc[0]);

    /*
    int n_pts = 30;
    Vector<double> pts(n_pts+1);
    for (j = 0; j <= n_pts; j++)
      pts(j) = 60 - 4 * j;
    */
    int n_pts = 29;
    Vector<double> pts(n_pts+1);
    for (j = 0; j <= n_pts; j++)
      pts(j) = sinh (4 * (1-2.0*j/n_pts));


    double R = pts(0);
    double S = pts(n_pts);

    // compute 2 integrals:
    double I1 = 0, I2 = 0;
    double tau = 0.001;
    for (double t = 0; t < 50; t += tau)
      {
	I1 += tau * (
	  1.0/6.0 * exp (-t) * rad / (t+rad*R) + 
	  2.0/3.0 * exp (-(t+tau/2)) * rad / (t+tau/2+rad*R) + 
	  1.0/6.0 * exp (-(t+tau)) * rad / (t+tau+rad*R) 
	  );
	I2 += tau * (
	  1.0/6.0 * exp (-t) * sqr (rad / (t+rad*R)) + 
	  2.0/3.0 * exp (-(t+tau/2)) * sqr (rad / (t+tau/2+rad*R)) + 
	  1.0/6.0 * exp (-(t+tau)) * sqr (rad / (t+tau+rad*R)) 
	  );
      }

    cout.precision(12);
    cout << "I1 = " << I1 << ", I2 = " << I2 << endl;

    int n = n_pts * rk_stages;

    Vector<double> int_pts(n), int_weights(n);

    for (j = 1; j <= n_pts; j++)
      for (k = 0; k < rk_stages; k++)
	{
	  int_pts((j-1)*rk_stages+k) = pts(j-1) + rk_vecc(k) * (pts(j)-pts(j-1));
	  int_weights((j-1)*rk_stages+k) = (pts(j)-pts(j-1))*rk_vecb(k);
	}

    mat_m.SetSize(n+2);
    mat_m = Complex(0.0);
    mat_k.SetSize(n+2);
    mat_k = Complex(0.0);

    mat_m(0,1) = -1;

    for (i = 0; i < n; i++)
      mat_m(1,i+2) = int_weights(i) / (int_pts(i)-z);
    mat_m(1,0) = 1.0/z * log( (R-z)*S / (R * (S-z)));
    mat_m(1,1) = 1.0/(z*R) - 1.0/(z*S) + 1.0/(z*z)*log ((R-z)*S / (R * (S-z)));


    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	{
	  mat_m(i+2,j+2) = (DIM-1) * int_pts(j) * exp (-rad*(int_pts(j)-int_pts(i)));
	  mat_k(i+2,j+2) = (int_pts(j)-int_pts(i)) * exp (-rad*(int_pts(j)-int_pts(i)));
	}

    for (i = 0; i < n_pts; i++)
      { 
	for (j = 0; j < i; j++)
	  for (k = 0; k < rk_stages; k++)
	    for (l = 0; l < rk_stages; l++)
	      {
		mat_m(i*rk_stages+k+2, j*rk_stages+l+2) *= -int_weights(j*rk_stages+l);
		mat_k(i*rk_stages+k+2, j*rk_stages+l+2) *= -int_weights(j*rk_stages+l);
	      }

	for (k = 0; k < rk_stages; k++)
	  for (l = 0; l < rk_stages; l++)
	    {
	      double tau = pts(i)-pts(i+1);
	      mat_m(i*rk_stages+k+2, i*rk_stages+l+2) *= rk_mata(k,l) * tau;
	      mat_k(i*rk_stages+k+2, i*rk_stages+l+2) *= rk_mata(k,l) * tau;
	    }

	for (j = i+1; j < n_pts; j++)
	  for (k = 0; k < rk_stages; k++)
	    for (l = 0; l < rk_stages; l++)
	      {
		mat_m(i*rk_stages+k+2, j*rk_stages+l+2) = 0;
		mat_k(i*rk_stages+k+2, j*rk_stages+l+2) = 0;
	      }
      }

    for (i = 0; i < n; i++)
      {
	mat_m(i+2,i+2) += int_pts(i) * int_pts(i) + 30*30;    // k^2

	mat_m(i+2,0) = - (int_pts(i) + (DIM-1)/rad * (1-exp (rad * (int_pts(i)-R))));
	mat_m(i+2,1) = -1 + I1 * (DIM-1) / rad * exp (rad * (int_pts(i)-R));

	mat_k(i+2,0) = -1.0/rad * exp(rad * (int_pts(i)-R)) * (I1 * int_pts(i)-1);
	mat_k(i+2,1) = -1.0/rad * exp(rad * (int_pts(i)-R)) * (I2 * int_pts(i)-I1);
      }


    (*testout) << "mat_m, now = " << endl << mat_m << endl;
    (*testout) << "mat_k, now = " << endl << mat_k << endl;
    cout << "************** YOU NEED inf_dof = " << mat_m.Height()-1 << " !!!" << endl;
  }



  template <int DIM>
  void RealAxis_IE_Laplace_Integrator<DIM> ::
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & locheap) const
  {
    FlatMatrix<double> lap_mat, mass_mat;

    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);

    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;

    elmat.AssignMemory (n*m, n*m, locheap);

    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = -mat_k(i,j)*lap_mat(k,l)+mat_m(i,j)*mass_mat(k,l);
  }


















  /* ******************************* Real-Axis ODE Laplace Integrator ******************** */

 

  template <int DIM>
  RealAxis_ODE_Laplace_Integrator<DIM> ::
  RealAxis_ODE_Laplace_Integrator (CoefficientFunction * acoef_lam,
				   CoefficientFunction * acoef_rho,
				   CoefficientFunction * acoef_rad)
    : lap_int(new ConstantCoefficientFunction (1)),
      mass_int (new ConstantCoefficientFunction (1))
  {

#ifdef divide_by_k

    int i, j, k, l;
    
    Complex z(0,1);

    lam = acoef_lam -> EvaluateConst();
    rad = acoef_rad -> EvaluateConst();

    double ma[3][3] =
      {
	{ (88-7*sqrt(6.))/360, (296-169*sqrt(6.))/1800, (-2+3*sqrt(6.))/225 },
	{ (296+169*sqrt(6.))/1800, (88+7*sqrt(6.))/360, (-2-3*sqrt(6.))/225 },
	{ (16-sqrt(6.))/36, (16+sqrt(6.))/36, 1.0/9 }
      };
    double vb[3] = {(16-sqrt(6.))/36, (16+sqrt(6.))/36, 1./9 };
    double vc[3] = { (4-sqrt(6.))/10, (4+sqrt(6.))/10, 1 };
    int rk_stages = 3;

    /*
    double ma[1][1] = {{ 0.5 } };
    double vb[1] = { 1.0 };
    double vc[1] = { 0.5 };
    int rk_stages = 1;
    */

    FlatMatrix<double> rk_mata(rk_stages, rk_stages,&ma[0][0]);
    FlatVector<double> rk_vecb(rk_stages, &vb[0]), rk_vecc(rk_stages, &vc[0]);


    int n_pts = 16;
    Vector<double> pts(n_pts+1);
    for (j = 0; j <= n_pts; j++)
      pts(j) = sinh (4 - 0.5 * j);

    /*
    int n_pts = 2;
    Vector<double> pts(n_pts+1);
    pts(0) = 0.8;
    pts(1) = 0.1;
    pts(2) = -0.5;
    */

    double R = pts(0);
    double S = pts(n_pts);

    // compute 2 integrals:
    double I1 = 0, I2 = 0;
    double tau = 0.001;
    for (double t = 0; t < 50; t += tau)
      {
	I1 += tau * (
	  1.0/6.0 * exp (-t) * rad / (t+rad*R) + 
	  2.0/3.0 * exp (-(t+tau/2)) * rad / (t+tau/2+rad*R) + 
	  1.0/6.0 * exp (-(t+tau)) * rad / (t+tau+rad*R) 
	  );
	I2 += tau * (
	  1.0/6.0 * exp (-t) * sqr (rad / (t+rad*R)) + 
	  2.0/3.0 * exp (-(t+tau/2)) * sqr (rad / (t+tau/2+rad*R)) + 
	  1.0/6.0 * exp (-(t+tau)) * sqr (rad / (t+tau+rad*R)) 
	  );
      }

    cout.precision(12);
    cout << "I1 = " << I1 << ", I2 = " << I2 << endl;

    int n = n_pts * rk_stages;

    Vector<double> int_pts(n), int_weights(n);

    for (j = 1; j <= n_pts; j++)
      for (k = 0; k < rk_stages; k++)
	{
	  int_pts((j-1)*rk_stages+k) = pts(j-1) + rk_vecc(k) * (pts(j)-pts(j-1));
	  int_weights((j-1)*rk_stages+k) = (pts(j)-pts(j-1))*rk_vecb(k);
	}

    mat_m.SetSize(2*n+4);
    mat_m = Complex(0.0);
    mat_k.SetSize(2*n+4);
    mat_k = Complex(0.0);

    // Neumann data
    mat_m(0,1) = -1;

    // pole condition
    for (i = 0; i < n; i++)
      mat_m(1,2*i+2) = int_weights(i) / (int_pts(i)-z);
    mat_m(1,0) = 1.0/z * log( (R-z)*S / (R * (S-z)));
    mat_m(1,1) = 1.0/(z*R) - 1.0/(z*S) + 1.0/(z*z)*log ((R-z)*S / (R * (S-z)));

    // initial conditions
    mat_m(2,2) = -1;
    mat_m(2,0) = 1/R;
    mat_m(2,1) = 1/(R*R);

    mat_m(3,3) = -1;
    mat_m(3,0) = 1/(rad*R) - 1/(rad*rad*R*R);
    mat_m(3,1) = 1/(rad*R*R);

    double wave = 30;

    for (i = 0; i < n_pts; i++)
      {
	int base = 2*i*rk_stages+4;
	for (j = 0; j < rk_stages; j++)
	  {
	    mat_m(base+2*j, base-2) = 1;
	    mat_m(base+2*j, base+2*j) = -1;
	    mat_m(base+2*j+1, base-2+1) = 1;
	    mat_m(base+2*j+1, base+2*j+1) = -1;
	  }

	for (j = 0; j < rk_stages; j++)
	  for (k = 0; k < rk_stages; k++)
	    {
	      double s = int_pts(i*rk_stages+k);
	      double fac = (pts(i+1) - pts(i)) * rk_mata(j,k);

	      mat_m(base+2*j, base+2*k) += fac * (rad + (DIM-3)*s / (s*s+wave*wave));
	      mat_m(base+2*j+1, base+2*k) -= fac;
	      mat_m(base+2*j+1, base+2*k+1) += fac * rad;

	      mat_k(base+2*j, base+2*k+1) += fac / (s*s+wave*wave);

	      mat_m(base+2*j, 0) += -fac * (rad*s+DIM-2) / (s*s+wave*wave);
	      mat_m(base+2*j, 1) += -fac * rad / (s*s+wave*wave);
	    }
      }

    (*testout) << "mat_m, now = " << endl << mat_m << endl;
    (*testout) << "mat_k, now = " << endl << mat_k << endl;
    cout << "************** YOU NEED inf_dof = " << mat_m.Height()-1 << " !!!" << endl;
#endif



    int i, j, k, l;
    
    Complex z(10,10);

    lam = acoef_lam -> EvaluateConst();
    rho = acoef_rho -> EvaluateConst();
    rad = acoef_rad -> EvaluateConst();


    // New: Needs Lobatto formula !!!
    // Runge-Kutte matrix: veca0...first column vector,  mata .... other cols of Butcher scheme
    // same for integration weights b = (b0, tk_vecb)

    /*
    double ma[3][3] =
      {
	{ (88-7*sqrt(6.))/360, (296-169*sqrt(6.))/1800, (-2+3*sqrt(6.))/225 },
	{ (296+169*sqrt(6.))/1800, (88+7*sqrt(6.))/360, (-2-3*sqrt(6.))/225 },
	{ (16-sqrt(6.))/36, (16+sqrt(6.))/36, 1.0/9 }
      };
    double ma0[3] = { 0, 0, 0 };
    double vb[3] = {(16-sqrt(6.))/36, (16+sqrt(6.))/36, 1./9 };
    double b0 = 0.0;
    double vc[3] = { (4-sqrt(6.))/10, (4+sqrt(6.))/10, 1 };
    int rk_stages = 3;
    */
    /*
    // lobatto4A
    double ma[2][2] =
      {
	{ 1.0/3.0, -1.0/24.0 },
	{ 2.0/3.0, 1.0/6.0 }
      };
    double ma0[2] = { 5.0/24.0, 1.0/6.0 };
    double vb[2] = { 2.0/3.0, 1.0/6.0 };
    double b0 = 1.0/6.0;
    double vc[2] = { 0.5, 1.0 };
    int rk_stages = 2;
    */

    // trapezoidal rule:
    double ma[1][1] = {{ 0.5 } };
    double ma0[1] = { 0.5  };  
    double vb[1] = { 0.5 };
    double b0 = 0.5;
    double vc[1] = { 1 };
    int rk_stages = 1;

    /*
    // implicit Euler rule:
    double ma[1][1] = {{ 1 } };
    double ma0[1] = { 0  };  
    double vb[1] = { 1 };
    double b0 = 0;
    double vc[1] = { 1 };
    int rk_stages = 1;
    */


    FlatMatrix<double> rk_mata(rk_stages, rk_stages,&ma[0][0]);
    FlatVector<double> rk_mata0(rk_stages, &ma0[0]);
    FlatVector<double> rk_vecb(rk_stages, &vb[0]), rk_vecc(rk_stages, &vc[0]);


    int n_pts = 16;
    Vector<double> pts(n_pts+1);
    for (j = 0; j <= n_pts; j++)
      pts(j) = sinh (4 - 0.5 * j);

    /*
    int n_pts = 2;
    Vector<double> pts(n_pts+1);
    pts(0) = 0.8;
    pts(1) = 0.1;
    pts(2) = -0.5;
    */

    double R = pts(0);
    double S = pts(n_pts);

    // compute 2 integrals:
    double I1 = 0, I2 = 0;
    double tau = 0.001;
    for (double t = 0; t < 50; t += tau)
      {
	I1 += tau * (
	  1.0/6.0 * exp (-t) * rad / (t+rad*R) + 
	  2.0/3.0 * exp (-(t+tau/2)) * rad / (t+tau/2+rad*R) + 
	  1.0/6.0 * exp (-(t+tau)) * rad / (t+tau+rad*R) 
	  );
	I2 += tau * (
	  1.0/6.0 * exp (-t) * sqr (rad / (t+rad*R)) + 
	  2.0/3.0 * exp (-(t+tau/2)) * sqr (rad / (t+tau/2+rad*R)) + 
	  1.0/6.0 * exp (-(t+tau)) * sqr (rad / (t+tau+rad*R)) 
	  );
      }

    cout.precision(12);
    cout << "I1 = " << I1 << ", I2 = " << I2 << endl;

    int n = n_pts * rk_stages;

    Vector<double> int_pts(n+1), int_weights(n+1);

    int_pts(0) = pts(0);
    int_weights = 0.0;

    for (j = 1; j <= n_pts; j++)
      {
	int_weights((j-1)*rk_stages) += (pts(j)-pts(j-1))*b0;
	for (k = 0; k < rk_stages; k++)
	  {
	    int_pts((j-1)*rk_stages+k+1) = pts(j-1) + rk_vecc(k) * (pts(j)-pts(j-1));
	    int_weights((j-1)*rk_stages+k+1) += (pts(j)-pts(j-1))*rk_vecb(k);
	  }
      }

    (*testout) << "int_pts = " << endl << int_pts << endl;
    (*testout) << "int_weights = " << endl << int_weights << endl;

    mat_m.SetSize(2*n+4);
    mat_m = Complex(0.0);
    mat_k.SetSize(2*n+4);
    mat_k = Complex(0.0);

    // Neumann data
    mat_m(0,1) = -lam;

    // pole condition
    for (i = 0; i <= n; i++)
      mat_m(1,2*i+2) = lam * int_weights(i) / (int_pts(i)-z);
    mat_m(1,0) = lam/z * log( (R-z)*S / (R * (S-z)));
    mat_m(1,1) = lam * (1.0/(z*R) - 1.0/(z*S) + 1.0/(z*z)*log ((R-z)*S / (R * (S-z))));

    // initial conditions
    mat_m(2,2) = -lam;
    mat_m(2,0) = lam/R;
    mat_m(2,1) = lam/(R*R);

    mat_m(3,3) = -lam;
    mat_m(3,0) = lam/(rad*R) - lam/(rad*rad*R*R);
    mat_m(3,1) = lam/(rad*R*R);

    // double wave = 30;

    for (i = 0; i < n_pts; i++)
      {
	int base = 2*i*rk_stages+4;
	for (j = 0; j < rk_stages; j++)
	  {
	    double s = int_pts(i*rk_stages+j+1);

	    mat_m(base+2*j, base-2) = lam*pts(i)*pts(i)-rho;
	    mat_m(base+2*j, base+2*j) = - (lam*s*s-rho);

	    mat_m(base+2*j+1, base-2+1) = lam;
	    mat_m(base+2*j+1, base+2*j+1) = -lam;
	  }

	for (j = 0; j < rk_stages; j++)
	  {
	    // first col for Butcher scheme:

	    double s = int_pts(i*rk_stages);
	    double fac = (pts(i+1) - pts(i)) * rk_mata0(j);
		
	    mat_m(base+2*j, base-2) += fac * (rad * (lam*s*s-rho) + lam*(DIM-1)*s);
	    mat_m(base+2*j+1, base-2) -= fac * lam;
	    mat_m(base+2*j+1, base-1) += fac * rad * lam;
		
	    mat_k(base+2*j, base-1) += fac * lam;
		
	    mat_m(base+2*j, 0) += -fac * lam * (rad*s+DIM-2);
	    mat_m(base+2*j, 1) += -fac * rad * lam;
	    

	    for (k = 0; k < rk_stages; k++)
	      {
		double s = int_pts(i*rk_stages+k+1);
		double fac = (pts(i+1) - pts(i)) * rk_mata(j,k);
		
		mat_m(base+2*j, base+2*k) += fac * (rad * (lam*s*s-rho) + lam*(DIM-1)*s);
		mat_m(base+2*j+1, base+2*k) -= fac * lam;
		mat_m(base+2*j+1, base+2*k+1) += fac * rad * lam;
		
		mat_k(base+2*j, base+2*k+1) += fac * lam;
		
		mat_m(base+2*j, 0) += -fac * lam * (rad*s+DIM-2);
		mat_m(base+2*j, 1) += -fac * rad * lam;
	      }
	  }
      }

    (*testout) << "mat_m, now = " << endl << mat_m << endl;
    (*testout) << "mat_k, now = " << endl << mat_k << endl;
    cout << "************** YOU NEED inf_dof = " << mat_m.Height()-1 << " !!!" << endl;
  }
  

  template <int DIM>
  void RealAxis_ODE_Laplace_Integrator<DIM> ::
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & locheap) const
  {
    FlatMatrix<double> lap_mat, mass_mat;

    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);

    (*testout) << "lap_mat = " << endl << lap_mat << endl;
    (*testout) << "mass_mat = " << endl << mass_mat << endl;




    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;

    elmat.AssignMemory (n*m, n*m, locheap);

    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = -mat_k(i,j)*lap_mat(k,l)+mat_m(i,j)*mass_mat(k,l);
  }












  Hardy_Integrator2D ::
  Hardy_Integrator2D (CoefficientFunction * acoef_k,
		      CoefficientFunction * acoef_faclaplace,
		      CoefficientFunction * acoef_rad,
		      CoefficientFunction * acoef_s0r,
		      CoefficientFunction * acoef_s0i)

    : lap_int(new ConstantCoefficientFunction (1)),
      mass_int (new ConstantCoefficientFunction (1))
  {
    int i, j;

    wave_k = acoef_k -> EvaluateConst();
    fac_laplace = acoef_faclaplace -> EvaluateConst();
    rad = acoef_rad -> EvaluateConst();

    s0 = Complex(acoef_s0r -> EvaluateConst(),
		 acoef_s0i -> EvaluateConst());
		 
    const int dimH = 4;

    Matrix<Complex> Diffop(dimH, dimH);
    Matrix<Complex> Intop(dimH, dimH);
    Matrix<Complex> T1(dimH, dimH);
    Matrix<Complex> T2(dimH, dimH);

    // compute T1
    T1 = Complex(0.0);
    T1(0,0) = - 1./sqrt(-2.*s0);
    for (i=1; i<dimH;i++)
      {
	T1(i-1,i) = 1.;
	T1(i,i) = 1.;
      }

    // compute T2
    T2 = Complex (0.0);
    T2(0,0) = - 1./sqrt(-2.0*s0);
    for (i=1; i<dimH;i++)
      {
	T2(i-1,i) = -1.;
	T2(i,i) = 1.;
      }

    SetupDiffOp(Diffop);

    Matrix<Complex> DiffopLong(2*dimH-1, 2*dimH-1);
    Matrix<Complex> IntopLong(2*dimH-1, 2*dimH-1);

    SetupDiffOp(DiffopLong);

    CalcInverse(DiffopLong,IntopLong);
    for (i=0;i<dimH;i++)
      for (j=0;i<dimH;i++)
	Intop(i,j) = IntopLong(i,j);

    *testout << "diffoplong = " << endl << DiffopLong << endl;

    *testout << "intoplong = " << endl << IntopLong << endl;

    mat_m.SetSize(dimH,dimH);
    mat_mkappa.SetSize(dimH,dimH);
    mat_k.SetSize(dimH,dimH);

    mat_m = (-s0*s0/rad)*Trans(T1)*Diffop*T1;
    mat_mkappa = (-1./rad)*Trans(T2)*Diffop*T2;
    mat_k = (1./rad)*Trans(T2)*Intop*T2;

    cout << "mat_m = " << endl << mat_m << endl;
    cout << "mat_mkappa = " << endl << mat_mkappa << endl;
    cout << "mat_k = " << endl << mat_k << endl;

    cout << "************** YOU NEED inf_per_node = " << mat_m.Height()-1 << " !!!" << endl;
  }

void Hardy_Integrator2D::SetupDiffOp(Matrix<Complex> &Diffop)
{
  Diffop = Complex (0.);
  int i,j,
    n = Diffop.Height();


  Diffop(0,0) = rad - 1./(2.*s0);
  Diffop(0,1) = 1./(2.*s0);
  for (i = 1; i<n-1; i++)
    {
      Diffop(i,i-1) = double (i)/(2.*s0);
      Diffop(i,i) = rad - (2.*i+1.)/(2.*s0);
      Diffop(i,i+1) = (i+1.)/(2.*s0);
    }
  i = n-1;
  Diffop(i,i-1) = double (i)/(2.*s0);
  Diffop(i,i) = rad - (2.*i+1.)/(2.*s0);
}

void Hardy_Integrator2D ::
  AssembleElementMatrix (const FiniteElement & fel, 
			 const ElementTransformation & eltrans, 
			 FlatMatrix<Complex> & elmat,
			 LocalHeap & locheap) const
  {
    FlatMatrix<double> lap_mat, mass_mat;

    lap_int.AssembleElementMatrix (fel, eltrans, lap_mat, locheap);
    mass_int.AssembleElementMatrix (fel, eltrans, mass_mat, locheap);

    int n = mat_m.Height();
    int m = lap_mat.Height();
    int i, j, k, l;

    elmat.AssignMemory (n*m, n*m, locheap);

    for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
	for (k = 0; k < m; k++)
	  for (l = 0; l < m; l++)
	    elmat(i*m+k, j*m+l) = 
	      fac_laplace * mat_k(i,j)*lap_mat(k,l)
	      - fac_laplace * mat_m(i,j) * mass_mat(k,l)
	      -wave_k * mat_mkappa(i,j) * mass_mat(k,l);
  }















  ///
  class NumProcGS : public NumProc
  {
  protected:
    ///
    BilinearForm * bfa;
    ///
    LinearForm * lff;
    ///
    GridFunction * gfu;
    ///
    int maxsteps;
  public:
    ///
    NumProcGS (PDE & apde, const Flags & flags);
    ///
    virtual ~NumProcGS();

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcGS (pde, flags);
    }


    ///
    virtual void Do();
    ///
    virtual string GetClassName () const
    {
      return "Polecondition - GS";
    }

  };



  NumProcGS :: NumProcGS (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {
    bfa = pde.GetBilinearForm (flags.GetStringFlag ("bilinearform", ""));
    lff = pde.GetLinearForm (flags.GetStringFlag ("linearform", ""));
    gfu = pde.GetGridFunction (flags.GetStringFlag ("gridfunction", ""));
    maxsteps = int(flags.GetNumFlag ("maxsteps", 200));
  }

  NumProcGS :: ~NumProcGS()
  {
    ;
  }


  void NumProcGS :: Do()
  {
    cout << "solve exterior Helmholtz problem" << endl;

    const BaseMatrix & mat = bfa->GetMatrix();
    const BaseVector & vecf = lff->GetVector();
    BaseVector & vecu = gfu->GetVector();

    const BaseMatrix & preabs = pde.GetPreconditioner("cabs") -> GetMatrix();

    cout << "has objects" << endl;

    Table<int> & blocks = *gfu->GetFESpace().CreateSmoothingBlocks();
    (*testout) << "smoothing blocks = " << endl << blocks << endl;

    BaseBlockJacobiPrecond & pre = 
      * dynamic_cast<const BaseSparseMatrix&> (mat) . CreateBlockJacobiPrecond (blocks);
 
    cout << "haspre" << endl;

    vecu = 0;

    BaseVector & d = *vecu.CreateVector();
    BaseVector & w = *vecu.CreateVector();
    
    for (int i = 0; i < maxsteps; i++)
      {
	d = vecf - mat * vecu;
	cout << "res = " << d.L2Norm() << endl;

	//	w = pre * d;
	//	vecu += 0.1 * w;
	pre.GSSmoothBack (vecu, vecf, 5);


	d = vecf - mat * vecu;
	w = 0;
	w = preabs * d;
	vecu += w;
      }

    
  }

  namespace {
    
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetIntegrators().AddBFIntegrator ("cut_ieq", 2, 2,
					Cut_Ieq_Integrator<2>::Create);
      GetIntegrators().AddBFIntegrator ("cut_ieq", 3, 2,
					Cut_Ieq_Integrator<3>::Create);
      GetIntegrators().AddBFIntegrator ("axis_ode", 2, 2,
					Axis_Ode_Integrator<2>::Create);
      GetIntegrators().AddBFIntegrator ("axis_ode", 3, 2,
					Axis_Ode_Integrator<3>::Create);

      GetIntegrators().AddBFIntegrator ("realaxis_ie_laplace", 2, 2,
					RealAxis_IE_Laplace_Integrator<2>::Create);
      GetIntegrators().AddBFIntegrator ("realaxis_ODE_laplace", 2, 3,
					RealAxis_ODE_Laplace_Integrator<2>::Create);

      GetIntegrators().AddBFIntegrator ("hardysymmetric", 2, 5,       // 2 = spacedim, 5 = number of parameters  
					Hardy_Integrator2D::Create);

      GetNumProcs().AddNumProc ("polecond", NumProcGS::Create, NumProcGS::PrintDoc);
    }

    Init init;
  }
  
}

