#ifndef FILE_FOURIER
#define FILE_FOURIER

//FourierFESpace****************************************************************************
class FourierFESpace : public FESpace
{
private:
  int ndof;
  int boundary;
  bool above;

public:

  FourierFESpace (const MeshAccess & ama, const Flags & aflags);

  virtual ~FourierFESpace () { ; }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    return new FourierFESpace(ma, flags);
  }

  virtual string GetClassName () const
  {
    return "FourierFESpace";
  }


  virtual void Update()
  {;}

  virtual int GetNDof () const
  {
    return ndof; 
  }

  virtual int GetBoundary () const
  {
    return boundary; 
  }


  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const
  { dnums.SetSize(0); return; }
  virtual void GetSDofNrs (int elnr, ARRAY<int> & dnums) const; 

  const FiniteElement &  GetFE (int elnr, LocalHeap & lh) const 
  { return * segm; }

};



//Periodic2DSpace*****************************************************************

class Periodic2DSpace : public H1HighOrderFESpace
{
  private:
    ARRAY<int> dofmap;
    mutable Complex gamma ;
  public:

    Periodic2DSpace (const MeshAccess & ama, const Flags & flags);
  ///
    virtual ~Periodic2DSpace ();

  void SetGamma(Complex agamma)const
  {
    gamma=agamma;
  }

  static FESpace * Create (const MeshAccess & ma, const Flags & flags)
  {
    return new Periodic2DSpace (ma, flags);
  }

  virtual string GetClassName () const
  {
    return "Periodic2DSpace";
  }

  virtual void Update ();
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;

  virtual void VTransformMC (int elnr, bool boundary,
			     FlatMatrix<Complex> & mat, TRANSFORM_TYPE type) const;
  virtual void VTransformVC (int elnr, bool boundary,
			     FlatVector<Complex> & vec, TRANSFORM_TYPE type) const;


  int DofMap (int i) const { return dofmap[i]; }
  Complex GetGamma () const { return gamma; }
};



//FESpaceDefract***************************************************************************

class FESpaceDefract : public CompoundFESpace
{
  public:
    FESpaceDefract (const MeshAccess & ama,
		    const ARRAY<const FESpace*> & aspaces,
		    const Flags & aflags)
      : CompoundFESpace (ama, aspaces, aflags)
    { ; }

    virtual ~FESpaceDefract () { ; } 

    static FESpace* Create (const MeshAccess & ma, const Flags & flags)
    {
     ARRAY<const FESpace*> spaces(3);

     cout << "Create Defract FE Space" << endl;
    
     int order = int (flags.GetNumFlag ("order", 1.0));

     Flags h1flags;
     h1flags.SetFlag ("order", flags.GetNumFlag ("order", 1.0));
     if (flags.GetDefineFlag ("eliminate_internal")) h1flags.SetFlag ("eliminate_internal");
     if (flags.GetDefineFlag ("complex")) h1flags.SetFlag ("complex");
     spaces[0] = new Periodic2DSpace (ma, h1flags);
    
     Flags fourierflags, fourier2flags;
     fourierflags.SetFlag ("fourierorder", flags.GetNumFlag ("fourier1order",0));
     fourierflags.SetFlag ("bc", flags.GetNumFlag ("bc1",1));
     fourierflags.SetFlag ("above");
     spaces[1] = new FourierFESpace (ma, fourierflags);

     fourier2flags.SetFlag ("fourierorder", flags.GetNumFlag ("fourier2order",0));
     fourier2flags.SetFlag ("bc", flags.GetNumFlag ("bc2",2));
     spaces[2] = new FourierFESpace (ma, fourier2flags);
      
     FESpaceDefract * fes = new FESpaceDefract (ma, spaces, flags);
     return fes;
   }

    virtual string GetClassName () const { return "Defract-FESpace"; }

  };


//****************************Special Elements**********************************
//FourrierElement******************************************************************
  class FourierElement : public SpecialElement
  {
    ARRAY<Vec<2,Complex> >& k;
    ARRAY<Complex[2]>& S;

    int normal;  /// 1 is above the grating, -1 is under the grating
    int ndof_k;
    int ndof;
    int ndof_f;
    int kSize;
    int elnr;

    double alpha;
    CoefficientFunction *coeff;
    ARRAY<int> dnums;
    ElementTransformation eltrans;
    const FESpaceDefract& space;

  public:
    FourierElement (ARRAY<Vec<2,Complex> >& ak,
                    ARRAY<Complex[2]>& aS,
		    int anormal,int aelnr,
		    CoefficientFunction *acoeff,
		    ARRAY<int>& adnums,
		    ElementTransformation aeltrans,
		    const FESpaceDefract& aspace);

    virtual ~FourierElement ()  {;}

    virtual void GetDofNrs (ARRAY<int> & nums) const
    {
      nums = dnums; 
    }


    void Assemble (FlatMatrix<Complex> & elmat,LocalHeap & lh) const;

    void DoIntegrationAbove (FlatMatrix<Complex> & elmat,Vector<Complex>& shapes, Vec<2> p,double det, double coef, double weight) const;

    void DoIntegrationBelow (FlatMatrix<Complex> & elmat,Vector<Complex>& shapes, Vec<2> p,double det, double coef, double weight) const;
  };

//IncomingWave************************************************************************************

class IncomingWave : public SpecialElement
{
    int pos;
    double penalty;


  public:
   IncomingWave (int apos,double apenalty) : pos(apos), penalty(apenalty) {;}

   virtual ~IncomingWave () {;} 

   virtual void GetDofNrs (ARRAY<int> & dnums) const;

   void Assemble (FlatMatrix<Complex> & elmat,LocalHeap & lh) const;

  };

//NumProcBVPFourier******************************************************
  class NumProcBVPFourier : public NumProc
  {
  protected:

    BilinearForm * bfa;
    LinearForm * lff;
    GridFunction * gfu;
 
    bool print;
    double lambda;
    double alpha;
    double penalty;

    ofstream * out1, *out2;

    ifstream *layerfile;

    int pol;
    bool layer;

  public:
    NumProcBVPFourier (PDE & apde, const Flags & flags);

    virtual ~NumProcBVPFourier() { ; }

    static NumProc * Create (PDE & pde, const Flags & flags)
    {
      return new NumProcBVPFourier (pde, flags);
    }

    void CalcSMatrix(int dim, ARRAY<Complex[2]>& S, ARRAY<Vec<2,Complex> >& kl,Complex& n_layer); 

    void Output(int h, BaseVector& v,ARRAY<Vec<2,Complex> >& ka,ARRAY<Vec<2,Complex> >& ku,Complex na, Complex nu);

    virtual void Do(LocalHeap & lh);

    virtual string GetClassName () const
    {
      return "Fourrier Problem";
    }

    virtual void PrintReport (ostream & ost)
    {
      ost << GetClassName() << endl
	  << "Bilinear-form = " << bfa->GetName() << endl
	  << "Linear-form   = " << lff->GetName() << endl
	  << "Gridfunction  = " << gfu->GetName() << endl;
    }

    ///
    static void PrintDoc (ostream & ost);
  };





#endif





