/**
   High Order Finite Element Space for Facets for EE with trig and quad order 
 */
 /* facetfespace.*pp
 * An fespace for functions living on the facets (=edges in 2D, faces in 3D)
 *   - the functions on different facets are independent (no continuity accross
 *     vertices ( or edges in 3D))
 *   - the functions on the facets are accessed via the elements
 *     (=FacetVolumeElements), which then can access their facets;
 *     see facetfe.hpp for details
 *   - the ordering of dofs is facet'wise
 *   - no low-order space
 *   - no prolongation so far (will probably never be needed)
 *   - some additional utility functions 
 */
#ifndef FACET_FESPACE_EE_HPP
#define FACET_FESPACE_EE_HPP

#include <fem.hpp>
#include <comp.hpp> 

namespace ngcomp
{
  using namespace ngcomp;

class FacetFESpaceEE : public FacetFESpace 
{
protected:  
  int trig_order; 
  int quad_order; 
  int trig_relorder; 
  int quad_relorder; 
  
public:
    ///
    FacetFESpaceEE (const MeshAccess & ama, const Flags & flags, bool parseflags=false);
    ///
    virtual ~FacetFESpaceEE ();
    ///
    static FESpace * Create (const MeshAccess & ma, const Flags & flags);
    ///
    virtual string GetClassName () const
    {
      return "FacetFESpaceEE";
    }
    ///
    virtual void Update();
};

}



#endif
