#ifndef FILE_HDIVHOFESZ_
#define FILE_HDIVHOFESZ_ 
/*********************************************************************/
/* File:   hdivhofe.hpp                                              */
/* Author: Almedin Becirovic                                         */
/* Date:   15. Feb. 2003                                             */
/*********************************************************************/


#define DIVTET

/// namespace for finite elements
namespace ngfem
{
 

///
template <int D>
class HDivHighOrderFiniteElementSZ : public HDivFiniteElement<D>
{

  //public:
  // enum { DIM = D };

protected:
public:
  int vnums[8];

  INT<3> order_inner_curl;
  INT<3> order_inner;
  INT<2> order_face[6];

  int order_edge[12];

  int ned; // number of edges in element
  int nv; // number of vertices in element
  int nf; // number of faces in element

  //SZ int ndof_edge;
  //SZ int ndof_face;
  //SZ int ndof_inner;

  bool augmented;
  bool outer_normal; 
  int discont; // 0 ... continuous, 1 ... discont (all internal), 
               // 2 ... disont (lowest-order are external) 

public:
  ///
  HDivHighOrderFiniteElementSZ (ELEMENT_TYPE aeltype);

  void SetVertexNumbers (FlatArray<int> & avnums);
  void SetOrderEdge(FlatArray<int> & oe);
  void SetOrderFace(FlatArray<int> & of);
  void SetOrderFace(FlatArray<INT<2> > & of);
  void SetOrderInner (int oi);
  void SetOrderInner (INT<3> oi);
  void SetOrderInnerCurl (int oi);
  void SetOrderInnerCurl (INT<3> oi);
  void SetRelOrderCurl (int rel); 
  void SetDiscontinuous(int d) {discont = d;}; 
  
  virtual void GetFacetDofs(int i, ARRAY<int> & dnums) const
  { *testout  << " GetFacetDofs for nothing " << endl; dnums.SetSize(0);}; 


  // virtual void GetInternalDofs (ARRAY<int> & idofs)  const; 
  virtual void ComputeNDof () = 0;

  int EdgeOrientation (int enr) const
  {
    const EDGE * edges = ElementTopology::GetEdges (this->eltype);
    return (vnums[edges[enr][1]] > vnums[edges[enr][0]]) ? 1 : -1;
  }

 
};


template <int D>
class HDivHighOrderNormalFiniteElementSZ : public HDivNormalFiniteElement<D>
{

  //public:
  // enum { DIM = D };

protected:
  int vnums[4];
  INT<2> order_inner;
  INT<2> order_inner_curl;
  int ned; // number of edges in element
  int nv; // number of vertices in element

  int discont; 
  bool augmented;

public:
  ///
  HDivHighOrderNormalFiniteElementSZ (ELEMENT_TYPE aeltype);


  void SetVertexNumbers (FlatArray<int> & avnums);
  void SetOrderInner (int oi);
  void SetOrderInner (INT<3> oi);
  void SetOrderInnerCurl (int oi);
  void SetOrderInnerCurl (INT<3> oi);

   void SetRelOrderCurl (int rel); 

  virtual void ComputeNDof () = 0;

  int EdgeOrientation (int enr) const
  {
    const EDGE * edges = ElementTopology::GetEdges (this->eltype);
    return (vnums[edges[enr][1]] > vnums[edges[enr][0]]) ? 1 : -1;
  }
};


template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderNormalSegmSZ : public HDivHighOrderNormalFiniteElementSZ<1>
{
public:

  HDivHighOrderNormalSegmSZ (int aorder);
  virtual void ComputeNDof();

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			   FlatVector<> shape) const;

};


template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderNormalTrigSZ : public HDivHighOrderNormalFiniteElementSZ<2>
{
private: 
  typedef TrigShapesInnerLegendre T_INNERSHAPES; 
public:

  HDivHighOrderNormalTrigSZ (int aorder);
  virtual void ComputeNDof();

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			   FlatVector<> shape) const;

  

};

template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderNormalQuadSZ : public HDivHighOrderNormalFiniteElementSZ<2>
{
public:

  HDivHighOrderNormalQuadSZ (int aorder);
  virtual void ComputeNDof();

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			   FlatVector<> shape) const;

};

///
template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderTrigSZ : public HDivHighOrderFiniteElementSZ<2>
{
private: 
typedef TrigShapesInnerLegendre T_INNERSHAPES; 
public:

  HDivHighOrderTrigSZ (int aorder);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			  FlatMatrixFixWidth<2> shape) const;

  /// compute Div of shape
  virtual void CalcDivShape (const IntegrationPoint & ip,
			     FlatVector<> shape) const;



  /// compute Div numerical diff
  //void CalcNumDivShape( const IntegrationPoint & ip,
  //			FlatVector<> divshape) const;
  virtual void GetFacetDofs(int i, ARRAY<int> & dnums) const; 

};

///

template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderQuadSZ : public HDivHighOrderFiniteElementSZ<2>
{
public:

  HDivHighOrderQuadSZ (int aorder);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			  FlatMatrixFixWidth<2> shape) const;


  /// compute Div of shape
  virtual void CalcDivShape (const IntegrationPoint & ip,
			     FlatVector<> shape) const;
  /// compute Div numerical diff
  void CalcNumDivShape( const IntegrationPoint & ip,
  		FlatVector<> divshape) const;

  virtual void GetFacetDofs(int i, ARRAY<int> & dnums) const; 

};

template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderTetSZ : public HDivHighOrderFiniteElementSZ<3>
{
private:
  typedef TetShapesInnerLegendre T_INNERSHAPES;
  typedef TetShapesFaceLegendre T_FACESHAPES; 
public:

  HDivHighOrderTetSZ (int aorder);
  virtual void ComputeNDof();
  virtual void GetInternalDofs (ARRAY<int> & idofs) const;

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			  FlatMatrixFixWidth<3> shape) const;
#ifdef DIVTET
  /// compute Div of shape
  virtual void CalcDivShape (const IntegrationPoint & ip, FlatVector<> shape) const;
#endif
  /// compute Div numerical diff
  //void CalcNumDivShape( const IntegrationPoint & ip,
  //			FlatVector<> divshape) const;

  
};

template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderPrismSZ : public HDivHighOrderFiniteElementSZ<3>
{
  typedef TrigShapesInnerLegendre T_TRIGFACESHAPES;
public:

  HDivHighOrderPrismSZ (int aorder);
  virtual void ComputeNDof();
  //   virtual void GetInternalDofs (ARRAY<int> & idofs) const;
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			  FlatMatrixFixWidth<3> shape) const;

  /// compute Div of shape
  // virtual void CalcDivShape (const IntegrationPoint & ip,
  //			     FlatVector<> shape) const;
  /// compute Div numerical diff
  //void CalcNumDivShape( const IntegrationPoint & ip,
  //			FlatVector<> divshape) const;


};

template <class T_ORTHOPOL = IntegratedLegendreMonomialExt>
class HDivHighOrderHexSZ : public HDivHighOrderFiniteElementSZ<3>
{
public:

  HDivHighOrderHexSZ (int aorder);
  virtual void ComputeNDof();
   virtual void GetInternalDofs (ARRAY<int> & idofs) const;
 
  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip,
			  FlatMatrixFixWidth<3> shape) const;

  /// compute Div of shape
  virtual void CalcDivShape (const IntegrationPoint & ip,
			     FlatVector<> shape) const;
  /// compute Div numerical diff
  //void CalcNumDivShape( const IntegrationPoint & ip,
  //			FlatVector<> divshape) const;


};

} // namespace

#endif





