 
#ifndef FILE_HDIVHOFESPACESZ

#define FILE_HDIVHOFESPACESZ

/*********************************************************************/
/* File:   hdivhofespace.hpp                                         */
/* Author: Sabine Zaglmayr                                           */
/* Date:   15. Dec. 2006                                             */
/*********************************************************************/

/**
   HDiv High Order Finite Element Space
*/

#include <comp.hpp>

namespace ngcomp
{
  

class HDivHighOrderFESpaceSZ : public FESpace
{
protected:
  
  // Number of Edges
  int ned;
  // Number of Faces
  int nfa;
  // Number of Elements
  int nel;
  // Number of Vertex
  int nv;
  // Number Dofs 
  int ndof;
  
  // tables of first facet and inner dofs 
  ARRAY<int> first_facet_dof;
  ARRAY<int> first_inner_dof;

  /// relative order to mesh-order
  int rel_order; 
  // relative curl order 
  int rel_curl_order; 
  // order of curls 
  int curl_order; 
  // bool if chosen variable space 
  bool var_order;
  // space is continuous/discontinuous 
  bool discont; 
 
  // arrays for various orders and fince facets 
  ARRAY<INT<3> > order_inner;
  ARRAY<INT<3> > order_inner_curl;
  ARRAY<INT<2> > order_facet; 
  ARRAY<bool> fine_facet; 
 

  ARRAY<int> ndlevel;

  // choose uniform fixed order on facets or elems 
  int uniform_order_inner; 
  int uniform_order_facet; 
 

  

  bool print; 
   
public:

  HDivHighOrderFESpaceSZ (const MeshAccess & ama, const Flags & flags);
  ///
  virtual ~HDivHighOrderFESpaceSZ ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

   virtual string GetClassName () const
  {
    return "HDivHighOrderFESpaceSZ";
  }

  ///
  virtual void Update();
  /// 
  virtual void UpdateDofTables(); 
  ///
  virtual int GetNDof () const;
  ///
  virtual int GetNDofLevel (int level) const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int selnr, LocalHeap & lh) const; // 2D: array =0.;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetExternalDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  ///
  virtual Table<int> * CreateSmoothingBlocks (const Flags & precflags) const;

  /// 
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;
  /// 
  virtual void GetFacetDofNrs(int fanr, ARRAY<int> & dnums) const 
  { 
    if (ma.GetDimension() == 2) GetEdgeDofNrs(fanr,dnums); 
    else if (ma.GetDimension() == 3) GetFaceDofNrs(fanr,dnums); 
  } 
			     
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;
  /// 
  void GetFacetOrder (ARRAY<INT<2> > & of, ARRAY<bool> & ff) const 
  {of = order_facet; ff = fine_facet;};
  /// 
  int GetFirstInnerDof(int elnr) const { return(first_inner_dof[elnr]);}; 

  int GetFirstFacetDof(int fanr) const; 
 
  
  int GetNElemDofs(int elnr) const 
  {
    if(discont) return(first_inner_dof[elnr+1] - first_inner_dof[elnr]); 
    else 
      { 
	ARRAY<int> dnums; 
	this->GetDofNrs(elnr,dnums); 
	return(dnums.Size()); 
      } 
  }
  
};
}
#endif





