/*********************************************************************/
/* File:   l2hofespace_ee.cpp                                        */
/* Author: SZ                                                        */
/* Date:   1. März 2007                                              */
/*********************************************************************/

/**
   High Order Finite Element Space for L2 with trig and quad (rel)order 
*/

/* ***********************************************
To do: *Internal External Dofs (eliminate internal) 
       *Flag for low-order dofs eliminated ...   
************************* */ 
#include <comp.hpp>
#include <multigrid.hpp>
#include "l2hofespace_ee.hpp" 

using namespace ngmg;

namespace ngcomp
{
  using namespace ngcomp;

  L2HighOrderFESpaceEE ::  
  L2HighOrderFESpaceEE (const MeshAccess & ama, const Flags & flags, bool parseflags)
    : L2HighOrderFESpace (ama, flags)
  {
    name="L2HighOrderFESpace_EE(l2ho_ee)";
    
    // defined flags
    DefineNumFlag("quad_relorder");
    DefineNumFlag("trig_relorder");
    DefineNumFlag("quad_order");
    DefineNumFlag("trig_order");
    
    if (parseflags) ParseFlags(flags);
    
    quad_relorder = int (flags.GetNumFlag("quad_relorder",rel_order)); 
    trig_relorder = int (flags.GetNumFlag("trig_relorder",rel_order)); 
    quad_order = int (flags.GetNumFlag("quad_order",order)); 
    trig_order = int (flags.GetNumFlag("trig_order",order)); 
    if(flags.NumFlagDefined("trig_relorder")||flags.NumFlagDefined("quad_relorder")) 
      var_order=1; 
  }
    
  L2HighOrderFESpaceEE :: ~L2HighOrderFESpaceEE ()
  {
    ;
  }


 
  void L2HighOrderFESpaceEE :: Update()
  {
    if(ma.GetDimension() == 3)
      {
	L2HighOrderFESpace :: Update(); 
	return;
      }

    if(low_order_space) low_order_space -> Update();
    nel = ma.GetNE();
    
    order_inner.SetSize(nel); 
    order_inner = INT<3> (0,0,0); 
    
    int dim = ma.GetDimension(); 
    
    for(int i=0;i<nel;i++) 
      {
	int r_ord;
	int ord; 
	INT<3> el_orders = ma.GetElOrders(i); 
	switch (ma.GetElType(i))
	  {
	  case ET_TRIG:
	    r_ord = trig_relorder; 
	    ord = trig_order;
	    break; 
	  case ET_QUAD: 
	    r_ord = quad_relorder; 
	    ord = quad_order;
	    break; 
	  default: 
	    r_ord = rel_order; 
	    ord = order; 
	    break;
	  }
	
	if(var_order)
	  {
	    INT<3> el_orders = ma.GetElOrders(i); 
	  	    
	    for(int j=0;j<dim;j++)
	      order_inner[i][j] = int(max2(el_orders[j]+r_ord,0));
	  }
	else 
	  order_inner[i] =  INT<3>(ord,ord,ord); 
      } 

    if(print) 
      *testout << " order_inner (l2ho_ee) " << order_inner << endl; 
    ndof = nel;
    
    UpdateDofTables();

    while (ma.GetNLevels() > ndlevel.Size())
      ndlevel.Append (ndof);
    ndlevel.Last() = ndof;

    if(low_order_space) prol->Update();

  } 
   
  // register FESpaces
  namespace
#ifdef MACOS
  l2hofespace_ee_cpp
#endif
  {
  
    class Init
    { 
    public: 
      Init ();
    };
    
    Init::Init()
    {
      GetFESpaceClasses().AddFESpace ("l2ho_ee", L2HighOrderFESpaceEE::Create);
    }
    
    Init init;
  }
}
