#ifndef FILE_L2HOFESPACESZ
#define FILE_L2HOFESPACESZ

/*********************************************************************/
/* File:   l2hofespace.hpp                                           */
/* Author: Start                                                     */
/* Date:   23.Feb. 2003 (Update(SZ): 2. Jan. 2007)                                             */
/*********************************************************************/
  
/**
   High Order Finite Element Space for L2 (element by element)
*/

class L2HighOrderFESpaceSZ : public FESpace
{
private:
  
  // Level
  int level;

  // Number of Elements
  int nel;

  ARRAY<int> first_element_dof;
  int ndof;
  ARRAY<int> ndlevel;

  // if order is relative to mesh order 
  bool var_order;
  // variable order is set to mesh_order + rel_order 
  int rel_order;
  // order of elements 
  ARRAY<INT<3> > order_inner;

  

public:

  L2HighOrderFESpaceSZ (const MeshAccess & ama, const Flags & flags);
  ///
  virtual ~L2HighOrderFESpaceSZ ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const
  {
    return "L2HighOrderFESpaceSZ";
  }

  ///
  virtual void Update();
  /// 
  // virtual void UpdateDofTables();
  ///
  virtual int GetNDof () const;
  ///
  virtual int GetNDofLevel (int level) const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  ///
  virtual Table<int> * CreateSmoothingBlocks (const Flags & precflags) const;
  /// 
 
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;

  int GetFirstInnerDof(int elnr) const  
  {return (first_element_dof[elnr]);}

  virtual int LowOrderDof() const { return(1);} 

  
};

class L2SurfaceHighOrderFESpaceSZ : public FESpace
{
private:
  
  // Level
  int level;

  // Number of Elements
  int nel;

  ARRAY<int> first_element_dof;
  int ndof;


  // if order is relative to mesh order 
  bool var_order;
  // variable order is set to mesh_order + rel_order 
  int rel_order;
  // order of elements 
  ARRAY<INT<3> > order_cell;

public:

  L2SurfaceHighOrderFESpaceSZ (const MeshAccess & ama, const Flags & flags);
  ///
  virtual ~L2SurfaceHighOrderFESpaceSZ ();

  static FESpace * Create (const MeshAccess & ma, const Flags & flags);

  virtual string GetClassName () const
  {
    return "L2SurfaceHighOrderFESpace";
  }

  ///
  virtual void Update();
  /// 
  //virtual void UpdateDofTables();
  ///
  virtual int GetNDof () const;
  ///
  virtual const FiniteElement & GetFE (int elnr, LocalHeap & lh) const;
  ///
  virtual const FiniteElement & GetSFE (int elnr, LocalHeap & lh) const;
  ///
  virtual void GetDofNrs (int elnr, ARRAY<int> & dnums) const;
  ///
  virtual void GetSDofNrs (int selnr, ARRAY<int> & dnums) const;
  
  virtual Table<int> * CreateSmoothingBlocks ( int type = 0) const;

  /// 
  virtual void GetInnerDofNrs (int elnr, ARRAY<int> & dnums) const;
  virtual void GetVertexDofNrs (int vnr, ARRAY<int> & dnums) const;
  virtual void GetEdgeDofNrs (int ednr, ARRAY<int> & dnums) const;
  virtual void GetFaceDofNrs (int fanr, ARRAY<int> & dnums) const;

  
};

#endif

