/*
  test.cpp: tests fuer den FacetFESpace 
*/


#include <solve.hpp>
#include <fem.hpp>
#include <comp.hpp>

#include "facetfespace.hpp"
#include "facetfe.hpp"

using namespace ngfem;
using namespace ngsolve;



/* *********************************************************** */
/* ************************* FESpace ************************* */
/* *********************************************************** */


class NumProcFacetTest : public NumProc
{
protected:

public:

  NumProcFacetTest (PDE & apde, const Flags & flags)
    : NumProc (apde)
  {}

  virtual ~NumProcFacetTest()
  { ; }

  static NumProc * Create (PDE & pde, const Flags & flags)
  {
    return new NumProcFacetTest (pde, flags);
  }




  virtual void Do(LocalHeap & lh)
  {
    cout << "Facet Test suite" << endl;
    cout << "================" << endl << endl;

    const FESpace * fes = pde.GetFESpace("v");
    const MeshAccess & ma = fes->GetMeshAccess();
    
    int ne = ma.GetNE();
    int ndof = fes->GetNDof();
    int nf;
    if (ma.GetDimension()==2)
      nf = ma.GetNEdges();
    else // dim=3
      nf = ma.GetNFaces();
    
    
    cout << "*** General info:" << endl;
    cout << "    ne = " << ne << endl;
    cout << "    nfacest = " << nf << endl;
    cout << "    ndof = " << ndof << endl;
    cout << endl;
    
    cout << "*** proceed over elements:" << endl;
    void * heapp = lh.GetPointer();
    for (int i=0; i<ne; i++)
    {
      const FacetVolumeFiniteElement & fel = 
          dynamic_cast<const FacetVolumeFiniteElement& >(fes->GetFE(i, lh));
      cout << "  - element " << i << endl;
      
      ARRAY<int> nums;
      fel.GetVertexNumbers(nums);
      cout << "    vertex numbers: "; 
      for (int j=0; j<nums.Size(); j++) 
        cout << nums[j] << " ";
      cout << endl;
       
      cout << "    facet orders: "; 
      fel.GetFacetOrders(nums);
      for (int j=0; j<nums.Size(); j++) 
      {
        cout << nums[i] <<  " ";
      }
      cout << endl;
      
      fes->GetDofNrs(i, nums);
      cout << "    global dof numbers: ";
      for (int j=0; j<nums.Size(); j++) 
        cout << nums[j] << " ";
      cout << endl;

      
      fes->GetExternalDofNrs(i, nums);
      cout << "    external dof numbers: ";
      for (int j=0; j<nums.Size(); j++) 
        cout << nums[j] << " ";
      cout << endl;

      
      int nf = fel.GetNF();
      cout << "    number of facets: " << nf << endl;
          
      void * heapp2 = lh.GetPointer();
      
      for (int j=0; j<nf; j++)
      {
        const FacetFacetFiniteElement & ffel = dynamic_cast<const FacetFacetFiniteElement& >(fel.GetFacetFE(j, lh));
        cout << "    + facet " << j << endl;  
        
        int order = ffel.GetOrder();
        cout << "      order = " << order << endl;
        fel.GetFacetDofNrs(j, nums);
        cout << "      semi-global dof numbers: ";
        for (int j=0; j<nums.Size(); j++) 
          cout << nums[j] << " ";
        cout << endl;
        
        cout << "      vertex numbers: "; 
        ffel.GetVertexNumbers(nums);
        for (int j=0; j<nums.Size(); j++) 
          cout << nums[j] << " ";
        cout << endl;
        
        cout << "      shapes in integration points:" << endl;
        
        const IntegrationRule *irp;
        if ( ma.GetDimension() == 2)
        {
          irp = &GetIntegrationRules().SelectIntegrationRule (ET_SEGM, 2*order);
        }
        else // dim == 3
        {
          if (ffel.ElementType() == ET_TRIG)
            irp = &GetIntegrationRules().SelectIntegrationRule (ET_TRIG, 2*order);
          else // quad
            irp = &GetIntegrationRules().SelectIntegrationRule (ET_QUAD, 2*order);
        }
        
        const IntegrationRule & ir=*irp;;

        void * heapp3=lh.GetPointer(); 
        for (int k=0; k<ir.GetNIP(); k++)
        {
          cout << "     ip(" << k << "): ";
          FlatVector<double> shape;
          if (ma.GetDimension() == 2)
            shape.AssignMemory(ffel.GetOrder()+1, lh);
          else // dim=3
          {
            if ( ffel.ElementType() == ET_TRIG )
              shape.AssignMemory( ((ffel.GetOrder()+1) * (ffel.GetOrder()+2))/2, lh);
            else //quad
              shape.AssignMemory( (ffel.GetOrder()+1) * (ffel.GetOrder()+1), lh);
          }   
              
          const IntegrationPoint & ip = ir[k];
          ffel.CalcShape(ip, shape);
          for (int l=0; l<shape.Size(); l++)
            cout << shape[l] << " ";
          cout << endl;
          
          lh.CleanUp(heapp3);
        }
      }
      lh.CleanUp(heapp2);
    }
    lh.CleanUp(heapp);

  }


  ///
  virtual string GetClassName () const
  {
    return "Facet Space Tester";
  }

  virtual void PrintReport (ostream & ost)
  { ; }
};







namespace  
{
  class Init
  { 
  public: 
    Init ();
  };
  
  Init::Init()
  {
    GetNumProcs().AddNumProc ("facettest", NumProcFacetTest::Create);
  }
  Init init;
}
