#ifndef DUFFY_TRAFO_HPP
#define DUFFY_TRAFO_HPP

#include <ngstd.hpp> 
#include <fem.hpp> 

namespace duffytrafo 
{
  //using namespace netgen; 
  using namespace ngstd;
  using namespace ngfem; 

  /*class DuffyTransformation 
{
private: 
  // int  sort[8]; 
  // int  vnums[8];

public:
  DuffyTransformation(const int * avnums); 
  
  }*/

class DuffyTransformationTet 
{
private:
  // 
  int sort[4]; 
  // 
  int vnums[4];
  
  // stores  DValues of lamir
  AutoDiff<3> lamir[4];
  // 
  Mat<3> dxi_dx_scaled; 
  Mat<3> trans_rot; 
  Mat<3> inv_jac; 
  
public:
  ///
  DuffyTransformationTet(const int * avnums)  
  { 
    // rotate coord. sys of tet acc. to global vertex-numbers 
    for(int l=0;l<4; l++) 
      sort[l] = l; 
    for(int l=0;l<4; l++) 
      vnums[l] = avnums[l]; 
        
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    
        
    lamir[0] = AutoDiff<3> (47, 0);
    lamir[1] = AutoDiff<3> (47, 1);
    lamir[2] = AutoDiff<3> (47, 2);
    lamir[3] = 1 - lamir[0] - lamir[1] - lamir[2];

    for(int k=0;k<3;k++)
      for(int l=0;l<3;l++) trans_rot(k,l) = lamir[sort[k]].DValue(l);  
        
    inv_jac(0,0) = 1;
    inv_jac(0,1) = inv_jac(0,2) = inv_jac(1,2) = 0;
  } 

  inline Mat<3> GetJacobianInverse(double * ip_hex) 
  {
    double invx = 1./(1-ip_hex[0]); 
    double invxy = invx/(1-ip_hex[1]); 
      
    inv_jac(0,0) = 1;
    inv_jac(0,1) = 0;
    inv_jac(0,2) = 0;
    
    inv_jac(1,0) = ip_hex[1]*invx;
    inv_jac(1,1) = 1*invx;
    inv_jac(1,2) = 0;
    
    inv_jac(2,0) = ip_hex[2]*invxy;
    inv_jac(2,1) = ip_hex[2]*invxy;
    inv_jac(2,2) = 1*invxy;

    return(inv_jac*trans_rot); 
  }


  inline Mat<3> GetJacobianInverse(double * ip_hex, bool scaled) 
  {
    double invx = 1.; 
    if(!scaled) invx/=(1-ip_hex[0]); 
    double invxy = invx; 
    if(!scaled) invxy/=(1-ip_hex[1]); 
      
    inv_jac(0,0) = 1;
    inv_jac(0,1) = 0;
    inv_jac(0,2) = 0;
    
    inv_jac(1,0) = ip_hex[1]*invx;
    inv_jac(1,1) = 1*invx;
    inv_jac(1,2) = 0;
    
    inv_jac(2,0) = ip_hex[2]*invxy;
    inv_jac(2,1) = ip_hex[2]*invxy;
    inv_jac(2,2) = 1*invxy;
 
    return(inv_jac*trans_rot); 
  }

  
  Mat<3> GetTransRot() 
  {
    return(trans_rot); 
  }
  Mat<3> GetInvJac() 
  {
    return(inv_jac); 
  }
  
  inline void SetEtaForJacobianInverse(double * ip_hex, bool scaled) 
  {
    double invx = 1.; 
    if(!scaled) 
      invx /= (1-ip_hex[0]);
        
    inv_jac(1,0) = ip_hex[1]*invx;
    inv_jac(1,1) = 1*invx;
    inv_jac(1,2) = 0;
    
    
  } 

  inline Mat<3> GetJacobianInverseZ(double * ip_hex, bool scaled) 
  {
    double invxy = 1.; 
    if(!scaled) 
      invxy /= ((1-ip_hex[0])* (1-ip_hex[1])); 
    
    inv_jac(2,0) = ip_hex[2]*invxy;
    inv_jac(2,1) = ip_hex[2]*invxy;
    inv_jac(2,2) = invxy;
    
    return(inv_jac*trans_rot); 
  }

  double GetJacobiDet(double * ip_hex)
  { 
    return(sqr(1-ip_hex[0])*(1-ip_hex[1])); 
  }

  void GetIPTet(double * ip_hex, double * x_tet)
  {
    double lami[3];   // { x, y, z, 1-x-y-z };
    lami[sort[0]] = ip_hex[0];
    lami[sort[1]] = ip_hex[1]* (1-ip_hex[0]);
    lami[sort[2]] = ip_hex[2]* (1-ip_hex[1]) * (1-ip_hex[0]);
        
    for(int l=0;l<3;l++)
      x_tet[l] = lami[l];  
  }

  inline AutoDiff<3> xi_x(double xi)
  {
    AutoDiff<3> xi_x = lamir[sort[0]]; 
    xi_x.Value() = xi;
    return(xi_x); 
  }

  inline AutoDiff<3> eta_x(AutoDiff<3> xi_x, double eta)
  {
    lamir[sort[1]].Value()=(1-xi_x.Value())*eta;
    AutoDiff<3> eta_x = lamir[sort[1]]/(1-xi_x); 
    eta_x.Value() = eta; 
    return(eta_x);  
  }

  inline AutoDiff<3> zeta_x(AutoDiff<3> xi_x, AutoDiff<3> eta_x, double zeta)
  {
    lamir[sort[2]].Value()=(1-xi_x.Value())*(1-eta_x.Value())*zeta;
    AutoDiff<3> zeta_x = lamir[sort[2]]/((1-xi_x)*(1-eta_x)); 
    zeta_x.Value() = zeta; 
    return(zeta_x);  
  }

  Mat<3> GetScaleOfJacobianInverse(double * xi)
  {
    Mat<3> s(0); 
   
    s(0,0) = 1.; 
    s(1,1) = 1./(1-xi[0]); 
    s(2,2) = s(1,1)/(1-xi[1]); 
    return(s); 
  }

};

class DuffyTransformationTrig 
{
private:
  // 
  int sort[3]; 
 
  // 
  int vnums[3];
  
  // stores  DValues of lamir
  AutoDiff<2> lamir[3];
  // 
  Mat<2> trans_rot; 
  Mat<2> inv_jac; 
  
public:
  ///
  DuffyTransformationTrig(const int * avnums)  
  { 
    // rotate coord. sys of tet acc. to global vertex-numbers 
    for(int l=0;l<2; l++) 
      sort[l] = l; 
    for(int l=0;l<2; l++) 
      vnums[l] = avnums[l]; 
        
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    
    lamir[0] = AutoDiff<2> (47, 0);
    lamir[1] = AutoDiff<2> (47, 1);
    lamir[2] = 1 - lamir[0] - lamir[1];

    for(int k=0;k<2;k++)
      for(int l=0;l<2;l++) 
	trans_rot(k,l) = lamir[sort[k]].DValue(l);  
       
    inv_jac(0,0) = 1;
    inv_jac(0,1) = 0;
  } 
  
  inline Mat<2> GetJacobianInverse(double * ip_hex, bool scaled) 
  {
    double invx = 1.; 
    if(!scaled) invx/=(1-ip_hex[0]); 
          
    inv_jac(0,0) = 1;
    inv_jac(0,1) = 0;
       
    inv_jac(1,0) = ip_hex[1]*invx;
    inv_jac(1,1) = 1*invx;
     
    return(inv_jac*trans_rot); 
  }

  
  Mat<2> GetTransRot() 
  {
    return(trans_rot); 
  }

  Mat<2> GetInvJac() 
  {
    return(inv_jac); 
  }
  
  double GetJacobiDet(double * ip_hex)
  { 
    return(1-ip_hex[0]); 
  }

  void GetIPTrig(double * ip_hex, double * x_tet)
  {
    double lami[2];
    lami[sort[0]] = ip_hex[0];
    lami[sort[1]] = ip_hex[1]* (1-ip_hex[0]);
        
    for(int l=0;l<2;l++)
      x_tet[l] = lami[l];  
  }

  inline AutoDiff<2> xi_x(double xi)
  {
    AutoDiff<2> xi_x = lamir[sort[0]]; 
    xi_x.Value() = xi;
    return(xi_x); 
  }

  inline AutoDiff<2> eta_x(AutoDiff<2> xi_x, double eta)
  {
    lamir[sort[1]].Value()=(1-xi_x.Value())*eta;
    AutoDiff<2> eta_x = lamir[sort[1]]/(1-xi_x); 
    eta_x.Value() = eta; 
    return(eta_x);  
  }

  Mat<2> GetScaleOfJacobianInverse(double * xi)
  {
    Mat<2> s(0); 
   
    s(0,0) = 1.; 
    s(1,1) = 1./(1-xi[0]); 
    return(s); 
  }

};

}



#endif
 
