/*********************************************************************/
/* File:   h1hofe.cpp                                                */
/* Author: Start                                                      */
/* Date:   6. Feb. 2003                                              */
/*********************************************************************/

// #define NOPROFILE
 
#include <fem.hpp>
#include "duffytrafo.hpp"

namespace ngfem
{
  using namespace ngfem;
  using namespace duffytrafo;

  // ************************** TRIG ************************************ 
  
  void H1HighOrderTrigTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                      FlatVector<> shape) const
  {
    T_CalcShape<double, double, FlatVector<> > (ip(0), ip(1), shape); 
  }
  
  void H1HighOrderTrigTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                       FlatMatrix<> dshape) const
  {
    AutoDiff<2> x(ip(0), 0);
    AutoDiff<2> y(ip(1), 1);

    /* 
    ArrayMem<AutoDiff<2>,40> sds(ndof);
    T_CalcShape<AutoDiff<2>, AutoDiff<2>, FlatArray<AutoDiff<2> > >
      (x,y,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 2; j++)
	dshape(i, j) = sds[i].DValue(j);
    
    return; 
    */ 


    AutoDiff<2> lami[3] = { x, y,  1-x-y };

    int sort[] = { 0, 1, 2 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] 
    
    AutoDiff<2> lamis[3];
    for (int i = 0; i < 3; i++)
      lamis[i] = lami[sort[i]];

    ArrayMem<AutoDiff<2>, 20> polx(order+1), poly(order+1), polyy(2); 

    ArrayMem<AutoDiff<1>, 20> polxi(order+1);   
 
    AutoDiff<1> * hp = &polxi[1];
    
    AutoDiff<1> xi(lamis[0].Value(),0); 
    AutoDiff<2> xi_x = lamis[0]; 
    LegendrePolynomialMult (order-2, 2*xi-1, xi, hp);
    polxi[0] = 1.0;
            
    for(int j=0;j<=order+1;j++)
      {
	for(int l=0;l<2;l++)
	  polx[j].DValue(l) = polxi[j].DValue(0)*xi_x.DValue(l);
	polx[j].Value() = polxi[j].Value(); 
      }
    // mit AutoDiff<1> und Dshape von Lamis[0]

    // l_j^S(lamis[1]-lamis[2],lamis[1]+lamis[2]) = l_j^S(eta,t)
    AutoDiff<2> eta_x = lamis[1] - lamis[2]; 
    AutoDiff<1> eta(eta_x.Value(),0); 
    AutoDiff<2> t_x = lamis[1] + lamis[2];
    double tgradt[] = {t_x.Value()*t_x.DValue(0),t_x.Value()*t_x.DValue(1)}; 
    
    ArrayMem<AutoDiff<1>, 15>  legs_eta(order+1);
    AutoDiff<1> * hp_eta = &legs_eta[1]; 
    ScaledLegendrePolynomial(order-2, eta, t_x.Value(), hp_eta); 
    
    legs_eta[0] = 1.0; 
    poly[0] = 1.0; 
    // poly[1] = lamis[1] * legs_eta[1].Value(); 
    for(int j=1;j<order+1;j++) 
      { 
	poly[j].Value() = lamis[1].Value() * legs_eta[j].Value();
	
	for(int l=0;l<2;l++)
	  poly[j].DValue(l) = lamis[1].DValue(l) * legs_eta[j].Value() 
	    + lamis[1].Value() * ( legs_eta[j].DValue(0) * eta_x.DValue(l) 
				   - legs_eta[j-1].DValue(0) * tgradt[l]);
      } 
    
    polyy[0] = 1.0;
    polyy[1] = lamis[2];

    ArrayMem<int[3], 100> trig2tensor(ndof);
    GetTrig2TensorMapping (trig2tensor);
        
    AutoDiff<2> adv; 
    for (int i = 0; i < ndof; i++)
      {
	adv =
	  polx[trig2tensor[i][0]] * 
	  poly[trig2tensor[i][1]] * 
	  polyy[trig2tensor[i][2]];
	for(int l=0;l<2;l++)
	  dshape(i,l) = adv.DValue(l);
      } 
     

    /* 
    for (int i = 0; i < ndof; i++)
      for(int l =0;l<3;l++) 
	dshape(i,l) = polx[trig2tensor[i][0]].Value()* (poly[trig2tensor[i][1]].Value() * polyy[trig2tensor[i][2]].DValue(l)+ 
							poly[trig2tensor[i][1]].DValue(l)*polyy[trig2tensor[i][2]].Value())
	  + polx[trig2tensor[i][0]].DValue(l) * poly[trig2tensor[i][1]].Value()*polyy[trig2tensor[i][2]].Value();
    */
    // Das sollte eventuell auch ueber Multiplikation mit Jacobian gehen? 
    // geht nicht da vroni trick 
  }


  template<typename Tx, typename Ty, typename TFA>  
  void  H1HighOrderTrigTPSZ :: T_CalcShape (Tx x, Ty y, TFA & sds) const
  {
    Tx lami[3] = { x, y,  1-x-y };

    int sort[] = { 0, 1, 2 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] 
    
    Tx lamis[3];
    for (int i = 0; i < 3; i++)
      lamis[i] = lami[sort[i]];

    ArrayMem<Tx, 20> polx(order+1), poly(order+1), polyy(2); 

    Tx * hp = &polx[1];
    LegendrePolynomialMult (order-2, 2*lamis[0]-1, lamis[0], hp);
    polx[0] = 1.0;
    // mit AutoDiff<1> und Dshape von Lamis[0]
    
    hp = &poly[1];
    ScaledLegendrePolynomialMult (order-2, lamis[1]-lamis[2], lamis[1] + lamis[2], lamis[1], hp);
    poly[0] = 1.0;

    polyy[0] = 1.0;
    polyy[1] = lamis[2];

    ArrayMem<int[3], 100> trig2tensor(ndof);
    GetTrig2TensorMapping (trig2tensor);

    for (int i = 0; i < ndof; i++)
      sds[i] =
        polx[trig2tensor[i][0]] * 
        poly[trig2tensor[i][1]] * 
        polyy[trig2tensor[i][2]];
  }


  void H1HighOrderTrigTPSZ :: GetTrig2TensorMapping(FlatArray<int[3]> trig2tensor) const 
  {
    // maps trig_shapes (0<=ii<=ndof) to the accoring tensor-product shapes (i,j,l) with i,j<=order, l<=2
    // (injective map) 

    int sort[] = { 0, 1, 2 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] 

    int isort[3];
    for (int i = 0; i < 3; i++) isort[sort[i]] = i;
    
    for (int i = 0; i < ndof; i++)
      {
	trig2tensor[i][0] = trig2tensor[i][1] =
	  trig2tensor[i][2] = 0;
      }
    
    // vertex shapes
    for (int i = 0; i < 3; i++)
      trig2tensor[i][isort[i]] = 1;
    
    int ii = 3;
    
    //  edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    for (int i = 0; i < 3; i++)
      if (order_edge[i] >= 2)
	{
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  int nde = order_edge[i]-1;
	  for (int j = 1; j <= nde; j++, ii++)
	    {
	      trig2tensor[ii][isort[es]] = j; // lami_s * ''pol(lami_s-lami_e,lami_s+lami_e)'' 
	      trig2tensor[ii][isort[ee]] = 1; // lami_e
	    }
	}
    
    // face dofs
    if (order_inner[0] >= 3)
      {
	int fav[3] = { 0, 1, 2 };
	
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
          
	int pi = order_inner[0];
	for (int j2 = 0; j2 <= pi-3; j2++)
	  for (int i2 = 0; i2 <= pi-3-j2; i2++, ii++)
	    {
	      trig2tensor[ii][isort[fav[0]]] = j2+1;
	      trig2tensor[ii][isort[fav[1]]] = i2+1;
	      trig2tensor[ii][isort[fav[2]]] = 1;
	    }
      }

  } 

  int H1HighOrderTrigTPSZ :: 
  GetTensor2SegmMapping(FlatArray<int[3] > trig2tensor, FlatArray<int> tensor2segm) const
  {
    // size (nd1d = 2 * order) 
    // mapping of 1d-tp points to either -1 (no segm point in trig) or to segm_dof_nr    
    
    tensor2segm = -1;
    
    for (int ii = 0; ii < ndof; ii++)
      {
	int j1d = trig2tensor[ii][1] * 2 + trig2tensor[ii][2];
	tensor2segm[j1d] = 0;
      }
    
    int ndsegm =0;
    int nd1d = 2*order;
    for (int i = 0; i < nd1d; i++)
      if (tensor2segm[i] != -1)
	{
	  tensor2segm[i] = ndsegm;
	  ndsegm++;
	}
    return(ndsegm); 
  } 


  // ******************************** TET **************************************
  /* 
  void H1HighOrderTetTP::
  SetVertexNumbers (FlatArray<int> & avnums)
  {
    for (int i = 0; i < avnums.Size(); i++)
      vnums[i] = avnums[i];
    
    for(int l=0;l<4;l++) sort[l]=l; 
 
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);

    }*/ 

  void H1HighOrderTetTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                      FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderTetTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                       FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    /* ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
    */
    
    AutoDiff<3> lami[4] = { x, y, z, 1-x-y-z };

    int sort[] = { 0, 1, 2, 3 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);

    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] < vnums[sort[3]]
    // eventuell SORT auch gleich mitabspeichern??? 
    
    AutoDiff<3> lamis[4];
    for (int i = 0; i < 4; i++)
      lamis[i] = lami[sort[i]];

    ArrayMem<AutoDiff<3>, 20> polx(order+1), poly(order+1), polz(order+1), polzz(2);
    ArrayMem<AutoDiff<1>, 20> polxi(order+1);
    
    AutoDiff<3> xi_x = lamis[0]; 
    AutoDiff<1> xi(xi_x.Value(),0); 
    AutoDiff<1> * hp_xi = &polxi[1];
    LegendrePolynomialMult (order-2, 2*xi-1, xi, hp_xi);
    
    polx[0] = 1.0;
    for(int i=1;i<order+1;i++)
      { 
	polx[i].Value() = polxi[i].Value(); 
	for(int l=0;l<3;l++) 
	  polx[i].DValue(l) = polxi[i].DValue(0)*xi_x.DValue(l);
      } 
    
    // l_j^S(lamis[1]-lamis[2],lamis[1]+lamis[2]) = l_j^S(eta,t)
    AutoDiff<3> eta_x = lamis[1] - lamis[2] - lamis[3]; 
    AutoDiff<1> eta(eta_x.Value(),0); 
    AutoDiff<3> t_x = lamis[1] + lamis[2] + lamis[3];
    double tgradt[] = 
      {t_x.Value()*t_x.DValue(0),t_x.Value()*t_x.DValue(1), 
       t_x.Value()*t_x.DValue(2)}; 
    
    ArrayMem<AutoDiff<1>, 15>  legs_eta(order+1);
    AutoDiff<1> * hp_eta = &legs_eta[1]; 
    ScaledLegendrePolynomial(order-2, eta, t_x.Value(), hp_eta); 
    
    legs_eta[0] = 1.0; 
    poly[0] = 1.0; 
    // poly[1] = lamis[1] * legs_eta[1].Value(); 
    for(int j=1;j<order+1;j++) 
      { 
	poly[j].Value() = lamis[1].Value() * legs_eta[j].Value();
	
	for(int l=0;l<3;l++)
	  poly[j].DValue(l) = lamis[1].DValue(l) * legs_eta[j].Value() 
	    + lamis[1].Value() * ( legs_eta[j].DValue(0) * eta_x.DValue(l) 
				   - legs_eta[j-1].DValue(0) * tgradt[l]);
      } 

    AutoDiff<3> zeta_x = lamis[2] - lamis[3]; 
    AutoDiff<1> zeta(zeta_x.Value(),0); 
    t_x = lamis[2] + lamis[3];
    for(int l=0;l<3;l++) tgradt[l] =  t_x.Value()*t_x.DValue(l);  
    
    ArrayMem<AutoDiff<1>, 15>  legs_zeta(order+1);
    AutoDiff<1> * hp_zeta = &legs_zeta[1]; 
    ScaledLegendrePolynomial(order-2, zeta, t_x.Value(), hp_zeta); 
    
    legs_zeta[0] = 1.0; 
    polz[0] = 1.0; 
    // poly[1] = lamis[1] * legs_eta[1].Value(); 
    for(int j=1;j<order+1;j++) 
      { 
	polz[j].Value() = lamis[2].Value() * legs_zeta[j].Value();
	
	for(int l=0;l<3;l++)
	  polz[j].DValue(l) = lamis[2].DValue(l) * legs_zeta[j].Value() 
	    + lamis[2].Value() * ( legs_zeta[j].DValue(0) * zeta_x.DValue(l) 
				   - legs_zeta[j-1].DValue(0) * tgradt[l]);
      } 

    polzz[0] = 1.0;
    polzz[1] = lamis[3];

    ArrayMem<int[4], 100> tet2tensor(ndof);
    GetTet2TensorMapping (tet2tensor);

    AutoDiff<3> sds; 
    for (int i = 0; i < ndof; i++)
      { 
	sds =
	  polx[tet2tensor[i][0]] * 
	  poly[tet2tensor[i][1]] * 
	  polz[tet2tensor[i][2]] *
	  polzz[tet2tensor[i][3]]; 
	
	for(int l=0;l<3;l++) 
	  dshape(i,l) = sds.DValue(l); 
      }
  }



  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderTetTPSZ :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx lami[4] = { x, y, z, 1-x-y-z };

    int sort[] = { 0, 1, 2, 3 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);

    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] < vnums[sort[3]]
    
    Tx lamis[4];
    for (int i = 0; i < 4; i++)
      lamis[i] = lami[sort[i]];

    ArrayMem<Tx, 20> polx(order+1), poly(order+1), polz(order+1), polzz(2);

    Tx * hp = &polx[1];
    LegendrePolynomialMult (order-2, 2*lamis[0]-1, lamis[0], hp);
    polx[0] = 1.0;
    // mit AutoDiff<1> und Dshape von Lamis[0]
    

    hp = &poly[1];
    ScaledLegendrePolynomialMult (order-2, lamis[1]-lamis[2]-lamis[3], 1-lamis[0], lamis[1], hp);
    poly[0] = 1.0;

    hp = &polz[1];
    ScaledLegendrePolynomialMult (order-2, lamis[2]-lamis[3], lamis[2]+lamis[3], lamis[2], hp);
    polz[0] = 1.0;
    // mit AutoDiff<2> Dshape von Lamis[2], Lamis[3]


    polzz[0] = 1.0;
    polzz[1] = lamis[3];

    ArrayMem<int[4], 100> tet2tensor(ndof);
    GetTet2TensorMapping (tet2tensor);

    for (int i = 0; i < ndof; i++)
      sds[i] =
        polx[tet2tensor[i][0]] * 
        poly[tet2tensor[i][1]] * 
        polz[tet2tensor[i][2]] *
        polzz[tet2tensor[i][3]];
  }





  void H1HighOrderTetTPSZ :: GetTet2TensorMapping(FlatArray<int[4]> tet2tensor) const 
  {
    int sort[] = {0,1,2,3};
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);

    int isort[4];
    for (int i = 0; i < 4; i++) isort[sort[i]] = i;
    
    for (int i = 0; i < ndof; i++)
      {
	tet2tensor[i][0] = tet2tensor[i][1] =
	  tet2tensor[i][2] = tet2tensor[i][3] = 0;
      }
    
    // vertex shapes
    for (int i = 0; i < 4; i++)
      tet2tensor[i][isort[i]] = 1;
    
    int ii = 4;
    
    //  edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  int nde = order_edge[i]-1;
	  for (int j = 1; j <= nde; j++, ii++)
	    {
	      tet2tensor[ii][isort[es]] = j;
	      tet2tensor[ii][isort[ee]] = 1;
	    }
	}
    
    // face dofs
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
      if (order_face[i][0] >= 3)
	{
	  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
	  
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
          
	  int pf = order_face[i][0];
	  for (int j2 = 0; j2 <= pf-3; j2++)
	    for (int i2 = 0; i2 <= pf-3-j2; i2++, ii++)
	      {
		tet2tensor[ii][isort[fav[0]]] = j2+1;
		tet2tensor[ii][isort[fav[1]]] = i2+1;
		tet2tensor[ii][isort[fav[2]]] = 1;
	      }
	}

    // inner dofs
    int pi = order_inner[0]; 
    if (pi >= 4)
      {
	for (int i = 0; i <= pi-4; i++)
	  for (int j = 0; j <= pi-4-i; j++)
	    for (int k = 0; k <= pi-4-i-j; k++, ii++)
	      {
		tet2tensor[ii][0] = k+1;
		tet2tensor[ii][1] = j+1;
		tet2tensor[ii][2] = i+1;
		tet2tensor[ii][3] = 1;
	      }
      }

  } 

  int H1HighOrderTetTPSZ :: 
  GetTrig2TensorMappings(FlatArray<int[4] >   tet2tensor, ARRAY<int[3]> & trig2tensor, FlatArray<int> tensor2trig) const
  {   
    int nd1d = 2 * (order);
    int nd2d = order * nd1d;
    
    tensor2trig = -1;
    
    for (int ii = 0; ii < ndof; ii++)
      {
	int j2d = tet2tensor[ii][1] * nd1d + tet2tensor[ii][2] * 2 + tet2tensor[ii][3];
	tensor2trig[j2d] = 0;
      }
    
    int ndtrig =0;
    for (int i = 0; i < nd2d; i++)
      if (tensor2trig[i] != -1)
	{
	  tensor2trig[i] = ndtrig;
	  ndtrig++;
	}
    trig2tensor.SetSize(ndtrig); 
    
    for (int i2 = 0; i2 < order; i2++)
      for (int i3 = 0; i3 < order; i3++)
	for (int i4 = 0; i4 < 2; i4++)
	  {
	    int i2d = i2 * nd1d + i3 * 2 + i4;
	    if (tensor2trig[i2d] != -1)
	      {
		trig2tensor[tensor2trig[i2d]][0] = i2;
		trig2tensor[tensor2trig[i2d]][1] = i3;
		trig2tensor[tensor2trig[i2d]][2] = i4;
	      }
	  }
    return(ndtrig);
  }


  int H1HighOrderTetTPSZ :: 
  GetTensor2SegmMapping(FlatArray<int[4] > tet2tensor, FlatArray<int> tensor2segm) const
  {
    tensor2segm = -1;
    
    for (int ii = 0; ii < ndof; ii++)
      {
	int j1d = tet2tensor[ii][2] * 2 + tet2tensor[ii][3];
	tensor2segm[j1d] = 0;
      }
    
    int ndsegm =0;
    int nd1d = 2*order;
    for (int i = 0; i < nd1d; i++)
      if (tensor2segm[i] != -1)
	{
	  tensor2segm[i] = ndsegm;
	  ndsegm++;
	}
    return(ndsegm); 
  } 



  // ******************************** PRISM **************************************




  void H1HighOrderPrismTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                        FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderPrismTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                         FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  } 




  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderPrismTPSZ :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

    Tx lami[6] = { x, y, 1-x-y, x, y, 1-x-y };
    Tx muz[6]  = { 1-z, 1-z, 1-z, z, z, z };

    
    Tx lamisum[6] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
        {
          if (vnums[j] > vnums[i]) lamisum[i] += lami[j];
          if (vnums[j+3] > vnums[i+3]) lamisum[i+3] += lami[j+3];
        }

    sds = 0.0;

    // vertex shapes
    for (int i = 0; i < 6; i++)
      sds[i] = lami[i] * muz[i];

    int ii = 6;

    // horizontal edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_PRISM);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int p = order_edge[i];
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];
          ScaledLegendrePolynomialMult (p-2, lami[es]-lamisum[es], lami[es]+lamisum[es], lami[es]*lami[ee]*muz[ee], hp);
	  ii += p-1;
	}
    
    // vertical edges
    for (int i = 6; i < 9; i++)
      if (order_edge[i] >= 2)
	{
	  int p = order_edge[i];
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];
          LegendrePolynomialMult (p-2, muz[es]-muz[ee], muz[es]*muz[ee]*lami[ee], hp);
	  ii += p-1;
	}
    

    ArrayMem<Tx,20> polx(order+1), poly(order+1), polz(order+1);

    const FACE * faces = ElementTopology::GetFaces (ET_PRISM);
    // trig face dofs

    int hv[2];

    for (int i = 0; i < 2; i++)
      if (order_face[i][0] >= 3)
	{
	  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
	  
          hv[i] = fav[0];

          LegendrePolynomialMult (order-2, lami[fav[0]]-lamisum[fav[0]], lami[fav[0]], polx);
          ScaledLegendrePolynomialMult (order-2, lami[fav[1]]-lamisum[fav[1]], lamisum[fav[0]], lami[fav[1]], poly);

          Tx bub = lami[fav[2]] * muz[fav[2]];
          
          int n = order_face[i][0];
	  for (int j2 = 0; j2 <= n-3; j2++)
            {
              Tx hv = bub * poly[j2];
              for (int i2 = 0; i2 <= n-3-j2; i2++, ii++)
                sds[ii] = hv * polx[i2];
            }
	}
    /*
    if (hv[1] == hv[0]+3)  cout << "good ";
                else       cout << "bad ";
    */
    // quad face dofs
    for (int i = 2; i < 5; i++)
      if (order_face[i][0] >= 2 && order_face[i][1] >= 2)
	{
	  INT<2> p = order_face[i];
	  int pp = int(max2(p[0], p[1]))+1;
	 
	  int fmax = 0;
	  for(int j = 1; j < 4; j++) 
	    if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;
	  
	  int fz = 3-fmax; 
	  int ftrig = fmax^1; 
	  
	  Tx xi = lami[faces[i][fmax]] - lami[faces[i][ftrig]]; 
	  Tx eta = 1-lami[faces[i][fmax]]-lami[faces[i][ftrig]]; 
	  Tx zeta = muz[faces[i][fmax]]-muz[faces[i][fz]]; 

          /*
            T_ORTHOPOL::CalcTrigExt (pp, xi, eta, polx);  
            T_ORTHOPOL::Calc (pp, zeta, polz);	
	  */
          Tx l1 = lami[faces[i][fmax]];
          Tx l2 = lami[faces[i][ftrig]];
          ScaledLegendrePolynomialMult (pp-2, l2-l1, l1+l2, l1*l2, polx);
          LegendrePolynomialMult (pp-2, -zeta, 0.25 * (1-zeta*zeta), polz);

	  // global x-direction is towards second-largest vertex (JS)
	  if (vnums[faces[i][ftrig]] > vnums[faces[i][fz]])  
	    for (int k = 0; k < p[0]-1; k++)
	      for (int j = 0; j < p[1]-1; j++)
		sds[ii++] = polx[k] * polz[j];
	  else
	    for (int j = 0; j < p[0]-1; j++)
	      for (int k = 0; k < p[1]-1; k++)
		sds[ii++] = polx[k] * polz[j];
	}
    
    // volume dofs:
    if (order_inner[0] > 2 && order_inner[2] > 1)
      {
	ArrayMem<Tx,20> pol_trig((order_inner[0]-1)*(order_inner[0]-2)/2);
	int nf = T_TRIGFACESHAPES::Calc (order_inner[0], x-y, 1-x-y, pol_trig);

	T_ORTHOPOL:: Calc(order_inner[2], 2*z-1, polz);
	for (int i = 0; i < nf; i++)
	  for (int k = 0; k < order_inner[2]-1; k++)
	    sds[ii++] = pol_trig[i] * polz[k];
      }
  }








  // ******************************** HEX **************************************




  void H1HighOrderHexTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                      FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderHexTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                       FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  } 




  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderHexTPSZ :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx xi[3] = { 2*x-1, 2*y-1, 2*z-1 };
    
    ArrayMem<Tx, 80> polxi( 3 * (order+1) );
    Tx * poli[3];
    for (int i = 0; i < 3; i++) poli[i] = &polxi[i*(order+1)];

    for (int i = 0; i < 3; i++)
      {
        Tx * hp = poli[i]+2;
        poli[i][0] = 0.5 * (1-xi[i]);
        poli[i][1] = 0.5 * (1+xi[i]);
        LegendrePolynomialMult (order-2, xi[i], 0.25 * (1-xi[i]*xi[i]), hp);
      }

    ArrayMem<int[3], 100> hex2tensor(ndof);
    ArrayMem<double, 100> fac(ndof);

    GetHex2TensorMapping (hex2tensor, fac);
    
    for (int i = 0; i < ndof; i++)
      sds[i] = fac[i] * 
        poli[0][hex2tensor[i][0]] * 
        poli[1][hex2tensor[i][1]] * 
        poli[2][hex2tensor[i][2]];
  }





  void H1HighOrderHexTPSZ :: 
  GetHex2TensorMapping (FlatArray<int[3]> hex2tensor, FlatArray<double> factor) const
  {

    static int hex_points[][3] = 
      { 
	{ 0, 0, 0 },
	{ 1, 0, 0 },
	{ 1, 1, 0 },
	{ 0, 1, 0 },
	{ 0, 0, 1 },
	{ 1, 0, 1 },
	{ 1, 1, 1 },
	{ 0, 1, 1 }
      };


    // vertex shapes

    for (int i = 0; i < 8; i++)
      {
        hex2tensor[i][0] = hex_points[i][0];
        hex2tensor[i][1] = hex_points[i][1];
        hex2tensor[i][2] = hex_points[i][2];
        factor[i] = 1;
      }
        
    int ii = 8;

    // edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_HEX);
    for (int i = 0; i < 12; i++)
      if (order_edge[i] >= 2)
	{
	  int p = order_edge[i];
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);


          int ind[3];
          double fac = 1, mult = 1;
          for (int l = 0; l <= p-2; l++)
            {
              for (int k = 0; k < 3; k++)
                {
                  if (hex_points[es][k] == hex_points[ee][k])
                    ind[k] = hex_points[es][k]; // --> Vertex-basis
                  else
                    {
                      ind[k] = l+2;
                      if (hex_points[es][k] < hex_points[ee][k])
                        mult = -1; // check if compatible with orientation on ref elem
                    }
                }

              factor[ii] = fac; // only ungerade mit -1 
              hex2tensor[ii][0] = ind[0];
              hex2tensor[ii][1] = ind[1];
              hex2tensor[ii][2] = ind[2];
              ii++;

              fac *= mult;
            }
	}


     
    const FACE * faces = ElementTopology::GetFaces (ET_HEX);
    for (int i = 0; i < 6; i++)
      if (order_face[i][0] >= 2 && order_face[i][1] >= 2)
	{
	  int px = order_face[i][0];
	  int py = order_face[i][1];
	  int p = max2(px, py);

	  int fmax = 0;
	  for (int j=1; j<4; j++) 
	    if (vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;  
	  
	  int f1 = faces[i][(fmax+3)%4];
	  int f2 = faces[i][(fmax+1)%4]; 
	  fmax = faces[i][fmax]; 
	  
	  if(vnums[f2]>vnums[f1]) swap(f1,f2);  // fmax > f1 > f2 
	  

          int ind[3];
          double fac1 = 1, fac2 = 1, mult1 = 1, mult2 = 1;
          for (int l = 0; l <= px-2; l++)
            {
              fac2 = 1;
              for (int m = 0; m <= py-2; m++)
                {
                  for (int k = 0; k < 3; k++)
                    {
                      ind[k] = hex_points[fmax][k];
                      if (hex_points[fmax][k] != hex_points[f2][k])
                        {
                          ind[k] = m+2;
                          if (hex_points[fmax][k] > hex_points[f2][k])
                            mult2 = -1;
                        }
                      if (hex_points[fmax][k] != hex_points[f1][k])
                        {
                          ind[k] = l+2;
                          if (hex_points[fmax][k] > hex_points[f1][k])
                            mult1 = -1;
                        }
                    }

                  factor[ii] = fac1*fac2;
                  hex2tensor[ii][0] = ind[0];
                  hex2tensor[ii][1] = ind[1];
                  hex2tensor[ii][2] = ind[2];
                  ii++;

                  fac2 *= mult2;
                }
              fac1 *= mult1;
            }

	}
    
    // volume dofs:
    if (order_inner[0] >= 2 && order_inner[1] >= 2 && order_inner[2] >= 2)
      {
	for (int i = 2; i <= order_inner[0]; i++)
	  for (int j = 2; j <= order_inner[1]; j++)
            for (int k = 2; k <= order_inner[2]; k++)
              {
                factor[ii] = 1;
                hex2tensor[ii][0] = i;
                hex2tensor[ii][1] = j;
                hex2tensor[ii][2] = k;
                ii++;
              }
      }    
  }




#ifndef SABINE 


  // ******************************** Laplace **************************************



  
  template <int DIM>
  class LaplaceIntegratorTPSZ : public BilinearFormIntegrator
  {
    CoefficientFunction * coef;
    BilinearFormIntegrator * orig;

    typedef DiffOpGradient<DIM> DIFFOP;
    enum { DIM_SPACE   = DIFFOP::DIM_SPACE };
    enum { DIM_ELEMENT = DIFFOP::DIM_ELEMENT };
    enum { DIM_DMAT    = DIFFOP::DIM_DMAT };
    // enum { DIM         = DIFFOP::DIM };

    DiagDMat<DIM> dmatop;
    
    // typedef H1HighOrderTetTPSZ FEL;
  public:
    LaplaceIntegratorTPSZ (CoefficientFunction * acoef)
      : BilinearFormIntegrator (), coef(acoef), dmatop (acoef)
    { 
      orig = new LaplaceIntegrator<DIM> (acoef); 
      fast = 1;
    }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new LaplaceIntegratorTPSZ (coeffs[0]);
    }

    virtual bool BoundaryForm () const { return 0; }
    

    void 
    AssembleElementMatrixTrig (const FiniteElement & bfel,
			       const ElementTransformation & eltrans, 
			       FlatMatrix<double> & elmat,
			       LocalHeap & lh, bool & scaled) const
    {
      typedef H1HighOrderTrigTPSZ FEL; 

      const FEL & fel = static_cast<const FEL&> (bfel);

      static int timer = NgProfiler::CreateTimer ("Fast assembling");
      static int timer1 = NgProfiler::CreateTimer ("Fast assembling - geometry");
      static int timer2a = NgProfiler::CreateTimer ("Fast assembling - transform x-j");
      static int timer2b = NgProfiler::CreateTimer ("Fast assembling - transform x-k");
      static int timer3 = NgProfiler::CreateTimer ("Fast assembling - transform y");
      static int timer4 = NgProfiler::CreateTimer ("Fast assembling - transform z");
      static int timeri = NgProfiler::CreateTimer ("Fast assembling - init");
      static int timertrans = NgProfiler::CreateTimer ("Fast assembling - transform y,z");



      NgProfiler::RegionTimer reg (timer);
      NgProfiler::StartTimer (timeri);
      int ndof = fel.GetNDof();
      
      elmat.AssignMemory (ndof, ndof, lh);
      elmat = 0;
      
      int order = fel.Order();
      const IntegrationRule & ir1dy = SelectIntegrationRule (ET_SEGM, 2*order-1);
      
      static IntegrationRule ir1dx;
      if (ir1dx.GetNIP() == 0)
	{
	  int np = order;
	  ARRAY<double> xi(np), wi(np);
	  ComputeGaussJacobiRule (np, xi, wi, 2, 0);
	  for (int i = 0; i < np; i++)
	    ir1dx.AddIntegrationPoint (new IntegrationPoint (xi[i], 0, 0, wi[i]));
	}
	  
      ArrayMem<AutoDiff<1>, 20> h_powx(order+2); 
      AutoDiff<1> * powx = &h_powx[1]; 
      h_powx[0] = 0.0; 
      ArrayMem<AutoDiff<1>, 20> polx(order+1);
      ArrayMem<AutoDiff<1>, 20> poly(order+1), polyy(2), powy(order+1);
            
      DuffyTransformationTrig duffy_trafo(fel.GetVNums()); 
      
      void * heapp = lh.GetPointer();
      
      FlatArray<int[3]> trig2tensor (ndof, lh);
      fel.GetTrig2TensorMapping(trig2tensor);  
      
      int nd1d = 2 * (order);
      	  	  
      FlatArray<int> tensor2segm ( nd1d, lh);
      int ndsegm = fel.GetTensor2SegmMapping(trig2tensor, tensor2segm); 
      
      FlatArray<int> trig2segm (ndof, lh);
      // maps trig_dofs(0<=ii<ndof) to segm_dofs (j,l) (surjective)
      for (int ii = 0; ii < ndof; ii++)
	trig2segm[ii] = tensor2segm[trig2tensor[ii][1] * 2 + trig2tensor[ii][2]];
      
      FlatMatrix<Mat<2> > sum1d(ndsegm, lh);
      FlatVector<Mat<2> > mult1d(ndsegm, lh);
      FlatMatrix<Vec<2> > mult2d(ndsegm, ndof, lh);
      
      FlatMatrix<> shapesy (ir1dy.GetNIP(), ndsegm, lh);
      FlatMatrix<> shapesdy (ir1dy.GetNIP(), ndsegm, lh);
      
      for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
	{
	  AutoDiff<1> yr1(ir1dy[i2](0), 0);
	  AutoDiff<1> * hp1 = &poly[1];
	  LegendrePolynomialMult (order-2, 2*yr1-1, yr1, hp1);
	  poly[0] = 1;
	  poly[order] = 0;
	  
	  polyy[0] = 1;
	  polyy[1] = 1-yr1; 
	  
	  for (int j2 = 0; j2 < order; j2++)
	    for (int j3 = 0; j3 < 2; j3++)
	      {
		int j1d = tensor2segm[j2 * 2 + j3];
		if (j1d == -1) continue;
		
		AutoDiff<1> shapejj1 = poly[j2] * polyy[j3];
		
		shapesy(i2, j1d) = shapejj1.Value();
		shapesdy(i2, j1d) = shapejj1.DValue(0);
	      }          
	}        
      
      int nip = ir1dx.GetNIP() * ir1dy.GetNIP();
      
      FlatVector<Mat<2> > tdmat(nip, lh);
      FlatMatrix<Vec<2> > fac_xdx(ir1dx.GetNIP(), ndof, lh);
      
      for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
	{
	  AutoDiff<1> xr1 (ir1dx[i1](0), 0);
	  AutoDiff<1> * hp1 = &polx[1];
	  LegendrePolynomialMult (order-2, 2*xr1-1, xr1, hp1);
	  polx[0] = 1;
	  
	  AutoDiff<1> hv = 1.0;
	  for (int i = 0; i <= order; i++)
	    {
	      powx[i] = hv;
	      hv *= (1-xr1);
	    }
	  
	  for (int jj = 0; jj < ndof; jj++)
	    {
	      // fx = polxi_i * (1-xi)^{j+l}
	      // if scaled fx.Val() = polxi_i * (1-xi)^{j+l-1}
	      AutoDiff<1> fx = polx[trig2tensor[jj][0]] 
		* powx[trig2tensor[jj][1]+trig2tensor[jj][2]-scaled];
	      
	      fac_xdx(i1, jj)(0) = fx.Value();
	      if(scaled)
		fx = polx[trig2tensor[jj][0]] 
		  * powx[trig2tensor[jj][1]+trig2tensor[jj][2]];
	      
	      fac_xdx(i1,jj)(1) = fx.DValue(0); 
	    }
	}

          NgProfiler::StopTimer (timeri);

          NgProfiler::StartTimer (timer1);

	  double tp_p[2], trig_p[2];
          int ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
	      tp_p[0] = ir1dx[i1](0);
	      double duffy_jac_det = duffy_trafo.GetJacobiDet(tp_p);

              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
		  tp_p[1] = ir1dy[i2](0);
	                        
		  duffy_trafo.GetIPTrig(tp_p,trig_p);
		  IntegrationPoint ip(trig_p[0], trig_p[1], 0, 
				      ir1dx[i1].Weight()*ir1dy[i2].Weight());
		  
		  SpecificIntegrationPoint<2,2> sip(ip, eltrans, lh);
		  
		  Mat<2> dmat, hmat, trans; 
		  dmatop.GenerateMatrix (fel, sip, dmat, lh);
		  dmat *= fabs (sip.GetJacobiDet()) * sip.IP().Weight() * duffy_jac_det;
		  		  
		  trans = duffy_trafo.GetJacobianInverse(tp_p,scaled) * sip.GetJacobianInverse(); 
		  
		  hmat = trans * dmat;
		  tdmat[ii] = hmat * Trans (trans); 
		  
                }
            }
	  

          NgProfiler::StopTimer (timer1);

          ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
	      sum1d = 0.0;
              NgProfiler::StartTimer (timertrans);
	      
              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
		  // NgProfiler::StartTimer (timer4);
                  
		  for (int j = 0; j < ndsegm; j++)
		    for (int l = 0; l < 2; l++)
		      {
			mult1d(j)(0,l) = tdmat[ii](0,l) * shapesy(i2,j);
			mult1d(j)(1,l) = tdmat[ii](1,l) * shapesdy(i2,j);
		      }
		  
		  for (int j = 0; j < ndsegm; j++)
		    for (int k = 0; k <= j; k++)
		      for (int l = 0; l < 2; l++)
			{
			  sum1d(j,k)(l,0) += mult1d(j)(l,0) * shapesy (i2,k);   
			  sum1d(j,k)(l,1) += mult1d(j)(l,1) * shapesdy (i2,k);   
			}
		}
		  
	      for (int j = 0; j < ndsegm; j++)
		for (int k = 0; k < j; k++)
		  sum1d(k,j) = Trans(sum1d(j,k));
	  	      
	      // NgProfiler::StopTimer (timer4);
	      
	      // NgProfiler::RegionTimer reg3 (timer3);
	      // NgProfiler::AddFlops (timer3, ndtrig * (5 * ndtrig + 7 * ndsegm));
	     
	      NgProfiler::StopTimer (timertrans);
	      
	      
              NgProfiler::StartTimer (timer2a);
              
	      
              // mult j
              for (int jj = 0; jj < ndof; jj++)
                {
                  int jsegm = trig2segm[jj];
                  Vec<2> fdxfx; 
		  fdxfx(0)= fac_xdx(i1,jj)(1);
		  fdxfx(1)= fac_xdx(i1,jj)(0);
                  for (int k = 0; k < ndsegm; k++)
                    mult2d(k,jj) = sum1d(jsegm, k) * fdxfx;
                }

              NgProfiler::StopTimer (timer2a);

              NgProfiler::StartTimer (timer2b);
              NgProfiler::AddFlops (timer2b, ndof * 2 * ndof );

              // multk
              for (int kk = 0; kk < ndof; kk++)
                {
                  int ksegm = trig2segm[kk];
		  Vec<2> fdxfx; 
		  fdxfx(0)= fac_xdx(i1,kk)(1);
		  fdxfx(1)= fac_xdx(i1,kk)(0);
		  for (int jj = 0; jj <= kk; jj++)
                    elmat(kk*ndof+jj) += InnerProduct (fdxfx,  mult2d(ksegm,jj));
                }

              NgProfiler::StopTimer (timer2b);
            }
      

          for (int j = 0; j < ndof; j++)
            for (int k = 0; k < j; k++)
              elmat(k,j) = elmat(j,k);

          lh.CleanUp (heapp);
	  
    }
 
    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
    


      typedef H1HighOrderTetTPSZ FEL;
      bool scaled=1; 
     
      if (dynamic_cast<const FEL*> (&bfel))
        {
          const FEL & fel = static_cast<const FEL&> (bfel);
          static int timer = NgProfiler::CreateTimer ("Fast assembling");
          static int timer1 = NgProfiler::CreateTimer ("Fast assembling - geometry");
          static int timer2a = NgProfiler::CreateTimer ("Fast assembling - transform x-j");
          static int timer2b = NgProfiler::CreateTimer ("Fast assembling - transform x-k");
          static int timer3 = NgProfiler::CreateTimer ("Fast assembling - transform y");
          static int timer4 = NgProfiler::CreateTimer ("Fast assembling - transform z");
          static int timeri = NgProfiler::CreateTimer ("Fast assembling - init");
          static int timertrans = NgProfiler::CreateTimer ("Fast assembling - transform y,z");
          
          NgProfiler::RegionTimer reg (timer);
          NgProfiler::StartTimer (timeri);
          int ndof = fel.GetNDof();
      
          elmat.AssignMemory (ndof, ndof, lh);
          elmat = 0;


          int order = fel.Order();
          // const IntegrationRule & ir1dx = SelectIntegrationRule (ET_SEGM, 2*order);
          const IntegrationRule & ir1dy = SelectIntegrationRule (ET_SEGM, 2*order-1);
          const IntegrationRule & ir1dz = SelectIntegrationRule (ET_SEGM, 2*order-2);
      
          static IntegrationRule ir1dx;
          if (ir1dx.GetNIP() == 0)
            {
              int np = order;
              ARRAY<double> xi(np), wi(np);
              ComputeGaussJacobiRule (np, xi, wi, 2, 0);
              for (int i = 0; i < np; i++)
                ir1dx.AddIntegrationPoint (new IntegrationPoint (xi[i], 0, 0, wi[i]));
            }

          ArrayMem<AutoDiff<1>, 20> polx(order+1), powx(order+1);
          ArrayMem<AutoDiff<1>, 20> poly(order+1), powy(order+1);
          ArrayMem<AutoDiff<1>, 20> polz(order+1), polzz(2);

	  DuffyTransformationTet duffy_trafo(fel.GetVNums()); 

          void * heapp = lh.GetPointer();

          FlatArray<int[4]> tet2tensor (ndof, lh);
	  fel.GetTet2TensorMapping(tet2tensor);  
	 
	  int nd1d = 2 * (order);
          int nd2d = order * nd1d; 
	  FlatArray<int> tensor2trig(nd2d, lh); 
	  ARRAY<int[3] > trig2tensor; 
         	  
	  int ndtrig = fel.GetTrig2TensorMappings(tet2tensor, trig2tensor, tensor2trig); 
	  
	  FlatArray<int> tet2trig (ndof, lh);
          for (int i = 0; i < ndof; i++)
            tet2trig[i] = tensor2trig[tet2tensor[i][1] * nd1d + tet2tensor[i][2] * 2 + tet2tensor[i][3]];
	  
          FlatArray<int> tensor2segm ( nd1d, lh);
	  int ndsegm = fel.GetTensor2SegmMapping(tet2tensor, tensor2segm); 
	  
	  FlatArray<int> trig2segm (ndtrig, lh);
          for (int i = 0; i < ndtrig; i++)
            trig2segm[i] = tensor2segm[trig2tensor[i][1] * 2 + trig2tensor[i][2]];
                
          FlatMatrix<Mat<2> > sum2d(ndtrig, lh);
          FlatMatrix<Mat<3> > sum1d(ndsegm, lh);

          FlatVector<Mat<3> > mult1d(ndsegm, lh);
          FlatMatrix<Vec<6> > mult2d(ndsegm, ndtrig, lh);
          FlatMatrix<Vec<2> > mult3d(ndtrig, ndof, lh);

          FlatMatrix<> shapesz (ir1dz.GetNIP(), ndsegm, lh);
          FlatMatrix<> shapesdz (ir1dz.GetNIP(), ndsegm, lh);
          FlatMatrix<> shapeszT (ndsegm, ir1dz.GetNIP(), lh);
          FlatMatrix<> shapesdzT (ndsegm, ir1dz.GetNIP(), lh);

          for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++)
            {
              AutoDiff<1> zr1(ir1dz[i3](0), 0);
              AutoDiff<1> * hp1 = &polz[1];
              LegendrePolynomialMult (order-2, 2*zr1-1, zr1, hp1);
              polz[0] = 1;
              polz[order] = 0;

              polzz[0] = 1;
              polzz[1] = (1-zr1); 

              for (int j3 = 0; j3 < order; j3++)
                for (int j4 = 0; j4 < 2; j4++)
                  {
                    int j1d = tensor2segm[j3 * 2 + j4];
                    if (j1d == -1) continue;

                    AutoDiff<1> shapejj1 = polz[j3] * polzz[j4];

                    shapesz(i3, j1d) = shapejj1.Value();
                    shapesdz(i3, j1d) = shapejj1.DValue(0);
                  }          
            }        
          shapeszT = Trans (shapesz);
          shapesdzT = Trans (shapesdz);

          int nip = ir1dx.GetNIP() * ir1dy.GetNIP() * ir1dz.GetNIP();
      
          FlatVector<Mat<3> > tdmat(nip, lh);
          FlatMatrix<Vec<2> > fac_ydy(ir1dy.GetNIP(), ndtrig, lh);
          FlatMatrix<> fac_y_s(ir1dy.GetNIP(), ndtrig, lh);
          
	  for (int iy = 0; iy < ir1dy.GetNIP(); iy++)
            {
              AutoDiff<1> yr1 (ir1dy[iy](0), 0);
              AutoDiff<1> * hp1 = &poly[1];
              LegendrePolynomialMult (order-2, 2*yr1-1, yr1, hp1);
              poly[0] = 1;
              
              AutoDiff<1> hv = 1.0;
              for (int i = 0; i <= order; i++)
                {
                  powy[i] = hv;
                  hv *= (1-yr1);
                }
          
              for (int jj = 0; jj < ndtrig; jj++)
                {
		  AutoDiff<1> fy = poly[trig2tensor[jj][0]] * powy[trig2tensor[jj][1]+trig2tensor[jj][2]];
                  fac_ydy(iy,jj)(0) = fy.Value();
                  fac_ydy(iy,jj)(1) = fy.DValue(0);
		  fac_y_s(iy,jj) = fy.Value();
		}

	      if(scaled) 
		for (int jj = 0; jj < ndtrig; jj++)
		  {
		    int ind_s = trig2tensor[jj][1]+trig2tensor[jj][2];
		    if(ind_s > 0)
		      fac_y_s(iy,jj) = poly[trig2tensor[jj][0]].Value() * powy[ind_s-1].Value();
		    else 
		      fac_y_s(iy,jj) = 1.; //poly[trig2tensor[jj][0]].Value() * powy[ind_s].Value() / (1-yr1.Value());
		  }
	    }




          FlatMatrix<Vec<2> > fac_xdx(ir1dx.GetNIP(), ndof, lh);

          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
              AutoDiff<1> xr1 (ir1dx[i1](0), 0);
              AutoDiff<1> * hp1 = &polx[1];
              LegendrePolynomialMult (order-2, 2*xr1-1, xr1, hp1);
              polx[0] = 1;

              AutoDiff<1> hv = 1.0;
              for (int i = 0; i <= order; i++)
                {
                  powx[i] = hv;
                  hv *= (1-xr1);
                }


              for (int jj = 0; jj < ndof; jj++)
                {
                  AutoDiff<1> fx = 
                    polx[tet2tensor[jj][0]] *
                    powx[tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3]];
		  	      
		  if(!scaled)
		    fac_xdx(i1, jj)(0) = fx.Value();
		  else 
		    {
		      int ind_s = tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3];
		      if(ind_s>0) 
			fac_xdx(i1,jj)(0) = polx[tet2tensor[jj][0]].Value() * powx[ind_s-1].Value();
		      else 
			fac_xdx(i1, jj)(0) = 1.; //polx[tet2tensor[jj][0]].Value() * powx[ind_s].Value()/(1-xr1.Value());
		    }
                  fac_xdx(i1, jj)(1) = fx.DValue(0);
                }
            }


          NgProfiler::StopTimer (timeri);

          NgProfiler::StartTimer (timer1);

	  double hex_p[3], tet_p[3];
          int ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
	      hex_p[0] = ir1dx[i1](0);

              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
		  hex_p[1] = ir1dy[i2](0);
                  double duffy_jac_det = duffy_trafo.GetJacobiDet(hex_p);
		  duffy_trafo.SetEtaForJacobianInverse(hex_p,scaled); 
		  
                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
		      hex_p[2] = ir1dz[i3](0);

		      duffy_trafo.GetIPTet(hex_p,tet_p);
                      IntegrationPoint ip(tet_p[0], tet_p[1], tet_p[2], 
                                          ir1dx[i1].Weight()*ir1dy[i2].Weight()*ir1dz[i3].Weight());

                      SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ip, eltrans, lh);

                      Mat<3> dmat, hmat, trans; 
                      dmatop.GenerateMatrix (fel, sip, dmat, lh);
                      dmat *= fabs (sip.GetJacobiDet()) * sip.IP().Weight() * duffy_jac_det;

		      trans = duffy_trafo.GetJacobianInverseZ(hex_p,scaled) * sip.GetJacobianInverse(); 
		      //trans = duffy_trafo.GetJacobianInverse(hex_p) * sip.GetJacobianInverse(); 
		     
		      hmat = trans * dmat;
                      tdmat[ii] = hmat * Trans (trans); 
		    }
                }
            }


          NgProfiler::StopTimer (timer1);


          ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
              sum2d = 0.0;

              NgProfiler::StartTimer (timertrans);

              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
                  // NgProfiler::StartTimer (timer4);

                  sum1d = 0.0;

                  // NgProfiler::AddFlops (timer4, ir1dz.GetNIP() * ndsegm * (ndsegm+1) * 6);

                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
                      for (int j = 0; j < ndsegm; j++)
                        for (int l = 0; l < 3; l++)
                          {
                            mult1d(j)(0,l) = tdmat[ii](0,l) * shapesz(i3,j);
                            mult1d(j)(1,l) = tdmat[ii](1,l) * shapesz(i3,j);
                            mult1d(j)(2,l) = tdmat[ii](2,l) * shapesdz(i3,j);
                          }

                      for (int j = 0; j < ndsegm; j++)
                        for (int k = 0; k <= j; k++)
                          for (int l = 0; l < 3; l++)
                            {
                              sum1d(j,k)(l,0) += mult1d(j)(l,0) * shapesz (i3,k);   
                              sum1d(j,k)(l,1) += mult1d(j)(l,1) * shapesz (i3,k);   
                              sum1d(j,k)(l,2) += mult1d(j)(l,2) * shapesdz(i3,k);  
                            }
                    }
                  for (int j = 0; j < ndsegm; j++)
                    for (int k = 0; k < j; k++)
                      sum1d(k,j) = Trans(sum1d(j,k));



                  // NgProfiler::StopTimer (timer4);

                  // NgProfiler::RegionTimer reg3 (timer3);
                  // NgProfiler::AddFlops (timer3, ndtrig * (5 * ndtrig + 7 * ndsegm));

                  // mult pj:
                  for (int jj = 0; jj < ndtrig; jj++)
                    {
                      int j1d = trig2segm[jj];

                      double fy = fac_ydy(i2,jj)(0);
                      double fdy = fac_ydy(i2,jj)(1);
		      double fy_s = (scaled) ? fac_y_s(i2,jj): fac_ydy(i2,jj)(0); 

                      for (int k = 0; k < ndsegm; k++)
                        {
                          /*
                            const Mat<3> & hsum1d = sum1d(j1d, k);
                            Vec<6> & hmult2d = mult2d(k,jj);

                            hmult2d(0) = fy * hsum1d(0,0);
                            hmult2d(1) = fy * hsum1d(0,1);
                            hmult2d(2) = fy * hsum1d(0,2);
                            hmult2d(3) = fy * hsum1d(2,0) + fdy * hsum1d(1,0);
                            hmult2d(4) = fy * hsum1d(2,1) + fdy * hsum1d(1,1);
                            hmult2d(5) = fy * hsum1d(2,2) + fdy * hsum1d(1,2);
                          */

                          mult2d(k,jj)(0) = fy * sum1d(j1d, k)(0,0);
                          mult2d(k,jj)(1) = fy * sum1d(j1d, k)(0,1);
                          mult2d(k,jj)(2) = fy * sum1d(j1d, k)(0,2);
                          mult2d(k,jj)(3) = fy_s * sum1d(j1d, k)(2,0) + fdy * sum1d(j1d,k)(1,0);
                          mult2d(k,jj)(4) = fy_s * sum1d(j1d, k)(2,1) + fdy * sum1d(j1d,k)(1,1);
                          mult2d(k,jj)(5) = fy_s * sum1d(j1d, k)(2,2) + fdy * sum1d(j1d,k)(1,2);
                        }
                    }

                  // mult pk:
                  for (int kk = 0; kk < ndtrig; kk++)
                    {
                      int k1d = trig2segm[kk];

                      double fy = fac_ydy(i2,kk)(0);
                      double fdy = fac_ydy(i2,kk)(1);
		      double fy_s = (scaled) ? fac_y_s(i2,kk): fac_ydy(i2,kk)(0); 
                  
                      for (int jj = 0; jj <= ndtrig; jj++)
                        {
                          Mat<2> & hsum2d = sum2d(kk,jj);
                          const Vec<6> & hmult2d = mult2d(k1d,jj);

                          hsum2d(0,0) += fy_s * hmult2d(5) + fdy * hmult2d(4);
                          hsum2d(1,0) += fy_s * hmult2d(2) + fdy * hmult2d(1);
                          hsum2d(0,1) += fy * hmult2d(3);
                          hsum2d(1,1) += fy * hmult2d(0);

                          /*
                            sum2d(kk,jj)(0,0) += fy * mult2d(k1d,jj)(5) + fdy * mult2d(k1d,jj)(4);
                            sum2d(kk,jj)(1,0) += fy * mult2d(k1d,jj)(2) + fdy * mult2d(k1d,jj)(1);
                            sum2d(kk,jj)(0,1) += fy * mult2d(k1d,jj)(3);
                            sum2d(kk,jj)(1,1) += fy * mult2d(k1d,jj)(0);
                          */
                        }
                    }
                }

              for (int j = 0; j < ndtrig; j++)
                for (int k = 0; k < j; k++)
                  sum2d(j,k) = Trans(sum2d(k,j));

              NgProfiler::StopTimer (timertrans);


              NgProfiler::StartTimer (timer2a);
              NgProfiler::AddFlops (timer2a, ndof * 4 * ndtrig);

              // mult j
              for (int jj = 0; jj < ndof; jj++)
                {
                  int jtrig = tet2trig[jj];
                  Vec<2> fxfdx = fac_xdx(i1,jj);
                  for (int k = 0; k < ndtrig; k++)
                    mult3d(k,jj) = sum2d(jtrig, k) * fxfdx;
                }

              NgProfiler::StopTimer (timer2a);

              NgProfiler::StartTimer (timer2b);
              NgProfiler::AddFlops (timer2b, ndof * 2 * ndof );

              // multk
              for (int kk = 0; kk < ndof; kk++)
                {
                  int ktrig = tet2trig[kk];
                  Vec<2> fxfdx = fac_xdx(i1,kk);
                  for (int jj = 0; jj <= kk; jj++)
                    elmat(kk*ndof+jj) += InnerProduct (fxfdx,  mult3d(ktrig,jj));
                }

              NgProfiler::StopTimer (timer2b);
            }
      

          for (int j = 0; j < ndof; j++)
            for (int k = 0; k < j; k++)
              elmat(k,j) = elmat(j,k);

          lh.CleanUp (heapp);

        }
#ifdef trigtpsz
      else if ((dynamic_cast<const H1HighOrderTrigTPSZ*> (&bfel))
	       AssembleElementMatrixTrig (bfel,eltrans, elmat,lh,scaled);         
#endif
      else  if (dynamic_cast<const H1HighOrderHexTPSZ*> (&bfel))
        {
          const H1HighOrderHexTPSZ & fel = static_cast<const H1HighOrderHexTPSZ&> (bfel);

          static int timerhex = NgProfiler::CreateTimer ("Fast assembling - Hex");
          static int timerhex1 = NgProfiler::CreateTimer ("Fast assembling - Hex yz");
          static int timerhex2a = NgProfiler::CreateTimer ("Fast assembling - Hex xa");
          static int timerhex2b = NgProfiler::CreateTimer ("Fast assembling - Hex xb");
          static int timerhexgeom = NgProfiler::CreateTimer ("Fast assembling - geometry");


          NgProfiler::RegionTimer reg (timerhex);

          // orig -> AssembleElementMatrix (fel, eltrans, elmat, lh);



          int ndof = fel.GetNDof();
      
          elmat.AssignMemory (ndof, ndof, lh);
          elmat = 0;

          

          int order = fel.Order();
          const IntegrationRule & ir1d = SelectIntegrationRule (ET_SEGM, 2*order);

          FlatMatrix<> shape1d(ir1d.GetNIP(), order+1, lh);
          FlatMatrix<> dshape1d(ir1d.GetNIP(), order+1, lh);
          FlatMatrix<> shape1dT(order+1, ir1d.GetNIP(), lh);
          FlatMatrix<> dshape1dT(order+1, ir1d.GetNIP(), lh);
          FlatMatrix<Vec<2> > sds1dT(order+1, ir1d.GetNIP(), lh);
          
          FlatArray<AutoDiff<1> > shapesad(order+1, lh);
          
          for (int i = 0; i < ir1d.GetNIP(); i++)
            {
              AutoDiff<1> adx (ir1d[i](0), 0);
              AutoDiff<1> xi = 2*adx-1;
              
              shapesad[0] = 1-adx;
              shapesad[1] = adx;
              
              AutoDiff<1> * hp = &shapesad[2];
              LegendrePolynomialMult (order-2, xi, 0.25 * (1-xi*xi), hp);

              for (int j = 0; j <= order; j++)
                {
                  shape1d(i,j) = shapesad[j].Value();
                  dshape1d(i,j) = shapesad[j].DValue(0);
                  sds1dT(j,i)(0) = shapesad[j].DValue(0);
                  sds1dT(j,i)(1) = shapesad[j].Value();
                }
            }
          shape1dT = Trans(shape1d);
          dshape1dT = Trans(dshape1d);

          NgProfiler::StartTimer (timerhexgeom);


          int nip = ir1d.GetNIP() * ir1d.GetNIP() * ir1d.GetNIP();
          FlatVector<Mat<3> > tdmat(nip, lh);

          int ii = 0;
          for (int i1 = 0; i1 < ir1d.GetNIP(); i1++)
            for (int i2 = 0; i2 < ir1d.GetNIP(); i2++)
              for (int i3 = 0; i3 < ir1d.GetNIP(); i3++, ii++)
                {
                  double x = ir1d[i1](0);
                  double y = ir1d[i2](0);
                  double z = ir1d[i3](0);
                  
                  IntegrationPoint ip(x, y, z,
                                      ir1d[i1].Weight()*ir1d[i2].Weight()*ir1d[i3].Weight());
                  
                  SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ip, eltrans, lh);
                  
                  Mat<3> dmat, trans, hmat;
                  
                  dmatop.GenerateMatrix (fel, sip, dmat, lh);
                  dmat *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
                  
                  trans = sip.GetJacobianInverse();
                  hmat = trans * dmat;
                  tdmat[ii] = hmat * Trans (trans); 
                }

          
          // ArrayMem<int[3], 100> hex2tensor(ndof);
          // ArrayMem<double, 100> fac(ndof);

          FlatArray<int[3]> hex2tensor(ndof, lh);
          FlatArray<double> fac(ndof, lh);
          FlatMatrix<> dshape(ndof, 3, lh);
          FlatMatrix<> ddshape(ndof, 3, lh);


          fel.GetHex2TensorMapping (hex2tensor, fac);
          
          FlatArray<int> tensor2hex(ndof, lh);
          tensor2hex = -1;
          for (int i = 0; i < ndof; i++)
            tensor2hex[ (hex2tensor[i][0] * (order+1) + hex2tensor[i][1]) * (order+1) + hex2tensor[i][2]] = i;


          ii = 0;

          /*
          for (int i1 = 0; i1 < ir1d.GetNIP(); i1++)
            for (int i2 = 0; i2 < ir1d.GetNIP(); i2++)
              for (int i3 = 0; i3 < ir1d.GetNIP(); i3++, ii++)
                {
                  IntegrationPoint ip(ir1d[i1](0), ir1d[i2](0), ir1d[i3](0), 1);
                  // fel.CalcDShape (ip, dshape);

                  for (int j = 0; j < ndof; j++)
                    {
                      dshape(j,0) = fac[j] * dshape1d(i1, hex2tensor[j][0]) *  shape1d(i2, hex2tensor[j][1]) *  shape1d(i3, hex2tensor[j][2]);
                      dshape(j,1) = fac[j] *  shape1d(i1, hex2tensor[j][0]) * dshape1d(i2, hex2tensor[j][1]) *  shape1d(i3, hex2tensor[j][2]);
                      dshape(j,2) = fac[j] *  shape1d(i1, hex2tensor[j][0]) *  shape1d(i2, hex2tensor[j][1]) * dshape1d(i3, hex2tensor[j][2]);
                    }


                  ddshape = dshape * tdmat[ii];

                  elmat += ddshape * Trans(dshape);
                }
          */

          int nd2d = (order+1)*(order+1);

          FlatMatrix<Mat<3> > sum1d( order+1, lh);
          FlatMatrix<Mat<2> > sum2d( nd2d, lh);

          FlatMatrix<Vec<6> > mult2d( order+1, nd2d, lh);

          FlatMatrix<Vec<2> > mult3d( nd2d, ndof, lh);
          FlatVector<Mat<2> > allsum2d (nd2d*nd2d*ir1d.GetNIP(), lh);

          NgProfiler::StopTimer (timerhexgeom);



          NgProfiler::StartTimer (timerhex1);

          for (int i1 = 0; i1 < ir1d.GetNIP(); i1++)
            {
              sum2d = 0.0;

              for (int i2 = 0; i2 < ir1d.GetNIP(); i2++)

                {
                  sum1d = 0.0;

                  for (int i3 = 0; i3 < ir1d.GetNIP(); i3++, ii++)
                    {
                      int ii = (i1*ir1d.GetNIP()+i2)*ir1d.GetNIP()+i3;
                      for (int j3 = 0; j3 <= order; j3++)
                        {
                          Mat<3> mult1d;
                          for (int l = 0; l < 3; l++)
                            {
                              mult1d(0,l) = tdmat[ii](0,l) *  shape1d(i3,j3);
                              mult1d(1,l) = tdmat[ii](1,l) *  shape1d(i3,j3);
                              mult1d(2,l) = tdmat[ii](2,l) * dshape1d(i3,j3);
                            }

                          for (int k3 = 0; k3 <= j3; k3++)
                            for (int l = 0; l < 3; l++)
                              {
                                sum1d(j3,k3) (l,0) += mult1d(l,0) *  shape1d(i3,k3);
                                sum1d(j3,k3) (l,1) += mult1d(l,1) *  shape1d(i3,k3);
                                sum1d(j3,k3) (l,2) += mult1d(l,2) *  dshape1d(i3,k3);
                              }
                        }
                    }

                  for (int k3 = 0; k3 <= order; k3++)
                    for (int j3 = k3+1; j3 <= order; j3++)
                      sum1d(k3,j3) = Trans (sum1d(j3,k3));
                  


                  for (int j2 = 0; j2 <= order; j2++)
                    for (int j3 = 0; j3 <= order; j3++)
                      {
                        int j2d = j2 * (order+1) + j3;                        
                        for (int k3 = 0; k3 <= order; k3++)
                          {
                            mult2d(k3, j2d)(0) = sum1d(j3,k3)(0,0) * shape1d(i2,j2);
                            mult2d(k3, j2d)(1) = sum1d(j3,k3)(0,1) * shape1d(i2,j2);
                            mult2d(k3, j2d)(2) = sum1d(j3,k3)(0,2) * shape1d(i2,j2);
                            mult2d(k3, j2d)(3) = sum1d(j3,k3)(1,0) * dshape1d(i2,j2) + sum1d(j3,k3)(2,0) * shape1d(i2,j2);
                            mult2d(k3, j2d)(4) = sum1d(j3,k3)(1,1) * dshape1d(i2,j2) + sum1d(j3,k3)(2,1) * shape1d(i2,j2);
                            mult2d(k3, j2d)(5) = sum1d(j3,k3)(1,2) * dshape1d(i2,j2) + sum1d(j3,k3)(2,2) * shape1d(i2,j2);
                          }
                      }

                  for (int k2 = 0; k2 <= order; k2++)
                    for (int k3 = 0; k3 <= order; k3++)
                      {
                        int k2d = k2 * (order+1) + k3;                        
                        for (int j2d = 0; j2d <= k2d; j2d++)
                          {
                            sum2d(k2d,j2d)(0,0) += mult2d(k3,j2d)(0) * shape1d(i2,k2);
                            sum2d(k2d,j2d)(1,0) += mult2d(k3,j2d)(1) * dshape1d(i2,k2) + mult2d(k3,j2d)(2) * shape1d(i2,k2);
                            sum2d(k2d,j2d)(0,1) += mult2d(k3,j2d)(3) * shape1d(i2,k2);
                            sum2d(k2d,j2d)(1,1) += mult2d(k3,j2d)(4) * dshape1d(i2,k2) + mult2d(k3,j2d)(5) * shape1d(i2,k2);
                          }
                      }
                }
              for (int k2d = 0; k2d < nd2d; k2d++)
                for (int j2d = k2d+1; j2d < nd2d; j2d++)
                  sum2d(k2d,j2d) = Trans (sum2d(j2d,k2d));

              for (int j2d = 0; j2d < nd2d; j2d++)
                for (int k2d = 0; k2d < nd2d; k2d++)
                  allsum2d( (j2d*nd2d+k2d) * ir1d.GetNIP() + i1) = sum2d(j2d,k2d);

            }
          NgProfiler::StopTimer (timerhex1);


          NgProfiler::StartTimer (timerhex2a);
          NgProfiler::AddFlops (timerhex2a, ir1d.GetNIP() * ndof * (ndof + nd2d*4));

          
          FlatMatrix<Vec<2> > help(ndof, ir1d.GetNIP(), lh);
          for (int k2d = 0; k2d < nd2d; k2d++)
            {
              for (int jj = 0; jj < ndof; jj++)
                {
                  int j1d = hex2tensor[jj][0];
                  int j2d = hex2tensor[jj][1] * (order+1) + hex2tensor[jj][2];
                  int ind = (k2d*nd2d+j2d) * ir1d.GetNIP();
                  
                  for (int i1 = 0; i1 < ir1d.GetNIP(); i1++)
                    help(jj, i1) = allsum2d(ind+i1) * sds1dT(j1d, i1);
                }
              
              for (int k1d = 0; k1d < order+1; k1d++)
                {
                  int kk = tensor2hex[k1d * nd2d + k2d];

                  Vec<2> * p_sds = &sds1dT(k1d,0);
                  for (int j = 0; j <= kk; j++)
                    {
                      Vec<2> * p_help = &help(j,0);

                      double sum = 0;

                      for (int i1 = 0; i1 < ir1d.GetNIP(); i1++)
                        sum +=  InnerProduct (p_sds[i1], p_help[i1]);

                      elmat(kk*ndof+j) = fac[kk]*fac[j]*sum;
                    }
                }
            }

          // upper half
          for (int j = 0; j < ndof; j++)
            for (int k = j+1; k < ndof; k++)
              elmat(j,k) = elmat(k,j);

          NgProfiler::StopTimer (timerhex2a);          

          return;
        }

      else


        // if (! dynamic_cast<const FEL*> (&bfel))
        {

          static int timeretc = NgProfiler::CreateTimer ("Fast assembling - etc");
          NgProfiler::RegionTimer reg (timeretc);


          orig -> AssembleElementMatrix (bfel, eltrans, elmat, lh);
          return;
        }
      
    }
    
  };

/**
   Element vector assembling.
   Assembling for linear-forms of type $\int D (B u) dx$.
 */

#ifdef SOURCETPSZ 
///
template <int DIM>
class SourceIntegratorTPSZ 
  : public LinearFormIntegrator
{
public:
 
  typedef DiffOpId<DIM> DIFFOP;
public:
  enum { DIM_SPACE = DIFFOP::DIM_SPACE };
  enum { DIM_ELEMENT = DIFFOP::DIM_ELEMENT };
  enum { DIM_DMAT = DIFFOP::DIM_DMAT };
  enum { DIM = DIFFOP::DIM }; 

 ///
  SourceIntegratorTPSZ (CoefficientFunction * coeff)
    : T_BIntegrator<DiffOpId<D>, DVec<1>, FEL> (DVec<1> (coeff))
  { ; }
  
  static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
  {
    return new SourceIntegratorTPSZ (coeffs[0]);
  }

  ///
  virtual string Name () const { return "SourceTPSZ"; }

  virtual void
  AssembleElementVector (const FiniteElement & bfel,
			 const ElementTransformation & eltrans, 
			 FlatVector<double> & elvec,
			 LocalHeap & locheap) const
  {
    typedef H1HighOrderTetTPSZ FEL;
       
    if (dynamic_cast<const FEL*> (&bfel))
      {
	const FEL & fel = static_cast<const FEL&> (bfel);
	static int timer = NgProfiler::CreateTimer ("Fast assembling");
	static int timer1 = NgProfiler::CreateTimer ("Fast assembling - geometry");
	static int timer2a = NgProfiler::CreateTimer ("Fast assembling - transform x-j");
	static int timer2b = NgProfiler::CreateTimer ("Fast assembling - transform x-k");
	static int timer3 = NgProfiler::CreateTimer ("Fast assembling - transform y");
	static int timer4 = NgProfiler::CreateTimer ("Fast assembling - transform z");
	static int timeri = NgProfiler::CreateTimer ("Fast assembling - init");
	static int timertrans = NgProfiler::CreateTimer ("Fast assembling - transform y,z");
        
	NgProfiler::RegionTimer reg (timer);
	NgProfiler::StartTimer (timeri);
	int ndof = fel.GetNDof();
	
        
	elvec.AssignMemory (ndof * DIM, lh); // ??? 
	elvec = 0;
		
	int order = fel.Order();
	// const IntegrationRule & ir1dx = SelectIntegrationRule (ET_SEGM, 2*order);
	const IntegrationRule & ir1dy = SelectIntegrationRule (ET_SEGM, 2*order-1);
	const IntegrationRule & ir1dz = SelectIntegrationRule (ET_SEGM, 2*order-2);
	
	static IntegrationRule ir1dx;
	if (ir1dx.GetNIP() == 0)
	  {
              int np = order;
              ARRAY<double> xi(np), wi(np);
              ComputeGaussJacobiRule (np, xi, wi, 2, 0);
              for (int i = 0; i < np; i++)
                ir1dx.AddIntegrationPoint (new IntegrationPoint (xi[i], 0, 0, wi[i]));
	  }
	
	ArrayMem<double, 20> polx(order+1), powx(order+1);
	ArrayMem<double, 20> poly(order+1), powy(order+1);
	ArrayMem<double, 20> polz(order+1), polzz(2);
	
	DuffyTransformationTet duffy_trafo(fel.GetVNums()); 
	  
	void * heapp = lh.GetPointer();
	
	FlatArray<int[4]> tet2tensor (ndof, lh);
	fel.GetTet2TensorMapping(tet2tensor);  
	
	int nd1d = 2 * (order);
	int nd2d = order * nd1d; 
	FlatArray<int> tensor2trig(nd2d, lh); 
	ARRAY<int[3] > trig2tensor; 
        
	int ndtrig = fel.GetTrig2TensorMappings(tet2tensor, trig2tensor, tensor2trig); 
	
	FlatArray<int> tet2trig (ndof, lh);
	for (int i = 0; i < ndof; i++)
	  tet2trig[i] = tensor2trig[tet2tensor[i][1] * nd1d + tet2tensor[i][2] * 2 + tet2tensor[i][3]];
	
	FlatArray<int> tensor2segm ( nd1d, lh);
	int ndsegm = fel.GetTensor2SegmMapping(tet2tensor, tensor2segm); 
	
	FlatArray<int> trig2segm (ndtrig, lh);
	for (int i = 0; i < ndtrig; i++)
	  trig2segm[i] = tensor2segm[trig2tensor[i][1] * 2 + trig2tensor[i][2]];
	
	FlatMatrix<> sum2d(ndtrig, lh);
	FlatMatrix<> sum1d(ndsegm, lh);
		
	FlatMatrix<> shapesz (ir1dz.GetNIP(), ndsegm, lh);
	FlatMatrix<> shapesdz (ir1dz.GetNIP(), ndsegm, lh);
        

	// compute shapez_kl(i3)
	for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++)
	  {
	    double zr1 = ir1dz[i3](0);
	    double * hp1 = &polz[1];
	    LegendrePolynomialMult (order-2, 2*zr1-1, zr1, hp1);
	    polz[0] = 1;
	    polz[order] = 0;
	    
	    polzz[0] = 1;
	    polzz[1] = (1-zr1); 
	    
	    for (int j3 = 0; j3 < order; j3++)
	      for (int j4 = 0; j4 < 2; j4++)
		{
		  int j1d = tensor2segm[j3 * 2 + j4];
		  if (j1d == -1) continue;
		  
		  shapesz(i3,j1d) = polz[j3] * polzz[j4];
		}          
	  }        
	
	  int nip = ir1dx.GetNIP() * ir1dy.GetNIP() * ir1dz.GetNIP();
	  
	  FlatVector<Vec<1> > tdvec(nip, lh);
	  FlatMatrix<> fac_y(ir1dy.GetNIP(), ndtrig, lh);
	  	  
	  for (int iy = 0; iy < ir1dy.GetNIP(); iy++)
            {
	      double yr = ir1dy[iy](0); 
	      double * hp1 = &poly[1];
	      LegendrePolynomialMult (order-2, 2*yr-1, yr, hp1);
	      poly[0] = 1;
	      
	      double hv = 1.0;
	      for (int i = 0; i <= order; i++)
		{
		  powy[i] = hv;
		  hv *= (1-yr1);
		}
	    
              for (int jj = 0; jj < ndtrig; jj++)
		fac_y[iy,jj] = poly[trig2tensor[jj][0]] * powy[trig2tensor[jj][1]+trig2tensor[jj][2]];
	    }
	  
          FlatMatrix<> fac_x(ir1dx.GetNIP(), ndof, lh);
	  
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
              double xr = ir1dx[i1](0);
              double * hp1 = &polx[1];
              LegendrePolynomialMult (order-2, 2*xr-1, xr, hp1);
              polx[0] = 1;
	      
              double hv = 1.0;
              for (int i = 0; i <= order; i++)
                {
                  powx[i] = hv;
                  hv *= (1-xr1);
                }
	      
              for (int jj = 0; jj < ndof; jj++)
                {
                  fac_x = polx[tet2tensor[jj][0]] *
                    powx[tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3]];
		}
            }
	  
          NgProfiler::StopTimer (timeri);
	  
          NgProfiler::StartTimer (timer1);
	  

	  // Compute tdvec for all tp_points (0<= ii <= nip) tdvec[ii] = coeff(i1,i2,i3) 
	  double hex_p[3], tet_p[3];
          int ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
	      hex_p[0] = ir1dx[i1](0);
              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
		  hex_p[1] = ir1dy[i2](0);
                  double duffy_jac_det = duffy_trafo.GetJacobiDet(hex_p);
		  duffy_trafo.SetEtaForJacobianInverse(hex_p,scaled); 
		  
                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
		      hex_p[2] = ir1dz[i3](0);
		      duffy_trafo.GetIPTet(hex_p,tet_p);
                      IntegrationPoint ip(tet_p[0], tet_p[1], tet_p[2], 
                                          ir1dx[i1].Weight()*ir1dy[i2].Weight()*ir1dz[i3].Weight());
		      		      
		      //SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ir[i], eltrans, locheap);
		      SpecificIntegrationPoint<3,3> sip(ir[i], eltrans, locheap);
		      
		      Vec<1> dvec; 
		      dvecop.GenerateVector (fel, sip, dvec, lh);
		      dvec * = fabs (sip.GetJacobiDet()) * sip.IP().Weight() * duffy_jac_det;
                      tdvec[ii] = dvec; 
		    }
                }
            }

          NgProfiler::StopTimer (timer1);


	  ii = 0; // runs through number of integration points !!! tp_points !! 
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
              sum2d = 0.0; // sum2d = 

              NgProfiler::StartTimer (timertrans);

              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
		  sum1d = 0.0; // sum1d = sum_i3 pz_kl(i3) * tdmat[ii] = sum_i3 pz_kl(i3) * tdmat(i1,i2,i3)

		  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
		    for (int j = 0; j < ndsegm; j++)
		      sum1d(j) += tdvec[ii](0) * shapesz(i3,j);
				  
		  for (int jj = 0; jj < ndtrig; jj++)
		    {
		      int j1d = trig2segm[jj];
		      
		      double fy = fac_y(i2,jj);
		      
		      for (int k = 0; k < ndsegm; k++)
			sum2d(k,jj) += fy * sum1d(j1d, k);
		    }
		}

	      NgProfiler::StopTimer (timertrans);
	      
	      
              NgProfiler::StartTimer (timer2a);
	      
              // mult j
              for (int jj = 0; jj < ndof; jj++)
                {
                  int jtrig = tet2trig[jj];
                  double fx = fac_x(i1,jj);
                  for (int k = 0; k < ndtrig; k++)
                    elvec(jj) += sum2d(jtrig, k) * fx;
                }

              NgProfiler::StopTimer (timer2a);
            }
	  lh.CleanUp (heapp);

      }

};

#endif
#ifndef MASSTPSZ

  template <int DIM>
  class MassIntegratorTPSZ : public BilinearFormIntegrator
  {
    CoefficientFunction * coef;
    BilinearFormIntegrator * orig;

    typedef DiffOpId<DIM> DIFFOP;
    enum { DIM_SPACE   = DIFFOP::DIM_SPACE };
    enum { DIM_ELEMENT = DIFFOP::DIM_ELEMENT };
    enum { DIM_DMAT    = DIFFOP::DIM_DMAT };
   
    DiagDMat<DIM> dmatop;
    
    // typedef H1HighOrderTetTPSZ FEL;
  public:
    MassIntegratorTPSZ (CoefficientFunction * acoef)
      : BilinearFormIntegrator (), coef(acoef), dmatop (acoef)
    { 
      orig = new LaplaceIntegrator<DIM> (acoef); 
      fast = 1;
    }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new MassIntegratorTPSZ (coeffs[0]);
    }

    virtual bool BoundaryForm () const { return 0; }
    

    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {

      typedef H1HighOrderTetTPSZ FEL;
      bool scaled=1; 
     
      if (dynamic_cast<const FEL*> (&bfel))
        {
          const FEL & fel = static_cast<const FEL&> (bfel);
          static int timer = NgProfiler::CreateTimer ("Fast assembling");
          static int timer1 = NgProfiler::CreateTimer ("Fast assembling - geometry");
          static int timer2a = NgProfiler::CreateTimer ("Fast assembling - transform x-j");
          static int timer2b = NgProfiler::CreateTimer ("Fast assembling - transform x-k");
          static int timer3 = NgProfiler::CreateTimer ("Fast assembling - transform y");
          static int timer4 = NgProfiler::CreateTimer ("Fast assembling - transform z");
          static int timeri = NgProfiler::CreateTimer ("Fast assembling - init");
          static int timertrans = NgProfiler::CreateTimer ("Fast assembling - transform y,z");
          
          NgProfiler::RegionTimer reg (timer);
          NgProfiler::StartTimer (timeri);
          int ndof = fel.GetNDof();
      
          elmat.AssignMemory (ndof, ndof, lh);
          elmat = 0;


          int order = fel.Order();
          // const IntegrationRule & ir1dx = SelectIntegrationRule (ET_SEGM, 2*order);
          const IntegrationRule & ir1dy = SelectIntegrationRule (ET_SEGM, 2*order-1);
          const IntegrationRule & ir1dz = SelectIntegrationRule (ET_SEGM, 2*order-2);
      
          static IntegrationRule ir1dx;
          if (ir1dx.GetNIP() == 0)
            {
              int np = order;
              ARRAY<double> xi(np), wi(np);
              ComputeGaussJacobiRule (np, xi, wi, 2, 0);
              for (int i = 0; i < np; i++)
                ir1dx.AddIntegrationPoint (new IntegrationPoint (xi[i], 0, 0, wi[i]));
            }

          ArrayMem<AutoDiff<1>, 20> polx(order+1), powx(order+1);
          ArrayMem<AutoDiff<1>, 20> poly(order+1), powy(order+1);
          ArrayMem<AutoDiff<1>, 20> polz(order+1), polzz(2);

	  DuffyTransformationTet duffy_trafo(fel.GetVNums()); 

          void * heapp = lh.GetPointer();

          FlatArray<int[4]> tet2tensor (ndof, lh);
	  fel.GetTet2TensorMapping(tet2tensor);  
	 
	  int nd1d = 2 * (order);
          int nd2d = order * nd1d; 
	  FlatArray<int> tensor2trig(nd2d, lh); 
	  ARRAY<int[3] > trig2tensor; 
         	  
	  int ndtrig = fel.GetTrig2TensorMappings(tet2tensor, trig2tensor, tensor2trig); 
	  
	  FlatArray<int> tet2trig (ndof, lh);
          for (int i = 0; i < ndof; i++)
            tet2trig[i] = tensor2trig[tet2tensor[i][1] * nd1d + tet2tensor[i][2] * 2 + tet2tensor[i][3]];
	  
          FlatArray<int> tensor2segm ( nd1d, lh);
	  int ndsegm = fel.GetTensor2SegmMapping(tet2tensor, tensor2segm); 
	  
	  FlatArray<int> trig2segm (ndtrig, lh);
          for (int i = 0; i < ndtrig; i++)
            trig2segm[i] = tensor2segm[trig2tensor[i][1] * 2 + trig2tensor[i][2]];
                
          FlatMatrix<Mat<2> > sum2d(ndtrig, lh);
          FlatMatrix<Mat<3> > sum1d(ndsegm, lh);

          FlatVector<Mat<3> > mult1d(ndsegm, lh);
          FlatMatrix<Vec<6> > mult2d(ndsegm, ndtrig, lh);
          FlatMatrix<Vec<2> > mult3d(ndtrig, ndof, lh);

          FlatMatrix<> shapesz (ir1dz.GetNIP(), ndsegm, lh);
          FlatMatrix<> shapesdz (ir1dz.GetNIP(), ndsegm, lh);
          FlatMatrix<> shapeszT (ndsegm, ir1dz.GetNIP(), lh);
          FlatMatrix<> shapesdzT (ndsegm, ir1dz.GetNIP(), lh);

          for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++)
            {
              AutoDiff<1> zr1(ir1dz[i3](0), 0);
              AutoDiff<1> * hp1 = &polz[1];
              LegendrePolynomialMult (order-2, 2*zr1-1, zr1, hp1);
              polz[0] = 1;
              polz[order] = 0;

              polzz[0] = 1;
              polzz[1] = (1-zr1); 

              for (int j3 = 0; j3 < order; j3++)
                for (int j4 = 0; j4 < 2; j4++)
                  {
                    int j1d = tensor2segm[j3 * 2 + j4];
                    if (j1d == -1) continue;

                    AutoDiff<1> shapejj1 = polz[j3] * polzz[j4];

                    shapesz(i3, j1d) = shapejj1.Value();
                    shapesdz(i3, j1d) = shapejj1.DValue(0);
                  }          
            }        
          shapeszT = Trans (shapesz);
          shapesdzT = Trans (shapesdz);

          int nip = ir1dx.GetNIP() * ir1dy.GetNIP() * ir1dz.GetNIP();
      
          FlatVector<Mat<3> > tdmat(nip, lh);
          FlatMatrix<Vec<2> > fac_ydy(ir1dy.GetNIP(), ndtrig, lh);
          FlatMatrix<> fac_y_s(ir1dy.GetNIP(), ndtrig, lh);
          
	  for (int iy = 0; iy < ir1dy.GetNIP(); iy++)
            {
              AutoDiff<1> yr1 (ir1dy[iy](0), 0);
              AutoDiff<1> * hp1 = &poly[1];
              LegendrePolynomialMult (order-2, 2*yr1-1, yr1, hp1);
              poly[0] = 1;
              
              AutoDiff<1> hv = 1.0;
              for (int i = 0; i <= order; i++)
                {
                  powy[i] = hv;
                  hv *= (1-yr1);
                }
          
              for (int jj = 0; jj < ndtrig; jj++)
                {
		  AutoDiff<1> fy = poly[trig2tensor[jj][0]] * powy[trig2tensor[jj][1]+trig2tensor[jj][2]];
                  fac_ydy(iy,jj)(0) = fy.Value();
                  fac_ydy(iy,jj)(1) = fy.DValue(0);
		  fac_y_s(iy,jj) = fy.Value();
		}

	      if(scaled) 
		for (int jj = 0; jj < ndtrig; jj++)
		  {
		    int ind_s = trig2tensor[jj][1]+trig2tensor[jj][2];
		    if(ind_s > 0)
		      fac_y_s(iy,jj) = poly[trig2tensor[jj][0]].Value() * powy[ind_s-1].Value();
		    else 
		      fac_y_s(iy,jj) = 1.; //poly[trig2tensor[jj][0]].Value() * powy[ind_s].Value() / (1-yr1.Value());
		  }
	    }




          FlatMatrix<Vec<2> > fac_xdx(ir1dx.GetNIP(), ndof, lh);

          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
              AutoDiff<1> xr1 (ir1dx[i1](0), 0);
              AutoDiff<1> * hp1 = &polx[1];
              LegendrePolynomialMult (order-2, 2*xr1-1, xr1, hp1);
              polx[0] = 1;

              AutoDiff<1> hv = 1.0;
              for (int i = 0; i <= order; i++)
                {
                  powx[i] = hv;
                  hv *= (1-xr1);
                }


              for (int jj = 0; jj < ndof; jj++)
                {
                  AutoDiff<1> fx = 
                    polx[tet2tensor[jj][0]] *
                    powx[tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3]];
		  	      
		  if(!scaled)
		    fac_xdx(i1, jj)(0) = fx.Value();
		  else 
		    {
		      int ind_s = tet2tensor[jj][1]+tet2tensor[jj][2]+tet2tensor[jj][3];
		      if(ind_s>0) 
			fac_xdx(i1,jj)(0) = polx[tet2tensor[jj][0]].Value() * powx[ind_s-1].Value();
		      else 
			fac_xdx(i1, jj)(0) = 1.; //polx[tet2tensor[jj][0]].Value() * powx[ind_s].Value()/(1-xr1.Value());
		    }
                  fac_xdx(i1, jj)(1) = fx.DValue(0);
                }
            }


          NgProfiler::StopTimer (timeri);

          NgProfiler::StartTimer (timer1);

	  double hex_p[3], tet_p[3];
          int ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
	      hex_p[0] = ir1dx[i1](0);

              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
		  hex_p[1] = ir1dy[i2](0);
                  double duffy_jac_det = duffy_trafo.GetJacobiDet(hex_p);
		  duffy_trafo.SetEtaForJacobianInverse(hex_p,scaled); 
		  
                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
		      hex_p[2] = ir1dz[i3](0);

		      duffy_trafo.GetIPTet(hex_p,tet_p);
                      IntegrationPoint ip(tet_p[0], tet_p[1], tet_p[2], 
                                          ir1dx[i1].Weight()*ir1dy[i2].Weight()*ir1dz[i3].Weight());

                      SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ip, eltrans, lh);

                      Mat<3> dmat, hmat, trans; 
                      dmatop.GenerateMatrix (fel, sip, dmat, lh);
                      dmat *= fabs (sip.GetJacobiDet()) * sip.IP().Weight() * duffy_jac_det;

		      trans = duffy_trafo.GetJacobianInverseZ(hex_p,scaled) * sip.GetJacobianInverse(); 
		      //trans = duffy_trafo.GetJacobianInverse(hex_p) * sip.GetJacobianInverse(); 
		     
		      hmat = trans * dmat;
                      tdmat[ii] = hmat * Trans (trans); 
		    }
                }
            }


          NgProfiler::StopTimer (timer1);


          ii = 0;
          for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
            {
              sum2d = 0.0;

              NgProfiler::StartTimer (timertrans);

              for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                {
                  // NgProfiler::StartTimer (timer4);

                  sum1d = 0.0;

                  // NgProfiler::AddFlops (timer4, ir1dz.GetNIP() * ndsegm * (ndsegm+1) * 6);

                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
                      for (int j = 0; j < ndsegm; j++)
                        for (int l = 0; l < 3; l++)
                          {
                            mult1d(j)(0,l) = tdmat[ii](0,l) * shapesz(i3,j);
                            mult1d(j)(1,l) = tdmat[ii](1,l) * shapesz(i3,j);
                            mult1d(j)(2,l) = tdmat[ii](2,l) * shapesdz(i3,j);
                          }

                      for (int j = 0; j < ndsegm; j++)
                        for (int k = 0; k <= j; k++)
                          for (int l = 0; l < 3; l++)
                            {
                              sum1d(j,k)(l,0) += mult1d(j)(l,0) * shapesz (i3,k);   
                              sum1d(j,k)(l,1) += mult1d(j)(l,1) * shapesz (i3,k);   
                              sum1d(j,k)(l,2) += mult1d(j)(l,2) * shapesdz(i3,k);  
                            }
                    }
                  for (int j = 0; j < ndsegm; j++)
                    for (int k = 0; k < j; k++)
                      sum1d(k,j) = Trans(sum1d(j,k));



                  // NgProfiler::StopTimer (timer4);

                  // NgProfiler::RegionTimer reg3 (timer3);
                  // NgProfiler::AddFlops (timer3, ndtrig * (5 * ndtrig + 7 * ndsegm));

                  // mult pj:
                  for (int jj = 0; jj < ndtrig; jj++)
                    {
                      int j1d = trig2segm[jj];

                      double fy = fac_ydy(i2,jj)(0);
                      double fdy = fac_ydy(i2,jj)(1);
		      double fy_s = (scaled) ? fac_y_s(i2,jj): fac_ydy(i2,jj)(0); 

                      for (int k = 0; k < ndsegm; k++)
                        {
                          /*
                            const Mat<3> & hsum1d = sum1d(j1d, k);
                            Vec<6> & hmult2d = mult2d(k,jj);

                            hmult2d(0) = fy * hsum1d(0,0);
                            hmult2d(1) = fy * hsum1d(0,1);
                            hmult2d(2) = fy * hsum1d(0,2);
                            hmult2d(3) = fy * hsum1d(2,0) + fdy * hsum1d(1,0);
                            hmult2d(4) = fy * hsum1d(2,1) + fdy * hsum1d(1,1);
                            hmult2d(5) = fy * hsum1d(2,2) + fdy * hsum1d(1,2);
                          */

                          mult2d(k,jj)(0) = fy * sum1d(j1d, k)(0,0);
                          mult2d(k,jj)(1) = fy * sum1d(j1d, k)(0,1);
                          mult2d(k,jj)(2) = fy * sum1d(j1d, k)(0,2);
                          mult2d(k,jj)(3) = fy_s * sum1d(j1d, k)(2,0) + fdy * sum1d(j1d,k)(1,0);
                          mult2d(k,jj)(4) = fy_s * sum1d(j1d, k)(2,1) + fdy * sum1d(j1d,k)(1,1);
                          mult2d(k,jj)(5) = fy_s * sum1d(j1d, k)(2,2) + fdy * sum1d(j1d,k)(1,2);
                        }
                    }

                  // mult pk:
                  for (int kk = 0; kk < ndtrig; kk++)
                    {
                      int k1d = trig2segm[kk];

                      double fy = fac_ydy(i2,kk)(0);
                      double fdy = fac_ydy(i2,kk)(1);
		      double fy_s = (scaled) ? fac_y_s(i2,kk): fac_ydy(i2,kk)(0); 
                  
                      for (int jj = 0; jj <= ndtrig; jj++)
                        {
                          Mat<2> & hsum2d = sum2d(kk,jj);
                          const Vec<6> & hmult2d = mult2d(k1d,jj);

                          hsum2d(0,0) += fy_s * hmult2d(5) + fdy * hmult2d(4);
                          hsum2d(1,0) += fy_s * hmult2d(2) + fdy * hmult2d(1);
                          hsum2d(0,1) += fy * hmult2d(3);
                          hsum2d(1,1) += fy * hmult2d(0);

                          /*
                            sum2d(kk,jj)(0,0) += fy * mult2d(k1d,jj)(5) + fdy * mult2d(k1d,jj)(4);
                            sum2d(kk,jj)(1,0) += fy * mult2d(k1d,jj)(2) + fdy * mult2d(k1d,jj)(1);
                            sum2d(kk,jj)(0,1) += fy * mult2d(k1d,jj)(3);
                            sum2d(kk,jj)(1,1) += fy * mult2d(k1d,jj)(0);
                          */
                        }
                    }
                }

              for (int j = 0; j < ndtrig; j++)
                for (int k = 0; k < j; k++)
                  sum2d(j,k) = Trans(sum2d(k,j));

              NgProfiler::StopTimer (timertrans);


              NgProfiler::StartTimer (timer2a);
              NgProfiler::AddFlops (timer2a, ndof * 4 * ndtrig);

              // mult j
              for (int jj = 0; jj < ndof; jj++)
                {
                  int jtrig = tet2trig[jj];
                  Vec<2> fxfdx = fac_xdx(i1,jj);
                  for (int k = 0; k < ndtrig; k++)
                    mult3d(k,jj) = sum2d(jtrig, k) * fxfdx;
                }

              NgProfiler::StopTimer (timer2a);

              NgProfiler::StartTimer (timer2b);
              NgProfiler::AddFlops (timer2b, ndof * 2 * ndof );

              // multk
              for (int kk = 0; kk < ndof; kk++)
                {
                  int ktrig = tet2trig[kk];
                  Vec<2> fxfdx = fac_xdx(i1,kk);
                  for (int jj = 0; jj <= kk; jj++)
                    elmat(kk*ndof+jj) += InnerProduct (fxfdx,  mult3d(ktrig*ndof+jj));
                }

              NgProfiler::StopTimer (timer2b);
            }
      

          for (int j = 0; j < ndof; j++)
            for (int k = 0; k < j; k++)
              elmat(k,j) = elmat(j,k);

          lh.CleanUp (heapp);

        }
    }
  };
#endif
  
  
  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };        
    
    Init::Init()
    {
      GetIntegrators().AddBFIntegrator ("laplacetpsz", 3, 1,
                                        LaplaceIntegratorTPSZ<3>::Create);
    } 
    
    Init init;
    
  }





#endif

}
