#ifndef FILE_H1HOFETPSZ
#define FILE_H1HOFETPSZ

/*********************************************************************/
/* File:   h1hofetp.hpp                                              */
/* Author: Start                                                     */
/* Date:   11 June 2007                                              */
/*********************************************************************/

/**
  High order finite elements for H^1  in full Tensor Product basis
*/

class H1HighOrderTrigTPSZ : public H1HighOrderTrig<>
{
public:

  H1HighOrderTrigTPSZ (int aorder)
    : H1HighOrderTrig<> (aorder) {;}

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename TFA>  
  void T_CalcShape (Tx x, Ty y, TFA & shape) const; 

  void GetTrig2TensorMapping(FlatArray<int[3]> trig2tensor) const; 

  int GetTensor2SegmMapping(FlatArray<int[3] > trig2tensor, FlatArray<int> tensor2segm) const; 

};



class H1HighOrderTetTPSZ : public H1HighOrderTet<>
{
  
public:

  // int sort[4]; 

  H1HighOrderTetTPSZ (int aorder)
    : H1HighOrderTet<> (aorder) {;}

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 

  void GetTet2TensorMapping (FlatArray<INT<4> > tet2tensor, FlatArray<int > tensor2tet, 
			     FlatArray<int > tensor2trig, int & ndtrig) const; 


  // virtual void SetVertexNumbers (FlatArray<int> & avnums); 
 
  void GetTet2TensorMapping(FlatArray<int[4]> tet2tensor) const; 
  int GetTrig2TensorMappings(FlatArray<int[4] >   tet2tensor, ARRAY<int[3]> & trig2tensor, FlatArray<int> tensor2trig) const; 
  int GetTensor2SegmMapping(FlatArray<int[4] > tet2tensor, FlatArray<int> tensor2segm) const; 

};




class H1HighOrderPrismTPSZ : public H1HighOrderPrism<>
{

public:
  H1HighOrderPrismTPSZ (int aorder)
    : H1HighOrderPrism<> (aorder) { ; }

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 

 
};



class H1HighOrderHexTPSZ : public H1HighOrderHex<>
{

public:
  H1HighOrderHexTPSZ (int aorder)
    : H1HighOrderHex<> (aorder) { ; }

  /// compute shape
  virtual void CalcShape (const IntegrationPoint & ip, 
			  FlatVector<> shape) const;
  
  /// compute gradient of shape
  virtual void CalcDShape (const IntegrationPoint & ip, 
			   FlatMatrix<> dshape) const;

  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void T_CalcShape (Tx x, Ty y, Tz z, TFA & shape) const; 


  void GetHex2TensorMapping (FlatArray<int[3]> hex2tensor, FlatArray<double> fac) const;
};











#endif



