/*********************************************************************/
/* File:   h1hofe.cpp                                                */
/* Author: Start                                                     */
/* Date:   1. June 2007                                              */
/*********************************************************************/

// #define NOPROFILE



#include <fem.hpp>
#include "h1hofetpsz_stage2.hpp" 
namespace ngfem 
{
  using namespace ngfem;
  
 

  // ************************** TRIG ************************************ 
  
  void H1HighOrderTrigTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                      FlatVector<> shape) const
  {
    T_CalcShape<double, double, FlatVector<> > (ip(0), ip(1), shape); 
  }
  
  void H1HighOrderTrigTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                       FlatMatrix<> dshape) const
  {
    AutoDiff<2> x(ip(0), 0);
    AutoDiff<2> y(ip(1), 1);

    ArrayMem<AutoDiff<2>,40> sds(ndof);
    T_CalcShape<AutoDiff<2>, AutoDiff<2>, FlatArray<AutoDiff<2> > >
      (x,y,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 2; j++)
	dshape(i, j) = sds[i].DValue(j);
    
    return; 
  }

  template<typename Tx, typename Ty, typename TFA>  
  void  H1HighOrderTrigTPSZ :: T_CalcShape (Tx x, Ty y, TFA & sds) const
  {
    Tx lami[3] = { x, y,  1-x-y };

    int sort[] = { 0, 1, 2 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] 
    
    Tx lamis[3];
    for (int i = 0; i < 3; i++)
      lamis[i] = lami[sort[i]];


#ifdef JACOBI

    ArrayMem<Tx, 100> memx((order+1)*order);
    ArrayMem<Tx, 100> memy(2*order);
    ArrayMem<Tx, 20> memyy(2);

    FlatMatrix<Tx> polsx(order+1, order, &memx[0]);
    FlatMatrix<Tx> polsy(2, order, &memy[0]);
    FlatMatrix<Tx> polsyy(1, 2, &memyy[0]);

    Tx * hp;
    
    for (int i = 0; i <= order; i++)
      {
        hp = &polsx(i, 1);
        JacobiPolynomial (order-2, 2*lamis[0]-1, max(0,2*i-2), 0, hp);
        for (int j = 1; j < order; j++)
          polsx(i,j) *= lamis[0];
        polsx(i,0) = 1.0;
      }

    for (int i = 0; i < 2; i++)
      {
        hp = &polsy(i,1);
        ScaledLegendrePolynomial (order-2, lamis[1]-lamis[2], lamis[1]+lamis[2], hp);
        for (int j = 1; j < order; j++)
          polsy(i,j) *= lamis[1];
        polsy(i,0) = 1.0;
      }

    polsyy(0,0) = 1.0;
    polsyy(0,1) = lamis[2];

    ArrayMem<int[3], 100> trig2tensor(ndof);
    GetTrig2TensorMapping (trig2tensor);

    for (int i = 0; i < ndof; i++)
      sds[i] =
        polsx(trig2tensor[i][1]+trig2tensor[i][2], trig2tensor[i][0]) * 
        polsy(0, trig2tensor[i][1]) *
        polsyy(0, trig2tensor[i][2]);

#else

    ArrayMem<Tx, 20> polx(order+1), poly(order+1), polyy(2); 

    Tx * hp = &polx[1];
    LegendrePolynomialMult (order-2, 2*lamis[0]-1, lamis[0], hp);
    polx[0] = 1.0;
    
    hp = &poly[1];
    ScaledLegendrePolynomialMult (order-2, lamis[1]-lamis[2], lamis[1] + lamis[2], lamis[1], hp);
    poly[0] = 1.0;

    polyy[0] = 1.0;
    polyy[1] = lamis[2];

    ArrayMem<int[3], 100> trig2tensor(ndof);
    GetTrig2TensorMapping (trig2tensor);

    for (int i = 0; i < ndof; i++)
      sds[i] =
        polx[trig2tensor[i][0]] * 
        poly[trig2tensor[i][1]] * 
        polyy[trig2tensor[i][2]];
#endif
  }


  void H1HighOrderTrigTPSZ :: GetTrig2TensorMapping(FlatArray<int[3]> trig2tensor) const 
  {
    // maps trig_shapes (0<=ii<=ndof) to the accoring tensor-product shapes (i,j,l) with i,j<=order, l<=2
    // (injective map) 

    int sort[] = { 0, 1, 2 };
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] 

    int isort[3];
    for (int i = 0; i < 3; i++) isort[sort[i]] = i;
    
    for (int i = 0; i < ndof; i++)
      {
	trig2tensor[i][0] = trig2tensor[i][1] =
	  trig2tensor[i][2] = 0;
      }
    
    // vertex shapes
    for (int i = 0; i < 3; i++)
      trig2tensor[i][isort[i]] = 1;
    
    int ii = 3;
    
    //  edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TRIG);
    for (int i = 0; i < 3; i++)
      if (order_edge[i] >= 2)
	{
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  int nde = order_edge[i]-1;
	  for (int j = 1; j <= nde; j++, ii++)
	    {
	      trig2tensor[ii][isort[es]] = j; // lami_s * ''pol(lami_s-lami_e,lami_s+lami_e)'' 
	      trig2tensor[ii][isort[ee]] = 1; // lami_e
	    }
	}
    
    // face dofs
    if (order_inner[0] >= 3)
      {
	int fav[3] = { 0, 1, 2 };
	
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
          
	int pi = order_inner[0];
	for (int j2 = 0; j2 <= pi-3; j2++)
	  for (int i2 = 0; i2 <= pi-3-j2; i2++, ii++)
	    {
	      trig2tensor[ii][isort[fav[0]]] = j2+1;
	      trig2tensor[ii][isort[fav[1]]] = i2+1;
	      trig2tensor[ii][isort[fav[2]]] = 1;
	    }
      }

  } 

  int H1HighOrderTrigTPSZ :: 
  GetTensor2SegmMapping(FlatArray<int[3] > trig2tensor, FlatArray<int> tensor2segm) const
  {
    // size (nd1d = 2 * order) 
    // mapping of 1d-tp points to either -1 (no segm point in trig) or to segm_dof_nr    
    
    tensor2segm = -1;
    
    for (int ii = 0; ii < ndof; ii++)
      {
	int j1d = trig2tensor[ii][1] * 2 + trig2tensor[ii][2];
	tensor2segm[j1d] = 0;
      }
    
    int ndsegm =0;
    int nd1d = 2*order;
    for (int i = 0; i < nd1d; i++)
      if (tensor2segm[i] != -1)
	{
	  tensor2segm[i] = ndsegm;
	  ndsegm++;
	}
    return(ndsegm); 
  } 









  // ******************************** TET **************************************

  void H1HighOrderTetTPSZ :: SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh)
  {
    H1HighOrderTet<> :: SetVertexNumbers (avnums, lh);

    for (int i = 0; i < 4; i++) sort[i] = i;

    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[2]] > vnums[sort[3]]) Swap (sort[2], sort[3]);
    if (vnums[sort[0]] > vnums[sort[2]]) Swap (sort[0], sort[2]);
    if (vnums[sort[1]] > vnums[sort[3]]) Swap (sort[1], sort[3]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);

    BuildTables(lh);

  } 

  void H1HighOrderTetTPSZ :: BuildTables(LocalHeap & lh)
  {
    // vnums[sort[0]] < vnums[sort[1]] < vnums[sort[2]] < vnums[sort[3]]
 
    tet2tensor = FlatArray<int[4]> (ndof, lh);

    int isort[4];
    for (int i = 0; i < 4; i++)
      isort[sort[i]] = i;

    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 4; j++)
        tet2tensor[i][j] = 0;

    // vertex shapes
    for (int i = 0; i < 4; i++)
      tet2tensor[i][isort[i]] = 1;
    
    int ii = 4;
    
    //  edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
        {
          int es = edges[i][0], ee = edges[i][1];
          if (vnums[es] > vnums[ee]) swap (es, ee);
          int nde = order_edge[i]-1;
          for (int j = 1; j <= nde; j++, ii++)
            {
              tet2tensor[ii][isort[es]] = j;
              tet2tensor[ii][isort[ee]] = 1;
            }
        }
    
    // face dofs
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
      if (order_face[i][0] >= 3)
        {
          int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
          
          if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
          if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
          if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
          
          int n = order_face[i][0];
          for (int j2 = 0; j2 <= n-3; j2++)
            for (int i2 = 0; i2 <= n-3-j2; i2++, ii++)
              {
                tet2tensor[ii][isort[fav[0]]] = j2+1;
                tet2tensor[ii][isort[fav[1]]] = i2+1;
                tet2tensor[ii][isort[fav[2]]] = 1;
              }
        }
    
    // inner dofs
    if (order_inner[0] >= 4)
      {
        int n = order_inner[0];
        for (int i = 0; i <= n-4; i++)
          for (int j = 0; j <= n-4-i; j++)
            for (int k = 0; k <= n-4-i-j; k++, ii++)
              {
                tet2tensor[ii][0] = k+1;
                tet2tensor[ii][1] = j+1;
                tet2tensor[ii][2] = i+1;
                tet2tensor[ii][3] = 1;
              }
      }

    
    int nd1d = 2 * (order);
    int nd2d = order * nd1d;


    FlatArray<int> tensor2trig ( nd2d, lh);
    tensor2trig = -1;

    for (int ii = 0; ii < ndof; ii++)
      {
        int j2d = tet2tensor[ii][1] * nd1d + tet2tensor[ii][2] * 2 + tet2tensor[ii][3];
        tensor2trig[j2d] = 0;
      }
      
    int ndtrig =0;
    for (int i = 0; i < nd2d; i++)
      if (tensor2trig[i] != -1)
        {
          tensor2trig[i] = ndtrig;
          ndtrig++;
        }
    
    
    // ArrayMem<int[3], 100> trig2tensor(ndtrig);
    trig2tensor = FlatArray<int[3]> (ndtrig, lh);

    for (int i2 = 0; i2 < order; i2++)
      for (int i3 = 0; i3 < order; i3++)
        for (int i4 = 0; i4 < 2; i4++)
          {
            int i2d = i2 * nd1d + i3 * 2 + i4;
            if (tensor2trig[i2d] != -1)
              {
                trig2tensor[tensor2trig[i2d]][0] = i2;
                trig2tensor[tensor2trig[i2d]][1] = i3;
                trig2tensor[tensor2trig[i2d]][2] = i4;
              }
          }

    FlatArray<int> tet2trig (ndof, lh);
    for (int i = 0; i < ndof; i++)
      tet2trig[i] = tensor2trig[tet2tensor[i][1] * nd1d + tet2tensor[i][2] * 2 + tet2tensor[i][3]];


    FlatArray<int> tensor2segm ( nd1d, lh);
    tensor2segm = -1;
    
    for (int ii = 0; ii < ndof; ii++)
      {
        int j1d = tet2tensor[ii][2] * 2 + tet2tensor[ii][3];
        tensor2segm[j1d] = 0;
      }
    
    int ndsegm =0;
    for (int i = 0; i < nd1d; i++)
      if (tensor2segm[i] != -1)
        {
          tensor2segm[i] = ndsegm;
          ndsegm++;
        }
    
    FlatArray<int> trig2segm (ndtrig, lh);
    for (int i = 0; i < ndtrig; i++)
      trig2segm[i] = tensor2segm[trig2tensor[i][1] * 2 + trig2tensor[i][2]];


    segm2tensor = FlatArray<int[2]> (ndsegm, lh);

    for (int i3 = 0; i3 < order; i3++)
      for (int i4 = 0; i4 < 2; i4++)
        {
          int i1d = i3 * 2 + i4;
          if (tensor2segm[i1d] != -1)
            {
              segm2tensor[tensor2segm[i1d]][0] = i3;
              segm2tensor[tensor2segm[i1d]][1] = i4;
            }
        }    
    
    map2dto1d = trig2segm;
    map3dto2d = tet2trig;
  }
    

  void H1HighOrderTetTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                      FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderTetTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                       FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  }




  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderTetTPSZ :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx lami[4] = { x, y, z, 1-x-y-z };

    Tx lamis[4];
    for (int i = 0; i < 4; i++)
      lamis[i] = lami[sort[i]];


#ifdef JACOBI

    ArrayMem<Tx, 100> memx((order+1)*order);
    ArrayMem<Tx, 100> memy((order+1)*order);
    ArrayMem<Tx, 20> memz(2*order);
    ArrayMem<Tx, 20> memzz(2);

    FlatMatrix<Tx> polsx(order+1, order, &memx[0]);
    FlatMatrix<Tx> polsy(order+1, order, &memy[0]);
    FlatMatrix<Tx> polsz(2, order, &memz[0]);
    FlatMatrix<Tx> polszz(1, 2, &memzz[0]);

    Tx * hp;
    
    for (int i = 0; i <= order; i++)
      {
        hp = &polsx(i, 1);
        JacobiPolynomial (order-2, 2*lamis[0]-1, max(0,2*i-2), 0, hp);
        for (int j = 1; j < order; j++)
          polsx(i,j) *= lamis[0];
        polsx(i,0) = 1.0;
      }

    for (int i = 0; i <= order; i++)
      {
        int alpha = max(0, 2*i-2);
        hp = &polsy(i,1);
        ScaledJacobiPolynomial (order-2, lamis[1]-lamis[2]-lamis[3], 1-lamis[0], max(0,2*i-2), 0, hp);
        for (int j = 1; j < order; j++)
          polsy(i,j) *= lamis[1];
        polsy(i, 0) = 1.0;
      }

    for (int i = 0; i < 2; i++)
      {
        hp = &polsz(i,1);
        ScaledLegendrePolynomial (order-2, lamis[2]-lamis[3], lamis[2]+lamis[3], hp);
        for (int j = 1; j < order; j++)
          polsz(i,j) *= lamis[2];
        polsz(i,0) = 1.0;
      }

    polszz(0,0) = 1.0;
    polszz(0,1) = lamis[3];

    /*
    ArrayMem<int[4], 100> tet2tensor(ndof);
    GetTet2TensorMapping (tet2tensor);
*/
    FlatArray<int[4]> tet2tensor = GetTet2TensorMapping ();

    for (int i = 0; i < ndof; i++)
      sds[i] =
        polsx(tet2tensor[i][1]+tet2tensor[i][2]+tet2tensor[i][3], tet2tensor[i][0]) * 
        polsy(tet2tensor[i][2]+tet2tensor[i][3], tet2tensor[i][1]) * 
        polsz(0, tet2tensor[i][2]) *
        polszz(0, tet2tensor[i][3]);


#else

    ArrayMem<Tx, 20> polx(order+1), poly(order+1), polz(order+1), polzz(2);

    Tx * hp = &polx[1];
    LegendrePolynomialMult (order-2, 2*lamis[0]-1, lamis[0], hp);
    polx[0] = 1.0;

    hp = &poly[1];
    ScaledLegendrePolynomialMult (order-2, lamis[1]-lamis[2]-lamis[3], 1-lamis[0], lamis[1], hp);
    poly[0] = 1.0;

    hp = &polz[1];
    ScaledLegendrePolynomialMult (order-2, lamis[2]-lamis[3], lamis[2]+lamis[3], lamis[2], hp);
    polz[0] = 1.0;

    polzz[0] = 1.0;
    polzz[1] = lamis[3];

    /*
    ArrayMem<int[4], 100> tet2tensor(ndof);
    GetTet2TensorMapping (tet2tensor);
    */

    FlatArray<int[4]> tet2tensor = GetTet2TensorMapping ();

    for (int i = 0; i < ndof; i++)
      sds[i] =
        polx[tet2tensor[i][0]] * 
        poly[tet2tensor[i][1]] * 
        polz[tet2tensor[i][2]] *
        polzz[tet2tensor[i][3]];


#endif
  }



  /*

  void H1HighOrderTetTPSZ :: GetTet2TensorMapping (FlatArray<int[4]> tet2tensor) const
  {
    int isort[4];
    for (int i = 0; i < 4; i++)
      isort[sort[i]] = i;

    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 4; j++)
        tet2tensor[i][j] = 0;

    // vertex shapes
    for (int i = 0; i < 4; i++)
      tet2tensor[i][isort[i]] = 1;
    
    int ii = 4;
    
    //  edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_TET);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
        {
          int es = edges[i][0], ee = edges[i][1];
          if (vnums[es] > vnums[ee]) swap (es, ee);
          int nde = order_edge[i]-1;
          for (int j = 1; j <= nde; j++, ii++)
            {
              tet2tensor[ii][isort[es]] = j;
              tet2tensor[ii][isort[ee]] = 1;
            }
        }
    
    // face dofs
    const FACE * faces = ElementTopology::GetFaces (ET_TET);
    for (int i = 0; i < 4; i++)
      if (order_face[i][0] >= 3)
        {
          int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
          
          if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
          if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
          if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
          
          int n = order_face[i][0];
          for (int j2 = 0; j2 <= n-3; j2++)
            for (int i2 = 0; i2 <= n-3-j2; i2++, ii++)
              {
                tet2tensor[ii][isort[fav[0]]] = j2+1;
                tet2tensor[ii][isort[fav[1]]] = i2+1;
                tet2tensor[ii][isort[fav[2]]] = 1;
              }
        }
    
    // inner dofs
    if (order_inner[0] >= 4)
      {
        int n = order_inner[0];
        for (int i = 0; i <= n-4; i++)
          for (int j = 0; j <= n-4-i; j++)
            for (int k = 0; k <= n-4-i-j; k++, ii++)
              {
                tet2tensor[ii][0] = k+1;
                tet2tensor[ii][1] = j+1;
                tet2tensor[ii][2] = i+1;
                tet2tensor[ii][3] = 1;
              }
      }
  }
  */


  // ******************************** PRISM **************************************


  void H1HighOrderPrismTPSZ :: SetVertexNumbers (FlatArray<int> & avnums, LocalHeap & lh)
  {
    H1HighOrderPrism<> :: SetVertexNumbers (avnums, lh);

    for (int i = 0; i < 6; i++) sort[i] = i;

    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);
    if (vnums[sort[1]] > vnums[sort[2]]) Swap (sort[1], sort[2]);
    if (vnums[sort[0]] > vnums[sort[1]]) Swap (sort[0], sort[1]);

    if (vnums[sort[3]] > vnums[sort[4]]) Swap (sort[3], sort[4]);
    if (vnums[sort[4]] > vnums[sort[5]]) Swap (sort[4], sort[5]);
    if (vnums[sort[3]] > vnums[sort[4]]) Swap (sort[3], sort[4]);

    BuildTables(lh);     
  }
    
  void H1HighOrderPrismTPSZ :: BuildTables(LocalHeap & lh)
  {
    prism2tensor = FlatArray<int[4]> (ndof,lh); 
    fac_pr2tp = FlatArray<double>(ndof,lh);  
    
    BuildPrism2TensorMapping ();

    int nd1d = order+1;      // zorder
    int nd2d = (order+1) * (order+1) * 2;  // poly, polyy, polz
    
    FlatArray<int > tensor2trig(nd2d, lh);
    tensor2trig = -1;
    
    for (int ii = 0; ii < ndof; ii++)
      {
	int j2d = prism2tensor[ii][1] * (2*nd1d) + prism2tensor[ii][2] * nd1d + prism2tensor[ii][3];
	tensor2trig[j2d] = 0;
      }
    
    int ndof2d =0;
    for (int i = 0; i < nd2d; i++)
      if (tensor2trig[i] != -1)
	{
	  tensor2trig[i] = ndof2d;
	  ndof2d++;
	}
    
    trig2tensor = FlatArray<int[3]> (ndof2d, lh);
    for (int i2 = 0; i2 <= order; i2++)
      for (int i3 = 0; i3 < 2; i3++)
	for (int i4 = 0; i4 <= order; i4++)
	  {
	    int i2d = i2 * 2*nd1d + i3 * nd1d + i4;
	    if (tensor2trig[i2d] != -1)
	      {
		trig2tensor[tensor2trig[i2d]][0] = i2;
		trig2tensor[tensor2trig[i2d]][1] = i3;
		trig2tensor[tensor2trig[i2d]][2] = i4;
	      }
	  }
    
    
    FlatArray<int> prism2trig (ndof, lh);
    for (int i = 0; i < ndof; i++)
      prism2trig[i] = tensor2trig[prism2tensor[i][1] * 2*nd1d + prism2tensor[i][2] * nd1d + prism2tensor[i][3]];
    
    
    FlatArray<int> tensor2segm ( nd1d, lh);
    for (int i = 0; i <= nd1d; i++)
      tensor2segm[i] = i;
    
    int ndof1d = order+1;
    
    FlatArray<int> trig2segm (ndof2d, lh);
    for (int i = 0; i < ndof2d; i++)
      trig2segm[i] = tensor2segm[trig2tensor[i][2]];
    
    
    map2dto1d = trig2segm;
    map3dto2d = prism2trig;
  }

  void H1HighOrderPrismTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                        FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderPrismTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                         FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  } 




  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderPrismTPSZ :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    if (sort[0]+3 == sort[3])  // same singular vertex at bottom and top
      {
        Tx lami2d[3] = { x, y, 1-x-y };

        Tx lami2ds[3];
        for (int i = 0; i < 3; i++)
          lami2ds[i] = lami2d[sort[i]];
        
        ArrayMem<Tx, 60> polx (order+1), poly(order+1), polyy(2);

        Tx * hp = &polx[1];
        LegendrePolynomialMult (order-2, 2*lami2ds[0]-1, lami2ds[0], hp);
        polx[0] = 1;
        
        hp = &poly[1];
        ScaledLegendrePolynomialMult (order-2, lami2ds[1]-lami2ds[2], 1-lami2ds[0], lami2ds[1], hp);
        poly[0] = 1;

        polyy[0] = 1;
        polyy[1] = lami2ds[2];


        ArrayMem<Tx,10> polz(order+1);
        polz[0] = 1-z;
        polz[1] = z;
        hp = &polz[2];
        LegendrePolynomialMult (order-2, 2*z-1, z*(1-z), hp);

	// ArrayMem<int[4], 100> prism2tensor(ndof);
        // ArrayMem<double, 100> fac(ndof);

        // GetPrism2TensorMapping (prism2tensor, fac);

        for (int i = 0; i < ndof; i++)
          sds[i] = fac_pr2tp[i] * 
            polx[prism2tensor[i][0]] * 
            poly[prism2tensor[i][1]] * 
            polyy[prism2tensor[i][2]] *
            polz[prism2tensor[i][3]];
        
        return;
      }



    typedef IntegratedLegendreMonomialExt T_ORTHOPOL;

    Tx lami[6] = { x, y, 1-x-y, x, y, 1-x-y };
    Tx muz[6]  = { 1-z, 1-z, 1-z, z, z, z };

    
    Tx lamisum[6] = { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 };
    for (int i = 0; i < 3; i++)
      for (int j = 0; j < 3; j++)
        {
          if (vnums[j] > vnums[i]) lamisum[i] += lami[j];
          if (vnums[j+3] > vnums[i+3]) lamisum[i+3] += lami[j+3];
        }

    sds = 0.0;

    // vertex shapes
    for (int i = 0; i < 6; i++)
      sds[i] = lami[i] * muz[i];

    int ii = 6;

    // horizontal edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_PRISM);
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
	{
	  int p = order_edge[i];
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];
          ScaledLegendrePolynomialMult (p-2, lami[es]-lamisum[es], lami[es]+lamisum[es], lami[es]*lami[ee]*muz[ee], hp);
	  ii += p-1;
	}
    
    // vertical edges
    for (int i = 6; i < 9; i++)
      if (order_edge[i] >= 2)
	{
	  int p = order_edge[i];
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);
	  
	  Tx * hp = &sds[ii];
          LegendrePolynomialMult (p-2, muz[es]-muz[ee], muz[es]*muz[ee]*lami[ee], hp);
	  ii += p-1;
	}
    

    ArrayMem<Tx,20> polx(order+1), poly(order+1), polz(order+1);

    const FACE * faces = ElementTopology::GetFaces (ET_PRISM);
    // trig face dofs

    for (int i = 0; i < 2; i++)
      if (order_face[i][0] >= 3)
	{
	  int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 
	  if(vnums[fav[1]] > vnums[fav[2]]) swap(fav[1],fav[2]);
	  if(vnums[fav[0]] > vnums[fav[1]]) swap(fav[0],fav[1]); 	
	  
          LegendrePolynomialMult (order-2, lami[fav[0]]-lamisum[fav[0]], lami[fav[0]], polx);
          ScaledLegendrePolynomialMult (order-2, lami[fav[1]]-lamisum[fav[1]], lamisum[fav[0]], lami[fav[1]], poly);

          Tx bub = lami[fav[2]] * muz[fav[2]];
          
          int n = order_face[i][0];
	  for (int j2 = 0; j2 <= n-3; j2++)
            {
              Tx hv = bub * polx[j2];
              for (int i2 = 0; i2 <= n-3-j2; i2++, ii++)
                sds[ii] = hv * poly[i2];
            }
	}

    // quad face dofs
    for (int i = 2; i < 5; i++)
      if (order_face[i][0] >= 2 && order_face[i][1] >= 2)
	{
	  INT<2> p = order_face[i];
	  int pp = int(max2(p[0], p[1]))+1;
	 
	  int fmax = 0;
	  for(int j = 1; j < 4; j++) 
	    if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;
	  
	  int fz = 3-fmax; 
	  int ftrig = fmax^1; 
	  
	  Tx xi = lami[faces[i][fmax]] - lami[faces[i][ftrig]]; 
	  Tx eta = 1-lami[faces[i][fmax]]-lami[faces[i][ftrig]]; 
	  Tx zeta = muz[faces[i][fmax]]-muz[faces[i][fz]]; 

          Tx l1 = lami[faces[i][fmax]];
          Tx l2 = lami[faces[i][ftrig]];
          ScaledLegendrePolynomialMult (pp-2, l2-l1, l1+l2, l1*l2, polx);
          LegendrePolynomialMult (pp-2, -zeta, 0.25 * (1-zeta*zeta), polz);

	  // global x-direction is towards second-largest vertex (JS)
	  if (vnums[faces[i][ftrig]] > vnums[faces[i][fz]])  
	    for (int k = 0; k < p[0]-1; k++)
	      for (int j = 0; j < p[1]-1; j++)
		sds[ii++] = polx[k] * polz[j];
	  else
	    for (int j = 0; j < p[0]-1; j++)
	      for (int k = 0; k < p[1]-1; k++)
		sds[ii++] = polx[k] * polz[j];
	}
    
    // volume dofs:
    if (order_inner[0] > 2 && order_inner[2] > 1)
      {
	ArrayMem<Tx,20> pol_trig((order_inner[0]-1)*(order_inner[0]-2)/2);
	int nf = T_TRIGFACESHAPES::Calc (order_inner[0], x-y, 1-x-y, pol_trig);

	T_ORTHOPOL:: Calc(order_inner[2], 2*z-1, polz);
	for (int i = 0; i < nf; i++)
	  for (int k = 0; k < order_inner[2]-1; k++)
	    sds[ii++] = pol_trig[i] * polz[k];
      }
  }

  
  void H1HighOrderPrismTPSZ :: BuildPrism2TensorMapping () 
  {
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 4; j++)
        prism2tensor[i][j] = 0;
    for (int i = 0; i < ndof; i++)
      fac_pr2tp[i] = 0;

    int isort[6];
    for (int i = 0; i < 6; i++) isort[sort[i]] = i;

    // vertex shapes
    for (int i = 0; i < 6; i++)
      {
        prism2tensor[i][isort[i%3]] = 1;
        prism2tensor[i][3] = i / 3;
        fac_pr2tp[i] = 1;
      }    

    int ii = 6;
    
    const EDGE * edges = ElementTopology::GetEdges (ET_PRISM);
    // horizontal edge dofs
    for (int i = 0; i < 6; i++)
      if (order_edge[i] >= 2)
        {
          int p = order_edge[i];
          int es = edges[i][0], ee = edges[i][1];
          if (vnums[es%3] > vnums[ee%3]) swap (es, ee);
          int mult = ( (vnums[es] > vnums[ee]) == (vnums[es%3] > vnums[ee%3])) ? 1 : -1;
            
          int fac = 1;
          int nde = order_edge[i]-1;
          for (int j = 1; j <= nde; j++, ii++)
            {
              prism2tensor[ii][isort[es%3]] = j;
              prism2tensor[ii][isort[ee%3]] = 1;
              prism2tensor[ii][3] = es / 3;
	      fac_pr2tp[ii] = fac;
              fac *= mult;
            }
        }
    
    // vertical edges
    for (int i = 6; i < 9; i++)
      if (order_edge[i] >= 2)
        {
          int p = order_edge[i];
          int es = edges[i][0], ee = edges[i][1];

          int mult = 1, fac = 1;
          if (vnums[es] < vnums[ee]) mult = -1;   
          
          int nde = order_edge[i]-1;
          for (int j = 1; j <= nde; j++, ii++)
            {
              prism2tensor[ii][isort[es%3]] = 1;
              prism2tensor[ii][3] = j+1;
              fac_pr2tp[ii] = fac;
              fac *= mult;
            }
        }

    
    const FACE * faces = ElementTopology::GetFaces (ET_PRISM);

    // trig faces
    for (int i = 0; i < 2; i++)
      if (order_face[i][0] >= 3)
        {
          int fav[3] = { faces[i][0], faces[i][1], faces[i][2] };

          if(vnums[fav[0]%3] > vnums[fav[1]%3]) swap(fav[0],fav[1]); 
          if(vnums[fav[1]%3] > vnums[fav[2]%3]) swap(fav[1],fav[2]);
          if(vnums[fav[0]%3] > vnums[fav[1]%3]) swap(fav[0],fav[1]); 	

          int mult = ( (vnums[fav[1]] > vnums[fav[2]]) == (vnums[fav[1]%3] > vnums[fav[2]%3])) ? 1 : -1;
          
          int n = order_face[i][0];

          for (int j2 = 0; j2 <= n-3; j2++)
            {
              int fac = 1;
              for (int i2 = 0; i2 <= n-3-j2; i2++, ii++)
                {
                  prism2tensor[ii][isort[fav[0]%3]] = j2+1;
                  prism2tensor[ii][isort[fav[1]%3]] = i2+1;
                  prism2tensor[ii][isort[fav[2]%3]] = 1;
                  prism2tensor[ii][3] = fav[0] / 3;                  
                  fac_pr2tp[ii] = fac;
                  fac *= mult;
                }
            }
        }


    

    // quad face dofs
    for (int i = 2; i < 5; i++)
      if (order_face[i][0] >= 2 && order_face[i][1] >= 2)
        {
          INT<2> p = order_face[i];
          int pp = int(max2(p[0], p[1]))+1;
          
          int fmax = 0;
          for(int j = 1; j < 4; j++) 
            if(vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;
          
          int fz = 3-fmax; 
          int ftrig = fmax^1; 

          fmax = faces[i][fmax];
          fz = faces[i][fz];
          ftrig = faces[i][ftrig];

          int multtrig = ( (vnums[fmax] > vnums[ftrig]) == (vnums[fmax%3] > vnums[ftrig%3])) ? 1 : -1;
          int multz = (fmax < 3) ? 1 : -1;

          int es = fmax%3, ee = ftrig%3;
          if (vnums[es] > vnums[ee]) swap (es, ee);

          // global x-direction is towards second-largest vertex (JS)
          if (vnums[ftrig] > vnums[fz])  
            {
              int factrig = 1;
              for (int k = 0; k < p[0]-1; k++)
                {
                  int facz = 1;
                  for (int j = 0; j < p[1]-1; j++, ii++)
                    {
                      prism2tensor[ii][isort[es]] = k+1;
                      prism2tensor[ii][isort[ee]] = 1;
                      prism2tensor[ii][3] = j+2;

		      fac_pr2tp[ii] = facz * factrig;
                      facz *= multz;
                    }
                  factrig *= multtrig;
                }

            }
          else
            {
              int facz = 1;
              for (int j = 0; j < p[0]-1; j++)
                {
                  int factrig = 1;
                  for (int k = 0; k < p[1]-1; k++, ii++)
                    {
                      prism2tensor[ii][isort[es]] = k+1;
                      prism2tensor[ii][isort[ee]] = 1;
                      prism2tensor[ii][3] = j+2;
		      fac_pr2tp[ii] = facz * factrig;
                      factrig *= multtrig;
                    }
                  facz *= multz;
                }
            }
        }

    // volume dofs:
    if (order_inner[0] > 2 && order_inner[2] > 1)
      {
        int pxy = order_inner[0];
        int pz = order_inner[2];

        for (int j = 0; j <= pxy-3; j++)
          for (int k = 0; k <= pxy-3-j; k++)
            for (int l = 0; l <= pz-2; l++, ii++)
              {
                prism2tensor[ii][0] = j+1;
                prism2tensor[ii][1] = k+1;
                prism2tensor[ii][2] = 1;
                prism2tensor[ii][3] = l+2;
		fac_pr2tp[ii] = 1;
              }
      }
  }






  // ******************************** HEX **************************************




  void H1HighOrderHexTPSZ :: CalcShape (const IntegrationPoint & ip, 
                                      FlatVector<> shape) const
  {
    T_CalcShape<double, double, double, FlatVector<> > (ip(0), ip(1), ip(2), shape); 
  }
  

  void H1HighOrderHexTPSZ :: CalcDShape (const IntegrationPoint & ip, 
                                       FlatMatrix<> dshape) const
  {
    AutoDiff<3> x(ip(0), 0);
    AutoDiff<3> y(ip(1), 1);
    AutoDiff<3> z(ip(2), 2);

    ArrayMem<AutoDiff<3>,40> sds(ndof);
    T_CalcShape<AutoDiff<3>, AutoDiff<3>, AutoDiff<3>, FlatArray<AutoDiff<3> > >
      (x,y,z,sds);
    for (int i = 0; i < ndof; i++)
      for (int j = 0; j < 3; j++)
	dshape(i, j) = sds[i].DValue(j);
  } 




  template<typename Tx, typename Ty, typename Tz, typename TFA>  
  void  H1HighOrderHexTPSZ :: T_CalcShape (Tx x, Ty y, Tz z, TFA & sds) const
  {
    Tx xi[3] = { 2*x-1, 2*y-1, 2*z-1 };
    
    ArrayMem<Tx, 80> polxi( 3 * (order+1) );
    Tx * poli[3];
    for (int i = 0; i < 3; i++) poli[i] = &polxi[i*(order+1)];

    for (int i = 0; i < 3; i++)
      {
        Tx * hp = poli[i]+2;
        poli[i][0] = 0.5 * (1-xi[i]);
        poli[i][1] = 0.5 * (1+xi[i]);
        LegendrePolynomialMult (order-2, xi[i], 0.25 * (1-xi[i]*xi[i]), hp);
      }

    ArrayMem<int[3], 100> hex2tensor(ndof);
    ArrayMem<double, 100> fac(ndof);

    GetHex2TensorMapping (hex2tensor, fac);

    for (int i = 0; i < ndof; i++)
      sds[i] = fac[i] * 
        poli[0][hex2tensor[i][0]] * 
        poli[1][hex2tensor[i][1]] * 
        poli[2][hex2tensor[i][2]];
  }





  void H1HighOrderHexTPSZ :: 
  GetHex2TensorMapping (FlatArray<int[3]> hex2tensor, FlatArray<double> factor) const
  {

    static int hex_points[][3] = 
      { 
	{ 0, 0, 0 },
	{ 1, 0, 0 },
	{ 1, 1, 0 },
	{ 0, 1, 0 },
	{ 0, 0, 1 },
	{ 1, 0, 1 },
	{ 1, 1, 1 },
	{ 0, 1, 1 }
      };


    // vertex shapes

    for (int i = 0; i < 8; i++)
      {
        hex2tensor[i][0] = hex_points[i][0];
        hex2tensor[i][1] = hex_points[i][1];
        hex2tensor[i][2] = hex_points[i][2];
        factor[i] = 1;
      }
        
    int ii = 8;

    // edge dofs
    const EDGE * edges = ElementTopology::GetEdges (ET_HEX);
    for (int i = 0; i < 12; i++)
      if (order_edge[i] >= 2)
	{
	  int p = order_edge[i];
	  int es = edges[i][0], ee = edges[i][1];
	  if (vnums[es] > vnums[ee]) swap (es, ee);

          double fac = 1, mult = 1;
          int dir = 0;

          for (int k = 0; k < 3; k++)
            if (hex_points[es][k] != hex_points[ee][k])
              {
                dir = k;
                if (hex_points[es][k] < hex_points[ee][k])
                  mult = -1;
              }

          for (int l = 0; l <= p-2; l++)
            {
              for (int k = 0; k < 3; k++)
                hex2tensor[ii][k] = hex_points[ee][k];
              hex2tensor[ii][dir] = l+2;

              factor[ii] = fac;
              fac *= mult;
              ii++;
            }
	}


     
    const FACE * faces = ElementTopology::GetFaces (ET_HEX);
    for (int i = 0; i < 6; i++)
      if (order_face[i][0] >= 2 && order_face[i][1] >= 2)
	{
	  int px = order_face[i][0];
	  int py = order_face[i][1];
	  int p = max2(px, py);

	  int fmax = 0;
	  for (int j=1; j<4; j++) 
	    if (vnums[faces[i][j]] > vnums[faces[i][fmax]]) fmax = j;  
	  
	  int f1 = faces[i][(fmax+3)%4];
	  int f2 = faces[i][(fmax+1)%4]; 
	  fmax = faces[i][fmax]; 
	  
	  if(vnums[f2]>vnums[f1]) swap(f1,f2);  // fmax > f1 > f2 
	  

          int ind[3];
          double fac1 = 1, fac2 = 1, mult1 = 1, mult2 = 1;
          for (int l = 0; l <= px-2; l++)
            {
              fac2 = 1;
              for (int m = 0; m <= py-2; m++)
                {
                  for (int k = 0; k < 3; k++)
                    {
                      ind[k] = hex_points[fmax][k];
                      if (hex_points[fmax][k] != hex_points[f2][k])
                        {
                          ind[k] = m+2;
                          if (hex_points[fmax][k] > hex_points[f2][k])
                            mult2 = -1;
                        }
                      if (hex_points[fmax][k] != hex_points[f1][k])
                        {
                          ind[k] = l+2;
                          if (hex_points[fmax][k] > hex_points[f1][k])
                            mult1 = -1;
                        }
                    }

                  factor[ii] = fac1*fac2;
                  hex2tensor[ii][0] = ind[0];
                  hex2tensor[ii][1] = ind[1];
                  hex2tensor[ii][2] = ind[2];
                  ii++;

                  fac2 *= mult2;
                }
              fac1 *= mult1;
            }

	}
    
    // volume dofs:
    if (order_inner[0] >= 2 && order_inner[1] >= 2 && order_inner[2] >= 2)
      {
	for (int i = 2; i <= order_inner[0]; i++)
	  for (int j = 2; j <= order_inner[1]; j++)
            for (int k = 2; k <= order_inner[2]; k++)
              {
                factor[ii] = 1;
                hex2tensor[ii][0] = i;
                hex2tensor[ii][1] = j;
                hex2tensor[ii][2] = k;
                ii++;
              }
      }    
  }








}


