/*********************************************************************/
/* File:   h1hofe.cpp                                                */
/* Author: Start                                                     */
/* Date:   1. June 2007                                              */
/*********************************************************************/


/*

High order integrators in tensor product form

*/
  


//using namespace tptransformation;

#include <fem.hpp>
#include "tptrafo.hpp"
#include "h1hofetpsz_stage2.hpp"

namespace ngfem
{

  
  using namespace ngfem;  
  using namespace tptransformation; 
  // ******************************** source integrator **************************************
  


  
  template <int D>
  class SourceIntegratorTPSZ : public SourceIntegrator<D>
  {
    typedef DiffOpId<D> DIFFOP;

    enum { DIM_SPACE   = DIFFOP::DIM_SPACE };
    enum { DIM_ELEMENT = DIFFOP::DIM_ELEMENT };
    enum { DIM_DMAT    = DIFFOP::DIM_DMAT };


  public:
    SourceIntegratorTPSZ (CoefficientFunction * acoef)
      : SourceIntegrator<D> (acoef)
    { 
      ;
    }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new SourceIntegratorTPSZ (coeffs[0]);
    }


    virtual void
    AssembleElementVector (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatVector<double> & elvec,
			   LocalHeap & lh) const
    {


      if (dynamic_cast<const H1HighOrderTetTP*> (&bfel))
        {
          const H1HighOrderTetTP & fel = static_cast<const H1HighOrderTetTP&> (bfel);

          int ndof = fel.GetNDof();
          
          elvec.AssignMemory (ndof, lh);
          elvec = 0;


          void * heapp = lh.GetPointer();

          
          int order = fel.Order();

          
          const IntegrationRule & irx = SelectIntegrationRule (ET_SEGM, 2*order+2);
          const IntegrationRule & iry = SelectIntegrationRule (ET_SEGM, 2*order+1);
          const IntegrationRule & irz = SelectIntegrationRule (ET_SEGM, 2*order  );

          int nip = irx.GetNIP() * iry.GetNIP() * irz.GetNIP();


          ArrayMem<Vec<3>, 100> axi(nip);
          ArrayMem<double, 100> aweight(nip);
          ArrayMem<Vec<3>, 100> ax(nip);
          ArrayMem<Mat<3,3>, 100> adxdxi(nip);

          int ii = 0;
          for (int i1 = 0; i1 < irx.GetNIP(); i1++)
            for (int i2 = 0; i2 < iry.GetNIP(); i2++)
              for (int i3 = 0; i3 < irz.GetNIP(); i3++, ii++)
                {
                  double x = irx[i1](0);
                  double y = iry[i2](0);
                  double z = irz[i3](0);
                  double lami[4];   // { x, y, z, 1-x-y-z };
                  
                  lami[fel.Sort(0)] = x;
                  lami[fel.Sort(1)] = y * (1-x);
                  lami[fel.Sort(2)] = z * (1-x) * (1-y);
                  lami[fel.Sort(3)] = 1 - lami[fel.Sort(0)] - lami[fel.Sort(1)] - lami[fel.Sort(2)];

                  axi[ii](0) = lami[0];
                  axi[ii](1) = lami[1];
                  axi[ii](2) = lami[2];
                  aweight[ii] = irx[i1].Weight()*iry[i2].Weight()*irz[i3].Weight() * sqr(1-x) * (1-y);
                }

          eltrans.CalcMultiPointJacobian (axi, ax, adxdxi, lh);

          Vec<DIM_DMAT> dvec;
          FlatVector<double> dvecs(nip, lh);

          for (int ii = 0; ii < nip; ii++)
            {
              IntegrationPoint ip(axi[ii], aweight[ii]);
              SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> 
                sip(ip, eltrans, ax[ii], adxdxi[ii], lh);

              (this->dvecop).GenerateVector (fel, sip, dvec, lh);

              double fac = fabs (sip.GetJacobiDet()) * ip.Weight();
              dvecs[ii] = dvec(0) * fac;
            }



          /*
          FlatArray<int[4]> tet2tensor (ndof, lh);
          fel.GetTet2TensorMapping (tet2tensor);
          */

          FlatArray<int[4]> tet2tensor = fel.GetTet2TensorMapping();

          int nd1d = 2 * (order);
          int nd2d = order * nd1d;

          FlatArray<int> tensor2trig ( nd2d, lh);
          tensor2trig = -1;

          for (int ii = 0; ii < ndof; ii++)
            {
              int j2d = tet2tensor[ii][1] * nd1d + tet2tensor[ii][2] * 2 + tet2tensor[ii][3];
              tensor2trig[j2d] = 0;
            }
      
          int ndtrig =0;
          for (int i = 0; i < nd2d; i++)
            if (tensor2trig[i] != -1)
              {
                tensor2trig[i] = ndtrig;
                ndtrig++;
              }

          ArrayMem<int[3], 100> trig2tensor(ndtrig);
          for (int i2 = 0; i2 < order; i2++)
            for (int i3 = 0; i3 < order; i3++)
              for (int i4 = 0; i4 < 2; i4++)
                {
                  int i2d = i2 * nd1d + i3 * 2 + i4;
                  if (tensor2trig[i2d] != -1)
                    {
                      trig2tensor[tensor2trig[i2d]][0] = i2;
                      trig2tensor[tensor2trig[i2d]][1] = i3;
                      trig2tensor[tensor2trig[i2d]][2] = i4;
                    }
                }

          FlatArray<int> tet2trig (ndof, lh);
          for (int i = 0; i < ndof; i++)
            tet2trig[i] = tensor2trig[tet2tensor[i][1] * nd1d + tet2tensor[i][2] * 2 + tet2tensor[i][3]];



          FlatArray<int> tensor2segm ( nd1d, lh);
          tensor2segm = -1;

          for (int ii = 0; ii < ndof; ii++)
            {
              int j1d = tet2tensor[ii][2] * 2 + tet2tensor[ii][3];
              tensor2segm[j1d] = 0;
            }
      
          int ndsegm =0;
          for (int i = 0; i < nd1d; i++)
            if (tensor2segm[i] != -1)
              {
                tensor2segm[i] = ndsegm;
                ndsegm++;
              }

          ArrayMem<int[2], 100> segm2tensor(ndsegm);
          for (int i3 = 0; i3 < order; i3++)
            for (int i4 = 0; i4 < 2; i4++)
              {
                int i1d = i3 * 2 + i4;
                if (tensor2segm[i1d] != -1)
                  {
                    segm2tensor[tensor2segm[i1d]][0] = i3;
                    segm2tensor[tensor2segm[i1d]][1] = i4;
                  }
              }





          FlatArray<int> trig2segm (ndtrig, lh);
          for (int i = 0; i < ndtrig; i++)
            trig2segm[i] = tensor2segm[trig2tensor[i][1] * 2 + trig2tensor[i][2]];


          
          ArrayMem<double, 20> polx(order+1), poly(order+1), polz(order+1), polzz(2);
          FlatVector<double> shape(ndof, lh);


          FlatMatrix<> powx(irx.GetNIP(), order+1, lh);
          FlatMatrix<> powy(iry.GetNIP(), order+1, lh);
          FlatMatrix<> powz(irz.GetNIP(), order+1, lh);

          FlatMatrix<> shapesx(irx.GetNIP(), order+1, lh);
          FlatMatrix<> shapesy(iry.GetNIP(), order+1, lh);
          FlatMatrix<> shapesz(irz.GetNIP(), order+1, lh);


          for (int i = 0; i < irx.GetNIP(); i++)
            {
              double fac = 1;
              double x = irx[i](0);

              double * hp = &shapesx(i, 1);
              LegendrePolynomialMult (order-2, 2*x-1, x, hp);
              shapesx(i,0) = 1.0;

              for (int j = 0; j <= order; j++, fac *= 1-x)
                powx(i,j) = fac;
            }

          for (int i = 0; i < iry.GetNIP(); i++)
            {
              double fac = 1;
              double y = iry[i](0);

              double * hp = &shapesy(i, 1);
              LegendrePolynomialMult (order-2, 2*y-1, y, hp);
              shapesy(i,0) = 1.0;

              for (int j = 0; j <= order; j++, fac *= 1-y)
                powy(i,j) = fac;
            }

          for (int i = 0; i < irz.GetNIP(); i++)
            {
              double fac = 1;
              double z = irz[i](0);

              double * hp = &shapesz(i, 1);
              LegendrePolynomialMult (order-2, 2*z-1, z, hp);
              shapesz(i,0) = 1.0;

              for (int j = 0; j <= order; j++, fac *= 1-z)
                powz(i,j) = fac;
            }
          
          FlatVector<> sum2d(ndtrig, lh);
          FlatVector<> sum1d(ndsegm, lh);

          for (int i1 = 0, ii = 0; i1 < irx.GetNIP(); i1++)
            {
              sum2d = 0;

              for (int i2 = 0; i2 < iry.GetNIP(); i2++)
                {
                  sum1d = 0;
                  for (int i3 = 0; i3 < irz.GetNIP(); i3++, ii++)
                    {
                      for (int i = 0; i < ndsegm; i++)
                        {
                          sum1d(i) += 
                            dvecs[ii] * 
                            shapesz(i3, segm2tensor[i][0]) *
                            powz(i3, segm2tensor[i][1]);
                          
                        }
                    }
                  
                  for (int i = 0; i < ndtrig; i++)
                    sum2d(i) +=
                      shapesy(i2, trig2tensor[i][0]) *
                      powy(i2, trig2tensor[i][1] + trig2tensor[i][2]) *
                      sum1d(trig2segm[i]);
                }

              for (int i = 0; i < ndof; i++)
                elvec(i) +=
                  shapesx(i1, tet2tensor[i][0]) * 
                  powx(i1, tet2tensor[i][1] + tet2tensor[i][2] + tet2tensor[i][3]) *
                  sum2d(tet2trig[i]);
            }



          
          lh.CleanUp();
          return;
        }



      SourceIntegrator<D>::AssembleElementVector(bfel, eltrans, elvec, lh);
    }
  };














  // ******************************** Laplace **************************************



  
  template <int D>
  class LaplaceIntegratorTPSZ : public LaplaceIntegrator<D>
  {
    typedef DiffOpGradient<D> DIFFOP;
    enum { DIM_SPACE   = DIFFOP::DIM_SPACE };
    enum { DIM_ELEMENT = DIFFOP::DIM_ELEMENT };
    enum { DIM_DMAT    = DIFFOP::DIM_DMAT };

  public:
    LaplaceIntegratorTPSZ (CoefficientFunction * acoef)
      : LaplaceIntegrator<D> (acoef)
    { ; }

    static Integrator * Create (ARRAY<CoefficientFunction*> & coeffs)
    {
      return new LaplaceIntegratorTPSZ (coeffs[0]);
    }

    virtual void
    AssembleElementMatrix (const FiniteElement & bfel,
			   const ElementTransformation & eltrans, 
			   FlatMatrix<double> & elmat,
			   LocalHeap & lh) const
    {
      if (dynamic_cast<const H1HighOrderTetTPSZ*> (&bfel) ||
          dynamic_cast<const H1HighOrderPrismTPSZ*> (&bfel) ||
          dynamic_cast<const H1HighOrderHexTPSZ*> (&bfel))
        {
          int ndof = bfel.GetNDof();
          
          elmat.AssignMemory (ndof, ndof, lh);

          void * heapp = lh.GetPointer();
          

          int order = bfel.Order();
          int nipx, nipy, nipz;

          int ndof2d, ndof1d;

          FlatVector<Mat<3> > tdmat;      // transformed D-matrix

          FlatArray<int> map3dto2d;       // tet2trig
          FlatArray<int> map2dto1d;       // trig2segm


          FlatMatrix<Vec<2> > fac_xdx;    // ndof   nipx
          FlatMatrix<Vec<2> > fac_ydy;    // ndof2d nipy
          FlatMatrix<Vec<2> > fac_zdz;    // ndof1d nipz



	  

        
          if (dynamic_cast<const H1HighOrderTetTPSZ*> (&bfel))
            {
	 
              const H1HighOrderTetTPSZ & fel = static_cast<const H1HighOrderTetTPSZ&> (bfel);
          
              static int timer = NgProfiler::CreateTimer ("Fast assembling - tet");
              static int timer1 = NgProfiler::CreateTimer ("Fast assembling - tet - geometry");

              NgProfiler::RegionTimer reg (timer);

              // const IntegrationRule & ir1dx = SelectIntegrationRule (ET_SEGM, 2*order);
              const IntegrationRule & ir1dy = SelectIntegrationRule (ET_SEGM, 2*order-1);
              const IntegrationRule & ir1dz = SelectIntegrationRule (ET_SEGM, 2*order-2);
      
              static IntegrationRule ir1dx;
              if (ir1dx.GetNIP() == 0)
                {
                  int np = order;
                  ARRAY<double> xi(np), wi(np);
                  ComputeGaussJacobiRule (np, xi, wi, 2, 0);
                  for (int i = 0; i < np; i++)
                    ir1dx.AddIntegrationPoint (new IntegrationPoint (xi[i], 0, 0, wi[i]));
                }

              nipx = ir1dx.GetNIP();
              nipy = ir1dy.GetNIP();
              nipz = ir1dz.GetNIP();

              map2dto1d = fel.Get2dTo1dMapping ();
              map3dto2d = fel.Get3dTo2dMapping ();

              ndof1d = fel.GetSegm2TensorMapping().Size();
              ndof2d = fel.GetTrig2TensorMapping().Size();


              fac_zdz.AssignMemory(ndof1d, ir1dz.GetNIP(), lh);
              fac_ydy.AssignMemory(ndof2d, ir1dy.GetNIP(), lh);
              fac_xdx.AssignMemory(ndof, ir1dx.GetNIP(), lh);

              FlatArray<AutoDiff<1> > facz(ndof1d, lh);
              for (int i1 = 0; i1 < ir1dz.GetNIP(); i1++)
                {
                  AutoDiff<1> z (ir1dz[i1](0), 0);
                  fel.CalcZFactor (z, facz, lh);
                  
                  for (int jj = 0; jj < ndof1d; jj++)
                    {
                      fac_zdz(jj, i1)(0) = facz[jj].Value();
                      fac_zdz(jj, i1)(1) = facz[jj].DValue(0);
                    }
                }

              FlatArray<AutoDiff<1> > facy(ndof2d, lh);
              for (int i1 = 0; i1 < ir1dy.GetNIP(); i1++)
                {
                  AutoDiff<1> y (ir1dy[i1](0), 0);
                  fel.CalcYFactor (y, facy, lh);
                  
                  for (int jj = 0; jj < ndof2d; jj++)
                    {
                      fac_ydy(jj, i1)(0) = facy[jj].Value();
                      fac_ydy(jj, i1)(1) = facy[jj].DValue(0);
                    }
                }

              FlatArray<AutoDiff<1> > facx(ndof, lh);
              for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
                  {
                  AutoDiff<1> x (ir1dx[i1](0), 0);
                  fel.CalcXFactor (x, facx, lh);
                  
                  for (int jj = 0; jj < ndof; jj++)
                    {
                      fac_xdx(jj, i1)(0) = facx[jj].Value();
                      fac_xdx(jj, i1)(1) = facx[jj].DValue(0);
                    }
                }
      
              // geometry calculation

 
              int nip = nipx * nipy * nipz;
              tdmat.AssignMemory(nip, lh); 

              NgProfiler::StartTimer (timer1);

 
	      TensorProductTransformation3D tp_trafo(fel.ElementType(),fel.GetVNums());

              ArrayMem<Vec<3>, 100> axi(nip);
              ArrayMem<double, 100> aweight(nip);
              ArrayMem<Vec<3>, 100> ax(nip);
              ArrayMem<Mat<3,3>, 100> adxdxi(nip);
	      ArrayMem<double[3], 100> axtp(nip); 

              int ii = 0;
              for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
                for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
		      double xtp[] = {ir1dx[i1](0),ir1dy[i2](0),ir1dz[i3](0)};
		      axi[ii] = tp_trafo.ElementPoint(xtp);                     
		      aweight[ii] = ir1dx[i1].Weight()*ir1dy[i2].Weight()*ir1dz[i3].Weight() 
		      		* tp_trafo.GetJacobiDet(xtp);
                    }
	      
	      /* 
	      int ii = 0;
              for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
		{ 
		  xtp[0] = ir1dx[i1](0); 
		  for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
		    { 
		      xtp[1] =  ir1dy[i2](0); 
		      for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
			{
			  xtp[2] = ir1dz[i3](0); 
			  axi[ii] = tp_trafo.ElementPoint(xtp);                     
			  aweight[ii] = ir1dx[i1].Weight()*ir1dy[i2].Weight()*ir1dz[i3].Weight() 
			    * tp_trafo.GetJacobiDet(xtp);
			  
			    } */ 
	      


              eltrans.CalcMultiPointJacobian (axi, ax, adxdxi, lh);

	      /*   
              for (int ii = 0; ii < nip; ii++)
                {

                  IntegrationPoint ip(axi[ii], aweight[ii]);
                  SpecificIntegrationPoint<3,3> sip(ip, eltrans, ax[ii], adxdxi[ii], lh);

                  Mat<3> dmat, hmat;
                  this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
                  dmat *= fabs (sip.GetJacobiDet()) * ip.Weight();

                  hmat = sip.GetJacobianInverse() * dmat;
                  tdmat[ii] = hmat * Trans (sip.GetJacobianInverse());                   
		  }*/ 
	      
	      /*   for (int ii = 0; ii < nip; ii++)
                {
                  IntegrationPoint ip(axi[ii], aweight[ii]);
                  SpecificIntegrationPoint<3,3> sip(ip, eltrans, ax[ii], adxdxi[ii], lh);

                  Mat<3> dmat, hmat, trans;
                  this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
                  dmat *= fabs (sip.GetJacobiDet()) * ip.Weight(); 
		  trans = tp_trafo.CalcJacobianInverse(axtp[ii])*sip.GetJacobianInverse(); 
		  		  
                  hmat = trans * dmat;
                  tdmat[ii] = hmat * Trans (trans);
		  } */ 


	    	     	
	      // compute D-Mats, reference element to mapped     
              // tet permutation plus hex -> tet mapping
	      // the all-in-one-version was faster on limette 
	       
              for (int i1 = 0, ii=0; i1 < nipx; i1++)
                {
		  tp_trafo.SetX(ir1dx[i1](0)); 
			
                  for (int i2 = 0; i2 < nipy; i2++)
                    {
		      tp_trafo.SetY(ir1dy[i2](0));

                      for (int i3 = 0; i3 < nipz; i3++, ii++)
                        {
			  tp_trafo.SetZ(ir1dz[i3](0)); 
			  
			  IntegrationPoint ip(axi[ii], aweight[ii]);
			  SpecificIntegrationPoint<3,3> sip(ip, eltrans, ax[ii], adxdxi[ii], lh);

			  Mat<3> dmat, hmat, trans;
			  this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
			  dmat *= fabs (sip.GetJacobiDet()) * ip.Weight();
			  
			  trans = tp_trafo.GetJacobianInverse() * sip.GetJacobianInverse();  
			  hmat = trans * dmat;
			  tdmat[ii] = hmat * Trans (trans);                  
			  
			  
			  //Mat<3> trans = tp_trafo.GetJacobianInverse();
			  //Mat<3> hmat = trans * tdmat[ii];
                          //tdmat[ii] = hmat * Trans (trans); 
                        }
                    }
      } 

              NgProfiler::StopTimer (timer1);
            }

          
          if (dynamic_cast<const H1HighOrderPrismTPSZ*> (&bfel))
            {
              const H1HighOrderPrismTPSZ & fel = static_cast<const H1HighOrderPrismTPSZ&> (bfel);

              if (fel.Sort(0)+3 != fel.Sort(3))  // same singular vertex at bottom and top
                {
                  static int timerprism = NgProfiler::CreateTimer ("Std assembling - Prism");
                  NgProfiler::RegionTimer reg (timerprism);
                  
                  LaplaceIntegrator<D> :: AssembleElementMatrix (bfel, eltrans, elmat, lh);
                  return;
                }
              
              static int timerprism = NgProfiler::CreateTimer ("Fast assembling - Prism");
              NgProfiler::RegionTimer reg (timerprism);


              const IntegrationRule & ir1dx = SelectIntegrationRule (ET_SEGM, 2*order+1);
              const IntegrationRule & ir1dy = SelectIntegrationRule (ET_SEGM, 2*order);
              const IntegrationRule & ir1dz = SelectIntegrationRule (ET_SEGM, 2*order);

              nipx = ir1dx.GetNIP();
              nipy = ir1dy.GetNIP();
              nipz = ir1dz.GetNIP();

              int sort[6];
              for (int i = 0; i  < 6; i++) sort[i] = fel.Sort(i);


              ArrayMem<AutoDiff<1>, 20> poly(order+1), powy(order+1);
              ArrayMem<AutoDiff<1>, 20> polz(order+1), polyy(2);



              FlatArray<int[4]> prism2tensor;
              FlatArray<int[3]> trig2tensor; // brauchma eigentlich net
              FlatArray<double> fac;

              prism2tensor =  fel.GetPrism2TensorMapping(); 
	      fac = fel.GetFactorPrism2TensorMapping(); 

              int nd1d = order+1;      // zorder
              int nd2d = (order+1) * (order+1) * 2;  // poly, polyy, polz

	      map2dto1d = fel.Get2dTo1dMapping ();
              map3dto2d = fel.Get3dTo2dMapping ();

	      ndof1d = order+1;
	      ndof2d = fel.GetTrig2TensorMapping().Size();

	      trig2tensor = fel.GetTrig2TensorMapping(); 

	      FlatArray<AutoDiff<1> > shapesad(order+1, lh);

              fac_zdz.AssignMemory(ndof1d, nipz, lh);


	      FlatArray<AutoDiff<1> > facz(ndof1d, lh);
              for (int i3 = 0; i3 < nipz; i3++)
                {
                  AutoDiff<1> adx (ir1dz[i3](0), 0);
                  AutoDiff<1> xi = 2*adx-1;
              
                  shapesad[0] = 1-adx;
                  shapesad[1] = adx;
                  
                  AutoDiff<1> * hp = &shapesad[2];
                  LegendrePolynomialMult (order-2, xi, 0.25 * (1-xi*xi), hp);
		  for (int j = 0; j <= order; j++)
                    {
                      fac_zdz(j, i3)(0) = shapesad[j].Value();
                      fac_zdz(j, i3)(1) = shapesad[j].DValue(0);
                    }
		  /* 
		    
                  AutoDiff<1> z (ir1dz[i3](0), 0);
                  fel.CalcZFactor (z, facz, lh);
                  for (int jj = 0; jj < ndof1d; jj++)
                    {
                      fac_zdz(jj, i3)(0) = facz[jj].Value();
                      fac_zdz(jj, i3)(1) = facz[jj].DValue(0);
                    }
		  */ 
                }
	      
                 
	      

              fac_ydy.AssignMemory(ndof2d, nipy, lh);
              for (int iy = 0; iy < ir1dy.GetNIP(); iy++)
                {
                  AutoDiff<1> yr1 (ir1dy[iy](0), 0);
                  AutoDiff<1> * hp1 = &poly[1];
                  LegendrePolynomialMult (order-2, 2*yr1-1, yr1, hp1);
                  poly[0] = 1;

                  polyy[0] = 1;
                  polyy[1] = (1-yr1); 
          
                  for (int jj = 0; jj < ndof2d; jj++)
                    {
                      AutoDiff<1> fy = 
                        poly[trig2tensor[jj][0]] * polyy[trig2tensor[jj][1]];
                      fac_ydy(jj, iy)(0) = fy.Value();
                      fac_ydy(jj, iy)(1) = fy.DValue(0);
                    }
                }


              // ArrayMem<AutoDiff<1>, 20> polx(order+1);
              ArrayMem<AutoDiff<1>, 20> powx(order+1);
              ArrayMem<AutoDiff<1>, 100> memx((order+1)*order);
              FlatMatrix<AutoDiff<1> > polsx(order+1, order, &memx[0]);

              fac_xdx.AssignMemory(ndof, nipx, lh);
              for (int i1 = 0; i1 < nipx; i1++)
                {
                  AutoDiff<1> xr1 (ir1dx[i1](0), 0);
                  /*
                  AutoDiff<1> * hp1 = &polx[1];
                  LegendrePolynomialMult (order-2, 2*xr1-1, xr1, hp1);
                  polx[0] = 1;
                  */

                  for (int i = 0; i <= order; i++)
                    {
                      AutoDiff<1> * hp = &polsx(i, 1);
#ifdef JACOBI
                      JacobiPolynomial (order-2, 2*xr1-1, max(0,2*i-2), 0, hp);
#else
                      LegendrePolynomial (order-2, 2*xr1-1, hp);
#endif
                      for (int j = 1; j < order; j++)
                        polsx(i,j) *= xr1;
                      polsx(i,0) = 1.0;
                    }

                  AutoDiff<1> hv = 1.0;
                  for (int i = 0; i <= order; i++)
                    {
                      powx[i] = hv;
                      hv *= (1-xr1);
                    }

                  for (int jj = 0; jj < ndof; jj++)
                    {
                      /*
                      AutoDiff<1> fx = 
                        polx[prism2tensor[jj][0]] *
                        powx[prism2tensor[jj][1]+prism2tensor[jj][2]];
                      */
                      int power = prism2tensor[jj][1]+prism2tensor[jj][2];
                      AutoDiff<1> fx =
                        polsx(power, prism2tensor[jj][0]) * powx[power];

                      fac_xdx(jj, i1)(0) = fx.Value()   * fac[jj];
                      fac_xdx(jj, i1)(1) = fx.DValue(0) * fac[jj];
                    }
                }
      
              // geometry calculation

              int nip = nipx * nipy * nipz;
              tdmat.AssignMemory(nip, lh);

              // tet permutation transformation
              double dlamdx[3][2] = 
                { { 1, 0 },
                  { 0, 1 },
                  { -1, -1 }};

              Mat<3> trans2;
              trans2 = 0.0;
              for (int k = 0; k < 2; k++)
                for (int l = 0; l < 2; l++)
                  trans2(l,k) = dlamdx[sort[l]][k];
              trans2(2,2) = 1.0;


              ArrayMem<Vec<3>, 100> axi(nip);
              ArrayMem<double, 100> aweight(nip);
              ArrayMem<Vec<3>, 100> ax(nip);
              ArrayMem<Mat<3,3>, 100> adxdxi(nip);

              int ii = 0;
              for (int i1 = 0; i1 < ir1dx.GetNIP(); i1++)
                for (int i2 = 0; i2 < ir1dy.GetNIP(); i2++)
                  for (int i3 = 0; i3 < ir1dz.GetNIP(); i3++, ii++)
                    {
                      double x = ir1dx[i1](0);
                      double y = ir1dy[i2](0);
                      double z = ir1dz[i3](0);
                      double lami[4];   // { x, y, z, 1-x-y-z };
                  
                      lami[sort[0]] = x;
                      lami[sort[1]] = y * (1-x);
                      lami[sort[2]] = 1 - lami[sort[0]] - lami[sort[1]];

                      axi[ii](0) = lami[0];
                      axi[ii](1) = lami[1];
                      axi[ii](2) = z;
                      aweight[ii] = ir1dx[i1].Weight()*ir1dy[i2].Weight()*ir1dz[i3].Weight() * (1-x);
                    }

              eltrans.CalcMultiPointJacobian (axi, ax, adxdxi, lh);

              ii = 0;
              for (int i1 = 0; i1 < nipx; i1++)
                {
                  double x = ir1dx[i1](0);
                  double invx = 1.0 / (1-x);

                  for (int i2 = 0; i2 < nipy; i2++)
                    {
                      double y = ir1dy[i2](0);

                      for (int i3 = 0; i3 < nipz; i3++, ii++)
                        {
                          double z = ir1dz[i3](0);
                  
                          IntegrationPoint ip(axi[ii], aweight[ii]);
                          SpecificIntegrationPoint<3,3> sip(ip, eltrans, ax[ii], adxdxi[ii], lh);

                          Mat<3> dmat, trans3, trans, hmat;
                          this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
                          dmat *= fabs (sip.GetJacobiDet()) * ip.Weight();
                  
                          // hex -> tet transform
                          trans3(0,0) = 1;
                          trans3(0,1) = 0;
                          trans3(0,2) = 0;

                          trans3(1,0) = y*invx;
                          trans3(1,1) = 1*invx;
                          trans3(1,2) = 0;

                          trans3(2,0) = 0;
                          trans3(2,1) = 0;
                          trans3(2,2) = 1;

                          hmat = trans2 * sip.GetJacobianInverse();
                          trans = trans3 * hmat;   // total transformation matrix

                          hmat = trans * dmat;
                          tdmat[ii] = hmat * Trans (trans); 
                        }
                    }
                }
            }




      
          else  if (dynamic_cast<const H1HighOrderHexTPSZ*> (&bfel))

            {
              const H1HighOrderHexTPSZ & fel = static_cast<const H1HighOrderHexTPSZ&> (bfel);

              static int timerhex = NgProfiler::CreateTimer ("Fast assembling - Hex");
              static int timerhexgeom = NgProfiler::CreateTimer ("Fast assembling - geometry");

              NgProfiler::RegionTimer reg (timerhex);


              const IntegrationRule & ir1d = SelectIntegrationRule (ET_SEGM, 2*order);

              nipx = nipy = nipz = ir1d.GetNIP();
              int nip = nipx * nipy * nipz;

              FlatMatrix<> shape1d(ir1d.GetNIP(), order+1, lh);
              FlatMatrix<> dshape1d(ir1d.GetNIP(), order+1, lh);
              FlatArray<AutoDiff<1> > shapesad(order+1, lh);
          
              for (int i = 0; i < ir1d.GetNIP(); i++)
                {
                  AutoDiff<1> adx (ir1d[i](0), 0);
                  AutoDiff<1> xi = 2*adx-1;
              
                  shapesad[0] = 1-adx;
                  shapesad[1] = adx;
              
                  AutoDiff<1> * hp = &shapesad[2];
                  LegendrePolynomialMult (order-2, xi, 0.25 * (1-xi*xi), hp);

                  for (int j = 0; j <= order; j++)
                    {
                      shape1d(i,j) = shapesad[j].Value();
                      dshape1d(i,j) = shapesad[j].DValue(0);
                    }
                }


              FlatArray<int[3]> hex2tensor(ndof, lh);
              FlatArray<double> fac(ndof, lh);

              fel.GetHex2TensorMapping (hex2tensor, fac);
          

              ndof2d = (order+1)*(order+1);
              ndof1d = order+1;

              map3dto2d = FlatArray<int> (ndof, lh);
              for (int jj = 0; jj < ndof; jj++)
                map3dto2d[jj] = hex2tensor[jj][1] * (order+1) + hex2tensor[jj][2];

              map2dto1d = FlatArray<int> (ndof2d, lh);
              for (int jj = 0; jj < ndof2d; jj++)
                map2dto1d[jj] = jj % (order+1);



              fac_xdx.AssignMemory (ndof, ir1d.GetNIP(), lh);
              for (int i = 0; i < ndof; i++)
                for (int j = 0; j < ir1d.GetNIP(); j++)
                  {
                    fac_xdx(i, j)(0) = fac[i] * shape1d(j, hex2tensor[i][0]);
                    fac_xdx(i, j)(1) = fac[i] * dshape1d(j, hex2tensor[i][0]);
                  }

              fac_ydy.AssignMemory (ndof2d, nipy, lh);
              for (int i = 0; i < ndof2d; i++)
                for (int j = 0; j < nipy; j++)
                  {
                    fac_ydy(i, j)(0) = shape1d(j, i / (order+1));
                    fac_ydy(i, j)(1) = dshape1d(j, i / (order+1));
                  }


              fac_zdz.AssignMemory (ndof1d, nipz, lh);
              for (int i = 0; i < ndof1d; i++)
                for (int j = 0; j < nipz; j++)
                  {
                    fac_zdz(i, j)(0) = shape1d(j, i);
                    fac_zdz(i, j)(1) = dshape1d(j, i);
                  }




              NgProfiler::StartTimer (timerhexgeom);


              tdmat.AssignMemory(nip, lh);

              int ii = 0;
              for (int i1 = 0; i1 < ir1d.GetNIP(); i1++)
                for (int i2 = 0; i2 < ir1d.GetNIP(); i2++)
                  for (int i3 = 0; i3 < ir1d.GetNIP(); i3++, ii++)
                    {
                      double x = ir1d[i1](0);
                      double y = ir1d[i2](0);
                      double z = ir1d[i3](0);
                  
                      IntegrationPoint ip(x, y, z,
                                          ir1d[i1].Weight()*ir1d[i2].Weight()*ir1d[i3].Weight());
                  
                      SpecificIntegrationPoint<DIM_ELEMENT,DIM_SPACE> sip(ip, eltrans, lh);
                  
                      Mat<3> dmat, trans, hmat;
                  
                      this->dmatop.GenerateMatrix (fel, sip, dmat, lh);
                      dmat *= fabs (sip.GetJacobiDet()) * sip.IP().Weight();
                  
                      trans = sip.GetJacobianInverse();
                      hmat = trans * dmat;
                      tdmat[ii] = hmat * Trans (trans); 
                    }

              NgProfiler::StopTimer (timerhexgeom);
            }


          
          static int timeryz = NgProfiler::CreateTimer ("Fast assembling - transform yz");
          static int timerx = NgProfiler::CreateTimer ("Fast assembling - transform x");



          FlatMatrix<Mat<2> > sum2d(ndof2d, lh);
          FlatMatrix<Mat<2> > allsum2d(ndof2d*ndof2d, nipx, lh);
          
          FlatMatrix<Mat<3> > sum1d(ndof1d, lh);
          
          FlatVector<Mat<3> > mult1d(ndof1d, lh);
          FlatMatrix<Vec<6> > mult2d(ndof1d, ndof2d, lh);

          NgProfiler::StartTimer (timeryz);


          NgProfiler::AddFlops (timeryz, 9*nipx*nipy*nipz*ndof1d);
          NgProfiler::AddFlops (timeryz, 9*nipx*nipy*ndof1d*ndof1d/2);
          NgProfiler::AddFlops (timeryz, 9*nipx*nipy*ndof1d*ndof2d);
          NgProfiler::AddFlops (timeryz, 3*nipx*nipy*ndof2d*ndof2d);

          // transform z and y

          for (int i1 = 0, ii = 0; i1 < nipx; i1++)
            {
              sum2d = 0.0;

              for (int i2 = 0; i2 < nipy; i2++)
                {
                  sum1d = 0.0;

                  for (int i3 = 0; i3 < nipz; i3++, ii++)
                    {
                      for (int j = 0; j < ndof1d; j++)
                        for (int l = 0; l < 3; l++)
                          {
                            mult1d(j)(0,l) = tdmat[ii](0,l) * fac_zdz(j, i3)(0);
                            mult1d(j)(1,l) = tdmat[ii](1,l) * fac_zdz(j, i3)(0);
                            mult1d(j)(2,l) = tdmat[ii](2,l) * fac_zdz(j, i3)(1);
                          }

                      for (int j = 0; j < ndof1d; j++)
                        for (int k = 0; k <= j; k++)
                          for (int l = 0; l < 3; l++)
                            {
                              sum1d(j,k)(l,0) += mult1d(j)(l,0) * fac_zdz (k, i3)(0);   
                              sum1d(j,k)(l,1) += mult1d(j)(l,1) * fac_zdz (k, i3)(0);   
                              sum1d(j,k)(l,2) += mult1d(j)(l,2) * fac_zdz (k, i3)(1);  
                            }
                    }
                  for (int j = 0; j < ndof1d; j++)
                    for (int k = 0; k < j; k++)
                      sum1d(k,j) = Trans(sum1d(j,k));

                  // mult pj:
                  for (int jj = 0; jj < ndof2d; jj++)
                    {
                      int j1d = map2dto1d[jj];

                      double fy = fac_ydy(jj, i2)(0);
                      double fdy = fac_ydy(jj, i2)(1);

                      for (int k = 0; k < ndof1d; k++)
                        {
                          mult2d(k,jj)(0) = fy * sum1d(j1d, k)(0,0);
                          mult2d(k,jj)(1) = fy * sum1d(j1d, k)(0,1);
                          mult2d(k,jj)(2) = fy * sum1d(j1d, k)(0,2);
                          mult2d(k,jj)(3) = fy * sum1d(j1d, k)(2,0) + fdy * sum1d(j1d,k)(1,0);
                          mult2d(k,jj)(4) = fy * sum1d(j1d, k)(2,1) + fdy * sum1d(j1d,k)(1,1);
                          mult2d(k,jj)(5) = fy * sum1d(j1d, k)(2,2) + fdy * sum1d(j1d,k)(1,2);
                        }
                    }

                  // mult pk:
                  for (int kk = 0; kk < ndof2d; kk++)
                    {
                      int k1d = map2dto1d[kk];

                      double fy = fac_ydy(kk, i2)(0);
                      double fdy = fac_ydy(kk, i2)(1);
                  
                      for (int jj = 0; jj <= kk; jj++)
                        {
                          Mat<2> & hsum2d = sum2d(kk,jj);
                          const Vec<6> & hmult2d = mult2d(k1d,jj);

                          hsum2d(0,0) += fy * hmult2d(5) + fdy * hmult2d(4);
                          hsum2d(0,1) += fy * hmult2d(2) + fdy * hmult2d(1);
                          hsum2d(1,0) += fy * hmult2d(3);
                          hsum2d(1,1) += fy * hmult2d(0);
                        }
                    }
                }

              for (int j2d = 0; j2d < ndof2d; j2d++)
                {
                  for (int k2d = 0; k2d <= j2d; k2d++)
                    allsum2d(j2d*ndof2d+k2d, i1) = sum2d(j2d, k2d);
                  for (int k2d = j2d+1; k2d < ndof2d; k2d++)
                    allsum2d(j2d*ndof2d+k2d, i1) = Trans (sum2d(k2d, j2d));
                }
            }

          NgProfiler::StopTimer (timeryz);

          NgProfiler::StartTimer (timerx);


          // transform x

          FlatMatrix<Vec<2> > help(ndof, nipx, lh);

          FlatArray<int> map2dto3d(ndof2d * (order+1), lh);
          map2dto3d = -1;
          for (int j = 0; j < ndof; j++)
            {
              int start = (order+1) * map3dto2d[j];
              while (map2dto3d[start] != -1) start++;
              map2dto3d[start] = j;
            }


          NgProfiler::AddFlops (timerx, 4*ndof2d*ndof*nipx);
          NgProfiler::AddFlops (timerx, ndof*ndof*nipx);

          for (int k2d = 0; k2d < ndof2d; k2d++)
            {
              for (int jj = 0; jj < ndof; jj++)
                {
                  int j2d = map3dto2d[jj];
                  int ind = k2d*ndof2d+j2d;

                  for (int i1 = 0; i1 < nipx; i1++)
                    help(jj, i1) = allsum2d(ind, i1) * fac_xdx(jj, i1); 
                }
              
              for (int k1d = 0; k1d <= order; k1d++)
                {
                  int kk = map2dto3d[k2d*(order+1)+k1d];
                  if (kk == -1) continue;

                  Vec<2> * p_sds = &fac_xdx(kk,0);
                  for (int j = 0; j <= kk; j++)
                    {
                      Vec<2> * p_help = &help(j,0);
                      
                      double sum = 0;
                      
                      for (int i1 = 0; i1 < nipx; i1++)
                        sum +=  InnerProduct (p_sds[i1], p_help[i1]);
                      
                      elmat(kk*ndof+j) = sum;
                    }
                }
            }              
        
          for (int j = 0; j < ndof; j++)
            for (int k = 0; k < j; k++)
              elmat(k,j) = elmat(j,k);

          NgProfiler::StopTimer (timerx);

          lh.CleanUp (heapp);

          return;
        }

      else

        {

          static int timeretc = NgProfiler::CreateTimer ("Fast assembling - etc");
          NgProfiler::RegionTimer reg (timeretc);


          LaplaceIntegrator<D> :: AssembleElementMatrix (bfel, eltrans, elmat, lh);
          return;
        }
      
    }
    
  };











  
  
  namespace
  {
    class Init
    { 
    public: 
      Init ();
    };        
    
    Init::Init()
    {
      GetIntegrators().AddLFIntegrator ("sourcetpsz", 3, 1,
                                        SourceIntegratorTPSZ<3>::Create);
      GetIntegrators().AddBFIntegrator ("laplacetpsz", 3, 1,
                                        LaplaceIntegratorTPSZ<3>::Create);

    }
    
    Init init;
    
  }




  







}

