geometry = mainz_rect.geo
mesh = mainz_rect.vol

define constant heapsize = 500000000

define constant geometryorder = 2 
define constant pi = 3.14159265
define constant frequ = 20000
define constant omega = (2*frequ*3.14159265) 

define constant mu0 = (4.E-7*pi)
define constant muinv = (1.0/mu0)
define constant eps0 = 8.8541878176E-12

define constant mur = 1.00007 
define constant epsr = 24 
define constant cond = 7E-2
 
define constant px = 1
define constant py = 0
define constant pz = 0


define coefficient muinv material		
ground (1.0 / (mu0*mur))
default (1.0/ mu0)	

define coefficient sigma  material
ground (cond * omega)
object (omega*omega*omega)
default 0

define coefficient eps material
ground (-1.0*omega*omega*epsr*eps0) 
default (-1.0*omega*omega*eps0) 

define coefficient reg material
default (0.001)

define coefficient polx material
sensor (px)
default (0)

define coefficient poly material
sensor (py)
default (0)

define coefficient polz material
sensor (pz)
default 0


# boundary conditions 
# sensor plane -bc=2  

define coefficient sensor_neumann_x 
0, 100, 0, 0, 0, 0, 0, 

define coefficient zero
0,0,0,0,0,0,0,0,0, 

define coefficient pen_obj 
# 0,0,1e12,0,
0,0,0,0,

define coefficient pen_kugel
1e12,0,0,0, 
# 0,0,0,0, 

#grad-doms in air !! (-> air penalty) 
# define fespace v -type=hcurlho -order=3 -complex -eliminate_internal #-gradientdomains=[0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]
# define fespace vr -type=hcurlho -order=3 -eliminate_internal
define fespace v -type=hcurlho -order=3 -complex  -eliminate_internal 
define fespace vr -type=hcurlho -order=3  -eliminate_internal

define gridfunction u -fespace=v

# define linearform f -fespace=v
# neumannedge sensor_neumann_x zero zero  ## (n x u) x n
# 
# define linearform fr -fespace=vr  
# neumannedge sensor_neumann_x zero zero 

define linearform f -fespace=v
curledge polx poly polz

define linearform fr -fespace=vr  

# numproc adddipole dip -linearform=f 
#   -x=[-0.1,0,0.1] 
#   -y=[-0.1,0,0.1] 
#   -z=[0.05] -pol=[1,0,0]

define bilinearform a -fespace=v -eliminate_internal -linearform=f
curlcurledge muinv 
massedge sigma -imag 
massedge eps
# massedge reg 
# robinedge pen_obj 
robinedge pen_kugel 
# robinedge pen_kugel -imag


define bilinearform ar -fespace=vr -eliminate_internal -linearform=fr
curlcurledge muinv
massedge sigma  
massedge eps
# massedge reg 
# robinedge pen_obj   ## E=-du/dt = -i omega u => n x E = n x (- i omega u) = 0
robinedge pen_kugel 


define preconditioner creal -fespace=v -type=multigrid -coarsetype=direct -smoother=block -bilinearform=ar 
define preconditioner c -type=complex -realpreconditioner=creal


numproc bvp np1 -solver=cg -bilinearform=a -linearform=f -gridfunction=u -preconditioner=c -maxsteps=30

numproc drawflux np2 -bilinearform=a -solution=u -applyd -label=H 
numproc drawflux np2 -bilinearform=a -solution=u -label=B

numproc visualization vis -scalarfunction=H 
  -clipsolution=scalar -clipvec=[0,0,-1] -clipdist=0.031

# numproc visualization vis -scalarfunction=H 
#   -clipsolution=scalar -clipvec=[0,1,0] -clipdist=0.0
