#ifndef FILE_PARALLELDOFS
#define FILE_PARALLELDOFS



using namespace ngcomp;

#ifndef PARALLEL
class ParallelDofs
{
  // dummy class
  ParallelDofs( const FESpace & afespace )
  { ; }
  ~ParallelDofs()
  { ; }

};

#else

class ParallelDofs
{

protected:
  ///
  const FESpace & fespace;
  /// number of degrees of freedom
  int ndof;
  /// number of degrees of freedom shared with processors
  ARRAY<int> nexdof;
  /// is dof i local or exchange-dof??
  BitArray isexchangedof;
  /// is dof i ghost dof or not??
  BitArray * isghostdof;
  /// number of dofs on distant processes
  ARRAY<int> distndof;
  /// mpi-datatype to send exchange dofs
  MPI_Datatype ** mpi_t;

public:
  
  /// arrays containing local dof nrs for exchangedof i, distant dof nrs for exchangedof i
  Table < int> * localexchangedof, *distantexchangedof;
  /// is this the master process??
  BitArray ismasterdof;

  ParallelDofs( const FESpace & afespace );
  ~ParallelDofs();

  const inline int GetNDof () const
  { return ndof; }

  const inline int GetNExDofs() const
  { return nexdof[id]; } 

  inline int GetNExDofs( const int sender ) const
  { return nexdof[sender]; } 

  inline void SetNExDof ( const ARRAY<int> & anexdof ) 
  { 
    nexdof.SetSize(ntasks);
    for ( int i = 0; i < ntasks; i++ ) 
      nexdof[i] = anexdof[i];
  }

  inline void SetExchangeDof ( const int dof )
  {  isexchangedof.Set ( (ntasks+1)*dof + ntasks );  }

  inline void SetExchangeDof ( const int proc, const int dof )
  {  isexchangedof.Set ( (ntasks+1)*dof + proc );  }

  void SetDistantDof ( const int proc, const int localdof, const int distantdof );

  void SetDistantDofs ( const int proc, const ARRAY<int> & dofs );

  void Update();

//   void UpdateLowOrder(LocalHeap & lh);

//   void UpdateLowOrder ()
//   {
//     LocalHeap lh ( 100000 );
//     UpdateLowOrder ( lh );
//   }
//   void UpdateGlobal(LocalHeap & lh);

//   void UpdateGlobal ()
//   {
//     LocalHeap lh ( 100000 );
//     UpdateGlobal ( lh );
//   }

  const inline  bool IsExchangeDof ( const int dof ) const
  {  return isexchangedof.Test( (ntasks+1)*dof + ntasks ) ;  }

  const inline bool IsExchangeDof ( const int proc, const int dof ) const
  {  return isexchangedof.Test ( (ntasks+1)*dof + proc ) ;  }

  const inline int ExDof2LocalDof ( const int proc, const int exdof ) const
  { return (*localexchangedof) [proc][exdof]; }
  const inline int ExDof2DistantDof ( const int proc, const int localexdof ) const
  { return (*distantexchangedof) [proc][localexdof]; }

  const int GetDistantDof ( const int proc, const int localdof ) const;
  
  const inline bool IsMasterDof ( const int localdof ) const
  { return ismasterdof.Test(localdof); }

  inline void SetMasterDof ( const int localdof ) 
  { ismasterdof.Set ( localdof ); }
  
  void GetDistantDofs ( const int localdof,  ARRAY <int> & distantdofs ) const;

  void GetExchangeProcs ( const int localdof, ARRAY<int> & procs ) const;

  void GetExchangeProcs ( ARRAY<int> & procs ) const;

  void GetHOExchangeProcs ( const int localdof, ARRAY<int> & procs ) const;

  void GetHOExchangeProcs ( ARRAY<int> & procs ) const;

  const bool IsExchangeProc ( const int proc ) const;

  const int GetReceiveDof ( const int proc, const int recvnum ) const;

  void Print() const;

  const FlatArray<int>  GetLocalExchangeDofs ( const int proc ) const
  { return (*localexchangedof)[proc]; }

  inline void SetGhostDof ( const int dof )
  { isghostdof -> Set(dof); }

  inline void ClearGhostDof ( const int dof )
  { isghostdof -> Clear(dof); }

  inline bool IsGhostDof ( const int dof ) const
  { return isghostdof->Test(dof); }

  bool ContainsParallelDof ( const FlatArray<int> & doftable ) const;

  void UpdateCompound (  );

  inline int GetDistNDof ( const int dest ) const { return distndof[dest]; }
  inline void SetDistNDof ( const int dest, const int dndof ) { distndof[dest] = dndof; }

  inline void ClearExchangeDof ( const int dof )
  {  isexchangedof.Clear ( (ntasks+1)*dof + ntasks );  }

  inline void ClearExchangeDof ( const int proc, const int dof )
  {  isexchangedof.Clear ( (ntasks+1)*dof + proc );  }
  inline void ClearMasterDof ( const int localdof ) 
  { ismasterdof.Clear ( localdof ); }

  MPI_Datatype * GetMPI_Type ( int dest )
  { return mpi_t[dest]; }

  void UpdateMPIType () ;
};


#endif //PARALLEL

#endif
