#ifndef FILE_PARALLELNGS
#define FILE_PARALLELNGS

#ifndef PARALLEL

#include <comp.hpp>
#include <la.hpp>

namespace ngparallel
{
  using namespace ngsolve;
  using namespace ngparallel;
  using namespace ngcomp;
  enum { id = 0 };
  enum { ntasks = 1 };
  extern ARRAY<int> hoprocs;
}

#else    // PARALLEL

#undef SEEK_SET
#undef SEEK_CUR
#undef SEEK_END
#include <mpi.h>



#include <comp.hpp>
#include <la.hpp>


namespace ngcomp
{
  class FESpace;
}


namespace ngparallel
{
  using namespace ngsolve;
  using namespace ngparallel;
  using namespace ngcomp;

  extern int id, ntasks;
  extern ParallelMeshAccess * parallelma;
  extern ARRAY<int> hoprocs;

  template <class T>
  MPI_Datatype * GetMPIType ( ) { cerr << "ERROR in GetMPIType() -- no type found" << endl;return 0;}

  
  inline void MyMPI_Send ( int i, int dest)
  {
    MPI_Send( &i, 1, MPI_INT, dest, 1, MPI_COMM_WORLD);
  }


  inline void MyMPI_Recv (int & i, int src)
  {
    MPI_Status status;
    MPI_Recv( &i, 1, MPI_INT, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  }

  inline void MyMPI_Send (double i, int dest)
  {
    MPI_Send( &i, 1, MPI_DOUBLE, dest, 1, MPI_COMM_WORLD);
  }


  inline void MyMPI_Recv (double & i, int src)
  {
    MPI_Status status;
    MPI_Recv( &i, 1, MPI_DOUBLE, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  }


  inline void MyMPI_Send (const string & s, int dest)
  {
    MPI_Send( const_cast<char*> (s.c_str()), s.length(), MPI_CHAR, dest, 1, MPI_COMM_WORLD);
  }

  inline void MyMPI_Recv (string & s, int src)
  {
    MPI_Status status;
    int len;
    MPI_Probe (src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    MPI_Get_count (&status, MPI_CHAR, &len);
    s.assign (len, ' ');
    MPI_Recv( &s[0], len, MPI_CHAR, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  }

 





  template <class T>
  inline void MyMPI_Send (const FlatArray<T> & s, int dest)
  {
    const MPI_Datatype * MPI_T  = GetMPIType<T> ();
    MPI_Send( &s[0], s.Size(), *MPI_T, dest, 22, MPI_COMM_WORLD);
    delete MPI_T;
  }

  template <class T>
  inline void MyMPI_Recv ( ARRAY <T> & s, int src)
  {
    MPI_Status status;
    int len;
    const MPI_Datatype * MPI_T  = GetMPIType<T> ();
    MPI_Probe (src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    MPI_Get_count (&status, *MPI_T, &len);

    s.SetSize (len);
    MPI_Recv( &s[0], len, *MPI_T, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    delete MPI_T;
  }


  template <class T>
  inline int MyMPI_Recv ( ARRAY <T> & s)
  {
    MPI_Status status;
    int len;
    const MPI_Datatype * MPI_T  = GetMPIType<T> ();
    MPI_Probe (MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

    int src = status.MPI_SOURCE;

    MPI_Get_count (&status, *MPI_T, &len);

    s.SetSize (len);
    MPI_Recv( &s[0], len, *MPI_T, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

    delete MPI_T;
    return src;
  }

template <class T>
void MyMPI_ISend ( const FlatArray<T> & s, int dest, MPI_Request & request ) 
{ 
  const MPI_Datatype * MPI_T  = GetMPIType<T> ();
  int len = s.Size();
  if ( MPI_T )
    MPI_Isend( const_cast<T*> (&s[0]), len, *MPI_T, dest, 22, MPI_COMM_WORLD, &request);
  else
    (*testout) << "????????" <<endl;
  if ( MPI_T )
    delete MPI_T;
}


template <class T>
void MyMPI_IRecv ( ARRAY<T> & s, int src, MPI_Request & request ) 
{ 
    MPI_Status status;
    int len;
    const MPI_Datatype *  MPI_T = GetMPIType<T> ();
    MPI_Probe (src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);

    MPI_Get_count (&status, *MPI_T, &len);

    s.SetSize (len);
    MPI_Irecv( &s[0], len, *MPI_T, src, MPI_ANY_TAG, MPI_COMM_WORLD, &request);

    delete MPI_T;


}




template <class T>
void MyMPI_Send ( const VVector<T> & s, FlatArray<int> * senddofs, const int dest ) ;


template <class T>
void MyMPI_RecvVec ( ARRAY<T> & s, const int src);

template <class T>
void MyMPI_IRecvVec ( ARRAY<T> & s, const int src, MPI_Request & request);

  
template <class TSCAL>
void MyMPI_Send ( const FlatVector<TSCAL> & s, const int entrysize, FlatArray<int> * senddofs, const int dest ) ;

template <class TSCAL>
void MyMPI_ISend ( const FlatVector<TSCAL> & s, const int entrysize, FlatArray<int> * senddofs, const int dest, MPI_Request & request ) ;

template <class T>
inline void MyMPI_Send (  T *& s, int & len,  int dest)
{
  MPI_Status status;
  const MPI_Datatype * MPI_T = GetMPIType<T> ();
//   MPI_Send( &len, 1, MPI_INT, dest, 1, MPI_COMM_WORLD);
  MPI_Send( s, len, MPI_T, dest, 1, MPI_COMM_WORLD);
  delete MPI_T;
}
  
  
template <class T>
inline void MyMPI_Recv ( T *& s, int & len, int src)
{
  MPI_Status status;
  const MPI_Datatype * MPI_T  = GetMPIType<T> ();
  MPI_Probe (src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  MPI_Get_count (&status, *MPI_T, &len);

  if ( s )
    delete [] s;
  s = new T [len];
  MPI_Recv( s, len, *MPI_T, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  delete MPI_T;
}

  /*

// old version 

  inline void MyMPI_Send (  int *& s, int & len,  int dest)
  {
     MPI_Send( &len, 1, MPI_INT, dest, 1, MPI_COMM_WORLD);
     MPI_Send( s, len, MPI_INT, dest, 1, MPI_COMM_WORLD);
//      (*testout) << "sent int[" << len << "] to " << dest << endl;
  }


  
  inline void MyMPI_Recv ( int *& s, int & len, int src)
  {
    MPI_Status status;
    MPI_Recv( &len, 1, MPI_INT, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    delete s;
    s = new int [len];
    MPI_Recv( s, len, MPI_INT, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
     (*testout) << "recv int[" << len << "] from " << src << endl;
  }



  inline void MyMPI_Send ( double *& s, int len,  int dest)
  {
     MPI_Send( &len, 1, MPI_INT, dest, 1, MPI_COMM_WORLD);
     MPI_Send( s, len, MPI_DOUBLE, dest, 1, MPI_COMM_WORLD);
  }


  inline void MyMPI_Recv ( double *& s, int & len, int src)
  {
    MPI_Status status;
    MPI_Recv( &len, 1, MPI_INT, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
    delete s;
    s = new double [len];
    MPI_Recv( s, len, MPI_DOUBLE, src, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
  }

  */
#include "parallelmeshaccess.hpp"
#include "paralleldofs.hpp"
}

namespace ngla
{
#include "superlu_dist.hpp"
 }

void Parallel_Exit ();

#endif

#endif
