########################################################################################
# solve poisson equation
#    - eps Delta u = f in Omega,  u=g on Gamma0
# by means of a mixed-hybrid method (see demo_habrid.cpp for details) 
########################################################################################


# geometry = square.in2d
# mesh = square.vol

geometry = cube.geo
mesh = cube.vol

define constant heapsize = 10000000

# ho-fespace: compound space for (sigma, u, lambda) ######
define fespace v -type=hybrid -order=3 -eliminate_internal

define gridfunction u -fespace=v

## boundary terms ########################################
define coefficient crob
1e12, 1e12, 0, 0, 0, 0,

define coefficient cneu
-1e12, 1e12, 0, 0, 0, 0, 


## some coefficients #####################################
define coefficient one
1, 1, 1, 1, 1, 1,

define coefficient minusf 
(-1),

## linearform : (-f, v), <g,mu>
define linearform f -fespace=v
neumann cneu -comp=3
source minusf -comp=2

## bilinearform: laplace and robin
define bilinearform a -fespace=v -eliminate_internal -linearform=f
hybridlaplace one
robin crob -comp=3


## elementary solver
numproc bvp np1 -bilinearform=a -linearform=f -gridfunction=u -solver=direct

# visualization
define bilinearform drawsigma -fespace=v -nonassemble
masshdiv one -comp=1

define bilinearform drawu -fespace=v -nonassemble
mass one -comp=2

define bilinearform drawlambda -fespace=v -nonassemble
robin one -comp=3


numproc drawflux df1 -bilinearform=drawsigma -solution=u -label=sigma
numproc drawflux df2 -bilinearform=drawu -solution=u -label=u
numproc drawflux df3 -bilinearform=drawlambda -solution=u -label=lambda

## automatic visualizaiton for 2D
# numproc visualization vis1 -scalarfunction=u -subdivision=2

## automatic visualization for 3D
numproc visualization vis2 -scalarfunction=u -subdivision=1 -clipvec=[0,0,-1] -clipsolution=scalar

